all: topaz tpv2ps tpvview 
clean: cleantopaz cleanview cleanps

# compiler
CC=g++

# Where are include files kept
INCLUDE=. -I/usr/openwin/include

# for Linux
CFLAGS= -O2 -Wall -DTOPAZSOLARIS -DTOPAZ_NCURSES -DTOPAZ_EUC
TOPAZLIB=  -lm -lncurses -lreadline -lhistory
PSLIB=   -lm
VIEWLIB= -lm -L/usr/openwin/lib -lX11

# obj
OBJ=topaz.o plobj.o graph.o axis.o frame.o frame2.o valbuff.o script.o funcs.o\
parsetoken.o conttoken.o buffarray.o system.o tokenbuff.o cui.o\
msgbuff.o styleline.o gdi.o data.o browse.o objectlist.o\
styledisp.o loaddata.o plotdata.o calc.o gamma.o dinvq.o\
pspline.o bspline.o parts.o filepoint.o pipecall.o naninf.o \
autoscale.o globalbuff.o genxdata.o linearfit.o nonlinearfit.o \
convtxt.o kanji.o paths.o topazprintf.o completion.o  regexpress.o

VIEWOBJ=tpvview.o convtxt.o kanji.o
PSOBJ=tpv2ps.o convtxt.o kanji.o

topaz: $(OBJ)
	$(CC) -o topaz $(OBJ) $(TOPAZLIB)

topaz.o: topaz.cc script.h frame.h menudef.h cui.h msgbuff.h filepoint.h jmenudef_euc.h topazcurses.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c topaz.cc

plobj.o: plobj.cc plobj.h convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c plobj.cc

graph.o: graph.cc graph.h frame.h data.h vect.h vectdefs.h col.h styledisp.h \
browse.h gdi.h filepoint.h linearfit.h system.h convtxt.h regexpress.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c graph.cc

axis.o: axis.cc axis.h frame.h graph.h col.h gdi.h minmax.h pointdefs.h naninf.h \
filepoint.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c axis.cc

frame.o: frame.cc frame.h graph.h axis.h data.h gdi.h vectdefs.h pointdefs.h \
filepoint.h system.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c frame.cc

frame2.o: frame2.cc frame.h graph.h axis.h data.h gdi.h vectdefs.h pointdefs.h \
topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c frame2.cc

valbuff.o: valbuff.cc valbuff.h buffarray.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c valbuff.cc

script.o: script.cc script.h system.h tokenbuff.h graph.h buffarray.h cui.h \
browse.h objectlist.h pipecall.h minmax.h gamma.h filepoint.h convtxt.h msgbuff.h \
topazvalues.h regexpress.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c script.cc

funcs.o: funcs.cc script.h system.h tokenbuff.h graph.h buffarray.h cui.h \
browse.h objectlist.h pipecall.h minmax.h gamma.h filepoint.h convtxt.h \
msgbuff.h kanji.h topazvalues.h paths.h topazprintf.h

	$(CC) -I$(INCLUDE) $(CFLAGS) -c funcs.cc

conttoken.o: conttoken.cc script.h system.h tokenbuff.h graph.h buffarray.h \
cui.h browse.h objectlist.h pipecall.h topazvalues.h convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c conttoken.cc

parsetoken.o: parsetoken.cc script.h system.h tokenbuff.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c parsetoken.cc

buffarray.o: buffarray.cc buffarray.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c buffarray.cc

system.o: system.cc convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c system.cc

tokenbuff.o: tokenbuff.cc tokenbuff.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c tokenbuff.cc

cui.o: cui.cc cui.h script.h buffarray.h gdi.h msgbuff.h filepoint.h kanji.h completion.h topazcurses.h frame.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c cui.cc

msgbuff.o: msgbuff.cc msgbuff.h kanji.h topazcurses.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c msgbuff.cc

styleline.o: styleline.cc vectdefs.h pointdefs.h col.h gdi.h graph.h naninf.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c styleline.cc

gdi.o: gdi.cc vectdefs.h pointdefs.h filepoint.h graph.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c gdi.cc

data.o: data.cc data.h gdi.h vectdefs.h pointdefs.h graph.h frame.h topazvalues.h \
convtxt.h paths.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c data.cc

browse.o: browse.cc browse.h vectdefs.h cui.h convtxt.h regexpress.h topazcurses.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c browse.cc

objectlist.o: objectlist.cc objectlist.h graph.h cui.h script.h vectdefs.h \
styledisp.h data.h parts.h gdi.h frame.h filepoint.h regexpress.h topazcurses.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c objectlist.cc

styledisp.o: styledisp.cc styledisp.h col.h data.h frame.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c styledisp.cc

loaddata.o: loaddata.cc data.h frame.h topazvalues.h convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c loaddata.cc

plotdata.o: plotdata.cc data.h frame.h axis.h vectdefs.h pointdefs.h gdi.h \
spline.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c plotdata.cc

calc.o: calc.cc calc.h data.h gamma.h graph.h frame.h naninf.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c calc.cc

gamma.o: gamma.cc gamma.h naninf.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c gamma.cc

dinvq.o: dinvq.cc
	$(CC) -I$(INCLUDE) $(CFLAGS) -c dinvq.cc

pspline.o: pspline.cc spline.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c pspline.cc

bspline.o: bspline.cc spline.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c bspline.cc

parts.o: parts.cc parts.h graph.h frame.h data.h defcharwidth.h gdi.h pointdefs.h \
vectdefs.h spline.h minmax.h filepoint.h kanji.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c parts.cc

filepoint.o: filepoint.cc filepoint.h pipecall.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c filepoint.cc

pipecall.o: pipecall.cc pipecall.h convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c pipecall.cc

naninf.o: naninf.cc naninf.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c naninf.cc

autoscale.o: autoscale.cc frame.h axis.h naninf.h topazvalues.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c autoscale.cc

globalbuff.o: globalbuff.cc globalbuff.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c globalbuff.cc

genxdata.o: genxdata.cc data.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c genxdata.cc

linearfit.o: linearfit.cc linearfit.h naninf.h calc.h graph.h frame.h data.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c linearfit.cc

nonlinearfit.o: nonlinearfit.cc data.h buffarray.h frame.h naninf.h graph.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c nonlinearfit.cc

convtxt.o: convtxt.cc convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c convtxt.cc

kanji.o: kanji.cc kanji.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c kanji.cc

paths.o: paths.cc paths.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c paths.cc

topazprintf.o: topazprintf.cc topazprintf.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c topazprintf.cc

completion.o: completion.cc completion.h vectdefs.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c completion.cc

regexpress.o: regexpress.cc regexpress.h valbuff.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c regexpress.cc

cleantopaz:
	-rm $(OBJ) topaz

tpv2ps: $(PSOBJ)
	$(CC) -o tpv2ps $(PSOBJ) $(PSLIB)

tpv2ps.o: tpv2ps.cc vectdefs.h pointdefs.h col.h kanji.h convtxt.h 
	$(CC) -I$(INCLUDE) $(CFLAGS) -c tpv2ps.cc

cleanps:
	-rm tpv2ps.o tpv2ps

tpvview: tpvview.o kanji.o convtxt.o
	$(CC) -o tpvview tpvview.o kanji.o convtxt.o $(VIEWLIB)

tpvview.o: tpvview.cc vectdefs.h pointdefs.h Xparray.h col.h kanji.h convtxt.h
	$(CC) -I$(INCLUDE) $(CFLAGS) -c tpvview.cc

cleanview:
	-rm tpvview.o tpvview




