/*----------------------------------------------------------------------------
                     Xparray.h (definitions of XPoint array)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__Xparray_h)
#define __Xparray_h

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <memory.h>

class Xparray
{
 public:
  
  Xparray();
  ~Xparray();
  int getarraylength();     
  bool setbuff(int length);
  bool add(XPoint* val);
  
  XPoint *array;

 private:

  int arraylength; 
  int bufflength;  
  int increment;
  bool changebufflength(int length);

};


Xparray::Xparray()
{
  arraylength = 0;
  increment = 10000;

  bufflength = (arraylength / increment + 1) * increment;

  array =   new XPoint[bufflength];
  memset(array, bufflength, 0);
}


Xparray::~Xparray()
{
  delete [] array;
}

int Xparray::getarraylength()
{
  return arraylength;
}

bool Xparray::setbuff(int length)
{
  delete [] array;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  array =   new XPoint[bufflength];
  if (array == 0)
    return false;
  
  memset(array, bufflength, 0);
  
  return true;
}


bool Xparray::add(XPoint* val)
{

  if (arraylength < bufflength)
    {
      arraylength++;
      array[arraylength - 1] = *val;
    }
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
      array[arraylength - 1] = *val;
    }
  
    return true;
}

bool Xparray::changebufflength(int newlength)
{
  
  XPoint *temparray;
  int temparraylength;

  if (newlength <= bufflength)
    return true;

  temparray =   new XPoint[bufflength];
  if (temparray == 0)
    return false;

  temparraylength = arraylength;
  
  memcpy(temparray, array, arraylength * sizeof(XPoint));

  delete[] array;

  array = temparray;

  return true;
}




#endif
