/*----------------------------------------------------------------------------
                       axis.h (definition of axis object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__axis_h)
#define __axis_h

#include "plobj.h"
#include "pointdefs.h"
#include "col.h"

#define DELTA_AXIS          1e-7

#define TICKPARITY_NULL     0
#define TICKPARITY_PLUS     1
#define TICKPARITY_MINUS    2
#define TICKPARITY_BOTH     3

#define SCALE_LINEAR        0
#define SCALE_LOG           1
#define SCALE_INV           2
#define SCALE_LOGLINEAR     3
#define SCALE_INVLINEAR     4

#define AXISTYPE_NILL       0
#define AXISTYPE_X          1
#define AXISTYPE_Y          2
#define AXISTYPE_T          3
#define AXISTYPE_R          4

#define LABEL_FORMAT_NULL     0
#define LABEL_FORMAT_DECIMAL  1
#define LABEL_FORMAT_EXP      2

#define LABELPARITY_PLUS     0
#define LABELPARITY_MINUS    1

#define CROSSPOINTPARITY_CENTER   0
#define CROSSPOINTPARITY_MINUS    1
#define CROSSPOINTPARITY_PLUS     2
#define CROSSPOINTPARITY_NULL     3


class axis:public plobj
{
 public:
  axis();
  ~axis();
  axis& operator = (const axis& ax);

  bool get(std::string* member,  tokenbuff* outvalue);  
  bool set(std::string* member,  std::string* setvalue);  
  bool exec(std::string* function, buffarray *argument, tokenbuff* outvalue);  

  bool paint(bool flag, FILE *f);
  bool paintmajortick(bool flag, FILE *f);
  bool paintminortick(bool flag, FILE *f);
  bool drawlabel(bool flag, FILE *f);

  doublepoint virtual_to_doubleworld(double x);
  intpoint virtual_to_world(double x);
  double virtual_to_axis(double x);
  doublepoint axis_to_world(double x, double y);

  double axis_to_virtual(double x);
  double world_to_axis(double x, double y);
  double doubleworld_to_virtual(double x, double y);

  /* axis read / write  */
  bool readaxis(FILE *fp);
  bool writeaxis(FILE *fp);


  int id;
  std::string *label;
  int selected;
  /*base*/
  int axistype;
  bool display;
  bool commonset;
  int scaling;
  bool invertscale;
  intpoint origin;
  int length;
  double angle;
  Caplinestyle baselinestyle;
  Capcolor basecolor;
  double tickarea_start;
  double tickarea_end;


  /*mumerical range*/
  double start;
  double end;
  double unit;
  double crosspoint;
  bool autocrosspoint;

  /*tick*/
  int tickparity;
  int tickwidth;
  int ticklength;           
  int division;

  /*tickline*/
  Capcolor majorticklinecolor, minorticklinecolor;
  Caplinestyle majorticklinestyle, minorticklinestyle;

  /*label*/
  Capfont fo;
  Capcolor fontcolor;
  int format;
  bool addplus;
  int offset;
  int skip;
  bool autoprecision;
  int intprecision;
  int decimalprecision;
  int labelparity;
  int crosspointparity;
  /* this member is impricit*/
  int crosspointoffset;  
};

#endif





