/*----------------------------------------------------------------------------
                           browse.cc (browse interfaces)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <time.h>
#include <sys/stat.h>
#include <string>

#include "vectdefs.h"
#include "browse.h"
#include "cui.h"
#include "convtxt.h"
#include "regexpress.h"
#include "topazcurses.h"

#define SORT_FILE        0
#define SORT_TIME        1
#define SORT_SIZE        2

extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;
extern int menuwinlines;
extern int maxrow;
extern int leftmargin;
extern int topmargin;
extern int linelength(char *str, int offset, int displength);
extern int topazgetch(WINDOW *win);
extern void clearmenuwin();

static int sortcond = 0;
static int reversed = 0;
std::string reg = std::string("");
char cdir[1024];

extern void message(const char *msg);


void browsehelp(WINDOW *win)
{
  int t = 0;
  int l = leftmargin + 5;

  if (win == 0)
    return;
  clearmenuwin();
  //wclear(win);

  wmove(win, t + 1, l);
  wprintw(win, "         [space]: file selection / unselection                        "); 
  wmove(win, t + 2, l);
  wprintw(win, "         [v]iew file                                                  "); 
  wmove(win, t + 3, l);
  wprintw(win, "         change show [m]ode                                           "); 
  wmove(win, t + 4, l);
  wprintw(win, "         select [a]ll files / [c]lear selection                       "); 
  wmove(win, t + 5, l);
  wprintw(win, "         sort by [n]ame, [t]ime, or [s]ize                            "); 
  wmove(win, t + 6, l);
  wprintw(win, "         show in incremental / decremental [o]rder                    "); 
  wmove(win, t + 7, l);
  wprintw(win, "         [g]rep with given regular expression                         "); 
  wmove(win, t + 8, l);
  wprintw(win, "         change current working [d]irectory                           "); 
  wmove(win, t + 9, l);
  wprintw(win, "         fix file selection and e[x]it file browsing                  "); 
  wmove(win, t + 10, l);
  wprintw(win, "         [ESC]: discard file selection and exit file browsing         "); 
  wmove(win, t + 11, l);
  wprintw(win, "                                                                      "); 
  wmove(win, t + 12, l);
  wprintw(win, "                           HIT ANY KEY !! "); 

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  topazgetch(win);
  wrefresh(win);
}

bool viewheader(WINDOW *win, char *filename)
{
  FILE *fp;
  char stemp[10000], stemp1[100], stemp2[10];
  int count, totalcount;
  char *p;
  int llength;
  char *endflag = 0;

  if (win == 0)
    return false;

  clearmenuwin();
  //wclear(win);
  totalcount = 0;
  //  wrefresh(win);
  if ((fp = fopen(filename, "r")) != 0)
    {
      do
	{
          clearmenuwin();
          //wclear(win);
	  count = 0;
	  while (count <  menuwinlines - 1 - topmargin 
		 && (endflag = topazfgets(stemp, sizeof(stemp), fp)) != NULL)
	    {
	      // remove '\r', '\n'
	      stemp[sizeof(stemp) - 1] = 0;
	      p = stemp;
	      strcpy(stemp1, "");
	      while (*p != '\r' && *p != '\n' && *p != 0)
		{
		  stemp2[0] = *p; stemp2[1] = 0;
		  strcat(stemp1, stemp2);
		  p++;
		}
	      llength = linelength(stemp1, 0, 72);
	      stemp1[llength] = 0;

	      wmove(win, topmargin + count, leftmargin);
	      wprintw(win, "%03d: %s", totalcount + 1, stemp1);
	      count++; totalcount++;
	    }
	  wattron(win, A_REVERSE);
	  wmove(win, 0,0);
	  wprintw(win, " ");
	  wmove(win, 0,0);
	  wattroff(win, A_REVERSE);

	  wrefresh(win);
	}while(topazgetch(win) == ' ' && endflag != NULL);
      fclose(fp);

      return true;
    }
  else
    return false;
}

bool canread(fileentry *e)
{
  u_short mode = e->status.st_mode;
  switch(mode & S_IFMT)
    {
    case S_IFSOCK:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFDIR:
    case S_IFIFO:
      return false;
    case S_IFLNK:
    default:
      return true;
    }
  return true;
}

bool isdir(fileentry *e)
{
  u_short mode = e->status.st_mode;
 if ((mode & S_IFMT) == S_IFDIR)
   return true;
 else
   return false;
}

int comparecond(const void *v1, const void *v2)
{
  int flag = false;
  fileentry** e1;
  fileentry** e2;

  e1 = (fileentry **)v1;
  e2 = (fileentry **)v2;

  if (strcmp((*e1)->filename, ".") == 0)
    return(-1);

  if (strcmp((*e2)->filename, ".") == 0)
    return(1);

  if (strcmp((*e1)->filename, "..") == 0)
    return(-1);

  if (strcmp((*e2)->filename, "..") == 0)
    return(1);

  if (!isdir(*e1) && isdir(*e2))
    return(1);

  if (isdir(*e1) && !isdir(*e2))
    return(-1);
  
  switch(sortcond)
    {
    case SORT_FILE:
      flag = strcmp((*e1)->filename, (*e2)->filename);
      break;
    case SORT_TIME:
      flag = (*e1)->status.st_mtime - (*e2)->status.st_mtime;
      break;
    case SORT_SIZE:
      flag = (int)((*e1)->status.st_size) - (int)((*e2)->status.st_size);
      break;
    }
  if (reversed)
    flag = -flag;

  return flag;
}


bool loadfilelist(char *dirname, fileentarray *ar)
{
  DIR *dp;
  struct dirent *entry;
  struct stat statbuff;
  std::string regex = reg;
  std::string str1;

  chdir (dirname);

  if ((dp = opendir(dirname)) == NULL)
    return false;

  while ((entry = readdir(dp)) != NULL)
    {
      str1 = std::string(entry->d_name);
      stat(entry->d_name, &statbuff);
      if (simplematch(str1.c_str(), regex.c_str())
          || ((statbuff.st_mode & S_IFMT) == S_IFDIR))
	{
	  ar->add(new fileentry);
	  strcpy((*ar)[ar->getarraylength() - 1].filename, entry->d_name);
	  (*ar)[ar->getarraylength() - 1].status = statbuff;
	}
    }

  if (ar->getarraylength() > 0)
    qsort((void *)ar->array, ar->getarraylength(), sizeof(fileentry *), comparecond);

  getcwd(cdir, sizeof(cdir));
  closedir (dp);
  return true;
}

void getmode(char *buf, u_short mode)
{
  switch(mode & S_IFMT) {
  case S_IFDIR:
    buf[0] = 'd';
    break;
  case S_IFBLK:
    buf[0] = 'b';
    break;
  case S_IFCHR:
    buf[0] = 'c';
    break;
  case S_IFLNK:
    buf[0] = 'l';
    break;
  case S_IFSOCK:
    buf[0] = 's';
    break;
  case S_IFIFO:
    buf[0] = 'p';
    break;
  default:
    buf[0] = '-';
    break;
  }
  
  buf[1] = (mode & S_IRUSR) ? 'r' : '-';
  buf[2] = (mode & S_IWUSR) ? 'w' : '-';
  buf[3] = (mode & S_ISUID) ? 's' : (mode & S_IXUSR) ? 'x' : '-';
  buf[4] = (mode & S_IRGRP) ? 'r' : '-';
  buf[5] = (mode & S_IWGRP) ? 'w' : '-';
  buf[6] = (mode & S_ISGID) ? 's' : (mode & S_IXGRP) ? 'x' : '-';
  buf[7] = (mode & S_IROTH) ? 'r' : '-';
  buf[8] = (mode & S_IWOTH) ? 'w' : '-';
  buf[9] = (mode & S_ISVTX) ? 't' : (mode & S_IXOTH) ? 'x' : '-';
  buf[10] = 0;

}

void filemenustatus(WINDOW *win)
{
  if (win == 0)
    return;

  wmove(win, menuwinlines - 2, leftmargin);
  wprintw(win, " current: %-70s", cdir); 
  wmove(win, menuwinlines - 1, leftmargin);
  //  wprintw(win, " [m]ode [a]ll [c]lear [n]ame [t]ime [s]ize [o]rder [g]rep ch[d]ir e[x]it"); 
  wprintw(win, " [space]select/unselect [v]iew [g]rep ch[d]ir e[x]it [ESC]cancel [h]elp"); 
}

void filelistdisp(WINDOW *win, int startline, fileentarray *ar)
{
  struct tm *tm;
  char buff[100];
  char fname[31];

  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  int length = ar->getarraylength();
  for (int i = 0; i < length; i++)
    {
      if (startline <= i && i < startline + menuwinlines - 3 - topmargin)
	{
	  wmove(win, i - startline + topmargin, leftmargin);
	  tm = localtime(&(*ar)[i].status.st_mtime);
	  getmode(buff, (*ar)[i].status.st_mode);
	  strncpy(fname, (*ar)[i].filename, sizeof(fname));
	  fname[sizeof(fname) - 1] = 0;
	  if ((*ar)[i].selected)
	    wprintw(win, " * %s   %-30s %8d %02d-%02d-%02d %02d:%02d\n",
		    buff,
		    fname,
		    (unsigned)(*ar)[i].status.st_size,
		    tm->tm_year % 100,
		    tm->tm_mon + 1,
		    tm->tm_mday,
		    tm->tm_hour,
		    tm->tm_min
		    ); 
	  else
	    wprintw(win, "   %s   %-30s %8d %02d-%02d-%02d %02d:%02d\n",
		    buff,
		    fname,
		    (unsigned)(*ar)[i].status.st_size,
		    tm->tm_year % 100,
		    tm->tm_mon + 1,
		    tm->tm_mday,
		    tm->tm_hour,
		    tm->tm_min
		    ); 
	}
    }
  filemenustatus(win);
  wrefresh(win);
}

void filelistcursor(WINDOW *win, int startline, fileentarray *ar, int no, bool flag)
{
  struct tm *tm;
  char buff[100];
  char fname[31];

  if (win == 0)
    return;

  if (ar->getarraylength() == 0)
    return;

  if (flag)
    wattron(win, A_REVERSE);

      if (startline <= no && no < startline + menuwinlines - 3 - topmargin)
	{
	  wmove(win, no - startline + topmargin, leftmargin);
	  tm = localtime(&(*ar)[no].status.st_mtime);
	  getmode(buff, (*ar)[no].status.st_mode);
	  strncpy(fname, (*ar)[no].filename, sizeof(fname));
	  fname[sizeof(fname) - 1] = 0;
	  if ((*ar)[no].selected)
	    wprintw(win, " * %s   %-30s %8d %02d-%02d-%02d %02d:%02d\n",
		    buff,
		    fname,
		    (unsigned)(*ar)[no].status.st_size,
		    tm->tm_year % 100,
		    tm->tm_mon + 1,
		    tm->tm_mday,
		    tm->tm_hour,
		    tm->tm_min
		    ); 
	  else
	    wprintw(win, "   %s   %-30s %8d %02d-%02d-%02d %02d:%02d\n",
		    buff,
		    fname,
		    (unsigned)(*ar)[no].status.st_size,
		    tm->tm_year % 100,
		    tm->tm_mon + 1,
		    tm->tm_mday,
		    tm->tm_hour,
		    tm->tm_min
		    ); 
	}
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);
  if (flag)
    wattroff(win, A_REVERSE);
  wrefresh(win);
}

void shortfilelistdisp(WINDOW *win, int startline, fileentarray *ar)
{
  char fname[17];
  int indx, indy;

  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  int length = ar->getarraylength();
  for (int i = 0; i < length; i++)
    {
      if (startline <= i && i < startline + (menuwinlines - 3 - topmargin) * 4)
	{
	  indy = (i - startline) / 4;
	  indx = (i - startline) % 4;

	  wmove(win, indy + topmargin, leftmargin + indx * 19);
	  strncpy(fname, (*ar)[i].filename, sizeof(fname));
	  fname[sizeof(fname) - 1] = 0;

	  if (((*ar)[i].status.st_mode & S_IFMT) == S_IFDIR)
	    {
	      wattron(win, A_BOLD);
	      if (strlen(fname) < sizeof(fname) - 1 &&
		  strcmp(fname, ".") != 0 &&
		  strcmp(fname, "..") != 0
		  )
		strcat(fname, "/");
	    }

	  if ((*ar)[i].selected)
	    wprintw(win, " * %-16s", fname); 
	  else
	    wprintw(win, "   %-16s", fname); 

	  if (((*ar)[i].status.st_mode & S_IFMT) == S_IFDIR)
	    	wattroff(win, A_BOLD);
	}
    }

  filemenustatus(win);
  wrefresh(win);
}

void shortfilelistcursor(WINDOW *win, int startline, fileentarray *ar, int no, bool flag)
{
  char fname[17];
  int indx, indy;

  if (win == 0)
    return;

  if (ar->getarraylength() == 0)
    return;

  if (flag)
    wattron(win, A_REVERSE);
  if (((*ar)[no].status.st_mode & S_IFMT) == S_IFDIR)
      wattron(win, A_BOLD);

  if (startline <= no && no < startline + (menuwinlines - 3 - topmargin) * 4)
    {
      indy = (no - startline) / 4;
      indx = (no - startline) % 4;
      wmove(win, indy + topmargin, leftmargin + indx * 19);
      strncpy(fname, (*ar)[no].filename, sizeof(fname));
      fname[sizeof(fname) - 1] = 0;

      if (((*ar)[no].status.st_mode & S_IFMT) == S_IFDIR)
	  if (strlen(fname) < sizeof(fname) - 1 &&
	      strcmp(fname, ".") != 0 &&
	      strcmp(fname, "..") != 0
	      )
	    strcat(fname, "/");
      
      if ((*ar)[no].selected)
	wprintw(win, " * %-16s", fname); 
      else
	wprintw(win, "   %-16s", fname); 

    }
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);

  if (((*ar)[no].status.st_mode & S_IFMT) == S_IFDIR)
    wattroff(win, A_BOLD);
  if (flag)
    wattroff(win, A_REVERSE);

  wrefresh(win);

}

 
int selectfile(WINDOW *win, std::string *dir, buffarray *out, bool ismulti, std::string* filter)
{
  int key;
  int startline = 0;
  int relativeline = 0;
  fileentarray *filearray;
  int length;
  char stemp[1024], fullpathdir[1024];
  std::string str1, str2;
  std::string outstring;
  bool shortmenu = true;

  if (win == 0)
    return false;

  filearray = new fileentarray;
  if (filearray == 0)
    return false;

  reg = *filter;
  loadfilelist((char *)dir->c_str(), filearray);

  if (shortmenu)
    shortfilelistdisp(win, startline, filearray);
  else
    filelistdisp(win, startline, filearray);

  length = filearray->getarraylength();

  while (1)
    {
      
      if (shortmenu)
	shortfilelistcursor(win, startline, filearray, startline + relativeline, true);
      else
	filelistcursor(win, startline, filearray, startline + relativeline, true);
      key = topazgetch(win);
      if (shortmenu)
	shortfilelistcursor(win, startline, filearray, startline + relativeline, false);
      else
	filelistcursor(win, startline, filearray, startline + relativeline, false);

      switch (key)
	{
	case LOCAL_LF_KEY:
	  if (filearray->getarraylength() > 0)
	    {
	      if (((*filearray)[startline + relativeline].status.st_mode & S_IFMT) == S_IFDIR
		  && strcmp((*filearray)[startline + relativeline].filename, ".") != 0)
		{
		  strcpy(stemp, (*filearray)[startline + relativeline].filename);
		  getcwd(fullpathdir, sizeof(fullpathdir));
                  if (strlen(fullpathdir) > 0)
		    {
		      if (fullpathdir[strlen(fullpathdir) - 1] != '/')
			strcat(fullpathdir, "/");
		    }
		  strcat(fullpathdir, stemp);

		  filearray->setbuff(0);
		  loadfilelist(fullpathdir, filearray);
		  length = filearray->getarraylength();
		  startline = 0;
		  relativeline = 0;
		  if (shortmenu)
		    shortfilelistdisp(win, startline, filearray);
		  else
		    filelistdisp(win, startline, filearray);
		}
	    }
	    break;
	case ' ':
	  if (filearray->getarraylength() > 0)
	    {
	      // multi select
	      if (ismulti)
		{
		  if (canread(&(*filearray)[startline + relativeline]))
		    {
		      if ((*filearray)[startline + relativeline].selected == true)
			(*filearray)[startline + relativeline].selected = 0;
		      else
			(*filearray)[startline + relativeline].selected = 1;
		      if (shortmenu)
			shortfilelistcursor(win, startline, filearray, startline + relativeline, false);
		      else
			filelistcursor(win, startline, filearray, startline + relativeline, false);
		    }
		}
	      // single select
	      else
		{
		  if (canread(&(*filearray)[startline + relativeline]))
		    {
		      // clear select
		      for (int i = 0; i < filearray->getarraylength(); i++)
			{
			  if (canread(&(*filearray)[i]))
			    (*filearray)[i].selected = false;
			}
		      (*filearray)[startline + relativeline].selected = 1;
		      if (shortmenu)
			shortfilelistdisp(win, startline, filearray);
		      else
			filelistdisp(win, startline, filearray);
		      if (shortmenu)
			shortfilelistcursor(win, startline, filearray, startline + relativeline, false);
		      else
			filelistcursor(win, startline, filearray, startline + relativeline, false);
		      break;

		    }
		}

	      // move to next position
	      if (ismulti)
		{
		  if (shortmenu)
		    {
		      if ((startline + relativeline + 1) < length)
			{
			  if (relativeline + 1 < (menuwinlines - 3 - topmargin) * 4)
			    relativeline += 1;
			  else
			    {
			      startline += (menuwinlines - 3 - topmargin) * 4;
			      relativeline = 0;
			      shortfilelistdisp(win, startline, filearray);
			    }
			}
		    }
		  else
		    {
		      if ((startline + relativeline + 1) < length)
			{
			  if (relativeline + 1 < menuwinlines - 3 - topmargin)
			    relativeline++;
			  else
			    {
			      startline += menuwinlines - 3 - topmargin;
			      relativeline = 0;
			      filelistdisp(win, startline, filearray);
			    }
			}
		    }
		}
	    }
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  if (shortmenu)
	    {
	      if ((startline + relativeline + 1) < length)
		{
		  if (relativeline + 1 < (menuwinlines - 3 - topmargin) * 4)
		    relativeline += 1;
		  else
		    {
		      startline += (menuwinlines - 3 - topmargin) * 4;
		      relativeline = 0;
		      shortfilelistdisp(win, startline, filearray);
		    }
		}
	    }
	  break;
	case KEY_LEFT:
	case CTRL('b'):
	  if (shortmenu)
	    {
	      if ((startline + relativeline - 1) >= 0)
		{
		  if (relativeline - 1 >= 0)
		    relativeline--;
		  else
		    {
		      startline -= (menuwinlines - 3 - topmargin) * 4;
		      relativeline = (menuwinlines - 3 - topmargin) * 4 - 1;
		      shortfilelistdisp(win, startline, filearray);
		    }
		}
	    }
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  if (shortmenu)
	    {
	      if ((startline + relativeline + 4) < length)
		{
		  if (relativeline + 4 < (menuwinlines - 3 - topmargin) * 4)
		    relativeline += 4;
		  else
		    {
		      startline += (menuwinlines - 3 - topmargin) * 4;
		      relativeline -= (menuwinlines - 3 - topmargin - 1) * 4;
		      shortfilelistdisp(win, startline, filearray);
		    }
		}
	    }
	  else
	    {
	      if ((startline + relativeline + 1) < length)
		{
		  if (relativeline + 1 < menuwinlines - 3 - topmargin)
		    relativeline++;
		  else
		    {
		      startline += menuwinlines - 3 - topmargin;
		      relativeline = 0;
		      filelistdisp(win, startline, filearray);
		    }
		}
	    }
	  break;
	case KEY_NPAGE:
	case CTRL('v'):
	  if (shortmenu)
	    {
	      if ((startline + relativeline + (menuwinlines - topmargin - 3) * 4) < length)
		{
		  startline += (menuwinlines - topmargin - 3) * 4;
		  shortfilelistdisp(win, startline, filearray);
		}
	      else if ((startline + (menuwinlines - topmargin - 3) * 4) < length)
		{
		  startline += (menuwinlines - topmargin - 3) * 4;
		  relativeline = 0;
		  shortfilelistdisp(win, startline, filearray);
		}
	      else
		{
		  relativeline = length - startline - 1;
		  shortfilelistdisp(win, startline, filearray);
		}
	    }
	  else
	    {
	      if ((startline + relativeline + menuwinlines - topmargin - 3) < length)
		{
		  startline += menuwinlines - topmargin - 3;
		  filelistdisp(win, startline, filearray);
		}
	      else if ((startline + menuwinlines - topmargin - 3) < length)
		{
		  startline += menuwinlines - topmargin - 3;
		  relativeline = 0;
		  filelistdisp(win, startline, filearray);
		}
	      else
		{
		  relativeline = length - startline - 1;
		  filelistdisp(win, startline, filearray);
		}
	    }
	  break;
	case KEY_UP:
	case CTRL('p'):
	  if (shortmenu)
	    {
	      if ((startline + relativeline - 4) >= 0)
		{
		  if (relativeline - 4 >= 0)
		    relativeline -= 4;
		  else
		    {
		      startline -= (menuwinlines - 3 - topmargin) * 4;
		      relativeline += (menuwinlines - 3 - topmargin - 1) * 4;
		      shortfilelistdisp(win, startline, filearray);
		    }
		}
	    }
	  else
	    {
	      if ((startline + relativeline - 1) >= 0)
		{
		  if (relativeline - 1 >= 0)
		    relativeline--;
		  else
		    {
		      startline -= menuwinlines - 3 - topmargin;
		      relativeline = menuwinlines - 3 - topmargin - 1;
		      filelistdisp(win, startline, filearray);
		    }
		}
	    }
	  break;
	case KEY_PPAGE:
	  if (shortmenu)
	    {
	      if ((startline  - (menuwinlines - topmargin - 3) * 4) >= 0)
		{
		  startline -= (menuwinlines - topmargin - 3) * 4;
		  shortfilelistdisp(win, startline, filearray);
		}
	      else
		{
		  startline = 0; relativeline = 0;
		  shortfilelistdisp(win, startline, filearray);
		}
	    }
	  else
	    {
	      if ((startline  - (menuwinlines - topmargin - 3)) >= 0)
		{
		  startline -= menuwinlines - topmargin - 3;
		  filelistdisp(win, startline, filearray);
		}
	      else
		{
		  startline = 0; relativeline = 0;
		  filelistdisp(win, startline, filearray);
		}
	    }
	  break;
	case CTRL('a'):
	  startline = 0; relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case CTRL('e'):
	  if (shortmenu)
	    {
	      startline = (length - 1) / ((menuwinlines - topmargin - 3) * 4);
	      startline *= (menuwinlines - topmargin - 3) * 4;
	      relativeline = length - startline - 1;
	      shortfilelistdisp(win, startline, filearray);
	    }
	  else
	    {
	      startline = (length - 1) / (menuwinlines - topmargin - 3);
	      startline *= menuwinlines - topmargin - 3;
	      relativeline = length - startline - 1;
	      filelistdisp(win, startline, filearray);
	    }
	  break;
	case 'n':
	  sortcond = SORT_FILE;
	  filearray->setbuff(0);
	  loadfilelist(".", filearray);
	  length = filearray->getarraylength();
	  startline = 0;
	  relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 't':
	  sortcond = SORT_TIME;
	  filearray->setbuff(0);
	  loadfilelist(".", filearray);
	  length = filearray->getarraylength();
	  startline = 0;
	  relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 's':
	  sortcond = SORT_SIZE;
	  filearray->setbuff(0);
	  loadfilelist(".", filearray);
	  length = filearray->getarraylength();
	  startline = 0;
	  relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 'a':
          if (ismulti)
            {
              if (filearray->getarraylength() > 0)
                {
                  for (int i = 0; i < filearray->getarraylength(); i++)
                    {
                      if (canread(&(*filearray)[i]))
                        (*filearray)[i].selected = true;
                    }
                  if (shortmenu)
                    shortfilelistdisp(win, startline, filearray);
                  else
                    filelistdisp(win, startline, filearray);
                }
            }
          break;
        case 'c':
          if (ismulti)
            {
              if (filearray->getarraylength() > 0)
                {
                  for (int i = 0; i < filearray->getarraylength(); i++)
                    {
                      if (canread(&(*filearray)[i]))
                        (*filearray)[i].selected = false;
                    }
                  if (shortmenu)
                    shortfilelistdisp(win, startline, filearray);
                  else
                    filelistdisp(win, startline, filearray);
                }
            }
          break;
	case 'o':
	  if (reversed)
	    reversed = false;
	  else
	    reversed = true;
	  filearray->setbuff(0);
	  loadfilelist(".", filearray);
	  length = filearray->getarraylength();
	  startline = 0;
	  relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 'm':
	  if (shortmenu)
	    shortmenu = false;
	  else
	    shortmenu = true;
	  startline = 0;
	  relativeline = 0;
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 'd':
	  getcwd(cdir, sizeof(cdir));
	  if (cdir[strlen(cdir) - 1] != '/')
	    strcat(cdir, "/");
	  str1 = std::string(cdir);
	  str2 = std::string("Change directory:");
	  if (edittextwindow(menuwin, &str1, &str2, true))
	    {
	      filearray->setbuff(0);
	      loadfilelist((char *)str1.c_str(), filearray);
	      length = filearray->getarraylength();
	      startline = 0;
	      relativeline = 0;
	    }
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 'g':
	  str1 = reg;
	  str2 = std::string("Input regular expression:");
	  if (edittextwindow(menuwin, &str1, &str2, false))
	    {
	      reg = str1;
	      filearray->setbuff(0);
	      loadfilelist(".", filearray);
	      length = filearray->getarraylength();
	      startline = 0;
	      relativeline = 0;
	    }
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case 'v':
	  if (filearray->getarraylength() > 0)
	    {
	      if (canread(&(*filearray)[startline + relativeline]))
		  viewheader(win, (*filearray)[startline + relativeline].filename);
	      if (shortmenu)
		shortfilelistdisp(win, startline, filearray);
	      else
		filelistdisp(win, startline, filearray);
	    }
	  break;
	case 'h':
	  browsehelp(win);
	  if (shortmenu)
	    shortfilelistdisp(win, startline, filearray);
	  else
	    filelistdisp(win, startline, filearray);
	  break;
	case CTRL('g'):
	  key = LOCAL_ESCAPE_KEY;
	  break;
	}

      if (key == LOCAL_ESCAPE_KEY)
	break;

      if (key == 'x' || key == CTRL(' ') || key == CTRL('@'))
	{
	  key = LOCAL_LF_KEY;
	  break;
	}
    }

  if (key == LOCAL_ESCAPE_KEY)
  {
    delete filearray;
    return false;
  }
  else if (key == LOCAL_LF_KEY)
  {
    if (filearray->getarraylength() > 0)
      {
	out->setbuff(0);
	for (int i = 0; i < filearray->getarraylength(); i++)
	{
	  if ((*filearray)[i].selected)
	    {
	      outstring = std::string(cdir) + std::string("/") + std::string((*filearray)[i].filename);
	      out->add(outstring.c_str());
	    }
	}
	delete filearray;
	return true;
      }
    else
      {
	delete filearray;
	return false;
      }
  }
  return true;

}





