/*----------------------------------------------------------------------------
                           bspline.cc (bspline curve)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include "spline.h"
#include "topazvalues.h"

#define OPENC     1                                 /* end of terminal */
#define CLOSEC   2
#define EPS  1e-5              /* Used to decide if two float are equal. */

int order_b;   /* Bspline order to use. */
int appnum;

int CalcBSpline(doublearray *INX, doublearray *INY
		 ,doublearray *OUTX, doublearray *OUTY
		 ,int end_terminal, int divno)
{

  doublearray *inxrow = 0, *inyrow = 0;
  int ErrorFlag = false;
  int	normalcount, i;
  int n;

  inxrow = new doublearray;
  if (inxrow == 0)
    {
      ErrorFlag = true;
      goto j11;
    }
  inxrow->setbuff(1000);

  inyrow = new doublearray;
  if (inyrow == 0)
    {
      ErrorFlag = true;
      goto j11;
    }
  inyrow->setbuff(1000);

  // remove invalid points
  normalcount = 0;
  n = INX->getarraylength() - 1;
  for (i = 0;i <= n; i++)
    {
      if (fabs((*INX)[i]) < HUGE_VAL 	&& fabs((*INY)[i]) < HUGE_VAL)
	{
	  if (!inxrow->add((*INX)[i]))
	    {
	      ErrorFlag = true;
	      goto j11;
	    }
	  if (!inyrow->add((*INY)[i]))
	    {
	      ErrorFlag = true;
	      goto j11;
	    }
	}
    }
  
  //calc b-Spline curve
  appnum = divno;
  
  if (inxrow->getarraylength() <= 1)
    return false;
  else if (order_b >= inxrow->getarraylength())
    order_b = inxrow->getarraylength() - 1;
  
  if (!BSplineAppr(inxrow, inyrow, OUTX, OUTY, inxrow->getarraylength(), end_terminal))
    ErrorFlag = true;


j11:;

  if (inxrow != 0)
    delete inxrow;
  if (inyrow != 0)
    delete inyrow;
  
  if (ErrorFlag == true)
    return false;
  else
    return true;

}



int BSplineAppr(doublearray *INX, doublearray *INY, doublearray *OUTX, doublearray *OUTY,
		int npoints,int end_terminal)
{
  int i,j;
  double dx[5], dy[5];
  double t,dt;
  double outx, outy;

  if (npoints < 8)
    return false;

  dt = 1.0 / (double)appnum;
  
  if (end_terminal == 1)  //open countor
    {
      //first region
      dx[1]=(*INX)[0];
      dx[2]=(*INX)[1];
      dx[3]=(*INX)[2];
      dx[4]=(*INX)[3];
      dy[1]=(*INY)[0];
      dy[2]=(*INY)[1];
      dy[3]=(*INY)[2];
      dy[4]=(*INY)[3];
      
      for (i = 0; i < appnum; i++)
	{
	  t = (double)i * dt;
	  GenBspline(1, dx, dy, &outx, &outy, t);
	  
	  if (!OUTX->add(outx))
	    return false;
	  if (!OUTY->add(outy))
	    return false;
				}
      
      //Second region
      if (npoints>=6)
	{
	  dx[1]=(*INX)[1];
	  dx[2]=(*INX)[2];
	  dx[3]=(*INX)[3];
	  dx[4]=(*INX)[4];
	  dy[1]=(*INY)[1];
	  dy[2]=(*INY)[2];
	  dy[3]=(*INY)[3];
	  dy[4]=(*INY)[4];
	  
	  for (i=0; i<appnum; i++)
	    {
	      t = (double)i * dt;
	      GenBspline(2, dx, dy, &outx, &outy, t);
	      
	      if (!OUTX->add(outx))
		return false;
	      if (!OUTY->add(outy))
		return false;
	    }
	}
      
      //periodic region
      for (j = 2; j <= npoints-6; j++)
	{
	  
	  dx[1]=(*INX)[j];
	  dx[2]=(*INX)[j+1];
	  dx[3]=(*INX)[j+2];
	  dx[4]=(*INX)[j+3];
	  dy[1]=(*INY)[j];
	  dy[2]=(*INY)[j+1];
	  dy[3]=(*INY)[j+2];
	  dy[4]=(*INY)[j+3];
	  
	  for (i=0; i<appnum; i++)
	    {
	      t = (double)i * dt;
	      GenBspline(0, dx, dy, &outx, &outy, t);
	      
	      if (!OUTX->add(outx))
		return false;
	      if (!OUTY->add(outy))
		return false;
	      
	    }
	}
      
      //last second region
      if (npoints>=6)
	{
	  dx[1]=(*INX)[npoints-5];
	  dx[2]=(*INX)[npoints-4];
	  dx[3]=(*INX)[npoints-3];
	  dx[4]=(*INX)[npoints-2];
	  dy[1]=(*INY)[npoints-5];
	  dy[2]=(*INY)[npoints-4];
	  dy[3]=(*INY)[npoints-3];
	  dy[4]=(*INY)[npoints-2];
	  
	  for (i=0; i<appnum; i++)
	    {
	      t = (double)i * dt;
	      GenBspline(3, dx, dy, &outx, &outy, t);
	      
	      if (!OUTX->add(outx))
		return false;
	      if (!OUTY->add(outy))
		return false;
	    }
	}
      
      //last region
      dx[1]=(*INX)[npoints-4];
      dx[2]=(*INX)[npoints-3];
      dx[3]=(*INX)[npoints-2];
      dx[4]=(*INX)[npoints-1];
      dy[1]=(*INY)[npoints-4];
      dy[2]=(*INY)[npoints-3];
      dy[3]=(*INY)[npoints-2];
      dy[4]=(*INY)[npoints-1];
      
      for (i=0; i<=appnum; i++)
	{
	  t = (double)i * dt;
	  GenBspline(4, dx, dy, &outx, &outy, t);
	  
	  if (!OUTX->add(outx))
	    return false;
	  if (!OUTY->add(outy))
	    return false;
	}
    }
  else if (end_terminal == 2)
    {
      
      //periodic region
      for (j = 1; j <= npoints; j++)
	{
	  dx[1]=(*INX)[(j-1)%npoints];
	  dx[2]=(*INX)[(j)%npoints];
	  dx[3]=(*INX)[(j+1)%npoints];
	  dx[4]=(*INX)[(j+2)%npoints];
	  dy[1]=(*INY)[(j-1)%npoints];
	  dy[2]=(*INY)[(j)%npoints];
	  dy[3]=(*INY)[(j+1)%npoints];
	  dy[4]=(*INY)[(j+2)%npoints];
	  
	  for (i=0; i<appnum; i++)
	    {
	      t = (double)i * dt;
	      GenBspline(0, dx, dy, &outx, &outy, t);
	      
	      if (!OUTX->add(outx))
		return false;
	      if (!OUTY->add(outy))
		return false;
	    }
	  
	  if (npoints == npoints)
	    {
	      t = (double)appnum * dt;
	      GenBspline(0, dx, dy, &outx, &outy, t);
	      
	      if (!OUTX->add(outx))
		return false;
	      if (!OUTY->add(outy))
		return false;
	    }
	}
    }
  return true;
}

void GenBspline(int mode, double *dx, double *dy,double *outx, double *outy, double t)
{
  double B[5];
  int i;
  double tempx,tempy;
  
  switch(mode)
    {
    case 0:
      B[1] = (1.0-3.0*t+3.0*t*t-t*t*t)/6.0;
      B[2] = (4.0-6.0*t*t+3.0*t*t*t)/6.0;
      B[3] = (1.0+3.0*t+3.0*t*t-3.0*t*t*t)/6.0;
      B[4] = t*t*t/6.0;
      break;
    case 1:
      B[1] = 1.0-3.0*t+3.0*t*t-t*t*t;
      B[2] = (36.0*t-54.0*t*t+21.0*t*t*t)/12.0;
      B[3] = (18.0*t*t-11.0*t*t*t)/12.0;
      B[4] = t*t*t/6.0;
      break;
    case 2:
      B[1] = (1.0-3.0*t+3.0*t*t-t*t*t)/4.0;
      B[2] = (7.0+3.0*t-15.0*t*t+7.0*t*t*t)/12.0;
      B[3] = (1.0+3.0*t+3.0*t*t-3.0*t*t*t)/6.0;
      B[4] = t*t*t/6.0;
      break;
    case 3:
      B[1] = (1.0-3.0*t+3.0*t*t-t*t*t)/6.0;
      B[2] = (1.0+3.0*(1.0-t)+3.0*(1.0-t)*(1.0-t)-3.0*(1.0-t)*(1.0-t)*(1.0-t))/6.0;
      B[3] = (7.0+3.0*(1.0-t)-15.0*(1.0-t)*(1.0-t)+7.0*(1-t)*(1-t)*(1-t))/12.0;
      B[4] = t*t*t/4.0;
      break;
    case 4:
      B[1] = (1.0-3.0*t+3.0*t*t-t*t*t)/6.0;
      B[2] = (18.0*(1.0-t)*(1.0-t)-11.0*(1.0-t)*(1.0-t)*(1.0-t))/12.0;
      B[3] = (36.0*(1.0-t)-54.0*(1.0-t)*(1.0-t)+21.0*(1-t)*(1-t)*(1-t))/12.0;
      B[4] = t*t*t;
      break;
      
    }
  
  
  tempx = 0.0;tempy = 0.0;
  for (i = 1; i <= 4; i++)
    {
      tempx += dx[i]*B[i];
      tempy += dy[i]*B[i];
    }
  
  *outx = tempx;
  *outy = tempy;
  
  return;
}
