/*----------------------------------------------------------------------------
                  buffarray.cc (definition of string array)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include "buffarray.h"
#include <stdio.h>
#include <string.h>
#include <string>

buffarray::buffarray()
{
  int i;
  arraylength = 0;
  increment = 10;

  bufflength = (arraylength / increment + 1) * increment;

  strarray =   new std::string * [bufflength];
  memset(strarray, 0, bufflength * sizeof(int));
  for (i = 0; i < arraylength; i++)
    strarray[i] = new std::string("");

}


buffarray::~buffarray()
{
  int i;
  for (i = 0; i < arraylength; i++)
    delete strarray[i];

  delete [] strarray;
}

bool buffarray::setscalar(const char *val)
{
  setbuff(0);
  return add(val);
}

bool buffarray::setdata(int pos, const char *val)
{
  if (pos < arraylength)
      *strarray[pos] =  std::string(val);
  else
    {
      changebufflength(pos + 1);
      for (int i = arraylength; i <= pos; i++)
	strarray[i] = new std::string("");

      *strarray[pos] =  std::string(val);
      arraylength = pos + 1;
    }
  return true;
}

std::string buffarray::getdata(int pos)
{
  std::string ret = std::string("");
  if (pos < arraylength)
    return *strarray[pos];
  else
    return ret;
}

int buffarray::getarraylength()
{
  return arraylength;
}

bool buffarray::setbuff(int length)
{
  int i;

  for (i = 0; i < arraylength; i++)
    delete strarray[i];
  delete [] strarray;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  strarray =  new std::string * [bufflength];
  if (strarray ==  0)
    return false;

  memset(strarray, 0, bufflength * sizeof(int));
  return true;
}

bool buffarray::setarray(int length)
{
  int i;

  if (setbuff(length) == false)
    return false;

  arraylength = length;

  for (i = 0; i < arraylength; i++)
    strarray[i] = new std::string("");
  
  return true;

}


bool buffarray::add(const char *val)
{

  if (arraylength < bufflength)
    {
      arraylength++;
      strarray[arraylength - 1] = new std::string(val);
    }
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
      strarray[arraylength - 1] = new std::string(val);
    }
  
  if (strarray[arraylength - 1] == 0)
    return false;
  else
    return true;
}

bool buffarray::addat(int pos, const char *val)
{
  int i;
  if (pos < arraylength)
    {
      if (arraylength >=  bufflength)
	{
	  if (changebufflength(++arraylength) == false)
	    return false;
	}
      else
	arraylength++;

      for (i = arraylength - 1; i > pos; i--)
	  strarray[i] = strarray[i - 1];
      
      strarray[pos] = new std::string(val);

      return true;
    }
  else
    return add(val);
}

bool buffarray::remove(int pos)
{
  int i;
  for (i = pos; i < arraylength - 1; i++)
    {
      if (setdata(i, getdata(i + 1).c_str()) == false)
	return false;
    }
  
  delete strarray[arraylength - 1];
  arraylength--;
  return true;
}

bool buffarray::changebufflength(int newlength)
{
  int i;
  std::string **tempstrarray;

  if (newlength <= bufflength)
    return true;

  bufflength = (newlength / increment + 1) * increment;
  tempstrarray =   new std::string * [bufflength];
  if (tempstrarray == 0)
    return false;

  memset(tempstrarray, 0, bufflength * sizeof(int));

  for (i = 0; i < arraylength; i++)
    tempstrarray[i] = strarray[i];

  delete [] strarray;

  strarray = tempstrarray;
  return true;
}

buffarray& buffarray::operator = (const buffarray&  buff)
{
  int i;  
  setbuff(buff.bufflength);
  for (i = 0; i < buff.arraylength; i++)
    add(buff.strarray[i]->c_str());
  return *this;
}

std::string& buffarray::operator [] (int pos)
{
    return *strarray[pos];
}





