/*----------------------------------------------------------------------------
                      calc.cc (numerical transformation)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include  <stdlib.h>
#include  <string.h>
#include <ctype.h>
#include <string>
#include  "calc.h"
#include "data.h"
#include "gamma.h"
#include "graph.h"
#include "frame.h"
#include "naninf.h"
#include "topazvalues.h"

extern bool dinvq(double in, double *out);
extern graph *gra;
extern void message(const char *msg);

int sp1, sp2;
double FormulaConst[27];
std::string TempFor;

TOKENREC stack[PARSERSTACKSIZE] , polish[PARSERSTACKSIZE] , curtoken;
TOKENREC v[PARSERSTACKSIZE];
int stacktop, tokentype, error;
char *input, isformula;

bool ValueUserConst(frame *fr, char *label, double *retvalue);


int nexttoken(data* dat, int cellno, int seriesno)
/* read next token from input stream */
{
  char *start, numstring[MAXINPUT+1];
  int decimal, len, numlen;
  char temp[20];
  int alphalen;
  char temp_alpha[21];
  double ftemp;
  bool Flag;
  char tempchar;
  frame *parentframe = (frame *)dat->parent;
  
  // reached end of data
  if (*input == 0)
    {
      curtoken.status = EOLN;
      return(EOLN);
    }
  
  if (strchr("0123456789.", *input))
    {
      start = input;
      len = 0;
      decimal = false;
      // is number?
      while ((isdigit((unsigned char)*input)) ||
	     ((*input == '.') && (!decimal)))
	{
	  if (*input == '.')
	    decimal = true;
	  input++;
	  len++;
	}
      
      // only '.' is invalid
      if ((len == 1) && (start[0] == '.'))
	return(BAD);
      
      // exponential expression
      if (*input == 'E' || *input == 'e')
	{
	  input++;
	  len++;
	  //if '+-' next to 'E', increment pointer
	  if (strchr("+-", *input) != 0)
	    {
	      input++;
	      len++;
	    }
	  numlen = 0;
	  // increment pointer to the end of the exponential expression
	  while ((isdigit((unsigned char)*input)) && (++numlen <= 3))
	    {
	      input++;
	      len++;
	    }
	}
      strncpy(numstring, start, len);
      numstring[len] = 0;
      // set current token
      curtoken.x.value = atof(numstring);
      curtoken.priority = PRI_OPERAND;
      curtoken.status = NUM;
      
      return(NUM);
    }
  // in case of alphabetic character
  else if (isalpha((unsigned char)*input))
    {
      Flag = false;
      for (int i = 0; i < 20; i++)
	{
	  tempchar = (unsigned char)*(input + i);
	  if (isalpha(tempchar) == 0 && isdigit(tempchar) == 0 && (tempchar != '_'))
	    {
	      alphalen = i;
	      strncpy(temp_alpha, input, alphalen);
	      temp_alpha[alphalen] = 0;
	      i = 21;
	      Flag = true;
	    }
	}
      if (Flag == false)
	return (BAD);
      
      if (strcmp("x", temp_alpha) == 0)
	{
	  curtoken.x.datarowno = 0;
	  curtoken.priority = PRI_DATAROW;
	  curtoken.status = DATAROW;
	  input ++;
	}
      else if (strcmp("y", temp_alpha) == 0)
	{
	  curtoken.x.datarowno = 1;
	  curtoken.priority = PRI_DATAROW;
	  curtoken.status = DATAROW;
	  input ++;
	}
      else if (strcmp("INF", temp_alpha) == 0)
	{
	  curtoken.x.value = infvalue();
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input +=3;
	}
      else if (strcmp("inf", temp_alpha) == 0)
	{
	  curtoken.x.value = infvalue();
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input +=3;
	}
      else if (strcmp("PI", temp_alpha) == 0)
	{
	  curtoken.x.value =3.141592653589793;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 2;
	}
      else if (strcmp("pi", temp_alpha) == 0)
	{
	  curtoken.x.value =3.141592653589793;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 2;
	}
      else if (strcmp("sinh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 7;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("cosh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 8;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("tanh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 9;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("sin", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 1;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("cos", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 2;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("tan", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 3;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("asin", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 4;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("acos", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 5;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("atan", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 6;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("abs", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 10;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("sqrt", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 11;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("exp", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 12;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("ln", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 13;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("log", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 14;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("max", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 15;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("min", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 16;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("theta", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 17;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("delta", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 18;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("sign", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 19;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("int", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 20;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("gauss", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 21;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("frac", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 22;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("round", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 23;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("gamma", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 24;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("beta", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 25;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("erf", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 26;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("rand", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 27;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("cell", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 28;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("eq", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 29;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("neq", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 30;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("ge", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 31;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("gt", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 32;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("le" ,temp_alpha) == 0)
	{
	  curtoken.x.funcno = 33;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("lt", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 34;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("not", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 35;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("or", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 36;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("and", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 37;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("xor", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 38;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("if", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 39;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("dx", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 40;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("dy", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 41;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 2;
	}
      else if (strcmp("nrx", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 42;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("nry", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 43;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 3;
	}
      else if (strcmp("invq", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 44;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("erfc", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 45;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("pnor", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 46;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("qnor", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 47;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("invp", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 48;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 4;
	}
      else if (strcmp("inverf", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 49;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 6;
	}
      else if (strcmp("inverfc", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 50;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 7;
	}
      else if (strcmp("asinh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 51;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("acosh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 52;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
      else if (strcmp("atanh", temp_alpha) == 0)
	{
	  curtoken.x.funcno = 53;
	  curtoken.priority = PRI_FUNCTION;
	  curtoken.status = FUNC;
	  input += 5;
	}
       else if (strcmp("xmin", temp_alpha) == 0)
	{
	  curtoken.x.value = dat->bound_xmin;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 4;
	}
      else if (strcmp("xmax", temp_alpha) == 0)
	{
	  curtoken.x.value = dat->bound_xmax;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 4;
	}
      else if (strcmp("ymin", temp_alpha) == 0)
	{
	  curtoken.x.value = dat->bound_ymin;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 4;
	}
      else if (strcmp("ymax", temp_alpha) == 0)
	{
	  curtoken.x.value = dat->bound_ymax;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += 4;
	}
      else if (ValueUserConst(parentframe, temp_alpha, &ftemp))
	{
	  curtoken.x.value = ftemp;
	  curtoken.priority = PRI_OPERAND;
	  curtoken.status = NUM;
	  input += strlen(temp_alpha);
	}
      else
	{
	  return (BAD);
	}
      return(FUNC);
    }
  
  // others
  else
    {
      switch(*(input++))
	{
	case '+' :
	  if (*(input - 2) == '('
	      || *(input - 2) == ','
	      || *(input - 2) == '^'
	      || *(input - 2) == '='
	      || *(input - 2) == '<'
	      || *(input - 2) == '>'
	      )
	    curtoken.status = PREPLUS;
	  else
	    curtoken.status = PLUS;
	  curtoken.priority = PRI_PRUSMINUS;	return(PLUS);
	case '-' :
	  if (*(input - 2) == '('
	      || *(input - 2) == ','
	      || *(input - 2) == '^'
	      || *(input - 2) == '='
	      || *(input - 2) == '<'
	      || *(input - 2) == '>'
	      )
	    curtoken.status = PREMINUS;
	  else
	    curtoken.status = MINUS;
	  curtoken.priority = PRI_PRUSMINUS;	return(MINUS);
	case '*' : curtoken.priority = PRI_MULDIV;		curtoken.status = TIMES;	return(TIMES);
	case '/' : curtoken.priority = PRI_MULDIV;		curtoken.status = DIVIDE;	return(DIVIDE);
	case '(' : curtoken.priority = PRI_OPAREN;		curtoken.status = OPAREN;	return(OPAREN);
	case ')' : curtoken.priority = PRI_CPAREN;		curtoken.status = CPAREN;	return(CPAREN);
	case '^' : curtoken.priority = PRI_EXP;  		curtoken.status = EXP;		return(EXP);
	case ',' : curtoken.priority = PRI_COMMA;		curtoken.status = COMMA; 	return(COMMA);
	case '|' :
	  if (input[0]=='|')     //or
	    {
	      curtoken.priority = PRI_OR;
	      curtoken.status = OR;
	      input ++;
	      return(OR);
	    }
	  else
	    return (BAD);
	  break;
	case '&' :
	  if (input[0]=='&')     //and
	    {
	      curtoken.priority = PRI_AND;
	      curtoken.status = AND;
	      input ++;
	      return(AND);
	    }
	  else
	    return (BAD);
	  break;
	case '=' :
	  if (input[0]=='=')     //equal
	    {
	      curtoken.priority = PRI_EQUAL;
	      curtoken.status = EQUAL;
	      input ++;
	      return(EQUAL);
	    }
	  else
	    return (BAD);
	  break;
	case '!' :
	  if (input[0]=='=')     //not equal
	    {
	      curtoken.priority = PRI_NOTEQUAL;
	      curtoken.status = NOTEQUAL;
	      input ++;
	      return(NOTEQUAL);
	    }
	  else
	    return (BAD);
	  break;
	case '<' :
	  if (input[0]=='=')     //smaller equal than
	    {
	      curtoken.priority = PRI_SMALLER_EQUAL_THAN;
	      curtoken.status = SMALLER_EQUAL_THAN;
	      input ++;
	      return(SMALLER_EQUAL_THAN);
	    }
	  else               		//smaller than
	    {
	      curtoken.priority = PRI_SMALLER_THAN;
	      curtoken.status = SMALLER_THAN;
	      return(SMALLER_THAN);
	    }
	  break;
	case '>' :
	  if (input[0]=='=')     //larger equal than
	    {
	      curtoken.priority = PRI_LARGER_EQUAL_THAN;
	      curtoken.status = LARGER_EQUAL_THAN;
	      input ++;
	      return(LARGER_EQUAL_THAN);
	    }
	  else               		//smaller than
	    {
	      curtoken.priority = PRI_LARGER_THAN;
	      curtoken.status = LARGER_THAN;
	      return(LARGER_THAN);
	    }
	  break;
	case '%' :
	  if (input[0]=='i' || input[0]=='I')     // data No.
	    {
	      if (seriesno == -1)
		return (BAD);
	      
	      curtoken.x.value =(double)seriesno;
	      curtoken.priority = PRI_OPERAND;
	      curtoken.status = NUM;
	      input ++;
	      return(NUM);
	    }
	  else if (input[0]=='c' || input[0]=='C')     // cell No.
	    {
	      curtoken.x.value =(double)cellno;
	      curtoken.priority = PRI_OPERAND;
	      curtoken.status = CELLNO;
	      input ++;
	      return(CELLNO);
	    }
	  else if (isdigit(input[0]) != 0 && isdigit(input[1]) != 0 && isdigit(input[2]) == 0)   //row
	    {
	      if (seriesno == -1)
		return (BAD);
	      
	      temp[0] = input[0];
	      temp[1] = input[1];
	      temp[2] = 0;
	      
	      curtoken.x.datarowno = atoi(temp) - 1;
	      if (curtoken.x.datarowno < 0 || curtoken.x.datarowno > 99)
		return BAD;

	      curtoken.priority = PRI_DATAROW;
	      curtoken.status = PARAMDATAROW;
	      input += 2;
	      return(DATAROW);
	    }
	  else if (isdigit(input[0]) != 0 && isdigit(input[1]) != 0 && isdigit(input[2]) != 0 &&
		   input[3] == ':' &&
		   isdigit(input[4]) != 0 && isdigit(input[5]) != 0 && isdigit(input[6]) == 0
		   )   //other data
	    {
	      if (seriesno == -1)
		return (BAD);
	      
	      // row no
	      temp[0] = input[4];
	      temp[1] = input[5];
	      temp[2] = 0;
	      curtoken.x.datarowno = atoi(temp) - 1;
	      
	      if (curtoken.x.datarowno < 0 || curtoken.x.datarowno > 99)
		return BAD;

	      // data object no
	      temp[0] = input[0];
	      temp[1] = input[1];
	      temp[2] = input[2];
	      temp[3] = 0;
	      curtoken.x.funcno = atoi(temp);

	      if (0 <= curtoken.x.funcno &&
		  curtoken.x.funcno < parentframe->darray->getarraylength())
		{
		  curtoken.priority = PRI_DATAROW;
		  curtoken.status = OTHERDATAROW;
		  input += 6;
		  return(DATAROW);
		}
	      else
		return BAD;
	    }
	  else if (isdigit(input[0]) != 0 && isdigit(input[1]) != 0 && isdigit(input[2]) != 0 &&
		   input[3] == ':' &&
		   (input[4] == 'x' || input[4] == 'X') && isdigit(input[5]) == 0
		   )   //other data
	    {
	      if (seriesno == -1)
		return (BAD);
	      
	      // row no
	      curtoken.x.datarowno = dat->col_x - 1; 
	      
	      // data object no
	      temp[0] = input[0];
	      temp[1] = input[1];
	      temp[2] = input[2];
	      temp[3] = 0;
	      curtoken.x.funcno = atoi(temp);

	      if (0 <= curtoken.x.funcno &&
		  curtoken.x.funcno < parentframe->darray->getarraylength())
		{
		  curtoken.priority = PRI_DATAROW;
		  curtoken.status = OTHER_X_DATAROW;
		  input += 5;
		  return(DATAROW);
		}
	      else
		return BAD;
	    }
	  else if (isdigit(input[0]) != 0 && isdigit(input[1]) != 0 && isdigit(input[2]) != 0 &&
		   input[3] == ':' &&
		   (input[4] == 'y' || input[4] == 'Y') && isdigit(input[5]) == 0
		   )   //other data
	    {
	      if (seriesno == -1)
		return (BAD);
	      
	      // row no
	      curtoken.x.datarowno = dat->col_y - 1;
	      
	      // data object no
	      temp[0] = input[0];
	      temp[1] = input[1];
	      temp[2] = input[2];
	      temp[3] = 0;
	      curtoken.x.funcno = atoi(temp);

	      if (0 <= curtoken.x.funcno &&
		  curtoken.x.funcno < parentframe->darray->getarraylength())
		{
		  curtoken.priority = PRI_DATAROW;
		  curtoken.status = OTHER_Y_DATAROW;
		  input += 5;
		  return(DATAROW);
		}
	      else
		return BAD;
	    }
	  else
	    return (BAD);
	  break;
	  default  : return(BAD);
	} /* switch */
      
    }
  
} /* nexttoken */




int SetFormula (data *dat, int cellno, int seriesno)
{
  
  int i;
  int flag;
  
  stack[0].priority=-1;
  stack[0].x.value=0;
  
  sp1=sp2=0;
  i=2;
  
  while ((flag = nexttoken(dat, cellno, seriesno))!=EOLN)
    {
      if (flag == BAD)
	return false;
      while (curtoken.priority <= stack[sp1].priority && stack[sp1].status != OPAREN && stack[sp1].status != COMMA)
			polish[++sp2] = stack[sp1--];
      if (curtoken.status != CPAREN && curtoken.status != COMMA)
	stack[++sp1]=curtoken;
      else if (curtoken.status == COMMA)
	sp1=sp1;
      else if (curtoken.status == CPAREN && stack[sp1].status == OPAREN)
	sp1--;
      
      i++;
    }
  
  // push all token to polsh
  for (i = sp1; i > 0; i--)
    polish[++sp2] = stack[i];
  
  return true;
}

int Calc (data* dat, int cellno, int Isminmax)
{
  int i,ix,itemp;
  double ftemp;
  int sizeX, sizeY;
  sizeX = dat->x->getarraylength() - 1;
  sizeY = dat->y->getarraylength() - 1;
  data* tmpdat;
  frame *parentframe = (frame *)dat->parent;
  
  sp1=0;
  for (i=1;i<=sp2;i++)
    {
      if (polish[i].status == NUM)
	v[++sp1].x.value = polish[i].x.value;
      else if (polish[i].status == DATAROW)
	{
	  if (polish[i].x.datarowno == 0)	//x-data
	    {
	      if (Isminmax == 0)
		v[++sp1].x.value = (*dat->x)[cellno].data;
	      else if (Isminmax == 1)		//min
		v[++sp1].x.value = (*dat->xmin)[cellno].data;
	      else if (Isminmax == 2)          //max
		v[++sp1].x.value = (*dat->xmax)[cellno].data;
	    }
	  else if (polish[i].x.datarowno == 1)    //y-data
	    {
	      if (Isminmax == 0)
		v[++sp1].x.value = (*dat->y)[cellno].data;
	      else if (Isminmax == 1)      //min
		v[++sp1].x.value = (*dat->ymin)[cellno].data;
	      else if (Isminmax == 2)      //max
		v[++sp1].x.value = (*dat->ymax)[cellno].data;
	    }
	}
      else if (polish[i].status == PARAMDATAROW) // row
	{
	  if (dat->extradatadispflag[polish[i].x.datarowno] == false)
	    return false;
	  v[++sp1].x.value = (*dat->extra[polish[i].x.datarowno])[cellno].data;
	}
      else if (polish[i].status == OTHERDATAROW) // other data object row
	{
	  tmpdat = &(*parentframe->darray)[polish[i].x.funcno];
	  if (tmpdat->extradatadispflag[polish[i].x.datarowno])
	    {
	      if ((tmpdat->extra)[polish[i].x.datarowno]->getarraylength() == sizeX + 1)
		  v[++sp1].x.value = (*tmpdat->extra[polish[i].x.datarowno])[cellno].data;
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (polish[i].status == OTHER_X_DATAROW) // other data object x
	{
	  tmpdat = &(*parentframe->darray)[polish[i].x.funcno];
	  if (tmpdat->x->getarraylength() == sizeX + 1)
	    v[++sp1].x.value = (*tmpdat->x)[cellno].data;
	  else
	    return false;
	}
      else if (polish[i].status == OTHER_Y_DATAROW) //other data object y
	{
	  tmpdat = &(*parentframe->darray)[polish[i].x.funcno];
	  if (tmpdat->y->getarraylength() == sizeX + 1)
	    v[++sp1].x.value = (*tmpdat->y)[cellno].data;
	  else
	    return false;
	}
      else if (polish[i].status == CELLNO)
	v[++sp1].x.value = cellno;
      else if (polish[i].status == FUNC)
	{
	  switch(polish[i].x.funcno)
	    {
	    case 1:	//sin
	      if (sp1>=1)
		v[sp1].x.value=sin(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 2: //cos
	      if (sp1>=1)
		v[sp1].x.value=cos(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 3: //tan
	      if (sp1>=1)
		v[sp1].x.value=tan(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 4: //asin
	      if (sp1>=1)
		v[sp1].x.value=asin(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 5: //acos
	      if (sp1>=1)
		v[sp1].x.value=acos(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 6: //atan
	      if (sp1>=1)
		v[sp1].x.value=atan(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 7:	//sinh
	      if (sp1>=1)
		v[sp1].x.value=sinh(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 8: //cosh
	      if (sp1>=1)
		v[sp1].x.value=cosh(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 9: //tanh
	      if (sp1>=1)
		v[sp1].x.value=tanh(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 10: //abs
	      if (sp1>=1)
		v[sp1].x.value=fabs(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 11: //sqrt
	      if (sp1>=1)
		v[sp1].x.value=sqrt(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 12: //exp
	      if (sp1>=1)
		v[sp1].x.value=exp(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 13: //ln
	      if (sp1>=1)
		v[sp1].x.value=log(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 14: //log
	      if (sp1>=1)
		v[sp1].x.value=log10(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 15: //max
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value > v[sp1].x.value)
		    v[sp1-1].x.value=v[sp1-1].x.value;
		  else
		    v[sp1-1].x.value=v[sp1].x.value;
		  
		  sp1--;
		  
		}
	      else
		return false;
	      break;
	    case 16: //min
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value < v[sp1].x.value)
		    v[sp1-1].x.value=v[sp1-1].x.value;
		  else
		    v[sp1-1].x.value=v[sp1].x.value;

		  sp1--;
		  
		}
	      else
		return false;
	      break;
	    case 17: //theta
	      if (sp1>=1)
		{
		  if (v[sp1].x.value>=0.0)
		    v[sp1].x.value=1.0;
		  else
		    v[sp1].x.value=0.0;
		}
	      else
		return false;
	      break;
	    case 18: //delta
	      if (sp1>=1)
		{
		  if (v[sp1].x.value==0.0)
		    v[sp1].x.value=1.0;
		  else
		    v[sp1].x.value=0.0;
		}
	      else
		return false;
	      break;
	    case 19:	//sign
	      if (sp1>=1)
		if (v[sp1].x.value>0.0)
		  v[sp1].x.value=1.0;
		else if (v[sp1].x.value<0.0)
		  v[sp1].x.value=-1.0;
		else
		  v[sp1].x.value=0.0;
	      else
		return false;
	      break;
	    case 20: //int
	      if (sp1>=1)
		if (v[sp1].x.value>0.0)
		  v[sp1].x.value=floor(v[sp1].x.value);
		else
		  v[sp1].x.value=-floor(fabs(v[sp1].x.value));
	      else
		return false;
	      break;
	    case 21: //Gauss
	      if (sp1>=1)
		v[sp1].x.value=floor(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 22: //frac
	      if (sp1>=1)
		{
		  ftemp=v[sp1].x.value;
		  modf(ftemp,&v[sp1].x.value);
		}
	      else
		return false;
	      break;
	    case 23: //round
	      if (sp1>=1)
		{
		  ftemp=fabs(v[sp1].x.value);
		  if (ftemp-floor(ftemp)>=.5)
		    {
		      if (v[sp1].x.value>0)
			v[sp1].x.value=floor(ftemp)+1;
		      else
			v[sp1].x.value=-(floor(ftemp)+1);
		      
		    }
		  else
		    {
		      if (v[sp1].x.value>0)
			v[sp1].x.value=floor(ftemp);
		      else
			v[sp1].x.value=-floor(ftemp);
		    }
		  
		}
	      else
		return false;
	      break;
	    case 24: //gamma function
	      if (sp1>=1)
		v[sp1].x.value=gamma(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 25: //beta function
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=beta(v[sp1-1].x.value,v[sp1].x.value);
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 26: //erf
	      if (sp1>=1)
		
		v[sp1].x.value=erf(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 27: //ramdom
	      if (sp1>=1)
		{
		  v[sp1].x.value = (double)rand()/(double)RAND_MAX*(v[sp1].x.value);
		}
	      else
		return false;
	      break;
	    case 29: //eq
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value == v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 30: //neq
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value != v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 31: //GE
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value >= v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 32: //GT
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value > v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 33: //LE
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value <= v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 34: //LT
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value < v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 35: //NOT
	      if (sp1>=1)
		{
		  if (v[sp1].x.value == 0.0)
		    v[sp1].x.value=1.0;
		  else
		    v[sp1].x.value=0.0;
		}
	      else
		return false;
	      break;
	    case 36: //OR
	      if (sp1>=2)
		{
		  if ((v[sp1-1].x.value!=0.0) || (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;

		  sp1--;
		}
	      else
		return false;
	      break;
	    case 37: //AND
	      if (sp1>=2)
		{
		  if ((v[sp1-1].x.value!=0.0) && (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 38: //XOR
	      if (sp1>=2)
		{
		  if ((v[sp1-1].x.value!=0.0) && (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=0.0;
		  else if ((v[sp1-1].x.value!=0.0) && (v[sp1].x.value==0.0))
		    v[sp1-1].x.value=1.0;
		  else if ((v[sp1-1].x.value==0.0) && (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case 39: //IF
	      if (sp1>=3)
		{
		  if (v[sp1-2].x.value!=0.0)
		    v[sp1-2].x.value=v[sp1-1].x.value;
		  else
		    v[sp1-2].x.value=v[sp1].x.value;
		  
		  sp1=sp1-2;
		}
	      else
		return false;
	      break;
	    case 40: //dx
	      if (sp1>=1)
		{
		  itemp = (int)floor(v[sp1].x.value + .5);
		  if (itemp >= 1 && itemp <= 4)
		    {
		      switch(itemp)
			{
			case 1:
			  if (cellno > 0)
			    v[sp1].x.value = (*dat->x)[cellno].data - (*dat->x)[cellno - 1].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 2:
			  if (cellno > 1 )
			    v[sp1].x.value = (*dat->x)[cellno].data
			      - 2 * (*dat->x)[cellno - 1].data
				+ (*dat->x)[cellno - 2].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 3:
			  if (cellno > 2 )
			    v[sp1].x.value = (*dat->x)[cellno].data
			      - 3 * (*dat->x)[cellno - 1].data
				+ 3 * (*dat->x)[cellno - 2].data
				  - (*dat->x)[cellno - 3].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 4:
			  if (cellno > 3 )
			    v[sp1].x.value = (*dat->x)[cellno].data
			      - 4 * (*dat->x)[cellno - 1].data
				+ 6 * (*dat->x)[cellno - 2].data
				  - 4 * (*dat->x)[cellno - 3].data
				    + (*dat->x)[cellno - 4].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			}
		    }
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 41: //dy
	      if (sp1>=1)
		{
		  itemp = (int)floor(v[sp1].x.value + .5);
		  if (itemp >= 1 && itemp <= 4)
		    {
		      switch(itemp)
			{
			case 1:
			  if (cellno > 0)
			    v[sp1].x.value = (*dat->y)[cellno].data - (*dat->y)[cellno - 1].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 2:
			  if (cellno > 1 )
			    v[sp1].x.value = (*dat->y)[cellno].data
			      - 2 * (*dat->y)[cellno - 1].data
				+ (*dat->y)[cellno - 2].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 3:
			  if (cellno > 2 )
			    v[sp1].x.value = (*dat->y)[cellno].data
			      - 3 * (*dat->y)[cellno - 1].data
				+ 3 * (*dat->y)[cellno - 2].data
				  - (*dat->y)[cellno - 3].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			case 4:
			  if (cellno > 3 )
			    v[sp1].x.value = (*dat->y)[cellno].data
			      - 4 * (*dat->y)[cellno - 1].data
				+ 6 * (*dat->y)[cellno - 2].data
				  - 4 * (*dat->y)[cellno - 3].data
				    + (*dat->y)[cellno - 4].data;
			  else
			    v[sp1].x.value = infvalue();
			  break;
			}
		    }
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 42: //nrx  noize reduction X
	      if (sp1>=1)
		{
		  itemp = (int)floor(v[sp1].x.value + .5);
		  if (itemp >= 1 && itemp <= 100)
		    {
		      if (cellno >= itemp && cellno <= sizeX - itemp)
			{
			  ftemp = 0.0;
			  for (ix = cellno - itemp; ix <= cellno + itemp; ix++)
			    ftemp += (*dat->x)[ix].data;
			  v[sp1].x.value = ftemp / (double)(2 * itemp + 1);
			}
		      else
			v[sp1].x.value = infvalue();
		    }
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 43: //nry  noize reduction Y
	      if (sp1>=1)
		{
		  itemp = (int)floor(v[sp1].x.value + .5);
		  if (itemp >= 1 && itemp <= 100)
		    {
		      if (cellno >= itemp && cellno <= sizeY - itemp)
			{
			  ftemp = 0.0;
			  for (ix = cellno - itemp; ix <= cellno + itemp; ix++)
			    ftemp += (*dat->y)[ix].data;
			  v[sp1].x.value = ftemp / (double)(2 * itemp + 1);
			}
		      else
			v[sp1].x.value = infvalue();
		    }
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 44: //inverse qnor = invq
	      if (sp1>=1)
		{
		  if (dinvq(v[sp1].x.value, &ftemp) == true)
		    v[sp1].x.value = ftemp;
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 45: //erfc
	      if (sp1>=1)
		
		v[sp1].x.value=erfc(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 46: //p_normal
	      if (sp1>=1)
		
		v[sp1].x.value=p_normal(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 47: //q_normal
	      if (sp1>=1)
		
		v[sp1].x.value=q_normal(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 48: //inverse pnor = invp
	      if (sp1>=1)
		{
		  if (dinvq(1 - v[sp1].x.value, &ftemp) == true)
		    v[sp1].x.value = ftemp;
		  else
		    return false;
		}
	      else
		return false;
	      break;
	    case 49: //inverse erf
	      if (sp1>=1)
		{
		  if (v[sp1].x.value <= -1 || v[sp1].x.value >= 1)
		    return false;
		  else
		    {
		      if (dinvq((1 + v[sp1].x.value) / 2.0, &ftemp) == true)
			v[sp1].x.value = - ftemp / sqrt(2.0);
		      else
			return false;
		    }
		}
	      else
		return false;
	      break;
	    case 50: //inverse erfc
	      if (sp1>=1)
		{
		  if (v[sp1].x.value <= 0 || v[sp1].x.value >= 2)
		    return false;
		  else
		    {
		      if (dinvq(v[sp1].x.value / 2.0, &ftemp) == true)
			v[sp1].x.value = ftemp / sqrt(2.0);
		      else
			return false;
		    }
		}
	      else
		return false;
	      break;
	    case 51:	//asinh
	      if (sp1>=1)
		v[sp1].x.value=asinh(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 52: //acosh
	      if (sp1>=1)
		v[sp1].x.value=acosh(v[sp1].x.value);
	      else
		return false;
	      break;
	    case 53: //atanh
	      if (sp1>=1)
		v[sp1].x.value=atanh(v[sp1].x.value);
	      else
		return false;
	      break;
	    default:
	      return false;
	    }
	}
      else
	{
	  switch (polish[i].status)
	    {
	    case PLUS:
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=v[sp1-1].x.value+v[sp1].x.value;
		  sp1--;
		}
	      else
		return false;
	      break;
	    case MINUS:
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=v[sp1-1].x.value-v[sp1].x.value;
		  sp1--;
		}
	      else
		return false;
	      break;
	    case TIMES:
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=v[sp1-1].x.value*v[sp1].x.value;
		  sp1--;
		}
	      else
		return false;
	      break;
	    case DIVIDE:
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=v[sp1-1].x.value/v[sp1].x.value;
		  sp1--;
		}
	      else
		return false;
	      break;
	    case EXP:
	      if (sp1>=2)
		{
		  v[sp1-1].x.value=pow(v[sp1-1].x.value,v[sp1].x.value);
		  sp1--;
		}
	      else
		return false;
	      break;
	    case PREPLUS:
	      if (sp1>=1)
		v[sp1].x.value=v[sp1].x.value;
	      else
		return false;
	      break;
	    case PREMINUS:
	      if (sp1>=1)
		v[sp1].x.value=-v[sp1].x.value;
	      else
		return false;
	      break;
	    case OR:
	      if (sp1>=2)
		{
		  if ((v[sp1-1].x.value!=0.0) || (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;

		  sp1--;
		}
	      else
		return false;
	      break;
	    case AND: //AND
	      if (sp1>=2)
		{
		  if ((v[sp1-1].x.value!=0.0) && (v[sp1].x.value!=0.0))
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case EQUAL: //equal
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value == v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case NOTEQUAL: //not equal
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value != v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case LARGER_EQUAL_THAN: //larger than
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value >= v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case LARGER_THAN: //larger than
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value > v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case SMALLER_EQUAL_THAN: //smaller equal than
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value <= v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    case SMALLER_THAN: //smaller than
	      if (sp1>=2)
		{
		  if (v[sp1-1].x.value < v[sp1].x.value)
		    v[sp1-1].x.value=1.0;
		  else
		    v[sp1-1].x.value=0.0;
		  
		  sp1--;
		}
	      else
		return false;
	      break;
	    default:
	      return false;
	    }
	}
      
      
    }
  if (sp1==1)
    return true;
  else
    return false;
  
}




int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax)
{
  char *p;
  char temp3[10];
  std::string str, str1;
  frame *parentframe = (frame *)dat->parent;
  
  // initialization
  if (InitFormula)
    {
      TempFor = std::string("(");
      p = s;
      
      // remove space
      do
	{
	  // user function
	  if (*p == '#')
	    {
	      if (*(p + 1) != 0)
		{
		  if (('0' <= *(p + 1)) && (*(p + 1) <= '9'))
		    {
		      str = std::string("#");
		      temp3[0] = *(p + 1); temp3[1] = 0;
		      str += std::string(temp3);
		      if (parentframe->usrfnc->GetScalar(&str, &str1))
			{
			  TempFor += str1;
			}
		      else
			return false;
		      p += 2;
		    }
		  else
		    p++;
		}
	      else
		p++;
	    }
	  else if (*p == ' ')
	    {
	      p++;
	    }
	  else
	    {
	      temp3[0] = *p; temp3[1] = '\0';
	      TempFor += std::string(temp3);
	      p++;
	    }
	}while(*p != 0) ;
      
      TempFor += std::string(")");
      input = (char *)TempFor.c_str();

      if (SetFormula(dat, cellno, seriesno) == false)
	return false;
      else
	{
	  if (isnormalvalue(v[1].x.value))
	    *n = v[1].x.value;
	  else
	    *n = infvalue();
	  return true;
	}
    }
  // calculate
  else
    {
      if (Calc(dat, cellno, Isminmax) == false)
	return false;
      else
	{
	  if (isnormalvalue(v[1].x.value))
	    *n = v[1].x.value;
	  else
	    *n = infvalue();
	  return true;
	}
    }
}

bool ValueUserConst(frame *fr, char *label, double *retvalue)
{
  std::string str1, str2;
  
  str2 = std::string(label);
  if (fr->usrconst->GetScalar(&str2, &str1))
    {
      *retvalue = atof(str1.c_str());
      return true;
    }
  else
    return false;

  return true;
}


bool IsValidConstant(std::string *str)
{
  char reservedstring[] = ",inf,pi,sinh,cosh,tanh,sin,cos,tan,asin,acos,atan,abs,sqrt,exp,ln,log,\
max,min,theta,delta,sign,int,gauss,frac,round,gamma,beta,erf,rand,cell,eq,neq,ge,gt,le,lt,not,or,\
and,xor,if,dx,dy,nrx,nry,invq,erfc,pnor,qnor,invp,inverf,inverfc,";

  std::string str1;
  
  str1 = std::string(",");
  str1 += *str;
  str1 += std::string(",");
  
  if (strlen(str1.c_str()) == 1)
    {
      if ('A' <= str1.c_str()[0] && str1.c_str()[0] <= 'Z')
	return false;
      else
	return true;
    }
  else
    {
      if (strstr(reservedstring, str1.c_str()) != NULL)
	return false;
      else
	return true;
    }
  return true;
}

bool IsValidFunction(std::string *str)
{
  if (strlen(str->c_str()) == 2)
    {
      if (str->c_str()[0] == '#' && '0' <= str->c_str()[1] && str->c_str()[1] <= '9')
	{
	  return true;
	}
      else
	return false;
    }
  else
    return false;
  return true;
}















