/*----------------------------------------------------------------------------
                         col.h (definitions of color)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__col_h)
#define __col_h

#include <string>

/* linecap */
#define LINECAPBUTT         0
#define LINECAPROUND        1
#define LINECAPPROJECTING   2

/* linejoin */
#define LINEJOINMITER       0
#define LINEJOINROUND       1
#define LINEJOINBEVEL       2

/* linestyle */
#define LINESTYLESOLID             0
#define LINESTYLEDASH              1
#define LINESTYLELONGDASH          2
#define LINESTYLESHORTDASH         3
#define LINESTYLEDASHDOTDASH       4
#define LINESTYLEDASHDOTDOTDASH    5
#define LINESTYLEDOT               6
#define LINESTYLENULL              7

/* fillstyle */
#define FILLSTYLESOLID             0
#define FILLSTYLENULL              1

/* fillrule */
#define FILLRULEEVENODD            0
#define FILLRULEWINDING            1

/* fontface */
#define FONT_TIMESROMAN            0 
#define FONT_TIMESBOLD             1 
#define FONT_TIMESITALIC           2 
#define FONT_TIMESBOLDITALIC       3 
#define FONT_HELVETICA             4 
#define FONT_HELVETICABOLD         5 
#define FONT_HELVETICAOBLIQUE      6 
#define FONT_HELVETICABOLDOBLIQUE  7 
#define FONT_COURIER               8 
#define FONT_COURIERBOLD           9 
#define FONT_COURIEROBLIQUE        10 
#define FONT_COURIERBOLDOBLIQUE    11 
#define FONT_SYMBOL                12

#define FONT_MINCHO                1000
#define FONT_GOTHIC                1004



class Capcolor
{
public:
  Capcolor(){init();}
  unsigned short red;
  unsigned short green;
  unsigned short blue;
  void init(){red = 0; green = 0; blue = 0;}
  void setwhite(){red = 65535; green = 65535; blue = 65535;}
  void setblack(){red = 0; green = 0; blue = 0;}
  void setred(){red = 65535; green = 0; blue = 0;}
  void setgreen(){green = 65535; red = 0; blue = 0;}
  void setblue(){blue = 65535; red = 0; green = 0;}

  Capcolor& operator = (const Capcolor& p)
    {red = p.red; green = p.green; blue = p.blue; return *this;}
  int operator == ( const Capcolor& obj ) const
    { return this == &obj;}
};

class Caplinestyle
{
public:
  Caplinestyle()
    {
      init();
    }
  void init(){width = 0; cap = LINECAPPROJECTING; join = LINEJOINMITER; style = LINESTYLESOLID;}

  int width;
  int cap;
  int join; 
  int style;

  Caplinestyle& operator = (const Caplinestyle& p)
    {width = p.width; cap = p.cap; join = p.join; style = p.style;  return *this;}
  int operator == ( const Caplinestyle& obj ) const
    { return this == &obj;}
};

class Capfillstyle
{
public:
  Capfillstyle()
    {
      init();
    }
  void init(){fillrule = FILLRULEEVENODD; style = FILLSTYLESOLID;}
  int fillrule;
  int style;

  Capfillstyle& operator = (const Capfillstyle& p)
    {fillrule = p.fillrule; style = p.style;  return *this;}
  int operator == ( const Capfillstyle& obj ) const
    { return this == &obj;}

};

class Capfont
{
public:
  Capfont()
    {
      init();
    }

  ~Capfont(){;}
  void init(){face = FONT_TIMESROMAN; jpface = FONT_MINCHO; size = 200; angle = 0.0;}
  int       face;
  int       jpface;
  int       size;
  double    angle;

  Capfont& operator = (const Capfont& p)
    {face = p.face; jpface = p.jpface; size = p.size; angle = p.angle;  return *this;}
  int operator == ( const Capfont& obj ) const
    { return this == &obj;}

};


#endif












