/*----------------------------------------------------------------------------
                     completion.cc (completion of filename)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include <string>
#include "vectdefs.h"

extern "C"{
#include <readline/readline.h>
#include <readline/history.h>
}


extern void message(const char *msg);

bool buildfullpath(char *in, char *out)
{
  char cdir[1024];
  char *homedir;

  switch(in[0])
    {
    case 0:
    case '/':
      strcpy(out, in);
      return true;
    case '~':
      if ((homedir = getenv("HOME")) == NULL)
	return false;
      strcpy(out, homedir);
      strcat(out, in + 1);
      return true;
    default:
      getcwd(cdir, sizeof(cdir));
      strcpy(out, cdir);
      if (cdir[strlen(cdir) - 1] != '/')
	strcat(out, "/");
      strcat(out, in);
      return true;
    }
  return true;
}

bool separatepath(char *fullpath, char *path, char *file)
{
  int lastslashpos = -1;
  char stemp[3];
  char tmpfullpath[1024];

  if(!buildfullpath(fullpath, tmpfullpath))
    return false;

  // getlast slash position
   for (int i = 0; i < (signed)strlen(tmpfullpath); i++)
     {
       if (tmpfullpath[i] == '/')
	 lastslashpos = i;
     }

  if (lastslashpos == -1)
    {
      strcpy(path, "./"); strcpy(file, tmpfullpath);
      return true;
    }

  strcpy(path, ""); strcpy(file, "");
  for (int i = 0; i <= lastslashpos; i++)
    {
      stemp[0] = tmpfullpath[i]; stemp[1] = 0;
      strcat(path, stemp);
    }
  strcpy(file, tmpfullpath + lastslashpos + 1);
  return true;

}

#define PAT_NOTMATCH    0
#define PAT_MATCH       1
#define PAT_MATCHALL    2

int patmatch(const char *s, const char *pat)
{
  unsigned int slen = strlen(s);
  unsigned int patlen = strlen(pat);

  if (slen < patlen)
    return PAT_NOTMATCH;
  else if (slen == patlen)
    {
      for (int i = 0; i < (signed)patlen; i++)
	{
	  if (s[i] != pat[i])
	    return PAT_NOTMATCH;
	}
      return PAT_MATCHALL;
    }
  else
    {
      for (int i = 0; i < (signed)patlen; i++)
	{
	  if (s[i] != pat[i])
	    return PAT_NOTMATCH;
	}
      return PAT_MATCH;
    }
}


bool completelist(char *dirname, char *diff)
{
  DIR *dp;
  struct dirent *entry;
  struct stat statbuff;
  char path[1024], file[1024], tmppath[1024], extmatch[1024], stemp[1024];
  std::string str1;
  int matchstatus;
  bool flag;
  int i;

  if(!separatepath(dirname, path, file))
    return false;

  if ((dp = opendir(path)) == NULL)
    return false;

  stringarray *sa = new stringarray;
  while ((entry = readdir(dp)) != NULL)
    {
      matchstatus = patmatch(entry->d_name, file);
      if (matchstatus == PAT_MATCH)
	{
	  strcpy(tmppath, path);
	  strcat(tmppath, entry->d_name);
	  stat(tmppath, &statbuff);
	  str1 = std::string(entry->d_name + strlen(file));
	  if(S_ISDIR(statbuff.st_mode))
	    str1 += std::string("/");
	  sa->add(new std::string(str1));
	}
      else if (matchstatus == PAT_MATCHALL)
	{
	  strcpy(tmppath, path);
	  strcat(tmppath, entry->d_name);
	  stat(tmppath, &statbuff);
	  str1 = std::string(entry->d_name + strlen(file));
	  if(S_ISDIR(statbuff.st_mode))
	    str1 += std::string("/");
	  strcpy(diff, str1.c_str());
	  delete sa;
	  closedir (dp);
	  return true;
	}
    }

  // in case of partial match
  if (sa->getarraylength() == 0)
    {
      delete sa;
      closedir (dp);
      return false;
    }
  else if (sa->getarraylength() == 1)
    {
      strcpy(diff, (*sa)[0].c_str());

      delete sa;
      closedir (dp);
      return true;
    }
  else
    {
      flag = false;
      for (i = 0; i < (signed)strlen((*sa)[0].c_str()); i++)
	{
	  for(int j = 1; j < sa->getarraylength(); j++)
	    {
	      if ((*sa)[0].c_str()[i] != (*sa)[j].c_str()[i])
		{
		  flag = true;
		  break;
		}
	    }
	  if (flag)
	    break;
	}
      strcpy(extmatch, "");
      for (int j = 0; j < i; j++)
	  extmatch[j] = (*sa)[0].c_str()[j];
      extmatch[i] = 0;
      strcpy(diff, extmatch);

      message("\n");
      for(int j = 0; j < sa->getarraylength(); j++)
	{
	  strcpy(stemp, file);
	  strcat(stemp, (*sa)[j].c_str());
	  strcat(stemp, " ");
	  message(stemp);
	}
      message("\n");
      delete sa;
      closedir (dp);
      return true;
    }
}

bool completion(char *input, char *diff)
{
  return completelist(input, diff);
}


char *line = (char *)NULL;
char *rline(char *prompt)
{
  if (line)
    {
      free(line);
      line = (char *)NULL;
    }
  line = readline(prompt);
  if (line)
    add_history(line);
  return line;
}

