/*----------------------------------------------------------------------------
                       conttoken.cc (control script flow)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <unistd.h>
#include  <math.h>
#include  <stdlib.h>
#include  <string>
#include <ctype.h>
#include <string.h>

#include "script.h"
#include "system.h"
#include "tokenbuff.h"
#include "graph.h"
#include "buffarray.h"
#include "cui.h"
#include "browse.h"
#include "objectlist.h"
#include "pipecall.h"
#include "topazvalues.h"
#include "convtxt.h"

extern void message(const char *msg);
extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;
extern bool fileexist(char *file);
extern char confpath[1000];
extern char tmppath[1000];
extern char pid[100];
extern bool isabort();
extern void resetabort();
extern char **envp;


int script::controltoken()
{
  int i, itemp, iOparenfound, iCparenfound, iObracefound, iCbracefound, iblocklast, if_or_ifelse_bracestatus = 0\
    ,commacount, iOcontbracefound, iCcontbracefound, iOUTblockstart;
  int commaposition[5];
  std::string conditionvalue;
  tokenbuff* curtoken;
  tokenbuff basetoken, commatoken, conditionvariant, setoperand, oparentoken, cparentoken;
  std::string str1, str2, str3;
  int length;
  int tempparenseslevel;
  int exectokenflag;
  int lin;

  //initialize

  basetoken.priority           = -100;
  basetoken.status             = 0;
  basetoken.value->add("0");
  *basetoken.label             = std::string("");
  basetoken.strstatus          = 0;
  basetoken.bracestatus        = 0;
  basetoken.bracenestlevel     = 0;
  basetoken.parentheseslevel   = 0;
  basetoken.jumpindex          = 0;
  basetoken.lineno             = 0;
  basetoken.count              = -1;

  stack->push1(&basetoken);

  commatoken.priority = PRI_SEMICOLON;
  commatoken.status   = SEMICOLON;

  conditionvariant.priority = PRI_OPERAND;
  conditionvariant.status = VARIANT;
  *conditionvariant.label = std::string("$reserved_by_system_for_evaluation");
  conditionvariant.value->add("");

  setoperand.priority = PRI_SET;
  setoperand.status = SET;
  *setoperand.label = std::string("");
  setoperand.value->add("");

  oparentoken.priority = PRI_OPAREN;
  oparentoken.status = OPAREN;
  *oparentoken.label = std::string("");
  oparentoken.value->add("");
  oparentoken.bracenestlevel = BRC_OPENPAREN;

  cparentoken.priority = PRI_CPAREN;
  cparentoken.status = CPAREN;
  *cparentoken.label = std::string("");
  cparentoken.value->add("");
  cparentoken.bracenestlevel = BRC_CLOSEDPAREN;


  resetabort();
  for (i = 0; i < buff->getarraylength(); i++)
    {
      if (isabort())
	{
	  message("User interruption. Macroscript aborted.\n");
	  return false;
	}

      curtoken = &(*buff)[i];
      lin = curtoken->lineno;
      
      //  '}'
      if (curtoken->status == CPAREN && curtoken->jumpindex != -1)
	{
	  i = curtoken->jumpindex;
	  curtoken = &(*buff)[i];
	}

      if (curtoken->status == LAST || curtoken->status == NEXT)
	{
	  //search start point of out block
	  //last is not in the any blocks
	  if (curtoken->parentheseslevel == 0)
	    return errormessage(lin, "invalid arguments", "next / last");
	  
	  //"last" is not a last token
	  if (i < buff->getarraylength() - 1)
	    {
	      // next token is in the while/until or for loop
	      if ((*buff)[i + 1].bracestatus == BRC_WHILE 
		  || (*buff)[i + 1].bracestatus == BRC_FOR 
		  || (*buff)[i + 1].bracestatus == BRC_FOREACH 
		  )
		{
		  iOUTblockstart = i + 1;
		}
	      else
		{
		  iOUTblockstart = -1;
		  tempparenseslevel = curtoken->parentheseslevel;
		  for (itemp = i + 1; itemp < buff->getarraylength(); itemp++)
		    {
		      if ((*buff)[itemp].parentheseslevel <= tempparenseslevel)
			{
			  if ((*buff)[itemp].bracestatus == BRC_WHILE 
			      || (*buff)[itemp].bracestatus == BRC_FOR 
			      || (*buff)[itemp].bracestatus == BRC_FOREACH 
			      )
			    {
			      iOUTblockstart = itemp;
			      break;
			    }
			  else
			    tempparenseslevel = (*buff)[itemp].parentheseslevel;
			}
		    }
		}

	      //start point is found
	      if (iOUTblockstart != -1)
		{
		  iCbracefound = -1;
		  //search the out position
		  for (itemp = iOUTblockstart; itemp < buff->getarraylength(); itemp++)
		    {
		      if ((*buff)[itemp].bracestatus == (*buff)[iOUTblockstart].bracestatus
			  && (*buff)[itemp].status == CPAREN
			  && (*buff)[itemp].parentheseslevel == (*buff)[iOUTblockstart].parentheseslevel
			  )
			{
			      iCbracefound = itemp;
			      break;
			}
		    }
		  
		  //closed brace found
		  if (iCbracefound != -1)
		    {
		      if (curtoken->status == LAST)
			{
			  
			  //search "continue"
			  //"continue" may exit
			  if((iCbracefound + 2) <= buff->getarraylength() - 1)
			    {
			      //"continue" exist
			      if ((*buff)[iCbracefound + 1].status == CONTINUE 
				  && (*buff)[iCbracefound + 2].status == OPAREN
				  && (*buff)[iOUTblockstart].bracestatus == BRC_WHILE
				  )
				{
				  iOcontbracefound = iCbracefound + 2;
				  //search the position of closed parensis
				  iCcontbracefound = -1;
				  for (itemp = iOcontbracefound; itemp < buff->getarraylength(); itemp++)
				    {
				      if ((*buff)[itemp].status == CPAREN \
					  && (*buff)[itemp].bracestatus == BRC_CONTINUE\
					  && (*buff)[itemp].parentheseslevel == (*buff)[iOcontbracefound].parentheseslevel)
					{
					  iCcontbracefound = itemp;
					  break;
					}
				    }
				  //closed brace of continue loop is found
				  if (iCcontbracefound != -1)
				      i = iCcontbracefound;
				  else
				    return errormessage(lin, "cannot find brace close for 'continue'", "next / last");
				}
			      //"continue" does not exist
			      else
				  i = iCbracefound;
			    }
			  //"continue" cannot exit
			  else
			      i = iCbracefound;
			}
		      else
			  i = iCbracefound;
		    }
		  else
		    return errormessage(lin, "cannot find brace close", "next / last");
		  
		}
	      else
		return errormessage(lin, "cannot find brace open", "next / last");
	    }
	  else
	    return errormessage(lin, "last token is not allowed", "next / last");
	}
      //   if,   elseif
      else if (curtoken->status == IF || curtoken->status == ELSEIF)            
	{
	  
	  if (curtoken->status == IF)
	    if_or_ifelse_bracestatus = BRC_IF;
	    else if (curtoken->status == ELSEIF)
	      if_or_ifelse_bracestatus = BRC_ELSEIF;
	  
	  //in case of 'elseif'
	  if (curtoken->status == ELSEIF)
	    {
	      if (i <= 0)
		return errormessage(lin, "invalid position", "elsif");
	      else
		{
		  // closed parentheses of 'if' must be exist just before 'elseif'
		  if (!((*buff)[i - 1].status == CPAREN \
			&& ((*buff)[i - 1].bracestatus == BRC_IF\
			    || (*buff)[i - 1].bracestatus == BRC_ELSEIF)))
		    return errormessage(lin, "cannot find brace close just before elsif", "elsif");
		}
	    }
	  
	  //check the nextoken to be open parentheses
	  //closed parentheses must be in buffer
	  if((i + 2) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot find parenthesis open / close", "if / elsif");
	  //open parentheses must be followed by 'if' or 'elseif'
	  if ((*buff)[i + 1].status == OPAREN)
	    {
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = i + 2; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == if_or_ifelse_bracestatus\
		      && (*buff)[itemp].parentheseslevel == (*buff)[i + 1].parentheseslevel)
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  //expression is executed
		  conditionvariant.lineno = i;
		  exectoken(&conditionvariant);
		  setoperand.lineno = i;
		  exectoken(&setoperand);
		  oparentoken.lineno = i;
		  exectoken(&oparentoken);
		  for (itemp = i + 2; itemp < iCparenfound; itemp++)
		    exectoken(&(*buff)[itemp]);
		  cparentoken.lineno = i;
		  exectoken(&cparentoken);
		  commatoken.lineno = i;
		  exectoken(&commatoken);
		  str3 = std::string("$reserved_by_system_for_evaluation");
		  vbuff->GetScalar(&str3, &conditionvalue);
		  
		  //if results true
		  if (atoi(conditionvalue.c_str()) != 0)
		    {
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot find brace open / close", "if / elsif");
		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == if_or_ifelse_bracestatus)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == if_or_ifelse_bracestatus\
				  && (*buff)[itemp].parentheseslevel\
				  == (*buff)[iObracefound].parentheseslevel)
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
			      // goto end of if block
			      iblocklast = -1;
			      for (itemp = iCbracefound; itemp <= buff->getarraylength() - 1; itemp++)
				{
				  if ((*buff)[itemp].status == CPAREN 
				      && ((*buff)[itemp].bracestatus == BRC_IF
					  || (*buff)[itemp].bracestatus == BRC_ELSEIF
					  || (*buff)[itemp].bracestatus == BRC_ELSE)
				      && (*buff)[itemp].parentheseslevel == (*buff)[iCbracefound].parentheseslevel
				      && (*buff)[itemp].bracenestlevel == BRC_CLOSEDBRACE 
				      && (
					  (
					   (itemp < buff->getarraylength() - 1) 
					  && ((*buff)[itemp + 1].status != ELSEIF 
					      && (*buff)[itemp + 1].status != ELSE)
					   )
					  || (itemp == buff->getarraylength() - 1)
					  )
				      )

				  {
				    iblocklast = itemp;
				    break;
				  }
				}
			      
			      //last block exist
			      if (iblocklast != -1)
				{
				  (*buff)[iCbracefound].jumpindex = iblocklast; 
				  i = iObracefound;
				}
			      else
				return errormessage(lin, "cannot find the end of 'if' block", "if / elsif");
			    }
			  else
			    return errormessage(lin, "cannot find brace close", "if / elsif");
			  
			}
		      else
			return errormessage(lin, "cannot find brace open", "if / elsif");
		    }
		  //if results false
		  else
		    {
		      //search closed brace of if block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == if_or_ifelse_bracestatus\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}
		      
		      
		      if (iCbracefound != -1)
			i = iCbracefound;
		      else
			return errormessage(lin, "cannot find the end of 'if' block", "if / elsif");
		    }			      
		}
	      else
		return errormessage(lin, "cannot find parenthesis close", "if / elsif");
	    }
	  else
	    return errormessage(lin, "cannot find parenthesis open", "if / elsif");
	}      
      //   else
      else if (curtoken->status == ELSE)
	{
	}
      else if (curtoken->status == WHILE || curtoken->status == UNTIL)
	{
	  //check the nextoken to be open parentheses
	  //closed parentheses must be in buffer
	  if((i + 2) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot find parenthesis close", "while / until");
	  //open parentheses must be followed by 'while'
	  if ((*buff)[i + 1].status == OPAREN)
	    {
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = i + 2; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == BRC_WHILE\
		      && (*buff)[itemp].parentheseslevel == (*buff)[i + 1].parentheseslevel)
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  //expression is executed
		  conditionvariant.lineno = i;
		  exectoken(&conditionvariant);
		  setoperand.lineno = i;
		  exectoken(&setoperand);
		  oparentoken.lineno = i;
		  exectoken(&oparentoken);
		  for (itemp = i + 2; itemp < iCparenfound; itemp++)
		    exectoken(&(*buff)[itemp]);
		  cparentoken.lineno = i;
		  exectoken(&cparentoken);
		  commatoken.lineno = i;
		  exectoken(&commatoken);
		  str3 = std::string("$reserved_by_system_for_evaluation");
		  vbuff->GetScalar(&str3, &conditionvalue);

		  itemp = atoi(conditionvalue.c_str());
		  if (curtoken->status == UNTIL)
		    {
		      if (itemp)
			itemp = 0;
		      else
			itemp = 1;
		    }
		    
		  //while results true / until results false
		  if (itemp != 0)
		    {
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot find brace open / close", "while / until");
		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == BRC_WHILE)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == BRC_WHILE\
				  && (*buff)[itemp].parentheseslevel\
				  == (*buff)[iObracefound].parentheseslevel)
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
			      //search "continue"
			      //"continue" may exit
			      if((iCbracefound + 2) <= buff->getarraylength() - 1)
				{
				  //"continue" exist
				  if ((*buff)[iCbracefound + 1].status == CONTINUE 
				      && (*buff)[iCbracefound + 2].status == OPAREN)
				    {
				      iOcontbracefound = iCbracefound + 2;
				      //search the position of closed parensis
				      iCcontbracefound = -1;
				      for (itemp = iOcontbracefound; itemp < buff->getarraylength(); itemp++)
					{
					  if ((*buff)[itemp].status == CPAREN \
					      && (*buff)[itemp].bracestatus == BRC_CONTINUE\
					      && (*buff)[itemp].parentheseslevel == (*buff)[iOcontbracefound].parentheseslevel)
					    {
					      iCcontbracefound = itemp;
					      break;
					    }
					}
				      //closed brace of continue loop is found
				      if (iCcontbracefound != -1)
					{
					  (*buff)[iCcontbracefound].jumpindex = i; 
					  i = iObracefound;
					}
				      else
					return errormessage(lin, "cannot find brace close for 'continue' loop", "while / until");
				    }
				  //"continue" does not exist
				  else
				    {
				      (*buff)[iCbracefound].jumpindex = i; 
				      i = iObracefound;
				    }
				}
			      //"continue" cannot exit
			      else
				{
				  (*buff)[iCbracefound].jumpindex = i; 
				  i = iObracefound;
				}
			    }
			  else
			    return errormessage(lin, "cannot find brace close", "while / until");
			}
		      else
			return errormessage(lin, "cannot find brace open", "while / until");
		    }
		  //while results false / until results true
		  else
		    {
		      //search closed brace of if block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == BRC_WHILE\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}

		      //closed brace found
		      if (iCbracefound != -1)
			{
			  //search "continue"
			  //"continue" may exit
			  if((iCbracefound + 2) <= buff->getarraylength() - 1)
			    {
			      //"continue" exist
			      if ((*buff)[iCbracefound + 1].status == CONTINUE 
				  && (*buff)[iCbracefound + 2].status == OPAREN)
				{
				  iOcontbracefound = iCbracefound + 2;
				  //search the position of closed parensis
				  iCcontbracefound = -1;
				  for (itemp = iOcontbracefound; itemp < buff->getarraylength(); itemp++)
				    {
				      if ((*buff)[itemp].status == CPAREN \
					  && (*buff)[itemp].bracestatus == BRC_CONTINUE\
					  && (*buff)[itemp].parentheseslevel == (*buff)[iOcontbracefound].parentheseslevel)
					{
					  iCcontbracefound = itemp;
					  break;
					}
				    }
				  
				  //closed brace of continue loop is found
				  if (iCcontbracefound != -1)
				      i = iCcontbracefound;
				  else
				    return errormessage(lin, "cannot find brace close for 'continue' loop", "while / until");
				}
			      //"continue" does not exist
			      else
				i = iCbracefound;
			    }
			  //"continue" cannot exit
			  else
			      i = iCbracefound;
			}
		      else
			return errormessage(lin, "cannot find brace close", "while / until");
		    }			      
		}
	      else 
		return errormessage(lin, "cannot find parenthesis close", "while / until");
	    }
	  else
	    return errormessage(lin, "cannot find parenthesis open", "while / until");
	}
      else if (curtoken->status == FOR)
	{
	  //closed parentheses must be in buffer
	  if((i + 3) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot find parenthesis close", "for");
	  //nextfor and open parentheses must be followed by 'for'
	  if ((*buff)[i + 1].status == NEXTFOR\
	      && (*buff)[i + 2].status == OPAREN)
	    {
	      iOparenfound = i + 2;
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = iOparenfound + 1; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == BRC_FOR\
		      && ((*buff)[itemp].parentheseslevel  == (*buff)[iOparenfound].parentheseslevel))
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  //condition expression must contain three 'comma's
		  commacount = 0;
		  for (itemp = iOparenfound; itemp < iCparenfound; itemp++) 
		    {
		      if ((*buff)[itemp].status == SEMICOLON)
			commacount++;
		    }

		  if (commacount != 2)
		    return errormessage(lin, "invalid 'for' condition", "for");

		  //search comma position
		  commacount = 0;
		  for (itemp = iOparenfound; itemp < iCparenfound; itemp++) 
		  {
		    if ((*buff)[itemp].status == SEMICOLON)
			commaposition[commacount++] = itemp;
		  }

		  //intialization (ex. i = 1;)
 		  for (itemp = iOparenfound + 1; itemp <=  commaposition[0]; itemp++)
		    exectoken(&(*buff)[itemp]);

		  //condition  (ex. i < 100)
		  conditionvariant.lineno = i;
		  exectoken(&conditionvariant);
		  setoperand.lineno = i;
		  exectoken(&setoperand);

 		  for (itemp = commaposition[0] + 1; itemp <=  commaposition[1]; itemp++)
		    exectoken(&(*buff)[itemp]);
		  str3 = std::string("$reserved_by_system_for_evaluation");
		  vbuff->GetScalar(&str3, &conditionvalue);

		  //for results true
		  if (atoi(conditionvalue.c_str()) != 0)
		    {
		      
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot find brace open / close", "for");

		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == BRC_FOR)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == BRC_FOR\
				  && ((*buff)[itemp].parentheseslevel == (*buff)[iObracefound].parentheseslevel))
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
  			      (*buff)[iCbracefound].jumpindex = i + 1; 
			      i = iObracefound;
			    }
			  else
			    return errormessage(lin, "cannot find brace close", "for");
			}
		      else
			return errormessage(lin, "cannot find brace open", "for");
		    }
		  //for results false
		  else
		    {
		      //search closed brace of for block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == BRC_FOR\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}

		      if (iCbracefound != -1)
			i = iCbracefound;
		      else
			return errormessage(lin, "cannot find brace close", "for");
		    }			      
		}
	      else 
		return errormessage(lin, "cannot find parenthesis close", "for");
	    }
	  else
	    return errormessage(lin, "cannot find parenthesis open", "for");
	}
      else if (curtoken->status == NEXTFOR)
	{
	  //closed parentheses must be in buffer
	  if((i + 2) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot find parenthesis close", "for");
	  //open parentheses must be followed by 'nextfor'
	  if ((*buff)[i + 1].status == OPAREN)
	    {
	      iOparenfound = i + 1;
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = i + 2; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == BRC_FOR\
		      && (*buff)[itemp].parentheseslevel == (*buff)[i + 1].parentheseslevel)
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  //condition expression must contain three 'comma's
		  commacount = 0;
		  for (itemp = iOparenfound; itemp <iCparenfound; itemp++) 
		  {
		    if ((*buff)[itemp].status == SEMICOLON)
		      commacount++;
		  }

		  if (commacount != 2)
		    return errormessage(lin, "invalid 'for' condition", "for");

		  //search comma position
		  commacount = 0;
		  for (itemp = iOparenfound; itemp <iCparenfound; itemp++) 
		  {
		    if ((*buff)[itemp].status == SEMICOLON)
		      commaposition[commacount++] = itemp;
		  }
		  
		  //incremet (ex. i = i + 1;)
 		  for (itemp = commaposition[1] + 1; itemp <  iCparenfound; itemp++)
		    exectoken(&(*buff)[itemp]);
		  commatoken.lineno = i;
		  exectoken(&commatoken);

		  //condition  (ex. i < 100)
		  conditionvariant.lineno = i;
		  exectoken(&conditionvariant);
		  setoperand.lineno = i;
		  exectoken(&setoperand);
 		  for (itemp = commaposition[0] + 1; itemp <=  commaposition[1]; itemp++)
		    exectoken(&(*buff)[itemp]);
		  str3 = std::string("$reserved_by_system_for_evaluation");
		  vbuff->GetScalar(&str3, &conditionvalue);

		  //for results true
		  if (atoi(conditionvalue.c_str()) != 0)
		    {
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot fine brace open / close", "for");

		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == BRC_FOR)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == BRC_FOR\
				  && (*buff)[itemp].parentheseslevel\
				  == (*buff)[iObracefound].parentheseslevel)
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
			      (*buff)[iCbracefound].jumpindex = i; 
			      i = iObracefound;
			    }
			  else
			    return errormessage(lin, "cannot fine brace close", "for");
			}
		      else
			return errormessage(lin, "cannot fine brace open", "for");
		    }
		  //for results false
		  else
		    {
		      //search closed brace of if block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == BRC_FOR\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}

		      if (iCbracefound != -1)
			i = iCbracefound;
		      else
			return errormessage(lin, "cannot fine brace close", "for");
		    }			      
		}
	      else 
		return errormessage(lin, "cannot fine parenthesis close", "for");
	    }
	  else
	    return errormessage(lin, "cannot fine parenthesis open", "for");
	}
      else if (curtoken->status == FOREACH)
	{
	  //closed parentheses must be in buffer
	  if((i + 4) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot fine parenthesis close", "foreach");

	  //"nextforeach" and "variant" and "open parentheses" must be followed by 'foreach'
	  if ((*buff)[i + 1].status == NEXTFOREACH
	      && (*buff)[i + 2].status == VARIANT
	      && (*buff)[i + 3].status == OPAREN)
	    {
	      iOparenfound = i + 3;
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = iOparenfound + 1; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == BRC_FOREACH\
		      && ((*buff)[itemp].parentheseslevel  == (*buff)[iOparenfound].parentheseslevel))
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  //exec between open parentheses and closed parenses
		  v->setbuff(0);
 		  for (itemp = iOparenfound; itemp <= iCparenfound; itemp++)
		    exectoken(&(*buff)[itemp]);
		  if (calc() == false)
		    return errormessage(lin, "invalid array or list", "foreach");

		  if ((*v)[v->getarraylength() - 1].status == ARRAY)
		    {
		      (*buff)[i + 1].count = 0;
		      (*buff)[i + 1].value->setbuff(0);
		      for (int j = 0; j < vbuff->ArrayLength((*v)[v->getarraylength() - 1].label); j++)
			{
			  vbuff->Get((*v)[v->getarraylength() - 1].label, j, &str1);
			  (*buff)[i + 1].value->add(str1.c_str());
			}
		      vbuff->Get((*v)[v->getarraylength() - 1].label, (*buff)[i + 1].count, &str1);
		      vbuff->SetScalar((*buff)[i + 2].label, &str1);
		      length = vbuff->ArrayLength((*v)[v->getarraylength() - 1].label);
		      v->pop();
		    }
		  else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
		    {
		      (*buff)[i + 1].count = 0;
		      (*buff)[i + 1].value->setbuff(0);
		      for (int j = 0; j < (*v)[v->getarraylength() - 1].value->getarraylength(); j++)
			{
			  str1 = (*v)[v->getarraylength() - 1].value->getdata(j);
			  (*buff)[i + 1].value->add(str1.c_str());
			}
		      str1 = (*v)[v->getarraylength() - 1].value->getdata((*buff)[i + 1].count);
		      vbuff->SetScalar((*buff)[i + 2].label, &str1);
		      length = (*v)[v->getarraylength() - 1].value->getarraylength();
		      v->pop();
		    }
		  else
		    return false;

		  //count is less than array length
		  if ((*buff)[i].count < length)
		    {
		      
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot fine brace open / close", "foreach");

		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == BRC_FOREACH)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == BRC_FOREACH\
				  && ((*buff)[itemp].parentheseslevel == (*buff)[iObracefound].parentheseslevel))
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
  			      (*buff)[iCbracefound].jumpindex = i + 1; 
			      i = iObracefound;
			    }
			  else
			    return errormessage(lin, "cannot fine brace close", "foreach");
			}
		      else
			return errormessage(lin, "cannot fine brace open", "foreach");
		    }
		  //count is greater than array length
		  else
		    {
		      //search closed brace of for block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == BRC_FOREACH\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}

		      if (iCbracefound != -1)
			i = iCbracefound;
		      else
			return errormessage(lin, "cannot fine brace close", "foreach");
		    }			      
		}
	      else 
		return errormessage(lin, "cannot fine parenthesis close", "foreach");
	    }
	  else
	    return errormessage(lin, "cannot fine parenthesis open", "foreach");
	}
      else if (curtoken->status == NEXTFOREACH)
	{

	  //closed parentheses must be in buffer
	  if((i + 3) > buff->getarraylength() - 1)
	    return errormessage(lin, "cannot fine parenthesis close", "foreach");

	  //"nextfor" and "variant" and "open parentheses" must be followed by 'for'
	  if ((*buff)[i + 1].status == VARIANT
	      && (*buff)[i + 2].status == OPAREN)
	    {
	      iOparenfound = i + 2;
	      //search the position of closed parensis
	      iCparenfound = -1;
	      for (itemp = iOparenfound + 1; itemp < buff->getarraylength(); itemp++)
		{
		  if ((*buff)[itemp].status == CPAREN \
		      && (*buff)[itemp].bracestatus == BRC_FOREACH\
		      && ((*buff)[itemp].parentheseslevel  == (*buff)[iOparenfound].parentheseslevel))
		    {
		      iCparenfound = itemp;
		      break;
		    }
		}

	      //closed parentheses is found
	      if (iCparenfound != -1)
		{
		  (*buff)[i].count++;
		  length = (*buff)[i].value->getarraylength();

		  //count is less than array length
		  if ((*buff)[i].count < length)
		    {
		      str1 = (*buff)[i].value->getdata((*buff)[i].count);
		      vbuff->SetScalar((*buff)[i + 1].label, &str1);
		      
		      //confirm open brace
		      iObracefound = iCparenfound + 1;
		      
		      //open brace must exist
		      if (iObracefound > buff->getarraylength() - 2)
			return errormessage(lin, "cannot find brace open / close", "foreach");

		      if ((*buff)[iObracefound].status == OPAREN\
			  && (*buff)[iObracefound].bracestatus == BRC_FOREACH)
			{
			  //search the position of closed brace
			  iCbracefound = -1;
			  for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			    {
			      if ((*buff)[itemp].status == CPAREN \
				  && (*buff)[itemp].bracestatus == BRC_FOREACH\
				  && ((*buff)[itemp].parentheseslevel == (*buff)[iObracefound].parentheseslevel))
				{
				  iCbracefound = itemp;
				  break;
				}
			    }
			  
			  //closed brace exist!!
			  if (iCbracefound != -1)
			    {
  			      (*buff)[iCbracefound].jumpindex = i; 
			      i = iObracefound;
			    }
			  else
			    return errormessage(lin, "cannot find brace close", "foreach");
			}
		      else
			return errormessage(lin, "cannot find brace open", "foreach");
		    }
		  //count is greater than array length
		  else
		    {
		      //search closed brace of for block
		      iObracefound = iCparenfound + 1;
		      iCbracefound = -1;
		      for (itemp = iObracefound; itemp <= buff->getarraylength() - 1; itemp++)
			{
			  if ((*buff)[itemp].status == CPAREN \
			      && (*buff)[itemp].bracestatus == BRC_FOREACH\
			      && (*buff)[itemp].parentheseslevel\
			      == (*buff)[iObracefound].parentheseslevel)
			    {
			      iCbracefound = itemp;
			      break;
			    }
			}

		      if (iCbracefound != -1)
			i = iCbracefound;
		      else
			return errormessage(lin, "cannot find brace close", "foreach");
		    }			      
		}
	      else 
		return errormessage(lin, "cannot find parenthesis close", "foreach");
	    }
	  else
	    return errormessage(lin, "cannot find parenthesis open", "foreach");
	}
      else
	{
	  exectokenflag = exectoken(curtoken);
	  if (exectokenflag != true)
	    return exectokenflag; 
	}
    }

  return true;
}



int script::exectoken(tokenbuff *token)
{
  int status;

  while (token->priority <= (*stack)[stack->getarraylength() - 1].priority \
	 && (*stack)[stack->getarraylength() - 1].status != OPAREN)
    {

      polish->push1(&(*stack)[stack->getarraylength() - 1]); stack->pop();
    }

  if (token->status != CPAREN)
    {
      if (token->status != SEMICOLON)
	stack->push1(token);
      else
	{
          status = calc();
	  if (status == false)
	      return false;
	  else if (status == exitscriptstatus)
	      return exitscriptstatus;
	  
	  v->setbuff(0);
	  polish->setbuff(0);
	}
    }
  else if (token->status == CPAREN && (*stack)[stack->getarraylength() - 1].status == OPAREN)
    stack->pop();
  
  return true;
}


int script::execscript(std::string *inputstr, std::string *out, buffarray* arg)
{
  std::string returnvalue, str;
  
  initparse(arg, 0);

  if (parsetoken(inputstr) == false)
      return false;

  switch (controltoken())
    {
    case exitscriptstatus:
      if (out == 0)
	return exitscriptstatus;
      else
	{
	  str = std::string("$_EXIT");
	  if (vbuff->GetScalar(&str, &returnvalue))
	    {
	      *out = returnvalue;
	      return exitscriptstatus;
	    }
	  else
	    return false;
	}
    case true:
      return true;
    case false:
      return false;
    }
  return true;
}

bool getmacroname(std::string *file, std::string *out)
{
  std::string str;
  if (getenv("TOPAZINIDIR") == NULL)
    return false;

  if (fileexist((char *)file->c_str()))
    {
      *out = *file;
      return true;
    }
  else
    {
      str = std::string(getenv("TOPAZINIDIR")) + std::string("/") + *file;
      if (fileexist((char *)str.c_str()))
	{
	  *out = str;
	  return true;
	}
      else
	{
	  if (getenv("TOPAZMACRODIR") == NULL)
	    str = std::string("");
	  else
	    str = std::string(getenv("TOPAZMACRODIR")) + std::string("/") + *file;
	  if (fileexist((char *)str.c_str()))
	    {
	      *out = str;
	      return true;
	    }
	  else
	    {
	      if (getenv("TOPAZDIR") == NULL)
		return false;
	      str = std::string(getenv("TOPAZDIR")) + std::string("/_topaz/") + *file;
	      if (fileexist((char *)str.c_str()))
		{
		  *out = str;
		  return true;
		}
	      else
		return false;
	    }
	}
    }
  return true;
}

int script::execscriptfile(std::string *inputfile, std::string *out, buffarray* arg)
{
  FILE *fp;
  char stemp[1000];
  std::string tempstr;
  std::string returnvalue,str;

  if (!getmacroname(inputfile, &tempstr))
    return NOTEXISTSCRIPT;

  *scriptname = *inputfile;
  if ((fp = fopen((char *)tempstr.c_str(), "r")) != 0)
    {
      initparse(arg, &tempstr);
      while (topazfgets(stemp, sizeof(stemp), fp) != NULL)
	{
	  tempstr = std::string(stemp);
	  if (!parsetoken(&tempstr))
	    return false;
	  lineno++;
	}
      fclose(fp);

      switch (controltoken())
	{
	case exitscriptstatus:
	  if (out == 0)
	    return exitscriptstatus;
	  else
	    {
	      str = std::string("$_EXIT");
	      if (vbuff->GetScalar(&str, &returnvalue))
		{
		  *out = returnvalue;
		  return exitscriptstatus;
		}
	      else
		return false;
	    }
	case true:
	  return true;
	case false:
	  return false;
	}
    }
  else
    return NOTEXISTSCRIPT;
  return true;
}


bool sepenvvalue(char *in, std::string *label, std::string *value)
{
  char *p = in;
  std::string str;
  char stemp[10];

  str = std::string("");
  while (*p != 0)
    {
      if (*p != '=')
	{
	  stemp[0] = *p; stemp[1] = 0;
	  str += std::string(stemp);
	}
      else
	{
	  *label = str;
	  *value = std::string(p + 1);
	  return true;
	}
      p++;
    }
  return false;
}

void script::getenvstring()
{
  std::string label, value, lbl;

  for (int i = 0; envp[i] != NULL; i++)
    {
      if (sepenvvalue(envp[i], &label, &value))
	{
	  lbl = std::string("$") + label;
	  vbuff->SetSystemScalar(&lbl, &value);  
	}
    }
}



void script::initparse(buffarray* arg, std::string *scrfile)
{
  int argno;
  char stemp[1000];
  std::string str1, str2;

  // initialize status
  strstatusbuffpoint = 0;
  bracestatusbuffpoint = 0;
  parentheseslevel = 0;
  lineno = 1;

  //special variables
  // separation of print 
  str1 = std::string("$,"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);         // print separator
  str1 = std::string("$\""); str2 = std::string(" ");
  vbuff->SetScalar(&str1, &str2);       // separator for array in "..."
  str1 = std::string("$/"); str2 = std::string("\n");
  vbuff->SetScalar(&str1, &str2);       // chomp character
  str1 = std::string("$_EXIT"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);     // exit  value
  str1 = std::string("$&"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);         // match pattern
  str1 = std::string("$`"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);         // match before
  str1 = std::string("$'"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);         // match after
  str1 = std::string("$|"); str2 = std::string("0");
  vbuff->SetScalar(&str1, &str2);         // print flush after "print"
  str1 = std::string("$_RET"); str2 = std::string("");
  vbuff->SetScalar(&str1, &str2);      
  str1 = std::string("$_CUR"); str2 = std::string("-2");
  vbuff->SetSystemScalar(&str1, &str2);  
  str1 = std::string("$_SEL"); str2 = std::string("-3");
  vbuff->SetSystemScalar(&str1, &str2);  
  str1 = std::string("$_ALL"); str2 = std::string("-4");
  vbuff->SetSystemScalar(&str1, &str2);  
  str1 = std::string("$_FSL"); str2 = std::string("-5");
  vbuff->SetSystemScalar(&str1, &str2);  
  sprintf(stemp, "%u", (unsigned int)getpid());
  str1 = std::string("$PID"); str2 = std::string(stemp);
  vbuff->SetSystemScalar(&str1, &str2);  
  str1 = std::string("@_RET");
  vbuff->SetArray(&str1, 0);

  // set environ values as readonly value
  getenvstring();

  // set initfile dir
  str1 = std::string("$TOPAZTMP"); str2 = std::string(tmppath);
  vbuff->SetSystemScalar(&str1, &str2);  

  // set numerical
  sprintf(stemp, "%.15g", M_PI);
  str1 = std::string("$PI"); str2 = std::string(stemp);
  vbuff->SetSystemScalar(&str1, &str2);  

  // set scriptfilename
  str1 = std::string("$0"); str2 = std::string("");
  if (scrfile != 0)
      vbuff->SetSystemScalar(&str1, scrfile);  
  else
      vbuff->SetSystemScalar(&str1, &str2);  

  // set argument
  if (arg != 0)
    {
      argno = arg->getarraylength();
      str1 = std::string("@ARGV");
      vbuff->SystemSetArray(&str1, argno);
      for (int i = 0; i < argno; i++)
	vbuff->SystemSet(&str1, i, &(*arg)[i]);
    }
}







