/*----------------------------------------------------------------------------
                           cui.cc (cui interfaces)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 99


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include "cui.h"
#include "script.h"
#include "buffarray.h"
#include "gdi.h"
#include "msgbuff.h"
#include "filepoint.h"
#include "kanji.h"
#include "completion.h"
#include "topazcurses.h"

extern FileHandleArray *fhvuff;


extern msgbuff *msgb;
extern graph *gra;
extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;
extern int menuwinlines;
extern int itemwidth;
extern int itemspace;
extern int maxrow;
extern int leftmargin;
extern int topmargin;
extern int truetitlelength(const char *s);
extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);
extern int topazgetch(WINDOW *win);

extern void showcurrentmenu(CUIMENU *menu, int maxitem, int addoffset);
extern void showcursor(CUIMENU *menu, int maxitem, int itemno, bool flag, int addoffset);
extern int cursormove (CUIMENU *menu, int *itemno, int maxitem, int addoffset);

extern CUIMENU Color[];
extern CUIMENU Yesno[];
extern CUIMENU Frametype[];
extern CUIMENU Tickdir[];
extern CUIMENU Scaling[];
extern CUIMENU MScaling[];
extern CUIMENU Linecap[];
extern CUIMENU Linejoin[];
extern CUIMENU Linestyle[];
extern CUIMENU Fillstyle[];
extern CUIMENU Fillrule[];
extern CUIMENU Mag[];
extern CUIMENU Interpolate[];
extern CUIMENU Markerstyle[];
extern CUIMENU Loaddata[];
extern CUIMENU Commentlineaction[];
extern CUIMENU Datasearchpath[];
extern CUIMENU Partsstyle[];
extern CUIMENU Textstyle[];
extern CUIMENU Partstype[];
extern CUIMENU Fontface[];
extern CUIMENU Jpfontface[];
extern CUIMENU Partsinterpolation[];
extern CUIMENU Labelparity[];
extern CUIMENU LabelCPparity[];
extern CUIMENU Labelformat[];
extern CUIMENU Orderlist[];
extern CUIMENU Papersize[];
extern CUIMENU MLinearfit[];

extern int hotkeyattr;

extern void clearmenuwin();

int smallskip = 20;
int largeskip = 200;

void execitem(CUIMENU *menu, int *selectedkey)
{
  script *scr = new script(gra);
  std::string str;
  str = std::string(menu[*selectedkey].macroscript);
  scr->execscript(&str, 0, 0);
  delete scr;
}

//
//   input window for text, value
//

int edittextwindow(WINDOW *win, std::string *input, std::string *caption, bool editfilename)
{
  if (win == 0)
    return false;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "+-------------------------------------------------------------------------+");
  wmove(win, 3, 2);
  wprintw(win, "|                                                                         |");
  wmove(win, 4, 2);
  wprintw(win, "+-------------------------------------------------------------------------+");
  wmove(win, 5, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 6, 2);
  wprintw(win, "                              KEY BINDINGS                                 ");
  wmove(win, 7, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 8, 2);
  wprintw(win, " Ctrl+f : move forward                  Ctrl+b : move back                 ");
  wmove(win, 9, 2);
  wprintw(win, " Ctrl+a : bring to top                  Ctrl+e : go to end                 ");
  wmove(win, 10, 2);
  wprintw(win, " Ctrl+d : delete                        Ctrl+h : backspace                 ");
  wmove(win, 11, 2);
  wprintw(win, " Ctrl+' ' or Ctrl+q: set mark           Ctrl+w : cut and save to buffer    ");
  wmove(win, 12, 2);
  wprintw(win, " Ctrl+y : paste from buffer             Ctrl+k : cut a line                ");

  if (editfilename)
  {
    wmove(win, 13, 2);
    wprintw(win, " Ctrl+i : complete file/directory name                                   ");
  }
  wrefresh(win);

  return textlineedit(win, input, 3, 3, 73, editfilename);
}

void printtext(WINDOW *win, char *input)
{
  char *q;
  q = input;
  char stemp[10];

  while (*q != 0)
    {
      stemp[0] = *q++; stemp[1] = 0;
      if (stemp[0] == '%')
	strcpy(stemp, "%%");
      wprintw(win, stemp);
    }
}

int linelength(char *str, int offset, int displength)
{
  int i;
  int total = 0;

  for (i = offset; i < (int)strlen(str); i++)
    {
      if (total + charlength(*(str + i)) <= displength)
	total += charlength(*(str + i));
      else
	return total;
      if (maykanji(str[i]))
	  i++;
    }
  return total;
}

int nextpage(char *str, int logicalpoint, int displength)
{
  int pos = logicalpoint - displength / 2;
  
  if (pos < 0)
    return 0;

  if (maykanji(str[pos]))
    {
      if (iskanji1(str + pos))
	return pos;
      else
	return pos - 1;
    }
  else
      return pos;
}

void disptext(WINDOW *win, char *input, int y, int x, int displength, int pos, int offset)
{
  std::string tempstr;
  char s[10], *p;
  int count = 0;

  if (win == 0)
    return;

  for (int i = 0; i < displength; i++)
    {
      wmove(win, y, x + i);
      wprintw(win, " ");
    }
  wrefresh(win);

  tempstr = std::string("");
  wmove(win, y, x);
  wrefresh(win);

  if ((int)strlen((const char*)(input + pos)) >= linelength(input, pos, displength))
    {
      p = input + pos; 
      while (count < linelength(input, pos, displength))
	{
	  if (maykanji(*p))
	    {
	      if (iskanji1(p))
		{
		  if (count + 2 <= linelength(input, pos, displength))
		    {
		      s[0] = *p++; s[1] = *p++; s[2] = 0;
		      tempstr += std::string(s);
		      count += 2;
		    }
		  else
		    break;
		}
	      else
		return;
	    }
	  else
	    {
	      s[0] = *p++; s[1] = 0;
	      tempstr += std::string(s);
	      count++;
	    }
	}

      while (count < displength)
	{
	  s[0] = ' '; s[1] = 0;
	  tempstr += std::string(s);
	  count++;
	}
      
      printtext(win, (char *)tempstr.c_str());
    }
  else
    {
      count = 0;
      while (count < displength)
	{
	  if (count < (int)strlen(input + pos))
	    {
	      s[0] = (input + pos)[count]; s[1] = 0;
	    }
	  else
	    {
	      s[0] = ' '; s[1] = 0;
	    }
	  tempstr += std::string(s);
	  count++;
	}
      printtext(win, (char *)tempstr.c_str());
    }

  // showarrow
  wmove(win, y + 2, x);
  if (offset > 0)
    wprintw(win, "<-");
  else
    wprintw(win, "  ");

  wmove(win, y + 2, x + displength -2);
  if (offset + displength < (int)strlen(input))
    wprintw(win, "->");
  else
    wprintw(win, "  ");

  wrefresh(win);
}


int textlineedit(WINDOW *win, std::string *input, int y, int x, int displength, bool editfilename)
{
  std::string tempstr1, tempstr2, tempstr = *input;
  int key, key1, key2;
  int strend = strlen(tempstr.c_str()) - 1;
  int dispoffset = 0; 
  int logicalpoint = 0;
  int tmplogicalpoint = 0;
  int disppoint = 0;
  int inc;
  char stemp[10];
  int mark = -1;
  static std::string buff;
  int itemp;
  char tmppath[1024];

  if (win == 0)
    return false;

  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
  wmove(win, y, x);
  wrefresh(win);


  while (1)
    {
      wmove(win, y, x + disppoint);
      wrefresh(win);

      key = topazgetch(win);

      switch (key)
	{
	  //mark
	case CTRL(' '):
	case CTRL('q'):
	  mark = logicalpoint;
	  break;
	  //cut
	case CTRL('w'):
	  if (mark != -1)
	    {
	      if (mark > logicalpoint)
		{
		  buff = tempstr.substr(logicalpoint, mark - logicalpoint);
		  tempstr.replace(logicalpoint, mark - logicalpoint, "" , 0);
		  strend = strlen(tempstr.c_str()) - 1;
		}
	      else
		{
		  buff = tempstr.substr(mark, logicalpoint - mark);
		  tempstr.replace(mark, logicalpoint - mark, "", 0);
		  strend = strlen(tempstr.c_str()) - 1;
		  logicalpoint = mark;
		}
	      if (logicalpoint < dispoffset)
		dispoffset = logicalpoint;
	      disppoint = logicalpoint - dispoffset;
	      mark = -1;
	      disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	    }
	  break;
	  //paste
	case CTRL('y'):
	  tempstr1 = tempstr.substr(0, logicalpoint);
	  tempstr2 = tempstr.substr(logicalpoint);
	  tempstr = tempstr1;
	  tempstr += buff + tempstr2;
	  strend = strlen(tempstr.c_str()) - 1;
	  
	  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);

	  mark = -1;
	  break;
	  //completion
	case CTRL('i'):
	  if (editfilename)
	    {
	      char diff[1024];
	      tempstr1 = tempstr.substr(0, logicalpoint);
	      tempstr2 = tempstr.substr(logicalpoint);
	      if (completion((char *)tempstr1.c_str(), diff))
		{
		  tempstr = tempstr1;
		  tempstr += std::string(diff) + tempstr2;
		  strend = strlen(tempstr.c_str()) - 1;

		  inc = strlen(diff);
		  if (logicalpoint + inc <= strend + 1)
		    {
		      itemp = linelength((char *)tempstr.c_str(), dispoffset, displength);
		      if (logicalpoint + inc <= dispoffset + itemp 
			  && 
			  (((logicalpoint + inc) <= strend
			    && disppoint + inc + charlength(tempstr.c_str()[logicalpoint + inc]) <=  displength)
			   || ((logicalpoint + inc) == strend + 1
			       && disppoint + inc  <=  displength)))
			{
			  logicalpoint += inc;
			  disppoint += inc;
			  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
			}
		      else
			{
			  logicalpoint += inc;
			  dispoffset = nextpage((char *)tempstr.c_str(), logicalpoint, displength);
			  disppoint = logicalpoint - dispoffset;
			  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
			}
		    }
		  mark = -1;
		}
	    }
	  break;
	  // left
	case KEY_LEFT:
	case CTRL('b'):
	  if (logicalpoint > 0)
	    {
	      if (logicalpoint > 1)
                inc = charlengthback(tempstr.c_str(), logicalpoint);
	      else
		inc = 1;

	      itemp = linelength((char *)tempstr.c_str(), dispoffset, displength);
	      
	      if (logicalpoint - inc >= dispoffset)
		  {
		    logicalpoint -= inc;
		    disppoint -= inc;
		  }
	      else
		  {
		    dispoffset = nextpage((char *)tempstr.c_str(), logicalpoint, displength);
		    logicalpoint -= inc;
		    disppoint = logicalpoint - dispoffset;
		    disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
		  }
	    }
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  inc = charlength(tempstr.c_str()[logicalpoint]);
	  if (logicalpoint + inc <= strend + 1)
	    {
	      itemp = linelength((char *)tempstr.c_str(), dispoffset, displength);
	      if (logicalpoint + inc <= dispoffset + itemp 
		  && 
		  (((logicalpoint + inc) <= strend
		    && disppoint + inc + charlength(tempstr.c_str()[logicalpoint + inc]) <=  displength)
		   || ((logicalpoint + inc) == strend + 1
		       && disppoint + inc  <=  displength)))
		{
		  logicalpoint += inc;
		  disppoint += inc;
		}
	      else
		{
		  logicalpoint += inc;
		  dispoffset = nextpage((char *)tempstr.c_str(), logicalpoint, displength);
		  disppoint = logicalpoint - dispoffset;
		  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
		}
	    }
	  break;
	case CTRL('a'):
	  logicalpoint = 0;
	  dispoffset = 0;
	  disppoint = logicalpoint - dispoffset;
	  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	  break;
	case CTRL('e'):
	  logicalpoint = strend + 1;
	  dispoffset = nextpage((char *)tempstr.c_str(), strend, displength);
	  disppoint = logicalpoint - dispoffset;
	  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	  break;
	case KEY_DC:
	case CTRL('d'):
	  inc = charlength(tempstr.c_str()[logicalpoint]);
	  tempstr.replace(logicalpoint, inc, "", 0);
	  strend = strlen(tempstr.c_str()) - 1;
	  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	  mark = -1;
	  break;
	case CTRL('k'):
	  if (logicalpoint <= strend)
	    {
	      buff = tempstr.substr(logicalpoint, strlen(tempstr.c_str()) - logicalpoint);

	      tempstr = tempstr.substr(0, logicalpoint);
	      strend = strlen(tempstr.c_str()) - 1;
	      disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	    }
	  mark = -1;
	  break;
	case CTRL('h'):
	case KEY_BACKSPACE:
	  if (logicalpoint > 0)
	    {
	      if (logicalpoint > 1)
		{
		  inc = charlengthback(tempstr.c_str(), logicalpoint);
		}
	      else
		inc = 1;
	      
	      tempstr.replace(logicalpoint - inc, inc, "", 0);
	      strend = strlen(tempstr.c_str()) - 1;
	      logicalpoint -= inc;

	      if (disppoint - inc >= 0)
		disppoint -= inc;
	      else
		{
		  dispoffset = nextpage((char *)tempstr.c_str(), logicalpoint, displength);
		  disppoint = logicalpoint - dispoffset;
		}
	      disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	    }
	  mark = -1;
	  break;
	}

      // input c_str()
      if ((' ' <= key  && key <= '~' && !maykanji(key)) || maykanji(key))
	{
	  key1 = key;
	  tmplogicalpoint = logicalpoint;
	  if (tmplogicalpoint <= strend)
	    {
	      tempstr1 = tempstr.substr(0, tmplogicalpoint);
	      stemp[0] = (unsigned char)key1; stemp[1] = 0;
	      tempstr1 += std::string(stemp);
	      stemp[0] = tempstr.c_str()[tmplogicalpoint]; stemp[1] = 0;
	      tempstr1 += std::string(stemp);
	      tempstr1 += tempstr.substr(tmplogicalpoint + 1);
	      tempstr = tempstr1;
	    }
	  else if (tmplogicalpoint == strend + 1)
	    {
	      stemp[0] = (unsigned char)key1; stemp[1] = 0;
	      tempstr += std::string(stemp);
	    }
	  tmplogicalpoint += 1;
	  strend = strlen(tempstr.c_str()) - 1;

	  if (maykanji(key1))
	    {
	      // second byte 
	      key2 = topazgetch(win);
	      if (tmplogicalpoint <= strend)
		{
		  tempstr1 = tempstr.substr(0, tmplogicalpoint);
		  stemp[0] = (unsigned char)key2; stemp[1] = 0;
		  tempstr1 += std::string(stemp);
		  stemp[0] = tempstr.c_str()[tmplogicalpoint]; stemp[1] = 0;
		  tempstr1 += std::string(stemp);
		  tempstr1 += tempstr.substr(tmplogicalpoint + 1);
		  tempstr = tempstr1;
		}
	      else if (tmplogicalpoint == strend + 1)
		{
		  stemp[0] = (unsigned char)key2; stemp[1] = 0;
		  tempstr += std::string(stemp);
		}
	      tmplogicalpoint += 1;
	    }
	  strend = strlen(tempstr.c_str()) - 1;

	  inc = charlength(tempstr.c_str()[logicalpoint]);
	  if (logicalpoint <= strend + 1)
	    {
	      itemp = linelength((char *)tempstr.c_str(), dispoffset, displength);
	      if (logicalpoint + inc <= dispoffset + itemp 
		  && 
		  (((logicalpoint + inc) <= strend
		    && disppoint + inc + charlength(tempstr.c_str()[logicalpoint + inc]) <=  displength)
		   || ((logicalpoint + inc) == strend + 1
		       && disppoint + inc  <=  displength)))
		{
		  logicalpoint += inc;
		  disppoint += inc;
		}
	      else
		{
		  logicalpoint += inc;
		  dispoffset = nextpage((char *)tempstr.c_str(), logicalpoint, displength);
		  disppoint = logicalpoint - dispoffset;
		  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
		}
	    }

	  disptext(win, (char *)tempstr.c_str(), y, x, displength, dispoffset, dispoffset);
	  mark = -1;
	}

      if (key == CTRL('g'))
	{
	  key = LOCAL_ESCAPE_KEY;
	  mark = -1;
	  break;
	}

      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	{
	  mark = -1;
	  break;
	}
    }

  if (key == LOCAL_LF_KEY)
    {
      if (editfilename)
	{
	  if(buildfullpath((char *)tempstr.c_str(), tmppath))
	    *input = std::string(tmppath);
	  else
	    return false;
	}
      else
	*input = tempstr;
    }
  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}



//
//  selection color
//
int execcolorselect(std::string *input, std::string *otherscaption)
{
  return execselect(Color, input, otherscaption);
}

//
//  yes/no window
//
int exec_yesno_select(std::string *input, std::string *otherscaption)
{
  return execselect(Yesno, input, otherscaption);
}

//
//  frametype window
//
int frametype_select(std::string *input, std::string *otherscaption)
{
  return execselect(Frametype, input, otherscaption);
}

//
//  tickdirection window
//
int tickdirection_select(std::string *input, std::string *otherscaption)
{
  return execselect(Tickdir, input, otherscaption);
}

//
//  scaling window
//
int scaling_select(std::string *input, std::string *otherscaption)
{
  return execselect(Scaling, input, otherscaption);
}

//
//  memory data scaling window
//
int mscaling_select(std::string *input, std::string *otherscaption)
{
  return execselect(MScaling, input, otherscaption);
}

//
//  linearfit function
//
int linearfitfunction_select(std::string *input, std::string *otherscaption)
{
  return execselect(MLinearfit, input, otherscaption);
}

//
//  linecap window
//
int linecap_select(std::string *input, std::string *otherscaption)
{
  return execselect(Linecap, input, otherscaption);
}

//
//  linejoin window
//
int linejoin_select(std::string *input, std::string *otherscaption)
{
  return execselect(Linejoin, input, otherscaption);
}

//
//  linestyle window
//
int linestyle_select(std::string *input, std::string *otherscaption)
{
  return execselect(Linestyle, input, otherscaption);
}

//
//  fillstyle window
//
int fillstyle_select(std::string *input, std::string *otherscaption)
{
  return execselect(Fillstyle, input, otherscaption);
}

//
//  fillrule window
//
int fillrule_select(std::string *input, std::string *otherscaption)
{
  return execselect(Fillrule, input, otherscaption);
}

//
//  mag window
//
int mag_select(std::string *input, std::string *otherscaption)
{
  return execselect(Mag, input, otherscaption);
}

//
//  interpolate window
//
int interpolate_select(std::string *input, std::string *otherscaption)
{
  return execselect(Interpolate, input, otherscaption);
}

//
//  markerstyle window
//
int markerstyle_select(std::string *input, std::string *otherscaption)
{
  return execselect(Markerstyle, input, otherscaption);
}

//
//  commentline window
//
int commentlineaction_select(std::string *input, std::string *otherscaption)
{
  return execselect(Commentlineaction, input, otherscaption);
}

//
//  datasearchpath window
//
int datasearchpath_select(std::string *input, std::string *otherscaption)
{
  return execselect(Datasearchpath, input, otherscaption);
}

//
//  partstype window
//
int partstype_select(std::string *input, std::string *otherscaption)
{
  return execselect(Partstype, input, otherscaption);
}

//
//  fontface window
//
int fontface_select(std::string *input, std::string *otherscaption)
{
  return execselect(Fontface, input, otherscaption);
}

//
//  jpfontface window
//
int jpfontface_select(std::string *input, std::string *otherscaption)
{
  return execselect(Jpfontface, input, otherscaption);
}

//
//  partsinterpolation window
//
int partsinterpolation_select(std::string *input, std::string *otherscaption)
{
  return execselect(Partsinterpolation, input, otherscaption);
}

//
//  labelparity window
//
int labelparity_select(std::string *input, std::string *otherscaption)
{
  return execselect(Labelparity, input, otherscaption);
}

//
//  labelCPparity window
//
int labelCPparity_select(std::string *input, std::string *otherscaption)
{
  return execselect(LabelCPparity, input, otherscaption);
}

//
//  labelformat window
//
int labelformat_select(std::string *input, std::string *otherscaption)
{
  return execselect(Labelformat, input, otherscaption);
}

//
//  orderlist window
//
int orderlist_select(std::string *input, std::string *otherscaption)
{
  return execselect(Orderlist, input, otherscaption);
}

//
//  papersize window
//
int papersizelist_select(std::string *input, std::string *otherscaption)
{
  return execselect(Papersize, input, otherscaption);
}


//
//  base of selection window
//
int execselect(CUIMENU *menu, std::string *input, std::string *otherscaption)
{
  CUIMENU *m;
  int itemlength = 0;
  int itemno;
  int flag1;
  std::string tempinput = *input;
  int selectedkey;
  int addoffset = 2;

  if (menuwin == 0)
    return false;

  // get number of items
  m = menu;
  for (; m->item; m++)
    {
      if (itemlength < (int)strlen(m->item))
	itemlength = strlen(m->item);
    }
  itemno = m - menu - 1;
  if (itemno < 1)
    return false;

  // search initially selected item and number
  for (int i = 1; i <= itemno; i++)
    {
      if (std::string(menu[i].macroscript) == *input)
	{
	  selectedkey = i;
	  break;
	}
      else
	selectedkey = 1;
    }


  showcurrentmenu(menu, itemno, addoffset);
  // show message
  wattron(menuwin, A_BOLD);
  wmove(menuwin, topmargin, leftmargin);
  wprintw(menuwin, "%s\n", otherscaption->c_str());
  wattroff(menuwin, A_BOLD);
  wrefresh(menuwin);

  flag1 = cursormove (menu, &selectedkey, itemno, addoffset);
  
  if (flag1 == LOCAL_LF_KEY  && menu[selectedkey].submenu)
    {
      if (edittextwindow(menuwin, &tempinput, otherscaption, false) == true)
	{
	  *input = tempinput;
	  return true;
	}
      else
	  return false;
     }
  else if (flag1 == LOCAL_LF_KEY && !menu[selectedkey].submenu)
    {
      *input = std::string(menu[selectedkey].macroscript);
      return true;
    }
  else
    return false;
} 


void showcor(WINDOW *win, int x, int y)
{
  double sx, sy;
  frame *fr;
  double fx1, fy1;
  char stemp[300];
  int len;

  if (win == 0)
    return;

  fr = &(*gra->farray)[gra->currentframe];

  wmove(win, 9, 2);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(wx, wy) = (%5d, %5d)", x, y);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 9, len);
  wprintw(win, stemp);

  wmove(win, 11, 2);
  fx1 = fr->vtow_x(x);
  fy1 = fr->vtow_y(y);
  sx = (*fr->axarray)[0].doubleworld_to_virtual(fx1, fy1);
  sy = (*fr->axarray)[1].doubleworld_to_virtual(fx1, fy1);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(x, y) = (%.8g, %.8g)", sx, sy);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 11, len);
  wprintw(win, stemp);

  wmove(win, 12, 2);
  fx1 = fr->vtow_x(x);
  fy1 = fr->vtow_y(y);
  sx = (*fr->axarray)[2].doubleworld_to_virtual(fx1, fy1);
  sy = (*fr->axarray)[3].doubleworld_to_virtual(fx1, fy1);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(t, r) = (%.8g, %.8g)", sx, sy);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 12, len);
  wprintw(win, stemp);

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

void shownodecor(WINDOW *win, buffarray *input, int node)
{
  double sx, sy;
  frame *fr;
  double fx1, fy1;
  char stemp[300];
  int len;

  if (win == 0)
    return;

  fr = &(*gra->farray)[gra->currentframe];

  wmove(win, 11, 2);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(wx, wy) = (%5d, %5d)",
	  atoi((*input)[2 * node + 0].c_str()),
	  atoi((*input)[2 * node + 1].c_str()));
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 11, len);
  wprintw(win, stemp);

  wmove(win, 13, 2);
  fx1 = fr->vtow_x(atoi((*input)[2 * node + 0].c_str()));
  fy1 = fr->vtow_y(atoi((*input)[2 * node + 1].c_str()));
  sx = (*fr->axarray)[0].doubleworld_to_virtual(fx1, fy1);
  sy = (*fr->axarray)[1].doubleworld_to_virtual(fx1, fy1);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(x, y) = (%.8g, %.8g)", sx, sy);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 13, len);
  wprintw(win, stemp);

  wmove(win, 14, 2);
  fx1 = fr->vtow_x(atoi((*input)[2 * node + 0].c_str()));
  fy1 = fr->vtow_y(atoi((*input)[2 * node + 1].c_str()));
  sx = (*fr->axarray)[2].doubleworld_to_virtual(fx1, fy1);
  sy = (*fr->axarray)[3].doubleworld_to_virtual(fx1, fy1);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(t, r) = (%.8g, %.8g)", sx, sy);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 14, len);
  wprintw(win, stemp);

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}


void showrects(buffarray *input)
{
  int x1, x2, y1, y2, xmin, xmax, ymin, ymax, itemp;
  FILE* f;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  pbegin(true, f);
  psetpaintmode(true, f, 1);
  psetforecolor(true, f, 0, 0, 0);
  plinestyle(true, f, 0, 0, 0, 0);
  
  psettempdashline(true, f);
  for (int i = 0; i < input->getarraylength(); i += 4)
    {
      x1 = atoi((*input)[i + 0].c_str());
      y1 = atoi((*input)[i + 1].c_str());
      x2 = atoi((*input)[i + 2].c_str());
      y2 = atoi((*input)[i + 3].c_str());
      
      if (x1 < x2)
	{
	  xmin = x1; xmax = x2;
	}
      else
	{
	  xmin = x2; xmax = x1;
	}
      if (y1 < y2)
	{
	  ymin = y1; ymax = y2;
	}
      else
	{
	  ymin = y2; ymax = y1;
	}
      
      pbeginpath(true, f);
      pmoveto(true, f, xmin, ymin);
      plineto(true, f, xmax, ymin);
      plineto(true, f, xmax, ymax);
      plineto(true, f, xmin, ymax);
      plineto(true, f, xmin, ymin);
      pstroke(true, f);
    }

  
  prestorelinestyle(true, f);
  psetpaintmode(true, f, 0);
  ppaint(true, f);
  
  pflush(true, f);

}

void incarray (buffarray *input, int inc_x1, int inc_y1, int inc_x2, int inc_y2)
{
  char stemp[100];

  for (int i = 0; i < input->getarraylength(); i++)
    {
      switch(i % 4)
	{
	case 0:
	  sprintf(stemp, "%d", atoi((*input)[i].c_str()) + inc_x1);
	  break;
	case 1:
	  sprintf(stemp, "%d", atoi((*input)[i].c_str()) + inc_y1);
	  break;
	case 2:
	  sprintf(stemp, "%d", atoi((*input)[i].c_str()) + inc_x2);
	  break;
	case 3:
	  sprintf(stemp, "%d", atoi((*input)[i].c_str()) + inc_y2);
	  break;
	}
      input->setdata(i, stemp);
    }
}



void movedisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                                UP (Ctrl+p)                                    ");
  wmove(win, 3, 2);
  wprintw(win, "               LEFT (Ctrl+b)                   RIGHT (Ctrl+f)                  ");
  wmove(win, 4, 2);
  wprintw(win, "                               DOWN (Ctrl+n)                                   ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                               ");
  wmove(win, 6, 2);
  wprintw(win, "                      Space : toggle small/large skip                          ");
  wmove(win, 7, 2);
  wprintw(win, "                            CR: exit  ESC: cancel                              ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

//
//   moveshape window
//
int movewindow(WINDOW *win, buffarray *input, std::string *caption)
{

  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;

  if (win == 0)
    return false;

  if ((input->getarraylength()) % 4 != 0 || input->getarraylength() == 0)
    return false;
  inc = smallinc;

  arr = new buffarray;
  *arr = *input;

  movedisp(win, caption);
  while (1)
    {
      
      showrects(arr);
      key = topazgetch(win);
      showrects(arr);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  incarray(arr, -inc, 0, -inc, 0);
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  incarray(arr, inc, 0, inc, 0);
	    break;
	case KEY_UP:
	case CTRL('p'):
	  incarray(arr, 0, -inc, 0, -inc);
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  incarray(arr, 0, inc, 0, inc);
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	}

      
      if (key == CTRL('g'))
	{
	  key = LOCAL_ESCAPE_KEY;
	  break;
	}

      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }

  if (key == LOCAL_LF_KEY)
    *input = *arr;

  delete arr;

  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}



void shownode(buffarray *input, int node)
{
  int x1, x2, y1, y2, x, y, itemp;
  int margin = (int)(50.0 / gra->mag);
  FILE* f;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  pbegin(true, f);
  psetpaintmode(true, f, 1);
  psetforecolor(true, f, 0, 0, 0);
  plinestyle(true, f, 0, 0, 0, 0);
  
  for (int i = 0; i < input->getarraylength(); i += 4)
    {
      x1 = atoi((*input)[i + 0].c_str());
      y1 = atoi((*input)[i + 1].c_str());
      x2 = atoi((*input)[i + 2].c_str());
      y2 = atoi((*input)[i + 3].c_str());
      
      if (node == 0)
	{
	  x = x1;
	  y = y1;
	}
      else
	{
	  x = x2;
	  y = y2;
	}

      pbeginpath(true, f);
      pmoveto(true, f, x - margin, y - margin);
      plineto(true, f, x + margin, y - margin);
      plineto(true, f, x + margin, y + margin);
      plineto(true, f, x - margin, y + margin);
      plineto(true, f, x - margin, y - margin);
      pstroke(true, f);
    }

  
  psetpaintmode(true, f, 0);
  ppaint(true, f);
  
  pflush(true, f);

}


void showstretchcor(WINDOW *win, buffarray *input, int node)
{
  char stemp[300];
  int len;

  if (win == 0)
    return;

  wmove(win, 11, 2);
  wprintw(win, "                                                                           ");

  sprintf(stemp, "(%5d, %5d) - (%5d, %5d)",
	  atoi((*input)[0].c_str()),
	  atoi((*input)[1].c_str()),
	  atoi((*input)[2].c_str()),
	  atoi((*input)[3].c_str()));
  
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 11, len);
  wprintw(win, stemp);

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

void stretchdisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                                UP (Ctrl+p)                                    ");
  wmove(win, 3, 2);
  wprintw(win, "               LEFT (Ctrl+b)                   RIGHT (Ctrl+f)                  ");
  wmove(win, 4, 2);
  wprintw(win, "                               DOWN (Ctrl+n)                                   ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                               ");
  wmove(win, 6, 2);
  wprintw(win, "                      Space : toggle small/large skip                          ");
  wmove(win, 7, 2);
  wprintw(win, "                       [n]ext node  [p]revious node                            ");
  wmove(win, 8, 2);
  wprintw(win, "                            CR: exit  ESC: cancel                              ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

//
//   stretch window
//
int stretchwindow(WINDOW *win, buffarray *input, std::string *caption)
{

  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;
  bool node = 0;

  if (win == 0)
    return false;

  if ((input->getarraylength()) % 4 != 0)
    return false;

  inc = smallinc;

  arr = new buffarray;
  *arr = *input;

  stretchdisp(win, caption);
  while (1)
    {
      showstretchcor(win, arr, node);
      showrects(arr);
      shownode(arr, node);
      key = topazgetch(win);
      showrects(arr);
      shownode(arr, node);
      showstretchcor(win, arr, node);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  if (node == 0)
	    incarray(arr, -inc, 0, 0, 0);
	  else
	    incarray(arr, 0, 0, -inc, 0);
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  if (node == 0)
	    incarray(arr, inc, 0, 0, 0);
	  else
	    incarray(arr, 0, 0, inc, 0);
	    break;
	case KEY_UP:
	case CTRL('p'):
	  if (node == 0)
	    incarray(arr, 0, -inc, 0, 0);
	  else
	    incarray(arr, 0, 0, 0, -inc);
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  if (node == 0)
	    incarray(arr, 0, inc, 0, 0);
	  else
	    incarray(arr, 0, 0, 0, inc);
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	case 'n':
	  if (node == 1)
	      node = 0;
	  else
	      node = 1;
	  break;
	case 'p':
	  if (node == 0)
	      node = 1;
	  else
	      node = 0;
	  break;
	}

      if (key == CTRL('g'))
	{
	  key = LOCAL_ESCAPE_KEY;
	  break;
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }

  if (key == LOCAL_LF_KEY)
    *input = *arr;

  delete arr;

  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}

/****************************************************************/
/*******************  change node  ******************************/
/****************************************************************/


void incnode (buffarray *input, int node, int inc_x, int inc_y)
{
  char stemp[100];

  sprintf(stemp, "%d", atoi((*input)[node * 2].c_str()) + inc_x);
  input->setdata(node * 2, stemp);

  sprintf(stemp, "%d", atoi((*input)[node * 2 + 1].c_str()) + inc_y);
  input->setdata(node * 2 + 1, stemp);
}



void showguide(buffarray *input, int type, int interpolate)
{
  int x1, x2, y1, y2, x, y, itemp;
  int length = input->getarraylength();
  FILE* f;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  if (length < 4 || length % 2 == 1)
    return;
  
  pbegin(true, f);
  psetpaintmode(true, f, 1);
  psetforecolor(true, f, 0, 0, 0);
  plinestyle(true, f, 0, 0, 0, 0);
  psettempdashline(true, f);
  
  if (type == 1 || (type == 2 && (interpolate == 0 || interpolate == 3)))
    {
      pbeginpath(true, f);
      x = atoi((*input)[0].c_str());
      y = atoi((*input)[1].c_str());
      pmoveto(true, f, x, y);
      for (int i = 2; i < length; i += 2)
	{
	  x = atoi((*input)[i + 0].c_str());
	  y = atoi((*input)[i + 1].c_str());
	  plineto(true, f, x, y);
	}
      pstroke(true, f);
    }
  else if (type == 2 && (interpolate == 1 || interpolate == 4))
    {
      pbeginpath(true, f);
      x = atoi((*input)[0].c_str());
      y = atoi((*input)[1].c_str());
      pmoveto(true, f, x, y);
      for (int i = 2; i < length; i += 2)
	{
	  x = atoi((*input)[i + 0].c_str());
	  y = atoi((*input)[i + 1].c_str());
	  plineto(true, f, x, y);
	}
      x = atoi((*input)[0].c_str());
      y = atoi((*input)[1].c_str());
      plineto(true, f, x, y);
      pstroke(true, f);
    }
  else
    {
      x1 = atoi((*input)[0].c_str());
      y1 = atoi((*input)[1].c_str());
      x2 = atoi((*input)[2].c_str());
      y2 = atoi((*input)[3].c_str());

      pbeginpath(true, f);
      pmoveto(true, f, x1, y1);
      plineto(true, f, x1, y2);
      plineto(true, f, x2, y2);
      plineto(true, f, x2, y1);
      plineto(true, f, x1, y1);
      pstroke(true, f);
    }
  prestorelinestyle(true, f);
  psetpaintmode(true, f, 0);
  ppaint(true, f);
  pflush(true, f);
}


void shownodes(buffarray *input, int node)
{
  int x, y, itemp;
  int margin = (int)(50.0 / gra->mag);
  FILE* f;
  std::string str, str1;
   
  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  pbegin(true, f);
  psetpaintmode(true, f, 1);
  psetforecolor(true, f, 0, 0, 0);
  plinestyle(true, f, 0, 0, 0, 0);

  x = atoi((*input)[2 * node + 0].c_str());
  y = atoi((*input)[2 * node + 1].c_str());
  
  pbeginpath(true, f);
  pmoveto(true, f, x - margin, y - margin);
  plineto(true, f, x + margin, y - margin);
  plineto(true, f, x + margin, y + margin);
  plineto(true, f, x - margin, y + margin);
  plineto(true, f, x - margin, y - margin);
  pstroke(true, f);

  psetpaintmode(true, f, 0);
  ppaint(true, f);
  
  pflush(true, f);

}


void changenodesdisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                                UP (Ctrl+p)                                    ");
  wmove(win, 3, 2);
  wprintw(win, "               LEFT (Ctrl+b)                   RIGHT (Ctrl+f)                  ");
  wmove(win, 4, 2);
  wprintw(win, "                               DOWN (Ctrl+n)                                   ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                               ");
  wmove(win, 6, 2);
  wprintw(win, "                      Space : toggle small/large skip                          ");
  wmove(win, 7, 2);
  wprintw(win, "                        [n]ext node  [p]revious node                           ");
  wmove(win, 8, 2);
  wprintw(win, "                Ctrl+y : append node  Ctrl+d : remove node                     ");
  wmove(win, 9, 2);
  wprintw(win, "                            CR: exit  ESC: cancel                              ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

int changenodeswindow(WINDOW *win, buffarray *input, std::string *caption)
{

  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;
  int node = 0;
  int length;
  int type;
  int interpolate;
  char stemp[100];
  int x, y;

  if (win == 0)
    return false;

  if ((input->getarraylength()) % 2 != 0 || input->getarraylength() < 4)
    return false;

  inc = smallinc;

  arr = new buffarray;
  *arr = *input;
  interpolate = atoi((*arr)[arr->getarraylength() - 1].c_str());
  arr->remove(arr->getarraylength() - 1);
  type = atoi((*arr)[arr->getarraylength() - 1].c_str());
  arr->remove(arr->getarraylength() - 1);

  length = arr->getarraylength() / 2;

  if (interpolate != 0 && interpolate != 1 &&  interpolate != 3 &&interpolate != 4)
    return false;

  if (type < 1 || type > 5)
    return false;

  if (type == 1 || type == 3 || type == 4 || type == 5)
    {
      if (length != 2)
	return false;
    }

  changenodesdisp(win, caption);
  while (1)
    {
      shownodecor(win, arr, node);
      shownodes(arr, node);
      showguide(arr, type, interpolate);
      key = topazgetch(win);
      shownodes(arr, node);
      showguide(arr, type, interpolate);
      shownodecor(win, arr, node);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  incnode (arr, node, -inc, 0);
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  incnode (arr, node, inc, 0);
	    break;
	case KEY_UP:
	case CTRL('p'):
	  incnode (arr, node, 0, -inc);
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  incnode (arr, node, 0, inc);
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	  //next node
	case 'n':
	  if (node < length - 1)
	    node++;
	  else
	    node = 0;
	  break;
	  //previouts node
	case 'p':
	  if (0 < node)
	    node--;
	  else
	    node = length - 1;
	  break;
	  //first node
	case CTRL('a'):
	  node = 0;
	  break;
	  //last node
	case CTRL('e'):
	  node = length - 1;
	  break;
	  //previouts node
	  //append node
	case CTRL('y'):
	  if (type == 2)
	    {
	      if (node < length - 1)
		{
		  x = (atoi((*arr)[2 * (node + 1) + 0].c_str()) + atoi((*arr)[2 * node + 0].c_str())) / 2;
		  y = (atoi((*arr)[2 * (node + 1) + 1].c_str()) + atoi((*arr)[2 * node + 1].c_str())) / 2;
		  sprintf(stemp, "%d", y);
		  arr->addat(2 * (node + 1), stemp);
		  sprintf(stemp, "%d", x);
		  arr->addat(2 * (node + 1), stemp);
		}
	      else
		{
		  x = (atoi((*arr)[2 * (length - 1) + 0].c_str()) + atoi((*arr)[0].c_str())) / 2;
		  y = (atoi((*arr)[2 * (length - 1) + 1].c_str()) + atoi((*arr)[1].c_str())) / 2;
		  sprintf(stemp, "%d", x);
		  arr->add(stemp);
		  sprintf(stemp, "%d", y);
		  arr->add(stemp);
		}
	      length++;
	      node++;
	    }
	  break;
	  //delete node
	case CTRL('d'):
	  if (type == 2 && length > 2)
	    {
	      arr->remove(2 * node);
	      arr->remove(2 * node);
	      length--;
	      node--;
	    }
	  break;
	}

      if (key == CTRL('g'))
	{
	  key = LOCAL_ESCAPE_KEY;
	  break;
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }

  if (key == LOCAL_LF_KEY)
      *input = *arr;

  delete arr;

  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}


/****************************************************************/
/*******************  input parts  ******************************/
/****************************************************************/


void showpointer(buffarray *input, int x, int y, int type)
{
  int margin = (int)(500.0 / gra->mag);
  int length = input->getarraylength();
  int x1, y1, itemp;
  FILE* f;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;

  pbegin(true, f);
  psetpaintmode(true, f, 1);
  psetforecolor(true, f, 0xffff, 0, 0);
  plinestyle(true, f, 0, 0, 0, 0);

  if (length % 2 == 0 && length > 0)
    {
      pbeginpath(true, f);
      pmoveto(true, f, x - margin, y);
      plineto(true, f, x + margin, y);
      pstroke(true, f);

      pbeginpath(true, f);
      pmoveto(true, f, x, y - margin);
      plineto(true, f, x, y + margin);
      pstroke(true, f);

      pbeginpath(true, f);
      if (type == 1 || type == 2)
	{
	  pmoveto(true, f, x, y);
	  plineto(true, f, atoi((*input)[length - 2].c_str()), atoi((*input)[length - 1].c_str()));
	}
      else
	{
	  x1 = atoi((*input)[length - 2].c_str());
	  y1 = atoi((*input)[length - 1].c_str());

	  pmoveto(true, f, x1, y1);
	  plineto(true, f, x, y1);
	  plineto(true, f, x, y);
	  plineto(true, f, x1, y);
	  plineto(true, f, x1, y1);
	}
      pstroke(true, f);
    }
  else if (length == 0)
    {
      pbeginpath(true, f);
      pmoveto(true, f, x - margin, y);
      plineto(true, f, x + margin, y);
      pstroke(true, f);
      
      pbeginpath(true, f);
      pmoveto(true, f, x, y - margin);
      plineto(true, f, x, y + margin);
      pstroke(true, f);
    }
  
  psetpaintmode(true, f, 0);
  ppaint(true, f);
  
  pflush(true, f);
}

void inputpartsdisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                                UP (Ctrl+p)                                    ");
  wmove(win, 3, 2);
  wprintw(win, "               LEFT (Ctrl+b)                   RIGHT (Ctrl+f)                  ");
  wmove(win, 4, 2);
  wprintw(win, "                               DOWN (Ctrl+n)                                   ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                               ");
  wmove(win, 6, 2);
  wprintw(win, "                      Space : toggle small/large skip                          ");
  wmove(win, 7, 2);
  wprintw(win, "                     m: fix node   e[x]it   ESC: cancel                        ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}


int inputpartswindow(WINDOW *win, buffarray *input, std::string *caption)
{

  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;
  int node = 0;
  int length;
  int type;
  int interpolate;
  char stemp[100];
  int x, y;
  frame *fr;

  if (win == 0)
    return false;

  if (input->getarraylength() != 2)
    return false;

  inc = smallinc;

  arr = new buffarray;
  interpolate = atoi((*input)[input->getarraylength() - 1].c_str());
  type = atoi((*input)[input->getarraylength() - 2].c_str());

  length = arr->getarraylength() / 2;

  if (interpolate != 0 && interpolate != 1 &&  interpolate != 3 &&interpolate != 4)
    return false;


  if (type < 1 || type > 5)
    return false;

  fr = &(*gra->farray)[gra->currentframe];
  x = fr->wtov_x((fr->p2.x - fr->p1.x) / 2);
  y = fr->wtov_y((fr->p2.y - fr->p1.y) / 2);
  inputpartsdisp(win, caption);
  while (1)
    {
      showcor(win, x, y);
      showpointer(arr, x, y, type);
      showguide(arr, type, interpolate);
      key = topazgetch(win);
      showpointer(arr, x, y, type);
      showguide(arr, type, interpolate);
      showcor(win, x, y);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  x += -inc;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  x += inc;
	    break;
	case KEY_UP:
	case CTRL('p'):
	  y += -inc;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  y += inc;
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	  //next node
	case 'n':
	  if (node < length - 1)
	      node++;
	  else
	      node = 0;
	  break;
	  //previouts node
	case 'p':
	  if (0 < node)
	      node--;
	  else
	      node = length - 1;
	  break;
	  //append node
	case LOCAL_LF_KEY:
	  if (type == 2)
	    {
	      sprintf(stemp, "%d", x);
	      arr->add(stemp);
	      sprintf(stemp, "%d", y);
	      arr->add(stemp);
	      length++;
	    }
	  else if (type != 2 && length < 2)
	    {
	      sprintf(stemp, "%d", x);
	      arr->add(stemp);
	      sprintf(stemp, "%d", y);
	      arr->add(stemp);
	      length++;

	      if (length == 2)
		{
		  key = LOCAL_LF_KEY;
		  goto j4;
		}
	    }
	  break;
	}

      if (key == CTRL('g') || key == LOCAL_ESCAPE_KEY)
	{
	  key = LOCAL_ESCAPE_KEY;
	  goto j4;
	}
      if (key == 'x' || key == '@')
	{
	  key = LOCAL_LF_KEY;
	  goto j4;
	}
    }
j4:;      

  if (key == LOCAL_LF_KEY)
      *input = *arr;
  delete arr;

  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}

/****************************************************************/
/*******************  trimming  *********************************/
/****************************************************************/


void trimmingdisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                               UP (Ctrl+p)                                    ");
  wmove(win, 3, 2);
  wprintw(win, "              LEFT (Ctrl+b)                   RIGHT (Ctrl+f)                  ");
  wmove(win, 4, 2);
  wprintw(win, "                              DOWN (Ctrl+n)                                   ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                              ");
  wmove(win, 6, 2);
  wprintw(win, "                     Space : toggle small/large skip                          ");
  wmove(win, 7, 2);
  wprintw(win, "                        CR: fix node   ESC: cancel                            ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}


void showcor(WINDOW *win, buffarray *arr, int x, int y)
{
  double sx, sy, ex, ey;
  frame *fr;
  double fx1, fy1, fx2, fy2;
  char stemp[300];
  int len;

  if (win == 0)
    return;

  fr = &(*gra->farray)[gra->currentframe];
  wmove(win, 9, 2);
  switch(arr->getarraylength())
    {
    case 0:
      fx1 = fr->vtow_x(x);
      fy1 = fr->vtow_y(y);
      sx = (*fr->axarray)[0].doubleworld_to_virtual(fx1, fy1);
      sy = (*fr->axarray)[1].doubleworld_to_virtual(fx1, fy1);
      wprintw(win, "                                                                           ");
      sprintf(stemp, "(%.8g, %.8g)", sx, sy);
      len = strlen(stemp); len = len / 2; len = 39 - len;
      wmove(win, 9, len);
      wprintw(win, stemp);
      break;
    case 2:
      fx1 = fr->vtow_x(atoi((*arr)[0].c_str()));
      fy1 = fr->vtow_y(atoi((*arr)[1].c_str()));
      sx = (*fr->axarray)[0].doubleworld_to_virtual(fx1, fy1);
      sy = (*fr->axarray)[1].doubleworld_to_virtual(fx1, fy1);
      fx2 = fr->vtow_x(x);
      fy2 = fr->vtow_y(y);
      ex = (*fr->axarray)[0].doubleworld_to_virtual(fx2, fy2);
      ey = (*fr->axarray)[1].doubleworld_to_virtual(fx2, fy2);
      wprintw(win, "                                                                           ");
      sprintf(stemp, "(%.8g, %.8g)  -  (%.8g, %.8g)", sx, sy, ex, ey);
      len = strlen(stemp); len = len / 2; len = 39 - len;
      wmove(win, 9, len);
      wprintw(win, stemp);
      break;
    }
  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}


int trimmingwindow(WINDOW *win, buffarray *input, std::string *caption)
{
  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;
  int length;
  int type = 3;
  char stemp[100], stemp1[100], stemp2[100];
  int x, y;
  double fx, fy;
  frame *fr;

  if (win == 0)
    return false;

  fr = &(*gra->farray)[gra->currentframe];
  inc = smallinc;

  arr = new buffarray;
  length = arr->getarraylength() / 2;

  x = fr->wtov_x((fr->p2.x - fr->p1.x) / 2);
  y = fr->wtov_y((fr->p2.y - fr->p1.y) / 2);
  trimmingdisp(win, caption);
  while (1)
    {
      showcor(win, arr, x, y);
      showpointer(arr, x, y, type);
      showguide(arr, type, 1);
      key = topazgetch(win);
      showcor(win, arr, x, y);
      showpointer(arr, x, y, type);
      showguide(arr, type, 1);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  x += -inc;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  x += inc;
	    break;
	case KEY_UP:
	case CTRL('p'):
	  y += -inc;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  y += inc;
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	  //append node
	case LOCAL_LF_KEY:
	  sprintf(stemp, "%d", x);
	  arr->add(stemp);
	  sprintf(stemp, "%d", y);
	  arr->add(stemp);
	  length++;
	  if (length == 2)
	    {
	      key = LOCAL_LF_KEY;
	      goto j4;
	    }
	  break;
	}

      if (key == CTRL('g') || key == LOCAL_ESCAPE_KEY)
	{
	  key = LOCAL_ESCAPE_KEY;
	  goto j4;
	}
    }
j4:;      

  if (key == LOCAL_LF_KEY)
    {
      fx = fr->vtow_x(atoi((*arr)[0].c_str()));
      fy = fr->vtow_y(atoi((*arr)[1].c_str()));
      sprintf(stemp1, "%.15g", (*fr->axarray)[0].doubleworld_to_virtual(fx, fy));
      sprintf(stemp2, "%.15g", (*fr->axarray)[1].doubleworld_to_virtual(fx, fy));
      (*arr)[0] = std::string(stemp1); 
      (*arr)[1] = std::string(stemp2); 
      fx = fr->vtow_x(atoi((*arr)[2].c_str()));
      fy = fr->vtow_y(atoi((*arr)[3].c_str()));
      sprintf(stemp1, "%.15g", (*fr->axarray)[0].doubleworld_to_virtual(fx, fy));
      sprintf(stemp2, "%.15g", (*fr->axarray)[1].doubleworld_to_virtual(fx, fy));
      (*arr)[2] = std::string(stemp1); 
      (*arr)[3] = std::string(stemp2); 
      *input = *arr;
    }
  delete arr;

  if (key == LOCAL_ESCAPE_KEY)
    return false;

  return true;

}

/****************************************************************/
/*******************  watch  ***********************************/
/****************************************************************/


void watchdisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                               UP (Ctrl+p)                                 ");
  wmove(win, 3, 2);
  wprintw(win, "              LEFT (Ctrl+b)                   RIGHT (Ctrl+f)               ");
  wmove(win, 4, 2);
  wprintw(win, "                              DOWN (Ctrl+n)                                ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 6, 2);
  wprintw(win, "                     Space : toggle small/large skip                       ");
  wmove(win, 7, 2);
  wprintw(win, "                              CR, ESC: exit                                ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}



int watchwindow(WINDOW *win, buffarray *input, std::string *caption)
{
  buffarray *arr;
  int key;
  int inc;
  int smallinc = (int)(smallskip / gra->mag);
  int largeinc = (int)(largeskip / gra->mag);
  bool skipstatus = 0;
  int type = 3;
  int x, y;
  frame *fr;
  char stemp[100];
  double fx, fy;

  if (win == 0)
    return false;

  fr = &(*gra->farray)[gra->currentframe];
  inc = smallinc;

  arr = new buffarray;

  x = fr->wtov_x((fr->p2.x - fr->p1.x) / 2);
  y = fr->wtov_y((fr->p2.y - fr->p1.y) / 2);
  watchdisp(win, caption);
  while (1)
    {
      showcor(win, x, y);
      showpointer(arr, x, y, type);
      key = topazgetch(win); 
      showpointer(arr, x, y, type);
      showcor(win, x, y);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  x += -inc;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  x += inc;
	    break;
	case KEY_UP:
	case CTRL('p'):
	  y += -inc;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  y += inc;
	  break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	case LOCAL_LF_KEY:
	  key = LOCAL_LF_KEY;
	  goto j5;
	  
	}
      
      if (key == CTRL('g') || key == LOCAL_ESCAPE_KEY)
	{
	  key = LOCAL_ESCAPE_KEY;
	  goto j5;
	}
    }

 j5:;
  if (key == LOCAL_LF_KEY)
    {
      fx = fr->vtow_x(x);
      fy = fr->vtow_y(y);
      input->setbuff(0);
      sprintf(stemp, "%.15g", (*fr->axarray)[0].doubleworld_to_virtual(fx, fy));
      input->add(stemp); 
      sprintf(stemp, "%.15g", (*fr->axarray)[1].doubleworld_to_virtual(fx, fy));
      input->add(stemp);
      sprintf(stemp, "%.15g", (*fr->axarray)[2].doubleworld_to_virtual(fx, fy));
      input->add(stemp); 
      sprintf(stemp, "%.15g", (*fr->axarray)[3].doubleworld_to_virtual(fx, fy));
      input->add(stemp); 
      sprintf(stemp, "%d", x);
      input->add(stemp); 
      sprintf(stemp, "%d", y);
      input->add(stemp); 
    }
  delete arr;

  if (key == LOCAL_LF_KEY)
    return true;
  else
    return false;
}

/****************************************************************/
/*******************  trace  ***********************************/
/****************************************************************/


void tracedisp(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  clearmenuwin();
  //wclear(win);

  wattron(menuwin, A_BOLD);
  wmove(win, 1, 2);
  wprintw(win, "%s\n", caption->c_str());
  wattroff(menuwin, A_BOLD);
  wmove(win, 2, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 3, 2);
  wprintw(win, "              LEFT (Ctrl+b)                   RIGHT (Ctrl+f)               ");
  wmove(win, 4, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 5, 2);
  wprintw(win, "                                                                           ");
  wmove(win, 6, 2);
  wprintw(win, "                     Space : toggle small/large skip                       ");
  wmove(win, 7, 2);
  wprintw(win, "                              CR, ESC: exit                                ");

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}


void showtracecor(WINDOW *win, double x, double y, int no)
{
  char stemp[300];
  int len;

  if (win == 0)
    return;

  wmove(win, 9, 2);
  wprintw(win, "                                                                           ");
  sprintf(stemp, "(%.15g, %.15g) No. = %d", x, y, no);
  len = strlen(stemp); len = len / 2; len = 39 - len;
  wmove(win, 9, len);
  wprintw(win, stemp);

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

int tracewindow(WINDOW *win, buffarray *input, std::string *caption, int datano)
{
  buffarray *arr;
  int key;
  int inc;
  int smallinc = 1;
  int largeinc = 100;
  bool skipstatus = 0;
  int type = 3;
  int x, y;
  frame *fr;
  data *dat;
  axis *ax; 
  char stemp1[100];
  double fx, fy, ftemp;
  int no, length;
  std::string temp_pre_formula;
  int refx, refy;

  if (win == 0)
    return false;

  fr = &(*gra->farray)[gra->currentframe];
  if (fr->darray->getarraylength() == 0)
    return false;

  if (datano < 0 || fr->darray->getarraylength() - 1 < datano)
    return false;

  dat = &(*fr->darray)[datano];

  length = dat->x->getarraylength();
  no = length / 2;

  if ((refx = fr->getaxisindex(dat->refaxis_x)) == -1)
    return false;
  if ((refy = fr->getaxisindex(dat->refaxis_y)) == -1)
    return false;

  inc = smallinc;

  arr = new buffarray;

  if (dat->trans_x)
    temp_pre_formula = *dat->format_x;
  else
    temp_pre_formula = std::string("x");
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fx, true ,dat, 0, fr->getdataindex(dat->id), 0);
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fx, false ,dat, no, fr->getdataindex(dat->id), 0);

  if (dat->trans_y)
    temp_pre_formula = *dat->format_y;
  else
    temp_pre_formula = std::string("y");
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fy, true ,dat, 0, fr->getdataindex(dat->id), 0);
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fy, false ,dat, no, fr->getdataindex(dat->id), 0);

  ax = &(*fr->axarray)[refx];
  ftemp = ax->virtual_to_doubleworld(fx).x;
  x = (int)fr->doubleworld_to_view_x(ftemp);
  ax = &(*fr->axarray)[refy];
  ftemp = ax->virtual_to_doubleworld(fy).y;
  y = (int)fr->doubleworld_to_view_y(ftemp);

  tracedisp(win, caption);
  while (1)
    {
      showtracecor(win, fx, fy, no);
      showpointer(arr, x, y, type);
      key = topazgetch(win); 
      showtracecor(win, fx, fy, no);
      showpointer(arr, x, y, type);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  if (no - inc <= 0)
	    no = 0;
	  else
	    no += -inc;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  if (no + inc >= length)
	    no = length - 1;
	  else
	    no += inc;
	    break;
	case CTRL('a'):
	    no = 0;
	    break;
	case CTRL('e'):
	    no = length - 1;
	    break;
	case ' ':
	  if (skipstatus == 1)
	    {
	      skipstatus = 0;
	      inc = smallinc;
	    }
	  else
	    {
	      skipstatus = 1;
	      inc = largeinc;
	    }
	  break;
	case LOCAL_LF_KEY:
	  key = LOCAL_LF_KEY;
	  goto j5;
	  
	}
      
      if (dat->trans_x)
	temp_pre_formula = *dat->format_x;
      else
	temp_pre_formula = std::string("x");
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fx, true ,dat, 0, fr->getdataindex(dat->id), 0);
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fx, false ,dat, no, fr->getdataindex(dat->id), 0);
      
      if (dat->trans_y)
	temp_pre_formula = *dat->format_y;
      else
	temp_pre_formula = std::string("y");
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fy, true ,dat, 0, fr->getdataindex(dat->id), 0);
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fy, false ,dat, no, fr->getdataindex(dat->id), 0);
      
      ax = &(*fr->axarray)[refx];
      ftemp = ax->virtual_to_doubleworld(fx).x;
      x = (int)fr->doubleworld_to_view_x(ftemp);
      ax = &(*fr->axarray)[refy];
      ftemp = ax->virtual_to_doubleworld(fy).y;
      y = (int)fr->doubleworld_to_view_y(ftemp);

      if (key == CTRL('g') || key == LOCAL_ESCAPE_KEY)
	{
	  key = LOCAL_ESCAPE_KEY;
	  goto j5;
	}

    }

 j5:;
  if (key == LOCAL_LF_KEY)
    {
      input->setbuff(0);
      sprintf(stemp1, "%.15g", fx);
      input->add(stemp1); 
      sprintf(stemp1, "%.15g", fy);
      input->add(stemp1); 
      sprintf(stemp1, "%d", no);
      input->add(stemp1); 
    }
  delete arr;

  if (key == LOCAL_LF_KEY)
    return true;
  else
    return false;
}


/****************************************************************/
/*******************  message  **********************************/
/****************************************************************/


void messagedisp(WINDOW *win, std::string *caption)
{
  char* p;
  char stemp[10];
  std::string str;
  buffarray *buff;
  buff = new buffarray;
  int len;

  if (win == 0)
    return;

  p = (char *)caption->c_str();
  str = std::string("");
  while (*p != 0)
    {
      if (*p == '\n')
	{
	  buff->add(str.c_str());
	  str = std::string("");
	}
      else
	{
	  stemp[0] = *p; stemp[1] = 0;
	  str += std::string(stemp);
	}
      p++;
    }
  buff->add(str.c_str());

  clearmenuwin();
  //wclear(win);

  for (int i = 0; i < buff->getarraylength(); i++)
    {
      len = strlen((*buff)[i].c_str()); len = len / 2; len = 39 - len;
      wmove(win, i + 3, len);
      wprintw(win, "%s\n", (*buff)[i].c_str());
    }

  wattron(win, A_REVERSE);
  wmove(win, 0,0);
  wprintw(win, " ");
  wmove(win, 0,0);
  wattroff(win, A_REVERSE);
  wrefresh(win);
}

void messagewindow(WINDOW *win, std::string *caption)
{
  if (win == 0)
    return;

  messagedisp(win, caption);
  topazgetch(win); 

  return;
}



/****************************************************************/
/*******************  selectitem ********************************/
/****************************************************************/

void showselectitem(WINDOW *win, buffarray *input, int addoffset)
{
  char mask[20], item[12];
  int x, y;
  int width = itemwidth + itemspace;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;

  if (win == 0)
    return;

  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");

  clearmenuwin();
  //wclear(win);
  
  for (int i = 1; i < input->getarraylength(); i++)
    {
      strncpy(item, (*input)[i].c_str(), sizeof(item));
      item[sizeof(item) - 1] = 0;

      keyflag = false;
      y = (i - 1) / maxrow;
      x = (i - 1) - maxrow * y;

      wmove(win, y + topmargin + addoffset, x * width + leftmargin);
      wprintw(win, mask);
      
      oset = (itemwidth - truetitlelength(item)) / 2;
      p = (char *)item;
      count = 0;

      while (*p != 0)
	{
	  if (*p != '&')
	    {
	      stemp[0] = *p; stemp[1] = 0;
	      wmove(win, y + topmargin + addoffset, x * width + leftmargin + oset + count++);
	      wprintw(win, stemp);

	      if (keyflag)
		{
		  wattroff(win, hotkeyattr);
		  keyflag = false;
		}
	    }
	  else
	    {
	      wattron(win, hotkeyattr);
	      keyflag = true;
	    }
	  p++;
	}

    }

  wrefresh(win);
}


void selectitemcursor(WINDOW *win, buffarray *input, int itemno, bool flag, int addoffset)
{
  char mask[20], item[12];
  int x, y;
  int width = itemwidth + itemspace;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;

  if (win == 0)
    return;

  y = (itemno - 1) / maxrow;
  x = (itemno - 1) - maxrow * y;

  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");

  if (flag)
      wattron(win, A_REVERSE);

  
  wmove(win, y + topmargin + addoffset, x * width + leftmargin);
  wprintw(win, mask);

  strncpy(item, (*input)[itemno].c_str(), sizeof(item));
  item[sizeof(item) - 1] = 0;

  oset = (itemwidth - truetitlelength(item)) / 2;
  p = (char *)item;
  count = 0;
  while (*p != 0)
    {
      if (*p != '&')
	{
	  stemp[0] = *p; stemp[1] = 0;
	  wmove(win, y + topmargin + addoffset, x * width + leftmargin + oset + count++);
	  wprintw(win, stemp);

	  if (keyflag)
	    {
	      if (!flag)
		wattroff(win, hotkeyattr);
	      keyflag = false;
	    }
	}
      else
	{
	  if (!flag)
	    wattron(win, hotkeyattr);
	  keyflag = true;
	}
      p++;
    }
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);

  if (flag)
      wattroff(win, A_REVERSE);

  wrefresh(win);

}


int selectitemmove (WINDOW *win, buffarray *input, int *itemno, int addoffset)
{
  int key, y;
  char *p;
  int maxitem = input->getarraylength() - 1;
  std::string str;
  script *scr;
  
  if (win == 0)
    return false;

  cbreak();
  noecho();
  if (*itemno < 1 || *itemno > maxitem)
    *itemno = 1;

  while (1)
    {

      y = *itemno;
      selectitemcursor(win, input, y, true, addoffset);
      key = topazgetch(win);
      selectitemcursor(win, input, y, false, addoffset);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  *itemno = (*itemno > 1) ? *itemno - 1: maxitem;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  *itemno = (*itemno < maxitem) ? *itemno + 1: 1;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  *itemno = (*itemno + maxrow <= maxitem) ? *itemno + maxrow: 1;
	  break;
	case KEY_UP:
	case CTRL('p'):
	  *itemno = (*itemno - maxrow >= 1) ? *itemno - maxrow: maxitem;
	  break;
	case CTRL('a'):
	  *itemno = 1;
	  break;
	case CTRL('e'):
	  *itemno = maxitem;
	  break;
	  //paint
	case CTRL('v'):
	case ' ':
	  scr = new script(gra);
	  str = std::string("$graph->paintall(V0);");
	  scr->execscript(&str, 0, 0);
	  delete scr;
	  break;
	  // return to top menu
	case CTRL('t'):
	  *itemno = -1;
	  return LOCAL_LF_KEY;
	  // scrolldown messagewindow
 	case '[':
	  msgb->previous(1);
	  break;
	  // scrollup messagewindow
	case ']':
	  msgb->next(1);
	  break;
	  // return to previous menu
	case CTRL('g'):
	  return LOCAL_ESCAPE_KEY;
	}

      if ( 'a' < key < 'z')
	{
	  for (int i = 1; i <= maxitem; i++)
	    {
	      p = (char *)(*input)[i].c_str();
	      while (*p != 0)
		{
		  if (*p == '&')
		    {
		      p++;
		      if ((*p == key) || ((*p - 'A' + 'a') == key))
			{
			  *itemno = i;
			  return (LOCAL_LF_KEY);
			}
		    }
		  p++;
		}
	    }
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }
  selectitemcursor(win, input, y, true, addoffset);

  return (key);
}

int selectitem(WINDOW *win, buffarray *input)
{
  int selected;
  std::string out;
  int addoffset = 2;

  if (win == 0)
    return false;

  if (input->getarraylength() < 1)
    return false;
  
  showselectitem(win, input, addoffset);

  // show message
  wattron(win, A_BOLD);
  wmove(win, topmargin, leftmargin);
  wprintw(win, "%s\n", (*input)[0].c_str());
  wattroff(win, A_BOLD);
  wrefresh(win);

  if (selectitemmove (win, input, &selected, addoffset) == LOCAL_LF_KEY)
    {
      out = (*input)[selected];
      input->setbuff(0);
      input->setscalar(out.c_str());
      return true;
    }
  else
    {
      out = std::string("");
      input->setbuff(0);
      input->setscalar(out.c_str());
      return false;
    }
  return false;
} 


/****************************************************************/
/*****************  selectlistitem ******************************/
/****************************************************************/

int selectitemundermargin = 1;

void showselectlistitem(WINDOW *win, buffarray *input, int addoffset, int startline)
{
  char mask[80], item[74];
  int  y;
  int count = 2;
  char stemp[10];
  char *p;

  if (win == 0)
    return;

  strcpy(mask, "");
  for (int i = 0; i < 74; i++)
    strcat(mask, " ");

  wclear(win);

  for (int i = 1; i < input->getarraylength(); i++)
    {
      if (startline <= i && i < startline + menuwinlines - topmargin - selectitemundermargin - addoffset)
	{
	  strncpy(item, (*input)[i].c_str(), sizeof(item));
	  item[linelength(item, 0, sizeof(item))] = 0;
	  y = (i - startline);
	  wmove(win, y + topmargin + addoffset, leftmargin);
	  wprintw(win, mask);
	  
	  p = (char *)item;
	  count = 0;
	  while (*p != 0)
	    {
	      stemp[0] = *p; stemp[1] = 0;
	      wmove(win, y + topmargin + addoffset, 3 + leftmargin + count++);
	      wprintw(win, stemp);
	      p++;
	    }
	  wmove(win, y + topmargin + addoffset, leftmargin);
	  sprintf(stemp, "%c:", i - startline + 'A');
	  wprintw(win, stemp);
	}
    }
  // show message
  wattron(win, A_BOLD);
  wmove(win, topmargin, leftmargin);
  wprintw(win, "%s\n", (*input)[0].c_str());
  wattroff(win, A_BOLD);
  wrefresh(win);

}


void selectlistitemcursor(WINDOW *win, buffarray *input, int relativeline, bool flag, int addoffset, int startline)
{
  char mask[80], item[74];
  int y, no;
  int count = 0;
  char stemp[10];
  char *p;

  if (win == 0)
    return;

  y = relativeline;
  no = startline + relativeline;

  strcpy(mask, "");
  for (int i = 0; i < 74; i++)
    strcat(mask, " ");

  if (flag)
      wattron(win, A_REVERSE);
  
  wmove(win, y + topmargin + addoffset, 3 + leftmargin);
  wprintw(win, mask);

  strncpy(item, (*input)[no].c_str(), sizeof(item));
  item[linelength(item, 0, sizeof(item))] = 0;

  p = (char *)item;
  count = 0;
  while (*p != 0)
    {
      stemp[0] = *p; stemp[1] = 0;
      wmove(win, y + topmargin + addoffset, 3 + leftmargin + count++);
      wprintw(win, stemp);
      p++;
    }
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);

  if (flag)
      wattroff(win, A_REVERSE);

  wrefresh(win);
}


int selectlistitemmove (WINDOW *win, buffarray *input, int *itemno, int addoffset)
{
  int key, y;
  int maxitem = input->getarraylength() - 1;
  std::string str;
  int startline = 2;
  int relativeline = 0;
  int length = input->getarraylength() - 1;
  
  if (win == 0)
    return false;

  cbreak();
  noecho();
  if (*itemno < 0 || *itemno > maxitem - 2)
    *itemno = 0;

  startline = *itemno / (menuwinlines - topmargin - selectitemundermargin - addoffset);
  startline *= menuwinlines - topmargin - selectitemundermargin - addoffset;
  startline += 2;
  relativeline = *itemno + 2 - startline;

  showselectlistitem(win, input, addoffset, startline);
  while (1)
    {

      y = relativeline;
      selectlistitemcursor(win, input, relativeline, true, addoffset, startline);
      key = topazgetch(win);
      selectlistitemcursor(win, input, relativeline, false, addoffset, startline);

      switch (key)
	{
	case KEY_UP:
	case CTRL('p'):
	  if ((startline + relativeline - 1) >= 2)
	    {
	      if (relativeline - 1 >= 0)
		  relativeline--;
	      else
		{
		  startline -= menuwinlines - topmargin - selectitemundermargin - addoffset;
		  relativeline = menuwinlines - topmargin - 1 - selectitemundermargin - addoffset;
		  showselectlistitem(win, input, addoffset, startline);
		}
	    }
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  if ((startline + relativeline + 1) <= length)
	    {
	      if (relativeline + 1 < menuwinlines - topmargin - 1 - addoffset)
		  relativeline++;
	      else
		{
		  startline += menuwinlines - topmargin - selectitemundermargin - addoffset;
		  relativeline = 0;
		  showselectlistitem(win, input, addoffset, startline);
		}
	    }
	  break;
	case CTRL('a'):
	  startline = 2; relativeline = 0;
	  showselectlistitem(win, input, addoffset, startline);
	  break;
	case CTRL('e'):
	  startline = (length - 2) / (menuwinlines - topmargin - selectitemundermargin - addoffset);
	  startline *= menuwinlines - topmargin - selectitemundermargin - addoffset;
	  startline += 2;
	  relativeline = length - startline;
	  showselectlistitem(win, input, addoffset, startline);
	  break;
	case KEY_NPAGE:
	case CTRL('v'):
	  if ((startline + relativeline + menuwinlines - topmargin - selectitemundermargin - addoffset) < length)
	    {
	      startline += menuwinlines - topmargin - selectitemundermargin - addoffset;
	      showselectlistitem(win, input, addoffset, startline);
	    }
	  else if ((startline + menuwinlines - topmargin - selectitemundermargin - addoffset) < length)
	    {
	      startline += menuwinlines - topmargin - selectitemundermargin - addoffset;
	      relativeline = 0;
	      showselectlistitem(win, input, addoffset, startline);
	    }
	  else
	    {
	      relativeline = length - startline;
	      showselectlistitem(win, input, addoffset, startline);
	    }
	  break;
	case KEY_PPAGE:
	  if ((startline  - (menuwinlines - topmargin - selectitemundermargin - addoffset)) >= 2)
	    {
	      startline -= menuwinlines - topmargin - selectitemundermargin - addoffset;
	      showselectlistitem(win, input, addoffset, startline);
	    }
	  else
	    {
	      startline = 2; relativeline = 0;
	      showselectlistitem(win, input, addoffset, startline);
	    }
	  break;
	case CTRL('g'):
	  return LOCAL_ESCAPE_KEY;
	}
      
      if ( 'a' <= key  && key <= ('a' + menuwinlines - topmargin - selectitemundermargin - addoffset))
	{
	  *itemno = key - 'a' + startline - 2;
	  return (LOCAL_LF_KEY);
	}

      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	{
	  *itemno = startline + relativeline - 2;
	  break;
	}
    }
  selectlistitemcursor(win, input, y, true, addoffset, startline);

  return (key);
}

int selectlistitem(WINDOW *win, buffarray *input)
{
  int selected;
  std::string out;
  int addoffset = 2;
  char stemp[100];

  if (win == 0)
    return false;

  if (input->getarraylength() < 3)
    return false;
  
  // show message
  wattron(win, A_BOLD);
  wmove(win, topmargin, leftmargin);
  wprintw(win, "%s\n", (*input)[0].c_str());
  wattroff(win, A_BOLD);
  wrefresh(win);

  selected = atoi((*input)[1].c_str());
  if (selected < 0 || selected > input->getarraylength() - 3)
    selected = 0;
    
    if (selectlistitemmove (win, input, &selected, addoffset) == LOCAL_LF_KEY)
    {
      sprintf(stemp, "%d", selected);
      out = std::string(stemp);
      input->setbuff(0);
      input->setscalar(out.c_str());
      return true;
    }
  else
    {
      out = std::string("");
      input->setbuff(0);
      input->setscalar(out.c_str());
      return false;
    }
  return false;
} 






