/*----------------------------------------------------------------------------
                        cui.h (cui interface definitions)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 99


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__cui_h)
#define __cui_h

#include <string>
#include "topazcurses.h"
#include "buffarray.h"

/*
   menu id definition
*/

#define M_SEP                  -1
#define M_MAIN                 0

#define M_FILE                 1
#define M_AXIS                 2
#define M_DATA                 3
#define M_PARTS                4
#define M_WINDOW               5
#define M_MACRO                6

#define M_QUIT                11
#define M_GRAPHOPEN           12
#define M_GRAPHSAVE           13
#define M_GRAPHSAVEOV         14
#define M_CHANGEDIR           15
#define M_PRINT               16
#define M_GRAPHNEW            17
#define M_GRAPHAPPEND         18
#define M_EXPORT              19
#define M_ABOUT               80
#define M_SHELL               81
#define M_HISTORY             82

#define M_FRAME               20
#define M_FRAME_X             21
#define M_FRAME_Y             22
#define M_FRAME_COLOR         23
#define M_FRAMEBODY_COLOR     24
#define M_FRAME_LENGTH_X      25
#define M_FRAME_LENGTH_Y      26
#define M_FRAMELINE_WIDTH     27
#define M_FRAMELINE_STYLE     28
#define M_FRAMEBODY_STYLE     29
#define M_FRAMETYPE           30
#define M_FRAMEDEL            31
#define M_FRAMENEW            32
#define M_FRAMEMOVE           33
#define M_FRAMECURRENT        34
#define M_FRAMESTRETCH        35
#define M_FRAMELIST           36
#define M_FRAMEORDER          37
#define M_FRAMEORDERUP        38
#define M_FRAMEORDERTOP       39
#define M_FRAMEORDERBOTTOM    40
#define M_FRAMEORDERDOWN      41
#define M_FRAMECLONE          42
#define M_FRAMEWITHDATA       43
#define M_AUTOSCALE           44
#define M_AUTOSCALE_ALL       45
#define M_AUTOSCALE_X         46
#define M_AUTOSCALE_Y         47
#define M_AUTOSCALE_T         48
#define M_AUTOSCALE_R         49
#define M_TRIMMING            50
#define M_WATCH               51
#define M_FRAMEEDIT           52
#define M_FRAMPAINTORDER      53


/*axis X*/
#define M_X                       100
#define M_X_RANGE                 101
#define M_X_MAJORTICK             102
#define M_X_MINORTICK             103
#define M_X_LABEL                 104
#define M_X_BASE                  105
#define M_X_DISP                  121
#define M_X_COMMONSET             122
#define M_X_SCALING               123
#define M_X_INVERT                124
#define M_X_ORIGINX               125
#define M_X_ORIGINY               126
#define M_X_LENGTH                127
#define M_X_ANGLE                 128
#define M_X_BASELINEWIDTH         129
#define M_X_BASELINECAP           130
#define M_X_BASELINEJOIN          131
#define M_X_BASELINESTYLE         132
#define M_X_BASELINECOLOR         133
#define M_X_START                 134
#define M_X_END                   135
#define M_X_UNIT                  136
#define M_X_CROSSPOINT            137
#define M_X_AUTOCROSSPOINT        138
#define M_X_TICKPARITY            139
#define M_X_TICKWIDTH             140
#define M_X_TICKLENGTH            141
#define M_X_DIVISION              142
#define M_X_MAJORTICKLINECOLOR    143
#define M_X_MINORTICKLINECOLOR    144
#define M_X_MAJORTICKLINEWIDTH    145
#define M_X_MAJORTICKLINECAP      146
#define M_X_MAJORTICKLINEJOIN     147
#define M_X_MAJORTICKLINESTYLE    148
#define M_X_MINORTICKLINEWIDTH    149
#define M_X_MINORTICKLINECAP      150
#define M_X_MINORTICKLINEJOIN     151
#define M_X_MINORTICKLINESTYLE    152
#define M_X_LABELFONTFACE         160             
#define M_X_LABELFONTSIZE         161             
#define M_X_LABELFONTANGLE        162             
#define M_X_LABELFORMAT           163             
#define M_X_LABELADDPLUS          164             
#define M_X_LABELPOSITION         165             
#define M_X_LABELSKIP             166             
#define M_X_LABELAUTOPRECISION    168             
#define M_X_LABELINTPRECISION     169             
#define M_X_LABELDECIMALPRECISION 170             
#define M_X_LABELCOLOR            171             
#define M_X_LABELPARITY           172             
#define M_X_CROSSPOINTPARITY      173             
#define M_X_BASETICKAREA_START    174             
#define M_X_BASETICKAREA_END      175             
#define M_X_CROSSPOINTOFFSET      176             

/*axis Y*/
#define M_Y                       200
#define M_Y_RANGE                 201
#define M_Y_MAJORTICK             202
#define M_Y_MINORTICK             203
#define M_Y_LABEL                 204
#define M_Y_BASE                  205
#define M_Y_DISP                  221
#define M_Y_COMMONSET             222
#define M_Y_SCALING               223
#define M_Y_INVERT                224
#define M_Y_ORIGINX               225
#define M_Y_ORIGINY               226
#define M_Y_LENGTH                227
#define M_Y_ANGLE                 228
#define M_Y_BASELINEWIDTH         229
#define M_Y_BASELINECAP           230
#define M_Y_BASELINEJOIN          231
#define M_Y_BASELINESTYLE         232
#define M_Y_BASELINECOLOR         233
#define M_Y_START                 234
#define M_Y_END                   235
#define M_Y_UNIT                  236
#define M_Y_CROSSPOINT            237
#define M_Y_AUTOCROSSPOINT        238
#define M_Y_TICKPARITY            239
#define M_Y_TICKWIDTH             240
#define M_Y_TICKLENGTH            241
#define M_Y_DIVISION              242
#define M_Y_MAJORTICKLINECOLOR    243
#define M_Y_MINORTICKLINECOLOR    244
#define M_Y_MAJORTICKLINEWIDTH    245
#define M_Y_MAJORTICKLINECAP      246
#define M_Y_MAJORTICKLINEJOIN     247
#define M_Y_MAJORTICKLINESTYLE    248
#define M_Y_MINORTICKLINEWIDTH    249
#define M_Y_MINORTICKLINECAP      250
#define M_Y_MINORTICKLINEJOIN     251
#define M_Y_MINORTICKLINESTYLE    252
#define M_Y_LABELFONTFACE         260             
#define M_Y_LABELFONTSIZE         261             
#define M_Y_LABELFONTANGLE        262             
#define M_Y_LABELFORMAT           263             
#define M_Y_LABELADDPLUS          264             
#define M_Y_LABELPOSITION         265             
#define M_Y_LABELSKIP             266             
#define M_Y_LABELAUTOPRECISION    268             
#define M_Y_LABELINTPRECISION     269             
#define M_Y_LABELDECIMALPRECISION 270             
#define M_Y_LABELCOLOR            271             
#define M_Y_LABELPARITY           272             
#define M_Y_CROSSPOINTPARITY      273             
#define M_Y_BASETICKAREA_START    274             
#define M_Y_BASETICKAREA_END      275             
#define M_Y_CROSSPOINTOFFSET      276             

/*axis T*/
#define M_T                       300
#define M_T_RANGE                 301
#define M_T_MAJORTICK             302
#define M_T_MINORTICK             303
#define M_T_LABEL                 304
#define M_T_BASE                  305
#define M_T_DISP                  321
#define M_T_COMMONSET             322
#define M_T_SCALING               323
#define M_T_INVERT                324
#define M_T_ORIGINX               325
#define M_T_ORIGINY               326
#define M_T_LENGTH                327
#define M_T_ANGLE                 328
#define M_T_BASELINEWIDTH         329
#define M_T_BASELINECAP           330
#define M_T_BASELINEJOIN          331
#define M_T_BASELINESTYLE         332
#define M_T_BASELINECOLOR         333
#define M_T_START                 334
#define M_T_END                   335
#define M_T_UNIT                  336
#define M_T_CROSSPOINT            337
#define M_T_AUTOCROSSPOINT        338
#define M_T_TICKPARITY            339
#define M_T_TICKWIDTH             340
#define M_T_TICKLENGTH            341
#define M_T_DIVISION              342
#define M_T_MAJORTICKLINECOLOR    343
#define M_T_MINORTICKLINECOLOR    344
#define M_T_MAJORTICKLINEWIDTH    345
#define M_T_MAJORTICKLINECAP      346
#define M_T_MAJORTICKLINEJOIN     347
#define M_T_MAJORTICKLINESTYLE    348
#define M_T_MINORTICKLINEWIDTH    349
#define M_T_MINORTICKLINECAP      350
#define M_T_MINORTICKLINEJOIN     351
#define M_T_MINORTICKLINESTYLE    352
#define M_T_LABELFONTFACE         360             
#define M_T_LABELFONTSIZE         361             
#define M_T_LABELFONTANGLE        362             
#define M_T_LABELFORMAT           363             
#define M_T_LABELADDPLUS          364             
#define M_T_LABELPOSITION         365             
#define M_T_LABELSKIP             366             
#define M_T_LABELAUTOPRECISION    368             
#define M_T_LABELINTPRECISION     369             
#define M_T_LABELDECIMALPRECISION 370             
#define M_T_LABELCOLOR            371             
#define M_T_LABELPARITY           372             
#define M_T_CROSSPOINTPARITY      373             
#define M_T_BASETICKAREA_START    374             
#define M_T_BASETICKAREA_END      375             
#define M_T_CROSSPOINTOFFSET      376             

/*axis R*/
#define M_R                       400
#define M_R_RANGE                 401
#define M_R_MAJORTICK             402
#define M_R_MINORTICK             403
#define M_R_LABEL                 404
#define M_R_BASE                  405
#define M_R_DISP                  421
#define M_R_COMMONSET             422
#define M_R_SCALING               423
#define M_R_INVERT                424
#define M_R_ORIGINX               425
#define M_R_ORIGINY               426
#define M_R_LENGTH                427
#define M_R_ANGLE                 428
#define M_R_BASELINEWIDTH         429
#define M_R_BASELINECAP           430
#define M_R_BASELINEJOIN          431
#define M_R_BASELINESTYLE         432
#define M_R_BASELINECOLOR         433
#define M_R_START                 434
#define M_R_END                   435
#define M_R_UNIT                  436
#define M_R_CROSSPOINT            437
#define M_R_AUTOCROSSPOINT        438
#define M_R_TICKPARITY            439
#define M_R_TICKWIDTH             440
#define M_R_TICKLENGTH            441
#define M_R_DIVISION              442
#define M_R_MAJORTICKLINECOLOR    443
#define M_R_MINORTICKLINECOLOR    444
#define M_R_MAJORTICKLINEWIDTH    445
#define M_R_MAJORTICKLINECAP      446
#define M_R_MAJORTICKLINEJOIN     447
#define M_R_MAJORTICKLINESTYLE    448
#define M_R_MINORTICKLINEWIDTH    449
#define M_R_MINORTICKLINECAP      450
#define M_R_MINORTICKLINEJOIN     451
#define M_R_MINORTICKLINESTYLE    452
#define M_R_LABELFONTFACE         460             
#define M_R_LABELFONTSIZE         461             
#define M_R_LABELFONTANGLE        462             
#define M_R_LABELFORMAT           463             
#define M_R_LABELADDPLUS          464             
#define M_R_LABELPOSITION         465             
#define M_R_LABELSKIP             466             
#define M_R_LABELAUTOPRECISION    468             
#define M_R_LABELINTPRECISION     469             
#define M_R_LABELDECIMALPRECISION 470             
#define M_R_LABELCOLOR            471             
#define M_R_LABELPARITY           472             
#define M_R_CROSSPOINTPARITY      473             
#define M_R_BASETICKAREA_START    474             
#define M_R_BASETICKAREA_END      475             
#define M_R_CROSSPOINTOFFSET      476             


/* macro */
#define M_MACROEXEC               600
#define M_MACROEXECFROMFILE       601
#define M_MACRORESISTERED         602


/* data */
#define M_DATAOPEN                700
#define M_PLOTMAPPING                 701
#define M_PLOTMAPPING_DISPLAY         702
#define M_PLOTMAPPING_TRANSX          703
#define M_PLOTMAPPING_TRANSFORMX      704
#define M_PLOTMAPPING_TRANSY          705
#define M_PLOTMAPPING_TRANSFORMY      706
#define M_PLOTMAPPING_REFAXISX        707
#define M_PLOTMAPPING_REFAXISY        708
#define M_PLOTMAPPING_CLIPPING        709
#define M_PLOTLINE                    720
#define M_PLOTLINE_STYLE              721
#define M_PLOTLINE_WIDTH              722
#define M_PLOTLINE_JOIN               723
#define M_PLOTLINE_CAP                724
#define M_PLOTLINE_COLOR              725
#define M_PLOTLINE_INTERPOLATE        726
#define M_PLOTLINE_INTERPOLATEDIV     727
#define M_PLOTMARKER                  740
#define M_PLOTMARKER_STYLE            741
#define M_PLOTMARKER_CENTERDOT        742
#define M_PLOTMARKER_SIZE             743
#define M_PLOTMARKER_EDGELINESTYLE    744
#define M_PLOTMARKER_EDGECOLOR        745
#define M_PLOTMARKER_BODYCOLOR        746
#define M_PLOTMARKER_BODYSTYLE        747
#define M_PLOTMARKER_BASEVALUE        748
#define M_PLOTMEMSET_MSTART           750
#define M_PLOTMEMSET_MEND             751
#define M_PLOTMEMSET_MDIV             752
#define M_PLOTMEMSET_MSCALING         753
#define M_PLOTMEMDATA                 754
#define M_LOADDATA                    760
#define M_LOADDATA_COLX               761
#define M_LOADDATA_COLXMIN            762
#define M_LOADDATA_COLXMAX            763
#define M_LOADDATA_COLY               764
#define M_LOADDATA_COLYMIN            765
#define M_LOADDATA_COLYMAX            766
#define M_LOADDATA_READSTART          767
#define M_LOADDATA_READEND            768
#define M_LOADDATA_READSKIP           769
#define M_LOADDATA_COMMENTLINEACTION  770
#define M_LOADDATA_BARXON             771
#define M_LOADDATA_BARYON             772
#define M_LOADDATA_SEARCHPATH         773
#define M_LOADDATA_EXTRADATA          774
#define M_DATALIST                    780
#define M_DATADELETE                  781
#define M_DATACORDER                  782
#define M_DATACLONE                   783
#define M_PLOTMARKER_EDGEWIDTH        784
#define M_DATAINFO                    785
#define M_DATATRACE                   786
#define M_LOADDATA_FILENAME           787
#define M_DATAEDIT                    788

/* parts */
#define M_PARTS_STYLE                 800
#define M_PARTS_TYPE                  801
#define M_PARTS_TEXT                  802
#define M_PARTS_FONTFACE              803
#define M_PARTS_FONTSIZE              804
#define M_PARTS_FONTANGLE             805
#define M_PARTS_LINESPACE             806
#define M_PARTS_LINECOLOR             807
#define M_PARTS_BODYCOLOR             808
#define M_PARTS_LINEWIDTH             809
#define M_PARTS_LINECAP               810
#define M_PARTS_LINEJOIN              811
#define M_PARTS_LINESTYLE             812
#define M_PARTS_BODYFILLRULE          813
#define M_PARTS_BODYFILLSTYLE         814
#define M_PARTS_LEGENDLENGTH          815
#define M_PARTS_INTERPORATE           816
#define M_PARTS_INTERPORATEDIV        817
#define M_PARTS_P1ARROW               818
#define M_PARTS_P2ARROW               819
#define M_PARTS_P1_X                  820
#define M_PARTS_P1_Y                  821
#define M_PARTS_P2_X                  822
#define M_PARTS_P2_Y                  823
#define M_PARTS_NODES                 824
#define M_PARTS_NEW                   840
#define M_PARTS_DEL                   841
#define M_PARTS_MOVE                  842
#define M_PARTS_MOVENODES             843
#define M_PARTS_INPUT                 844
#define M_PARTS_FONTCOLOR             845
#define M_PARTSLIST                   846
#define M_PARTSORDER                  847
#define M_PARTSCLONE                  848
#define M_PARTS_ARROWLENGTH           849
#define M_PARTS_ARROWHALFANGLE        850
#define M_PARTS_JPFONTFACE            851
#define M_PARTSEDIT                   852

#define M_DATA_CREATE                     900
#define M_DATA_CREATE_FUNC                901
#define M_DATA_CREATE_LINEAR              902
#define M_DATA_CREATE_NONLINEAR           903
#define M_DATA_CREATE_FUNC_START          910
#define M_DATA_CREATE_FUNC_END            911
#define M_DATA_CREATE_FUNC_DIV            912
#define M_DATA_CREATE_FUNC_SCALING        913
#define M_DATA_CREATE_FUNC_FUNCTION       914
#define M_DATA_CREATE_LINEAR_START        920
#define M_DATA_CREATE_LINEAR_END          921
#define M_DATA_CREATE_LINEAR_DIV          922
#define M_DATA_CREATE_LINEAR_SCALING      923
#define M_DATA_CREATE_LINEAR_FUNCTION     924
#define M_DATA_CREATE_LINEAR_STARTL       925
#define M_DATA_CREATE_LINEAR_ENDL         926
#define M_DATA_CREATE_LINEAR_FIXED        927
#define M_DATA_CREATE_LINEAR_FIXEDX       928
#define M_DATA_CREATE_LINEAR_FIXEDY       929
#define M_DATA_CREATE_NONLINEAR_START     940
#define M_DATA_CREATE_NONLINEAR_END       941
#define M_DATA_CREATE_NONLINEAR_DIV       942
#define M_DATA_CREATE_NONLINEAR_SCALING   943
#define M_DATA_CREATE_NONLINEAR_FUNCTION  944
#define M_DATA_CREATE_NONLINEAR_STARTL    945
#define M_DATA_CREATE_NONLINEAR_ENDL      946
#define M_DATA_CREATE_NONLINEAR_ALLOW     947
#define M_DATA_CREATE_NONLINEAR_INI_A     950
#define M_DATA_CREATE_NONLINEAR_INI_B     951
#define M_DATA_CREATE_NONLINEAR_INI_C     952
#define M_DATA_CREATE_NONLINEAR_INI_D     953
#define M_DATA_CREATE_NONLINEAR_INI_E     954
#define M_DATA_CREATE_NONLINEAR_INI_F     955
#define M_DATA_CREATE_NONLINEAR_INI_G     956
#define M_DATA_CREATE_NONLINEAR_INI_H     957
#define M_DATA_CREATE_NONLINEAR_INI_I     958
#define M_DATA_CREATE_NONLINEAR_FUN_A     960
#define M_DATA_CREATE_NONLINEAR_FUN_B     961
#define M_DATA_CREATE_NONLINEAR_FUN_C     962
#define M_DATA_CREATE_NONLINEAR_FUN_D     963
#define M_DATA_CREATE_NONLINEAR_FUN_E     964
#define M_DATA_CREATE_NONLINEAR_FUN_F     965
#define M_DATA_CREATE_NONLINEAR_FUN_G     966
#define M_DATA_CREATE_NONLINEAR_FUN_H     967
#define M_DATA_CREATE_NONLINEAR_FUN_I     968


/*  color id definition */
#define C_OTHERS          1000 
#define C_BLACK           1001 
#define C_WHITE           1002 
#define C_RED             1003 
#define C_GREEN           1004 
#define C_BLUE            1005 

/*  mag */
#define M_MAG             1200 
#define MAG_OTHERS        1201 
#define MAG_10            1202 
#define MAG_20            1203 
#define MAG_30            1204 
#define MAG_40            1205 
#define MAG_50            1206 
#define MAG_60            1207 
#define MAG_70            1208 
#define MAG_80            1209 
#define MAG_90            1210 
#define MAG_100           1211 
#define MAG_125           1212 
#define MAG_150           1213 
#define MAG_175           1214 
#define MAG_200           1215 
#define MAG_300           1216 
#define MAG_400           1217 


#define M_VIEWER_DRAW     1300
#define M_VIEWER_SHIFT    1301
#define M_MSGWIN_CLEAR    1302
#define M_WIN_HOME        1303


/* yes / no */
#define A_YES             2000 
#define A_NO              2001 

/* frame type */
#define FRAMETYPE_BOX     3000
#define FRAMETYPE_XY      3001
#define FRAMETYPE_IND     3002

/*tick direction */
#define TICK_NULL         3100
#define TICK_IN           3101
#define TICK_OUT          3102
#define TICK_BOTH         3103

/* scaling */
#define SCALING_LINEAR    3200
#define SCALING_LOG       3201
#define SCALING_INV       3202
#define SCALING_LOGLINEAR 3203
#define SCALING_INVLINEAR 3204

/* linecap */
#define LINECAP_BUTT      3300
#define LINECAP_ROUND     3301
#define LINECAP_PROJECTING 3302

/* linejoin */
#define LINEJOIN_MITER     3400
#define LINEJOIN_ROUND     3401
#define LINEJOIN_BEVEL     3402

/* linestyle */
#define LINESTYLE_SOLID             3500
#define LINESTYLE_DASH              3501
#define LINESTYLE_LONGDASH          3502
#define LINESTYLE_SHORTDASH         3503
#define LINESTYLE_DASHDOTDASH       3504
#define LINESTYLE_DASHDOTDOTDASH    3505
#define LINESTYLE_DOT               3506
#define LINESTYLE_NULL              3507

/* fillstyle */
#define FILLSTYLE_SOLID             3600
#define FILLSTYLE_NULL              3601

/* fillrule */
#define FILLRULE_EVENODD            3700
#define FILLRULE_WINDING            3701

/* commentline action */
#define M_COMMENT_ABORT             3800
#define M_COMMENT_IGNORE            3801
#define M_COMMENT_SEPARTION         3802

/* data search path */
#define M_PATH_ABSOLUTE             3810
#define M_PATH_RELATIVE             3811

/* parts type */
#define M_PARTS_TYPE_TEXT                3830
#define M_PARTS_TYPE_LINE                3831
#define M_PARTS_TYPE_POLYLINE            3832
#define M_PARTS_TYPE_RECTANGLE           3833
#define M_PARTS_TYPE_ELLIPSE             3834
#define M_PARTS_TYPE_CUT                 3835
#define M_PARTS_TYPE_LEGEND              3836

/* font face */
#define M_FONT_TIMESROMAN            3850 
#define M_FONT_TIMESBOLD             3851 
#define M_FONT_TIMESITALIC           3852 
#define M_FONT_TIMESBOLDITALIC       3853 
#define M_FONT_HELVETICA             3854 
#define M_FONT_HELVETICABOLD         3855 
#define M_FONT_HELVETICAOBLIQUE      3856 
#define M_FONT_HELVETICABOLDOBLIQUE  3857 
#define M_FONT_COURIER               3858 
#define M_FONT_COURIERBOLD           3859 
#define M_FONT_COURIEROBLIQUE        3860 
#define M_FONT_COURIERBOLDOBLIQUE    3861 
#define M_FONT_SYMBOL                3862

/* jp-font face */
#define M_FONT_MINCHO                3910 
#define M_FONT_GOTHIC                3914 

/* label parity */
#define LABEL_PARITY_PLUS            3900 
#define LABEL_PARITY_MINUS           3901 

/* label crosspoint parity */
#define LABEL_CPPARITY_ZERO          3920 
#define LABEL_CPPARITY_PLUS          3921 
#define LABEL_CPPARITY_MINUS         3922 
#define LABEL_CPPARITY_NULL          3923 

/* order list */
#define ORDERLIST_TOP                4000 
#define ORDERLIST_UP                 4001 
#define ORDERLIST_DOWN               4002 
#define ORDERLIST_BOTTOM             4003 

/* order list */
#define M_PAPERSIZE                  1304 
#define PAPERSIZE_A4P                4101 
#define PAPERSIZE_A4L                4102 
#define PAPERSIZE_A3P                4103 
#define PAPERSIZE_A3L                4104 
#define PAPERSIZE_A5P                4105 
#define PAPERSIZE_A5L                4106 
#define PAPERSIZE_B4P                4107 
#define PAPERSIZE_B4L                4108 
#define PAPERSIZE_B5P                4109 
#define PAPERSIZE_B5L                4110 
#define PAPERSIZE_LETTERP            4111 
#define PAPERSIZE_LETTERL            4112 

/* memory data scaling */
#define M_MSCALING_LINEAR              4300
#define M_MSCALING_LOG                 4301
#define M_MSCALING_INV                 4302

/* linear fit function */
#define M_LINEARFIT_POLY0            4400
#define M_LINEARFIT_POLY1            4401
#define M_LINEARFIT_POLY2            4402
#define M_LINEARFIT_POLY3            4403
#define M_LINEARFIT_POLY4            4404
#define M_LINEARFIT_POLY5            4405
#define M_LINEARFIT_POLY6            4406
#define M_LINEARFIT_POLY7            4407
#define M_LINEARFIT_POLY8            4408
#define M_LINEARFIT_POLY9            4409
#define M_LINEARFIT_POLY10           4410
#define M_LINEARFIT_EXP              4411
#define M_LINEARFIT_LOG              4412
#define M_LINEARFIT_POW1             4413
#define M_LINEARFIT_POW2             4414




#define WINDOWCOL         78

typedef struct cuimenu
{
  int id;                        /* menu id */
  struct cuimenu *submenu;       /* pointer to submenu*/
  char *item;                    /* caption */
  int ismacro;                   /* macro or execute */
  char *macroscript;             /* macroscript */
  char *getvalue;                 /* macroscript to show value*/
}CUIMENU;

/* execute menu item */
void execitem(CUIMENU *menu, int *selectedkey);

/* window for editing text */
int edittextwindow(WINDOW *win, std::string *input, std::string *caption, bool editfilename);
int textlineedit(WINDOW *win, std::string *input, int y, int x, int displength, bool editfilename); 

/* selection window */
int execselect(CUIMENU *menu, std::string *input, std::string *otherscaption);
int execcolorselect(std::string *input, std::string *otherscaption);
int exec_yesno_select(std::string *input, std::string *otherscaption);
int frametype_select(std::string *input, std::string *otherscaption);
int tickdirection_select(std::string *input, std::string *otherscaption);
int scaling_select(std::string *input, std::string *otherscaption);
int mscaling_select(std::string *input, std::string *otherscaption);
int linecap_select(std::string *input, std::string *otherscaption);
int linejoin_select(std::string *input, std::string *otherscaption);
int linestyle_select(std::string *input, std::string *otherscaption);
int fillstyle_select(std::string *input, std::string *otherscaption);
int fillrule_select(std::string *input, std::string *otherscaption);
int mag_select(std::string *input, std::string *otherscaption);
int interpolate_select(std::string *input, std::string *otherscaption);
int markerstyle_select(std::string *input, std::string *otherscaption);
int commentlineaction_select(std::string *input, std::string *otherscaption);
int datasearchpath_select(std::string *input, std::string *otherscaption);
int partstype_select(std::string *input, std::string *otherscaption);
int fontface_select(std::string *input, std::string *otherscaption);
int jpfontface_select(std::string *input, std::string *otherscaption);
int partsinterpolation_select(std::string *input, std::string *otherscaption);
int labelparity_select(std::string *input, std::string *otherscaption);
int labelCPparity_select(std::string *input, std::string *otherscaption);
int labelformat_select(std::string *input, std::string *otherscaption);
int orderlist_select(std::string *input, std::string *otherscaption);
int papersizelist_select(std::string *input, std::string *otherscaption);
int linearfitfunction_select(std::string *input, std::string *otherscaption);
int framemenu();
int xbasemenu();
int xrangemenu();
int xtickmenu();
int xsubtickmenu();
int xlabelmenu();
int ybasemenu();
int yrangemenu();
int ytickmenu();
int ysubtickmenu();
int ylabelmenu();
int tbasemenu();
int trangemenu();
int ttickmenu();
int tsubtickmenu();
int tlabelmenu();
int rbasemenu();
int rrangemenu();
int rtickmenu();
int rsubtickmenu();
int rlabelmenu();

int plotmappingmenu();
int plotlinemenu();
int plotmarkermenu();
int plotmemdatamenu();
int loaddatamenu();
int partsstylemenu();
int textstylemenu();
int functionmenu();
int linearfitmenu();
int nonlinearfitmenu();




int movewindow(WINDOW *win, buffarray *input, std::string *caption);
int stretchwindow(WINDOW *win, buffarray *input, std::string *caption);
void printtext(WINDOW *win, char *input);
int changenodeswindow(WINDOW *win, buffarray *input, std::string *caption);
int inputpartswindow(WINDOW *win, buffarray *input, std::string *caption);
int trimmingwindow(WINDOW *win, buffarray *input, std::string *caption);
int watchwindow(WINDOW *win, buffarray *input, std::string *caption);
int tracewindow(WINDOW *win, buffarray *input, std::string *caption, int datano);
void messagewindow(WINDOW *win, std::string *caption);
int selectitem(WINDOW *win, buffarray *input);
int selectlistitem(WINDOW *win, buffarray *input);

#define CTRL(ch)        ((ch)&0x1F)

/*
#define LOCAL_ESCAPE_KEY KEY_BACKSPACE //27
*/
/*
#define LOCAL_ESCAPE_KEY 0411
*/
#define LOCAL_ESCAPE_KEY 27
#define LOCAL_LF_KEY 10

#define MENU_LEFT        1000
#define MENU_RIGHT       1001
#define MENU_UP          1002
#define MENU_DOWN        1003

#endif


