/*----------------------------------------------------------------------------
                             data.cc (data object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <string>
#include "data.h"
#include "gdi.h"
#include "vectdefs.h"
#include "pointdefs.h"
#include "filepoint.h"
#include "graph.h"
#include "frame.h"
#include "topazvalues.h"
#include "convtxt.h"
#include "paths.h"

extern void message(const char *msg);
extern void StyleLinePath(bool flag, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);
extern char *toshortfilename(char *in);
extern char confpath[1000];
extern char tmppath[1000];
extern char pid[100];
extern FileHandleArray *fhvuff;
extern bool genxdata(double start, double end, int div, int scaling, earray* outx, earray* outy);
extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);

data::data():plobj()
{
  /* plot style */
  version             = 2;
  id                  = 0;
  label               = new std::string("");
  virgin              = true;
  selected            = false;
  display             = true;  
  trans_x             = false;
  format_x            = new std::string("");
  trans_y             = false;
  format_y            = new std::string("");
  linecolor.setblack();
  markeredgecolor.setblack();
  markerbodycolor.setwhite();
  linestyle.init();
  markeredgestyle.init();
  markerbodyfillstyle.init();
  interpolate         = INTER_POLYLINE;
  interpolatediv      = 50;
  markerstyle         = MARKER_SQUARE;
  markercenterdot     = false;
  markersize          = 100; 
  clipping            = true;
  refaxis_x           = 0;
  refaxis_y           = 1;
  basevalue           = 0.0;
  m_start              = 0.0;
  m_end                = 10.0;
  m_div                = 10;
  m_scaling            = MSCALING_LINEAR;

  /* read data */
  col_x               = 1;
  col_y               = 2;
  col_xmin            = 1;
  col_xmax            = 1;
  col_ymin            = 1;
  col_ymax            = 1;
  readstart           = 1;
  readend             = -1;
  skip                = 1;
  commentlineaction   = COMMENT_ABORT;
  bar_x               = false;
  bar_y               = false;
  datasearchpath      = PATH_RELATIVE;
  for (int i = 0; i < 99; i++)
    extradatadispflag[i] = false;
  extradata           = new std::string("");

  /* data */
  x                   = new earray;
  y                   = new earray;
  xmax                = new earray;
  xmin                = new earray;
  ymax                = new earray;
  ymin                = new earray;;
  for (int i = 0; i < 99; i++)
    extra[i]          = new earray;
  bound_xmin          = MAXDOUBLE;
  bound_xmax          = -MAXDOUBLE;
  bound_ymin          = MAXDOUBLE;
  bound_ymax          = -MAXDOUBLE;
  strcpy(filename, "");
  isfiledata          = true;
  info                = new std::string("");
}

data::~data()
{
  delete label;

  /* plot style */
  delete format_x;
  delete format_y;

  /* read data */
  delete extradata;

  /* data */
  delete x;
  delete y;
  delete xmax;
  delete xmin;
  delete ymax;
  delete ymin;
  for (int i = 0; i < 99; i++)
    delete extra[i];
  delete info;
  
}

data& data::operator = (const data&  da)
{
//  parent              = da.parent;

  /* plot style */
//  id                  = da.id;
  version             = da.version;
  virgin              = da.virgin;
  selected            = da.selected;
  display             = da.display;  
  trans_x             = da.trans_x;
  *format_x           = *da.format_x;
  trans_y             = da.trans_y;
  *format_y           = *da.format_y;
  linecolor           = da.linecolor;
  markeredgecolor     = da.markeredgecolor;
  markerbodycolor     = da.markerbodycolor;
  linestyle           = da.linestyle;
  markeredgestyle     = da.markeredgestyle;
  markerbodyfillstyle = da.markerbodyfillstyle;
  interpolate         = da.interpolate;
  interpolatediv      = da.interpolatediv;
  markerstyle         = da.markerstyle;
  markercenterdot     = da.markercenterdot;
  markersize          = da.markersize; 
  clipping            = da.clipping;
  refaxis_x           = da.refaxis_x;
  refaxis_y           = da.refaxis_y;
  basevalue           = da.basevalue;
  m_start              = da.m_start;
  m_end                = da.m_end;
  m_div                = da.m_div;
  m_scaling            = da.m_scaling;

  // read data
  col_x               = da.col_x;
  col_y               = da.col_y;
  col_xmin            = da.col_xmin;
  col_xmax            = da.col_xmax;
  col_ymin            = da.col_ymin;
  col_ymax            = da.col_ymax;
  readstart           = da.readstart;
  readend             = da.readend;
  skip                = da.skip;
  commentlineaction   = da.commentlineaction;
  bar_x               = da.bar_x;
  bar_y               = da.bar_y;
  datasearchpath      = da.datasearchpath;
  for (int i = 0; i < 99; i++)
    extradatadispflag[i] = da.extradatadispflag[i];
  *extradata          = *da.extradata;

  x->setbuff(0);
  for(int i = 0; i < da.x->getarraylength(); i++)
    {
      x->add(new dataelement);
      (*x)[i] = (*da.x)[i];
    }
  y->setbuff(0);
  for(int i = 0; i < da.y->getarraylength(); i++)
    {
      y->add(new dataelement);
      (*y)[i] = (*da.y)[i];
    }
  xmin->setbuff(0);
  for(int i = 0; i < da.xmin->getarraylength(); i++)
    {
      xmin->add(new dataelement);
      (*xmin)[i] = (*da.xmin)[i];
    }
  xmax->setbuff(0);
  for(int i = 0; i < da.xmax->getarraylength(); i++)
    {
      xmax->add(new dataelement);
      (*xmax)[i] = (*da.xmax)[i];
    }
  ymin->setbuff(0);
  for(int i = 0; i < da.ymin->getarraylength(); i++)
    {
      ymin->add(new dataelement);
      (*ymin)[i] = (*da.ymin)[i];
    }
  ymax->setbuff(0);
  for(int i = 0; i < da.ymax->getarraylength(); i++)
    {
      ymax->add(new dataelement);
      (*ymax)[i] = (*da.ymax)[i];
    }

  for (int i = 0; i < 99; i++)
    {
      extra[i]->setbuff(0);
      for (int j = 0; j < da.extra[i]->getarraylength(); j++)
	{
	  extra[i]->add(new dataelement);
	  (*(extra[i]))[j] = (*(da.extra[i]))[j];
	}
    }

  bound_xmin          = da.bound_xmin;
  bound_xmax          = da.bound_xmax;
  bound_ymin          = da.bound_ymin;
  bound_ymax          = da.bound_ymax;
  strcpy(filename, da.filename);

  isfiledata          = da.isfiledata;
  *info               = *da.info;

  return *this;
}

bool data::ispatternequal(data* dat1, data* dat2)
{
  return (
	  dat1->linecolor           == dat2->linecolor             &&
	  dat1->markeredgecolor     == dat2->markeredgecolor       &&
	  dat1->markerbodycolor     == dat2->markerbodycolor       &&
	  dat1->linestyle           == dat2->linestyle             &&
	  dat1->markeredgestyle     == dat2->markeredgestyle       &&
	  dat1->markerbodyfillstyle == dat2->markerbodyfillstyle   &&
	  dat1->markerstyle         == dat2->markerstyle           &&
	  dat1->markercenterdot     == dat2->markercenterdot       &&
	  dat1->markersize          == dat2->markersize
	  );
}


bool data::get(std::string* member, tokenbuff* outvalue)
{
  std::string out, shrunk;
  int iid;
  char stemp[1000];
  int itemp;
  std::string temp_pre_formula;
  double ftemp;
  frame* fr = (frame *)parent;

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  sprintf(stemp, "%d", id);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  outvalue->value->add(label->c_str());
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  sprintf(stemp, "%d", selected);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("show") && iid == -1)
	{
	  sprintf(stemp, "%d", display);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("virgin") && iid == -1)
	{
	  sprintf(stemp, "%d", virgin);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("trans_x") && iid == -1)
	{
	  sprintf(stemp, "%d", trans_x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("format_x") && iid == -1)
	{
	  outvalue->value->add(format_x->c_str());
	  return true;
	}
      else if (out == std::string("trans_y") && iid == -1)
	{
	  sprintf(stemp, "%d", trans_y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("format_y") && iid == -1)
	{
	  outvalue->value->add(format_y->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", markeredgecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", markeredgecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", markeredgecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerbodycolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", markerbodycolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerbodycolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", markerbodycolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerbodycolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", markerbodycolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecap") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", markeredgestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgecap") && iid == -1)
	{
	  sprintf(stemp, "%d", markeredgestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", markeredgestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markeredgestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", markeredgestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerfillstyle") && iid == -1)
	{
	  sprintf(stemp, "%d", markerbodyfillstyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerfillrule") && iid == -1)
	{
	  sprintf(stemp, "%d", markerbodyfillstyle.fillrule);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("interpolate") && iid == -1)
	{
	  sprintf(stemp, "%d", interpolate);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("interpolatediv") && iid == -1)
	{
	  sprintf(stemp, "%d", interpolatediv);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markerstyle") && iid == -1)
	{
	  sprintf(stemp, "%d", markerstyle);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markercenterdot") && iid == -1)
	{
	  sprintf(stemp, "%d", markercenterdot);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("markersize") && iid == -1)
	{
	  sprintf(stemp, "%d", markersize);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("clipping") && iid == -1)
	{
	  sprintf(stemp, "%d", clipping);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("refaxis_x") && iid == -1)
	{
	  sprintf(stemp, "%d", refaxis_x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("refaxis_y") && iid == -1)
	{
	  sprintf(stemp, "%d", refaxis_y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("basevalue") && iid == -1)
	{
	  sprintf(stemp, "%.15g", basevalue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("m_start") && iid == -1)
	{
	  sprintf(stemp, "%.15g", m_start);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("m_end") && iid == -1)
	{
	  sprintf(stemp, "%.15g", m_end);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("m_div") && iid == -1)
	{
	  sprintf(stemp, "%d", m_div);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("m_scaling") && iid == -1)
	{
	  sprintf(stemp, "%d", m_scaling);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_x") && iid == -1)
	{
	  sprintf(stemp, "%d", col_x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_xmin") && iid == -1)
	{
	  sprintf(stemp, "%d", col_xmin);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_xmax") && iid == -1)
	{
	  sprintf(stemp, "%d", col_xmax);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_y") && iid == -1)
	{
	  sprintf(stemp, "%d", col_y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_ymin") && iid == -1)
	{
	  sprintf(stemp, "%d", col_ymin);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("col_ymax") && iid == -1)
	{
	  sprintf(stemp, "%d", col_ymax);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("readstart") && iid == -1)
	{
	  sprintf(stemp, "%d", readstart);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("readend") && iid == -1)
	{
	  sprintf(stemp, "%d", readend);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("skip") && iid == -1)
	{
	  sprintf(stemp, "%d", skip);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("commentlineaction") && iid == -1)
	{
	  sprintf(stemp, "%d", commentlineaction);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bar_x") && iid == -1)
	{
	  sprintf(stemp, "%d", bar_x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bar_y") && iid == -1)
	{
	  sprintf(stemp, "%d", bar_y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("datasearchpath") && iid == -1)
	{
	  sprintf(stemp, "%d", datasearchpath);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("extradatadispflag") && 1 <= iid && iid <= 99)
	{
	  sprintf(stemp, "%d", extradatadispflag[iid - 1]);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("extradata") && iid == -1)
	{
	  sprintf(stemp, "%s", extradata->c_str());
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("X") && 0 <= iid)
	{
	  if (iid < x->getarraylength())
	    {
	      if (trans_x)
		temp_pre_formula = *format_x;
	      else
		temp_pre_formula = std::string("x");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 0);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 0);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("XMIN") && 0 <= iid)
	{
	  if (iid < xmin->getarraylength())
	    {
	      if (trans_x)
		temp_pre_formula = *format_x;
	      else
		temp_pre_formula = std::string("x");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 1);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 1);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("XMAX") && 0 <= iid)
	{
	  if (iid < xmax->getarraylength())
	    {
	      if (trans_x)
		temp_pre_formula = *format_x;
	      else
		temp_pre_formula = std::string("x");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 2);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 2);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("Y") && 0 <= iid)
	{
	  if (iid < y->getarraylength())
	    {
	      if (trans_y)
		temp_pre_formula = *format_y;
	      else
		temp_pre_formula = std::string("y");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 0);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 0);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("YMIN") && 0 <= iid)
	{
	  if (iid < ymin->getarraylength())
	    {
	      if (trans_y)
		temp_pre_formula = *format_y;
	      else
		temp_pre_formula = std::string("y");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 1);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 1);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("YMAX") && 0 <= iid)
	{
	  if (iid < ymax->getarraylength())
	    {
	      if (trans_y)
		temp_pre_formula = *format_y;
	      else
		temp_pre_formula = std::string("y");
	      
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, true ,this, 0, fr->getdataindex(id), 2);
	      CalcFormulastring((char *)temp_pre_formula.c_str(), &ftemp, false ,this, iid, fr->getdataindex(id), 2);

	      sprintf(stemp, "%.15g", ftemp);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("x") && 0 <= iid)
	{
	  if (iid < x->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*x)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmin") && 0 <= iid)
	{
	  if (iid < xmin->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*xmin)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmax") && 0 <= iid)
	{
	  if (iid < xmax->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*xmax)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("y") && 0 <= iid)
	{
	  if (iid < y->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*y)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymin") && 0 <= iid)
	{
	  if (iid < ymin->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*ymin)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymax") && 0 <= iid)
	{
	  if (iid < ymax->getarraylength())
	    {
	      sprintf(stemp, "%.15g", (*ymax)[iid].data);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (
	       (out.c_str())[0] == 'd'
	       && ('0' <= (out.c_str())[1] && (out.c_str())[1] <= 9)
	       && ('0' <= (out.c_str())[2] && (out.c_str())[2] <= 9)
	       && (out.c_str())[3] == 0
	       && 0 <= iid)
	{
	  itemp = 10 * ((out.c_str())[1] + out.c_str())[2] - 1;
	  if (itemp >= 0 && itemp < 99)
	    {
	      if (0 <= iid && iid < extra[itemp]->getarraylength())
		{
		  sprintf(stemp, "%.15g", (*extra[itemp])[iid].data);
		  outvalue->value->add(stemp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("x_attr") && 0 <= iid)
	{
	  if (iid < x->getarraylength())
	    {
	      sprintf(stemp, "%d", (*x)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmin_attr") && 0 <= iid)
	{
	  if (iid < xmin->getarraylength())
	    {
	      sprintf(stemp, "%d", (*xmin)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmax_attr") && 0 <= iid)
	{
	  if (iid < xmax->getarraylength())
	    {
	      sprintf(stemp, "%d", (*xmax)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("y_attr") && 0 <= iid)
	{
	  if (iid < y->getarraylength())
	    {
	      sprintf(stemp, "%d", (*y)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymin_attr") && 0 <= iid)
	{
	  if (iid < ymin->getarraylength())
	    {
	      sprintf(stemp, "%d", (*ymin)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymax_attr") && 0 <= iid)
	{
	  if (iid < ymax->getarraylength())
	    {
	      sprintf(stemp, "%d", (*ymax)[iid].status);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (
	       (out.c_str())[0] == 'a'
	       && ('0' <= (out.c_str())[1] && (out.c_str())[1] <= 9)
	       && ('0' <= (out.c_str())[2] && (out.c_str())[2] <= 9)
	       && (out.c_str())[3] == 0
	       && 0 <= iid)
	{
	  itemp = 10 * ((out.c_str())[1] + out.c_str())[2] - 1;
	  if (itemp >= 0 && itemp < 99)
	    {
	      if (0 <= iid && iid < extra[itemp]->getarraylength())
		{
		  sprintf(stemp, "%d", (*extra[itemp])[iid].status);
		  outvalue->value->add(stemp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("bound_xmin") && iid == -1)
	{
	  sprintf(stemp, "%.15g", bound_xmin);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bound_xmax") && iid == -1)
	{
	  sprintf(stemp, "%.15g", bound_xmax);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bound_ymin") && iid == -1)
	{
	  sprintf(stemp, "%.15g", bound_ymin);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bound_ymax") && iid == -1)
	{
	  sprintf(stemp, "%.15g", bound_ymax);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("filename") && iid == -1)
	{
	  sprintf(stemp, "%s", filename);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("isfiledata") && iid == -1)
	{
	  sprintf(stemp, "%d", isfiledata);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("info") && iid == -1)
	{
	  outvalue->value->add(info->c_str());
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}

bool data::set(std::string* member, std::string* setvalue)
{
  std::string out, shrunk;
  int iid;
  int itemp;

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  id = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  *label = *setvalue;
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  selected = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("show") && iid == -1)
	{
	  display = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("virgin") && iid == -1)
	{
	  virgin = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("trans_x") && iid == -1)
	{
	  trans_x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("format_x") && iid == -1)
	{
	  *format_x = *setvalue;
	  return true;
	}
      else if (out == std::string("trans_y") && iid == -1)
	{
	  trans_y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("format_y") && iid == -1)
	{
	  *format_y = *setvalue;
	  return true;
	}
      else if (out == std::string("linecolor_red") && iid == -1)
	{
	  linecolor.red = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_green") && iid == -1)
	{
	  linecolor.green = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_blue") && iid == -1)
	{
	  linecolor.blue = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgecolor_red") && iid == -1)
	{
	  markeredgecolor.red = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgecolor_green") && iid == -1)
	{
	  markeredgecolor.green = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgecolor_blue") && iid == -1)
	{
	  markeredgecolor.blue = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerbodycolor_red") && iid == -1)
	{
	  markerbodycolor.red = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerbodycolor_green") && iid == -1)
	{
	  markerbodycolor.green = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerbodycolor_blue") && iid == -1)
	{
	  markerbodycolor.blue = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linewidth") && iid == -1)
	{
	  linestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecap") && iid == -1)
	{
	  linestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linejoin") && iid == -1)
	{
	  linestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linestyle") && iid == -1)
	{
	  linestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgewidth") && iid == -1)
	{
	  markeredgestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgecap") && iid == -1)
	{
	  markeredgestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgejoin") && iid == -1)
	{
	  markeredgestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markeredgestyle") && iid == -1)
	{
	  markeredgestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerfillstyle") && iid == -1)
	{
	  markerbodyfillstyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerfillrule") && iid == -1)
	{
	  markerbodyfillstyle.fillrule = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("interpolate") && iid == -1)
	{
	  interpolate = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("interpolatediv") && iid == -1)
	{
	  interpolatediv = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markerstyle") && iid == -1)
	{
	  markerstyle = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markercenterdot") && iid == -1)
	{
	  markercenterdot = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("markersize") && iid == -1)
	{
	  markersize = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("clipping") && iid == -1)
	{
	  clipping = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("refaxis_x") && iid == -1)
	{
	  refaxis_x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("refaxis_y") && iid == -1)
	{
	  refaxis_y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("basevalue") && iid == -1)
	{
	  basevalue = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("m_start") && iid == -1)
	{
	  m_start = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("m_end") && iid == -1)
	{
	  m_end = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("m_div") && iid == -1)
	{
	  m_div = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("m_scaling") && iid == -1)
	{
	  m_scaling = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_x") && iid == -1)
	{
	  col_x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_xmin") && iid == -1)
	{
	  col_xmin = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_xmax") && iid == -1)
	{
	  col_xmax = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_y") && iid == -1)
	{
	  col_y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_ymin") && iid == -1)
	{
	  col_ymin = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("col_ymax") && iid == -1)
	{
	  col_ymax = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("readstart") && iid == -1)
	{
	  readstart = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("readend") && iid == -1)
	{
	  readend = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("skip") && iid == -1)
	{
	  skip = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("commentlineaction") && iid == -1)
	{
	  commentlineaction = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bar_x") && iid == -1)
	{
	  bar_x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bar_y") && iid == -1)
	{
	  bar_y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("datasearchpath") && iid == -1)
	{
	  datasearchpath = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("extradatadispflag") && 1 <= iid && iid <= 99)
	{
	  extradatadispflag[iid - 1] = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("extradata") && iid == -1)
	{
	  *extradata = *setvalue;
	  return true;
	}
      else if (out == std::string("x") && 0 <= iid)
	{
	  if (iid < x->getarraylength())
	    {
	      (*x)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmin") && 0 <= iid)
	{
	  if (iid < xmin->getarraylength())
	    {
	      (*xmin)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmax") && 0 <= iid)
	{
	  if (iid < xmax->getarraylength())
	    {
	      (*xmax)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("y") && 0 <= iid)
	{
	  if (iid < y->getarraylength())
	    {
	      (*y)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymin") && 0 <= iid)
	{
	  if (iid < ymin->getarraylength())
	    {
	      (*ymin)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymax") && 0 <= iid)
	{
	  if (iid < ymax->getarraylength())
	    {
	      (*ymax)[iid].data = atof(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (
	       (out.c_str())[0] == 'd'
	       && ('0' <= (out.c_str())[1] && (out.c_str())[1] <= 9)
	       && ('0' <= (out.c_str())[2] && (out.c_str())[2] <= 9)
	       && (out.c_str())[3] == 0
	       && 0 <= iid)
	{
	  itemp = 10 * ((out.c_str())[1] + out.c_str())[2] - 1;
	  if (itemp >= 0 && itemp < 99)
	    {
	      if (0 <= iid && iid < extra[itemp]->getarraylength())
		{
		  (*extra[itemp])[iid].data = atof(setvalue->c_str());
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("x_attr") && 0 <= iid)
	{
	  if (iid < x->getarraylength())
	    {
	      (*x)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmin_attr") && 0 <= iid)
	{
	  if (iid < xmin->getarraylength())
	    {
	      (*xmin)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("xmax_attr") && 0 <= iid)
	{
	  if (iid < xmax->getarraylength())
	    {
	      (*xmax)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("y_attr") && 0 <= iid)
	{
	  if (iid < y->getarraylength())
	    {
	      (*y)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymin_attr") && 0 <= iid)
	{
	  if (iid < ymin->getarraylength())
	    {
	      (*ymin)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("ymax_attr") && 0 <= iid)
	{
	  if (iid < ymax->getarraylength())
	    {
	      (*ymax)[iid].status = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (
	       (out.c_str())[0] == 'a'
	       && ('0' <= (out.c_str())[1] && (out.c_str())[1] <= 9)
	       && ('0' <= (out.c_str())[2] && (out.c_str())[2] <= 9)
	       && (out.c_str())[3] == 0
	       && 0 <= iid)
	{
	  itemp = 10 * ((out.c_str())[1] + out.c_str())[2] - 1;
	  if (itemp >= 0 && itemp < 99)
	    {
	      if (0 <= iid && iid < extra[itemp]->getarraylength())
		{
		  (*extra[itemp])[iid].status = atoi(setvalue->c_str());
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("bound_xmin") && iid == -1)
	{
	  bound_xmin = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bound_xmax") && iid == -1)
	{
	  bound_xmax = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bound_ymin") && iid == -1)
	{
	  bound_ymin = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bound_ymax") && iid == -1)
	{
	  bound_ymax = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("filename") && iid == -1)
	{
	  if (strlen(setvalue->c_str()) < 1000)
            {
              strcpy(filename, setvalue->c_str());
              return true;
            }
          else
            return false;
	}
      else if (out == std::string("isfiledata") && iid == -1)
	{
	  isfiledata = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("info") && iid == -1)
	{
	  *info = *setvalue;
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}

bool data::exec(std::string* function, buffarray *argument, tokenbuff* outvalue)
{
  std::string out, shrunk, str1, str2;
  int iid, arglength;
  char stemp[100];
  FILE* fp;
  int itemp;
  buffarray ini, fun;
  graph* gra = (graph *)(((frame *)parent)->parent);

  outvalue->value->setbuff(0);
  if (topobject(function, &out, &iid) == true)
    {
      arglength = argument->getarraylength();

      // paint
      if (out == std::string("paint") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (fhvuff->get(&(*argument)[0], &fp, &itemp, &str1))
		{
		  paint(true, fp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // remakebuffer
      else if (out == std::string("remakebuffer") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      if (!isfiledata)
		if (!genxdata(m_start, m_end, m_div, m_scaling, x, y))
		  return false;
	      return true;
	    }
	  else
	    return false;
	}
      // setbbox
      else if (out == std::string("setbbox") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      paint(false, fp);
	      return true;
	    }
	  else
	    return false;
	}
      // datalength
      else if (out == std::string("datalength") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      sprintf(stemp, "%d", x->getarraylength());
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // linearfit
      else if (out == std::string("linearfit") && iid == -1)
	{
	  if (arglength == 6)
	    {
	      gra->updated = true;
	      if (ExecLinearSQ(atoi((*argument)[0].c_str()),
			   atoi((*argument)[1].c_str()),
			   atof((*argument)[2].c_str()),
			   atof((*argument)[3].c_str()),
			   atoi((*argument)[4].c_str()),
			   atoi((*argument)[5].c_str()),
			   &str1,
			   &str2))
		{
		  outvalue->value->add(str1.c_str());
		  outvalue->value->add(str2.c_str());
		}
	      else
		{
		  outvalue->value->add("");
		  outvalue->value->add("");
		}
	      return true;
	    }
	  else
	    return false;
	}
      // nonlinearfit
      else if (out == std::string("nonlinearfit") && iid == -1)
	{
	  if (arglength == 22)
	    {
	      ini.setbuff(0);
	      ini.add("");
	      for (int i = 3; i <= 11; i++)
		ini.add((*argument)[i].c_str());

	      fun.setbuff(0);
	      fun.add("");
	      for (int i = 12; i <= 20; i++)
		fun.add((*argument)[i].c_str());

	      gra->updated = true;
	      if (OutMar(atoi((*argument)[0].c_str())
			 ,atoi((*argument)[1].c_str())
			 ,&(*argument)[2]
			 ,&ini
			 ,&fun
			 ,atof((*argument)[21].c_str())
			 , &str1
			 , &str2
			 ))
		{
		  outvalue->value->add(str1.c_str());
		  outvalue->value->add(str2.c_str());
		}
	      else
		{
		  outvalue->value->add("");
		  outvalue->value->add("");
		}
	      return true;
	    }
	  else
	    return false;
	}
      // loaddata
      else if (out == std::string("loaddata") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      gra->updated = true;
	      if (loaddata())
		sprintf(stemp, "1");
	      else
		sprintf(stemp, "0");

	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;
}


bool data::paint(bool flag, FILE* f)
{
  drawplot(flag, f);
  return true;
}


bool data::readdata(FILE *fp)
{
  char stemp[10000];
  std::string str;
  char *p;
  int readversion;
  
//  virgin = false;

  // version
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      readversion = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // id
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      id = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // virgin
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      virgin = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (readversion >= 2)
    {
      // display
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  if (nextitem(stemp, &str) != 0)
	    display = atoi(str.c_str());
	  else
	    return false;
	}
      else
	return false;
    }

  // trans_x, trans_y
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // trans_x
    p = nextitem(p, &str);
    if (p != 0)
      trans_x = atoi(str.c_str());
    else
      return false;
    // trans_y
    p = nextitem(p, &str);
    if (p != 0)
      trans_y = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // format_x
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      *format_x = str;
    else
      return false;
  }

  // format_y
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      *format_y = str;
    else
      return false;
  }

  // linecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // markeredgecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      markeredgecolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      markeredgecolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      markeredgecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // markerbodycolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      markerbodycolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      markerbodycolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      markerbodycolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // linestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // width
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.width = atoi(str.c_str());
    else
      return false;
    // cap
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.cap = atoi(str.c_str());
    else
      return false;
    // join
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.join = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // markeredgestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // width
    p = nextitem(p, &str);
    if (p != 0)
      markeredgestyle.width = atoi(str.c_str());
    else
      return false;
    // cap
    p = nextitem(p, &str);
    if (p != 0)
      markeredgestyle.cap = atoi(str.c_str());
    else
      return false;
    // join
    p = nextitem(p, &str);
    if (p != 0)
      markeredgestyle.join = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      markeredgestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // markerbodyfillstyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // fillrule
    p = nextitem(p, &str);
    if (p != 0)
      markerbodyfillstyle.fillrule = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      markerbodyfillstyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // interpolate
    p = nextitem(p, &str);
    if (p != 0)
      interpolate = atoi(str.c_str());
    else
      return false;
    // interpolatediv
    p = nextitem(p, &str);
    if (p != 0)
      interpolatediv = atoi(str.c_str());
    else
      return false;
    // markerstyle
    p = nextitem(p, &str);
    if (p != 0)
      markerstyle = atoi(str.c_str());
    else
      return false;
    // markercenterdot
    p = nextitem(p, &str);
    if (p != 0)
      markercenterdot = atoi(str.c_str());
    else
      return false;
    // markersize
    p = nextitem(p, &str);
    if (p != 0)
      markersize = atoi(str.c_str());
    else
      return false;
    // clipping
    p = nextitem(p, &str);
    if (p != 0)
      clipping = atoi(str.c_str());
    else
      return false;
    // refaxis_x
    p = nextitem(p, &str);
    if (p != 0)
      refaxis_x = atoi(str.c_str());
    else
      return false;
    // refaxis_y
    p = nextitem(p, &str);
    if (p != 0)
      refaxis_y = atoi(str.c_str());
    else
      return false;
    // basevalue
    p = nextitem(p, &str);
    if (p != 0)
      basevalue = atof(str.c_str());
    else
      return false;
  }
  else
    return false;

  // memory data
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // m_start
    p = nextitem(p, &str);
    if (p != 0)
      m_start = atof(str.c_str());
    else
      return false;
    // m_end
    p = nextitem(p, &str);
    if (p != 0)
      m_end = atof(str.c_str());
    else
      return false;
    // m_div
    p = nextitem(p, &str);
    if (p != 0)
      m_div = atoi(str.c_str());
    else
      return false;
    // m_scaling
    p = nextitem(p, &str);
    if (p != 0)
      m_scaling = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;




  // col_x
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // col_x
    p = nextitem(p, &str);
    if (p != 0)
      col_x = atoi(str.c_str());
    else
      return false;
    // col_xmin
    p = nextitem(p, &str);
    if (p != 0)
      col_xmin = atoi(str.c_str());
    else
      return false;
    // col_xmax
    p = nextitem(p, &str);
    if (p != 0)
      col_xmax = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // col_y
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // col_y
    p = nextitem(p, &str);
    if (p != 0)
      col_y = atoi(str.c_str());
    else
      return false;
    // col_ymin
    p = nextitem(p, &str);
    if (p != 0)
      col_ymin = atoi(str.c_str());
    else
      return false;
    // col_ymax
    p = nextitem(p, &str);
    if (p != 0)
      col_ymax = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // readposition
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // readstart
    p = nextitem(p, &str);
    if (p != 0)
      readstart = atoi(str.c_str());
    else
      return false;
    // readend
    p = nextitem(p, &str);
    if (p != 0)
      readend = atoi(str.c_str());
    else
      return false;
    // skip
    p = nextitem(p, &str);
    if (p != 0)
      skip = atoi(str.c_str());
    else
      return false;
    // commentlineaction
    p = nextitem(p, &str);
    if (p != 0)
      commentlineaction = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // bar
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // bar_x
    p = nextitem(p, &str);
    if (p != 0)
      bar_x = atoi(str.c_str());
    else
      return false;
    // bar_y
    p = nextitem(p, &str);
    if (p != 0)
      bar_y = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // datasearchpath
    p = nextitem(p, &str);
    if (p != 0)
      datasearchpath = atoi(str.c_str());
    else
      return false;
    // isfiledata
    p = nextitem(p, &str);
    if (p != 0)
      isfiledata = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // extradata
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      *extradata = str;
    else
      return false;
  }

  // filename
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      strcpy(filename, str.c_str());
    else
      return false;
  }

  // info
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      *info = str;
    else
      return false;
  }

  return true;
}

bool data::writedata(FILE *fp)
{
  std::string str;
  fprintf(fp, "[data]\n");

  // style
  fprintf(fp, "%d\n", version);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d\n", virgin);
  fprintf(fp, "%d\n", display);
  fprintf(fp, "%d %d\n", trans_x, trans_y);
  fprintf(fp, "\"%s\"\n", format_x->c_str());
  fprintf(fp, "\"%s\"\n", format_y->c_str());
  fprintf(fp, "%u %u %u\n", linecolor.red, linecolor.green, linecolor.blue);
  fprintf(fp, "%u %u %u\n", markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
  fprintf(fp, "%u %u %u\n", markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
  fprintf(fp, "%d %d %d %d\n", linestyle.width, linestyle.cap, linestyle.join, linestyle.style);
  fprintf(fp, "%d %d %d %d\n", markeredgestyle.width, markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
  fprintf(fp, "%d %d\n", markerbodyfillstyle.fillrule, markerbodyfillstyle.style);
  fprintf(fp, "%d %d %d %d %d %d %d %d %.15g\n", interpolate, interpolatediv, markerstyle, markercenterdot, markersize,
	  clipping, refaxis_x, refaxis_y, basevalue);
  fprintf(fp, "%.15g %.15g %d %d\n", m_start, m_end, m_div, m_scaling);

  // read data
  fprintf(fp, "%d %d %d\n", col_x, col_xmin, col_xmax);
  fprintf(fp, "%d %d %d\n", col_y, col_ymin, col_ymax);
  fprintf(fp, "%d %d %d %d\n", readstart, readend, skip, commentlineaction);
  fprintf(fp, "%d %d\n", bar_x, bar_y);
  fprintf(fp, "%d %d\n", datasearchpath, isfiledata);
  fprintf(fp, "\"%s\"\n", extradata->c_str());
  fprintf(fp, "\"%s\"\n", filename);
  fprintf(fp, "\"%s\"\n", info->c_str());

  return true;
}

bool data::appendoriginaldata(FILE *fp, int dataid)
{
  FILE *dest;
  char openfilename[1000];
  char stemp[10000];
  char header[1000];
  frame* fr = (frame *)parent;

  sprintf(header, "sod[%s]\n", filename);

  // load from tmpdir
  if (!virgin)
    {
      strcpy(openfilename, tmppath);
      strcat(openfilename, "/");
      strcat(openfilename, pid);
      sprintf(stemp, "%d_", fr->id);
      strcat(openfilename, stemp);
      sprintf(stemp, "%d_", dataid);
      strcat(openfilename, stemp);
      strcat(openfilename, filename);
      
      if ((dest = fopen(openfilename, "r")) != NULL)
	{
	  fputs(header, fp);
	  while (topazfgets(stemp, sizeof(stemp), dest) != NULL)
	    fputs(stemp, fp);
	  fputs("eod\n", fp);
	  fclose(dest);
	}
      else
	return false;
    }
  return true;
}

void data::gettmpdataname(std::string* file)
{
  frame* fr = (frame *)parent;
  char stemp[10000];

  *file = std::string(pid);
  sprintf(stemp, "%d_", fr->id);
  *file += std::string(stemp);
  sprintf(stemp, "%d_", id);
  *file += std::string(stemp);
  *file += std::string(toshortfilename(filename));
}


bool data::setrefpath(char *refpath)
{
  frame* fr = (frame *)parent;

  if ((!fr->withdata) && isfiledata && datasearchpath == 1)
    strcpy(filename, relpath((const char *)refpath, (const char *)filename));
  return true;
}


bool data::restoreabspath(char *refpath)
{
  frame* fr = (frame *)parent;

  if ((!fr->withdata) && isfiledata && datasearchpath == 1)
    return abspath((const char *)filename, (const char *)refpath, filename);

  return true;
}



