/*----------------------------------------------------------------------------
                       data.h (definitions of data object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__data_h)
#define __data_h

#include <stdio.h>

#include "plobj.h"
#include <string>
#include "col.h"
#include "vect.h"
#include "vectdefs.h"

/* marker style */
#define MARKER_NULL          0
#define MARKER_SQUARE        1
#define MARKER_TRIANGLE      2 
#define MARKER_REVTRIANGLE   3
#define MARKER_DIAMOND       4
#define MARKER_CIRCLE        5
#define MARKER_CROSS         6
#define MARKER_DCROSS        7
#define MARKER_VAREA         8
#define MARKER_HAREA         9
#define MARKER_VBAR          10
#define MARKER_HBAR          11
#define MARKER_VSTEP         12
#define MARKER_HSTEP         13
#define MARKER_VBAR2         14
#define MARKER_HBAR2         15

/* action of reading comment line */
#define COMMENT_ABORT        0
#define COMMENT_IGNORE       1
#define COMMENT_SEPARTION    2

/* data search path */
#define PATH_ABSOLUTE        0
#define PATH_RELATIVE        1

/* data interpolation */
#define INTER_POLYLINE           0
#define INTER_CLOSEDPOLYLINE     1
#define INTER_SPLINE             2
#define INTER_PSPLINE            3
#define INTER_CLOSEDPSPLINE      4
#define INTER_BSPLINE            5
#define INTER_CLOSEDBSPLINE      6

/* dataelement status*/
#define DATA_NORMAL               0
#define DATA_INVALID              2
#define DATA_IGNORE               4

/* maxreadline */
#define MAXREADLINE             2000;

/* mscaling */
#define MSCALING_LINEAR          0
#define MSCALING_LOG             1
#define MSCALING_INV             2


class dataelement
{
 public:
  dataelement(){data = 0; status = DATA_NORMAL;}
  dataelement(double d, int s){data = d; status = s;}
  ~dataelement(){;}
  
  double data;
  int    status;
    
  int operator == (const dataelement& de ) const;
  dataelement& operator =(const dataelement& de);
};

inline int dataelement::operator == ( const dataelement& de ) const
{
        return this == &de;
}

inline dataelement& dataelement::operator =(const dataelement& de)
{
        data                    = de.data;
        status                  = de.status;

        return *this;
}

typedef vect<dataelement> earray;
/* typedef vect<earray> eearray; */

class data: public plobj
{
 public:
  data();
  ~data();
  data& operator = (const data& dat);
  bool ispatternequal(data* dat1, data* dat2);
  
  bool get(std::string* member,  tokenbuff* outvalue);  
  bool set(std::string* member,  std::string* setvalue);  
  bool exec(std::string* function, buffarray *argument, tokenbuff* outvalue);  
  bool paint(bool flag, FILE *f);
  
  /* plot style */
  int id;
  std::string *label;
  int selected;
  bool virgin;
  bool display;
  bool trans_x;
  std::string *format_x;
  bool trans_y;
  std::string *format_y;
  Capcolor linecolor, markeredgecolor, markerbodycolor;
  Caplinestyle linestyle, markeredgestyle;
  Capfillstyle markerbodyfillstyle;
  int interpolate;
  int interpolatediv;
  int markerstyle;
  bool markercenterdot;
  int markersize;
  bool clipping;
  int refaxis_x;
  int refaxis_y;
  double basevalue;

  /* memory data */
  double m_start;
  double m_end;
  int    m_div;
  int    m_scaling;

  /* data read / write  */
  bool readdata(FILE *fp);
  bool writedata(FILE *fp);
  bool appendoriginaldata(FILE *fp, int dataid);
  void gettmpdataname(std::string* file);

  /* read data */
  int col_x, col_y, col_xmin, col_xmax, col_ymin, col_ymax;
  int readstart, readend, skip;
  int commentlineaction;
  bool bar_x, bar_y;
  int datasearchpath;
  bool extradatadispflag[99];
  std::string *extradata;

  /* data */
  earray *x, *y, *xmax, *xmin, *ymax, *ymin;
  earray *extra[99];
  double bound_xmin, bound_xmax, bound_ymin, bound_ymax;
  char filename[1000];
  bool isfiledata;
  std::string *info;


  bool loaddata();
  bool loaddataproc(FILE *fp);
  bool setrefpath(char *refpath);
  bool restoreabspath(char *refpath);
  bool extradataformat(const char *s, bool CheckMode);
  void drawplot(int flag, FILE* f);
  void drawmark(int flag, FILE* f, int x,int y, intpointarray *ia);
  bool ExecLinearSQ(int fittingCurve, bool Fixed, double fixed_x, double fixed_y, int startline, int endline
		    , std::string* outfnstr, std::string *infostr);
  bool UserFuncs(data *dat, int Ndata, double *a, int na, int Npara);
  bool MarqSetPara(data *dat, int Ndata, double *a, int Na, int *list,
		   int Npara, double **alpha, double *beta, double *X2);
  bool ExecMarquadt(data *dat, int Ndata, double *a, int *list,
		    int Npara, double init_chisq, double *out_chisq);
  bool OutMar(int StartLine, int EndLine, std::string* func, buffarray* ini, buffarray* dydA, double seido, 
	      std::string* outfn, std::string* info);
  void createfunction(std::string *in, std::string *out);

};

#endif








