/*----------------------------------------------------------------------------
                  datadef.h (definitions of point, rect)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__datadef_h)
#define __datadef_h


template <class T>
class imppoint
{
 public:
  
  imppoint(){;}
  imppoint(T px, T py){x = px, y = py;}
  ~imppoint(){;}

  imppoint& operator = (const imppoint& p)
    {x = p.x; y = p.y; return *this;}
  int operator == ( const imppoint& obj ) const
    { return this == &obj;}


  T x;
  T y;

};

template <class T>
class imprect
{
 public:
  
  imprect(){;}
  imprect(T pleft, T ptop, T pright, T pbottom){left = pleft; top = ptop; right = pright; bottom = pbottom;}
  ~imprect(){;}

  imprect& operator = (const imprect& r)
    {left = r.left; top = r.top; right = r.right; bottom = r.bottom; return *this;}
  int operator == ( const imprect& obj ) const
    { return this == &obj;}

  bool contains(T x, T y);
  
  T left;
  T top;
  T right;
  T bottom;

  imprect operator & (const imprect& r);
  void normalized();
  bool isempty()
    {
      if (left < right && top < bottom)
	return false;
      else
	return true;
    }
  
};

template <class T>
bool imprect<T>::contains(T x, T y)
{
  T minx, maxx, miny, maxy;

  if (left < right)
    {
      minx = left; maxx = right;
    }
  else
    {
      minx = right; maxx = left;
    }

  if (top < bottom)
    {
      miny = top; maxy = bottom;
    }
  else
    {
      miny = bottom; maxy = top;
    }

  if (minx <= x && x <= maxx && miny <= y && y <= maxy)
    return true;
  else
    return false;
}


template <class T>
void imprect<T>::normalized()
{
  T minx, maxx, miny, maxy;

  if (left < right)
    {
      minx = left; maxx = right;
    }
  else
    {
      minx = right; maxx = left;
    }

  if (top < bottom)
    {
      miny = top; maxy = bottom;
    }
  else
    {
      miny = bottom; maxy = top;
    }

  left = minx; right  = maxx;
  top  = miny; bottom = maxy;

}

template <class T>
imprect<T> imprect<T>::operator & (const imprect& r)
{
  imprect<T> outr;
  imprect<T> tempr = r;
  imprect<T> temp = *this;
  tempr.normalized();
  temp.normalized();
  
  if (temp.left < tempr.left)
    outr.left = tempr.left;
  else 
    outr.left = temp.left;
  
  if (temp.right > tempr.right)
    outr.right = tempr.right;
  else 
    outr.right = temp.right;
  
  if (temp.top < tempr.top)
    outr.top = tempr.top;
  else 
    outr.top = temp.top;
  
  if (temp.bottom > tempr.bottom)
    outr.bottom = tempr.bottom;
  else 
	outr.bottom = temp.bottom;
  
  return outr;
}

#endif











