/*----------------------------------------------------------------------------
                         dinvq.cc (inverse erf function)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#define GOSA    1e-15

double kou(double x, int i)
{
	double ftemp = x;
    int j;

    if (i == 0)
    	return x;
    else
    {
		ftemp = x;
	    for (j = 1; j <= i; j++)
    		ftemp *= x * x /(2.0 * j + 1.0);
    }
    return ftemp;
}

bool kyuusuu(double in, double *out)
{
    double sumn, addsum;
    int count = 0;
    int limitcount = 10000;

    sumn = 0.0;
    do
	{
    	addsum = kou(in, count);
    	sumn = sumn + addsum;
        count++;

        if (count > limitcount)
        	return false;

    }while(fabs(addsum/sumn) > GOSA);

    *out = sumn;
    return true;
}


bool iterfunction2(double yn, double x, double *out)
{
    double ftemp;

    if (kyuusuu(yn, &ftemp) == true)
	{
    	*out = yn - ftemp + sqrt(2.0 * M_PI) * x * exp(yn * yn / 2.0);
    	return true;
    }
    else
    	return false;
}

bool dinvq_calc2(double in, double *out)
{

    double x1, y0;
    double yn, yn1;
    x1 = 0.5 - in;
    y0 = sqrt(2.0 * M_PI) * x1 * 1.5;
    int count = 0;
    int limitcount = 10000;
    double ftemp;

    yn1 = y0;
    do
	{
	    yn = yn1;

    	if (iterfunction2(yn, x1, &ftemp) == true)
    		yn1 = ftemp;
        else
        	return false;

        count++;

        if (count > limitcount)
        	return false;

//    }while(fabs((yn1 - yn)/yn1) > 1e-15);
    }while(fabs(yn1 / yn - 1.0) > GOSA);

    *out = yn1;
    return true;
}

bool renbun(double in, double *out)
{
   int niter;  		//iteration number
    double ftemp;
    int i;

    if (in < 7.5)
    	niter = (int)(110.0 / (in - 1.0));
    else if ((7.5 <= in) && (in < 12.5))
    	niter = (int)(-1.6 * in + 30.0);
    else if (12.5 <= in)
    	niter = 10;
    else
    	return false;

    ftemp = in + (double)(niter + 1.0) / in;
    for (i = niter; i >= 1; i--)
    	ftemp = in + (double)i / ftemp;

    ftemp = 1.0 / ftemp;

    *out = ftemp;
    return true;
}

bool iterfunction1(double yn, double x, double *out)
{
    double ftemp;

    if (renbun(yn, &ftemp) == true)
	{
    	*out = yn + ftemp - sqrt(2.0 * M_PI) * x * exp(yn * yn / 2.0);
    	return true;
    }
    else
    	return false;
}
bool dinvq_calc1(double in, double *out)
{
    double c0, c1, y0, yn, yn1;
    int count = 0;
    int limitcount = 10000;
    double ftemp;

    c0 = sqrt(-2.0 * log(sqrt(2.0 * M_PI) * 3.0 * in));
    c1 = c0 + 1.0 / c0;
    y0 = sqrt(-2.0 * log(sqrt(2.0 * M_PI) * c1 * in));

    yn1 = y0;
    do
	{
        yn = yn1;
    	if (iterfunction1(yn, in, &ftemp) == true)
    		yn1 = ftemp;
        else
        	return false;

        count++;

        if (count > limitcount)
        	return false;

    }while(fabs((yn1-yn)/yn) > GOSA);

    *out = yn1;
    return true;
}


bool dinvq(double in, double *out)
{

    double ftemp, xstar;

	if ((in <= 0.0) || (in > 1.0))
    	return false;
    else if ((0.0 < in ) && (in <= 0.01))
	{
    	if (dinvq_calc1(in, &ftemp) == true)
        {
        	*out = ftemp;
            return true;
    	}
        else
        	return false;
    }
    else if ((0.01 < in ) && (in <= 0.5))
	{
    	if (dinvq_calc2(in, &ftemp) == true)
        {
        	*out = ftemp;
            return true;
    	}
        else
        	return false;
    }
    else if ((0.5 < in ) && (in < 1.0))
	{
    	xstar = 1.0 - in;
        if ((0.0 < xstar) && (xstar <= 0.01))
        {
	    	if (dinvq_calc1(xstar, &ftemp) == true)
    	    {
        		*out = -ftemp;
            	return true;
	    	}
    	    else
        		return false;
        }
        else if ((0.01 < xstar) && (xstar < 0.5))
        {
	    	if (dinvq_calc2(xstar, &ftemp) == true)
    	    {
        		*out = -ftemp;
            	return true;
	    	}
    	    else
        		return false;
    	}
    }
    else if (in == 1.0)
    {
    	*out = 0.0;
        return true;
    }
    return true;
}
