/*----------------------------------------------------------------------------
                      filepoint.cc (objectlist interfaces)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string>
#include "filepoint.h"
#include "pipecall.h"

HandleElement::HandleElement()
{label = std::string(""); handle = 0;}

HandleElement& HandleElement::operator = (const HandleElement&  ele)
{
  label         = ele.label;
  handle        = ele.handle;
  handletype    = ele.handletype;
  filename      = ele.filename;
  return *this;
}

FileHandleArray::FileHandleArray()
{
  int i;
  arraylength = 0;
  increment = 100;

  bufflength = (arraylength / increment + 1) * increment;
  element =   new HandleElement * [bufflength];

  for (i = 0; i < bufflength; i++)
    element[i] = 0;

  for (i = 0; i < arraylength; i++)
    element[i] = new HandleElement();
}

FileHandleArray::~FileHandleArray()
{
  int i;
  for (i = 0; i < arraylength; i++)
    {
      switch(element[i]->handletype)
	{
	case READFROMFILE:
	case WRITETOFILE:
	case APPENDTOFILE:
	  fclose(element[i]->handle);
	  break;
	case READFROMPIPE:
	case WRITETOPIPE:
	  pclose(element[i]->handle);
	  break;
	default:
	  return;
	}
      delete element[i];
    }
  delete [] element;

}

bool FileHandleArray::search(std::string *label, int *SeachedIndex)
{
  int i;

  for (i = 0; i < arraylength; i++)
    {
      if (*label == element[i]->label)
	{
	  *SeachedIndex = i;
	  return true;
	}
    }
  return false;

}

bool FileHandleArray::set(std::string *label, FILE **h, int *handletype, std::string *filename)
{
  int itemp;

  if (search(label, &itemp) == true)    //overwrite
    {
      element[itemp]->handle = *h;
      element[itemp]->handletype = *handletype;
      element[itemp]->filename = *filename;
      return true;
    }
  else                                  //append to last
    {
      add(label, h, handletype, filename);
      return true;
    }
}

bool FileHandleArray::get(std::string *label, FILE **h, int *handletype, std::string *filename)
{
  int itemp;

  if (search(label, &itemp) == true)
    {
      *h          = element[itemp]->handle;
      *handletype = element[itemp]->handletype;
      *filename   = element[itemp]->filename;
      return true;
    }
  else if (*label == std::string("STDIN"))
    {
      *h          = stdin;
      *handletype = READFROMFILE;
      *filename   = std::string("");
      return true;
    }
  else if (*label == std::string("STDOUT"))
    {
      *h          = stdout;
      *handletype = WRITETOFILE;
      *filename   = std::string("");
      return true;
    }
  else if (*label == std::string("STDERR"))
    {
      *h          = stderr;
      *handletype = WRITETOFILE;
      *filename   = std::string("");
      return true;
    }
  else
     return false;
}

bool FileHandleArray::getinfo(int i, std::string *label, int *handletype, std::string *filename)
{
  if (0 <= i && i < getarraylength())
    {
      *label      = element[i]->label;
      *handletype = element[i]->handletype;
      *filename   = element[i]->filename;
      return true;
    }
  else
     return false;
}


int FileHandleArray::getarraylength()
{
  return arraylength;
}

bool FileHandleArray::setbuff(int length)
{
  int i;

  for (i = 0; i < arraylength; i++)
    delete element[i];
  delete [] element;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  element =  new HandleElement * [bufflength];
  for (i = 0; i < bufflength; i++)
    element[i] = 0;
  return true;
}

bool FileHandleArray::add(std::string *label, FILE **h, int *handletype, std::string *filename)
{
  int itemp;
  if (search(label, &itemp))
    return false;
  if (arraylength + 1 >= bufflength)
      changebufflength(arraylength + 1);

  arraylength++;
  element[arraylength - 1] = new HandleElement();
  element[arraylength - 1]->label        = *label;
  element[arraylength - 1]->handle       = *h;
  element[arraylength - 1]->handletype   = *handletype;
  element[arraylength - 1]->filename     = *filename;

  return true;
}

bool FileHandleArray::remove(int pos)
{
  int i;

  for (i = pos; i < arraylength - 1; i++)
    *element[i] = *element[i + 1];

  delete element[arraylength - 1];
  arraylength--;
  return true;
}

bool FileHandleArray::changebufflength(int newlength)
{

  int i;
  HandleElement **tempstrarray;
  int temparraylength;
  
  if (newlength <= bufflength)
    return true;

  tempstrarray =   new HandleElement * [bufflength];
  temparraylength = arraylength;
  
  for (i = 0; i < arraylength; i++)
    tempstrarray[i] = element[i];
  
  for (i = arraylength; i < bufflength; i++)
    tempstrarray[i] = 0;

  delete[] element;

  bufflength = (newlength / increment + 1) * increment;
  element =  new HandleElement * [bufflength];
  for (i = 0; i < bufflength; i++)
    element[i] = 0;

  if (newlength > temparraylength)
    {
      for (i = 0; i < temparraylength; i++)
	element[i] = tempstrarray[i];
      for (i = temparraylength; i < newlength; i++)
	element[i] = 0;

    }
  else
    {
      for (i = 0; i < newlength; i++)
	element[i] = tempstrarray[i];
    }

  delete [] tempstrarray;

  return true;
}




