/*----------------------------------------------------------------------------
                           frame.cc (frame object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "frame.h"
#include "graph.h"
#include "axis.h"
#include "data.h"
#include "gdi.h"
#include "vectdefs.h"
#include "pointdefs.h"
#include "filepoint.h"
#include "system.h"

extern void StyleLinePath(bool flag, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);
extern bool IsValidConstant(std::string *str);
extern bool IsValidFunction(std::string *str);
extern FileHandleArray *fhvuff;
extern bool isabort();
extern void resetabort();
extern char tmppath[1000];
extern void message(const char *msg);

frame::frame():plobj()
{
  version = 2;

  axarray = new axisarray;
  axarray->add(new axis);
  (*axarray)[axarray->getarraylength() - 1].parent = (void *)this;
  (*axarray)[axarray->getarraylength() - 1].id = axarray->getarraylength() - 1;
  (*axarray)[axarray->getarraylength() - 1].axistype = AXISTYPE_X;
  (*axarray)[axarray->getarraylength() - 1].tickparity = TICKPARITY_MINUS;
  (*axarray)[axarray->getarraylength() - 1].angle = 0.0;
  (*axarray)[axarray->getarraylength() - 1].fo.angle = 0.0;
  (*axarray)[axarray->getarraylength() - 1].labelparity = LABELPARITY_MINUS;
  (*axarray)[axarray->getarraylength() - 1].crosspointparity = CROSSPOINTPARITY_PLUS;
  setaxisid(axarray->getarraylength() - 1);

  axarray->add(new axis);
  (*axarray)[axarray->getarraylength() - 1].parent = (void *)this;
  (*axarray)[axarray->getarraylength() - 1].id = axarray->getarraylength() - 1;
  (*axarray)[axarray->getarraylength() - 1].axistype = AXISTYPE_Y;
  (*axarray)[axarray->getarraylength() - 1].tickparity = TICKPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].angle = 90.0;
  (*axarray)[axarray->getarraylength() - 1].fo.angle = 270.0;
  (*axarray)[axarray->getarraylength() - 1].labelparity = LABELPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].crosspointparity = CROSSPOINTPARITY_PLUS;
  setaxisid(axarray->getarraylength() - 1);

  axarray->add(new axis);
  (*axarray)[axarray->getarraylength() - 1].parent = (void *)this;
  (*axarray)[axarray->getarraylength() - 1].id = axarray->getarraylength() - 1;
  (*axarray)[axarray->getarraylength() - 1].axistype = AXISTYPE_T;
  (*axarray)[axarray->getarraylength() - 1].tickparity = TICKPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].angle = 0.0;
  (*axarray)[axarray->getarraylength() - 1].fo.angle = 0.0;
  (*axarray)[axarray->getarraylength() - 1].labelparity = LABELPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].crosspointparity = CROSSPOINTPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].format = 0;
  setaxisid(axarray->getarraylength() - 1);

  axarray->add(new axis);
  (*axarray)[axarray->getarraylength() - 1].parent = (void *)this;
  (*axarray)[axarray->getarraylength() - 1].id = axarray->getarraylength() - 1;
  (*axarray)[axarray->getarraylength() - 1].axistype = AXISTYPE_R;
  (*axarray)[axarray->getarraylength() - 1].tickparity = TICKPARITY_MINUS;
  (*axarray)[axarray->getarraylength() - 1].angle = 90.0;
  (*axarray)[axarray->getarraylength() - 1].fo.angle = 270.0;
  (*axarray)[axarray->getarraylength() - 1].labelparity = LABELPARITY_MINUS;
  (*axarray)[axarray->getarraylength() - 1].crosspointparity = CROSSPOINTPARITY_PLUS;
  (*axarray)[axarray->getarraylength() - 1].format = 0;
  setaxisid(axarray->getarraylength() - 1);

  darray = new dataarray;
  parray = new partsarray;

  usrconst = new VariantBuff;
  usrfnc  = new VariantBuff;

  id = 0;
  label = new std::string("");
  paintorder = new std::string("frame,data,axis,axisline,parts");

  p1.x = 0;
  p1.y = 0;
  p2.x = 5000;
  p2.y = 5000;

  framelinecolor.setblack();
  framebodycolor.setwhite();
  framelinestyle.init();
  framebodyfillstyle.init();

  frametype = FRAME_BOX;
  selected = false;
  withdata = true;

  arrangeaxes();
}

frame::~frame()
{
  delete axarray;
  delete darray;
  delete parray;
  delete usrconst;
  delete usrfnc;
  delete label;
  delete paintorder;
}

frame& frame::operator = (const frame&  buff)
{

  // axis
  axarray->setbuff(0);
  for(int i = 0; i < buff.axarray->getarraylength(); i++)
    {
      axarray->add(new axis);
      (*axarray)[axarray->getarraylength() - 1].parent = (void *)this;
      setaxisid(axarray->getarraylength() - 1);
      (*axarray)[i] = (*buff.axarray)[i];
    }

  // data
  darray->setbuff(0);
  for(int i = 0; i < buff.darray->getarraylength(); i++)
    {
      darray->add(new data);
      (*darray)[darray->getarraylength() - 1].parent = (void *)this;
      setdataid(darray->getarraylength() - 1);
      (*darray)[i] = (*buff.darray)[i];
    }

  // parts
  parray->setbuff(0);
  for(int i = 0; i < buff.parray->getarraylength(); i++)
    {
      parray->add(new parts);
      (*parray)[parray->getarraylength() - 1].parent = (void *)this;
      setpartsid(parray->getarraylength() - 1);
      (*parray)[i] = (*buff.parray)[i];
    }

  // usrcost
  *usrconst = *buff.usrconst;

  // usrfnc
  *usrfnc = *buff.usrfnc;


  version = buff.version;
  p1 = buff.p1;
  p2 = buff.p2;
  framelinecolor = buff.framelinecolor;
  framebodycolor = buff.framebodycolor;
  framelinestyle = buff.framelinestyle;
  framebodyfillstyle = buff.framebodyfillstyle;
  frametype = buff.frametype;
  selected = buff.selected;
  withdata = buff.withdata;

  return *this;
}

bool frame::get(std::string* member, tokenbuff* outvalue)
{
  std::string out, shrunk;
  int iid;
  char stemp[100];

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("axis"))
	{
	  if (shrinkobject(member, &shrunk))
	    {
	      if (iid < axarray->getarraylength() && iid >= 0)
		{
		  if ((*axarray)[iid].get(&shrunk, outvalue))
		    return true;
		  else
		    return false;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if ((*axarray)[i].selected)
			{
			  if ((*axarray)[i].get(&shrunk, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  outvalue->value->add("");
		  return false;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("data"))
	{
	  if (shrinkobject(member, &shrunk))
	    {

	      if (iid < darray->getarraylength() && iid >= 0)
		{
		  if ((*darray)[iid].get(&shrunk, outvalue))
		    return true;
		  else
		    return false;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if ((*darray)[i].selected)
			{
			  if ((*darray)[i].get(&shrunk, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  outvalue->value->add("");
		  return false;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("parts"))
	{
	  if (shrinkobject(member, &shrunk))
	    {

	      if (iid < parray->getarraylength() && iid >= 0)
		{
		  if ((*parray)[iid].get(&shrunk, outvalue))
		    return true;
		  else
		    return false;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if ((*parray)[i].selected)
			{
			  if ((*parray)[i].get(&shrunk, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  outvalue->value->add("");
		  return false;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("id") && iid == -1)
	{
	  sprintf(stemp, "%d", id);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  outvalue->value->add(label->c_str());
	  return true;
	}
      else if (out == std::string("paintorder") && iid == -1)
	{
	  outvalue->value->add(paintorder->c_str());
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  sprintf(stemp, "%d", selected);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("withdata") && iid == -1)
	{
	  sprintf(stemp, "%d", withdata);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("x1") && iid == -1)
	{
	  sprintf(stemp, "%d", p1.x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("y1") && iid == -1)
	{
	  sprintf(stemp, "%d", p1.y);
	  outvalue->value->add(stemp);
	  return true;
	}
      if (out == std::string("x2") && iid == -1)
	{
	  sprintf(stemp, "%d", p2.x);
	  outvalue->value->add(stemp);
	  return true;
	}
      if (out == std::string("y2") && iid == -1)
	{
	  sprintf(stemp, "%d", p2.y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framebodycolor_red") && iid == -1)
	{
	  sprintf(stemp, "%d", framebodycolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framebodycolor_green") && iid == -1)
	{
	  sprintf(stemp, "%d", framebodycolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framebodycolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%d", framebodycolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinecap") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framelinestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", framelinestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framebodyfillstyle") && iid == -1)
	{
	  sprintf(stemp, "%d", framebodyfillstyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("framebodyfillrule") && iid == -1)
	{
	  sprintf(stemp, "%d", framebodyfillstyle.fillrule);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("frametype") && iid == -1)
	{
	  sprintf(stemp, "%d", frametype);
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}

bool frame::set(std::string* member, std::string* setvalue)
{
  std::string out, shrunk;
  int iid;

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("axis"))
	{
	  if (shrinkobject(member, &shrunk))
	    {
	      if (iid < axarray->getarraylength() && iid >= 0)
		{
		  if ((*axarray)[iid].set(&shrunk, setvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if ((*axarray)[i].selected)
			{
			  if (!(*axarray)[i].set(&shrunk, setvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if (!(*axarray)[i].set(&shrunk, setvalue))
			return false;
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("data"))
	{
	  if (shrinkobject(member, &shrunk))
	    {
	      if (iid < darray->getarraylength() && iid >= 0)
		{
		  if ((*darray)[iid].set(&shrunk, setvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if ((*darray)[i].selected)
			{
			  if (!(*darray)[i].set(&shrunk, setvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if (!(*darray)[i].set(&shrunk, setvalue))
			return false;
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("parts"))
	{
	  if (shrinkobject(member, &shrunk))
	    {
	      if (iid < parray->getarraylength() && iid >= 0)
		{
		  if ((*parray)[iid].set(&shrunk, setvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if ((*parray)[i].selected)
			{
			  if (!(*parray)[i].set(&shrunk, setvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if (!(*parray)[i].set(&shrunk, setvalue))
			return false;
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else if (out == std::string("id") && iid == -1)
	{
	  id = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  *label = *setvalue;
	  return true;
	}
      else if (out == std::string("paintorder") && iid == -1)
	{
	  *paintorder = *setvalue;
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  selected = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("withdata") && iid == -1)
	{
	  withdata = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("x1") && iid == -1)
	{
	  p1.x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("y1") && iid == -1)
	{
	  p1.y = atoi(setvalue->c_str());
	  return true;
	}
      if (out == std::string("x2") && iid == -1)
	{
	  p2.x = atoi(setvalue->c_str());
	  return true;
	}
      if (out == std::string("y2") && iid == -1)
	{
	  p2.y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinecolor_red") && iid == -1)
	{
	  framelinecolor.red = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinecolor_green") && iid == -1)
	{
	  framelinecolor.green = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinecolor_blue") && iid == -1)
	{
	  framelinecolor.blue = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framebodycolor_red") && iid == -1)
	{
	  framebodycolor.red = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framebodycolor_green") && iid == -1)
	{
	  framebodycolor.green = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framebodycolor_blue") && iid == -1)
	{
	  framebodycolor.blue = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinewidth") && iid == -1)
	{
	  framelinestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinecap") && iid == -1)
	{
	  framelinestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinejoin") && iid == -1)
	{
	  framelinestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framelinestyle") && iid == -1)
	{
	  framelinestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framebodyfillstyle") && iid == -1)
	{
	  framebodyfillstyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("framebodyfillrule") && iid == -1)
	{
	  framebodyfillstyle.fillrule = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("frametype") && iid == -1)
	{
	  frametype = atoi(setvalue->c_str());
	  return true;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}


bool fr_paint(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str1, str3;
  FILE* fp;
  int itemp;
  std::string buf, str;
  char *p, stemp[3];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      if (fhvuff->get(&(*argument)[0], &fp, &itemp, &str1))
	{
	  resetabort();

	  buf = std::string("");
	  //remove space
	  p = (char *)fr->paintorder->c_str();
	  while (*p != 0)
	    {
	      if (*p != ' ')
		{
		  stemp[0] = *p;
		  stemp[1] = 0;
		  buf += std::string(stemp);
		}
	      p++;
	    }

	  str = std::string("");
	  p = (char *)buf.c_str();
	  do
	    {
	      if (*p != ',' && *p != 0)
		{
		  stemp[0] = *p;
		  stemp[1] = 0;
		  str += std::string(stemp);
		}
	      else
		{
		  if (strcmp((char *)str.c_str(), "frame") == 0)
		    {
		      //paint frame itself
		      if (!isabort())
			fr->paint(true, fp);
		    }
		  else if (strcmp((char *)str.c_str(), "data") == 0)
		    {
		      // data
		      for (int i = 0; i < fr->darray->getarraylength(); i++)
			{
			  if (!isabort())
			    {
			      str3 = std::string("paint");
			      if (!(*fr->darray)[i].exec(&str3, argument, outvalue))
				return false;
			    }
			}
		    }
		  else if (strcmp((char *)str.c_str(), "axis") == 0)
		    {		  
		      // axis
		      for (int i = 0; i < fr->axarray->getarraylength(); i++)
			{
			  str3 = std::string("paint");
			  if (!(*fr->axarray)[i].exec(&str3, argument, outvalue))
			    return false;
			}
		    }		  
		  else if (strcmp((char *)str.c_str(), "axisline") == 0)
		    {
		      // axis line
		      for (int i = 0; i < 2; i++)
			{
			  if (!isabort())
			    fr->axispaintmajorline(true, fp, &(*fr->axarray)[i]);
			  
			  if (!isabort())
			    fr->axispaintminorline(true, fp, &(*fr->axarray)[i]);
			}
		    }		  
		  else if (strcmp((char *)str.c_str(), "parts") == 0)
		    {
		      // parts
		      for (int i = 0; i < fr->parray->getarraylength(); i++)
			{
			  str3 = std::string("paint");
			  if (!(*fr->parray)[i].exec(&str3, argument, outvalue))
			    return false;
			}
		    }

		  str = std::string("");
		}

	      if (*p == 0)
		return true;

	      p++;
	    }while (1);

	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_setbbox(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str3;

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      resetabort();
      //paint frame itself
      if (!isabort())
	fr->paint(false, 0);
      
      // data
      for (int i = 0; i < fr->darray->getarraylength(); i++)
	{
	  if (!isabort())
	    {
	      str3 = std::string("setbbox");
	      if (!(*fr->darray)[i].exec(&str3, argument, outvalue))
		return false;
	    }
	}
      
      // axis
      for (int i = 0; i < fr->axarray->getarraylength(); i++)
	{
	  if (!isabort())
	    {
	      str3 = std::string("setbbox");
	      if (!(*fr->axarray)[i].exec(&str3, argument, outvalue))
		return false;
	    }
	}
      
      // axis line
      for (int i = 0; i < 2; i++)
	{
	  if (!isabort())
	    fr->axispaintmajorline(false, 0, &(*fr->axarray)[i]);
	  
	  if (!isabort())
	    fr->axispaintminorline(false, 0, &(*fr->axarray)[i]);
	}
      
      // parts
      for (int i = 0; i < fr->parray->getarraylength(); i++)
	{
	  if (!isabort())
	    {
	      str3 = std::string("setbbox");
	      if (!(*fr->parray)[i].exec(&str3, argument, outvalue))
		return false;
	    }
	}
      
      return true;
    }
  else
    return false;
}

bool fr_setframelinecolor(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 3 && iid == -1)
    {
      gra->updated = true;
      fr->framelinecolor.red = atoi((*argument)[0].c_str());
      fr->framelinecolor.green = atoi((*argument)[1].c_str());
      fr->framelinecolor.blue = atoi((*argument)[2].c_str());
      return true;
    }
  else
    return false;
}

bool fr_setframebodycolor(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 3 && iid == -1)
    {
      gra->updated = true;
      fr->framebodycolor.red = atoi((*argument)[0].c_str());
      fr->framebodycolor.green = atoi((*argument)[1].c_str());
      fr->framebodycolor.blue = atoi((*argument)[2].c_str());
      return true;
    }
  else
    return false;
}

bool fr_arrangeaxes(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      gra->updated = true;
      fr->arrangeaxes();
      return true;
    }
  else
    return false;
}

bool fr_wtov_x(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      itemp = (int)fr->doubleworld_to_view_x(atof((*argument)[0].c_str()));
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_wtov_y(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      itemp = (int)fr->doubleworld_to_view_y(atof((*argument)[0].c_str()));
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_vtow_x(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      itemp = (int)fr->doubleview_to_world_x(atof((*argument)[0].c_str()));
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_vtow_y(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      itemp = (int)fr->doubleview_to_world_y(atof((*argument)[0].c_str()));
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_autoscale(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  int itemp;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      itemp = atoi((*argument)[1].c_str());
      switch(atoi((*argument)[0].c_str()))
	{
	case 0:
	  fr->autoscale(true, false, false, false, 1, itemp);
	  break;
	case 1:
	  fr->autoscale(false, true, false, false, 1, itemp);
	  break;
	case 2:
	  fr->autoscale(false, false, true, false, 1, itemp);
	  break;
	case 3:
	  fr->autoscale(false, false, false, true, 1, itemp);
	  break;
	case -4:
	  fr->autoscale(true, true, true, true, 1, itemp);
	  break;
	}
      return true;
    }
  else
    return false;
}


bool fr_autounit(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  int itemp;
  if (arglength == 1 && iid == -1)
    {
      gra->updated = true;
      itemp  = atoi((*argument)[0].c_str());
      if (0 <= itemp && itemp < fr->axarray->getarraylength())
	  fr->autoscaleunit(&(*fr->axarray)[itemp], (*fr->axarray)[itemp].start, 
			  (*fr->axarray)[itemp].end, 0, 0);
      return true;
    }
  else
    return false;
}

bool fr_getindex(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 2 && iid == -1)
    {
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  if (fr->getaxisindex(atoi((*argument)[1].c_str())) != -1)
	    {
	      sprintf(stemp, "%d", fr->getaxisindex(atoi((*argument)[1].c_str())));
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (fr->getdataindex(atoi((*argument)[1].c_str())) != -1)
	    {
	      sprintf(stemp, "%d", fr->getdataindex(atoi((*argument)[1].c_str())));
	      outvalue->value->add(stemp);
		      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (fr->getpartsindex(atoi((*argument)[1].c_str())) != -1)
	    {
	      sprintf(stemp, "%d", fr->getpartsindex(atoi((*argument)[1].c_str())));
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_delete(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis")
	  && atoi((*argument)[1].c_str()) < fr->axarray->getarraylength()
	  && 3 < atoi((*argument)[1].c_str())
	  )
	{
	  fr->axarray->remove(atoi((*argument)[1].c_str()));
	  return true;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->darray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->darray->remove(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //selected
	  else if (atoi((*argument)[1].c_str()) == -3)
	    {
	      for (int i = fr->darray->getarraylength() - 1; i >= 0; i--)
		if ((*fr->darray)[i].selected)
		  fr->darray->remove(i);
	      return true;
	    }
	  //all
	  else if (atoi((*argument)[1].c_str()) == -4)
	    {
	      for (int i = fr->darray->getarraylength() - 1; i >= 0; i--)
		fr->darray->remove(i);
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->parray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->parray->remove(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //selected
	  else if (atoi((*argument)[1].c_str()) == -3)
	    {
	      for (int i = fr->parray->getarraylength() - 1; i >= 0; i--)
		if ((*fr->parray)[i].selected)
		  fr->parray->remove(i);
	      return true;
	    }
	  //all
	  else if (atoi((*argument)[1].c_str()) == -4)
	    {
	      for (int i = fr->parray->getarraylength() - 1; i >= 0; i--)
		fr->parray->remove(i);
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_new(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  char stemp[100];
  if (arglength == 1 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  fr->axarray->add(new axis);
	  (*fr->axarray)[fr->axarray->getarraylength() - 1].parent = (void *)fr;
	  fr->setaxisid(fr->axarray->getarraylength() - 1);
	  sprintf(stemp, "%d", fr->axarray->getarraylength());  
	  outvalue->value->add(stemp);
	  return true;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  fr->darray->add(new data);
	  (*fr->darray)[fr->darray->getarraylength() - 1].parent = (void *)fr;
	  fr->setdataid(fr->darray->getarraylength() - 1);
	  sprintf(stemp, "%d", fr->darray->getarraylength());  
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else if (arglength == 2)
    {
      // parts object
      if ((*argument)[0] == std::string("parts"))
	{
	  fr->parray->add(new parts);
	  (*fr->parray)[fr->parray->getarraylength() - 1].parent = (void *)fr;
	  fr->setpartsid(fr->parray->getarraylength() - 1);
	  
	  if ((*argument)[1] == std::string("text"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_TEXT;
	  else if ((*argument)[1] == std::string("line"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_LINE;
	  else if ((*argument)[1] == std::string("polyline"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_POLYLINE;
	  else if ((*argument)[1] == std::string("rectangle"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_RECTANGLE;
	  else if ((*argument)[1] == std::string("ellipse"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_ELLIPSE;
	  else if ((*argument)[1] == std::string("cutline"))
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_CUT;
	  else if (atoi((*argument)[1].c_str()) == PARTS_TEXT)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_TEXT;
	  else if (atoi((*argument)[1].c_str()) == PARTS_LINE)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_LINE;
	  else if (atoi((*argument)[1].c_str()) == PARTS_POLYLINE)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_POLYLINE;
	  else if (atoi((*argument)[1].c_str()) == PARTS_RECTANGLE)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_RECTANGLE;
	  else if (atoi((*argument)[1].c_str()) == PARTS_ELLIPSE)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_ELLIPSE;
	  else if (atoi((*argument)[1].c_str()) == PARTS_CUT)
	    (*fr->parray)[fr->parray->getarraylength() - 1].type = PARTS_CUT;
	  else
	    return false;
	  
	  sprintf(stemp, "%d", fr->parray->getarraylength());  
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_flushselect(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  for (int i = 0; i <fr->axarray->getarraylength(); i++)
	    (*fr->axarray)[i].selected = false;
	  return true;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  for (int i = 0; i < fr->darray->getarraylength(); i++)
	    (*fr->darray)[i].selected = false;
	  return true;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  for (int i = 0; i < fr->parray->getarraylength(); i++)
	    (*fr->parray)[i].selected = false;
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_length(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  sprintf(stemp, "%d", fr->axarray->getarraylength());
	  outvalue->value->add(stemp);
	  return true;
	}
      // data  object
      else if ((*argument)[0] == std::string("data"))
	{
	  sprintf(stemp, "%d", fr->darray->getarraylength());
	  outvalue->value->add(stemp);
	  return true;
	}
      // parts  object
      else if ((*argument)[0] == std::string("parts"))
	{
	  sprintf(stemp, "%d", fr->parray->getarraylength());
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_selectednumber(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  itemp = 0;
	  for (int i = 0; i < fr->axarray->getarraylength(); i++)
	    {
	      if ((*fr->axarray)[i].selected)
		itemp++;
	    }
	  sprintf(stemp, "%d", itemp);
	  outvalue->value->add(stemp);
	  return true;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  itemp = 0;
	  for (int i = 0; i < fr->darray->getarraylength(); i++)
	    {
	      if ((*fr->darray)[i].selected)
		itemp++;
	    }
	  sprintf(stemp, "%d", itemp);
	  outvalue->value->add(stemp);
	  return true;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  itemp = 0;
	  for (int i = 0; i < fr->parray->getarraylength(); i++)
	    {
	      if ((*fr->parray)[i].selected)
		itemp++;
	    }
	  sprintf(stemp, "%d", itemp);
	  outvalue->value->add(stemp);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_top(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->axarray->getarraylength() 
	      && 3 < atoi((*argument)[1].c_str())
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->axarray->top(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      fr->axarray->top(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->darray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->darray->top(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      fr->darray->top(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->parray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->parray->top(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      fr->parray->top(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_up(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->axarray->getarraylength() 
	      && 3 < atoi((*argument)[1].c_str())
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->axarray->up(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      fr->axarray->up(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->darray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->darray->up(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      fr->darray->up(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->parray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->parray->up(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      fr->parray->up(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_down(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->axarray->getarraylength() 
	      && 3 < atoi((*argument)[1].c_str())
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->axarray->down(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      fr->axarray->down(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->darray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->darray->down(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      fr->darray->down(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->parray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->parray->down(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      fr->parray->down(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_bottom(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->axarray->getarraylength() 
	      && 3 < atoi((*argument)[1].c_str())
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->axarray->bottom(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      fr->axarray->bottom(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->darray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->darray->bottom(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      fr->darray->bottom(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  if (atoi((*argument)[1].c_str()) < fr->parray->getarraylength() 
	      && 0 <= atoi((*argument)[1].c_str()))
	    {
	      fr->parray->bottom(atoi((*argument)[1].c_str()));
	      return true;
	    }
	  //first selected
	  else if (atoi((*argument)[1].c_str()) == -5)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      fr->parray->bottom(i);
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_copy(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  int itemp1, itemp2;
  std::string str1, str2;
  if (arglength == 3 && iid == -1)
    {
      gra->updated = true;
      // axis object
      if ((*argument)[0] == std::string("axis"))
	{
	  itemp1 = atoi((*argument)[1].c_str());
	  itemp2 = atoi((*argument)[2].c_str());
	  
	  if (itemp1 < fr->axarray->getarraylength() 
	      && 0 <= itemp1
	      && itemp2 < fr->axarray->getarraylength() 
	      && 0 <= itemp2)
	    {
	      (*fr->axarray)[itemp1] = (*fr->axarray)[itemp2];
	      return true;
	    }
	  //first selected
	  else if (itemp1 == -5
		   && itemp2 < fr->axarray->getarraylength() 
		   && 0 <= itemp2)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      (*fr->axarray)[i] = (*fr->axarray)[itemp2];
		      return true;
		    }
		}
	      return true;
	    }
	  else if (itemp1 < fr->axarray->getarraylength() 
		   && 0 <= itemp1
		   && itemp2  == -5)
	    {
	      for (int i = 0; i < fr->axarray->getarraylength(); i++)
		{
		  if ((*fr->axarray)[i].selected)
		    {
		      (*fr->axarray)[itemp1] = (*fr->axarray)[i];
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // data object
      else if ((*argument)[0] == std::string("data"))
	{
	  itemp1 = atoi((*argument)[1].c_str());
	  itemp2 = atoi((*argument)[2].c_str());
	  
	  if (itemp1 < fr->darray->getarraylength() 
	      && 0 <= itemp1
	      && itemp2 < fr->darray->getarraylength() 
	      && 0 <= itemp2)
	    {
	      (*fr->darray)[itemp1] = (*fr->darray)[itemp2];
	      
	      (*fr->darray)[itemp1].gettmpdataname(&str1);
	      str1 = std::string(tmppath) + std::string("/") + str1;
	      (*fr->darray)[itemp2].gettmpdataname(&str2);
	      str2 = std::string(tmppath) + std::string("/") + str2;
	      if (isfileexist((char *)str2.c_str()) && !(*fr->darray)[itemp2].virgin && fr->withdata)
		filecp((char *)str1.c_str(), (char *)str2.c_str());
	      return true;
	    }
	  //first selected
	  else if (itemp1 == -5
		   && itemp2 < fr->darray->getarraylength() 
		   && 0 <= itemp2)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      (*fr->darray)[i] = (*fr->darray)[itemp2];
		      
		      (*fr->darray)[i].gettmpdataname(&str1);
		      str1 = std::string(tmppath) + std::string("/") + str1;
		      (*fr->darray)[itemp2].gettmpdataname(&str2);
		      str2 = std::string(tmppath) + std::string("/") + str2;
		      if (isfileexist((char *)str2.c_str()) && !(*fr->darray)[itemp2].virgin && fr->withdata)
			filecp((char *)str1.c_str(), (char *)str2.c_str());
		      return true;
		    }
		}
	      return true;
	    }
	  else if (itemp1 < fr->darray->getarraylength() 
		   && 0 <= itemp1
		   && itemp2  == -5)
	    {
	      for (int i = 0; i < fr->darray->getarraylength(); i++)
		{
		  if ((*fr->darray)[i].selected)
		    {
		      (*fr->darray)[itemp1] = (*fr->darray)[i];
		      
		      (*fr->darray)[itemp1].gettmpdataname(&str1);
		      str1 = std::string(tmppath) + std::string("/") + str1;
		      (*fr->darray)[i].gettmpdataname(&str2);
		      str2 = std::string(tmppath) + std::string("/") + str2;
		      if (isfileexist((char *)str2.c_str()) && !(*fr->darray)[i].virgin && fr->withdata)
			filecp((char *)str1.c_str(), (char *)str2.c_str());
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      // parts object
      else if ((*argument)[0] == std::string("parts"))
	{
	  itemp1 = atoi((*argument)[1].c_str());
	  itemp2 = atoi((*argument)[2].c_str());
	  
	  if (itemp1 < fr->parray->getarraylength() 
	      && 0 <= itemp1
	      && itemp2 < fr->parray->getarraylength() 
	      && 0 <= itemp2)
	    {
	      (*fr->parray)[itemp1] = (*fr->parray)[itemp2];
	      return true;
	    }
	  //first selected
	  else if (itemp1 == -5
		   && itemp2 < fr->parray->getarraylength() 
		   && 0 <= itemp2)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      (*fr->parray)[i] = (*fr->parray)[itemp2];
		      return true;
		    }
		}
	      return true;
	    }
	  else if (itemp1 < fr->parray->getarraylength() 
		   && 0 <= itemp1
		   && itemp2  == -5)
	    {
	      for (int i = 0; i < fr->parray->getarraylength(); i++)
		{
		  if ((*fr->parray)[i].selected)
		    {
		      (*fr->parray)[itemp1] = (*fr->parray)[i];
		      return true;
		    }
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_usrconstlen(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      itemp = fr->usrconst->getarraylength();
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_setusrconst(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      if (IsValidConstant(&(*argument)[0]))
	{
	  gra->updated = true;
	  fr->usrconst->SetScalar(&(*argument)[0], &(*argument)[1]);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_getusrconst(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str1;

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      if (fr->usrconst->GetScalar(&(*argument)[0], &str1))
	{
	  outvalue->value->add(str1.c_str());
	  return true;
	}
      else
	return false;
      return true;
    }
  else
    return false;
}

bool fr_getallusrconst(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str1;

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      for (int i = 0; i < fr->usrconst->getarraylength(); i++)
	{
	  str1 = std::string("");
	  str1 += (*fr->usrconst)[i].label;
	  str1 += std::string(" : ");
	  str1 += (*(*fr->usrconst)[i].value)[0];
	  outvalue->value->add(str1.c_str());
	}
      return true;
    }
  else
    return false;
}

bool fr_rmusrconst(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  int itemp;
  if (arglength == 1 && iid == -1)
    {
      if (fr->usrconst->search(&(*argument)[0], &itemp))
	{
	  gra->updated = true;
	  fr->usrconst->remove(itemp);
	  outvalue->value->add("1");
	}
      else
	outvalue->value->add("0");
      return true;
    }
  else
    return false;
}

bool fr_flushusrconst(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      gra->updated = true;
      fr->usrconst->setbuff(0);
      return true;
    }
  else
    return false;
}

bool fr_usrfnclen(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  int itemp;
  char stemp[100];

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      itemp = fr->usrfnc->getarraylength();
      sprintf(stemp, "%d", itemp);  
      outvalue->value->add(stemp);
      return true;
    }
  else
    return false;
}

bool fr_setusrfnc(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 2 && iid == -1)
    {
      if (IsValidFunction(&(*argument)[0]))
	{
	  gra->updated = true;
	  fr->usrfnc->SetScalar(&(*argument)[0], &(*argument)[1]);
	  return true;
	}
      else
	return false;
    }
  else
    return false;
}

bool fr_getusrfnc(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str1;

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1)
    {
      if (fr->usrfnc->GetScalar(&(*argument)[0], &str1))
	{
	  outvalue->value->add(str1.c_str());
	  return true;
	}
      else
	return false;
      return true;
    }
  else
    return false;
}

bool fr_getallusrfnc(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{
  std::string str1;

  if (fr == 0 || outvalue == 0)
    return false;

  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      for (int i = 0; i < fr->usrfnc->getarraylength(); i++)
	{
	  str1 = std::string("");
	  str1 += (*fr->usrfnc)[i].label;
	  str1 += std::string(" : ");
	  str1 += (*(*fr->usrfnc)[i].value)[0];
	  outvalue->value->add(str1.c_str());
	}
      return true;
    }
  else
    return false;
}

bool fr_rmusrfnc(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  int itemp;
  if (arglength == 1 && iid == -1)
    {
      if (fr->usrfnc->search(&(*argument)[0], &itemp))
	{
	  gra->updated = true;
	  fr->usrfnc->remove(itemp);
	  outvalue->value->add("1");
	}
      else
	outvalue->value->add("0");
      return true;
    }
  else
    return false;
}

bool fr_flushusrfnc(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue)
{

  if (fr == 0 || outvalue == 0)
    return false;

  graph* gra = (graph *)fr->parent;
  if (arglength == 1 && iid == -1 && (*argument)[0] == std::string("null"))
    {
      gra->updated = true;
      fr->usrfnc->setbuff(0);
      return true;
    }
  else
    return false;
}

typedef struct frfuncs
{
  char *function;
  bool (*func)(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue);
}FRFUNCS;

FRFUNCS frf[] = {
  {"paint", fr_paint},
  {"setbbox", fr_setbbox},
  {"setframelinecolor", fr_setframelinecolor},
  {"setframebodycolor", fr_setframebodycolor},
  {"arrangeaxes", fr_arrangeaxes},
  {"wtov_x", fr_wtov_x},
  {"wtov_y", fr_wtov_y},
  {"vtow_x", fr_vtow_x},
  {"vtow_y", fr_vtow_y},
  {"autoscale", fr_autoscale},
  {"autounit", fr_autounit},
  {"getindex", fr_getindex},
  {"delete", fr_delete},
  {"new", fr_new},
  {"flushselect", fr_flushselect},
  {"length", fr_length},
  {"selectednumber", fr_selectednumber},
  {"top", fr_top},
  {"up", fr_up},
  {"down", fr_down},
  {"bottom", fr_bottom},
  {"copy", fr_copy},
  {"usrconstlen", fr_usrconstlen},
  {"setusrconst", fr_setusrconst},
  {"getusrconst", fr_getusrconst},
  {"getallusrconst", fr_getallusrconst},
  {"rmusrconst", fr_rmusrconst},
  {"flushusrconst", fr_flushusrconst},
  {"usrfnclen", fr_usrfnclen},
  {"setusrfnc", fr_setusrfnc},
  {"getusrfnc", fr_getusrfnc},
  {"getallusrfnc", fr_getallusrfnc},
  {"rmusrfnc", fr_rmusrfnc},
  {"flushusrfnc", fr_flushusrfnc},
  {0, 0}
}; 


bool frame::exec(std::string* function, buffarray *argument, tokenbuff* outvalue)
{
  std::string out, shrunk, str1, str2, str3;
  int iid, arglength;

  FRFUNCS *frfncs;
  int itemno;
  bool (*execfunc)(frame *fr, int arglength, int iid, buffarray *argument, tokenbuff* outvalue);

  outvalue->value->setbuff(0);
  if (topobject(function, &out, &iid) == true)
    {
      arglength = argument->getarraylength();

      // transfer message to axis
      if (out == std::string("axis"))
	{
	  if (shrinkobject(function, &shrunk))
	    {
	      if (iid < axarray->getarraylength() && iid >= 0)
		{
		  if ((*axarray)[iid].exec(&shrunk, argument, outvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if ((*axarray)[i].selected)
			{
			  if (!(*axarray)[i].exec(&shrunk, argument, outvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if (!(*axarray)[i].exec(&shrunk, argument, outvalue))
			return false;
		    }
		  return true;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < axarray->getarraylength(); i++)
		    {
		      if ((*axarray)[i].selected)
			{
			  if ((*axarray)[i].exec(&shrunk, argument, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // transfer message to data
      else if (out == std::string("data"))
	{
	  if (shrinkobject(function, &shrunk))
	    {
	      if (iid < darray->getarraylength() && iid >= 0)
		{
		  if ((*darray)[iid].exec(&shrunk, argument, outvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if ((*darray)[i].selected)
			{
			  if (!(*darray)[i].exec(&shrunk, argument, outvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if (!(*darray)[i].exec(&shrunk, argument, outvalue))
			return false;
		    }
		  return true;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < darray->getarraylength(); i++)
		    {
		      if ((*darray)[i].selected)
			{
			  if ((*darray)[i].exec(&shrunk, argument, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // transfer message to parts
      else if (out == std::string("parts"))
	{
	  if (shrinkobject(function, &shrunk))
	    {
	      if (iid < parray->getarraylength() && iid >= 0)
		{
		  if ((*parray)[iid].exec(&shrunk, argument, outvalue))
		    return true;
		  else
		    return false;
		}
	      // selected
	      else if (iid == -3)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if ((*parray)[i].selected)
			{
			  if (!(*parray)[i].exec(&shrunk, argument, outvalue))
			    return false;
			}
		    }
		  return true;
		}
	      // all
	      else if (iid == -4)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if (!(*parray)[i].exec(&shrunk, argument, outvalue))
			return false;
		    }
		  return true;
		}
	      // first selected
	      else if (iid == -5)
		{
		  for (int i = 0; i < parray->getarraylength(); i++)
		    {
		      if ((*parray)[i].selected)
			{
			  if ((*parray)[i].exec(&shrunk, argument, outvalue))
			    return true;
			  else
			    return false;
			}
		    }
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else
	{
	  frfncs = frf;
	  for (; frfncs->function; frfncs++)
	    {
	    }
	  itemno = frfncs - frf;
	  for (int i = 0; i < itemno; i++)
	    {
	      if (strcmp(out.c_str(), frf[i].function) == 0)
		{
		  execfunc = frf[i].func;
		  return execfunc(this, arglength, iid, argument, outvalue);
		}
	    }
	  return false;
	}
    }
  else
    return false;
      
  return true;
}











