/*----------------------------------------------------------------------------
                      frame.h (definitions of frame object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__frame_h)
#define __frame_h

#include "plobj.h"
#include "pointdefs.h"
#include <string>
#include "col.h"
#include "vect.h"
#include "axis.h"
#include "data.h"
#include "parts.h"
#include "valbuff.h"

typedef vect<axis> axisarray;
typedef vect<data> dataarray;
typedef vect<parts> partsarray;

#define FRAME_BOX       0
#define FRAME_XY        1
#define FRAME_INDIV     2
 

class frame: public plobj
{
 public:
  frame();
  ~frame();
  
  frame& operator = (const frame& fr);
  
  bool get(std::string* member,  tokenbuff* outvalue);  
  bool set(std::string* member,  std::string* setvalue);  
  bool exec(std::string* function, buffarray *argument, tokenbuff* outvalue);  
  void arrangeaxes();

  bool paint(bool flag, FILE* f);
  int wtov_x(int wx);
  int wtov_y(int wy);
  int scale(int x);
  double doubleworld_to_view_x(double wx);
  double doubleworld_to_view_y(double wy);
  int doubleworldtoshortworld(double x);

  int vtow_x(int vx);
  int vtow_y(int vy);
  double doubleview_to_world_x(double vx);
  double doubleview_to_world_y(double vy);


  bool axispaintmajorline(bool flag, FILE* f, axis *axis);
  bool axispaintminorline(bool flag, FILE* f, axis *axis);

  bool setaxisid(int arrayno);
  bool setdataid(int arrayno);
  bool setpartsid(int arrayno);
  
  int getaxisindex(int id);
  int getdataindex(int id);
  int getpartsindex(int id);

  void getBBbox(intrect *r);

  void autoscaleunit(axis *AUnit, double Start, double End, bool Spacing, int Spacingfactor);
  void autoscale(bool BFlag, bool LFlag, bool TFlag, bool RFlag, bool Spacing, int Spacingfactor);


  /* frame read / write  */
  bool readframe(FILE *fp);
  bool writeframe(FILE *fp);
  void setwithdata();

  int id;
  std::string *label;
  std::string *paintorder;
  axisarray *axarray;
  dataarray *darray;
  partsarray *parray;
  VariantBuff *usrconst, *usrfnc;

  intpoint p1, p2;
  Capcolor framelinecolor, framebodycolor;
  Caplinestyle framelinestyle;
  Capfillstyle framebodyfillstyle;
  int frametype;
  int selected;
  bool withdata;
};

#endif



