/*----------------------------------------------------------------------------
                 frame2.cc (functions related to frame objects)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "frame.h"
#include "graph.h"
#include "axis.h"
#include "data.h"
#include "gdi.h"
#include "vectdefs.h"
#include "pointdefs.h"
#include "topazvalues.h"

extern void message(const char *msg);
extern void StyleLinePath(bool flag, FILE *f, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);
extern bool IsValidConstant(std::string *str);
extern bool IsValidFunction(std::string *str);

bool frame::paint(bool flag, FILE* f)
{
  intpointarray *ia;
  ia = new intpointarray;
  
  plinestyle(flag, f, scale(framelinestyle.width), framelinestyle.cap, framelinestyle.join, framelinestyle.style);
  pfillstyle(flag, f, framebodyfillstyle.fillrule, framebodyfillstyle.style);
  psetforecolor(flag, f, framelinecolor.red, framelinecolor.green, framelinecolor.blue);
  psetbackcolor(flag, f, framebodycolor.red, framebodycolor.green, framebodycolor.blue);
  
  ia->setbuff(0);
  ia->add(new intpoint(wtov_x(0), wtov_y(0)));
  ia->add(new intpoint(wtov_x(p2.x - p1.x), wtov_y(0)));
  ia->add(new intpoint(wtov_x(p2.x - p1.x), wtov_y(p2.y - p1.y)));
  ia->add(new intpoint(wtov_x(0), wtov_y(p2.y - p1.y)));
  ppolygon(flag, f, ia);

  delete ia;
  return true;
}

int frame::wtov_x(int wx)
{
  return (wx + p1.x);
}

int frame::wtov_y(int wy)
{
  return (wy + p1.y);
}

int frame::scale(int x)
{
  return x;
}

double frame::doubleworld_to_view_x(double wx)
{
  return (wx + (double)p1.x);
}

double frame::doubleworld_to_view_y(double wy)
{
  return (wy + (double)p1.y);
}

int frame::doubleworldtoshortworld(double x)
{
  if (fabs(x) >= (double)MAXSHORT)
    return MAXSHORT;
  else
    return ((int)x);
}


int frame::vtow_x(int vx)
{
  return (vx - p1.x);
}

int frame::vtow_y(int vy)
{
  return (vy - p1.y);
}

double frame::doubleview_to_world_x(double vx)
{
  return ((double)vx - (double)p1.x);
}

double frame::doubleview_to_world_y(double vy)
{
  return ((double)vy - (double)p1.y);
}

void frame::arrangeaxes()
{
  axis *ax, *ay, *at, *ar;
  intpoint ip;
  int itemp;

  ax = &(*axarray)[0];
  ay = &(*axarray)[1];
  at = &(*axarray)[2];
  ar = &(*axarray)[3];

  switch(frametype)
    {
    case FRAME_BOX:
      ax->origin.x      = 0;
      ax->origin.y      = p2.y - p1.y;
      ax->length        = p2.x - p1.x;
      ax->angle         = 0.0;
      ax->fo.angle      = 0.0;

      ay->display       = ax->display;
      ay->origin.x      = 0;
      ay->origin.y      = p2.y - p1.y;
      ay->length        = p2.y - p1.y;
      ay->angle         = 90.0;
      ay->tickwidth     = ax->tickwidth;
      ay->ticklength    = ax->ticklength;
      ay->fo.angle      = 270.0;


      at->display       = ax->display;
      at->origin.x      = 0;
      at->origin.y      = 0;
      at->length        = p2.x - p1.x;
      at->angle         = 0.0;
      at->crosspoint    = ax->crosspoint;
      at->autocrosspoint = ax->autocrosspoint;
      at->tickwidth     = ax->tickwidth;
      at->ticklength    = ax->ticklength;
      at->fo.angle      = 0.0;
      
      if (ax->commonset)
	{
	  at->scaling      = ax->scaling;
	  at->invertscale  = ax->invertscale;
	  at->start        = ax->start;
	  at->end          = ax->end;
	  at->unit         = ax->unit;
	  
	  switch(ax->tickparity)
	    {
	    case TICKPARITY_NULL:
	    case TICKPARITY_BOTH:
	      at->tickparity   = ax->tickparity;
	      break;
	    case TICKPARITY_PLUS:
	      at->tickparity   = TICKPARITY_MINUS;
	      break;
	    case TICKPARITY_MINUS:
	      at->tickparity   = TICKPARITY_PLUS;
	      break;
	    }

	  at->division     = ax->division;
	  at->baselinestyle = ax->baselinestyle;
	  at->basecolor     = ax->basecolor;
	  at->tickarea_start = ax->tickarea_start;
	  at->tickarea_end  = ax->tickarea_end;
	  
	  at->majorticklinecolor = ax->majorticklinecolor;
	  at->minorticklinecolor = ax->minorticklinecolor;
	  at->majorticklinestyle = ax->majorticklinestyle;
	  at->minorticklinestyle = ax->minorticklinestyle;
	  at->fo.face       = ax->fo.face;
	  at->fo.size       = ax->fo.size;
	  at->fontcolor     = ax->fontcolor;
	  at->addplus       = ax->addplus;
	  at->offset        = ax->offset;
	  at->skip          = ax->skip;
	  at->autoprecision = ax->autoprecision;
	  at->intprecision  = ax->intprecision;
	  at->decimalprecision = ax->decimalprecision;
	}

      ar->display       = ax->display;
      ar->origin.x      = p2.x - p1.x;
      ar->origin.y      = p2.y - p1.y;
      ar->length        = p2.y - p1.y;
      ar->angle         = 90.0;
      ar->crosspoint    = ay->crosspoint;
      ar->autocrosspoint = ay->autocrosspoint;
      ar->tickwidth     = ax->tickwidth;
      ar->ticklength    = ax->ticklength;
      ar->fo.angle      = 270.0;

      if (ay->commonset)
	{
	  ar->scaling      = ay->scaling;
	  ar->invertscale  = ay->invertscale;
	  ar->start        = ay->start;
	  ar->end          = ay->end;
	  ar->unit         = ay->unit;

	  switch(ay->tickparity)
	    {
	    case TICKPARITY_NULL:
	    case TICKPARITY_BOTH:
	      ar->tickparity   = ay->tickparity;
	      break;
	    case TICKPARITY_PLUS:
	      ar->tickparity   = TICKPARITY_MINUS;
	      break;
	    case TICKPARITY_MINUS:
	      ar->tickparity   = TICKPARITY_PLUS;
	      break;
	    }

	  ar->division     = ay->division;
	  ar->baselinestyle = ay->baselinestyle;
	  ar->basecolor     = ay->basecolor;
	  ar->tickarea_start = ay->tickarea_start;
	  ar->tickarea_end  = ay->tickarea_end;
	  ar->majorticklinecolor = ay->majorticklinecolor;
	  ar->minorticklinecolor = ay->minorticklinecolor;
	  ar->majorticklinestyle = ay->majorticklinestyle;
	  ar->minorticklinestyle = ay->minorticklinestyle;
	  ar->fo.face       = ay->fo.face;
	  ar->fo.size       = ay->fo.size;
	  ar->fontcolor     = ay->fontcolor;
	  ar->addplus       = ay->addplus;
	  ar->offset        = ay->offset;
	  ar->skip          = ay->skip;
	  ar->autoprecision = ay->autoprecision;
	  ar->intprecision  = ay->intprecision;
	  ar->decimalprecision = ay->decimalprecision;
	}

      break;
    case FRAME_XY:

      if (ay->autocrosspoint)
	{
	  itemp = p2.y - p1.y;
	}
      else
	{
	  ip = ay->virtual_to_world(ay->crosspoint);
	  itemp = ip.y;
	}
      
      if (itemp < 0 || itemp > p2.y - p1.y)
	itemp = 0;


      ax->origin.x      = 0;
      ax->origin.y      = itemp;
      ax->length        = p2.x - p1.x;
      ax->angle         = 0.0;
      ax->crosspointoffset = ay->offset;
      ax->fo.angle      = 0.0;

      if (ax->autocrosspoint)
	{
	  itemp = 0;
	}
      else
	{
	  ip = ax->virtual_to_world(ax->crosspoint);
	  itemp = ip.x;
	}

      if (itemp < 0 || itemp > p2.x - p1.x)
	itemp = 0;

      ay->display       = ax->display;
      ay->origin.x      = itemp;
      ay->origin.y      = p2.y - p1.y;
      ay->length        = p2.y - p1.y;
      ay->angle         = 90.0;
      ay->baselinestyle = ax->baselinestyle;
      ay->basecolor     = ax->basecolor;
      ay->tickwidth     = ax->tickwidth;
      ay->ticklength    = ax->ticklength;
      ay->crosspointoffset = ax->offset;
      ay->fo.angle      = 270.0;

      at->display       = false;
      at->origin.x      = 0;
      at->origin.y      = 0;
      at->length        = p2.x - p1.x;
      at->angle         = 0.0;
      at->baselinestyle = ax->baselinestyle;
      at->basecolor     = ax->basecolor;
      at->crosspoint    = ax->crosspoint;
      at->autocrosspoint = ax->autocrosspoint;
      at->tickwidth     = ax->tickwidth;
      at->ticklength    = ax->ticklength;
      at->fo.angle      = 0.0;
      
      if (ax->commonset)
	{
	  at->scaling      = ax->scaling;
	  at->tickarea_start = ax->tickarea_start;
	  at->tickarea_end  = ax->tickarea_end;
	  at->invertscale  = ax->invertscale;
	  at->start        = ax->start;
	  at->end          = ax->end;
	  at->unit         = ax->unit;
	  at->tickparity   = ax->tickparity;
	  at->division     = ax->division;
	  at->majorticklinecolor = ax->majorticklinecolor;
	  at->minorticklinecolor = ax->minorticklinecolor;
	  at->majorticklinestyle = ax->majorticklinestyle;
	  at->minorticklinestyle = ax->minorticklinestyle;
	  at->fo.face       = ax->fo.face;
	  at->fo.size       = ax->fo.size;
	  at->fontcolor     = ax->fontcolor;
	  at->addplus       = ax->addplus;
	  at->offset        = ax->offset;
	  at->skip          = ax->skip;
	  at->autoprecision = ax->autoprecision;
	  at->intprecision  = ax->intprecision;
	  at->decimalprecision = ax->decimalprecision;
	}

      ar->display       = false;
      ar->origin.x      = p2.x - p1.x;
      ar->origin.y      = p2.y - p1.y;
      ar->length        = p2.y - p1.y;
      ar->angle         = 90.0;
      ar->baselinestyle = ax->baselinestyle;
      ar->basecolor     = ax->basecolor;
      ar->crosspoint    = ay->crosspoint;
      ar->autocrosspoint = ay->autocrosspoint;
      ar->tickwidth     = ax->tickwidth;
      ar->ticklength    = ax->ticklength;
      ar->fo.angle      = 270.0;

      if (ay->commonset)
	{
	  ar->scaling      = ay->scaling;
	  ar->tickarea_start = ay->tickarea_start;
	  ar->tickarea_end  = ay->tickarea_end;
	  ar->invertscale  = ay->invertscale;
	  ar->start        = ay->start;
	  ar->end          = ay->end;
	  ar->unit         = ay->unit;
	  ar->tickparity   = ay->tickparity;
	  ar->division     = ay->division;
	  ar->majorticklinecolor = ay->majorticklinecolor;
	  ar->minorticklinecolor = ay->minorticklinecolor;
	  ar->majorticklinestyle = ay->majorticklinestyle;
	  ar->minorticklinestyle = ay->minorticklinestyle;
	  ar->fo.face       = ay->fo.face;
	  ar->fo.size       = ay->fo.size;
	  ar->fontcolor     = ay->fontcolor;
	  ar->addplus       = ay->addplus;
	  ar->offset        = ay->offset;
	  ar->skip          = ay->skip;
	  ar->autoprecision = ay->autoprecision;
	  ar->intprecision  = ay->intprecision;
	  ar->decimalprecision = ay->decimalprecision;
	}
      break;
    case FRAME_INDIV:
      /*
      ax->angle         = 0.0;
      ay->angle         = 90.0;
      at->angle         = 0.0;
      ar->angle         = 90.0;
      ax->fo.angle      = 0.0;
      ay->fo.angle      = 270.0;
      at->fo.angle      = 0.0;
      ar->fo.angle      = 270.0;
      */
      break;
    }
}

bool frame::axispaintmajorline(bool flag, FILE *f, axis *axis)
{
  int linestart, lineend;
  double virstart, virend, startval, endval;
  double delta;
  int itercount;
  intpoint valpos, fixedpos;
  doublepoint tempdpoint;
  double fa;
  doublepointarray *da;
  intrect axisrect, cliprect;

  if (axis->display == false)
    return true;

  da = new doublepointarray;

  // determine tickstart, tickend
  switch(axis->axistype)
    {
    case AXISTYPE_X:
    case AXISTYPE_T:
      linestart = wtov_y(0); lineend = wtov_y(p2.y - p1.y);
      break;
    case AXISTYPE_Y:
    case AXISTYPE_R:
      linestart = wtov_x(0); lineend = wtov_x(p2.x - p1.x);
      break;
    default:
      return false;
    }

  // majorline style
  plinestyle(flag, f, scale(axis->majorticklinestyle.width), axis->majorticklinestyle.cap,
	     axis->majorticklinestyle.join, LINESTYLESOLID);

  //majorlinecolor
  psetforecolor(flag, f, axis->majorticklinecolor.red, axis->majorticklinecolor.green, axis->majorticklinecolor.blue);

  switch(axis->scaling)
    {
    case SCALE_LINEAR:
    case SCALE_LOGLINEAR:
    case SCALE_INVLINEAR:

      if (axis->unit == 0.0)
	{
	  message("Invalid unit!!\n");
	  delete da;
	  return false;
	}

      virstart = floor(axis->start / axis->unit) * axis->unit;
      virend = floor(axis->end / axis->unit) * axis->unit;
      virend += axis->unit;
      delta           = DELTA_AXIS * (axis->end - axis->start);
      startval        = axis->start      - delta;
      endval          = axis->end        + delta;
      itercount = (int)((virend - virstart) / axis->unit) + 2;

      if (itercount >100)
	{
	  message("Too much ticks!!\n");
	  delete da;
	  return false;
	}

      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart + (double)i * axis->unit;
	  if ((fa >= startval) && (fa <= endval))
	    {
	      switch(axis->axistype)
		{
		case AXISTYPE_X:
		case AXISTYPE_T:
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  valpos.x = wtov_x((int)tempdpoint.x);
		  valpos.y = linestart;
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  fixedpos.x = wtov_x((int)tempdpoint.x);
		  fixedpos.y = lineend;
		  break;
		case AXISTYPE_Y:
		case AXISTYPE_R:
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  valpos.x = linestart;
		  valpos.y = wtov_y((int)tempdpoint.y);
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  fixedpos.x = lineend;
		  fixedpos.y = wtov_y((int)tempdpoint.y);
		  break;
	      }


	      pbeginpath(flag, f); 
	      da->setbuff(0);
	      da->add(new doublepoint(fixedpos.x, fixedpos.y));
	      da->add(new doublepoint(valpos.x, valpos.y));
	      StyleLinePath(flag, f, da, da->getarraylength() - 1,
			    axis->majorticklinestyle.style, axis->majorticklinestyle.width,
			    false, &axisrect, &cliprect);
	      pstroke(flag, f); 
	    }
	}
      break;
    case SCALE_LOG:
    case SCALE_INV:
      virstart = pow(10.0, floor(log10(axis->start)));
      virend   = pow(10.0, floor(log10(axis->end)));
      virend += 1.0;
      delta           = DELTA_AXIS * (log10(axis->end) - log10(axis->start));
      startval        = axis->start      / pow(10.0, delta);
      endval          = axis->end        * pow(10.0, delta);
      
      itercount = (int)(log10(virend / virstart)) + 2;
      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart * pow(10.0, (double)i);
	  if ((fa >= startval) && (fa <= endval))
	    {
	      switch(axis->axistype)
		{
		case AXISTYPE_X:
		case AXISTYPE_T:
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  valpos.x = wtov_x((int)tempdpoint.x);
		  valpos.y = linestart;
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  fixedpos.x = wtov_x((int)tempdpoint.x);
		  fixedpos.y = lineend;
		  break;
		case AXISTYPE_Y:
		case AXISTYPE_R:
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  valpos.x = linestart;
		  valpos.y = wtov_y((int)tempdpoint.y);
		  tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		  fixedpos.x = lineend;
		  fixedpos.y = wtov_y((int)tempdpoint.y);
		  break;
	      }

	      pbeginpath(flag, f); 
	      da->setbuff(0);
	      da->add(new doublepoint(fixedpos.x, fixedpos.y));
	      da->add(new doublepoint(valpos.x, valpos.y));
	      StyleLinePath(flag, f, da, da->getarraylength() - 1, 
			    axis->majorticklinestyle.style, axis->majorticklinestyle.width,
			    false, &axisrect, &cliprect);
	      pstroke(flag, f); 
	    }
	}
      break;
    }

  delete da;
  return true;
}

bool frame::axispaintminorline(bool flag, FILE* f, axis *axis)
{
  int linestart, lineend;
  double virstart, virend, startval, endval;
  double delta;
  int itercount;
  intpoint valpos, fixedpos;
  doublepoint tempdpoint;
  double fa;
  int logminortickstart, logminortickend, logminortickunit;
  doublepointarray *da;
  intrect axisrect, cliprect;

  if (axis->display == false)
    return true;

  da = new doublepointarray;

  // determine tickstart, tickend
  switch(axis->axistype)
    {
    case AXISTYPE_X:
    case AXISTYPE_T:
      linestart = wtov_y(0); lineend = wtov_y(p2.y - p1.y);
      break;
    case AXISTYPE_Y:
    case AXISTYPE_R:
      linestart = wtov_x(0); lineend = wtov_x(p2.x - p1.x);
      break;
    default:
      return false;
    }

  // majorline style
  plinestyle(flag, f, scale(axis->minorticklinestyle.width), axis->minorticklinestyle.cap,
	     axis->minorticklinestyle.join, LINESTYLESOLID);

  //majorlinecolor
  psetforecolor(flag, f, axis->minorticklinecolor.red, axis->minorticklinecolor.green, axis->minorticklinecolor.blue);

  switch(axis->scaling)
    {
    case SCALE_LINEAR:
    case SCALE_LOGLINEAR:
    case SCALE_INVLINEAR:

      if (axis->unit == 0.0)
	{
	  message("Invalid unit!!\n");
	  delete da;
	  return false;
	}

      virstart = floor(axis->start / axis->unit) * axis->unit;
      virend = floor(axis->end / axis->unit) * axis->unit;
      virend += axis->unit;
      delta           = DELTA_AXIS * (axis->end - axis->start);
      startval        = axis->start      - delta;
      endval          = axis->end        + delta;
      itercount = (int)((virend - virstart) / axis->unit * (double)axis->division) + 2;

      if (itercount >100)
	{
	  message("Too much ticks!!\n");
	  delete da;
	  return false;
	}

      for (int i = 0; i <= itercount; i++)
	{
	  fa = virstart + (double)i * axis->unit / (double)axis->division;
	  if ((fa >= startval) && (fa <= endval))
	    {
	      if ((i % axis->division) != 0)         // except tick position
		{
		  switch(axis->axistype)
		    {
		    case AXISTYPE_X:
		    case AXISTYPE_T:
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      valpos.x = wtov_x((int)tempdpoint.x);
		      valpos.y = linestart;
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      fixedpos.x = wtov_x((int)tempdpoint.x);
		      fixedpos.y = lineend;
		      break;
		    case AXISTYPE_Y:
		    case AXISTYPE_R:
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      valpos.x = linestart;
		      valpos.y = wtov_y((int)tempdpoint.y);
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      fixedpos.x = lineend;
		      fixedpos.y = wtov_y((int)tempdpoint.y);
		      break;
		    }
		  
		  pbeginpath(flag, f); 
		  da->setbuff(0);
		  da->add(new doublepoint(fixedpos.x, fixedpos.y));
		  da->add(new doublepoint(valpos.x, valpos.y));
		  StyleLinePath(flag, f, da, da->getarraylength() - 1, 
				axis->minorticklinestyle.style, axis->minorticklinestyle.width,
				false, &axisrect, &cliprect);
		  pstroke(flag, f); 
		}
	    }
	}
      break;
    case SCALE_LOG:
    case SCALE_INV:
      switch(axis->division)
	{
	case 1:
	  logminortickstart       = 1;
	  logminortickend         = -1;
	  logminortickunit        = 1;
	  break;
	case 2:
	  logminortickstart       = 5;
	  logminortickend         = 5;
	  logminortickunit        = 1;
	  break;
	case 5:
	  logminortickstart       = 2;
	  logminortickend         = 8;
	  logminortickunit        = 2;
	  break;
	case 10:
	default:
	  logminortickstart       = 2;
	  logminortickend         = 9;
	  logminortickunit        = 1;
	  break;
	}

      virstart = pow(10.0, floor(log10(axis->start)));
      virend   = pow(10.0, floor(log10(axis->end)));
      virend += 1.0;
      delta           = DELTA_AXIS * (log10(axis->end) - log10(axis->start));
      startval        = axis->start      / pow(10.0, delta);
      endval          = axis->end        * pow(10.0, delta);
      
      itercount = (int)(log10(virend / virstart)) + 2;
      for (int i = 0; i <= itercount; i++)
	{
	  for (int j = logminortickstart; j <= logminortickend; j = j + logminortickunit)
	    {
	      fa = virstart * pow(10.0, (double)i) * (double)j;
	      if ((fa >= startval) && (fa <= endval))
		{
		  switch(axis->axistype)
		    {
		    case AXISTYPE_X:
		    case AXISTYPE_T:
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      valpos.x = wtov_x((int)tempdpoint.x);
		      valpos.y = linestart;
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      fixedpos.x = wtov_x((int)tempdpoint.x);
		      fixedpos.y = lineend;
		      break;
		    case AXISTYPE_Y:
		    case AXISTYPE_R:
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      valpos.x = linestart;
		      valpos.y = wtov_y((int)tempdpoint.y);
		      tempdpoint = axis->axis_to_world(axis->virtual_to_axis(fa), 0);
		      fixedpos.x = lineend;
		      fixedpos.y = wtov_y((int)tempdpoint.y);
		      break;
		    }
		  
		  pbeginpath(flag, f); 
		  da->setbuff(0);
		  da->add(new doublepoint(fixedpos.x, fixedpos.y));
		  da->add(new doublepoint(valpos.x, valpos.y));
		  StyleLinePath(flag, f, da, da->getarraylength() - 1, 
				axis->minorticklinestyle.style, axis->minorticklinestyle.width,
				false, &axisrect, &cliprect);
		  pstroke(flag, f); 
		}
	    }
	}
      break;
    }
  delete da;
  return true;
}

bool frame::setaxisid(int arrayno)
{
  // calc max id
  int maxid = -1;
  for (int i = 0; i < axarray->getarraylength(); i++)
    {
      if (i != arrayno && maxid < (*axarray)[i].id)
	maxid = (*axarray)[i].id;
    }

  if (maxid == -1)
    {
      (*axarray)[arrayno].id = 0;
      return true;
    }

  // create table
  bool *buf = new bool [(maxid + 1) * sizeof(bool)];
  if (buf == 0)
    return false;
  memset(buf, 0, (maxid + 1) * sizeof(bool));

  // set table
  for (int i = 0; i < axarray->getarraylength(); i++)
    {
      if (i != arrayno)
	buf[(*axarray)[i].id] = true;
    }

  // search id
  for (int i = 0; i <= maxid; i++)
    {
      if (buf[i] == 0)
	{
	  (*axarray)[arrayno].id = i;
	  delete buf;
	  return true;
	}
    }
  delete buf;
  (*axarray)[arrayno].id = maxid + 1;
  return true;
}

int frame::getaxisindex(int id)
{
  for (int i = 0; i < axarray->getarraylength(); i++)
    {
      if (id == (*axarray)[i].id)
	return i;
    }
  return -1;
}

bool frame::setdataid(int arrayno)
{
  // calc max id
  int maxid = -1;
  for (int i = 0; i < darray->getarraylength(); i++)
    {
      if (i != arrayno && maxid < (*darray)[i].id)
	maxid = (*darray)[i].id;
    }

  if (maxid == -1)
    {
      (*darray)[arrayno].id = 0;
      return true;
    }

  // create table
  bool *buf = new bool [(maxid + 1) * sizeof(bool)];
  if (buf == 0)
    return false;
  memset(buf, 0, (maxid + 1) * sizeof(bool));

  // set table
  for (int i = 0; i < darray->getarraylength(); i++)
    {
      if (i != arrayno)
	buf[(*darray)[i].id] = true;
    }

  // search id
  for (int i = 0; i <= maxid; i++)
    {
      if (buf[i] == 0)
	{
	  (*darray)[arrayno].id = i;
	  delete buf;
	  return true;
	}
    }
  delete buf;
  (*darray)[arrayno].id = maxid + 1;
  return true;
}

int frame::getdataindex(int id)
{
  for (int i = 0; i < darray->getarraylength(); i++)
    {
      if (id == (*darray)[i].id)
	return i;
    }
  return -1;
}

bool frame::setpartsid(int arrayno)
{
  // calc max id
  int maxid = -1;
  for (int i = 0; i < parray->getarraylength(); i++)
    {
      if (i != arrayno && maxid < (*parray)[i].id)
	maxid = (*parray)[i].id;
    }

  if (maxid == -1)
    {
      (*parray)[arrayno].id = 0;
      return true;
    }

  // create table
  bool *buf = new bool [(maxid + 1) * sizeof(bool)];
  if (buf == 0)
    return false;
  memset(buf, 0, (maxid + 1) * sizeof(bool));

  // set table
  for (int i = 0; i < parray->getarraylength(); i++)
    {
      if (i != arrayno)
	buf[(*parray)[i].id] = true;
    }

  // search id
  for (int i = 0; i <= maxid; i++)
    {
      if (buf[i] == 0)
	{
	  (*parray)[arrayno].id = i;
	  delete buf;
	  return true;
	}
    }
  delete buf;
  (*parray)[arrayno].id = maxid + 1;
  return true;
}


int frame::getpartsindex(int id)
{
  for (int i = 0; i < parray->getarraylength(); i++)
    {
      if (id == (*parray)[i].id)
	return i;
    }
  return -1;
}

void frame::getBBbox(intrect *r)
{
  r->left   = wtov_x(0);
  r->top    = wtov_y(0);
  r->right  = wtov_x(p2.x - p1.x);
  r->bottom = wtov_y(p2.y - p1.y);
}

bool frame::readframe(FILE *fp)
{
  char stemp[10000];
  std::string str, label, value;
  char *p;
  int no;
  int readversion;

  // version
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      readversion = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // id
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      id = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // frametype
  if (getnextline(stemp, sizeof(stemp), fp))
  {

    if (nextitem(stemp, &str) != 0)
      frametype = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // p1, p2
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // p1.x
    p = nextitem(p, &str);
    if (p != 0)
      p1.x = atoi(str.c_str());
    else
      return false;
    // p1.y
    p = nextitem(p, &str);
    if (p != 0)
      p1.y = atoi(str.c_str());
    else
      return false;
    // p2.x
    p = nextitem(p, &str);
    if (p != 0)
      p2.x = atoi(str.c_str());
    else
      return false;
    // p2.y
    p = nextitem(p, &str);
    if (p != 0)
      p2.y = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // framelinecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // framelinecolor.red
    p = nextitem(p, &str);
    if (p != 0)
      framelinecolor.red = atoi(str.c_str());
    else
      return false;
    // framelinecolor.green
    p = nextitem(p, &str);
    if (p != 0)
      framelinecolor.green = atoi(str.c_str());
    else
      return false;
    // framelinecolor.blue
    p = nextitem(p, &str);
    if (p != 0)
      framelinecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

    
  // framebodycolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // framebodycolor.red
    p = nextitem(p, &str);
    if (p != 0)
      framebodycolor.red = atoi(str.c_str());
    else
      return false;
    // framebodycolor.green
    p = nextitem(p, &str);
    if (p != 0)
      framebodycolor.green = atoi(str.c_str());
    else
      return false;
    // framebodycolor.blue
    p = nextitem(p, &str);
    if (p != 0)
      framebodycolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // framelinestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // framelinestyle.width
    p = nextitem(p, &str);
    if (p != 0)
      framelinestyle.width = atoi(str.c_str());
    else
      return false;
    // framelinestyle.cap
    p = nextitem(p, &str);
    if (p != 0)
      framelinestyle.cap = atoi(str.c_str());
    else
      return false;
    // framelinestyle.join
    p = nextitem(p, &str);
    if (p != 0)
      framelinestyle.join = atoi(str.c_str());
    else
      return false;
    // framelinestyle.style
    p = nextitem(p, &str);
    if (p != 0)
      framelinestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // framebodyfillstyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // framebodyfillstyle.fillrule
    p = nextitem(p, &str);
    if (p != 0)
      framebodyfillstyle.fillrule = atoi(str.c_str());
    else
      return false;
    // framebodyfillstyle.style
    p = nextitem(p, &str);
    if (p != 0)
      framebodyfillstyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // withdata
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      withdata = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // paintorder
  if (readversion < 2)
      *paintorder = std::string("frame,data,axis,axisline,parts");
  else
    {
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  if (nextitem(stemp, &str) != 0)
	    *paintorder = str;
	  else
	    return false;
	}
      else
	return false;
    }

  // user const
  // number
  usrconst->setbuff(0);
  if (getnextline(stemp, sizeof(stemp), fp))
    no = atoi(stemp);
  else
    return false;
  for (int i = 0; i < no; i++)
    {
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  // label
	  if (nextitem(stemp, &str) != 0)
	    label = str;
	  else
	    return false;
	}
      else
	return false;
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  // value
	  if (nextitem(stemp, &str) != 0)
	    value = str;
	  else
	    return false;
	}
      else
	return false;
      usrconst->add(&label, &value);
    }
  
  // user function
  // number
  usrfnc->setbuff(0);
  if (getnextline(stemp, sizeof(stemp), fp))
    no = atoi(stemp);
  else
    return false;
  for (int i = 0; i < no; i++)
    {
      // label
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  if (gettxt(stemp, &str))
	    label = str;
	  else
	    return false;
	}
      else
	return false;
      //value
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  if (gettxt(stemp, &str))
	    value = str;
	  else
	    return false;
	}
      else
	return false;
      usrfnc->add(&label, &value);
    }
  return true;
}

bool frame::writeframe(FILE *fp)
{
  fprintf(fp, "[frame]\n");
  fprintf(fp, "%d\n", version);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d\n", frametype);
  fprintf(fp, "%d %d %d %d\n", p1.x, p1.y, p2.x, p2.y);
  fprintf(fp, "%u %u %u\n", framelinecolor.red, framelinecolor.green, framelinecolor.blue);
  fprintf(fp, "%u %u %u\n", framebodycolor.red, framebodycolor.green, framebodycolor.blue);
  fprintf(fp, "%d %d %d %d\n", framelinestyle.width, framelinestyle.cap, framelinestyle.join, framelinestyle.style);
  fprintf(fp, "%d %d\n", framebodyfillstyle.fillrule, framebodyfillstyle.style);
  fprintf(fp, "%d\n", withdata);
  fprintf(fp, "%s\n", paintorder->c_str());
  fprintf(fp, "%d\n", usrconst->getarraylength());
  for (int i = 0; i < usrconst->getarraylength(); i++)
    {
      fprintf(fp, "%s\n", (*usrconst)[i].label.c_str());
      fprintf(fp, "%.15e\n", atof((*(*usrconst)[i].value)[0].c_str()));
    }
  fprintf(fp, "%d\n", usrfnc->getarraylength());
  for (int i = 0; i < usrfnc->getarraylength(); i++)
    {
      fprintf(fp, "\"%s\"\n", (*usrfnc)[i].label.c_str());
      fprintf(fp, "\"%s\"\n",(*(*usrfnc)[i].value)[0].c_str());
    }

  return true;
}

void frame::setwithdata()
{
  for (int i = 0; i < darray->getarraylength(); i++)
    (*darray)[i].virgin = !withdata;
}
