/*----------------------------------------------------------------------------
                  funcs.cc (sub routine of script execution)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include  <stdlib.h>
#include  <string>
#include  <ctype.h>
#include  <string.h>
#include  <unistd.h>
#include  <time.h>
#include  <sys/utsname.h>

#include "script.h"
#include "system.h"
#include "tokenbuff.h"
#include "graph.h"
#include "buffarray.h"
#include "cui.h"
#include "browse.h"
#include "objectlist.h"
#include "pipecall.h"
#include "minmax.h"
#include "gamma.h"
#include "filepoint.h"
#include "convtxt.h"
#include "msgbuff.h"
#include "kanji.h"
#include "topazvalues.h"
#include "paths.h"
#include "topazprintf.h"
#include "completion.h"

extern bool dinvq(double in, double *out);
extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;
extern FileHandleArray *fhvuff;
extern void cuimenu();
extern void showsepwin();
extern void message(const char *msg);
extern msgbuff* msgb;
extern void exitcuimenu();
extern void restorecuimenu();
extern bool getmacroname(std::string *file, std::string *out);


//*****************************************
//************* print *********************
//*****************************************
int f_print(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, str4, str5, flushstr;
  bool isexistprinthandle;
  tokenbuff printseparator;
  FILE *h;
  int itemp;
  bool b1 = false;;

  // printseparator $,
  printseparator.priority = PRI_OPERAND;
  printseparator.status = VARIANT;
  *printseparator.label = std::string("$,");
  printseparator.value->add("");
  
  // flushstr
  str5 = std::string("$|");
  scr->vbuff->GetScalar(&str5, &flushstr);
  
  if (v->getarraylength() > 1)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 2], &str1))
	{
	  if (fhvuff->get(&str1, &h, &itemp, &str2))
	    {
	      isexistprinthandle = true;
	      v->remove(v->getarraylength() - 2);
	    }
	  else
	    isexistprinthandle = false;
	}
      else
	isexistprinthandle = false;
    }
  else if (v->getarraylength() == 1)
    isexistprinthandle = false;
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
    {
      if (isexistprinthandle)
	fprintf(h, "%s", str1.c_str());
      else
	message(str1.c_str());
      v->pop();
    }
  else
    {
      if ((*v)[v->getarraylength() - 1].status == ARRAY)
	{
	  b1 = scr->GetValue(&printseparator, &str1);
	  if (b1)
	    {
	      if (scr->vbuff->search((*v)[v->getarraylength() - 1].label, &itemp))
		{
		  if ( scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) > 0)
		    {
		      str3 = std::string("");
		      for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1; itemp++)
			{
			  if (scr->vbuff->Get((*v)[v->getarraylength() - 1].label, itemp, &str2))
			    str3 += str2 + str1;
			  else
			    return scr->errormessage(lineno, "invalid arguments", fname);
			}
		      if (scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 
				     scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1, &str2))
			{
			  str3 += str2;
			  if (isexistprinthandle)
			    {
			      fprintf(h, "%s", str3.c_str());
			      if (atoi(flushstr.c_str()) != 0)
				fflush(h);
			    }
			  else
			    message(str3.c_str());
			  v->pop();
			}
		      else
			return scr->errormessage(lineno, "invalid arguments", fname);
		    }
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else
	    return scr->errormessage(lineno, "undefined print separator $", fname);
	}
      else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  b1 = scr->GetValue(&printseparator, &str1);
	  if (b1)
	    {
	      if ((*v)[v->getarraylength() - 1].value->getarraylength() > 0)
		{
		  str3 = std::string("");
		  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength() - 1; itemp++)
		    {
		      str4  = (*v)[v->getarraylength() - 1].value->getdata(itemp);
		      str3 += str4 + str1;
		    }
		  str4 = (*v)[v->getarraylength() - 1].value->getdata((*v)[v->getarraylength() - 1].value
								      ->getarraylength() - 1);
		  str3 += str4;
		  if (isexistprinthandle)
		    {
		      fprintf(h, "%s", str3.c_str());
		      if (atoi(flushstr.c_str()) != 0)
			fflush(h);
		    }
		  else
		    message(str3.c_str());
		}
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "undefined print separator $", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }

  return true;
}

//*****************************************
//************* printf *******************
//*****************************************
int f_printf(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, flushstr;
  bool isexistprinthandle;
  FILE *h;
  int itemp;

  // flushstr
  str3 = std::string("$|");
  scr->vbuff->GetScalar(&str3, &flushstr);

  if (v->getarraylength() > 1)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 2], &str1))
	{
	  if (fhvuff->get(&str1, &h, &itemp, &str2))
	    {
	      isexistprinthandle = true;
	      v->remove(v->getarraylength() - 2);
	    }
	  else
	    isexistprinthandle = false;
	}
      else
	isexistprinthandle = false;
    }
  else if (v->getarraylength() == 1)
    isexistprinthandle = false;
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
    {
      if (topazprintf((*v)[v->getarraylength() - 1].value, &str1))
	{
	  if (isexistprinthandle)
	    {
	      fprintf(h, "%s", str1.c_str());
	      if (atoi(flushstr.c_str()) != 0)
		fflush(h);
	    }
	  else
	    message(str1.c_str());

	  v->pop();
	}
      else
	return scr->errormessage(lineno, "", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}


//*****************************************
//************* sprintf *******************
//*****************************************
int f_sprintf(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool isexistprinthandle;
  FILE *h;
  int itemp;

  if (v->getarraylength() > 1)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 2], &str1))
	{
	  if (fhvuff->get(&str1, &h, &itemp, &str2))
	    {
	      isexistprinthandle = true;
	      v->remove(v->getarraylength() - 2);
	    }
	  else
	    isexistprinthandle = false;
	}
      else
	isexistprinthandle = false;
    }
  else if (v->getarraylength() == 1)
    isexistprinthandle = false;
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
    {
      if (topazprintf((*v)[v->getarraylength() - 1].value, &str1))
	{
	  (*v)[v->getarraylength() - 1].value->setscalar(str1.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}


//*****************************************
//********* chop, chomp *******************
//*****************************************
int f_chop(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  VariantBuff *vbuff = scr->vbuff;
  std::string str1, str2, str3, str4, str5, chompstr;
  int itemp, itemp1, itemp2;
  char stemp[1024];

  // chompstr
  str5 = std::string("$/");
  scr->vbuff->GetScalar(&str5, &chompstr);
  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  itemp2 = (int)strlen(str1.c_str());
	  itemp1 = itemp2 - charlength(str1.c_str()[itemp2 - 1]);
	  
	  if (strcmp(fname, "chop") == 0)
	    str2 = str1.substr(0, itemp1);
	  if (strcmp(fname, "chomp") == 0)
	    {
	      stemp[0] = str1.c_str()[itemp1]; stemp[1] = 0;
	      if (chompstr == std::string(stemp))
		str2 = str1.substr(0, itemp1);
	      else
		str2 = str1;
	    }
	  
	  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (vbuff->search((*v)[v->getarraylength() - 1].label, &itemp))
		{
		  (*v)[v->getarraylength() - 1].value->setbuff(0);
		  for (int i = 0; i < vbuff->ArrayLength((*v)[v->getarraylength() - 1].label); i++)
		    {
		      vbuff->Get((*v)[v->getarraylength() - 1].label, i, &str2);

		      itemp2 = (int)strlen(str2.c_str());
		      itemp1 = itemp2 - charlength(str2.c_str()[itemp2 - 1]);
		      
		      if (strcmp(fname, "chop") == 0)
			str3 = str2.substr(0, itemp1);
		      if (strcmp(fname, "chomp") == 0)
			{
			  stemp[0] = str2.c_str()[itemp1]; stemp[1] = 0;
			  if (chompstr == std::string(stemp))
			    str3 = str2.substr(0, itemp1);
			  else
			    str3 = str2;
			}
		      
		      (*v)[v->getarraylength() - 1].value->add(str3.c_str());
		    }
		  *(*v)[v->getarraylength() - 1].label = std::string("");
		  (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      for (int i = 0; i < (*v)[v->getarraylength() - 1].value->getarraylength(); i++)
		{
		  itemp2 = (int)strlen((*(*v)[v->getarraylength() - 1].value)[i].c_str());
		  itemp1 = itemp2 - charlength((*(*v)[v->getarraylength() - 1].value)[i].c_str()[itemp2 - 1]);

		  
		  if (strcmp(fname, "chop") == 0)
		    str3 = (*(*v)[v->getarraylength() - 1].value)[i].substr(0, itemp1);
		  if (strcmp(fname, "chomp") == 0)
		    {
		      stemp[0] = str2.c_str()[itemp1]; stemp[1] = 0;
		      if (chompstr == std::string(stemp))
			str3 = (*(*v)[v->getarraylength() - 1].value)[i].substr(0, itemp1);
		      else
			str3 = str2;
		    }
		  
		  (*(*v)[v->getarraylength() - 1].value)[i] = str3;
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//*************** substr ******************
//*****************************************
int f_substr(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, str4;
  int itemp, itemp1;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
	    {
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
	      itemp = atoi(str2.c_str());
	      
	      if (itemp <= (int)strlen(str1.c_str()) && 0 < itemp)
		str4 = str1.substr(atoi(str2.c_str()) - 1);
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else if ((*v)[v->getarraylength() - 1].value->getarraylength() == 3)
	    {
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
	      itemp = atoi(str2.c_str());
	      str3 = (*v)[v->getarraylength() - 1].value->getdata(2);
	      itemp1 = atoi(str3.c_str());
	      
	      if (itemp <= (int)strlen(str1.c_str()) && 0 < itemp)
		{
		  if (itemp1 <= ((int)strlen(str1.c_str()) - itemp + 1) && 0 < itemp1)
		    str4 = str1.substr(itemp - 1, itemp1);
		  else
		    return scr->errormessage(lineno, "invalid arguments", fname);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	  
	  (*v)[v->getarraylength() - 1].value->setscalar(str4.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//********* max, min, beta ****************
//*****************************************
int f_max(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  double ftemp = 0.0;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
	    {
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	  
	  if (strcmp(fname, "min") == 0)
	    ftemp = min(atof(str1.c_str()), atof(str2.c_str()));
	  else if (strcmp(fname, "max") == 0)
	    ftemp = max(atof(str1.c_str()), atof(str2.c_str()));
	  else if (strcmp(fname, "beta") == 0)
	    ftemp = beta(atof(str1.c_str()), atof(str2.c_str()));
	  
	  sprintf(stemp, "%.15g", ftemp);
	  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* exit **********************
//*****************************************
int f_exit(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  str2 = std::string("$_EXIT");
	  if(scr->vbuff->SetScalar(&str2, &str1))
	    return exitscriptstatus;
	  else
	    return false;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* toplaintext ***************
//*****************************************
int f_toplaintext(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if(tosavetxt((char *)str1.c_str(), &str2))
	    {
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* fromplaintext *************
//*****************************************
int f_fromplaintext(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if(fromsavetxt((char *)str1.c_str(), &str2))
	    {
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//*********** toplainfilename *************
//*****************************************
int f_toplainfilename(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if(toplainfilename((char *)str1.c_str(), &str2))
	    {
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* updateseparator ***********
//*****************************************
int f_updateseparator(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      showsepwin();
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* localtime *****************
//*****************************************
int f_localtime(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  time_t timeval;
  struct tm *ptm;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      (void)time(&timeval);
	      ptm = localtime(&timeval);
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      strftime(stemp, sizeof(stemp), "%S", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%M", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%H", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%d", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%m", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%Y", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%u", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* localtimename *************
//*****************************************
int f_localtimename(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  time_t timeval;
  struct tm *ptm;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      (void)time(&timeval);
	      ptm = localtime(&timeval);
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      strftime(stemp, sizeof(stemp), "%S", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%M", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%H", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%d", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%B", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%Y", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      strftime(stemp, sizeof(stemp), "%A", ptm);
	      (*v)[v->getarraylength() - 1].value->add(stemp);
	      
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* flushmsg ******************
//*****************************************
int f_flushmsg(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      msgb->flush();
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* cuimenu *******************
//*****************************************
int f_cuimenu(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    cuimenu();
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* system ********************
//*****************************************
int f_system(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  int ret;
  char stemp[100];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  ret = system(str1.c_str());
	  sprintf(stemp, "%d", ret);

	  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* open **********************
//*****************************************
int f_open(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  FILE *h;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
	    {
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
	      
	      if (str1 == std::string("STDIN")
		  || str1 == std::string("STDOUT")
		  || str1 == std::string("STDERR"))
		return scr->errormessage(lineno, "could not open file handle", fname);
	      
	      if (!fhvuff->search(&str1, &itemp))
		{
		  h = PipeOpen(&str2, &itemp);
		  if (h != NULL)
		    {
		      fhvuff->add(&str1, &h, &itemp, &str2);
		      (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
		      *(*v)[v->getarraylength() - 1].label = std::string("");
		      (*v)[v->getarraylength() - 1].status = NUM;
		    }
		  else
		    {
		      (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
		      *(*v)[v->getarraylength() - 1].label = std::string("");
		      (*v)[v->getarraylength() - 1].status = NUM;
		    }
		}
	      else
		return scr->errormessage(lineno, "this handle is already opened", fname);
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* close *********************
//*****************************************
int f_close(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  FILE *h;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if (str1 == std::string("STDIN")
	      || str1 == std::string("STDOUT")
	      || str1 == std::string("STDERR"))
	    return scr->errormessage(lineno, "could not close file handle", fname);
	  
	  if (fhvuff->get(&str1, &h, &itemp, &str2))
	    {
	      PipeClose(h, &itemp);
	      fhvuff->search(&str1, &itemp);
	      fhvuff->remove(itemp);
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* listhandle ****************
//*****************************************
int f_listhandle(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  int itemp;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  sprintf(stemp, "id  label  type  file/pipe\n");
	  message(stemp);
	  for(int i = 0; i < fhvuff->getarraylength(); i++)
	    {
	      fhvuff->getinfo(i, &str1, &itemp, &str2);
	      sprintf(stemp, "%d  %s  %d  '%s'\n", i, str1.c_str(), itemp, str2.c_str());
	      message(stemp);
	    }
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* isfileexist ***************
//*****************************************
int f_isfileexist(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  FILE *h;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if ((h = fopen(str1.c_str(), "r")) != NULL)
	    {
	      fclose(h);
	      (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
	    }
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* getmacroname **************
//*****************************************
int f_getmacroname(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if (getmacroname(&str1, &str2))
	      (*(*v)[v->getarraylength() - 1].value)[0] = str2;
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* getcwd ********************
//*****************************************
int f_getcwd(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if (str1 == std::string("null"))
	    {
	      if (getcwd(stemp, sizeof(stemp)) == 0)
		(*(*v)[v->getarraylength() - 1].value)[0] = std::string("");
	      else
		{
		  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* uname *********************
//*****************************************
int f_uname(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  char stemp[1024];
  struct utsname utsn;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if (str1 == std::string("null"))
	    {
	      if (uname(&utsn) == 0)
		{
		  (*v)[v->getarraylength() - 1].value->setbuff(0);
		  sprintf(stemp, "%s", utsn.sysname);
		  (*v)[v->getarraylength() - 1].value->add(stemp);
		  sprintf(stemp, "%s", utsn.nodename);
		  (*v)[v->getarraylength() - 1].value->add(stemp);
		  sprintf(stemp, "%s", utsn.release);
		  (*v)[v->getarraylength() - 1].value->add(stemp);
		  sprintf(stemp, "%s", utsn.version);
		  (*v)[v->getarraylength() - 1].value->add(stemp);
		  sprintf(stemp, "%s", utsn.machine);
		  (*v)[v->getarraylength() - 1].value->add(stemp);

		  *(*v)[v->getarraylength() - 1].label = std::string("");
		  (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
		}
	      else
		{
		  *(*v)[v->getarraylength() - 1].label = std::string("");
		  (*v)[v->getarraylength() - 1].status = NUM;
		  return scr->errormessage(lineno, "invalid arguments", fname);
		}
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* chdir *********************
//*****************************************
int f_chdir(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  if (chdir(str1.c_str()) == -1)
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  else
	    {
	      (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
	    }
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* sleep *********************
//*****************************************
int f_sleep(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  sprintf(stemp, "%d", sleep(atoi(str1.c_str())));
	  (*v)[v->getarraylength() - 1].value->setscalar(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* eval **********************
//*****************************************
int f_eval(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  script *scr1 = new script(scr->gra);
	  switch (scr1->execscript(&str1, &str2, 0))
	    {
	    case true:
	      delete scr1;
	      v->pop();
	      break;
	    case false:
	      delete scr1;
	      return false;
	    default:
	      delete scr1;
	      (*(*v)[v->getarraylength() - 1].value)[0] = str2;
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* execscr *******************
//*****************************************
int f_execscr(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  buffarray tempbuffarray;
  script *scr1;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() > 0)
	    {
	      // script
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      tempbuffarray.setbuff(0);
	      
	      for (int i = 1; i < (*v)[v->getarraylength() - 1].value->getarraylength(); i++)
		tempbuffarray.add((*v)[v->getarraylength() - 1].value->getdata(i).c_str());
	      
	      scr1 = new script(scr->gra);
	      switch (scr1->execscript(&str1, &str2, &tempbuffarray))
		{
		case false:
		  delete scr1;
		  return false;
		case true:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar("1");
		  break;
		default:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "too few arguments", fname);
	}
      else if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  scr1 = new script(scr->gra);
	  switch (scr1->execscript(&str1, &str2, 0))
	    {
	    case false:
	      delete scr1;
	      return false;
	    case true:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	      break;
	    default:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	    }
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* execfile ******************
//*****************************************
int f_execfile(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  buffarray tempbuffarray;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() > 0)
	    {
	      // script
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      tempbuffarray.setbuff(0);
	      
	      for (int i = 1; i < (*v)[v->getarraylength() - 1].value->getarraylength(); i++)
		tempbuffarray.add((*v)[v->getarraylength() - 1].value->getdata(i).c_str());
	      
	      script *scr1 = new script(scr->gra);
	      switch (scr1->execscriptfile(&str1, &str2, &tempbuffarray))
		{
		case NOTEXISTSCRIPT:
		  delete scr1;
		  str3 = std::string("cannot execute ");
		  str3 += str1;
		  return scr->errormessage(lineno, (char *)str3.c_str(), fname);
		case false:
		  delete scr1;
		  return false;
		case true:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar("1");
		  break;
		default:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "too few arguments", fname);
	}
      else if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  script *scr1 = new script(scr->gra);
	  switch (scr1->execscriptfile(&str1, &str2, 0))
	    {
	    case NOTEXISTSCRIPT:
	      delete scr1;
	      str3 = std::string("cannot execute ");
	      str3 += str1;
	      return scr->errormessage(lineno, (char *)str3.c_str(), fname);
	    case false:
	      delete scr1;
	      return false;
	    case true:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	      break;
	    default:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	    }
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* includefile ***************
//*****************************************
int f_includefile(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  buffarray tempbuffarray;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() > 0)
	    {
	      // script
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      tempbuffarray.setbuff(0);
	      
	      for (int i = 1; i < (*v)[v->getarraylength() - 1].value->getarraylength(); i++)
		tempbuffarray.add((*v)[v->getarraylength() - 1].value->getdata(i).c_str());
	      
	      script *scr1 = new script(scr->gra, scr);
	      switch (scr1->execscriptfile(&str1, &str2, &tempbuffarray))
		{
		case NOTEXISTSCRIPT:
		  delete scr1;
		  str3 = std::string("cannot execute ");
		  str3 += str1;
		  return scr->errormessage(lineno, (char *)str3.c_str(), fname);
		case false:
		  delete scr1;
		  return false;
		case true:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar("1");
		  break;
		default:
		  delete scr1;
		  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "too few arguments", fname);
	}
      else if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  script *scr1 = new script(scr->gra, scr);
	  switch (scr1->execscriptfile(&str1, &str2, 0))
	    {
	    case NOTEXISTSCRIPT:
	      delete scr1;
	      str3 = std::string("cannot execute ");
	      str3 += str1;
	      return scr->errormessage(lineno, (char *)str3.c_str(), fname);
	    case false:
	      delete scr1;
	      return false;
	    case true:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	      break;
	    default:
	      delete scr1;
	      (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	      break;
	    }
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}


//*****************************************
//************* edittext ******************
//*****************************************
int f_edittext(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) == 2)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 0, &str1);
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 1, &str2);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else if((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
		{
		  str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
		  str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  
	  if (edittextwindow(menuwin, &str1, &str2, false))
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	  str3 = std::string("$_RET");
	  scr->vbuff->SetScalar(&str3, &str1);  
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}


//*****************************************
//************* editfilename **************
//*****************************************
int f_editfilename(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) == 2)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 0, &str1);
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 1, &str2);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else if((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
		{
		  str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
		  str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  
	  if (edittextwindow(menuwin, &str1, &str2, true))
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	  str3 = std::string("$_RET");
	  scr->vbuff->SetScalar(&str3, &str1);  
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* objectmove ****************
//*****************************************
int f_objectmove(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      str1 = *(*v)[v->getarraylength() - 1].label;  
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength(&str1); itemp++)
		{
		  scr->vbuff->Get(&str1, itemp, &str2);
		  (*v)[v->getarraylength() - 1].value->add(str2.c_str());
		}
	    }
	  
	  str3 = std::string("move objects");
	  if (movewindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
	    {
	      str3 = std::string("@_RET");
	      scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
	      for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else 
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* inputparts ****************
//*****************************************
int f_inputparts(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      str1 = *(*v)[v->getarraylength() - 1].label;  
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength(&str1); itemp++)
		{
		  scr->vbuff->Get(&str1, itemp, &str2);
		  (*v)[v->getarraylength() - 1].value->add(str2.c_str());
		}
	    }
	  
	  str3 = std::string("input parts");
	  if (inputpartswindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
	    {
	      str3 = std::string("@_RET");
	      scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
	      for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else 
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* trimming ******************
//*****************************************
int f_trimming(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      str3 = std::string("trimming");
	      if (trimmingwindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
		{
		  str3 = std::string("@_RET");
		  scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
		  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		    scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
		  (*v)[v->getarraylength() - 1].value->setscalar("1");
		}
	      else 
		(*v)[v->getarraylength() - 1].value->setscalar("0");
	      
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* watch *********************
//*****************************************
int f_watch(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      str3 = std::string("watch");
	      if (watchwindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
		{
		  str3 = std::string("@_RET");
		  scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
		  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		    scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
		  (*v)[v->getarraylength() - 1].value->setscalar("1");
		}
	      else
		(*v)[v->getarraylength() - 1].value->setscalar("0");
	      
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = NUM;
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* trace *********************
//*****************************************
int f_trace(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  str3 = std::string("trace");
	  if (tracewindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3, atoi(str1.c_str())))
	    {
	      str3 = std::string("@_RET");
	      scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
	      for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* message *******************
//*****************************************
int f_message(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  messagewindow(menuwin, &str1);
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* changenodes ***************
//*****************************************
int f_changenodes(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      str1 = *(*v)[v->getarraylength() - 1].label;  
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength(&str1); itemp++)
		{
		  scr->vbuff->Get(&str1, itemp, &str2);
		  (*v)[v->getarraylength() - 1].value->add(str2.c_str());
		}
	    }
	  
	  str3 = std::string("move node");
	  if (changenodeswindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
	    {
	      str3 = std::string("@_RET");
	      scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
	      for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else 
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* objectstretch *************
//*****************************************
int f_objectstretch(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  int itemp;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      str1 = *(*v)[v->getarraylength() - 1].label;  
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength(&str1); itemp++)
		{
		  scr->vbuff->Get(&str1, itemp, &str2);
		  (*v)[v->getarraylength() - 1].value->add(str2.c_str());
		}
	    }
	  
	  str3 = std::string("stretch objects");
	  if (stretchwindow(menuwin, (*v)[v->getarraylength() - 1].value, &str3))
	    {
	      str3 = std::string("@_RET");
	      scr->vbuff->SetArray(&str3, (*v)[v->getarraylength() - 1].value->getarraylength());
	      for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		scr->vbuff->Set(&str3, itemp, &(*(*v)[v->getarraylength() - 1].value)[itemp]);
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else 
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* yesnoname *****************
//*****************************************
int f_yesnoname(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  int itemp;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  itemp = atoi(str1.c_str());
	  if (itemp == 0)
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("false");
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("true");
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* umenu **********************
//*****************************************
int f_umenu(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	    {
	      if(strcmp(fname, "framemenu") == 0)
		framemenu();
	      else if(strcmp(fname, "xbasemenu") == 0)
		xbasemenu();
	      else if(strcmp(fname, "xrangemenu") == 0)
		xrangemenu();
	      else if(strcmp(fname, "xtickmenu") == 0)
		xtickmenu();
	      else if(strcmp(fname, "xsubtickmenu") == 0)
		xsubtickmenu();
	      else if(strcmp(fname, "xlabelmenu") == 0)
		xlabelmenu();
	      else if(strcmp(fname, "ybasemenu") == 0)
		ybasemenu();
	      else if(strcmp(fname, "yrangemenu") == 0)
		yrangemenu();
	      else if(strcmp(fname, "ytickmenu") == 0)
		ytickmenu();
	      else if(strcmp(fname, "ysubtickmenu") == 0)
		ysubtickmenu();
	      else if(strcmp(fname, "ylabelmenu") == 0)
		ylabelmenu();
	      else if(strcmp(fname, "tbasemenu") == 0)
		tbasemenu();
	      else if(strcmp(fname, "trangemenu") == 0)
		trangemenu();
	      else if(strcmp(fname, "ttickmenu") == 0)
		ttickmenu();
	      else if(strcmp(fname, "tsubtickmenu") == 0)
		tsubtickmenu();
	      else if(strcmp(fname, "tlabelmenu") == 0)
		tlabelmenu();
	      else if(strcmp(fname, "rbasemenu") == 0)
		rbasemenu();
	      else if(strcmp(fname, "rrangemenu") == 0)
		rrangemenu();
	      else if(strcmp(fname, "rtickmenu") == 0)
		rtickmenu();
	      else if(strcmp(fname, "rsubtickmenu") == 0)
		rsubtickmenu();
	      else if(strcmp(fname, "rlabelmenu") == 0)
		rlabelmenu();
	      else if(strcmp(fname, "plotmappingmenu") == 0)
		plotmappingmenu();
	      else if(strcmp(fname, "plotlinemenu") == 0)
		plotlinemenu();
	      else if(strcmp(fname, "plotmarkermenu") == 0)
		plotmarkermenu();
	      else if(strcmp(fname, "plotrangemenu") == 0)
		plotmemdatamenu();
	      else if(strcmp(fname, "functionmenu") == 0)
		functionmenu();
	      else if(strcmp(fname, "linearfitmenu") == 0)
		linearfitmenu();
	      else if(strcmp(fname, "nonlinearfitmenu") == 0)
		nonlinearfitmenu();
	      else if(strcmp(fname, "loaddatamenu") == 0)
		loaddatamenu();
	      else if(strcmp(fname, "partsstylemenu") == 0)
		partsstylemenu();
	      else if(strcmp(fname, "textstylemenu") == 0)
		textstylemenu();
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* arith **********************
//*****************************************
int f_arith(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  char stemp[1024];
  double ftemp, ftemp1;


  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  ftemp = atof(str1.c_str());
	  if (strcmp(fname, "sin") == 0)
	    sprintf(stemp, "%.15g", sin(ftemp));
	  else if (strcmp(fname, "cos") == 0)
	    sprintf(stemp, "%.15g", cos(ftemp));
	  else if (strcmp(fname, "abs") == 0)
	    sprintf(stemp, "%.15g", fabs(ftemp));
	  else if (strcmp(fname, "tan") == 0)
	    sprintf(stemp, "%.15g", tan(ftemp));
	  else if (strcmp(fname, "asin") == 0)
	    sprintf(stemp, "%.15g", asin(ftemp));
	  else if (strcmp(fname, "acos") == 0)
	    sprintf(stemp, "%.15g", acos(ftemp));
	  else if (strcmp(fname, "atan") == 0)
	    sprintf(stemp, "%.15g", atan(ftemp));
	  else if (strcmp(fname, "sinh") == 0)
	    sprintf(stemp, "%.15g", sinh(ftemp));
	  else if (strcmp(fname, "cosh") == 0)
	    sprintf(stemp, "%.15g", cosh(ftemp));
	  else if (strcmp(fname, "tanh") == 0)
	    sprintf(stemp, "%.15g", tanh(ftemp));
	  else if (strcmp(fname, "asinh") == 0)
	    sprintf(stemp, "%.15g", asinh(ftemp));
	  else if (strcmp(fname, "acosh") == 0)
	    sprintf(stemp, "%.15g", acosh(ftemp));
	  else if (strcmp(fname, "atanh") == 0)
	    sprintf(stemp, "%.15g", atanh(ftemp));
	  else if (strcmp(fname, "exp") == 0)
	    sprintf(stemp, "%.15g", exp(ftemp));
	  else if (strcmp(fname, "ln") == 0)
	    sprintf(stemp, "%.15g", log(ftemp));
	  else if (strcmp(fname, "log") == 0)
	    sprintf(stemp, "%.15g", log10(ftemp));
	  else if (strcmp(fname, "int") == 0)
	    {
	      if (ftemp >= 0)
		sprintf(stemp, "%.15g", floor(ftemp));
	      else
		sprintf(stemp, "%.15g", -floor(fabs(ftemp)));
	    }
	  else if (strcmp(fname, "gauss") == 0)
	    sprintf(stemp, "%.15g", floor(ftemp));
	  else if (strcmp(fname, "sqrt") == 0)
	    sprintf(stemp, "%.15g", sqrt(ftemp));
	  else if (strcmp(fname, "rand") == 0)
	    sprintf(stemp, "%.15g", (double)rand() / (double)RAND_MAX * ftemp);
	  else if (strcmp(fname, "frac") == 0)
	    sprintf(stemp, "%.15g", modf(ftemp, &ftemp1));
	  else if (strcmp(fname, "theta") == 0)
	    {
	      if (ftemp >= 0)
		strcpy(stemp, "1");
	      else
		strcpy(stemp, "0");
	    }
	  else if (strcmp(fname, "sign") == 0)
	    {
	      if (ftemp < 0.0)
		strcpy(stemp, "-1");
	      else if (ftemp == 0.0)
		strcpy(stemp, "0");
	      else
		strcpy(stemp, "1");
	    }
	  else if (strcmp(fname, "round") == 0)
	    {
	      if (fabs(ftemp) - floor(fabs(ftemp)) >= .5)
		{
		  if (ftemp > 0)
		    sprintf(stemp, "%.15g", floor(fabs(ftemp)) + 1);
		  else
		    sprintf(stemp, "%.15g", -(floor(fabs(ftemp)) + 1));
		}
	      else
		{
		  if (ftemp > 0)
		    sprintf(stemp, "%.15g", floor(fabs(ftemp)));
		  else
		    sprintf(stemp, "%.15g", -floor(fabs(ftemp)));
		}
	    }
	  else if (strcmp(fname, "gamma") == 0)
		sprintf(stemp, "%.15g", gamma(ftemp));
	  else if (strcmp(fname, "erf") == 0)
	    sprintf(stemp, "%.15g", erf(ftemp));
	  else if (strcmp(fname, "erfc") == 0)
	    sprintf(stemp, "%.15g", erfc(ftemp));
	  else if (strcmp(fname, "invq") == 0)
	    {
	      if (dinvq(ftemp, &ftemp1))
		sprintf(stemp, "%.15g", ftemp1);
	      else
		return scr->errormessage(lineno, "calculation failure", fname);
	    }
	  else if (strcmp(fname, "invp") == 0)
	    {
	      if (dinvq(1.0 - ftemp, &ftemp1))
		sprintf(stemp, "%.15g", ftemp1);
	      else
		return scr->errormessage(lineno, "calculation failure", fname);
	    }
	  else if (strcmp(fname, "p_normal") == 0)
	    sprintf(stemp, "%.15g", p_normal(ftemp));
	  else if (strcmp(fname, "q_normal") == 0)
	    sprintf(stemp, "%.15g", q_normal(ftemp));
	  else if (strcmp(fname, "inverf") == 0)
	    {
	      if (ftemp <= -1 || ftemp >= 1)
		return scr->errormessage(lineno, "invalid argument", fname);
	      else
		{
		  if (dinvq((1 + ftemp) / 2.0, &ftemp1) == true)
		    sprintf(stemp, "%.15g", - ftemp1 / sqrt(2.0));
		  else
		    return scr->errormessage(lineno, "calculation failure", fname);
		}
	    }
	  else if (strcmp(fname, "inverfc") == 0)
	    {
	      if (ftemp <= 0 || ftemp >= 2)
		return scr->errormessage(lineno, "invalid argument", fname);
	      else
		{
		  if (dinvq(ftemp / 2.0, &ftemp1) == true)
		    sprintf(stemp, "%.15g", ftemp1 / sqrt(2.0));
		  else
		    return scr->errormessage(lineno, "calculation failure", fname);
		}
	    }
	  else if (strcmp(fname, "strlength") == 0)
	    sprintf(stemp, "%u", str1.length());
	  
	  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* selectfile ****************
//*****************************************
int f_selectfile(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  buffarray tempbuffarray;
  int itemp;

  if (v->getarraylength() > 1)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
	    {
	      str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	  
	  tempbuffarray.setbuff(0);
	  str3 = std::string(".");
	  if (selectfile(menuwin, &str3, &tempbuffarray, atoi(str1.c_str()), &str2))
	    {
	      if (tempbuffarray.getarraylength() > 0)
		{
		  str3 = std::string("@_RET");
		  (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
		  scr->vbuff->SetArray(&str3, tempbuffarray.getarraylength());
		  for (itemp = 0; itemp < tempbuffarray.getarraylength(); itemp++)
		    scr->vbuff->Set(&str3, itemp, &tempbuffarray[itemp]);
		}
	      else
		(*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	    }
	  else
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* selectobject **************
//*****************************************
int f_selectobject(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == ARRAY
	  || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) == 2)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 0, &str1);
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 1, &str2);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	      
	    }
	  else if((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
		{
		  str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
		  str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  
	  if (objectlist(menuwin, atoi(str2.c_str()), &str1))
	    (*v)[v->getarraylength() - 1].value->setscalar("1");
	  else
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* selectitem ****************
//*****************************************
int f_selectitem(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str3, str4;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if (selectitem(menuwin, (*v)[v->getarraylength() - 1].value))
	    {
	      str3 = std::string("$_RET");
	      str4 = (*v)[v->getarraylength() - 1].value->getdata(0); 
	      scr->vbuff->SetScalar(&str3, &str4);  
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else
	    {
	      str4 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      scr->vbuff->SetScalar(&str3, &str4);  
	      (*v)[v->getarraylength() - 1].value->setscalar("0");
	    }
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* selectlistitem ****************
//*****************************************
int f_selectlistitem(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str3, str4;

  if (v->getarraylength() > 0)
    {
      if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if (selectlistitem(menuwin, (*v)[v->getarraylength() - 1].value))
	    {
	      str3 = std::string("$_RET");
	      str4 = (*v)[v->getarraylength() - 1].value->getdata(0); 
	      scr->vbuff->SetScalar(&str3, &str4);  
	      (*v)[v->getarraylength() - 1].value->setscalar("1");
	    }
	  else
	    {
	      str4 = (*v)[v->getarraylength() - 1].value->getdata(0);
	      scr->vbuff->SetScalar(&str3, &str4);  
	      (*v)[v->getarraylength() - 1].value->setscalar("0");
	    }
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* uselect *******************
//*****************************************
int f_uselect(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  int itemp = 0;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) == 2)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 0, &str1);
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 1, &str2);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  else if((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      if ((*v)[v->getarraylength() - 1].value->getarraylength() == 2)
		{
		  str1 = (*v)[v->getarraylength() - 1].value->getdata(0);
		  str2 = (*v)[v->getarraylength() - 1].value->getdata(1);
		}
	      else
		return scr->errormessage(lineno, "invalid arguments", fname);
	    }
	  
	  if(strcmp(fname,"selectcoloritem") == 0)
	    itemp = execcolorselect(&str1, &str2);
	  else if(strcmp(fname, "selectyesno") == 0)
	    itemp = exec_yesno_select(&str1, &str2);
	  else if(strcmp(fname, "selectframetype") == 0)
	    itemp = frametype_select(&str1, &str2);
	  else if(strcmp(fname, "selecttickdirection") == 0)
	    itemp = tickdirection_select(&str1, &str2);
	  else if(strcmp(fname, "selectscaling") == 0)
	    itemp = scaling_select(&str1, &str2);
	  else if(strcmp(fname, "selectmscaling") == 0)
	    itemp = mscaling_select(&str1, &str2);
	  else if(strcmp(fname, "selectlinecap") == 0)
	    itemp = linecap_select(&str1, &str2);
	  else if(strcmp(fname, "selectlinejoin") == 0)
	    itemp = linejoin_select(&str1, &str2);
	  else if(strcmp(fname, "selectlinestyle") == 0)
	    itemp = linestyle_select(&str1, &str2);
	  else if(strcmp(fname, "selectfillstyle") == 0)
	    itemp = fillstyle_select(&str1, &str2);
	  else if(strcmp(fname, "selectfillrule") == 0)
	    itemp = fillrule_select(&str1, &str2);
	  else if(strcmp(fname, "selectmag") == 0)
	    itemp = mag_select(&str1, &str2);
	  else if(strcmp(fname, "selectinterpolate") == 0)
	    itemp = interpolate_select(&str1, &str2);
	  else if(strcmp(fname, "selectmarkerstyle") == 0)
	    itemp = markerstyle_select(&str1, &str2);
	  else if(strcmp(fname, "selectcommentlineaction") == 0)
	    itemp = commentlineaction_select(&str1, &str2);
	  else if(strcmp(fname, "selectdatasearchpath") == 0)
	    itemp = datasearchpath_select(&str1, &str2);
	  else if(strcmp(fname, "selectpartstype") == 0)
	    itemp = partstype_select(&str1, &str2);
	  else if(strcmp(fname, "selectfontface") == 0)
	    itemp = fontface_select(&str1, &str2);
	  else if(strcmp(fname, "selectjpfontface") == 0)
	    itemp = jpfontface_select(&str1, &str2);
	  else if(strcmp(fname, "selectpartsinterpolation") == 0)
	    itemp = partsinterpolation_select(&str1, &str2);
	  else if(strcmp(fname, "selectlabelparity") == 0)
	    itemp = labelparity_select(&str1, &str2);
	  else if(strcmp(fname, "selectlabelcpparity") == 0)
	    itemp = labelCPparity_select(&str1, &str2);
	  else if(strcmp(fname, "selectlabelformat") == 0)
	    itemp = labelformat_select(&str1, &str2);
	  else if(strcmp(fname, "selectorder") == 0)
	    itemp = orderlist_select(&str1, &str2);
	  else if(strcmp(fname, "selectpapersize") == 0)
	    itemp = papersizelist_select(&str1, &str2);
	  else if(strcmp(fname, "selectlinearfitfunction") == 0)
	    itemp = linearfitfunction_select(&str1, &str2);
	  
	  if (itemp)
	    (*v)[v->getarraylength() - 1].value->setscalar("1");
	  else
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	  str3 = std::string("$_RET");
	  scr->vbuff->SetScalar(&str3, &str1);  
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************* toarray *******************
//*****************************************
int f_toarray(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  char *p;
  tokenbuff temptoken;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      //argument must be scalar
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  str2 = std::string("");
	  p = (char *)str1.c_str();
	  temptoken.value->setbuff(0);
	  while (1)
	    {
	      if (*p == ',')
		{
		  temptoken.value->add(str2.c_str());
		  str2 = std::string("");
		}
	      else if (*p == 0)
		{
		  temptoken.value->add(str2.c_str());
		  str2 = std::string("");
		  break;
		}
	      else
		{
		  stemp[0] = *p; stemp[1] = 0;
		  str2 += std::string(stemp);
		}
	      p++;
	    }
	  (*v)[v->getarraylength() - 1] = temptoken;
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* toscalar ******************
//*****************************************
int f_toscalar(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 2 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      str2 = std::string();
	      for (int i = 0; i < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1; i++)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, i, &str1);
		  
		  str2 += str1; str2 += std::string(",");
		}
	      scr->vbuff->Get((*v)[v->getarraylength() - 1].label,
			 scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1, &str1);
	      
	      str2 += str1;
	    }
	  else if((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      str2 = std::string();
	      for (int i = 0; i < (*v)[v->getarraylength() - 1].value->getarraylength() - 1; i++)
		{
		  str1 = (*v)[v->getarraylength() - 1].value->getdata(i);
		  str2 += str1; str2 += std::string(",");
		}
	      str1 = (*v)[v->getarraylength() - 1].value->
		getdata((*v)[v->getarraylength() - 1].value->getarraylength() - 1);
	      str2 += str1;
	    }
	  
	  
	  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}


//*****************************************
//************* defined  ******************
//*****************************************
int f_defined(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str;
  int itemp;

  if (v->getarraylength() > 0)
    {
      // argument must be imparray or array with 1 member
      if ((*v)[v->getarraylength() - 1].status == ARRAY || (*v)[v->getarraylength() - 1].status == VARIANT)
	{
	  if(scr->vbuff->search((*v)[v->getarraylength() - 1].label, &itemp))
	    str = std::string("1");
	  else
	    str = std::string("0");
	}
      else
	str = std::string("0");
	
      (*v)[v->getarraylength() - 1].value->setscalar(str.c_str());
      *(*v)[v->getarraylength() - 1].label = std::string("");
      (*v)[v->getarraylength() - 1].status = NUM;
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}


//*****************************************
//************* appendabspath ****************
//*****************************************
int f_appendabspath(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str))
	{
	  strcpy(stemp, toabsolutepath((char *)str.c_str()));
	  str = std::string(stemp);
	  (*v)[v->getarraylength() - 1].value->setscalar(str.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************* removepath ****************
//*****************************************
int f_removepath(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str;
  char stemp[1024];

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str))
	{
	  strcpy(stemp, tofilename((char *)str.c_str()));
	  str = std::string(stemp);
	  (*v)[v->getarraylength() - 1].value->setscalar(str.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//************ removectrcode ***************
//*****************************************
int f_removectrcode(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  char stemp[3];
  char *p;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1))
	{
	  str2 = std::string("");
	  p = (char *)str1.c_str();
          while (*p != 0)
	    {
	      if (iskanji1(p))
		{
		  stemp[0] = *p++; stemp[1] = *p;stemp[2] = 0;
		  str2 += std::string(stemp);
		}
	      else if (*p == '#' && strchr("HMRBIOSGFpPYXTt", *(p + 1)) != NULL)
		p++;
	      else if  (*p == '#' && strchr("#|^_", *(p + 1)) != NULL)
		{
		  stemp[0] = *(p + 1); stemp[1] = 0;
		  str2 += std::string(stemp);
		  p++;
		}
	      else if  (*p == '\\' && *(p + 1) == 'n')
		{
		  stemp[0] = '\n'; stemp[1] = 0;
		  str2 += std::string(stemp);
		  p++;
		}
	      else if  (*p == '\\' && *(p + 1) == 't')
		{
		  stemp[0] = '\t'; stemp[1] = 0;
		  str2 += std::string(stemp);
		  p++;
		}
	      else if (strchr("|^_", *p) != NULL)
		{
		}
	      else
		{
		  stemp[0] = *p; stemp[1] = 0;
		  str2 += std::string(stemp);
		}
	      p++;
	    }
	  (*v)[v->getarraylength() - 1].value->setscalar(str2.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);
  
  return true;
}

//*****************************************
//*********** escapecuimenu *****************
//*****************************************
int f_escapecuimenu(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	      exitcuimenu();
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//*********** restorecuimenu *****************
//*****************************************
int f_restorecuimenu(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
	{
	  if (str1 == std::string("null"))
	      exitcuimenu();
	  else
	    return scr->errormessage(lineno, "invalid arguments", fname);
	}
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************** unlink *******************
//*****************************************
int f_unlink(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str))
	{
	  if (unlink(str.c_str()) == 0)
	    str = std::string("1");
	  else
	    str = std::string("0");

	  (*v)[v->getarraylength() - 1].value->setscalar(str.c_str());
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}

//*****************************************
//************** readline *****************
//*****************************************

int f_readline(script *scr, char *fname, int lineno)
{
  stackarray *v = scr->v;
  std::string str, str3;
  char *ret;

  if (v->getarraylength() > 0)
    {
      if (scr->GetValue(&(*v)[v->getarraylength() - 1], &str))
	{
	  ret = rline((char *)str.c_str());
	  if (ret == (char *)NULL)
	    (*v)[v->getarraylength() - 1].value->setscalar("0");
	  else
	    (*v)[v->getarraylength() - 1].value->setscalar("1");

	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	  
	  if (ret != (char *)NULL)
	    {
	      str = std::string(ret);
	      str3 = std::string("$_RET");
	      scr->vbuff->SetScalar(&str3, &str);  
	    }
	}
      else
	return scr->errormessage(lineno, "invalid arguments", fname);
    }
  else
    return scr->errormessage(lineno, "invalid arguments", fname);

  return true;
}


typedef struct funcsmap
{
  char *label;
  int (*func)(script *scr, char *fname, int lineno);
}FMAP;

FMAP fmap[] = {
  {"print", f_print},
  {"printf", f_printf},
  {"sprintf", f_sprintf},
  {"abs", f_arith},
  {"sin", f_arith},
  {"cos", f_arith},
  {"tan", f_arith},
  {"asin", f_arith},
  {"acos", f_arith},
  {"atan", f_arith},
  {"sinh", f_arith},
  {"cosh", f_arith},
  {"tanh", f_arith},
  {"asinh", f_arith},
  {"acosh", f_arith},
  {"atanh", f_arith},
  {"exp", f_arith},
  {"ln", f_arith},
  {"log", f_arith},
  {"int", f_arith},
  {"gauss", f_arith},
  {"sqrt", f_arith},
  {"rand", f_arith},
  {"frac", f_arith},
  {"theta", f_arith},
  {"delta", f_arith},
  {"sign", f_arith},
  {"round", f_arith},
  {"gamma", f_arith},
  {"erf", f_arith},
  {"erfc", f_arith},
  {"invq", f_arith},
  {"invp", f_arith},
  {"p_normal", f_arith},
  {"q_normal", f_arith},
  {"inverf", f_arith},
  {"inverfc", f_arith},
  {"strlength", f_arith},
  {"toarray", f_toarray},
  {"toscalar", f_toscalar},
  {"chop", f_chop},
  {"chomp", f_chop},
  {"substr", f_substr},
  {"max", f_max},
  {"min", f_max},
  {"beta", f_max},
  {"exit", f_exit},
  {"unlink", f_unlink},
  {"toplaintext", f_toplaintext},
  {"fromplaintext", f_fromplaintext},
  {"toplainfilename", f_toplainfilename},
  {"updateseparator", f_updateseparator},
  {"localtime", f_localtime},
  {"localtimename", f_localtimename},
  {"flushmsg", f_flushmsg},
  {"cuimenu", f_cuimenu},
  {"system", f_system},
  {"open", f_open},
  {"close", f_close},
  {"listhandle", f_listhandle},
  {"isfileexist", f_isfileexist},
  {"getcwd", f_getcwd},
  {"chdir", f_chdir},
  {"sleep", f_sleep},
  {"getmacroname", f_getmacroname},
  {"eval", f_eval},
  {"execscr", f_execscr},
  {"execfile", f_execfile},
  {"edittext", f_edittext},
  {"editfilename", f_editfilename},
  {"objectmove", f_objectmove},
  {"inputparts", f_inputparts},
  {"trimming", f_trimming},
  {"watch", f_watch},
  {"trace", f_trace},
  {"message", f_message},
  {"changenodes", f_changenodes},
  {"objectstretch", f_objectstretch},
  {"yesnoname", f_yesnoname},
  {"framemenu", f_umenu},
  {"xbasemenu", f_umenu},
  {"xrangemenu", f_umenu},
  {"xtickmenu", f_umenu},
  {"xsubtickmenu", f_umenu},
  {"xlabelmenu", f_umenu},
  {"ybasemenu", f_umenu},
  {"yrangemenu", f_umenu},
  {"ytickmenu", f_umenu},
  {"ysubtickmenu", f_umenu},
  {"ylabelmenu", f_umenu},
  {"tbasemenu", f_umenu},
  {"trangemenu", f_umenu},
  {"ttickmenu", f_umenu},
  {"tsubtickmenu", f_umenu},
  {"tlabelmenu", f_umenu},
  {"rbasemenu", f_umenu},
  {"rrangemenu", f_umenu},
  {"rtickmenu", f_umenu},
  {"rsubtickmenu", f_umenu},
  {"rlabelmenu", f_umenu},
  {"plotmappingmenu", f_umenu},
  {"plotlinemenu", f_umenu},
  {"plotmarkermenu", f_umenu},
  {"plotrangemenu", f_umenu},
  {"functionmenu", f_umenu},
  {"linearfitmenu", f_umenu},
  {"nonlinearfitmenu", f_umenu},
  {"loaddatamenu", f_umenu},
  {"partsstylemenu", f_umenu},
  {"textstylemenu", f_umenu},
  {"selectfile", f_selectfile},
  {"selectobject", f_selectobject},
  {"selectitem", f_selectitem},
  {"selectlistitem", f_selectlistitem},
  {"selectcoloritem", f_uselect},
  {"selectyesno", f_uselect},
  {"selectframetype", f_uselect},
  {"selecttickdirection", f_uselect},
  {"selectscaling", f_uselect},
  {"selectmscaling", f_uselect},
  {"selectlinecap", f_uselect},
  {"selectlinejoin", f_uselect},
  {"selectlinestyle", f_uselect},
  {"selectfillstyle", f_uselect},
  {"selectfillrule", f_uselect},
  {"selectmag", f_uselect},
  {"selectinterpolate", f_uselect},
  {"selectmarkerstyle", f_uselect},
  {"selectcommentlineaction", f_uselect},
  {"selectdatasearchpath", f_uselect},
  {"selectpartstype", f_uselect},
  {"selectfontface", f_uselect},
  {"selectjpfontface", f_uselect},
  {"selectpartsinterpolation", f_uselect},
  {"selectlabelparity", f_uselect},
  {"selectlabelcpparity", f_uselect},
  {"selectlabelformat", f_uselect},
  {"selectorder", f_uselect},
  {"selectpapersize", f_uselect},
  {"selectlinearfitfunction", f_uselect},
  {"includefile", f_includefile},
  {"uname", f_uname},
  {"defined", f_defined},
  {"appendabspath", f_appendabspath},
  {"removepath", f_removepath},
  {"removectrcode", f_removectrcode},
  {"escapecuimenu", f_escapecuimenu},
  {"restorecuimenu", f_restorecuimenu},
  {"readline", f_readline},
  {0, 0}
}; 



int script::funcs (int parsepoint)
{
  int lineno;
  std::string fname;
  FMAP *fm;
  int itemno;
  int (*execfunc)(script *scr, char *s, int linenumber);

  lineno = (*polish)[parsepoint].lineno;
  fname = (*polish)[parsepoint].value->getdata(0);

  fm = fmap;
  for (; fm->label; fm++)
    {
    }
  itemno = fm - fmap;
  for (int i = 0; i < itemno; i++)
    {
      if (strcmp(fname.c_str(), fmap[i].label) == 0)
	{
	  execfunc = fmap[i].func;
	  return execfunc(this, (char *)fname.c_str(), lineno);
	}
    }
  return errormessage(lineno, "invalid command", (char *)fname.c_str());
}


