/*----------------------------------------------------------------------------
                            gamma.cc (gamma function)
                       This file is a part of topaz systems
       This module is based on the codes freely distributed by Mr. Okumura.
                 This module is free from copyright protection.


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include "naninf.h"

double lngamma(double x)
{
  double fa, fb, fc;
  
  fa = 1.0;
  while (x < 8)
    {
      fa *= x;
      x++;
    }
  fb = 1.0 / x / x;
  
  fc = - 3617.0 / 510.0 / 16.0 / 15.0 * fb; 
  fc = (fc + 7.0 / 6.0 / 14.0 / 13.0) * fb;
  fc = (fc - 691.0 / 2730.0 / 12.0 / 11.0) * fb;
  fc = (fc + 5.0 / 66.0 / 10.0 / 9.0) * fb;
  fc = (fc - 1.0 / 30.0 / 8.0 / 7.0) * fb;
  fc = (fc + 1.0 / 42.0 / 6.0 / 5.0) * fb;
  fc = (fc - 1.0 / 30.0 / 4.0 / 3.0) * fb;
  fc = (fc + 1.0 / 6.0 / 2.0 / 1.0) / x;
  fc += 0.5 * log(2.0 * M_PI) - log(fa) - x + (x - 0.5) * log(x);
  
  return fc;

}


double q_gamma(double a, double x, double b);

double p_gamma(double a, double x, double b)
{
  double sum, value, osum;
  
  // trivial
  if (x == 0.0)
    return 0.0;
  
  if (x >= 1.0 + a)
    return 1.0 - q_gamma(a, x, b);
  
  value = exp(a * log(x) - x - b) / a;
  sum = value;
  
  for (int i = 1; i < 10000; i++)
    {
      osum = sum;
      value *= x / (a + (double)i);
      sum += value;
      // converged !!
      if (sum == osum)
	return sum;
    }
  // not converged
  return infvalue();
}

double q_gamma(double a, double x, double b)
{
  double sum, osum, fa, fb, fc, fd;
  
  fc = 1.0;
  fd = 1.0 + x - a;
  
  if (x < 1 + a)
    return 1 - p_gamma(a, x, b);
  
  fa = exp(a * log(x) - x - b);
  sum = fa / fd;
  
  for (int i = 2; i < 10000; i++)
    {
      osum = sum;
      fa *= ((double)i - 1.0 - a) / (double)i;
      fb = (((double)i + x) * fd + ((double)i - 1.0 - a) * (fd - fc)) / (double)i;
      fc = fd;
      fd = fb;
      fb = fa / (fc * fd);
      sum += fb;
      
      // converged !!
      if (sum == osum)
	return sum;
    }
  // not converged
  return infvalue();
}

double erf(double x)
{
  if (x >= 0.0)
    return   p_gamma(0.5, x * x, log(M_PI) / 2.0);
  else
    return - p_gamma(0.5, x * x, log(M_PI) / 2.0);
}

double erfc(double x)
{
  if (x >= 0.0)
    return  q_gamma(0.5, x * x, log(M_PI) / 2.0);
  else
    return  1.0 + p_gamma(0.5, x * x, log(M_PI) / 2.0);
}

double p_normal(double x)
{
  if (x >= 0.0)
    return 0.5 * (1.0 + p_gamma(0.5, 0.5 * x * x, log(M_PI) / 2.0));
  else
    return 0.5 * q_gamma(0.5, 0.5 * x * x, log(M_PI) / 2.0);
}

double q_normal(double x)
{
  if (x >= 0.0)
    return 0.5 * q_gamma(0.5, 0.5 * x * x, log(M_PI) / 2.0);
  else
    return 0.5 * (1.0 + p_gamma(0.5, 0.5 * x * x, log(M_PI) / 2.0));
}

double gamma(double x)
{
  if (x >= 0.0)
    return exp(lngamma(x));
  else
    return M_PI / (sin(M_PI * x) * exp(lngamma(1.0 - x)));
  
}

double beta(double x, double y)
{
  return exp(lngamma(x) + lngamma(y) - lngamma(x + y));
}
