/*----------------------------------------------------------------------------
                           gdi.cc (generate tpv data)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include "vectdefs.h"
#include "pointdefs.h"
#include "filepoint.h"
#include "graph.h"
#include <time.h>
#include "topazvalues.h"

extern graph *gra;
extern FileHandleArray *fhvuff;
extern intpoint RotatePoint(intpoint P, intpoint OrgP, double Theta);

// following flag    true:paint   false:getrect 
bool movelineflag = false;
int prevx, prevy;
intrect bb;
int linew = 0;
int lineno = 0;

void updateBB(int x, int y)
{
  if (x < bb.left)
    bb.left = x;
  if (x > bb.right)
    bb.right = x;
  if (y < bb.top)
    bb.top = y;
  if (y > bb.bottom)
    bb.bottom = y;
}

void psetmag(bool flag, FILE *f)
{
  if (flag)
    {
      fprintf(f, "/MAG\n"); 
      fprintf(f, "%f\n", gra->mag); 
    }
}

void psetshift(bool flag, FILE *f)
{
  if (flag)
    {
      fprintf(f, "/SFT\n"); 
      fprintf(f, "%d,%d\n", gra->offset.x, gra->offset.y); 
    }
}

void pmoveto(bool flag, FILE *f, int x, int y)
{
  if (flag)
    {
      if (movelineflag)
	{
	  fprintf(f, "/SRP\n"); 
	  fprintf(f, "/BGP\n"); 
	}
      movelineflag = false;
      prevx = x; prevy = y;
    }
  else
    {
      movelineflag = false;
      prevx = x; prevy = y;
    }
}

void plineto(bool flag, FILE *f, int x, int y)
{
  if (flag)
    {
      if (movelineflag == false)
	{
	  fprintf(f, "/MVT\n"); 
	  fprintf(f, "%d,%d\n", prevx, prevy); 
	}

      fprintf(f, "/LNT\n"); 
      fprintf(f, "%d,%d\n", x, y); 
      movelineflag = true;
      lineno++;

      if (lineno > 1000)
	{
	  fprintf(f, "/SRP\n"); 
	  fprintf(f, "/BGP\n"); 
	  movelineflag = false;
	  prevx = x; prevy = y;
	  lineno = 0;
	}
    }
  else
    {
      if (movelineflag == false)
	{
	  updateBB(prevx - gra->scale(linew / 2), prevy - gra->scale(linew / 2));
	  updateBB(prevx + gra->scale(linew / 2), prevy + gra->scale(linew / 2));
	}

      updateBB(x - gra->scale(linew / 2), y - gra->scale(linew / 2));
      updateBB(x + gra->scale(linew / 2), y + gra->scale(linew / 2));
      movelineflag = true;
    }
}


void pstroke(bool flag, FILE *f)
{
  if (flag)
    {
      fprintf(f, "/SRP\n"); 
    }
}

void pbeginpath(bool flag, FILE *f)
{
  if (flag)
      fprintf(f, "/BGP\n"); 
}

void pclosepath(bool flag, FILE *f)
{
  if (flag)
    {
      fprintf(f, "/CSP\n"); 
    }
}

void plinestyle(bool flag, FILE *f, int width, int cap, int join, int style)
{
  if (flag)
    {
      fprintf(f, "/LSY\n"); 
      fprintf(f, "%d\n", width);
      linew = width;
      fprintf(f, "%d\n", cap);
      fprintf(f, "%d\n", join);
      fprintf(f, "%d\n", style);
    }
  else
    linew = width;
}

void pfillstyle(bool flag, FILE *f, int fillrule, int style)
{
  if (flag)
    {
      fprintf(f, "/FSY\n"); 
      fprintf(f, "%d\n", fillrule);
      fprintf(f, "%d\n", style);
    }
}

void ppolygon(bool flag, FILE *f, intpointarray *ia)
{
  if (ia->getarraylength() > 1000)
    return;

  if (flag)
    {
      fprintf(f, "/PLY\n"); 
      fprintf(f, "%d\n", ia->getarraylength()); 
      for (int i = 0; i < ia->getarraylength(); i++)
	fprintf(f, "%d,%d\n", (*ia)[i].x, (*ia)[i].y); 
    }
  else
    {
      for (int i = 0; i < ia->getarraylength(); i++)
	{
	  updateBB((*ia)[i].x - gra->scale(linew / 2), (*ia)[i].y - gra->scale(linew / 2));
	  updateBB((*ia)[i].x + gra->scale(linew / 2), (*ia)[i].y + gra->scale(linew / 2));
	}
    }
}

void pellipse(bool flag, FILE *f, int x1, int y1, int x2, int y2)
{
  if (flag)
    {
      fprintf(f, "/ELP\n"); 
      fprintf(f, "%d,%d\n", x1, y1);
      fprintf(f, "%d,%d\n", x2, y2);
    }
  else
    {
      updateBB(x1 - gra->scale(linew / 2), y1 - gra->scale(linew / 2));
      updateBB(x1 + gra->scale(linew / 2), y1 + gra->scale(linew / 2));
      updateBB(x2 - gra->scale(linew / 2), y2 - gra->scale(linew / 2));
      updateBB(x2 + gra->scale(linew / 2), y2 + gra->scale(linew / 2));
    }
}

void psetforecolor(bool flag, FILE *f, unsigned int red, unsigned int green, unsigned int blue)
{
  if (flag)
    {
      fprintf(f, "/FCL\n"); 
      fprintf(f, "%u,%u,%u\n",red, green, blue);
    }
}

void psetbackcolor(bool flag, FILE *f, unsigned int red, unsigned int green, unsigned int blue)
{
  if (flag)
    {
      fprintf(f, "/BCL\n"); 
      fprintf(f, "%u,%u,%u\n",red, green, blue);
    }
}

void psetpaintmode(bool flag, FILE *f, int mode)
{
  if (flag)
    {
      fprintf(f, "/PMD\n"); 
      fprintf(f, "%d\n", mode); 
    }
}

void psavegdc(bool flag, FILE *f)
{
  if (flag)
    fprintf(f, "\n"); 
}

void presoregdc(bool flag, FILE *f)
{
  if (flag)
    fprintf(f, "\n"); 
}

void psetcliprect(bool flag, FILE *f, intrect *rect)
{
  if (flag)
    {
      fprintf(f, "/CRT\n"); 
      fprintf(f, "%d,%d\n", rect->left, rect->top);
      fprintf(f, "%d,%d\n", rect->right, rect->bottom);
    }
}

void pinitcliprect(bool flag, FILE *f)
{
  if (flag)
    {
      fprintf(f, "/RCP\n"); 
    }
}

void ploadfont(bool flag, FILE *f, int face, int size, double angle)
{
  if (flag)
    {
      fprintf(f, "/LFT\n"); 
      fprintf(f, "%d,%d,%f\n", face, size, angle);
    }
}

void pdrawtext(bool flag, FILE *f, int x, int y, const char *text, int size, int width, double angle, int ascend)
{
  intpoint lt, lb, rt, rb;

  if (flag)
    {
      fprintf(f, "/DTX\n"); 
      fprintf(f, "%d,%d\n", x, y);
      fprintf(f, "%s\n", text);
    }
  else
    {
      lt = RotatePoint(intpoint(x, y - ascend), intpoint (x, y), angle);
      rt = RotatePoint(intpoint(x + width, y - ascend), intpoint (x, y), angle);
      rb = RotatePoint(intpoint(x + width, y - ascend + size), intpoint (x, y), angle);
      lb = RotatePoint(intpoint(x, y - ascend + size), intpoint (x, y), angle);

      updateBB(lt.x, lt.y);
      updateBB(rt.x, rt.y);
      updateBB(rb.x, rb.y);
      updateBB(lb.x, lb.y);
    }
}

void psettempdashline(bool flag, FILE *f)
{
  if (flag)
      fprintf(f, "/STD\n"); 
}

void prestorelinestyle(bool flag, FILE *f)
{
  if (flag)
      fprintf(f, "/RLS\n"); 
}

void poutBBox(FILE *f)
{
      fprintf(f, "/PAR\n"); 
      fprintf(f, "%d,%d\n", gra->papersize.x, gra->papersize.y); 
      fprintf(f, "/BBX\n"); 
      fprintf(f, "%d,%d\n", bb.left, bb.top); 
      fprintf(f, "%d,%d\n", bb.right, bb.bottom); 
}

void pbegin(bool flag, FILE *f)
{
  time_t timeval;
  if (flag)
    {
      (void)time(&timeval);
      fprintf(f, "# topaz primitive vector format\n"); 
      fprintf(f, "# created by topaz %s", ctime(&timeval)); 
      fprintf(f, "# Date : %s", ctime(&timeval)); 
      poutBBox(f);
      psetmag(flag, f);
      psetshift(flag, f);
      fprintf(f, "BGN\n"); 
    }
  else
    {
      bb.left      =  MAXINT;
      bb.top       =  MAXINT;
      bb.right     = -MAXINT;
      bb.bottom    = -MAXINT;
    }
}

void ppaint(bool flag, FILE *f)
{
  if (flag)
    fprintf(f, "PNT\n"); 
}

void pclearviewer(bool flag, FILE *f)
{
  if (flag)
    fprintf(f, "/CLR\n"); 
}

void pflush(bool flag, FILE *f)
{
  if (flag)
    fflush(f);
}










