/*----------------------------------------------------------------------------
                topazxdata.cc (generate x-data for function plotting)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include "data.h"

bool genxdata(double start, double end, int div, int scaling, earray* outx, earray* outy)
{
  double s, val;
  int datanum = div + 1;

  if (div <= 0 || div >= 30000)
    return false;

  outx->setbuff(0);
  outy->setbuff(0);
  switch(scaling)
    {
    case MSCALING_LINEAR:
      s = (end - start) / (double)div;
      for (int i = 0; i < datanum; i++)
	{
	  val = start + s * (double)i;
	  outx->add(new dataelement(val, DATA_NORMAL));
	  outy->add(new dataelement(val, DATA_NORMAL));
	}
      break;
      
    case MSCALING_LOG:
      if (start > 0.0 && end > 0.0)
	{
	  s = (log(end) - log(start)) / div;
	  for (int i = 0; i < datanum; i++)
	    {
	      val = start * exp(s * (double)i);
	      outx->add(new dataelement(val, DATA_NORMAL));
	      outy->add(new dataelement(val, DATA_NORMAL));
	    }
	}
      else if (start < 0.0 && end < 0.0)
	{
	  s = (log(-end) - log(-start)) / div;
	  for (int i = 0; i < datanum; i++)
	    {
	      val = start * exp(s * (double)i);
	      outx->add(new dataelement(val, DATA_NORMAL));
	      outy->add(new dataelement(val, DATA_NORMAL));
	    }
	}
      else
	return false;
      break;
    case MSCALING_INV:
      if (start > 0.0 && end > 0.0)
	{
	  s = (1 / end - 1 / start) / div;
	  for (int i = 0; i < datanum; i++)
	    {
	      val = 1 / (1 / start + s * (double)i);
	      outx->add(new dataelement(val, DATA_NORMAL));
	      outy->add(new dataelement(val, DATA_NORMAL));
	    }
	}
      else if (start < 0.0 && end < 0.0)
	{
	  s = (1 / end - 1 / start) / div;
	  for (int i = 0; i < datanum; i++)
	    {
	      val = 1 / (1 / start + s * (double)i);
	      outx->add(new dataelement(val, DATA_NORMAL));
	      outy->add(new dataelement(val, DATA_NORMAL));
	    }
	}
      else
	return false;
      break;
    }
  return true;
}
