/*----------------------------------------------------------------------------
                       globalbuff.cc (global buff object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string>
#include <stdio.h>
#include <stdlib.h>
#include "globalbuff.h"
extern void message(const char *msg);

globalbuff :: globalbuff() : plobj()
{
  buffer = new VariantBuff;
}

globalbuff :: ~globalbuff()
{
  delete buffer;
}


bool globalbuff::exec(std::string* function, buffarray *argument, tokenbuff* outvalue)
{
  std::string out, shrunk, str1;
  int iid, arglength, itemp;
  char stemp[1000];

  outvalue->value->setbuff(0);
  if (topobject(function, &out, &iid) == true)
    {
      arglength = argument->getarraylength();

      // remove
      if (out == std::string("remove") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (buffer->search(&(*argument)[0], &itemp))
		  return buffer->remove(itemp);
	      else
		return false;
	    }
	  else
	    return false;
	}
      // removeat
      else if (out == std::string("removeat") && iid == -1)
	{
	  if (arglength == 1)
	    return buffer->remove(atoi((*argument)[0].c_str()));
	  else
	    return false;
	}
      // set
      else if (out == std::string("set") && iid == -1)
	{
	  if (arglength == 2)
	      return buffer->SetScalar(&(*argument)[0], &(*argument)[1]);
	  else
	    return false;
	}
      // setat
      else if (out == std::string("setat") && iid == -1)
	{
	  if (arglength == 2)
	    {
	      itemp = atoi((*argument)[0].c_str());
	      if (itemp < buffer->getarraylength())
		{
		  (*(*buffer)[itemp].value)[0] = (*argument)[1];
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // get
      else if (out == std::string("get") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (buffer->search(&(*argument)[0], &itemp))
		{
		  if(buffer->GetScalar(&(*argument)[0], &str1))
		    return outvalue->value->add(str1.c_str());
		  else
		    return false;
		}
	      return false;
	    }
	  else
	    return false;
	}
      // getat
      else if (out == std::string("getat") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      itemp = atoi((*argument)[0].c_str());
	      if (itemp < buffer->getarraylength())
		  return outvalue->value->add((*(*buffer)[itemp].value)[0].c_str());
	      else
		return false;
	    }
	  else
	    return false;
	}
      // flush
      else if (out == std::string("flush") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      buffer->setbuff(0);
	      return true;
	    }
	  else
	    return false;
	}
      // length
      else if (out == std::string("length") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      sprintf(stemp, "%d", buffer->getarraylength());
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      // search
      else if (out == std::string("search") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (buffer->search(&(*argument)[0], &itemp))
		{
		  sprintf(stemp, "%d", itemp);
		  outvalue->value->add(stemp);
		}
	      else
		outvalue->value->add("-1");
	      return true;
	    }
	  else
	    return false;
	}
      // listbuffer
      else if (out == std::string("listbuffer") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      for (int i = 0; i < buffer->getarraylength(); i++)
		{
		  sprintf(stemp, "%d: %s = %s\n", i, (*buffer)[i].label.c_str(), (*(*buffer)[i].value)[0].c_str());
		  message(stemp);
		}
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;
}



