/*----------------------------------------------------------------------------
                     graph.h (definitions of graph object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__graph_h)
#define __graph_h

#include <stdio.h>
#include "plobj.h"
#include "frame.h"
#include "vect.h"
#include "buffarray.h"
#include "vectdefs.h"
#include "globalbuff.h"

typedef vect<frame> framearray;

class graph: public plobj
{
 public:
  graph();
  ~graph();

  bool get(std::string* member, tokenbuff* outvalue);  
  bool set(std::string* member, std::string* setvalue);  
  bool exec(std::string* function, buffarray *argument, tokenbuff* outvalue);  

  bool setframeid(int arrayno);

  double mag;
  int autoscalemargin;
  int currentframe;
  char graphname[1024];
  intpoint offset;
  intpoint papersize;
  bool updated;

  framearray *farray;
  globalbuff *gbuffer;
  globalbuff *init;

  /* graph read / write  */
  bool readgraph(FILE *fp);
  bool appendgraph(FILE *fp);
  bool writegraph(FILE *fp);
  bool checkfileexist();
  bool loadtobuff(char *infile, int frameno, int dataid);
  void flushtmpfiles();
  bool drawpaper(bool flag, FILE* f);
  int scale(int x);

  void versionstring(char *s);
};

#endif








