/*----------------------------------------------------------------------------
               kanji.cc (several routines related to kanji character)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include "kanji.h"

bool maykanji(char c)
{
#if defined(TOPAZ_SJIS)
    if (((unsigned char)c >= 0x81 && (unsigned char)c <= 0x9f) ||
        ((unsigned char)c >= 0xe0 && (unsigned char)c <= 0xfc))
      return true;
    else
      return false;
#else
    if ((unsigned char)c >= 0xa1)
      return true;
    else
      return false;
#endif

}

int charlength(char c)
{
  if (maykanji(c))
    return 2;
  else
    return 1;
}

int charlengthback(const char *s, int pos)
{
#if defined(TOPAZ_SJIS)
  int p = 0;
  int last = 1;
  
  while(p < pos)
    {
      if (maykanji(s[p]))
        {
          last = 2;
          p += 2;
        }
      else
        {
          last = 1;
          p++;
        }
    }
  return last;
#else
  if (maykanji(s[pos-1]))
    return 2;
  else
    return 1;
#endif
}

bool iskanji1(char *pos)
{
  char *p;
  p = pos;

  unsigned char u, u1;
  u = *pos;
  u1 = *(pos + 1);

#if defined(TOPAZ_SJIS)
    if (
         ((u >= 0x81 && u <= 0x9f) || (u >= 0xe0 && u <= 0xfc))
         &&
         (u1 >= 0x40 && u1 <= 0xfc && u1 != 0x7f)
       )
      return true;
    else
      return false;

#else

  int count = 0;
  if (maykanji(*p))
    {
      while (*p != 0)
	  {
	    if (maykanji(*p))
	      count++;
	    p++;
	  }
      return ((count % 2 + 1) % 2);
    }
  else
    return false;
#endif
}

void tojis(unsigned int *s)
{

  unsigned int h, l;

  l = (*s >> 8) & 0xff;
  h = *s & 0xff;

  *s = ((h << 8) | l) - 0x8080 ;

}

void tosjis(unsigned int *s)
{
  unsigned int h, l;

  l = (*s >> 8) & 0xff;
  h = *s & 0xff;


  if (h & 1)
    l += 0x1f;
  else
    l += 0x7d;

  if (l >= 0x7f)
    l++;

  h = ((h - 0x21) >> 1) + 0x81;

  if (h > 0x9f)
    h += 0x40;

  *s = (h << 8) | l;
  return;
}

void sjistojis(unsigned int *s)
{
  unsigned int h, l;

  l = (*s >> 8) & 0xff;
  h = *s & 0xff;

  if (h <= 0x9f)
    h -= 0x71;
  else
    h -= 0xb1;
  
  h = (h << 1) + 1;


  if (l >= 0x7f)
    l -= 0x01;

  if (l >= 0x9e)
  {
    l -= 0x7d;
    h++;
  }
  else
  l -= 0x1f;

  *s = (h << 8) | l;
  return;
}
