/*----------------------------------------------------------------------------
                   loaddata.cc (load numerical data from disk)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include "data.h"
#include "frame.h"
#include "topazvalues.h"
#include "convtxt.h"

extern char confpath[1000];
extern char tmppath[1000];
extern char pid[100];
extern char *toshortfilename(char *in);
extern void message(const char *msg);

int NumhyphnInToken(const char *s)
{
  char *p;
  int count = 0;
  p = (char *)s;
  while (*p != 0)
    {
      if (*p == '-')
	count++;
      p++;
    }
  return count;
}


bool data::extradataformat(const char *s, bool CheckMode)
{
  std::string str, buf, str1;
  char *p, *q, stemp[3];
  std::string nul = std::string("");
  bool IsSecondPara = false;
  int value, startvalue = 0, endvalue, i;

  if (CheckMode == false)
    {
      for (i = 0; i < 99; i++)
	extradatadispflag[i] = false;
    }
  
  buf = std::string("");
  //remove space
  p = (char *)s;
  while (*p != 0)
    {
      if (*p != ' ')
	{
	  stemp[0] = *p;
	  stemp[1] = 0;
	  buf += std::string(stemp);
	}
      p++;
    }
  if (buf == nul)
    return true;
  
  buf += std::string(",");
  
  str = std::string("");
  p = (char *)buf.c_str();
  while (*p != 0)
    {
      if (*p != ',' && *p != 0)
	{
	  if (strchr("0123456789-", *p) == 0)
	    return false;
	  
	  stemp[0] = *p;
	  stemp[1] = 0;
	  str += std::string(stemp);
	}
      else
	{
	  if (str == nul)
	    return false;
	  else
	    {
	      if (NumhyphnInToken(str.c_str()) == 0)
		{
		  if ((value = atoi(str.c_str())) == 0)
		    return false;
		  else
		    {
		      if (value < 1 || value > 99)
			return false;
		      
		      if (CheckMode == false)
			extradatadispflag[value - 1] = true;
		    }
		}
	      else if (NumhyphnInToken(str.c_str()) == 1)
		{
		  IsSecondPara = false;
		  str += std::string("-");
		  q = (char *)str.c_str();
		  str1 = std::string("");
		  while (*q != 0)
		    {
		      if (*q != '-')
			{
			  stemp[0] = *q;
			  stemp[1] = 0;
			  str1 += std::string(stemp);
			}
		      else
			{
			  if (str1 == nul)
			    return false;
			  else
			    {
			      if (IsSecondPara == true)
				{
				  if ((endvalue = atoi(str1.c_str())) == 0)
				    return false;
				  else
				    {
				      if (endvalue < 1 || endvalue > 99)
					return false;
				      
				      if(startvalue > endvalue)
					return false;
				      else
					{
					  if (CheckMode == false)
					    {
					      for (i = startvalue - 1; i <= endvalue - 1; i++)
						extradatadispflag[i] = true;
					    }
					  IsSecondPara = false;
					}
				    }
				}
			      else
				{
				  if ((startvalue = atoi(str1.c_str())) == 0)
				    return false;
				  else
				    {
				      if (startvalue < 1 || startvalue > 99)
					return false;
				    }
				}
			      
			      if (IsSecondPara == false)
				IsSecondPara = true;
			      str1 = std::string("");
			    }
			}
		      q++;
		    }
		}
	      else
		return false;
	    }
	  
	  str = std::string("");
	}
      p++;
    }
  return true;
}

bool iscommentlinedata(char c)
{
  if (c == 39   || 
      c == '\0' ||
      c == '"'  ||
      c == ';'  ||
      c == '/'  || 
      c == '#'  || 
      c == '%'  || 
      c == 13   || 
      c == 10)
    return true;
  else
    return false;
}

bool data::loaddata()
{
  FILE *fp;
  bool flag;
  char filen[1000];
  char stemp[1000];
  frame* fr = (frame *)parent;

  if (!isfiledata)
    return true;

  if (!virgin && fr->withdata)
    {
      strcpy(filen, tmppath);
      strcat(filen, "/");
      strcat(filen, pid);
      sprintf(stemp, "%d_", fr->id);
      strcat(filen, stemp);
      sprintf(stemp, "%d_", id);
      strcat(filen, stemp);
      strcat(filen, toshortfilename(filename));
    }
  else
    strcpy(filen, filename);
  /*  else
    {
      if (datasearchpath == PATH_ABSOLUTE)
	strcpy(filen, filename);
      else if (datasearchpath == PATH_CURRENT)
	strcpy(filen, toshortfilename(filename));
      else
	return false;
    }
    */
  // file open
  if ((fp = fopen(filen, "r")) == NULL)
    {
      sprintf(stemp, "Error!!: Cannot open '%s'\n", filen);
      message(stemp);
      return false;
    }

  flag = loaddataproc(fp);
  // file close
  fclose(fp);

  return flag;
}

bool data::loaddataproc(FILE *fp)
{
  int i, j;
  char b[100];
  int nkomma;
  int initnkomma = 0;
  char *p;
  int flag;
  double ftemp;
  char *ctemp, *start;
  char tempstr[100];
  int tokcount;
  int StartLine, EndLine, Skip;
  int count;
  int effectcount = 0;
  dataelement TempElement;
  bool IsDataStart = false;
  int maxcolno = -1;
  char datastr[2000], ndatastr[200];

  //check file or memory data
  if (!isfiledata)
    return (true);
  
  StartLine = readstart;
  EndLine = readend;
  Skip = skip;
  
  if (EndLine == -1)
    EndLine = MAXINT;
  
  // to readstartline
  for (i = 1; i < StartLine; i++)
    {
      if (topazfgets(datastr, 2000, fp) == NULL)
	  return false;
    }

  //clear data buffer
  x->setbuff(0);
  xmax->setbuff(0);
  xmin->setbuff(0);
  y->setbuff(0);
  ymax->setbuff(0);
  ymin->setbuff(0);
  x->setbuff(0);
  for (i = 0; i < 99; i++)
    extra[i]->setbuff(0);

  // set boundvalues
  bound_xmin  = MAXDOUBLE;
  bound_xmax  = -MAXDOUBLE;
  bound_ymin  = MAXDOUBLE;
  bound_ymax  = -MAXDOUBLE;
  
  extradataformat(extradata->c_str(), false);
  
  // get maxcolno
  if (col_x >= maxcolno)  maxcolno = col_x;
  if (col_y >= maxcolno)  maxcolno = col_y;
  if (bar_x)
    {
      if (col_xmin >= maxcolno) maxcolno = col_xmin;
      if (col_xmax >= maxcolno) maxcolno = col_xmax;
    }
  if (bar_y)
    {
      if (col_ymin >= maxcolno) maxcolno = col_ymin;
      if (col_ymax >= maxcolno) maxcolno = col_ymax;
    }
  for (i = 0; i < 99; i++)
    if (i + 1 >= maxcolno && extradatadispflag[i]) maxcolno = i + 1;

  count = -1;  //initialization of countor
  for (i = 1; i <= EndLine - StartLine + 1; i++)
    {
      /* start reading */
      if (topazfgets(datastr, 2000, fp) == NULL)
	{
	  //read error / end of file
	  return true;
	}
      
      // remove '\n' at the end of the line
      if (datastr[strlen(datastr) - 1] == '\n')
	datastr[strlen(datastr) - 1] = 0;

      // remove space
      if (strlen(datastr) > 0)
	{
	  // remove end space
	  for (p = datastr + strlen(datastr) - 1; p >= datastr; p--) 
	    {
	      if (*p != ' ')
		{
		  *(p + 1) = 0;
		  break;
		}
	    }
	  // remove top space
	  for (p = datastr; *p != 0; p++) 
	    {
	      if (*p != ' ')
		  break;
	    }
	  start = p;
	}
      else
	start = datastr;

      // remove comment lines
      if (IsDataStart == false)
	{
	  if (iscommentlinedata(*start))
	    goto j1;
	  else
	    IsDataStart = true;
	}
      
      // change tab("\t") and  space(" ")  "," to ","
      strcpy(ndatastr, "");
      nkomma = 0;
      flag = false;
      
      if ((strchr(start, ',') != 0) || (strchr(start, '\t') != 0))
        {
	  for (p = start; *p != 0; p++)
	    {
	      if (*p == '\t' || *p == ',')
		{
		  strcpy(b, " , ");
		  strcat(ndatastr, b);
		  nkomma++;
		  if (count == -1)
		    initnkomma++;
		}
	      else
		{
		  b[0] = *p;
		  b[1] = '\0';
		  if (b[0] == 'd' || b[0] == 'D')
		    strcat(ndatastr, "e");
		  else
		    strcat(ndatastr, b);
		}
	    }
        }
      else
        {
	  for (p = start; *p != 0; p++)
	    {
	      if (*p == ' ')
		{
		  // continuous space
		  if (flag)
		    flag = true;
		  // meet first space
		  else
		    {
		      flag = true;
		      strcpy(b, " , ");
		      strcat(ndatastr, b);
		      nkomma++;
		      
		      if (count == -1)
			initnkomma++;
		    }
		}
	      else
		{
		  flag = false;
		  
		  b[0] = *p;
		  b[1] = '\0';

		  if (b[0] == 'd' || b[0] == 'D')
		    strcat(ndatastr, "e");
		  else
		    strcat(ndatastr, b);
		}
	    }
        }

      // all selected columns are valid?
      if (count == -1)
	{
	  if ((col_x > initnkomma + 1) || (col_y > initnkomma + 1)
	      )
	      return false;

	  if (bar_x)
	    {
	      if ((col_xmax > initnkomma + 1)
		  || (col_xmin > initnkomma + 1)
		  )
		  return false;
	    }
	  if (bar_y)
	    {
	      if ((col_ymax > initnkomma + 1)
		  || (col_ymin > initnkomma + 1)
		  )
		  return false;
	    }
	}


      // is commentline?
      if (iscommentlinedata(*ndatastr))
	{
	  switch(commentlineaction)
	    {
	    case COMMENT_ABORT:
	      return true;
	      break;
	    case COMMENT_IGNORE:
	      effectcount = 0;
	      goto j1;
	      break;
	    case COMMENT_SEPARTION:
	      strcpy(ndatastr, "");
	      for (j = 1; j <= initnkomma ; j++)
		strcat(ndatastr, "*,");
	      strcat(ndatastr, "*");
	      effectcount = 0;
	      break;
	    }
	}
      else
	{
	  effectcount++;
	  // skip
	  if (Skip != 1)
	    {
	      if (effectcount % Skip != 1)
		goto j1;
	    }
	  
	  // valid comma number?
	  if (nkomma != initnkomma)
	    {
	      // if not, abort.
	      return true;
	    }
	}

      if (initnkomma + 1 <= 99)
	{
	  for (j = initnkomma + 1 ; j < 99; j++)
	    extradatadispflag[j] = false;
	}
      
      p = strtok(ndatastr, ",");
      tokcount = 1;
      while (p)
	{
	  if (maxcolno < tokcount)
	    goto j0;

	  strncpy(tempstr, p, 38);
	  ftemp = strtod(p, &ctemp);

	  //input data
	  if (col_x == 0 && tokcount == 1) //count only 1 times
	    {
	      if (!x->add(new dataelement((double)count + 2.0, DATA_NORMAL)))
		goto j2;
	    }
	  else
	    {
	      if (tokcount == col_x)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!x->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!x->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!x->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		      if (ftemp < bound_xmin)
			bound_xmin = ftemp;
		      if (ftemp > bound_xmax)
			bound_xmax = ftemp;
		    }
		}
	    }
	  if (bar_x)
	    {
	      if (tokcount == col_xmin)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!xmin->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!xmin->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!xmin->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		    }
		}
	      if (tokcount == col_xmax)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!xmax->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!xmax->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!xmax->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		    }
		}
	    }

	  if (col_y == 0 && tokcount == 1) //count only 1 times
	    {
	      if (!y->add(new dataelement((double)count + 2.0, DATA_NORMAL)))
		goto j2;
	    }
	  else
	    {
	      if (tokcount == col_y)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!y->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!y->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!y->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		      if (ftemp < bound_ymin)
			bound_ymin = ftemp;
		      if (ftemp > bound_ymax)
			bound_ymax = ftemp;
		    }
		}
	    }
	  
	  if (bar_y)
	    {
	      if (tokcount == col_ymin)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!ymin->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!ymin->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!ymin->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		    }
		}
	      if (tokcount == col_ymax)
		{
		  if (strchr(tempstr,'*') != 0)
		    {
		      if (!ymax->add(new dataelement(0.0, DATA_INVALID)))
			goto j2;
		    }
		  else if (strchr(tempstr,'|') != 0)
		    {
		      if (!ymax->add(new dataelement(0.0, DATA_IGNORE)))
			goto j2;
		    }
		  else
		    {
		      if (!ymax->add(new dataelement(ftemp, DATA_NORMAL)))
			goto j2;
		    }
		}
	    }
	  
	  if (extradatadispflag[tokcount - 1])
	    {
	      if (strchr(tempstr,'*') != 0)
		{
		  if (!extra[tokcount - 1]->add(new dataelement(0.0, DATA_INVALID)))
		    goto j2;
		}
	      else if (strchr(tempstr,'|') != 0)
		{
		  if (!extra[tokcount - 1]->add(new dataelement(0.0, DATA_IGNORE)))
		    goto j2;
		}
	      else
		{
		  if (!extra[tokcount - 1]->add(new dataelement(ftemp, DATA_NORMAL)))
		    goto j2;
		}
	    }
	  
	  tokcount++;
	  p = strtok(0,",");
	}
    j0:;
      count++;
    j1:;
      
    }
  //end of for
  return true;
  
  
 j2:
  
  //clear data buffer
  x->setbuff(0);
  xmin->setbuff(0);
  xmax->setbuff(0);
  y->setbuff(0);
  ymin->setbuff(0);
  ymax->setbuff(0);
  for (j = 0; j < 99; j++)
    extra[j]->setbuff(0);
  return true;
}








