/*----------------------------------------------------------------------------
                           msgbuff.cc (mesage area)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <memory.h>
#include <string.h>
#include "topazcurses.h"
#include "msgbuff.h"
#include "kanji.h"

extern int tabnumber;
extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;

void printlinetext(WINDOW *win, char *text)
{
  char *p = text;
  char line[1000], stemp[3];

  strcpy(line, "");
  while(*p != 0)
    {
      if (iskanji1(p))
	{
	  stemp[0] = *p++; stemp[1] = *p++; stemp[2] = 0;
	}
      else if (*p == '%')
	{
	  stemp[0] = '%'; stemp[1] = '%'; stemp[2] = 0; p++;
	}
      else
	{
	  stemp[0] = *p++; stemp[1] = 0;
	}
      strcat(line, stemp);
    }
  wprintw(win, line);
}

msgbuff::msgbuff(WINDOW *msgwin, int msgwinlines, int buffline)
{
  win = msgwin;
  winlines = msgwinlines;

  width = 78;
  buff = new char[width * buffline];
  memset(buff, 0, width * buffline);
  pointer = 0;
  topline = 0;
  maxline = buffline - 1;
  flush();
}

msgbuff::~msgbuff()
{
  delete[] buff;
}

void msgbuff::copytomsgwin()
{
  if (win == 0)
    return;

  if (pointer + winlines - 1 <= topline)
    {
      wclear(win);
      wrefresh(win);
      for (int i = 0; i < winlines; i++)
	{
	  wmove(win, i, 0);
	  printlinetext(win, buff + width * (i + pointer));
	}
      wrefresh(win);
    }
  else
    {
      wclear(win);
      wrefresh(win);
      for (int i = topline - winlines + 1; i <= topline; i++)
	{
	  wmove(win, i - topline + winlines - 1, 0);
	  printlinetext(win, buff + width * i);
	}
      wrefresh(win);
    }
}

void msgbuff::addtext(const char *str)
{
  char *p;
  p = (char *)str;
  char buf[100], stemp[5];
  int itemp;
  int inc;
  bool ispercent = false;

  if (win == 0)
    return;

  while (*p != 0)
    {
      strcpy(stemp, "");
      strcpy(buf, buff + width * topline);

      if (iskanji1(p))
	{
	  inc = 2;
	  ispercent = false;
	}
      else
	{
	  inc = 1;
	  if (*p == '%' && *(p + 1) == '%')
	    ispercent = true;
	  else
	    ispercent = false;
	}

      if (inc == 1)
	{
	  if (ispercent)
	    {
	      stemp[0] = *p++; p++;
	      stemp[1] = 0;
	    }
	  else
	    {
	      stemp[0] = *p++;
	      stemp[1] = 0;
	    }
	}
      else
	{
	  stemp[0] = *p++;
	  stemp[1] = *p++;
	  stemp[2] = 0;
	}
      
      if ((signed int)strlen(buf) < width - inc)
	{
	  if (*stemp == '\n')
	    {
	      add("");
	    }
	  else if (*stemp == '\t')
	    {
	      itemp = (strlen(buf) / tabnumber + 1) * tabnumber;
	      if (itemp < width - 1)
		{
		  if (topline - pointer > winlines)
		    {
		      pointer = topline - winlines + 1;
		      copytomsgwin();
		      strcpy(buf, buff + width * topline);
		      wmove(win, winlines - 1, strlen(buf));
		    }
		  else
		    wmove(win, topline, strlen(buf));

		  for (int i = strlen(buf); i < itemp; i++)
		    {
		      strcat(buf, " ");
		      strcat(buff + width * topline, " ");
		      wprintw(win, " ");
		    }
		}
	      else
		{
		  add("");
		}
	    }
	  else
	    {
	      strcat(buf, stemp);
	      strcat(buff + width * topline, stemp);

	      if (topline - pointer > winlines)
		{
		  pointer = topline - winlines + 1;
		  copytomsgwin();
		  strcpy(buf, buff + width * topline);
		  wmove(win, winlines - 1, strlen(buf) - inc);
		}
	      else
		wmove(win, topline, strlen(buf) - inc);

	      if (ispercent)
		wprintw(win, "%%");
	      else
		wprintw(win, stemp);
	    }
	}
      else
	{
	  add(stemp);
	}
    }
   wrefresh(win);
 }


void msgbuff::add(const char *str)
{

  if (win == 0)
    return;

  if (topline == maxline)
    {
      memcpy(buff, buff + width, width * maxline);
      strcpy(buff + width * topline, str);

      if (topline - pointer <  winlines)
	{
	  wscrl(win, 1); 
	  wmove(win, topline - pointer, 0);
	  printlinetext(win, buff + width * topline);
	  wrefresh(win);
	}
      else if (topline - pointer == winlines)
	{
	  pointer = topline - winlines + 1;
	  wscrl(win, 1); 
	  wmove(win, winlines - 1, 0);
	  printlinetext(win, buff + width * (winlines - 1 + pointer));
	  wrefresh(win);
	}
      else if (topline - pointer > winlines)
	{
	  pointer = topline - winlines + 1;
	  copytomsgwin();
	  wrefresh(win);
	}
    }
  else
    {
      topline++;
      strcpy(buff + width * topline, str);

      if (topline - pointer <  winlines)
	{
	  pointer = 0;
	  wmove(win, topline, 0);
	  printlinetext(win, buff + width * topline);
	  wrefresh(win);
	}
      else if (topline - pointer == winlines)
	{
	  pointer = topline - winlines + 1;
	  wscrl(win, 1); 
	  wmove(win, winlines - 1, 0);
	  printlinetext(win, buff + width * (winlines - 1 + pointer));
	  wrefresh(win);
	}
      else if (topline - pointer > winlines)
	{
	  pointer = topline - winlines + 1;
	  copytomsgwin();
	  wrefresh(win);
	}
    }
}

bool msgbuff::next(int paging)
{
  if (win == 0)
    return true;

  if (pointer + winlines - 1 + paging <= topline)
  {
    pointer += paging;
    wscrl(win, paging); 
    wmove(win, winlines - 1, 0);
    printlinetext(win, buff + width * (winlines - 1 + pointer));
    wrefresh(win);
    return true;
  }
  else
    return false;
}

bool msgbuff::previous(int paging)
{
  if (win == 0)
    return true;

  if (pointer - paging >= 0)
    {
      pointer -= paging;
      wscrl(win, -paging); 
      wmove(win, 0, 0);
      printlinetext(win, buff + width * pointer);
      wrefresh(win);
      return true;
    }
  else
    return false;
}

void msgbuff::print(const char *str)
{
  if (win == 0)
    fprintf(stderr, str);
  else
    addtext(str);
}

void msgbuff::flush()
{
  if (win == 0)
    return;
  
  pointer = 0;
  topline = 0;
  wclear(win);
  wrefresh(win);
  wmove(win, 0, 0);
  memset(buff, 0, width * (maxline + 1));
  wrefresh(win);

}


int msgbuff::ems(int lineno, char *mes, const char *com)
{
  std::string tempstr;
  char stemp[100];

  sprintf(stemp, "%d: ", lineno);
  tempstr = std::string(stemp);
  tempstr += std::string("Error!: ");
  tempstr += std::string(mes);
  tempstr += std::string(" '");
  tempstr += std::string(com);
  tempstr += std::string("'\n");
  print (tempstr.c_str());

  return false;
}



