/*----------------------------------------------------------------------------
                   nonlinearfit.cc (non-linear fitting routine)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include "data.h"
#include "buffarray.h"
#include "frame.h"
#include "naninf.h"
#include "graph.h"

extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);
extern graph* gra;
extern bool isabort();
doublearray* dyda[10];
int list[10];
int NumPar;
double formulaconst[27];
std::string MarDyDX[10];
std::string MarFuncs;


bool matinv(int n, double **a, double **ia)
{
	double dv, diag, f;
	
	for (int i = 1; i <= n; i++)
		for (int j = 1; j <= n; j++)
			ia[i][j] = a[i][j];

	dv = 1.0;
	for (int i = 1; i <= n; i++)
	{
		diag = ia[i][i];
		if (diag == 0.0)
			return false;
		dv *= diag;

		for (int j = 0; j <= n; j++)
			ia[i][j] /= diag;

		ia[i][i] = 1.0 / diag;
			
		for (int j = 1; j <= n; j++)
		{
			if (j != i)
			{
				f = ia[j][i];
				for (int k = 1; k <= n; k++)
				{
					if (k == i)
						ia[j][k] = -f / diag;
					else
						ia[j][k] -= ia[i][k] * f;
				}

			}

		}

	}
	return true;
}


bool solve(double **a, int n, double *b)
{
  double **ta, *tb;
  int i , j;
  bool ErrorFlag = false;
  
  ta = new double *[10];
  for (i = 0;i <= 9; i++)
    ta[i] = new double [10];
  
  tb = new double [10];
  for (i = 1; i <= n; i++)
    tb[i] = b[i];


  if (matinv(n, a, ta) == false)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  for (i = 1; i <= n; i++)
    {
      b[i] = 0.0;
      for (j = 1; j <= n; j++)
	b[i] += ta[i][j] * tb[j];
    }
  
j1:;
  delete tb;
  for (i = 0;i <= 9; i++)
    delete ta[i];
  delete [] ta;

  if (ErrorFlag)
    return false;
  else
    return true;
}

bool data::UserFuncs(data *dat, int Ndata, double *a, int na, int Npara)
{
  int i, j, ii, index;
  bool Flag;
  double ftemp;
  double temp_a[10];
  double temp_dyda2;
  double da;
  frame *fr = (frame *)parent;
  std::string label, value, str1, str2;
  char stemp[100], stemp1[100];
  
  for (i = 1; i <= na; i++)
    {
      sprintf(stemp, "%c", i - 1 + 'A');
      sprintf(stemp1, "%.15g", a[i]);
      str1 = std::string(stemp);
      str2 = std::string(stemp1);
      fr->usrconst->SetSystemScalar(&str1, &str2);
    }

  // get value
  Flag = CalcFormulastring((char *)MarFuncs.c_str(), &ftemp, true, dat, 0, fr->getdataindex(id), 0);
  if (Flag == false)
    return false;
  
  for (ii = 1; ii <= Ndata; ii++)
    {
      Flag = CalcFormulastring((char *)MarFuncs.c_str(), &ftemp, false, dat, ii - 1, fr->getdataindex(id), 0);
      if (Flag)
	(*dyda[0])[ii] = ftemp;
      else
	return false;
      
    }

  // get derivative value

  for (j = 0; j <= 9; j++)
    temp_a[j] = a[j];
  for (i = 1; i <= Npara; i++)
    {
      index = list[i];
      // derivative is given
      if (MarDyDX[index] != std::string(""))
	{
	  Flag = CalcFormulastring((char *)MarDyDX[index].c_str(), &ftemp, true, dat, 0, fr->getdataindex(id), 0);
	  if (Flag == false)
	    return false;
	  for (ii = 1; ii <= Ndata; ii++)
	    {
	      Flag = CalcFormulastring((char *)MarDyDX[index].c_str(), &ftemp, false, dat, ii - 1, fr->getdataindex(id), 0);
	      if (Flag)
		(*dyda[index])[ii] = ftemp;
	      else
		return false;
	    }
	}
      // not given
      else
	{
	  da = a[index] * 1e-5;
	  if (da == 0.0)
	    da = 1e-5;

	  formulaconst[index] = temp_a[index] + da;

	  sprintf(stemp, "%c", index - 1 + 'A');
	  sprintf(stemp1, "%.15g", formulaconst[index]);
	  str1 = std::string(stemp);
	  str2 = std::string(stemp1);
	  fr->usrconst->SetSystemScalar(&str1, &str2);

	  Flag = CalcFormulastring((char *)MarFuncs.c_str(), &ftemp, true, dat, 0, fr->getdataindex(id), 0);
	  if (Flag == false)
	    return false;
	  
	  for (ii = 1; ii <= Ndata; ii++)
	    {
	      Flag = CalcFormulastring((char *)MarFuncs.c_str(), &ftemp, false, dat, ii - 1, fr->getdataindex(id), 0);
	      if (Flag)
		temp_dyda2 = ftemp;
	      else
		return false;
	      
	      (*dyda[index])[ii] = (temp_dyda2 - (*dyda[0])[ii]) / da;
	    }

	  // restore parameters
	  formulaconst[index] = temp_a[index];
	  sprintf(stemp, "%c", index - 1 + 'A');
	  sprintf(stemp1, "%.15g", formulaconst[index]);
	  str1 = std::string(stemp);
	  str2 = std::string(stemp1);
	  fr->usrconst->SetSystemScalar(&str1, &str2);
	}
    }

  return true;
}



bool data::MarqSetPara(data *dat, int Ndata, double *a, int Na, int *list,
			int Npara, double **alpha, double *beta, double *X2)
{
  int i, j, k;
  double dy;
  bool ErrorFlag = false;

  for (j = 1; j <= Npara; j++)
    {
      for (k = 1; k <= j; k++)
	alpha[j][k] = 0.0;
      beta[j] = 0.0;
    }

  *X2 = 0.0;

  if (UserFuncs(dat, Ndata, a, Na, Npara) == false)
    {
      ErrorFlag = true;
      goto j1;
    }

  for (i = 1; i <= Ndata; i++)
    {
      dy = (*dat->y)[i - 1].data - (*dyda[0])[i];
      
      for (j = 1; j <= Npara; j++)
	{
	  for (k = 1; k <= j; k++)
	    {
	      alpha[j][k] += (*dyda[list[j]])[i] * (*dyda[list[k]])[i];
	      alpha[k][j] = alpha [j][k];
	    }
	  
	  beta[j] += dy * (*dyda[list[j]])[i];
	}
      (*X2) += dy * dy;
    }
 j1:;
  
  if (ErrorFlag == false)
    return true;
  else
    return false;
}


bool data::ExecMarquadt(data *dat, int Ndata, double *a, int *list,
			int Npara, double init_chisq, double *out_chisq)
{
  double **temp_gauss, **alpha, X2, lamda;
  int i, count = 0;
  int j, k, Na = 9;
  bool ErrorFlag = false;
  double initchisq, oldchisq;
  double *da, *temp_a, *temp_da, *beta, OX2;

  temp_gauss = new double *[10];

  for (i = 0;i <= 9; i++)
    temp_gauss[i] = new double [10];

  alpha = new double *[10];

  for (i = 0;i <= 9; i++)
    alpha[i] = new double [10];


  temp_da = new double [Npara + 1];
  temp_a = new double [Na + 1];
  da   = new double [Na + 1];
  beta = new double [Na + 1];

  // initial
  
  lamda = .001;

  if (MarqSetPara(dat, Ndata, a, Na, list, Npara, alpha, beta, &X2) == false)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  OX2 = X2;
  
  for (j = 1; j <= Npara; j++)
    {
      for ( k = 1; k <= Npara; k++)
	temp_gauss[j][k] = alpha[j][k];
      temp_gauss[j][j] = alpha[j][j] * (1.0 + lamda);
      temp_da[j] = beta[j];
    }
  
  if (solve(temp_gauss, Npara, temp_da) == false)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  for (j = 1; j <= Npara; j++)
    da[j] = temp_da[j];

  for (j = 1; j <= Na; j++)
    temp_a[j] = a[j];

  for (j = 1; j <= Npara; j++)
    temp_a[list[j]] = a[list[j]] + da[j];

  if (MarqSetPara(dat, Ndata, temp_a, Na, list, Npara, temp_gauss, da, &X2) == false)
    {
      ErrorFlag = true;
      goto j1;
    }

  if (X2 < OX2)
    {
      lamda *= 0.1;
      OX2 = X2;
      for (j = 1; j <= Npara; j++)
	{
	  for (k=1; k <=Npara; k++)
	    alpha[j][k] = temp_gauss[j][k];
	  
	  beta[j] = da[j];
	  a[list[j]] = temp_a[list[j]];
	}
    }
  else
    {
      lamda *=10.0;
      X2 = OX2;
    }

  
  //iteration routine
  initchisq = X2;
  
  do
    {
      if (isabort())
	goto j1;

      oldchisq = X2;
      count++;
      
      for (j = 1; j <= Npara; j++)
	{
	  for (k = 1; k <= Npara; k++)
	    temp_gauss[j][k] = alpha[j][k];
	  temp_gauss[j][j] = alpha[j][j] * (1.0 + lamda);
	  temp_da[j] = beta[j];
	}
      if (solve(temp_gauss, Npara, temp_da) == false)
	{
	  ErrorFlag = true;
	  goto j1;
	}

      for (j = 1; j <= Npara; j++)
	da[j] = temp_da[j];


      for (j = 1; j <= Na; j++)
	temp_a[j] = a[j];

      for (j = 1; j <= Npara; j++)
	temp_a[list[j]] = a[list[j]] + da[j];

      if (MarqSetPara(dat, Ndata, temp_a, Na, list, Npara, temp_gauss, da, &X2) == false)
	{
	  ErrorFlag = true;
	  goto j1;
	}

      if (X2 < OX2)
	{
	  lamda *= 0.1;
	  OX2 = X2;
	  for (j = 1; j <= Npara; j++)
	    {
	      for (k=1; k <=Npara; k++)
		alpha[j][k] = temp_gauss[j][k];
	      
	      beta[j] = da[j];
	      a[list[j]] = temp_a[list[j]];
	    }
	}
      else
	{
	  lamda *=10.0;
	  X2 = OX2;
	}
      
      //iteraton error
      if (count >= 100)
	{
	  ErrorFlag = true;
	  goto j1;
	}
    }while (fabs(X2 -oldchisq) > initchisq * init_chisq);


  //final try
  for (j = 1; j <= Npara; j++)
    {
      for (k = 1; k <= Npara; k++)
	temp_gauss[j][k] = alpha[j][k];
      temp_gauss[j][j] = alpha[j][j] * (1.0 + lamda);
      temp_da[j] = beta[j];
    }
  
  if (solve(temp_gauss, Npara, temp_da) == false)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  for (j = 1; j <= Npara; j++)
    da[j] = temp_da[j];
  

j1:;
  delete beta;
  delete da;
  delete temp_a;
  delete temp_da;

  for (i = 0;i <= 9; i++)
    delete temp_gauss[i];

  delete [] temp_gauss;

  for (i = 0;i <= 9; i++)
    delete alpha[i];

  delete [] alpha;

  *out_chisq = X2;

  if (ErrorFlag == false)
    return true;
  else
    return false;
}

bool data::OutMar(int StartLine
		  , int EndLine
		  , std::string* func
		  , buffarray* ini
		  , buffarray* dydA
		  , double seido
		  , std::string* outfn
		  , std::string* info
		  )
{
  bool ErrorFlag = false;
  int i, normalcount;
  double X2;
  int tempstartline,tempendline;
  std::string temp_pre_formula, temp_pre_formula_x, temp_pre_formula_y;
  double fa, fb;
  int NumPoint;
  std::string mes;
  data *dat;
  doublearray *in2x = 0, *in2y = 0;
  frame *fr = (frame *)parent;
  char stemp[100], stemp1[1000];
  std::string label, value, str, str1;
  int itemp;

  dat = new data;
  if (dat == 0)
    {
      dat = 0;
      ErrorFlag = true;
      goto j1;
    }
  dat->id = id;
  dat->parent = parent;

  in2x = new doublearray;
  if (in2x == 0)
    {
      in2x = 0;
      ErrorFlag = true;
      goto j1;
    }

  in2y = new doublearray;
  if (in2y == 0)
    {
      in2y = 0;
      ErrorFlag = true;
      goto j1;
    }


  //calclation
  NumPoint = x->getarraylength() - 1;
  if (trans_x)
    temp_pre_formula_x = *format_x;
  else
    temp_pre_formula_x = std::string("x");
  if (trans_y)
    temp_pre_formula_y = *format_y;
  else
    temp_pre_formula_y = std::string("y");
  
  for (i = 0; i <= NumPoint; i++)
    {
      CalcFormulastring((char *)temp_pre_formula_x.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_x.c_str(), &fa, false ,this, i, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_y.c_str(), &fb, true ,this, 0, fr->getdataindex(id), 0);
      CalcFormulastring((char *)temp_pre_formula_y.c_str(), &fb, false ,this, i, fr->getdataindex(id), 0);
      if (isnormalvalue(fa) && (*x)[i].status == 0 &&
	  isnormalvalue(fb) && (*y)[i].status == 0
	  )
	{
	  if (!in2x->add(fa))
	    goto j1;
	  if (!in2y->add(fb))
	    goto j1;
	}
    }

  // remove invalid data
  tempstartline = StartLine;
  if (EndLine == -1)
    tempendline = in2x->getarraylength();
  else
    {
      if (EndLine <= in2x->getarraylength())
	tempendline = EndLine;
      else
	tempendline = in2x->getarraylength();
    }

  normalcount = 0;

  for (i = tempstartline; i <= tempendline; i++)
    {
      if (isnormalvalue((*in2x)[i - 1]) && isnormalvalue((*in2y)[i - 1]))
	{
	  dat->x->add(new dataelement((*in2x)[i - 1], 0));
	  dat->y->add(new dataelement((*in2y)[i - 1], 0));
	  normalcount++;
	}
    }

  if (normalcount == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  for (i = 0; i < 10; i++)
    {
      dyda[i] = new doublearray;
      dyda[i]->setarray(normalcount + 1);
    }

  NumPar = 0;
  for (i = 1; i <= 9; i++)
    if ((*ini)[i] != std::string(""))
      NumPar++;

  itemp = 0;
  for (i = 1; i <= 9; i++)
    {
      if ((*ini)[i] != std::string(""))
	{
	  itemp++;
	  list[itemp] = i;
	}
    }

  // set formulaconst
  for (i = 1; i <= 9; i++)
    formulaconst[i] = atof((*ini)[i].c_str());

  // set formulaconst
  for (i = 1; i <= 9; i++)
    MarDyDX[i] = (*dydA)[i];

  // set function
  MarFuncs = *func;

  if (ExecMarquadt(dat, normalcount, formulaconst, list,
		   NumPar, seido / 100.0, &X2) == false)
    return false;

  for (i = 0; i < 10; i++)
    delete dyda[i];

  // create info
  mes = std::string("");
  mes = std::string("--------- fitting results ---------\\n");
  mes += std::string("\\n");
  mes += std::string("Original data: ");
  mes += std::string(filename);
  mes += std::string("\\n");
  mes += std::string("\\n");
  mes += std::string("Function: y = ");
  mes += MarFuncs;
  mes += std::string("\\n");
  mes += std::string("\\n");

  for (i = 1; i <= itemp; i++)
    {
      sprintf(stemp, "%c", list[i] + 'A' - 1);
      str1 = std::string(stemp);
      fr->usrconst->GetScalar(&str1, &str);
      sprintf(stemp1, "%c = %s\\n", list[i] + 'A' - 1, str.c_str());
      mes += std::string(stemp1);
    }
  mes += std::string("\\n");
  
  sprintf(stemp, "Allowance: %.15g %%\\n", seido);
  sprintf(stemp, "Fitting line: %d -> %d\\n", StartLine, EndLine);
  mes += std::string(stemp);

  mes += std::string("\\n");
  mes += std::string("--------------- end ---------------\\n");
  
  *info = mes;
  createfunction(&MarFuncs, outfn);
 j1:;
  
  if (dat != 0)
    delete dat;
  if (in2x != 0)
    delete in2x;
  if (in2y != 0)
    delete in2y;
  
  if (ErrorFlag == true)
    return false;
  else
    return true;
}


bool iscoef(char c)
{
  if ('A' <= c && c <= 'I')
    return true;
  else
    return false;
}

bool isseparator(char c)
{
  if (isalpha(c) == 0 && isdigit(c) == 0 && (c != '_'))
    return true;
  else
    return false;
}

void data::createfunction(std::string *in, std::string *out)
{
  char* p = (char *)in->c_str();
  bool prevsep = false;
  std::string str = std::string(""), str1, str2;
  char stemp[100];
  frame *fr = (frame *)parent;
  
  while (*p != 0)
    {
      if (isseparator(*p))
	{
	  prevsep = true;
	  stemp[0] = *p; stemp[1] = 0; str += std::string(stemp);
       	}
      else
	{
	  if (iscoef(*p))
	    {
	      if (isseparator(*(p + 1)))
		{
		  sprintf(stemp, "%c", *p);
		  str2 = std::string(stemp);
		  fr->usrconst->GetScalar(&str2, &str1);
		  str += std::string("(") + str1 + std::string(")");
		}
	      else
		{
                  prevsep = false;
                  stemp[0] = *p; stemp[1] = 0; str += std::string(stemp);
                }
	    }
	  else
	    {
              prevsep = false;
              stemp[0] = *p; stemp[1] = 0; str += std::string(stemp);
            }
	}
      p++;
    }
  *out = str;
}








