/*----------------------------------------------------------------------------
                      objectlist.cc (objectlist interfaces)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 99


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include "topazcurses.h"
#include "objectlist.h"
#include "graph.h"
#include "cui.h"
#include "script.h"
#include "vectdefs.h"
#include "styledisp.h"
#include "data.h"
#include "parts.h"
#include "gdi.h"
#include "frame.h"
#include "filepoint.h"
#include "topazvalues.h"
#include "regexpress.h"

extern FileHandleArray *fhvuff;
extern int menuwinlines;
extern graph *gra;
extern int maxrow;
extern int leftmargin;
extern int topmargin;
extern int topazgetch(WINDOW *win);

extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;

#define OBJ_FRAME   1
#define OBJ_AXIS    2
#define OBJ_DATA    3
#define OBJ_PARTS   4

int undermargin = 4;
double nodesize = 100.0;

extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);

std::string reducestr(std::string *str, int len)
{
  std::string out;
  char stemp[100];

  strncpy(stemp, str->c_str(), len);
  stemp[len] = 0;
  out = std::string(stemp);
  return out;
}

char *toshortfilename(char *in)
{
  char *p, *slash;
  char buff[1000];

  strcpy(buff, in);
  p = buff;
  slash = p;
  bool flag = false;

  while(*p != 0)
    {
      if (*p == '/')
	{
	  slash = p;
	  flag = true;
	}
      p++;
    }
  if (flag)
    return slash + 1;
  else
    return (slash = buff);
}

void showselectedframe(intarray *ia)
{
  intrect bbox;
  frame* fr;
  FILE* f;
  int itemp;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;

  
  for (int i = 0; i < ia->getarraylength(); i++)
    {
      if ((*ia)[i])
	{
	  fr = &(*gra->farray)[i];
	  fr->getBBbox(&bbox);
	  
	  pbegin(true, f);
	  psetpaintmode(true, f, 1);
	  psetforecolor(true, f, 0, 0, 0);
	  plinestyle(true, f, 0, 0, 0, 0);
	  psettempdashline(true, f);
	  
	  pbeginpath(true, f);
	  pmoveto(true, f, bbox.left, bbox.top);
	  plineto(true, f, bbox.right, bbox.top);
	  plineto(true, f, bbox.right, bbox.bottom);
	  plineto(true, f, bbox.left, bbox.bottom);
	  plineto(true, f, bbox.left, bbox.top);
	  pstroke(true, f);
	  
	  prestorelinestyle(true, f);
	  psetpaintmode(true, f, 0);
	  ppaint(true, f);
	  pflush(true, f);
	}
    }
}


void showselectedparts(intarray *ia)
{
  intrect bbox;
  parts* par;
  FILE* f;
  int itemp;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;

  
  for (int i = 0; i < ia->getarraylength(); i++)
    {
      if ((*ia)[i])
	{
	  par = &(*((*gra->farray)[gra->currentframe]).parray)[i];
	  par->getBBbox(&bbox);
	  
	  pbegin(true, f);
	  psetpaintmode(true, f, 1);
	  psetforecolor(true, f, 0, 0, 0);
	  plinestyle(true, f, 0, 0, 0, 0);
	  psettempdashline(true, f);
	  
	  pbeginpath(true, f);
	  pmoveto(true, f, bbox.left, bbox.top);
	  plineto(true, f, bbox.right, bbox.top);
	  plineto(true, f, bbox.right, bbox.bottom);
	  plineto(true, f, bbox.left, bbox.bottom);
	  plineto(true, f, bbox.left, bbox.top);
	  pstroke(true, f);
	  
	  prestorelinestyle(true, f);
	  psetpaintmode(true, f, 0);
	  ppaint(true, f);
	  pflush(true, f);
	}
    }
}


void showselected(intarray *ia, int objectid)
{
  switch(objectid)
    {
    case OBJ_FRAME:
      showselectedframe(ia);
      break;
    case OBJ_PARTS:
      showselectedparts(ia);
      break;
    }
}

void showcurrentframe(int no)
{
  intrect bbox;
  int m = (int)(nodesize / gra->mag);
  FILE* f;
  int itemp;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;

  
  if (0 <= no && no < gra->farray->getarraylength())
    {
      frame* par = &(*gra->farray)[no];
      par->getBBbox(&bbox);
      
      pbegin(true, f);
      psetpaintmode(true, f, 1);
      psetforecolor(true, f, 0, 0, 0);
      plinestyle(true, f, 0, 0, 0, 0);
  
      pbeginpath(true, f);
      pmoveto(true, f, bbox.left    , bbox.top);
      plineto(true, f, bbox.left - m, bbox.top);
      plineto(true, f, bbox.left - m, bbox.top - m);
      plineto(true, f, bbox.left    , bbox.top - m);
      plineto(true, f, bbox.left    , bbox.top);
      pstroke(true, f);

      pbeginpath(true, f);
      pmoveto(true, f, bbox.right    , bbox.top);
      plineto(true, f, bbox.right + m, bbox.top);
      plineto(true, f, bbox.right + m, bbox.top - m);
      plineto(true, f, bbox.right    , bbox.top - m);
      plineto(true, f, bbox.right    , bbox.top);
      pstroke(true, f);
  
      pbeginpath(true, f);
      pmoveto(true, f, bbox.right    , bbox.bottom);
      plineto(true, f, bbox.right + m, bbox.bottom);
      plineto(true, f, bbox.right + m, bbox.bottom + m);
      plineto(true, f, bbox.right    , bbox.bottom + m);
      plineto(true, f, bbox.right    , bbox.bottom);
      pstroke(true, f);

      pbeginpath(true, f);
      pmoveto(true, f, bbox.left    , bbox.bottom);
      plineto(true, f, bbox.left - m, bbox.bottom);
      plineto(true, f, bbox.left - m, bbox.bottom + m);
      plineto(true, f, bbox.left    , bbox.bottom + m);
      plineto(true, f, bbox.left    , bbox.bottom);
      pstroke(true, f);

      prestorelinestyle(true, f);
      psetpaintmode(true, f, 0);
      ppaint(true, f);
      pflush(true, f);
    }
}


void showcurrentparts(int no)
{
  intrect bbox;
  parts* par;
  int m = (int)(nodesize / gra->mag);
  FILE* f;
  int itemp;
  std::string str, str1;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  if (0 <= no && no < (*gra->farray)[gra->currentframe].parray->getarraylength())
    {
      par = &(*((*gra->farray)[gra->currentframe]).parray)[no];
      par->getBBbox(&bbox);
      
      pbegin(true, f);
      psetpaintmode(true, f, 1);
      psetforecolor(true, f, 0, 0, 0);
      plinestyle(true, f, 0, 0, 0, 0);
  
      pbeginpath(true, f);
      pmoveto(true, f, bbox.left    , bbox.top);
      plineto(true, f, bbox.left - m, bbox.top);
      plineto(true, f, bbox.left - m, bbox.top - m);
      plineto(true, f, bbox.left    , bbox.top - m);
      plineto(true, f, bbox.left    , bbox.top);
      pstroke(true, f);

      pbeginpath(true, f);
      pmoveto(true, f, bbox.right    , bbox.top);
      plineto(true, f, bbox.right + m, bbox.top);
      plineto(true, f, bbox.right + m, bbox.top - m);
      plineto(true, f, bbox.right    , bbox.top - m);
      plineto(true, f, bbox.right    , bbox.top);
      pstroke(true, f);
  
      pbeginpath(true, f);
      pmoveto(true, f, bbox.right    , bbox.bottom);
      plineto(true, f, bbox.right + m, bbox.bottom);
      plineto(true, f, bbox.right + m, bbox.bottom + m);
      plineto(true, f, bbox.right    , bbox.bottom + m);
      plineto(true, f, bbox.right    , bbox.bottom);
      pstroke(true, f);

      pbeginpath(true, f);
      pmoveto(true, f, bbox.left    , bbox.bottom);
      plineto(true, f, bbox.left - m, bbox.bottom);
      plineto(true, f, bbox.left - m, bbox.bottom + m);
      plineto(true, f, bbox.left    , bbox.bottom + m);
      plineto(true, f, bbox.left    , bbox.bottom);
      pstroke(true, f);

      psetpaintmode(true, f, 0);
      ppaint(true, f);
      pflush(true, f);
    }
}


void showcurrentdata(int no)
{
  intrect bbox;
  frame *fr;
  axis *ax;
  data* dat;
  int m = (int)(nodesize / gra->mag * .5);
  FILE* f;
  int itemp;
  std::string str, str1;
  int inc;
  std::string temp_pre_formula;
  double x, y;
  int ix, iy;

  str1 = std::string("V0");
  if (!fhvuff->get(&str1, &f, &itemp, &str))
    return;
  
  fr = &(*gra->farray)[gra->currentframe];
  if (0 <= no && no < fr->darray->getarraylength())
    {
      dat = &(*fr->darray)[no];

      if (dat->x->getarraylength() > 25)
	inc = dat->x->getarraylength() / 25;
      else
	inc = 1;

      pbegin(true, f);
      psetpaintmode(true, f, 1);
      psetforecolor(true, f, 0, 0, 0);
      plinestyle(true, f, 0, 0, 0, 0);

      for (int i = 0; i < dat->x->getarraylength();i += inc)
	{
	  if (dat->trans_x)
	    temp_pre_formula = *dat->format_x;
	  else
	    temp_pre_formula = std::string("x");
	      
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &x,
			    true ,dat, 0, fr->getdataindex(no), 0);
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &x,
			    false ,dat, i, fr->getdataindex(no), 0);

	  if (dat->trans_y)
	    temp_pre_formula = *dat->format_y;
	  else
	    temp_pre_formula = std::string("y");
	      
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &y,
			    true ,dat, 0, fr->getdataindex(no), 0);
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &y,
			    false ,dat, i, fr->getdataindex(no), 0);

	  ax = &(*fr->axarray)[dat->refaxis_x];
	  x = ax->virtual_to_doubleworld(x).x;
	  x = fr->doubleworld_to_view_x(x);

	  ax = &(*fr->axarray)[dat->refaxis_y];
	  y = ax->virtual_to_doubleworld(y).y;
	  y = fr->doubleworld_to_view_y(y);

	  ix = (int)x; iy = (int)y;

	  if (-MAXSHORT + 1 < ix && ix < MAXSHORT - 1 &&
	      -MAXSHORT + 1 < iy && iy < MAXSHORT - 1)
	    {
	      pbeginpath(true, f);
	      pmoveto(true, f, ix - m, iy - m);
	      plineto(true, f, ix + m, iy - m);
	      plineto(true, f, ix + m, iy + m);
	      plineto(true, f, ix - m, iy + m);
	      plineto(true, f, ix - m, iy - m);
	      pstroke(true, f);
	    }
	} 
      
      psetpaintmode(true, f, 0);
      ppaint(true, f);
      pflush(true, f);

    }
}

void flistdisp(WINDOW *win, int startline, intarray *ia, int mode)
{
  char cur[4];
  std::string type, withdata;

  wclear(win);
  //  wrefresh(win);

  wmove(win, topmargin, leftmargin);
  wprintw(win, "S aid iid C D frametype    position                                         \n");
  wmove(win, topmargin + 1, leftmargin);
  wprintw(win, "============================================================================\n");
  wmove(win, menuwinlines - 1, leftmargin);
  if (mode)
    wprintw(win, "[a]ll [c]lear [r]everse\n");

  int framelength = gra->farray->getarraylength();
  for (int i = 0; i < framelength; i++)
    {
      if (startline <= i && i < startline + menuwinlines - topmargin - undermargin)
	{
	  if (i == gra->currentframe)
	    strcpy(cur, "C");
	  else
	    strcpy(cur, " ");
	  if ((*gra->farray)[i].withdata)
	    withdata = std::string("D");
	  else
	    withdata = std::string(" ");

	  getframetypename((*gra->farray)[i].frametype, &type);

	  wmove(win, i - startline + topmargin + 2, leftmargin);
	  if ((*ia)[i])
	    wprintw(win, "* %03d %03d %-1s %-1s %-11s (%5d, %5d) - (%5d, %5d)                   \n",
		    i,
		    (*gra->farray)[i].id,
		    cur,
		    withdata.c_str(),
		    type.c_str(),
		    (*gra->farray)[i].p1.x, (*gra->farray)[i].p1.y,
		    (*gra->farray)[i].p2.x, (*gra->farray)[i].p2.y
		    ); 
	  else
	    wprintw(win, "  %03d %03d %-1s %-1s %-11s (%5d, %5d) - (%5d, %5d)                   \n",
		    i,
		    (*gra->farray)[i].id,
		    cur,
		    withdata.c_str(),
		    type.c_str(),
		    (*gra->farray)[i].p1.x, (*gra->farray)[i].p1.y,
		    (*gra->farray)[i].p2.x, (*gra->farray)[i].p2.y
		    ); 

	}
    }
  
  wmove(win, menuwinlines - 1, leftmargin);
  if (mode)
    wprintw(win, "[a]ll [c]lear [r]everse\n");
  wrefresh(win);
}

void flistcursor(WINDOW *win, int startline, intarray *ia, int no, bool flag)
{
  char cur[4];
  std::string type, withdata;

  if (flag)
    wattron(win, A_REVERSE);

      if (startline <= no && no < startline + menuwinlines - topmargin - undermargin)
	{
	  if (no == gra->currentframe)
	    strcpy(cur, "C");
	  else
	    strcpy(cur, " ");
	  if ((*gra->farray)[no].withdata)
	    withdata = std::string("D");
	  else
	    withdata = std::string(" ");

	  getframetypename((*gra->farray)[no].frametype, &type);

	  wmove(win, no - startline + topmargin + 2, leftmargin);
	  if ((*ia)[no])
	    wprintw(win, "* %03d %03d %-1s %-1s %-11s (%5d, %5d) - (%5d, %5d)                   \n",
		    no,
		    (*gra->farray)[no].id,
		    cur,
		    withdata.c_str(),
		    type.c_str(),
		    (*gra->farray)[no].p1.x, (*gra->farray)[no].p1.y,
		    (*gra->farray)[no].p2.x, (*gra->farray)[no].p2.y
		    ); 
	  else
	    wprintw(win, "  %03d %03d %-1s %-1s %-11s (%5d, %5d) - (%5d, %5d)                   \n",
		    no,
		    (*gra->farray)[no].id,
		    cur,
		    withdata.c_str(),
		    type.c_str(),
		    (*gra->farray)[no].p1.x, (*gra->farray)[no].p1.y,
		    (*gra->farray)[no].p2.x, (*gra->farray)[no].p2.y
		    ); 
	}
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);
  if (flag)
    wattroff(win, A_REVERSE);
  wrefresh(win);


}

void showcurrent(int no, int objectid)
{
  switch(objectid)
    {
    case OBJ_FRAME:
      showcurrentframe(no);
      break;
    case OBJ_DATA:
      showcurrentdata(no);
      break;
    case OBJ_PARTS:
      showcurrentparts(no);
      break;
    }
}


void partslistdisp(WINDOW *win, int startline, intarray *ia, int mode)
{
  int length;
  std::string type, font, linecolor, bodycolor;
  parts *par;
  char stemp[17];

  wclear(win);
  //  wrefresh(win);
  
  length = (*gra->farray)[gra->currentframe].parray->getarraylength();

  wmove(win, topmargin, leftmargin);
  wprintw(win, "S aid iid type      text             font                  L-color  B-color \n");
  wmove(win, topmargin + 1, leftmargin);
  wprintw(win, "============================================================================\n");
  wmove(win, menuwinlines - 1, leftmargin);
  if (mode)
    wprintw(win, "[a]ll [c]lear [r]everse\n");

  for (int i = 0; i < length; i++)
    {
      if (startline <= i && i < startline + menuwinlines - topmargin - undermargin)
	{
	  par = &(*((*gra->farray)[gra->currentframe]).parray)[i];

	  getpartsname(par->type, &type);
	  type = reducestr(&type, 9);
	  gettextfacename(par->fo.face, &font);
	  font = reducestr(&font, 21);
	  getcolorname(par->linecolor.red,
		       par->linecolor.green, 
		       par->linecolor.blue,
		       &linecolor);
	  linecolor = reducestr(&linecolor, 8);
	  getcolorname(par->bodycolor.red,
		       par->bodycolor.green, 
		       par->bodycolor.blue,
		       &bodycolor);
	  bodycolor = reducestr(&bodycolor, 8);
	  strncpy(stemp, par->text->c_str(), sizeof(stemp));
	  stemp[sizeof(stemp) - 1] = 0;

	  wmove(win, i - startline + topmargin + 2, leftmargin);
	  if ((*ia)[i])
	    wprintw(win, "* %03d %03d %-9s %-16s %-21s %-8s %-8s\n",
		    i,
		    par->id,
		    type.c_str(),
		    stemp,
		    font.c_str(),
		    linecolor.c_str(),
		    bodycolor.c_str()
		    ); 
	  else
	    wprintw(win, "  %03d %03d %-9s %-16s %-21s %-8s %-8s\n",
		    i,
		    par->id,
		    type.c_str(),
		    stemp,
		    font.c_str(),
		    linecolor.c_str(),
		    bodycolor.c_str()
		    ); 
	}
    }
  
  wrefresh(win);
}

void partslistcursor(WINDOW *win, int startline, intarray *ia, int no, bool flag)
{
  std::string type, font, linecolor, bodycolor;
  parts *par;
  char stemp[17];

  if (flag)
    wattron(win, A_REVERSE);

      if (startline <= no && no < startline + menuwinlines - topmargin - undermargin)
	{
	  par = &(*((*gra->farray)[gra->currentframe]).parray)[no];

	  getpartsname(par->type, &type);
	  type = reducestr(&type, 9);
	  gettextfacename(par->fo.face, &font);
	  font = reducestr(&font, 21);
	  getcolorname(par->linecolor.red,
		       par->linecolor.green, 
		       par->linecolor.blue,
		       &linecolor);
	  linecolor = reducestr(&linecolor, 8);
	  getcolorname(par->bodycolor.red,
		       par->bodycolor.green, 
		       par->bodycolor.blue,
		       &bodycolor);
	  bodycolor = reducestr(&bodycolor, 8);
	  strncpy(stemp, par->text->c_str(), sizeof(stemp));
	  stemp[sizeof(stemp) - 1] = 0;

	  wmove(win, no - startline + topmargin + 2, leftmargin);
	  if ((*ia)[no])
	    wprintw(win, "* %03d %03d %-9s %-16s %-21s %-8s %-8s\n",
		    no,
		    par->id,
		    type.c_str(),
		    stemp,
		    font.c_str(),
		    linecolor.c_str(),
		    bodycolor.c_str()
		    ); 
	  else
	    wprintw(win, "  %03d %03d %-9s %-16s %-21s %-8s %-8s\n",
		    no,
		    par->id,
		    type.c_str(),
		    stemp,
		    font.c_str(),
		    linecolor.c_str(),
		    bodycolor.c_str()
		    ); 

	}
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);
  if (flag)
    wattroff(win, A_REVERSE);
  wrefresh(win);
}


void datalistdisp(WINDOW *win, int startline, intarray *ia, int mode)
{
  int length;
  std::string isfile, linestyle, linecolor, markerstyle, markerbodycolor, display, fname;
  data *d;

  wclear(win);
  //  wrefresh(win);
  
  length = (*gra->farray)[gra->currentframe].darray->getarraylength();

  wmove(win, topmargin, leftmargin);
  wprintw(win, "S aid iid T rx ry points D filename            line    color  marker color  \n");
  wmove(win, topmargin + 1, leftmargin);
  wprintw(win, "============================================================================\n");
  wmove(win, menuwinlines - 1, leftmargin);
  if (mode)
    wprintw(win, "[a]ll [c]lear [r]everse [g]rep\n");
  else
    wprintw(win, "[g]rep\n");
	    
  for (int i = 0; i < length; i++)
    {
      if (startline <= i && i < startline + menuwinlines - topmargin - undermargin)
	{
	  d = &(*((*gra->farray)[gra->currentframe]).darray)[i];

	  getlinestylename(d->linestyle.style, &linestyle);
	  linestyle = reducestr(&linestyle, 7);
	  getcolorname(d->linecolor.red,
		       d->linecolor.green, 
		       d->linecolor.blue,
		       &linecolor);
	  linecolor = reducestr(&linecolor, 6);
	  getmarkerstylename(d->markerstyle, &markerstyle);
	  markerstyle = reducestr(&markerstyle, 6);
	  getcolorname(d->markerbodycolor.red,
		       d->markerbodycolor.green, 
		       d->markerbodycolor.blue,
		       &markerbodycolor);
	  markerbodycolor = reducestr(&markerbodycolor, 6);
	  if (d->isfiledata)
	    isfile = std::string("F");
	  else
	    isfile = std::string("M");
	  if (d->display)
	    display = std::string("S");
	  else
	    display = std::string(" ");
	  fname = std::string(toshortfilename(d->filename));
	  fname = reducestr(&fname, 19);

	  wmove(win, i - startline + topmargin + 2, leftmargin);
	  if ((*ia)[i])
	    wprintw(win, "* %03d %03d %1s %02d %02d %06d %1s %-19s %-7s %-6s %-6s %-6s\n",
		    i,
		    d->id,
		    isfile.c_str(),
		    d->refaxis_x,
		    d->refaxis_y,
		    d->x->getarraylength(),
		    display.c_str(),
		    fname.c_str(),
		    linestyle.c_str(),
		    linecolor.c_str(),
		    markerstyle.c_str(),
		    markerbodycolor.c_str()
		    ); 
	  else
	    wprintw(win, "  %03d %03d %1s %02d %02d %06d %1s %-19s %-7s %-6s %-6s %-6s\n",
		    i,
		    d->id,
		    isfile.c_str(),
		    d->refaxis_x,
		    d->refaxis_y,
		    d->x->getarraylength(),
		    display.c_str(),
		    fname.c_str(),
		    linestyle.c_str(),
		    linecolor.c_str(),
		    markerstyle.c_str(),
		    markerbodycolor.c_str()
		    ); 
	}
    }
  
  wrefresh(win);
}

void datalistcursor(WINDOW *win, int startline, intarray *ia, int no, bool flag)
{
  std::string isfile, linestyle, linecolor, markerstyle, markerbodycolor, display, fname;
  data *d;

  if (flag)
    wattron(win, A_REVERSE);

      if (startline <= no && no < startline + menuwinlines - topmargin - undermargin)
	{
	  d = &(*((*gra->farray)[gra->currentframe]).darray)[no];

	  getlinestylename(d->linestyle.style, &linestyle);
	  linestyle = reducestr(&linestyle, 7);
	  getcolorname(d->linecolor.red,
		       d->linecolor.green, 
		       d->linecolor.blue,
		       &linecolor);
	  linecolor = reducestr(&linecolor, 6);
	  getmarkerstylename(d->markerstyle, &markerstyle);
	  markerstyle = reducestr(&markerstyle, 6);
	  getcolorname(d->markerbodycolor.red,
		       d->markerbodycolor.green, 
		       d->markerbodycolor.blue,
		       &markerbodycolor);
	  markerbodycolor = reducestr(&markerbodycolor, 6);
	  if (d->isfiledata)
	    isfile = std::string("F");
	  else
	    isfile = std::string("M");
	  if (d->display)
	    display = std::string("S");
	  else
	    display = std::string(" ");
	  fname = std::string(toshortfilename(d->filename));
	  fname = reducestr(&fname, 19);

	  wmove(win, no - startline + topmargin + 2, leftmargin);
	  if ((*ia)[no])
	    wprintw(win, "* %03d %03d %1s %02d %02d %06d %1s %-19s %-7s %-6s %-6s %-6s\n",
		    no,
		    d->id,
		    isfile.c_str(),
		    d->refaxis_x,
		    d->refaxis_y,
		    d->x->getarraylength(),
		    display.c_str(),
		    fname.c_str(),
		    linestyle.c_str(),
		    linecolor.c_str(),
		    markerstyle.c_str(),
		    markerbodycolor.c_str()
		    ); 
	  else
	    wprintw(win, "  %03d %03d %1s %02d %02d %06d %1s %-19s %-7s %-6s %-6s %-6s\n",
		    no,
		    d->id,
		    isfile.c_str(),
		    d->refaxis_x,
		    d->refaxis_y,
		    d->x->getarraylength(),
		    display.c_str(),
		    fname.c_str(),
		    linestyle.c_str(),
		    linecolor.c_str(),
		    markerstyle.c_str(),
		    markerbodycolor.c_str()
		    ); 
	  if (flag)
	    wattroff(win, A_REVERSE);
	  wmove(win, menuwinlines - 2, leftmargin);
	  wprintw(win, "%-75s\n", d->filename);
	  if (flag)
	    wattron(win, A_REVERSE);

	}
  
  wmove(win, 0, 0);
  wprintw(win, " ");
  wmove(win, 0, 0);
  if (flag)
    wattroff(win, A_REVERSE);
  wrefresh(win);
}


void objectlistdisp(WINDOW *win, int startline, intarray *ia, int objectid, int mode)
{
  switch(objectid)
    {
    case OBJ_FRAME:
      flistdisp(win, startline, ia, mode);
      break;
    case OBJ_DATA:
      datalistdisp(win, startline, ia, mode);
      break;
    case OBJ_PARTS:
      partslistdisp(win, startline, ia, mode);
      break;
    }
}

void objectlistcursor(WINDOW *win, int startline, intarray *ia, int no, bool flag, int objectid)
{
  switch(objectid)
    {
    case OBJ_FRAME:
      flistcursor(win, startline, ia, no, flag);
      break;
    case OBJ_DATA:
      datalistcursor(win, startline, ia, no, flag);
      break;
    case OBJ_PARTS:
      partslistcursor(win, startline, ia, no, flag);
      break;
    }
}

bool loadlist(intarray *iarray, int objectid, int* length)
{
  iarray->setbuff(0);
  switch(objectid)
    {
    case OBJ_FRAME:
      *length = gra->farray->getarraylength();
      if (*length == 0)
	return false;
      for (int i = 0; i < *length; i++)
	iarray->add((*gra->farray)[i].selected);
      break;
    case OBJ_AXIS:
      *length = (*gra->farray)[gra->currentframe].axarray->getarraylength();
      if (*length == 0)
	return false;
      for (int i = 0; i < *length; i++)
	iarray->add((*((*gra->farray)[gra->currentframe]).axarray)[i].selected);
      break;
    case OBJ_DATA:
      *length = (*gra->farray)[gra->currentframe].darray->getarraylength();
      if (*length == 0)
	return false;
      for (int i = 0; i < *length; i++)
	iarray->add((*((*gra->farray)[gra->currentframe]).darray)[i].selected);
      break;
    case OBJ_PARTS:
      *length = (*gra->farray)[gra->currentframe].parray->getarraylength();
      if (*length == 0)
	return false;
      for (int i = 0; i < *length; i++)
	iarray->add((*((*gra->farray)[gra->currentframe]).parray)[i].selected);
      break;
    default:
      return false;
    }
  return true;
}

int objectlist(WINDOW *win, int mode, std::string *objtype)
{
  // mode = 0  single select
  // mode = 1  multi  select
  //
  // objtype = data
  // object =  parts

  int key;
  intarray *iarray;
  int startline = 0;
  int relativeline = 0;
  int length;
  int objectid;
  static std::string reg = std::string("");
  std::string str1, str2;
  data *d;

  static int f_startline = 0, a_startline = 0, d_startline = 0, p_startline = 0;
  static int f_relativeline = 0, a_relativeline = 0, d_relativeline = 0, p_relativeline = 0;
  static int f_length = 0, a_length = 0, d_length = 0, p_length = 0;


  if (win == 0)
    return false;

  if(mode != 0 && mode != 1)
    return false;

  if (strcmp(objtype->c_str(), "frame") == 0)
    objectid = OBJ_FRAME;
  else if (strcmp(objtype->c_str(), "axis") == 0)
    objectid = OBJ_AXIS;
  else if (strcmp(objtype->c_str(), "data") == 0)
    objectid = OBJ_DATA;
  else if (strcmp(objtype->c_str(), "parts") == 0)
    objectid = OBJ_PARTS;
  else
    return false;

  iarray = new intarray;
  if (!loadlist(iarray, objectid, &length))
    {
      delete iarray;
      return false;
    }

  switch(objectid)
    {
    case OBJ_FRAME:
      if (length == f_length)
	{
	  startline = f_startline; relativeline = f_relativeline;
	}
      break;
    case OBJ_AXIS:
      if (length == a_length)
	{
	  startline = a_startline; relativeline = a_relativeline;
	}
      break;
    case OBJ_DATA:
      if (length == d_length)
	{
	  startline = d_startline; relativeline = d_relativeline;
	}
      break;
    case OBJ_PARTS:
      if (length == p_length)
	{
	  startline = p_startline; relativeline = p_relativeline;
	}
      break;
    default:
      break;
    }
  
  objectlistdisp(win, startline, iarray, objectid, mode);

  while (1)
    {
      
      objectlistcursor(win, startline, iarray, startline + relativeline, true, objectid);
      showcurrent(startline + relativeline, objectid);
      showselected(iarray, objectid);
      key = topazgetch(win);
      objectlistcursor(win, startline, iarray, startline + relativeline, false, objectid);
      showcurrent(startline + relativeline, objectid);
      showselected(iarray, objectid);

      switch (key)
	{
	case ' ':
	  if (mode == 0)
	    {
	      for (int i = 0; i < length; i++)
		{
		  if ((*iarray)[i] == true)
		  {
		    (*iarray)[i] = false;
		    objectlistcursor(win, startline, iarray, i, false, objectid);
		  }
		}
	    }

	  if ((*iarray)[startline + relativeline] == 1)
	    (*iarray)[startline + relativeline] = 0;
	  else
	    (*iarray)[startline + relativeline] = 1;

	  objectlistcursor(win, startline, iarray, startline + relativeline, false, objectid);
	case KEY_DOWN:
	case CTRL('n'):
	  if ((startline + relativeline + 1) < length)
	    {
	      if (relativeline + 1 < menuwinlines - topmargin - undermargin)
		  relativeline++;
	      else
		{
		  startline += menuwinlines - topmargin - undermargin;
		  relativeline = 0;
		  objectlistdisp(win, startline, iarray, objectid, mode);
		}
	    }
	  break;
	case KEY_NPAGE:
	case CTRL('v'):
	  if ((startline + relativeline + menuwinlines - topmargin - undermargin) < length)
	    {
	      startline += menuwinlines - topmargin - undermargin;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  else if ((startline + menuwinlines - topmargin - undermargin) < length)
	    {
	      startline += menuwinlines - topmargin - undermargin;
	      relativeline = 0;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  else
	    {
	      relativeline = length - startline - 1;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	case KEY_UP:
	case CTRL('p'):
	  if ((startline + relativeline - 1) >= 0)
	    {
	      if (relativeline - 1 >= 0)
		  relativeline--;
	      else
		{
		  startline -= menuwinlines - topmargin - undermargin;
		  relativeline = menuwinlines - topmargin - 1 - undermargin;
		  objectlistdisp(win, startline, iarray, objectid, mode);
		}
	    }
	  break;
	case KEY_PPAGE:
	  if ((startline  - (menuwinlines - topmargin - undermargin)) >= 0)
	    {
	      startline -= menuwinlines - topmargin - undermargin;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  else
	    {
	      startline = 0; relativeline = 0;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	case CTRL('a'):
	  startline = 0; relativeline = 0;
	  objectlistdisp(win, startline, iarray, objectid, mode);
	  break;
	case CTRL('e'):
	  startline = (length - 1) / (menuwinlines - topmargin - undermargin);
	  startline *= menuwinlines - topmargin - undermargin;
	  relativeline = length - startline - 1;
	  objectlistdisp(win, startline, iarray, objectid, mode);
	  break;
	case CTRL('g'):
	  key = LOCAL_ESCAPE_KEY;
	  break;
	case 'a':
	  if (mode == 1)
	    {
	      for (int i = 0; i < length; i++)
		(*iarray)[i] = true;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	case 'c':
	  if (mode == 1)
	    {
	      for (int i = 0; i < length; i++)
		(*iarray)[i] = false;
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	case 'r':
	  if (mode == 1)
	    {
	      for (int i = 0; i < length; i++)
		{
		  if ((*iarray)[i])
		    (*iarray)[i] = false;
		  else
		    (*iarray)[i] = true;
		}
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	case 'g':
	  if (objectid == OBJ_DATA)
	    {
	      str1 = reg;
	      str2 = std::string("Input regular expression:");
	      if (edittextwindow(menuwin, &str1, &str2, false))
		{
		  reg = str1;
		  for (int i = 0; i < length; i++)
		    {
		      d = &(*((*gra->farray)[gra->currentframe]).darray)[i];
		      str1 =  std::string(toshortfilename(d->filename));
		      if (simplematch(str1.c_str(), reg.c_str()))
			(*iarray)[i] = true;
		      else
			(*iarray)[i] = false;
		    }
		}
	      objectlistdisp(win, startline, iarray, objectid, mode);
	    }
	  break;
	}

      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }


  if (key == LOCAL_ESCAPE_KEY)
    {
      delete iarray;
      return false;
    }
  else if (key == LOCAL_LF_KEY)
    {
      switch(objectid)
	{
	case OBJ_FRAME:
	  for (int i = 0; i < length; i++)
	    (*gra->farray)[i].selected = (*iarray)[i];
	  break;
	case OBJ_AXIS:
	  for (int i = 0; i < length; i++)
	    (*((*gra->farray)[gra->currentframe]).axarray)[i].selected = (*iarray)[i];
	  break;
	case OBJ_DATA:
	  for (int i = 0; i < length; i++)
	    (*((*gra->farray)[gra->currentframe]).darray)[i].selected = (*iarray)[i];
	  break;
	case OBJ_PARTS:
	  for (int i = 0; i < length; i++)
	    (*((*gra->farray)[gra->currentframe]).parray)[i].selected = (*iarray)[i];
	  break;
	default:
	  break;
	}

      switch(objectid)
	{
	case OBJ_FRAME:
	  f_startline = startline; f_relativeline = relativeline; f_length = length;
	  break;
	case OBJ_AXIS:
	  a_startline = startline; a_relativeline = relativeline; a_length = length;
	  break;
	case OBJ_DATA:
	  d_startline = startline; d_relativeline = relativeline; d_length = length;
	  break;
	case OBJ_PARTS:
	  p_startline = startline; p_relativeline = relativeline; p_length = length;
	  break;
	default:
	  break;
	}
    }
  
  delete iarray;
  return true;

}



