
/*----------------------------------------------------------------------------
                          parsetoken.cc (parse script)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string>
#include <ctype.h>
#include <string.h>
#include "script.h"
#include "system.h"
#include "tokenbuff.h"
#include "topazvalues.h"

extern void message(const char *msg);

char spchrs[] = "\"&'`,/|";

bool script::addbuff(int priority, int status, const char *value, const char *label)
{
  std::string str1, str2;
  str1 = std::string(value); str2 = std::string(label);
  return buff->push(priority, status, &str1, &str2,
		    strstatusbuff[strstatusbuffpoint],
		    bracestatusbuff[bracestatusbuffpoint],
		    bracetype,
		    //	     bracestatusbuffpoint,
		    parentheseslevel,	     
		    -1,
		    lineno,
		    -1
		    );
}


char * nextchar(char *in)
{
  char *tempin;
  tempin = in;
  while(*tempin != 0)
    {
      if (*(unsigned char*)tempin != ' '
	  && *(unsigned char*)tempin != '\t')
	  return tempin;
      else
	  tempin ++;
    }
  return tempin;
}

typedef struct funcnamemap
{
  char *label;
}FNMAP;

FNMAP function_pri_map[] = {
  {"abs"},
  {"sin"},
  {"cos"},
  {"tan"},
  {"asin"},
  {"acos"},
  {"atan"},
  {"sinh"},
  {"cosh"},
  {"tanh"},
  {"asinh"},
  {"acosh"},
  {"atanh"},
  {"exp"},
  {"ln"},
  {"log"},
  {"int"},
  {"gauss"},
  {"sqrt"},
  {"rand"},
  {"frac"},
  {"theta"},
  {"delta"},
  {"sign"},
  {"round"},
  {"gamma"},
  {"erf"},
  {"erfc"},
  {"invp"},
  {"invq"},
  {"p_normal"},
  {"q_normal"},
  {"inverf"},
  {"inverfc"},
  {"strlength"},
  {"flushmsg"},
  {"system"},
  {"exit"},
  {"localtime"},
  {"localtimename"},
  {"toplaintext"},
  {"fromplaintext"},
  {"toplainfilename"},
  {"close"},
  {"listhandle"},
  {"isfileexist"},
  {"chdir"},
  {"sleep"},
  {"toarray"},
  {"eval"},
  {"yesnoname"},
  {"selectfile"},
  {"listhandle"},
  {"includefile"},
  {"defined"},
  {"appendabspath"},
  {"removepath"},
  {"unlink"},
  {"removectrcode"},
  {"getmacroname"},
  {"readline"},
  {0}
}; 

FNMAP print_pri_map[] = {
  {"beta"},
  {"min"},
  {"max"},
  {"substr"},
  {"open"},
  {"chomp"},
  {"chop"},
  {"print"},
  {"printf"},
  {"sprintf"},
  {"toscalar"},
  {"edittext"},
  {"editfilename"},
  {"selectcoloritem"},
  {"selectyesno"},
  {"selectitem"},
  {"selectlistitem"},
  {"selectframetype"},
  {"inputparts"},
  {"changenodes"},
  {"objectstretch"},
  {"selecttickdirection"},
  {"selectscaling"},
  {"selectmscaling"},
  {"selectlinecap"},
  {"selectlinejoin"},
  {"selectlinestyle"},
  {"selectfillstyle"},
  {"selectfillrule"},
  {"execscr"},
  {"execfile"},
  {"selectmag"},
  {"selectinterpolate"},
  {"selectmarkerstyle"},
  {"selectcommentlineaction"},
  {"selectdatasearchpath"},
  {"selectpartstype"},
  {"selectfontface"},
  {"selectjpfontface"},
  {"selectpartsinterpolation"},
  {"selectlabelparity"},
  {"selectlabelcpparity"},
  {"selectlabelformat"},
  {"selectorder"},
  {"selectpapersize"},
  {"selectlinearfitfunction"},
  {"selectobject"},
  {0}
}; 

FNMAP parenfunc_pri_map[] = {
  {"getcwd"},
  {"uname"},
  {"objectmove"},
  {"trimming"},
  {"watch"},
  {"trace"},
  {"message"},
  {"cuimenu"},
  {"framemenu"},
  {"xbasemenu"},
  {"xrangemenu"},
  {"xtickmenu"},
  {"xsubtickmenu"},
  {"xlabelmenu"},
  {"ybasemenu"},
  {"yrangemenu"},
  {"ytickmenu"},
  {"ysubtickmenu"},
  {"ylabelmenu"},
  {"tbasemenu"},
  {"trangemenu"},
  {"ttickmenu"},
  {"tsubtickmenu"},
  {"tlabelmenu"},
  {"rbasemenu"},
  {"rrangemenu"},
  {"rtickmenu"},
  {"rsubtickmenu"},
  {"rlabelmenu"},
  {"plotmappingmenu"},
  {"plotlinemenu"},
  {"plotmarkermenu"},
  {"plotrangemenu"},
  {"functionmenu"},
  {"nonlinearfitmenu"},
  {"linearfitmenu"},
  {"loaddatamenu"},
  {"partsstylemenu"},
  {"textstylemenu"},
  {"updateseparator"},
  {"escapecuimenu"},
  {"restorecuimenu"},
  {0}
}; 

bool is_pri_function(char *s)
{
  FNMAP *fm;
  int itemno;

  fm = function_pri_map;
  for (; fm->label; fm++)
    {
    }
  itemno = fm - function_pri_map;
  for (int i = 0; i < itemno; i++)
    {
      if (strcmp(s, function_pri_map[i].label) == 0)
	return true;
    }
  return false;
}

bool is_pri_print(char *s)
{
  FNMAP *fm;
  int itemno;

  fm = print_pri_map;
  for (; fm->label; fm++)
    {
    }
  itemno = fm - print_pri_map;
  for (int i = 0; i < itemno; i++)
    {
      if (strcmp(s, print_pri_map[i].label) == 0)
	return true;
    }
  return false;
}

bool is_pri_parenfunction(char *s)
{
  FNMAP *fm;
  int itemno;

  fm = parenfunc_pri_map;
  for (; fm->label; fm++)
    {
    }
  itemno = fm - parenfunc_pri_map;
  for (int i = 0; i < itemno; i++)
    {
      if (strcmp(s, parenfunc_pri_map[i].label) == 0)
	return true;
    }
  return false;
}


bool script::parsetoken(std::string *inputstring)
/* $BF~NO%9%H%j!<%`$+$i<!$N%H!<%/%s$rFI$_9~$`(B */
{
  char *input;
  char *start, numstring[MAXINPUT+1];
  int decimal, len, numlen;
  char temp[20];
  int alphalen;
  std::string temp_alpha, temp_alpha_1;
  bool Flag;
  char tempchar;
  int i;
  int plusstatus = 0;
  char quoteseparator = '"';
  int funcpriority, printpriority;

  input = (char *)inputstring->c_str();

  
  //repeat until pointer reaches end of string
  while(*input != 0)
    {
      switch (strstatusbuff[strstatusbuffpoint])
	{
	case STR_NORMAL:

	  //skip space
	  if ((unsigned char)*input == ' ')
	    {
	      Flag = false;
	      while (*input != 0)
		{
		  if (*input != ' ')
		    Flag = true;

		  //get out of while loop
		  if (Flag == true)
		    break;

		  input++;
		} 
	    }
	  // skip tab
	  else if ((unsigned char)*input == '\t')
	    {
	      input++;
	    }
	  // skip \n
	  else if ((unsigned char)*input == '\n')
	    {
	      input++;
	    }
	  // digit
	  else if (strchr("0123456789.", *input))
	    {
	      start = input;
	      len = 0;
	      decimal = false;
	      //calc length of numerical string
	      while ((isdigit((unsigned char)*input)) ||
		     ((*input == '.') && (!decimal)))
		{
		  if (*input == '.')
		    decimal = true;
		  input++;
		  len++;
		}
	      
	      //only comma -> addition operation for string
	      if ((len == 1) && (start[0] == '.'))
		{
		  if (*(start + 1) == '=')
		    {
		      if (addbuff(PRI_SET, STRPLUSEQUAL\
				  ,"", "") == false)
			return false;
		      input++;
		      len = 0;
		    }
		  else
		    {
		      if (addbuff(PRI_PRUSMINUS, STR_PLUS\
				  ,"", "") == false)
			return false;
		    }
		}
	      else
		{      
		  //exponential expression
		  if (*input == 'E' || *input == 'e')
		    {
		      input++;
		      len++;
		      //if there exist "+=" following "E"$B!"(Bpointer is added by 1
		      if (strchr("+-", *input) != NULL)
			{
			  input++;
			  len++;
			}
		      numlen = 0;
		      //pointer is set at the end of the exponential description
		      while ((isdigit((unsigned char)*input)) && (++numlen <= 3))
			{
			  input++;
			  len++;
			}
		    }
		  
		  strncpy(numstring, start, len);
		  numstring[len] = 0;
		  
		  //add to tokenbuff
		  if (addbuff(PRI_OPERAND, NUM,\
			      numstring, "") == false)
		    return false;
		}      
	    }
	  
	  // '"' or '\'' or '`'
	    else if ((unsigned char)*input == '"')
	      {
		if (strstatusbuffpoint < STRSTATUS_BUFF_LEN - 1)
		  {
		    parentheseslevel++;
		    strstatusbuffpoint++;
		    strstatusbuff[strstatusbuffpoint] = STR_DQUOTE;

		    if (addbuff(PRI_OPAREN, OPAREN\
				, "", "") == false)
		      return false;

		    if (addbuff(PRI_OPERAND, NUM,\
				"", "") == false)
		      return false;

		    input++;
		  }
		else
		  return false;
	      }
	    else if ((unsigned char)*input == '\'')
	      {
		if (strstatusbuffpoint < STRSTATUS_BUFF_LEN - 1)
		  {
		    parentheseslevel++;
		    strstatusbuffpoint++;
		    strstatusbuff[strstatusbuffpoint] = STR_SQUOTE;

		    if (addbuff(PRI_OPAREN, OPAREN\
				, "", "") == false)
		      return false;

		    if (addbuff(PRI_OPERAND, NUM,\
				"", "") == false)
		      return false;

		    input++;
		  }
		else
		  return false;
	      }
	    else if ((unsigned char)*input == '`')
	      {
		if (strstatusbuffpoint < STRSTATUS_BUFF_LEN - 1)
		  {
		    parentheseslevel++;
		    strstatusbuffpoint++;
		    strstatusbuff[strstatusbuffpoint] = STR_UQUOTE;

		    if (addbuff(PRI_OPAREN, OPAREN\
				, "", "") == false)
		      return false;

		    if (addbuff(PRI_OPERAND, NUM,\
				"", "") == false)
		      return false;

		    input++;
		  }
		else
		  return false;
	      }
	    else if ((unsigned char)*input == '/')
	      {
		if (buff->getarraylength() > 0)
		  {
		    if (strstatusbuffpoint < STRSTATUS_BUFF_LEN - 1
			&& ((*buff)[buff->getarraylength() - 1].status == MATCH || (*buff)[buff->getarraylength() - 1].status == NOTMATCH))
		      {
			parentheseslevel++;
			strstatusbuffpoint++;
			strstatusbuff[strstatusbuffpoint] = STR_REGEX;
			
			if (addbuff(PRI_OPAREN, OPAREN\
				    , "", "") == false)
			  return false;
			
			if (addbuff(PRI_OPERAND, NUM,\
				    "", "") == false)
			  return false;
			
			input++;
		      }
		    else if (*(input + 1) == '=')     //  /=
		      {
			if (addbuff(PRI_SET, DIVIDEEQUAL\
				    , "", "") == false)
			  return false;
			input += 2;
		      }
		    else                   // a / b
		      {
			if (addbuff(PRI_MULDIV, DIVIDE\
				    , "", "") == false)
			  return false;
			input ++;
		      }
		  }
		else
		  return errormessage(lineno, "parse error", "/");
	      }

	  //  alphabetic character
	  else if (isalpha((unsigned char)*input))
	    {
	      Flag = false;
	      alphalen = -1;
	      tempchar = (unsigned char)*input;
	      i = 0;
	      temp_alpha = std::string("");
	      while (tempchar != 0)
		{
		  if (isalpha(tempchar) == 0 \
		      && isdigit(tempchar) == 0 \
		      && (tempchar != '_'))
		    Flag = true;
		  else
		    {
		      alphalen++;
		      temp[0] = tempchar; temp[1] = 0;
		      temp_alpha += std::string(temp);
		    }
		  //get out of while loop
		  if (Flag == true)
		    break;
		  
		  i++;
		  tempchar = (unsigned char)*(input + i);
		} 

	      if (*(unsigned char*)nextchar(input + i) == '(')
		funcpriority = PRI_PARENFUNCTION;
	      else
		funcpriority = PRI_FUNCTION;

	      if (*(unsigned char*)nextchar(input + i) == '(')
		printpriority = PRI_PARENFUNCTION;
	      else
		printpriority = PRI_PRINT;

	      //add to tokenbuff
	      //if
	      if (temp_alpha == std::string("if"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_IF;
		  if (addbuff(PRI_FUNCTION, IF\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //elseif
	      else if (temp_alpha == std::string("elsif"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_ELSEIF;
		  if (addbuff(PRI_FUNCTION, ELSEIF\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //else
	      else if (temp_alpha == std::string("else"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_ELSE;
		  if (addbuff(PRI_FUNCTION, ELSE\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //while
	      else if (temp_alpha == std::string("while"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_WHILE;
		  if (addbuff(PRI_FUNCTION, WHILE\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //until
	      else if (temp_alpha == std::string("until"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_WHILE;
		  if (addbuff(PRI_FUNCTION, UNTIL\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //for
	      else if (temp_alpha == std::string("for"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_FOR;
		  if (addbuff(PRI_FUNCTION, FOR\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		  if (addbuff(PRI_FUNCTION, NEXTFOR\
			      , "nextfor", "") == false)
		    return false;
		}
	      //foreach
	      else if (temp_alpha == std::string("foreach"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_FOREACH;
		  if (addbuff(PRI_FUNCTION, FOREACH\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		  if (addbuff(PRI_FUNCTION, NEXTFOREACH\
			      , "nextforeach", "") == false)
		    return false;
		}
	      //continue
	      else if (temp_alpha == std::string("continue"))
		{
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_CONTINUE;
		  if (addbuff(PRI_FUNCTION, CONTINUE\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //last
	      else if (temp_alpha == std::string("last"))
		{
		  if (addbuff(PRI_FUNCTION, LAST\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //next
	      else if (temp_alpha == std::string("next"))
		{
		  if (addbuff(PRI_FUNCTION, NEXT\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string eq
	      else if (temp_alpha == std::string("eq"))
		{
		  if (addbuff(PRI_STR_EQUAL, STR_EQUAL\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string not eq
	      else if (temp_alpha == std::string("ne"))
		{
		  if (addbuff(PRI_STR_EQUAL, STR_NOTEQUAL\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string less than 
	      else if (temp_alpha == std::string("lt"))
		{
		  if (addbuff(PRI_STR_LESSTHAN, STR_LESSTHAN\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string greater than 
	      else if (temp_alpha == std::string("gt"))
		{
		  if (addbuff(PRI_STR_LESSTHAN, STR_GREATERTHAN\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string less equal than 
	      else if (temp_alpha == std::string("le"))
		{
		  if (addbuff(PRI_STR_LESSTHAN, STR_LESSEQUALTHAN\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      //string greater equal than 
	      else if (temp_alpha == std::string("ge"))
		{
		  if (addbuff(PRI_STR_LESSTHAN, STR_GREATEREQUALTHAN\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      else if (is_pri_function((char *)temp_alpha.c_str()) == true)
		{
		  if (addbuff(funcpriority, FUNC\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      else if (is_pri_print((char *)temp_alpha.c_str()) == true)
		{
		  if (addbuff(printpriority, FUNC\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      else if (is_pri_parenfunction((char *)temp_alpha.c_str()) == true)
		{
		  if (addbuff(PRI_PARENFUNCTION, FUNC\
			      , temp_alpha.c_str(), "") == false)
		    return false;
		}
	      else
		{
		  if (buff->getarraylength() > 0)
		    {
		      if ((*buff)[buff->getarraylength() - 1].status == OBJ_POINTER)
			{
			  if (*(unsigned char*)nextchar(input + i) == '[')  //  object toarray operator
			    {
			      if (addbuff(PRI_PARENFUNCTION, OBJ_TOARRAY\
					  , "", temp_alpha.c_str()) == false)
				return false;
			    }
			  else if (*(unsigned char*)nextchar(input + i) == '(')  //  object method
			    {
			      if (addbuff(PRI_PARENFUNCTION, OBJ_METHOD\
					  , "", temp_alpha.c_str()) == false)
				return false;
			    }
			  else
			    {
			      if (addbuff(PRI_OPERAND, OBJ_MEMBER\
					  , "", temp_alpha.c_str()) == false)
				return false;
			    }
			}
		      else
			{
			  parentheseslevel++;
			  bracetype = BRC_OPENPAREN;
			  if (addbuff(PRI_OPAREN, OPAREN\
				      ,"", "") == false)
			    return false;

			  if (addbuff(PRI_OPERAND, NUM\
				      , temp_alpha.c_str(), "") == false)
			    return false;

			  bracetype = BRC_CLOSEDPAREN;
			  if (addbuff(PRI_CPAREN, CPAREN\
				      ,"", "") == false)
			    return false;
			  parentheseslevel--;
			}
		    }
		  else
		    {
		      parentheseslevel++;
		      bracetype = BRC_OPENPAREN;
		      if (addbuff(PRI_OPAREN, OPAREN\
				  ,"", "") == false)
			return false;

		      if (addbuff(PRI_OPERAND, NUM\
				  , temp_alpha.c_str(), "") == false)
			return false;

		      bracetype = BRC_CLOSEDPAREN;
		      if (addbuff(PRI_CPAREN, CPAREN\
				  ,"", "") == false)
			return false;
		      parentheseslevel--;
		    }
		}
	      
	      input += strlen(temp_alpha.c_str());
	      
	    }
	  //variant
	  else if ((unsigned char)*input == '$')
	    {
	      Flag = false;
	      alphalen = -1;
	      tempchar = (unsigned char)*input;
	      i = 0;
	      temp_alpha = std::string("");
	      
	      tempchar = (unsigned char)*(input);
	      alphalen++;
	      temp[0] = tempchar; temp[1] = 0;
	      temp_alpha += std::string(temp);

	      if (*(input + 1) == 0)
		return errormessage(lineno, "parse error", "$");

	      i++;
	      tempchar = (unsigned char)*(input + i);

	      //$a{
	      if (tempchar == '{')
		{
		  input = nextchar(input + i + 1);
		  strstatusbuffpoint++;
		  strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_TOOPERAND;
		  
		  if (addbuff(PRI_PARENFUNCTION, TO_OPERAND\
			      ,"", "") == false)
		    return false;
		  
		  if (addbuff(PRI_OPAREN, OPAREN\
			      ,"", "") == false)
		    return false;
		}
	      else
		{
		  tempchar = (unsigned char)*(input + i);
		  alphalen++;
		  temp[0] = tempchar; temp[1] = 0;
		  temp_alpha += std::string(temp);
		  i++;
		  tempchar = (unsigned char)*(input + i);
		  
		  while (tempchar != 0)
		    {
		      if (isalpha(tempchar) == 0
			  && isdigit(tempchar) == 0
			  && (tempchar != '_')
			  )
			Flag = true;
		      else
			{
			  alphalen++;
			  temp[0] = tempchar; temp[1] = 0;
			  temp_alpha += std::string(temp);
			}
		      //get out of while loop
		      if (Flag == true)
			break;
		      i++;
		      tempchar = (unsigned char)*(input + i);
		    } 
		  
		  //add to tokenbuff
		  if (*(unsigned char*)nextchar(input + i) == '[')  //  to array operator
		    {
		      input = nextchar(input + i);
		      strstatusbuffpoint++;
		      strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
		      bracestatusbuffpoint++;
		      bracestatusbuff[bracestatusbuffpoint] = BRC_TOARRAYVARIANT;
		      
		      if (addbuff(PRI_PARENFUNCTION, TOARRAY\
				  ,"", temp_alpha.c_str()) == false)
			return false;
		    }
		  else
		    {
		      if (temp_alpha == std::string("$graph"))
			{
			  if (addbuff(PRI_OPERAND, OBJ_MEMBER\
				      , "", temp_alpha.c_str()) == false)
			    return false;
			}
		      else
			{
			  if (strlen(temp_alpha.c_str()) == 2)
			    {
			      if (strchr(spchrs, temp_alpha.c_str()[1]) == NULL &&
				  isalpha(temp_alpha.c_str()[1]) == 0 &&
				  isdigit(temp_alpha.c_str()[1]) == 0
				  )
				return false;
			    }
			  if (addbuff(PRI_OPERAND, VARIANT\
				      , "", temp_alpha.c_str()) == false)
			    return false;
			}
		      input += strlen(temp_alpha.c_str());
		    }
		}
	    }  
	  //array
	  else if ((unsigned char)*input == '@')
	    {
	      Flag = false;
	      alphalen = -1;
	      tempchar = (unsigned char)*input;
	      i = 0;
	      temp_alpha = std::string("");
	      
	      tempchar = (unsigned char)*(input);
	      alphalen++;
	      temp[0] = tempchar; temp[1] = 0;
	      temp_alpha += std::string(temp);

	      if (*(input + 1) == 0)
		return errormessage(lineno, "parse error", "@");

	      i++;
	      tempchar = (unsigned char)*(input + i);


	      //@a{
	      if (tempchar == '{')
		{
		  input = nextchar(input + i + 1);
		  strstatusbuffpoint++;
		  strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
		  bracestatusbuffpoint++;
		  bracestatusbuff[bracestatusbuffpoint] = BRC_TOARRAY;
		  
		  if (addbuff(PRI_PARENFUNCTION, TO_ARRAY\
			      ,"", "") == false)
		    return false;
		  
		  if (addbuff(PRI_OPAREN, OPAREN\
			      ,"", "") == false)
		    return false;
		}
	      else
		{
		  tempchar = (unsigned char)*(input + i);
		  alphalen++;
		  temp[0] = tempchar; temp[1] = 0;
		  temp_alpha += std::string(temp);
		  i++;
		  tempchar = (unsigned char)*(input + i);
		  
		  while (tempchar != 0)
		    {
		      if (isalpha(tempchar) == 0
			  && isdigit(tempchar) == 0
			  && (tempchar != '_')
			  )
			Flag = true;
		      else
			{
			  alphalen++;
			  temp[0] = tempchar; temp[1] = 0;
			  temp_alpha += std::string(temp);
			}
		      //get out of while loop
		      if (Flag == true)
			break;
		      i++;
		      tempchar = (unsigned char)*(input + i);
		    } 
		  
		  if (temp_alpha == std::string("@graph"))
		    {
		      if (addbuff(PRI_OPERAND, OBJ_MEMBER\
				  , "", temp_alpha.c_str()) == false)
			return false;
		    }
		  //add to buffer
		  else
		    {
		      if (strlen(temp_alpha.c_str()) == 2)
			{
			  if (isalpha(temp_alpha.c_str()[1]) == 0 &&
			      isdigit(temp_alpha.c_str()[1]) == 0
			      )
			    return false;
			}
		      if (addbuff(PRI_OPERAND, ARRAY\
				  , "", temp_alpha.c_str()) == false)
			return false;
		    }
		  input += strlen(temp_alpha.c_str());
		}
	    }  
	  //others (operands)
	  else
	    {
	      switch(*(input++))
		{
		case '+' :
		  if (*input == '=')     //  +=
		    {
		      if (addbuff(PRI_SET, PLUSEQUAL\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '+')     //  ++
		    {
		      if (buff->getarraylength() > 0 &&
			  ((*buff)[buff->getarraylength() - 1].status == VARIANT
			  || (*buff)[buff->getarraylength() - 1].status == CPAREN)
			  )  // after
			{
			  if (addbuff(PRI_INCDEC, AFTER_INC\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		      else                                                        // may be before
			{
			  if (addbuff(PRI_INCDEC, BEFORE_INC\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		    }
		  else
		    {
		      if (buff->getarraylength() > 0 && (*buff)[buff->getarraylength() - 1].status >= 1000)  // +a
			{
			  if (addbuff(PRI_NOT, PREPLUS\
				      , "", "") == false)
			    return false;
			}
		      else    // a + b
			{
			  if (addbuff(PRI_PRUSMINUS, PLUS\
				      , "", "") == false)
			    return false;
			}
		    }
		  break;
		case '-' :
		  if (*input == '=')     //  -=
		    {
		      if (addbuff(PRI_SET, MINUSEQUAL\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '>')     //  ->
		    {
		      if (addbuff(PRI_OBJPOINTER, OBJ_POINTER\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '-')     //  --
		    {
		      if (buff->getarraylength() > 0 &&
			  ((*buff)[buff->getarraylength() - 1].status == VARIANT
			  || (*buff)[buff->getarraylength() - 1].status == CPAREN))  // after
			{
			  if (addbuff(PRI_INCDEC, AFTER_DEC\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		      else                                                        // may be before
			{
			  if (addbuff(PRI_INCDEC, BEFORE_DEC\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		    }
		  else          // -a
		    {
		      if (buff->getarraylength() > 0 && (*buff)[buff->getarraylength() - 1].status >= 1000) //  -a
			{
			  if (addbuff(PRI_NOT, PREMINUS\
				      , "", "") == false)
			    return false;
			}
		      else      // a - b
			{
			  if (addbuff(PRI_PRUSMINUS, MINUS\
				      , "", "") == false)
			    return false;
			}
		    }
		  break;
		case '*' :
		  if (*input == '*')     //exp
		    {
		      if (*(input + 1) == '=')     //  **=
			{
			  if (addbuff(PRI_SET, EXPEQUAL\
				      , "", "") == false)
			    return false;
			  input += 2;
			}
		      else                   // a ** b
			{
			  if (addbuff(PRI_EXP, EXP\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		    }
		  else
		    {
		      if (*input == '=')     //  *=
			{
			  if (addbuff(PRI_SET, TIMESEQUAL\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		      else                   // a * b
			{
			  if (addbuff(PRI_MULDIV, TIMES\
				      , "", "") == false)
			    return false;
			}
		    }
		  break;
		case '/' :
		      if (*input == '=')     //  /=
			{
			  if (addbuff(PRI_SET, DIVIDEEQUAL\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		      else                   // a / b
			{
			  if (addbuff(PRI_MULDIV, DIVIDE\
				      , "", "") == false)
			    return false;
			}
		  break;
		case '%' :
		      if (*input == '=')     //  %=
			{
			  if (addbuff(PRI_SET, INTDIVIDEEQUAL\
				      , "", "") == false)
			    return false;
			  input ++;
			}
		      else                   // a % b
			{
			  if (addbuff(PRI_MULDIV, INTDIVIDE\
				      , "", "") == false)
			    return false;
			}
		  break;
		case '^' :
		  if (addbuff(PRI_EXP, EXP\
			      , "", "") == false)
		    return false;
		  break;
		case '(' :
		  parentheseslevel++;
		  bracetype = BRC_OPENPAREN;
		  if (addbuff(PRI_OPAREN, OPAREN\
			      , "", "") == false)
		    return false;
		  break;
		case ')' : 

		  if (buff->getarraylength() > 0 && (*buff)[buff->getarraylength() - 1].status == OPAREN)
		    {
		      if (addbuff(PRI_OPERAND, NUM\
				  , "null", "") == false)
			return false;
		    }

		  bracetype = BRC_CLOSEDPAREN;
		  if (addbuff(PRI_CPAREN, CPAREN\
			      , "", "") == false)
		    return false;
		  parentheseslevel--;
		  break;
		case '[' :
		  parentheseslevel++;
		  bracetype = BRC_BRA;
		  if (addbuff(PRI_OPAREN, OPAREN\
			      , "", "") == false)
		    return false;
		  break;
		case ']' : 
		  bracetype = BRC_KET;
		  if (addbuff(PRI_CPAREN, CPAREN\
			      , "", "") == false)
		    return false;
		  if (bracestatusbuff[bracestatusbuffpoint] == BRC_TOARRAYVARIANT)
		    {
		      strstatusbuffpoint--;
		      bracestatusbuffpoint--;
		    }
		  parentheseslevel--;
		  break;
		case '{' : 
		  parentheseslevel++;
		  bracetype = BRC_OPENBRACE;
		  if (addbuff(PRI_CPAREN, OPAREN\
			      , "", "") == false)
		    return false;
		  if (bracestatusbuff[bracestatusbuffpoint] == BRC_TOOPERAND
		      || bracestatusbuff[bracestatusbuffpoint] == BRC_TOARRAY)
		    strstatusbuffpoint++;
		  break;
		case '}' : 
		  bracetype = BRC_CLOSEDBRACE;
		  if (addbuff(PRI_CPAREN, CPAREN\
			      , "", "") == false)
		    return false;
		  if (bracestatusbuff[bracestatusbuffpoint] == BRC_TOOPERAND
		      || bracestatusbuff[bracestatusbuffpoint] == BRC_TOARRAY)
		    strstatusbuffpoint--;
		  bracestatusbuffpoint--;
		  parentheseslevel--;
		  break;
		case '>' :
		  if (*input == '=')     //larger equal than
		    {
		      if (addbuff(PRI_LARGER_EQUAL_THAN, LARGER_EQUAL_THAN\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else                  //larger than
		    {
		      if (addbuff(PRI_LARGER_THAN, LARGER_THAN\
				  , "", "") == false)
			return false;
		    }
		  break;
		case '<' :
		  if (*input == '=')     //smaller equal than
		    {
		      if (addbuff(PRI_SMALLER_EQUAL_THAN, SMALLER_EQUAL_THAN\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else
		    {

		      Flag = false;
		      alphalen = -1;
		      tempchar = (unsigned char)*input;
		      i = 0;
		      temp_alpha = std::string("");
		      while (tempchar != 0)
			{
			  if (isalpha(tempchar) == 0 \
			      && isdigit(tempchar) == 0 \
			      && (tempchar != '_'))
			    Flag = true;
			  else
			    {
			      alphalen++;
			      temp[0] = tempchar; temp[1] = 0;
			      temp_alpha += std::string(temp);
			    }
			  //get out of while loop
			  if (Flag == true)
			    break;
			  
			  i++;
			  tempchar = (unsigned char)*(input + i);
			} 
		      
		      // read from file
		      if (tempchar == '>')
			{
			  if (addbuff(PRI_OPERAND, READPIPE\
				      , "", temp_alpha.c_str()) == false)
			    return false;
			  input = input + i + 1;
			}
		      else
			{
			  //smaller than
			  if (addbuff(PRI_SMALLER_THAN, SMALLER_THAN\
				      , "", "") == false)
			    return false;
			}


		    }
		  break;
		case '=' :
		  if (*input == '>')     //larger equal than
		    {
		      if (addbuff(PRI_LARGER_EQUAL_THAN, LARGER_EQUAL_THAN\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '<')     //smaller equal than
		    {
		      if (addbuff(PRI_SMALLER_EQUAL_THAN, SMALLER_EQUAL_THAN\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '=')     //equal
		    {
		      if (addbuff(PRI_EQUAL, EQUAL\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '~')     //pattern match
		    {
		      if (addbuff(PRI_MATCH, MATCH\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else                  //set
		    {
		      if (addbuff(PRI_SET, SET\
				  , "", "") == false)
			return false;
		    }
		  break;
		case '!' :
		  if (*input == '=')     //not equal
		    {
		      if (addbuff(PRI_NOTEQUAL, NOTEQUAL\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else if (*input == '~')     //not pattern match
		    {
		      if (addbuff(PRI_MATCH, NOTMATCH\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else                 // not
		    {
		      if (addbuff(PRI_NOT, NOT\
				  , "", "") == false)
			return false;
		    }
		  break;
		case '|' :
		  if (*input == '|')     // or
		    {
		      if (addbuff(PRI_OR, OR\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else
		    return false;
		  break;
		case '&' :
		  if (*input == '&')     // and
		    {
		      if (addbuff(PRI_AND, AND\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else
		    return false;
		  break;
		case ';' :
		  if (addbuff(PRI_SEMICOLON, SEMICOLON\
			      , "", "") == false)
		    return false;
		  break;
		case ',' :
		  if (*input == '=')     //  ,=
		    {
		      if (addbuff(PRI_SET, COMMAEQUAL\
				  , "", "") == false)
			return false;
		      input ++;
		    }
		  else                   // ,
		    {
		      if (addbuff(PRI_COMMA, COMMA\
				  , "", "") == false)
			return false;
		    }
		  break;
		case '#':     //comment
		  return true;
		  break;
		default: 
		  return errormessage(lineno, "parse error", "");
		} /* switch */
	    }
	  break;

	  // in the single quotation
	case STR_SQUOTE:
	  if ((unsigned char)*input == '\'')
	  {
	    input++;
	    strstatusbuffpoint--;

	    if (addbuff(PRI_CPAREN, CPAREN\
			, "", "") == false)
	      return false;
	    parentheseslevel--;
	  }
	  else 
	    {
	      temp_alpha = std::string("");
	      while(*input != '\'' && *input != 0)
		{
		  //escape sequence
		  if (*input == '\\')
		    {
		      //octagonal 
		      if (isdigit(*(input + 1)) != 0)
			{
			  if (*(input + 2) != 0)
			    {
			      if (*(input + 3) != 0)
				{
				  temp[0] = (*(input + 1) - '0') * 64 + (*(input + 2) - '0') * 8 + (*(input + 3) - '0');
				  temp[1] = 0;
				  input += 4;
				}
			      else
				return errormessage(lineno, "parse error", "\\");
			    }
			  else
			    return errormessage(lineno, "parse error", "\\");
			}
		      else
			{
			  switch (*(input + 1))
			    {
			    case '\'':
			      temp[0] = '\''; temp[1] = 0;
			      input += 2;
			      break;
			    case '\\':
			      temp[0] = '\\'; temp[1] = 0;
			      input += 2;
			      break;
			    default:
			      temp[0] = *input; temp[1] = 0;
			      input ++;
			      break;
			    }
			}
		    }
		  else
		    {
		      temp[0] = *input; temp[1] = 0;
		      input++;
		    }
		  temp_alpha += std::string(temp);
		}
	      
	      if (addbuff(PRI_PRUSMINUS, STR_PLUS\
			  , "", "") == false)
		return false;
	      
	      if (addbuff(PRI_OPERAND, NUM,\
			  temp_alpha.c_str(), "") == false)
		return false;
	    }
	  break;
	  
	  // in the double / upper quotation
	case STR_UQUOTE:
	case STR_DQUOTE:
	case STR_REGEX:
	  if (strstatusbuff[strstatusbuffpoint] == STR_UQUOTE)
	    {
	      plusstatus = IMPLICITSTRPLUS;
	      quoteseparator = '`';
	    }
	  else if (strstatusbuff[strstatusbuffpoint] == STR_DQUOTE)
	    {
	      plusstatus = STR_PLUS;
	      quoteseparator = '"';
	    }
	  else if (strstatusbuff[strstatusbuffpoint] == STR_REGEX)
	    {
	      plusstatus = STR_PLUS;
	      quoteseparator = '/';
	    }

	  if ((unsigned char)*input == '"'\
	      && strstatusbuff[strstatusbuffpoint] == STR_DQUOTE)
	  {
	      input++;
	      strstatusbuffpoint--;

	      if (addbuff(PRI_CPAREN, CPAREN\
			  , "", "") == false)
		return false;
	      parentheseslevel--;
	  }
	  else if ((unsigned char)*input == '`'\
		   && strstatusbuff[strstatusbuffpoint] == STR_UQUOTE)
	  {
	    input++;
	    strstatusbuffpoint--;
	    
	    if (addbuff(PRI_CPAREN, CPAREN\
			, "", "") == false)
	      return false;
	    parentheseslevel--;
	  }
	  else if ((unsigned char)*input == '/'\
		   && strstatusbuff[strstatusbuffpoint] == STR_REGEX)
	  {
	    input++;
	    strstatusbuffpoint--;
	    
	    if (addbuff(PRI_CPAREN, CPAREN\
			, "", "") == false)
	      return false;
	    parentheseslevel--;
	  }
	  else
	    {
	      if (*input  == ' ')
		{
		  temp_alpha = std::string("");
		  while(*input == ' ')
		    {
		      temp[0] = *input; temp[1] = 0;
		      temp_alpha += std::string(temp);
		      input++;
		    }

		  if (addbuff(PRI_PRUSMINUS, plusstatus\
			      , "", "") == false)
		    return false;
		  
		  if (addbuff(PRI_OPERAND, NUM,\
			      temp_alpha.c_str(), "") == false)
		    return false;

		}
	      else if (*input  == '$')
		{
		  //  such as ${abs}		  
		  if (*(unsigned char*)nextchar(input + 1) == '{')
		    {
		      input = nextchar(input + 1) + 1;
		      strstatusbuffpoint++;
		      strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
		      bracestatusbuffpoint++;
		      bracestatusbuff[bracestatusbuffpoint] = BRC_TOOPERAND;
		      parentheseslevel++;

		      //
		      if (addbuff(PRI_PRUSMINUS, plusstatus\
				  , "", "") == false)
			return false;
		      
		      if (addbuff(PRI_PARENFUNCTION, TO_OPERAND\
				  ,"", "") == false)
			return false;

		      if (addbuff(PRI_OPAREN, OPAREN\
				  ,"", "") == false)
			return false;

		    }
		  //  such as $abc
		  else
		    {
		      temp_alpha = std::string("$");
		      input = nextchar(input + 1);
		      if (*input == 0)
			return errormessage(lineno, "parse error", "$");
		      else if (*input == '/' && strstatusbuff[strstatusbuffpoint] == STR_REGEX)  // last $ for regrep
			{
			  if (addbuff(PRI_PRUSMINUS, plusstatus\
				      , "", "") == false)
			    return false;
			  
			  if (addbuff(PRI_OPERAND, NUM\
				      , "$", "") == false)
			    return false;
			}
		      else
			{
			  temp[0] = *input; temp[1] = 0;
			  temp_alpha += std::string(temp);
			  input++;
			  while(*input != ' ' 
				&& *input != quoteseparator
				&& *input != '$'
				&& *input != 0
				&& (
				    isalpha(*input) != 0
				    || isdigit(*input) != 0
				    || *input == '_'
				    )
				)
			    {
			      temp[0] = *input; temp[1] = 0;
			      temp_alpha += std::string(temp);
			      input++;
			    }

			  //such as $ab[1]
			  if (*input == '[')
			    {
			      input = nextchar(input);
			      strstatusbuffpoint++;
			      strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
			      bracestatusbuffpoint++;
			      bracestatusbuff[bracestatusbuffpoint] = BRC_TOARRAYVARIANT;
			      
			      //
			      if (addbuff(PRI_PRUSMINUS, plusstatus\
					  , "", "") == false)
				return false;
			      
			      if (addbuff(PRI_PARENFUNCTION, TOARRAY\
					  ,"", temp_alpha.c_str()) == false)
				return false;
			      
			    }
			  else
			    {
			      //add to tokenbuff
			      if (addbuff(PRI_PRUSMINUS, plusstatus\
					  , "", "") == false)
				return false;
			      
			      if (addbuff(PRI_OPERAND, VARIANT\
					  , "", temp_alpha.c_str()) == false)
				return false;
			    }
			}
		      
		    }
		}
	      else if (*input  == '@')
		{
		  //  such as @{abs}		  
		  if (*(unsigned char*)nextchar(input + 1) == '{')
		    {
		      input = nextchar(input + 1) + 1;
		      strstatusbuffpoint++;
		      strstatusbuff[strstatusbuffpoint] = STR_NORMAL;
		      bracestatusbuffpoint++;
		      bracestatusbuff[bracestatusbuffpoint] = BRC_TOARRAY;
		      parentheseslevel++;

		      //
		      if (addbuff(PRI_PRUSMINUS, plusstatus\
				  , "", "") == false)
			return false;
		      
		      if (addbuff(PRI_PARENFUNCTION, TO_ARRAY\
				  ,"", "") == false)
			return false;

		      if (addbuff(PRI_OPAREN, OPAREN\
				  ,"", "") == false)
			return false;

		    }
		  //  such as @abc
		  else
		    {
		      temp_alpha = std::string("@");
		      input = nextchar(input + 1);
		      if (*input == 0)
			return errormessage(lineno, "parse error", "@");
		      else
			{
			  temp[0] = *input; temp[1] = 0;
			  temp_alpha += std::string(temp);
			  input++;
			  while(*input != ' ' 
				&& *input != quoteseparator
				&& *input != '$'
				&& *input != 0
				&& (
				    isalpha(*input) != 0
				    || isdigit(*input) != 0
				    || *input == '_'
				    )
				)
			    {
			      temp[0] = *input; temp[1] = 0;
			      temp_alpha += std::string(temp);
			      input++;
			    }

			  //add to tokenbuff
			  if (addbuff(PRI_PRUSMINUS, plusstatus\
				      , "", "") == false)
			    return false;
			  
			  if (addbuff(PRI_OPERAND, ARRAY\
				      , "", temp_alpha.c_str()) == false)
			    return false;
			  
			}
		    }
		}
	      else
		{
		  temp_alpha = std::string("");
		  while(*input != ' ' 
			&& *input != quoteseparator
			&& *input != '$'
			&& *input != 0)
		    {
		      //escape sequence
		      if (*input == '\\')
			{
			  //octagonal 
			  if (isdigit(*(input + 1)) != 0)
			    {
			      if (*(input + 2) != 0 && *(input + 3) != 0)
				{
				  temp[0] = (*(input + 1) - '0') * 64 + (*(input + 2) - '0') * 8 + (*(input + 3) - '0');
				  temp[1] = 0;
				  input += 4;
				}
			      else
				return errormessage(lineno, "parse error", "\\");
			    }
			  else
			    {
			      switch (*(input + 1))
				{
				case 'a':
				  temp[0] = '\a'; temp[1] = 0;
				  input += 2;
				  break;
				case 'b':
				  temp[0] = '\b'; temp[1] = 0;
				  input += 2;
				  break;
				case 'e':
				  temp[0] = '\e'; temp[1] = 0;
				  input += 2;
				  break;
				case 'f':
				  temp[0] = '\f'; temp[1] = 0;
				  input += 2;
				  break;
				case 'n':
				  temp[0] = '\n'; temp[1] = 0;
				  input += 2;
				  break;
				case 'r':
				  temp[0] = '\r'; temp[1] = 0;
				  input += 2;
				  break;
				case 't':
				  temp[0] = '\t'; temp[1] = 0;
				  input += 2;
				  break;
				case '\'':
				  temp[0] = '\''; temp[1] = 0;
				  input += 2;
				  break;
				case '"':
				  temp[0] = '"'; temp[1] = 0;
				  input += 2;
				  break;
				case '$':
				  temp[0] = '$'; temp[1] = 0;
				  input += 2;
				  break;
				case '/':
				  temp[0] = '/'; temp[1] = 0;
				  input += 2;
				  break;
				case '@':
				  temp[0] = '@'; temp[1] = 0;
				  input += 2;
				  break;
				case '\\':
				  temp[0] = '\\'; temp[1] = 0;
				  input += 2;
				  break;
				default:
				  temp[0] = *input; temp[1] = 0;
				  input ++;
				  break;
				}
			    }
			}
		      else
			{
			  temp[0] = *input; temp[1] = 0;
			  input++;
			}
		      temp_alpha += std::string(temp);
		    }
		  
		  if (addbuff(PRI_PRUSMINUS, plusstatus\
			      , "", "") == false)
		    return false;
		  
		  if (addbuff(PRI_OPERAND, NUM,\
			      temp_alpha.c_str(), "") == false)
		    return false;
		}	      
	      
	    }	  
	  break;
	}
    }
  return true;
  
} /* nexttoken */
