/*----------------------------------------------------------------------------
                            parts.cc (parts object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string>
#include <time.h>
#include "graph.h"
#include "frame.h"
#include "parts.h"
#include "data.h"
#include "defcharwidth.h"
#include "gdi.h"
#include "pointdefs.h"
#include "vectdefs.h"
#include "spline.h"
#include "minmax.h"
#include "filepoint.h"
#include "kanji.h"
#include "topazvalues.h"
#include "convtxt.h"

extern void StyleLinePath(bool flag, FILE* f, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);

extern FileHandleArray *fhvuff;

parts::parts():plobj()
{
  version                 = 1;
  id                      = 0;
  label                   = new std::string("");
  selected                = false;
  type                    = 0;

  /* font */
  text = new std::string("");
  fo.init();
  fontcolor.setblack();
  linespacing             = 1.0;
  
  /* common */
  linecolor.setblack();
  bodycolor.setwhite();
  linestyle.init();
  bodystyle.init();
  legendlength            = 100;
  interpolate             = PARTS_INTER_POLYLINE;
  interpolatediv          = 50;
  p1.x                    = 0;
  p1.y                    = 0;
  p2.x                    = 0;
  p2.y                    = 0;
  p1_arrow                = false;
  p2_arrow                = false;
  arrowlength             = 250;
  arrowhalfangle          = 10; 

  nodes                   = new intpointarray;

}

parts::~parts()
{
  delete label;
  delete text;
  delete nodes;
}

parts& parts::operator = (const parts&  parts)
{
//  id                   = parts.id;
  selected             = parts.selected;
  type                 = parts.type;
  *text                = *parts.text;
  fo                   = parts.fo;
  fontcolor            = parts.fontcolor;
  linespacing          = parts.linespacing;
  linecolor            = parts.linecolor;
  bodycolor            = parts.bodycolor;
  linestyle            = parts.linestyle;
  bodystyle            = parts.bodystyle;
  legendlength         = parts.legendlength;
  interpolate          = parts.interpolate;
  interpolatediv       = parts.interpolatediv;
  p1.x                 = parts.p1.x;
  p1.y                 = parts.p1.y;
  p2.x                 = parts.p2.x;
  p2.y                 = parts.p2.y;
  p1_arrow             = parts.p1_arrow;
  p2_arrow             = parts.p2_arrow;
  arrowlength          = parts.arrowlength;
  arrowhalfangle       = parts.arrowhalfangle; 
  *nodes               = *parts.nodes;

  return *this;
}


bool parts::get(std::string* member, tokenbuff* outvalue)
{
  std::string out;
  int iid;
  char stemp[100];

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  sprintf(stemp, "%d", id);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  outvalue->value->add(label->c_str());
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  sprintf(stemp, "%d", selected);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("type") && iid == -1)
	{
	  sprintf(stemp, "%d", type);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("text") && iid == -1)
	{
	  outvalue->value->add(text->c_str());
	  return true;
	}
      else if (out == std::string("fontface") && iid == -1)
	{
	  sprintf(stemp, "%d", fo.face);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("jpfontface") && iid == -1)
	{
	  sprintf(stemp, "%d", fo.jpface);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontsize") && iid == -1)
	{
	  sprintf(stemp, "%d", fo.size);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontangle") && iid == -1)
	{
	  sprintf(stemp, "%.15g", fo.angle);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("fontcolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", fontcolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linespacing") && iid == -1)
	{
	  sprintf(stemp, "%.15g", linespacing);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", linecolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bodycolor_red") && iid == -1)
	{
	  sprintf(stemp, "%u", bodycolor.red);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bodycolor_green") && iid == -1)
	{
	  sprintf(stemp, "%u", bodycolor.green);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bodycolor_blue") && iid == -1)
	{
	  sprintf(stemp, "%u", bodycolor.blue);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linewidth") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.width);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linecap") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.cap);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linejoin") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.join);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("linestyle") && iid == -1)
	{
	  sprintf(stemp, "%d", linestyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bodyfillstyle") && iid == -1)
	{
	  sprintf(stemp, "%d", bodystyle.style);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("bodyfillrule") && iid == -1)
	{
	  sprintf(stemp, "%d", bodystyle.fillrule);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("legendlength") && iid == -1)
	{
	  sprintf(stemp, "%d", legendlength);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("interpolate") && iid == -1)
	{
	  sprintf(stemp, "%d", interpolate);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("interpolatediv") && iid == -1)
	{
	  sprintf(stemp, "%d", interpolatediv);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("x1") && iid == -1)
	{
	  sprintf(stemp, "%d", p1.x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("y1") && iid == -1)
	{
	  sprintf(stemp, "%d", p1.y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("x2") && iid == -1)
	{
	  sprintf(stemp, "%d", p2.x);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("y2") && iid == -1)
	{
	  sprintf(stemp, "%d", p2.y);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("arrow1") && iid == -1)
	{
	  sprintf(stemp, "%d", p1_arrow);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("arrow2") && iid == -1)
	{
	  sprintf(stemp, "%d", p2_arrow);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("arrowlength") && iid == -1)
	{
	  sprintf(stemp, "%d", arrowlength);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("arrowhalfangle") && iid == -1)
	{
	  sprintf(stemp, "%.15g", arrowhalfangle);
	  outvalue->value->add(stemp);
	  return true;
	}
      else if (out == std::string("node_x") && 0 <= iid)
	{
	  if (iid < nodes->getarraylength())
	    {
	      sprintf(stemp, "%d", (*nodes)[iid].x);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("node_y") && 0 <= iid)
	{
	  if (iid < nodes->getarraylength())
	    {
	      sprintf(stemp, "%d", (*nodes)[iid].y);
	      outvalue->value->add(stemp);
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}

bool parts::set(std::string* member, std::string* setvalue)
{
  std::string out;
  int iid;

  if (topobject(member, &out, &iid) == true)
    {
      if (out == std::string("id") && iid == -1)
	{
	  id = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("label") && iid == -1)
	{
	  *label = *setvalue;
	  return true;
	}
      else if (out == std::string("selected") && iid == -1)
	{
	  selected = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("type") && iid == -1)
	{
	  type = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("text") && iid == -1)
	{
	  *text = *setvalue;
	  return true;
	}
      else if (out == std::string("fontface") && iid == -1)
	{
	  fo.face = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("jpfontface") && iid == -1)
	{
	  fo.jpface = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontsize") && iid == -1)
	{
	  fo.size = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontangle") && iid == -1)
	{
	  fo.angle = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_red") && iid == -1)
	{
	  fontcolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_green") && iid == -1)
	{
	  fontcolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("fontcolor_blue") && iid == -1)
	{
	  fontcolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linespacing") && iid == -1)
	{
	  linespacing = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_red") && iid == -1)
	{
	  linecolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_green") && iid == -1)
	{
	  linecolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecolor_blue") && iid == -1)
	{
	  linecolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bodycolor_red") && iid == -1)
	{
	  bodycolor.red = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bodycolor_green") && iid == -1)
	{
	  bodycolor.green = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bodycolor_blue") && iid == -1)
	{
	  bodycolor.blue = atol(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linewidth") && iid == -1)
	{
	  linestyle.width = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linecap") && iid == -1)
	{
	  linestyle.cap = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linejoin") && iid == -1)
	{
	  linestyle.join = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("linestyle") && iid == -1)
	{
	  linestyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bodyfillstyle") && iid == -1)
	{
	  bodystyle.style = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("bodyfillrule") && iid == -1)
	{
	  bodystyle.fillrule = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("legendlength") && iid == -1)
	{
	  legendlength = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("interpolate") && iid == -1)
	{
	  interpolate = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("interpolatediv") && iid == -1)
	{
	  interpolatediv = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("x1") && iid == -1)
	{
	  p1.x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("y1") && iid == -1)
	{
	  p1.y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("x2") && iid == -1)
	{
	  p2.x = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("y2") && iid == -1)
	{
	  p2.y = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("arrow1") && iid == -1)
	{
	  p1_arrow = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("arrow2") && iid == -1)
	{
	  p2_arrow = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("arrowlength") && iid == -1)
	{
	  arrowlength = atoi(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("arrowhalfangle") && iid == -1)
	{
	  arrowhalfangle = atof(setvalue->c_str());
	  return true;
	}
      else if (out == std::string("node_x") && 0 <= iid)
	{
	  if (iid < nodes->getarraylength())
	    {
	      (*nodes)[iid].x = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else if (out == std::string("node_y") && 0 <= iid)
	{
	  if (iid < nodes->getarraylength())
	    {
	      (*nodes)[iid].y = atoi(setvalue->c_str());
	      return true;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;

}

bool parts::exec(std::string* function, buffarray *argument, tokenbuff* outvalue)
{
  std::string out, shrunk, str1;
  int iid, arglength;
  char stemp[100];
  intrect bbox;
  FILE* fp;
  int itemp;
  graph* gra = (graph *)(((frame *)parent)->parent);

  outvalue->value->setbuff(0);
  if (topobject(function, &out, &iid) == true)
    {
      arglength = argument->getarraylength();

      // paint
      if (out == std::string("paint") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if (fhvuff->get(&(*argument)[0], &fp, &itemp, &str1))
		{
		  paint(true, fp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // setbbox
      else if (out == std::string("setbbox") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      paint(false, fp);
	      return true;
	    }
	  else
	    return false;
	}
      // length
      else if (out == std::string("length") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if ((*argument)[0] == std::string("nodes"))
		{
		  sprintf(stemp, "%d", nodes->getarraylength());
		  outvalue->value->add(stemp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      // addnodepoint
      else if (out == std::string("addnodepoint") && iid == -1)
	{
	  if (arglength == 2)
	    {
	      gra->updated = true;
	      if (nodes->add(new intpoint(atoi((*argument)[0].c_str()), atoi((*argument)[1].c_str()))))
		return true;
	      else
		return false;
	    }
	  else
	    return false;
	}
      // flushnode
      else if (out == std::string("flushnode") && iid == -1)
	{
	  if (arglength == 1 && (*argument)[0] == std::string("null"))
	    {
	      gra->updated = true;
	      nodes->setbuff(0);
		return true;
	    }
	  else
	    return false;
	}
      // bounding box
      else if (out == std::string("bbox") && iid == -1)
	{
	  if (arglength == 1)
	    {
	      if ((*argument)[0] == std::string("null"))
		{
		  getBBbox(&bbox);
		  outvalue->value->setbuff(0);
		  sprintf(stemp, "%d", bbox.left);
		  outvalue->value->add(stemp);
		  sprintf(stemp, "%d", bbox.top);
		  outvalue->value->add(stemp);
		  sprintf(stemp, "%d", bbox.right);
		  outvalue->value->add(stemp);
		  sprintf(stemp, "%d", bbox.bottom);
		  outvalue->value->add(stemp);
		  return true;
		}
	      else
		return false;
	    }
	  else
	    return false;
	}
      else
	return false;
    }
  else
    return false;
      
  return true;
}

void parts::getBBbox(intrect *r)
{
  int maxx, minx, maxy, miny;

//  intpoint temp[50];
  int i, no, interporatedno;
  doublearray *InX = 0, *InY = 0, *OutX = 0, *OutY = 0;
  
  frame *fr = (frame *)parent;

  minx =  MAXINT;  maxx = -MAXINT;
  miny =  MAXINT;  maxy = -MAXINT;

  switch(type)
    {
    case PARTS_TEXT:
      drawtext(false, 0, r, 0);
      break;
    case PARTS_LINE:
    case PARTS_RECTANGLE:
    case PARTS_ELLIPSE:
    case PARTS_CUT:
      minx = min(p1.x, p2.x); maxx = max(p1.x, p2.x);
      miny = min(p1.y, p2.y); maxy = max(p1.y, p2.y);
      r->left = fr->wtov_x(minx); r->right  = fr->wtov_x(maxx);
      r->top  = fr->wtov_y(miny); r->bottom = fr->wtov_y(maxy);
      break;
    case PARTS_POLYLINE:
      no = nodes->getarraylength() - 1;
      if (no >= 1)
	{
	  if (interpolate == PARTS_INTER_POLYLINE || interpolate == PARTS_INTER_CLOSEDPOLYLINE)
	    {
	      for (i = 0; i <= no; i++)
		{
		  minx = min(minx, fr->wtov_x((*nodes)[i].x + p1.x));
		  maxx = max(maxx, fr->wtov_x((*nodes)[i].x + p1.x));
		  miny = min(miny, fr->wtov_y((*nodes)[i].y + p1.y));
		  maxy = max(maxy, fr->wtov_y((*nodes)[i].y + p1.y));
		}

	      r->left = minx; r->right = maxx;
	      r->top  = miny; r->bottom = maxy;
	    }
	  else if (interpolate == PARTS_INTER_PSPLINE || interpolate == PARTS_INTER_CLOSEDPSPLINE)
	    {
	      InX = new doublearray;
	      if (InX == 0)
		{
		  InX = 0;
		  goto j1;
		}
	      
	      InY = new doublearray;
	      if (InY == 0)
		{
		  InY = 0;
		  goto j1;
		}
	      
 	      OutX = new doublearray;
	      if (OutX == 0)
		{
		  OutX = 0;
		  goto j1;
		}
	      
 	      OutY = new doublearray;
	      if (OutY == 0)
		{
		  OutY = 0;
		  goto j1;
		}
	      
	      for (i = 0; i <= no; i++)
		{
		  InX->add((*nodes)[i].x + p1.x);
		  InY->add((*nodes)[i].y + p1.y);
		}
	      
	      if (interpolate ==  PARTS_INTER_PSPLINE)
		CalcPSpline(InX, InY, OutX, OutY, 1, interpolatediv);
	      else
		CalcPSpline(InX, InY, OutX, OutY, 2, interpolatediv);
	      
	      interporatedno = OutX->getarraylength() - 1;
	      
	      for (i = 0; i <= interporatedno; i++)
		{
		  minx = min(minx, fr->wtov_x((int)(*OutX)[i]));
		  maxx = max(maxx, fr->wtov_x((int)(*OutX)[i]));
		  miny = min(miny, fr->wtov_y((int)(*OutY)[i]));
		  maxy = max(maxy, fr->wtov_y((int)(*OutY)[i]));
		}
	    j1:
	      if (InX != 0)
		delete InX;
	      if (InY != 0)
		delete InY;
	      if (OutX != 0)
		delete OutX;
	      if (OutY != 0)
		delete OutY;

	      r->left = minx; r->right  = maxx;
	      r->top  = miny; r->bottom = maxy;
	    }
	}
      else
	{
	  r->left = 0; r->right  = 0;
	  r->top  = 0; r->bottom = 0;
	}
      break;
    }
}


bool parts::paint(bool flag, FILE* f)
{
  switch(type)
    {
    case PARTS_TEXT:
      drawtext(flag, f, 0, 0);
      break;
    case PARTS_LINE:
    case PARTS_POLYLINE:
    case PARTS_RECTANGLE:
    case PARTS_ELLIPSE:
    case PARTS_CUT:
      drawparts(flag, f);
      break;
    }

  return true;
}

bool getpartsname(int id, std::string *out)
{
  switch(id)
    {
    case PARTS_TEXT:
      *out = std::string("text");
      break;
    case PARTS_LINE:
      *out = std::string("line");
      break;
    case PARTS_POLYLINE:
      *out = std::string("polyline");
      break;
    case PARTS_RECTANGLE:
      *out = std::string("rectangle");
      break;
    case PARTS_ELLIPSE:
      *out = std::string("ellipse");
      break;
    case PARTS_CUT:
      *out = std::string("cutline");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

bool gettextfacename(int id, std::string *out)
{
  switch(id)
    {
    case FONT_TIMESROMAN:
      *out = std::string("Times-Roman");
      break;
    case FONT_TIMESBOLD:
      *out = std::string("Times-Bold");
      break;
    case FONT_TIMESITALIC:
      *out = std::string("Times-Italic");
      break;
    case FONT_TIMESBOLDITALIC:
      *out = std::string("Times-BoldItalic");
      break;
    case FONT_HELVETICA:
      *out = std::string("Helvetica");
      break;
    case FONT_HELVETICABOLD:
      *out = std::string("Helvetica-Bold");
      break;
    case FONT_HELVETICAOBLIQUE:
      *out = std::string("Helvetica-Oblique");
      break;
    case FONT_HELVETICABOLDOBLIQUE:
      *out = std::string("Helvetica-BoldOblique");
      break;
    case FONT_COURIER:
      *out = std::string("Courier");
      break;
    case FONT_COURIERBOLD:
      *out = std::string("Courier-Bold");
      break;
    case FONT_COURIEROBLIQUE:
      *out = std::string("Courier-Oblique");
      break;
    case FONT_COURIERBOLDOBLIQUE:
      *out = std::string("Courier-BoldOblique");
      break;
    case FONT_SYMBOL:
      *out = std::string("Symbol");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

bool gettextjpfacename(int id, std::string *out)
{
  switch(id)
    {
    case FONT_MINCHO:
      *out = std::string("Mincho");
      break;
    case FONT_GOTHIC:
      *out = std::string("Gothic");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

intpoint RotatePoint(intpoint P, intpoint OrgP, double Theta)
{
  intpoint TempPoint1, TempPoint2;
  double fx, fy, Radian;
  
  TempPoint1.x = P.x - OrgP.x;
  TempPoint1.y = P.y - OrgP.y;
  Radian = Theta / 180.0 * M_PI;
  
  fx = cos(Radian) * (double)TempPoint1.x + sin(Radian) * (double)TempPoint1.y + (double)OrgP.x;
  fy = -sin(Radian) * (double)TempPoint1.x + cos(Radian) * (double)TempPoint1.y + (double)OrgP.y;
  
  TempPoint2.x = (int)fx;
  TempPoint2.y = (int)fy;
  
  return TempPoint2;
}


int parts::getcharwidth(Capfont *lf, unsigned int c, bool iskanji)
{
  double ftemp;
  
  // ascii
  if (!iskanji)
    {
      switch(lf->face)
	{
	case FONT_TIMESROMAN:
	  ftemp = TR[c];
	  break;
	case FONT_TIMESBOLD:
	  ftemp = TB[c];
	  break;
	case FONT_TIMESITALIC:
	  ftemp = TI[c];
	  break;
	case FONT_TIMESBOLDITALIC:
	  ftemp = TO[c];
	  break;
	case FONT_HELVETICA:
	  ftemp = HR[c];
	  break;
	case FONT_HELVETICABOLD:
	  ftemp = HB[c];
	  break;
	case FONT_HELVETICAOBLIQUE:
	  ftemp = HI[c];
	  break;
	case FONT_HELVETICABOLDOBLIQUE:
	  ftemp = HO[c];
	  break;
	case FONT_COURIER:
	  ftemp = CR[c];
	  break;
	case FONT_COURIERBOLD:
	  ftemp = CB[c];
	  break;
	case FONT_COURIEROBLIQUE:
	  ftemp = CI[c];
	  break;
	case FONT_COURIERBOLDOBLIQUE:
	  ftemp = CO[c];
	  break;
	case FONT_SYMBOL:
	  ftemp = SR[c];
	  break;
	default:
	  ftemp = 0;
	}
    }
  // jp
  else
    {
      ftemp = 10.0;
    }
  return (int)(((double)ftemp) * 20 * ((double)lf->size) / 200.0);
}

int parts::getascend(Capfont *lf, bool iskanji)
{
  double ftemp;

  if (!iskanji)
    {
      switch(lf->face)
	{
	case FONT_TIMESROMAN:
	  ftemp = DTR;
	  break;
	case FONT_TIMESBOLD:
	  ftemp = DTB;
	  break;
	case FONT_TIMESITALIC:
	  ftemp = DTI;
	  break;
	case FONT_TIMESBOLDITALIC:
	  ftemp = DTO;
	  break;
	case FONT_HELVETICA:
	  ftemp = DHR;
	  break;
	case FONT_HELVETICABOLD:
	  ftemp = DHB;
	  break;
	case FONT_HELVETICAOBLIQUE:
	  ftemp = DHI;
	  break;
	case FONT_HELVETICABOLDOBLIQUE:
	  ftemp = DHO;
	  break;
	case FONT_COURIER:
	  ftemp = DCR;
	  break;
	case FONT_COURIERBOLD:
	  ftemp = DCB;
	  break;
	case FONT_COURIEROBLIQUE:
	  ftemp = DCI;
	  break;
	case FONT_COURIERBOLDOBLIQUE:
	  ftemp = DCO;
	  break;
	case FONT_SYMBOL:
	  ftemp = DSR;
	  break;
	default:
	  ftemp = 0;
	}
    }
  else
    ftemp = 1.2;
  return (int)((10.0 - (double)ftemp) * 20 *((double)lf->size) / 200.0);
}



char * nextpoint(char *in)
{
  if (maykanji(*in))
    return (in + 2);
  else
    return (in + 1);
}


void parts::drawtext(int flag, FILE* f, intrect *rect, int* textwidth)
{
  /*
     Large						small
     A
     B       font bold
     C
     D       datafilename (full path)         datafilename (only filename)
     E
     F       restore font
     G       greek
     H       plots filename(full path)        plotsfilename (only filename)
     I       italic
     J
     K
     L
     M       date(month:day:year)			date(year/month/day)
     N       
     O       bold italic
     P       pitch           				legend pattern
     Q
     R       roman
     S       font size
     T       Large TAB          				small tabbing
     U
     V
     W
     X       shift x-direction
     Y       shift y-direction
     Z
     */

  
  Capfont lf, lf0;
  int i;
  int CX,CY,CX0,CY0, maxCX;
  char *lpszTemp1;
  char temp3[10];
  unsigned char ustemp[10];
  int itemp;
  int dx,dy,tdx;
  int Tabspace, SmallTabspace;
  int h;
  int KaigyoNo = 0;
  char char_D[2]; char_D[0] = 13; char_D[1] = '\0';
  char char_A[2]; char_A[0] = 10; char_A[1] = '\0';
  intpoint RotatedPoint;		//position of text out after rotated
  std::string str1, str2;
  frame *fr = (frame *)parent;
  char *lpszTemp;
  int charw = 0;
  int lastcharwidth = 0;
  intrect bbrect;
  intpoint RotatedLeftTop, RotatedRightTop, RotatedRightBottom, RotatedLeftBottom;
  int MinX, MaxX, MinY, MaxY;
  int linewidth = 0, maxlinewidth = 0;
  int cascend;
  graph* gra = (graph *)(((frame *)parent)->parent);
  time_t timeval;
  bool iskanji = false;

  //pattern
  intrect PatternRect;
  int DataNo = 0;
  double LengthRate = 3.0 * legendlength;  //pattertn lenght / text height
  
  std::string str;
  
  dx = 0; dy = 0;
  
  // select initial font
  lf = fo;
  
  // save initial font
  lf0 = fo;

  // select textcolor
  if (rect == 0 && textwidth == 0)
    psetforecolor(flag, f, fontcolor.red, fontcolor.green, fontcolor.blue);

  // set initial font
  if (rect == 0 && textwidth == 0)
    ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
  h = getascend(&lf0, false);
  cascend = h;

  
  //get text ascent
  h = getascend(&lf0, false);
  Tabspace = h * 2;
  SmallTabspace = h;
  
  CX = (int)fr->doubleworld_to_view_x(p1.x);
  CY = (int)fr->doubleworld_to_view_y(p1.y);
  CX0 = CX;
  CY0 = CY;
  maxCX = CX;
  RotatedPoint.x = CX0;
  RotatedPoint.y = CY0;

  
  lpszTemp = (char *)text->c_str();
  str = std::string("");
  //Insert text settings
  while (*lpszTemp != 0)
    {
      switch (*lpszTemp)
	{
	case '#':
	  switch(*(lpszTemp + 1))
	    {
	    case 'H': // graph filename
	      str1 = std::string(gra->graphname);
	      toplainfilename((char *)str1.c_str(), &str2);
	      str += str2;
	      lpszTemp += 2;
	      break;
	    case 'M': // date (month:day:year)
	      (void)time(&timeval);
	      str += std::string(ctime(&timeval));
	      lpszTemp += 2;
	      break;

	    default:
	      lpszTemp1 = nextpoint(lpszTemp);
	      if ((strlen(lpszTemp)-strlen(lpszTemp1))==2)
		{
		  temp3[0] = *lpszTemp;
		  temp3[1] = *(lpszTemp + 1);
		  temp3[2] = 0;
		}
	      else
		{
		  temp3[0] = *lpszTemp;
		  temp3[1] = 0;
		}
	      str += std::string(temp3);
	      lpszTemp = nextpoint(lpszTemp);
	      break;
	    }
	  break;
	default:
	  lpszTemp1 = nextpoint(lpszTemp);
	  if ((strlen(lpszTemp)-strlen(lpszTemp1))==2)
	    {
	      temp3[0] = *lpszTemp;
	      temp3[1] = *(lpszTemp + 1);
	      temp3[2] = 0;
	    }
	  else
	    {
	      temp3[0] = *lpszTemp;
	      temp3[1] = 0;
	    }
	  str += std::string(temp3);
	  lpszTemp = nextpoint(lpszTemp);
	  break;
	}
    }
  i = 0;
  lpszTemp = (char *)str.c_str();
  while (*lpszTemp != 0)
    {
      i++;
      tdx = 0;
      
      // setfont
      switch (*lpszTemp)
	{
	case '#':
	  switch(*(lpszTemp + 1))
	    {
		case 'R':
	      switch(lf.face)
		{
		case FONT_TIMESROMAN:
		case FONT_TIMESBOLD:
		case FONT_TIMESITALIC:
		case FONT_TIMESBOLDITALIC:
		  lf.face = FONT_TIMESROMAN;
		  break;
		    case FONT_HELVETICA:
		case FONT_HELVETICABOLD:
		case FONT_HELVETICAOBLIQUE:
		case FONT_HELVETICABOLDOBLIQUE:
		  lf.face = FONT_HELVETICA;
		  break;
		case FONT_COURIER:
		case FONT_COURIERBOLD:
		case FONT_COURIEROBLIQUE:
		case FONT_COURIERBOLDOBLIQUE:
		  lf.face = FONT_COURIER;
		  break;
		}
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	    case 'B':
	      switch(lf.face)
		{
		case FONT_TIMESROMAN:
		case FONT_TIMESBOLD:
		case FONT_TIMESITALIC:
		case FONT_TIMESBOLDITALIC:
		  lf.face = FONT_TIMESBOLD;
		  break;
		case FONT_HELVETICA:
		case FONT_HELVETICABOLD:
		case FONT_HELVETICAOBLIQUE:
		case FONT_HELVETICABOLDOBLIQUE:
		  lf.face = FONT_HELVETICABOLD;
		  break;
		case FONT_COURIER:
		case FONT_COURIERBOLD:
		case FONT_COURIEROBLIQUE:
		case FONT_COURIERBOLDOBLIQUE:
		  lf.face = FONT_COURIERBOLD;
		  break;
		}
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	    case 'I':
	      switch(lf.face)
		{
		case FONT_TIMESROMAN:
		case FONT_TIMESBOLD:
		case FONT_TIMESITALIC:
		case FONT_TIMESBOLDITALIC:
		  lf.face = FONT_TIMESITALIC;
		  break;
		case FONT_HELVETICA:
		case FONT_HELVETICABOLD:
		case FONT_HELVETICAOBLIQUE:
		case FONT_HELVETICABOLDOBLIQUE:
		  lf.face = FONT_HELVETICAOBLIQUE;
		  break;
		case FONT_COURIER:
		case FONT_COURIERBOLD:
		case FONT_COURIEROBLIQUE:
		case FONT_COURIERBOLDOBLIQUE:
		  lf.face = FONT_COURIEROBLIQUE;
		  break;
		}
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	    case 'O':
	      switch(lf.face)
		{
		case FONT_TIMESROMAN:
		case FONT_TIMESBOLD:
		case FONT_TIMESITALIC:
		case FONT_TIMESBOLDITALIC:
		  lf.face = FONT_TIMESBOLDITALIC;
		  break;
		case FONT_HELVETICA:
		case FONT_HELVETICABOLD:
		case FONT_HELVETICAOBLIQUE:
		case FONT_HELVETICABOLDOBLIQUE:
		  lf.face = FONT_HELVETICABOLDOBLIQUE;
		  break;
		case FONT_COURIER:
		case FONT_COURIERBOLD:
		case FONT_COURIEROBLIQUE:
		case FONT_COURIERBOLDOBLIQUE:
		  lf.face = FONT_COURIERBOLDOBLIQUE;
		  break;
		}
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	    case 'S':
	      if (isdigit(*(lpszTemp + 2)) != 0 && isdigit(*(lpszTemp+3)) != 0 && isdigit(*(lpszTemp+4)) != 0)
		{
		  strncpy(temp3,lpszTemp + 2,3);
		  temp3[3] = 0;
		  itemp = atoi(temp3);
		  
		  lf.size = (int)(((double)lf0.size) * ((double)itemp) / 100.0);

		  if (rect == 0 && textwidth == 0)
		    ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
		  cascend = getascend(&lf, true);
		}
	      break;
	    case 'G':
	      lf.face = FONT_SYMBOL;
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	    case 'F':
	      lf.face = lf0.face;
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      break;
	      //legend
	    case 'p':
	      if (isdigit(*(lpszTemp + 2)) != 0 && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  strncpy(temp3, lpszTemp+2, 3);
		  temp3[3] = 0;
		  DataNo = atoi(temp3);
		}
	      break;
	    }
	  break;
	case '^':
	case '_':
	  lf.size = (int)(((double)lf0.size) * 2.0 / 3.0);
	  if (rect == 0 && textwidth == 0)
	    ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	  cascend = getascend(&lf, true);
	  break;
	case '|':
	  lf.size = lf0.size;
	  if (rect == 0 && textwidth == 0)
	    ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	  cascend = getascend(&lf, true);
	  break;
	}


      if (maykanji(*lpszTemp))
	{
	  if (!iskanji)
	    {
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.jpface, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, false);
	      iskanji = true;
	    }
	}
      else
	{
	  if (iskanji)
	    {
	      if (rect == 0 && textwidth == 0)
		ploadfont(flag, f, lf.face, fr->scale(lf.size), lf.angle);
	      cascend = getascend(&lf, true);
	      iskanji = false;
	    }
	}

      
      // set position
      switch (*lpszTemp)
	{
	case '^':
	  dy = h / 2;
	  break;
	case '_':
	  dy = -h / 4;
	  break;
	case '|':
	  dx = 0;
	  dy = 0;
	  break;
	case '#':
	  switch(*(lpszTemp + 1))
	    {
	    case 'P':
	      if ((*(lpszTemp + 2) == '+' || *(lpszTemp + 2) == '-') 
		  && isdigit(*(lpszTemp + 3))!= 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  strncpy(temp3, lpszTemp + 2, 3);
		  temp3[3] = '\0';
		  itemp = atoi(temp3);
		  
		  dx = (int)(((double)h) * ((double)itemp) / 100.0);
		}
	      break;
	    case 'Y':
	      if ((*(lpszTemp + 2) == '+' || *(lpszTemp + 2) == '-') 
		  && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  strncpy(temp3, lpszTemp + 2, 3);
		  temp3[3] = 0;
		  itemp = atoi(temp3);
		  
		  dy = (int)(((double)h) * ((double)itemp) / 100.0);
		}
	      break;
	    case 'X':
	      if ((*(lpszTemp + 2) == '+' || *(lpszTemp + 2) == '-') 
		  && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  strncpy(temp3, lpszTemp + 2, 3);
		  temp3[3] = 0;
		  itemp = atoi(temp3);
		  
		  tdx = (int)(((double)h) * ((double)itemp) / 100.0);
		}
	      break;
	    case 'T':
	      CX = ((CX - CX0) / Tabspace+1) * Tabspace + CX0;
	      break;
	    case 't':
	      CX = ((CX - CX0) / SmallTabspace + 1) * SmallTabspace + CX0;
	      break;
	    }
	  break;
	}
      
      // escape sequence
      if (*lpszTemp == '\\')
	{
	  switch(*(lpszTemp + 1))
	    {
	    case 'n':
	      KaigyoNo++;
	      CX = CX0;
	      charw = 0;
	      linewidth = 0;
	      break;
	    case 'b':
	      charw = -lastcharwidth;
	      linewidth -= lastcharwidth;
	      break;
	    }
	}
      

      // draw text
      switch (*lpszTemp)
	{
	case  '#':
	  switch(*(lpszTemp + 1))
	    {
	    case 'R':
	    case 'B':
	    case 'I':
	    case 'O':
	      lpszTemp = lpszTemp + 2;
	      charw = 0;
	      i++;
	      break;
	    case 'S':
	      if (isdigit(*(lpszTemp + 2)) != 0 && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  lpszTemp = lpszTemp + 5;
		  i += 4;
		}
	      else
		lpszTemp++;
	      charw = 0;
	      break;
	    case 'G':
	      lpszTemp = lpszTemp + 2;
	      i++;
	      charw = 0;
	      break;
	    case 'F':
	      lpszTemp = lpszTemp + 2;
	      i++;
	      charw = 0;
	      break;
	    case 'P':
	    case 'Y':
	      if ((*(lpszTemp + 2) == '+' || *(lpszTemp + 2) == '-') 
		  && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  lpszTemp = lpszTemp + 5;
		  i += 4;
		}
	      else
		lpszTemp++;
	      charw = 0;
	      break;
	    case 'X':
	      if ((*(lpszTemp + 2) == '+' || *(lpszTemp + 2) == '-')
		  && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  lpszTemp = lpszTemp + 5;
		  i += 4;
		}
	      else
		lpszTemp++;
	      charw = 0;
	      break;
	    case 't':
	    case 'T':
	      lpszTemp = lpszTemp + 2;
	      i++;
	      charw = 0;
	      break;
	    case '#':
	      charw = getcharwidth(&lf, '#', false);
	      if (rect == 0 && textwidth == 0)
		pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, "#", fr->scale(lf.size), charw, lf.angle, cascend);
	      lpszTemp = lpszTemp + 2;
	      lastcharwidth = charw;
	      linewidth += charw;
	      if (maxlinewidth < linewidth)
		maxlinewidth = linewidth;
	      i++;
	      break;
	    case '|':
	      charw = getcharwidth(&lf, '|', false);
	      if (rect == 0 && textwidth == 0)
		pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, "|", fr->scale(lf.size), charw, lf.angle, cascend);
	      lpszTemp = lpszTemp + 2;
	      lastcharwidth = charw;
	      linewidth += charw;
	      if (maxlinewidth < linewidth)
		maxlinewidth = linewidth;
	      i++;
	      break;
	    case '^':
	      charw = getcharwidth(&lf, '^', false);
	      if (rect == 0 && textwidth == 0)
		pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, "^", fr->scale(lf.size), charw, lf.angle, cascend);
	      lpszTemp = lpszTemp + 2;
	      lastcharwidth = charw;
	      linewidth += charw;
	      if (maxlinewidth < linewidth)
		maxlinewidth = linewidth;
	      i++;
	      break;
	    case '_':
	      if (rect == 0 && textwidth == 0)
		pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, "_", fr->scale(lf.size), charw, lf.angle, cascend);
	      lpszTemp = lpszTemp + 2;
	      charw = getcharwidth(&lf, '_', false);
	      lastcharwidth = charw;
	      linewidth += charw;
	      if (maxlinewidth < linewidth)
		maxlinewidth = linewidth;
	      i++;
	      break;
	      //legend pattern
	    case 'p':
	      if (isdigit(*(lpszTemp + 2)) != 0 && isdigit(*(lpszTemp + 3)) != 0 && isdigit(*(lpszTemp + 4)) != 0)
		{
		  lpszTemp = lpszTemp + 5;
		  i += 4;

		  drawpattern(flag, f, CX, CY + (fr->scale(lf0.size) - h)
			      , (int)(LengthRate * fr->scale(lf0.size) / 100.0), -fr->scale(lf0.size), DataNo);

		  charw = (int)(LengthRate * fr->scale(lf0.size) / 100.0);
		  lastcharwidth = charw;
		  linewidth += charw;
		  if (maxlinewidth < linewidth)
		    maxlinewidth = linewidth;

		}
	      else
		{
		  lpszTemp++;
		  charw = 0;
		}
	      break;
	    default:
	      lpszTemp++;
	      charw = 0;
	    }
	  
	  break;
	  
	case '\\':
	  if (*(lpszTemp + 1) == 'n')
	      lpszTemp += 2;
	  else if (*(lpszTemp + 1) == 'b')
	      lpszTemp += 2;
	  else if (*(lpszTemp + 1) == '\\')
	    {
	      temp3[0] = '\\';
	      temp3[1] = 0;
	      charw = getcharwidth(&lf, (unsigned int)*temp3, false);
	      if (rect == 0 && textwidth == 0)
		pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, temp3, fr->scale(lf.size), charw, lf.angle, cascend);
	      lastcharwidth = charw;
	      linewidth += charw;
	      if (maxlinewidth < linewidth)
		maxlinewidth = linewidth;
	      lpszTemp += 2;
	    }
	  else if (isdigit(*(lpszTemp + 1)) != 0)
	    {
	      if (isdigit(*(lpszTemp + 2)) != 0)
		{
		  if (isdigit(*(lpszTemp + 3)) != 0)
		    {
		      ustemp[0] = (*(lpszTemp + 1) - '0') * 64 + (*(lpszTemp + 2) - '0') * 8 + (*(lpszTemp + 3) - '0');
		      ustemp[1] = 0;
		      charw = getcharwidth(&lf, (unsigned int)*ustemp, false);
		      if (rect == 0 && textwidth == 0)
			pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, (const char *)ustemp, fr->scale(lf.size), charw, lf.angle, cascend);
		      lastcharwidth = charw;
		      linewidth += charw;
		      if (maxlinewidth < linewidth)
			maxlinewidth = linewidth;
		      lpszTemp += 4;
		    }
		  else
		    lpszTemp += 3;
		}
	      else
		lpszTemp += 2;
	    }
	  else
	    lpszTemp++;

	  break;
	case '^':
	case '_':
	case '|':
	case 10 :
	case 13 :
	  lpszTemp = lpszTemp + 1;
	  charw = 0;
	  break;
	default:
	  // dwaw text
	  lpszTemp1 = nextpoint(lpszTemp);
	  
	  if ((strlen(lpszTemp)-strlen(lpszTemp1)) == 2)
	    {
	      temp3[0] = *lpszTemp;
	      temp3[1] = *(lpszTemp + 1);
	      temp3[2] = 0;
	      charw = getcharwidth(&lf, (unsigned int)*temp3, true);
	    }
	  else
	    {
	      temp3[0] = *lpszTemp;
	      temp3[1] = 0;
	      charw = getcharwidth(&lf, (unsigned int)*temp3, false);
	    }

	  if (rect == 0 && textwidth == 0)
	    pdrawtext(flag, f, RotatedPoint.x, RotatedPoint.y, temp3, fr->scale(lf.size), charw, lf.angle, cascend);
	  lastcharwidth = charw;
	  linewidth += charw;
	  if (maxlinewidth < linewidth)
	    maxlinewidth = linewidth;
	  lpszTemp = nextpoint(lpszTemp);
	  break;
	}
      
      // update character position
      CY = (int)fr->doubleworld_to_view_y(p1.y) - dy;
      CX = CX + charw + dx + tdx;
      CY = CY + (int)(((double)lf0.size) * ((double)KaigyoNo) * ((double)linespacing));
      RotatedPoint = RotatePoint(intpoint(CX, CY), intpoint(CX0, CY0), (double)lf.angle);
      if (CX > maxCX)
	maxCX = CX;
    }

  //return bounding box
  if (rect != 0)
    {
      RotatedLeftTop = RotatePoint(intpoint(CX0, CY0 - h), intpoint(CX0, CY0), (double)lf.angle);
      RotatedRightTop = RotatePoint(intpoint(maxCX, CY0 - h), intpoint(CX0, CY0), (double)lf.angle);
      RotatedRightBottom = RotatePoint(intpoint(maxCX, CY + fr->scale(lf0.size) - h),intpoint(CX0, CY0), (double)lf.angle);
      RotatedLeftBottom = RotatePoint(intpoint(CX0, CY + fr->scale(lf0.size) - h), intpoint(CX0, CY0), (double)lf.angle);

      //MinX
      if (RotatedLeftTop.x <= RotatedRightTop.x)
	MinX = RotatedLeftTop.x;
      else
	MinX = RotatedRightTop.x;
      
      if (RotatedRightBottom.x <= MinX)
	MinX = RotatedRightBottom.x;
      
      if (RotatedLeftBottom.x <= MinX)
	MinX = RotatedLeftBottom.x;
      
      //MaxX
      if (RotatedLeftTop.x >= RotatedRightTop.x)
	MaxX = RotatedLeftTop.x;
      else
	MaxX = RotatedRightTop.x;
      
      if (RotatedRightBottom.x >= MaxX)
	MaxX = RotatedRightBottom.x;
      
      if (RotatedLeftBottom.x >= MaxX)
	MaxX = RotatedLeftBottom.x;
      
      //MinY
      if (RotatedLeftTop.y <= RotatedRightTop.y)
	MinY = RotatedLeftTop.y;
      else
	MinY = RotatedRightTop.y;
      
      if (RotatedRightBottom.y <= MinY)
	MinY = RotatedRightBottom.y;
      
      if (RotatedLeftBottom.y <= MinY)
	MinY = RotatedLeftBottom.y;
      
      //MaxY
      if (RotatedLeftTop.y >= RotatedRightTop.y)
	MaxY = RotatedLeftTop.y;
      else
	MaxY = RotatedRightTop.y;
      
      if (RotatedRightBottom.y >= MaxY)
	MaxY = RotatedRightBottom.y;
      
      if (RotatedLeftBottom.y >= MaxY)
	MaxY = RotatedLeftBottom.y;
      
      bbrect.left       = MinX;
      bbrect.right      = MaxX;
      bbrect.top        = MinY;
      bbrect.bottom     = MaxY;

      *rect = bbrect;
    }
  else if (textwidth != 0)
      *textwidth = CX - CX0;
}

int parts::gettextwidth()
{
  int width;
  drawtext(true, 0, 0, &width);
  return width;
}

void parts::drawpattern(int flag, FILE* f, int x, int y, int w, int h, int dataid)
{
  int rad;
  int recrad;
  int trirad;
  int dotrad;
  int datano = -1;
  int cx, cy;
  intpointarray *ia;
  doublepointarray *dpa;

  frame *fr = (frame *)parent;
  data *dat = 0;

  // search data object
  for (int i = 0; i < fr->darray->getarraylength(); i++)
    {
      dat = &(*fr->darray)[i];
      if (dat->id == dataid)
	{
	  datano = i;
	  break;
	}
    }
  if (datano == -1)
    return;

  rad = fr->scale(dat->markersize);
  recrad = rad * 14 / 20;
  trirad = rad * 17 / 20;
  dotrad = 8;
  cx = x + w / 2;
  cy = y + h / 2;

  // line
  dpa = new doublepointarray;
  dpa->add(new doublepoint(x, y + h / 2));
  dpa->add(new doublepoint(x + w, y + h / 2));
  plinestyle(flag, f, fr->scale(dat->linestyle.width), dat->linestyle.cap, dat->linestyle.join, dat->linestyle.style);
  psetforecolor(flag, f, dat->linecolor.red, dat->linecolor.green, dat->linecolor.blue);
  pbeginpath(flag, f); 
  StyleLinePath(flag, f, dpa, dpa->getarraylength() - 1,
		dat->linestyle.style, fr->scale(dat->linestyle.width),
		false, 0, 0);
  pstroke(flag, f); 
  delete dpa;

  // marker
  ia = new intpointarray;
  psetforecolor(flag, f, dat->markeredgecolor.red, dat->markeredgecolor.green, dat->markeredgecolor.blue);
  psetbackcolor(flag, f, dat->markerbodycolor.red, dat->markerbodycolor.green, dat->markerbodycolor.blue);
  plinestyle(flag, f, fr->scale(dat->markeredgestyle.width),
	     dat->markeredgestyle.cap, dat->markeredgestyle.join, dat->markeredgestyle.style);
  pfillstyle(flag, f, dat->markerbodyfillstyle.fillrule, dat->markerbodyfillstyle.style);
  switch (dat->markerstyle)
    {
      //NULL
    case MARKER_NULL:
      break;
      //rectangle
    case MARKER_SQUARE:
      ia->add(new intpoint(cx - recrad, cy - recrad));
      ia->add(new intpoint(cx + recrad, cy - recrad));
      ia->add(new intpoint(cx + recrad, cy + recrad));
      ia->add(new intpoint(cx - recrad, cy + recrad));
      ppolygon(flag, f, ia);
      break;
      //rev.triangle
    case MARKER_REVTRIANGLE:
      ia->add(new intpoint(cx         , cy + rad));
      ia->add(new intpoint(cx + trirad, cy - rad / 2));
      ia->add(new intpoint(cx - trirad, cy - rad / 2));
      ppolygon(flag, f, ia);
      break;
      
      //triangle
    case MARKER_TRIANGLE:
      ia->add(new intpoint(cx         , cy - rad));
      ia->add(new intpoint(cx + trirad, cy + rad / 2));
      ia->add(new intpoint(cx - trirad, cy + rad / 2));
      ppolygon(flag, f, ia);
      break;
      
      //diamond
    case MARKER_DIAMOND:
      ia->add(new intpoint(cx         , cy - rad));
      ia->add(new intpoint(cx +    rad, cy));
      ia->add(new intpoint(cx         , cy + rad));
      ia->add(new intpoint(cx -    rad, cy));
      ppolygon(flag, f, ia);
      break;
      
      //circle
    case MARKER_CIRCLE:
      pellipse(flag, f, cx - rad, cy - rad, cx + rad, cy + rad);
      break;
      //cross
    case MARKER_CROSS:
      pbeginpath(flag, f);
      pmoveto(flag, f, cx - rad, cy);
      plineto(flag, f, cx + rad, cy),
      pmoveto(flag, f, cx      , cy - rad);
      plineto(flag, f, cx      , cy + rad);
      pstroke(flag, f);
      break;
      //diag. cross
    case MARKER_DCROSS:
      pbeginpath(flag, f);
      pmoveto(flag, f, cx - rad, cy - rad);
      plineto(flag, f, cx + rad, cy + rad),
      pmoveto(flag, f, cx - rad, cy + rad);
      plineto(flag, f, cx + rad, cy - rad);
      pstroke(flag, f);
      break;
    case MARKER_VAREA:
    case MARKER_HAREA:
    case MARKER_VBAR:
    case MARKER_HBAR:
    case MARKER_VSTEP:
    case MARKER_HSTEP:
    case MARKER_VBAR2:
    case MARKER_HBAR2:
      ia->add(new intpoint(x, y + h / 4));
      ia->add(new intpoint(x + w, y + h / 4));
      ia->add(new intpoint(x + w, y + 3* h / 4));
      ia->add(new intpoint(x, y + 3 * h / 4));
      ppolygon(flag, f, ia);
      break;
    }

  
  //dot
  if (dat->markercenterdot && dat->markerstyle)
      pellipse(flag, f, cx - dotrad, cy - dotrad, cx + dotrad, cy + dotrad);

  delete ia;

  psetforecolor(flag, f, fontcolor.red, fontcolor.green, fontcolor.blue);
}




void parts::drawarrow(int flag, FILE* f, doublepoint *start, doublepoint *end)
{
  double fa,fb,fc;
  double sx, sy, ex, ey;
  intpointarray *PArray;
  intpoint *TempPoint;
  frame *fr = (frame *)parent;
  double alength = (double)fr->scale(arrowlength);

  PArray = new intpointarray;

  plinestyle(flag, f, 0, 0, 0, 0);
  pfillstyle(flag, f, 0, 0);
  psetforecolor(flag, f, linecolor.red, linecolor.green, linecolor.blue);
  psetbackcolor(flag, f, linecolor.red, linecolor.green, linecolor.blue);
  
  sx = start->x;
  sy = start->y;
  ex = end->x;
  ey = end->y;
  
  fa = (double)(ex - sx);
  fa = fa * fa;
  fb = (double)(ey - sy);
  fb = fb * fb;
  fa = fa + fb;
  fa = sqrt(fa);
  
  if ((int)fa != 0)
    {
      if((ex - sx) > 0)
	fa = asin((double)(ey - sy) / fa);
      else
	fa = M_PI - asin((double)(ey - sy) / fa);
      
      TempPoint = new intpoint;
      TempPoint->x = (int)sx;
      TempPoint->y = (int)sy;
      PArray->add(TempPoint);
      
      fb = alength * cos(fa + arrowhalfangle / 180.0 * M_PI);
      fc = alength * sin(fa + arrowhalfangle / 180.0 * M_PI);

      TempPoint = new intpoint;
      TempPoint->x = (int)(sx + fb);
      TempPoint->y = (int)(sy + fc);
      PArray->add(TempPoint);
      
      fb = alength * cos(fa - arrowhalfangle / 180.0 * M_PI);
      fc = alength * sin(fa - arrowhalfangle / 180.0 * M_PI);

      TempPoint = new intpoint;
      TempPoint->x = (int)(sx + fb);
      TempPoint->y = (int)(sy + fc);
      PArray->add(TempPoint);

      ppolygon(flag, f, PArray);
    }
  delete PArray;


}


void parts::drawparts(int flag, FILE* f)
{
//  intpoint temp[50];
  int sx, sy, ex, ey;
  int asx = 0, asy = 0, aex = 0, aey = 0;
  double len, projected_arrowlen = 0.0, cost = 0.0, sint = 0.0;
  int i, no, interporatedno;
  doublearray *InX = 0, *InY = 0, *OutX = 0, *OutY = 0;
  intpointarray *parray = 0;
  intpoint *temppoint;
  
  doublepointarray *PArray;
  doublepoint *TempPoint;
  doublepoint tdp1, tdp2;
  frame *fr = (frame *)parent;

  sx = fr->wtov_x(p1.x);
  sy = fr->wtov_y(p1.y);
  ex = fr->wtov_x(p2.x);
  ey = fr->wtov_y(p2.y);
  
  plinestyle(flag, f, fr->scale(linestyle.width),
	     linestyle.cap, linestyle.join, linestyle.style);
  pfillstyle(flag, f, bodystyle.fillrule, bodystyle.style);
  psetforecolor(flag, f, linecolor.red, linecolor.green, linecolor.blue);
  psetbackcolor(flag, f, bodycolor.red, bodycolor.green, bodycolor.blue);
  
  //draw parts
  switch(type)
    {
    case PARTS_LINE:
      
      len = (p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y); len = sqrt(len); 
      if (len == 0.0)
	return;

      if (p1_arrow || p2_arrow)
	{
	  projected_arrowlen = arrowlength * cos(arrowhalfangle / 180.0 * M_PI);
	  if (len != 0.0)
	    {
	      cost = (p2.x - p1.x) / len;
	      sint = (p2.y - p1.y) / len;
	    }
	}
      if (len != 0.0)
	{
	  if (p1_arrow)
	    {
	      asx = (int)(p1.x + projected_arrowlen * cost);
	      asy = (int)(p1.y + projected_arrowlen * sint);
	      asx = fr->wtov_x(asx);
	      asy = fr->wtov_y(asy);
	    }
	  else
	    {
	      asx = sx;
	      asy = sy;
	    }

	  if (p2_arrow)
	    {
	      aex = (int)(p1.x + (len - projected_arrowlen) * cost);
	      aey = (int)(p1.y + (len - projected_arrowlen) * sint);
	      aex = fr->wtov_x(aex);
	      aey = fr->wtov_y(aey);
	    }
	  else
	    {
	      aex = ex;
	      aey = ey;
	    }
	}

      PArray = new doublepointarray;
      PArray->add(new doublepoint(asx, asy));
      PArray->add(new doublepoint(aex, aey));
      
      pbeginpath(flag, f); 
      StyleLinePath(flag, f, PArray, PArray->getarraylength() - 1,
		    linestyle.style, linestyle.width,
		    false, 0, 0);
      pstroke(flag, f); 
      delete PArray;
      
      if (len != 0)
	{
	  if(p1_arrow)
	    {
	      tdp1 = doublepoint(sx, sy);
	      tdp2 = doublepoint(asx, asy);
	      drawarrow(flag, f, 
			&tdp1,
			&tdp2);
	    }

	  if(p2_arrow)
	    {
	      tdp1 = doublepoint(ex, ey);
	      tdp2 = doublepoint(aex, aey);
	      drawarrow(flag, f,
		      &tdp1,
		      &tdp2);
	    }
	}

      break;
      
    case PARTS_RECTANGLE:
      parray = new intpointarray;
      parray->add(new intpoint (sx, sy));
      parray->add(new intpoint (ex, sy));
      parray->add(new intpoint (ex, ey));
      parray->add(new intpoint (sx, ey));
      ppolygon(flag, f, parray);
      delete parray;
      break;
    case PARTS_ELLIPSE:
      pellipse(flag, f, sx, sy, ex, ey);
      break;
    case PARTS_CUT:
      plinestyle(flag, f, fr->scale(linestyle.width),
		 linestyle.cap, linestyle.join, LINESTYLENULL);
      parray = new intpointarray;
      parray->add(new intpoint (sx, sy + (ey - sy) * 3 / 4));
      parray->add(new intpoint (sx + (ex - sx) * 3 / 4, sy));
      parray->add(new intpoint (ex, sy + (ey - sy) / 4));
      parray->add(new intpoint (sx + (ex - sx) / 4, ey));
      ppolygon(flag, f, parray);
      delete parray;

      plinestyle(flag, f, fr->scale(linestyle.width),
		 linestyle.cap, linestyle.join, linestyle.style);
      pbeginpath(flag, f); 
      pmoveto(flag, f, sx, sy + (ey - sy) * 3 / 4);
      plineto(flag, f, sx + (ex - sx) * 3 / 4, sy);
      pmoveto(flag, f, sx + (ex - sx) / 4, ey);
      plineto(flag, f, ex, sy + (ey - sy) / 4);
      pstroke(flag, f); 
      break;
    case PARTS_POLYLINE:
      no = nodes->getarraylength() - 1;
      if (no >= 1)
	{
	  if (interpolate == PARTS_INTER_POLYLINE || interpolate == PARTS_INTER_CLOSEDPOLYLINE)
	    {
	      if (interpolate == PARTS_INTER_POLYLINE)
		{
		  PArray = new doublepointarray;
		  for (i = 0; i <= no; i++)
		    {
		      TempPoint = new doublepoint;
		      TempPoint->x = fr->wtov_x((*nodes)[i].x + p1.x);
		      TempPoint->y = fr->wtov_y((*nodes)[i].y + p1.y);
		      PArray->add(TempPoint);
		    }
		  
		  pbeginpath(flag, f); 
		  StyleLinePath(flag, f, PArray, PArray->getarraylength() - 1,
			       linestyle.style, linestyle.width,
			       false, 0, 0);
		  pstroke(flag, f); 
		  delete PArray;
		  
		}
	      //closed polyline
	      else
		{
		  parray = new intpointarray;
		  for (i = 0; i <= no; i++)
		    {
		      temppoint = new intpoint(fr->wtov_x((*nodes)[i].x + p1.x), fr->wtov_y((*nodes)[i].y + p1.y));
		      parray->add(temppoint);
		    }
		  temppoint = new intpoint(fr->wtov_x((*nodes)[0].x + p1.x), fr->wtov_y((*nodes)[0].y + p1.y));
		  parray->add(temppoint);
		  
		  ppolygon(flag, f, parray);
		  delete parray;

		}
	    }
	  else if (interpolate == PARTS_INTER_PSPLINE || interpolate == PARTS_INTER_CLOSEDPSPLINE)
	    {
	      InX = new doublearray;
	      if (InX == 0)
		{
		  InX = 0;
		  goto j1;
		}
	      
	      InY = new doublearray;
	      if (InY == 0)
		{
		  InY = 0;
		  goto j1;
		}
	      
 	      OutX = new doublearray;
	      if (OutX == 0)
		{
		  OutX = 0;
		  goto j1;
		}
	      
 	      OutY = new doublearray;
	      if (OutY == 0)
		{
		  OutY = 0;
		  goto j1;
		}
	      
	      for (i = 0; i <= no; i++)
		{
		  InX->add((*nodes)[i].x + p1.x);
		  InY->add((*nodes)[i].y + p1.y);
		}
	      
	      if (interpolate == PARTS_INTER_PSPLINE)
		CalcPSpline(InX, InY, OutX, OutY, 1, interpolatediv);
	      else
		CalcPSpline(InX, InY, OutX, OutY, 2, interpolatediv);
	      
	      interporatedno = OutX->getarraylength() - 1;
	      
	      if (interpolate == PARTS_INTER_PSPLINE)
		{
		  PArray = new doublepointarray;
		  for (i = 0; i <= interporatedno; i++)
		    PArray->add(new doublepoint(fr->wtov_x((int)(*OutX)[i]), fr->wtov_y((int)(*OutY)[i])));
		  
		  pbeginpath(flag, f); 
		  StyleLinePath(flag, f, PArray, PArray->getarraylength() - 1,
				linestyle.style, linestyle.width,
				false, 0, 0);
		  pstroke(flag, f); 
		  delete PArray;
		}
	      //close pspline
	      else
		{
		  parray = new intpointarray;
		  for (i = 0; i <= interporatedno; i++)
		    {
		      temppoint = new intpoint(fr->wtov_x((int)(*OutX)[i]), fr->wtov_y((int)(*OutY)[i]));
		      parray->add(temppoint);
		    }
		  plinestyle(flag, f, fr->scale(linestyle.width),
			     linestyle.cap, linestyle.join, LINESTYLENULL);
		  ppolygon(flag, f, parray);
		  delete parray;
		  
		  PArray = new doublepointarray;
		  for (i = 0; i <= interporatedno; i++)
		    PArray->add(new doublepoint(fr->wtov_x((int)(*OutX)[i]), fr->wtov_y((int)(*OutY)[i])));
		  
		  plinestyle(flag, f, fr->scale(linestyle.width),
			     linestyle.cap, linestyle.join, linestyle.style);
		  pbeginpath(flag, f); 
		  StyleLinePath(flag, f, PArray, PArray->getarraylength() - 1,
				linestyle.style, linestyle.width,
				false, 0, 0);
		  pstroke(flag, f); 
		  delete PArray;
		}
	      
	    j1:
	      if (InX != 0)
		delete InX;
	      if (InY != 0)
		delete InY;
	      if (OutX != 0)
		delete OutX;
	      if (OutY != 0)
		delete OutY;
	    }
	}
      break;
    }
}

bool parts::readparts(FILE *fp)
{
  char stemp[10000];
  std::string str;
  char *p;
  int len, x, y;
  int readversion;
  
//  virgin = false;

  // version
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      readversion = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // id
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      id = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // type
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (nextitem(stemp, &str) != 0)
      type = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // text
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    if (gettxt(stemp, &str))
      *text = str;
    else
      return false;
  }

  // fo
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // face
    p = nextitem(p, &str);
    if (p != 0)
      fo.face = atoi(str.c_str());
    else
      return false;

    if (readversion >= 1)
      {
	// jpface
	p = nextitem(p, &str);
	if (p != 0)
	  fo.jpface = atoi(str.c_str());
	else
	  return false;
      }

    // size
    p = nextitem(p, &str);
    if (p != 0)
      fo.size = atoi(str.c_str());
    else
      return false;
    // angle
    p = nextitem(p, &str);
    if (p != 0)
      fo.angle = atof(str.c_str());
    else
      return false;
  }
  else
    return false;

  // fontcolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      fontcolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // linespacing
    p = nextitem(p, &str);
    if (p != 0)
      linespacing = atof(str.c_str());
    else
      return false;
  }
  else
    return false;

  // linecolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      linecolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // bodycolor
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // red
    p = nextitem(p, &str);
    if (p != 0)
      bodycolor.red = atoi(str.c_str());
    else
      return false;
    // green
    p = nextitem(p, &str);
    if (p != 0)
      bodycolor.green = atoi(str.c_str());
    else
      return false;
    // blue
    p = nextitem(p, &str);
    if (p != 0)
      bodycolor.blue = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // linestyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // width
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.width = atoi(str.c_str());
    else
      return false;
    // cap
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.cap = atoi(str.c_str());
    else
      return false;
    // join
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.join = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      linestyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // bodystyle
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // fillrule
    p = nextitem(p, &str);
    if (p != 0)
      bodystyle.fillrule = atoi(str.c_str());
    else
      return false;
    // style
    p = nextitem(p, &str);
    if (p != 0)
      bodystyle.style = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // legendlength
    p = nextitem(p, &str);
    if (p != 0)
      legendlength = atoi(str.c_str());
    else
      return false;
    // interpolate
    p = nextitem(p, &str);
    if (p != 0)
      interpolate = atoi(str.c_str());
    else
      return false;
    // interpolatediv
    p = nextitem(p, &str);
    if (p != 0)
      interpolatediv = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // p1_arrow
    p = nextitem(p, &str);
    if (p != 0)
      p1_arrow = atoi(str.c_str());
    else
      return false;
    // p2_arrow
    p = nextitem(p, &str);
    if (p != 0)
      p2_arrow = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  // p1, p2
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // p1.x
    p = nextitem(p, &str);
    if (p != 0)
      p1.x = atoi(str.c_str());
    else
      return false;
    // p1.y
    p = nextitem(p, &str);
    if (p != 0)
      p1.y = atoi(str.c_str());
    else
      return false;
    // p2.x
    p = nextitem(p, &str);
    if (p != 0)
      p2.x = atoi(str.c_str());
    else
      return false;
    // p2.y
    p = nextitem(p, &str);
    if (p != 0)
      p2.y = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    // arrowlength
    p = nextitem(p, &str);
    if (p != 0)
      arrowlength = atoi(str.c_str());
    else
      return false;
    // arrowhalfangle
    p = nextitem(p, &str);
    if (p != 0)
      arrowhalfangle = atof(str.c_str());
    else
      return false;
  }
  else
    return false;


  // nodes
  if (getnextline(stemp, sizeof(stemp), fp))
  {
    p = stemp;
    p = nextitem(p, &str);
    if (p != 0)
      len = atoi(str.c_str());
    else
      return false;
  }
  else
    return false;

  nodes->setbuff(0);
  for (int i = 0; i < len; i++)
    {
      if (getnextline(stemp, sizeof(stemp), fp))
	{
	  // x
	  p = stemp;
	  p = nextitem(p, &str);
	  if (p != 0)
	    x = atoi(str.c_str());
	  else
	    return false;
	  // y
	  p = nextitem(p, &str);
	  if (p != 0)
	    y = atoi(str.c_str());
	  else
	    return false;
	}
      else
	return false;

      nodes->add(new intpoint(x ,y));
    }

  return true;
}

bool parts::writeparts(FILE *fp)
{
  fprintf(fp, "[parts]\n");

  // text
  fprintf(fp, "%d\n", version);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d\n", type);
  fprintf(fp, "\"%s\"\n", text->c_str());
  fprintf(fp, "%d %d %d %.15g\n", fo.face, fo.jpface, fo.size, fo.angle);
  fprintf(fp, "%u %u %u\n", fontcolor.red, fontcolor.green, fontcolor.blue);
  fprintf(fp, "%.15g\n", linespacing);

  // parts
  fprintf(fp, "%u %u %u\n", linecolor.red, linecolor.green, linecolor.blue);
  fprintf(fp, "%u %u %u\n", bodycolor.red, bodycolor.green, bodycolor.blue);
  fprintf(fp, "%d %d %d %d\n", linestyle.width, linestyle.cap, linestyle.join, linestyle.style);
  fprintf(fp, "%d %d\n", bodystyle.fillrule, bodystyle.style);
  fprintf(fp, "%d %d %d\n", legendlength, interpolate, interpolatediv);
  fprintf(fp, "%d %d\n", p1_arrow, p2_arrow);
  fprintf(fp, "%d %d %d %d\n", p1.x, p1.y, p2.x, p2.y);
  fprintf(fp, "%d %.15g\n", arrowlength, arrowhalfangle);
  fprintf(fp, "%d\n", nodes->getarraylength());
  for (int i = 0; i < nodes->getarraylength(); i++)
    fprintf(fp, "%d %d\n", (*nodes)[i].x, (*nodes)[i].y);

  return true;
}




