/*----------------------------------------------------------------------------
                      parts.h (definitions of parts object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__parts_h)
#define __parts_h

#include <string>
#include "plobj.h"
#include "vectdefs.h"
#include "col.h"
#include "pointdefs.h"

#define  PARTS_TEXT               0
#define  PARTS_LINE               1
#define  PARTS_POLYLINE           2
#define  PARTS_RECTANGLE          3
#define  PARTS_ELLIPSE            4
#define  PARTS_CUT                5

#define PARTS_INTER_POLYLINE           0
#define PARTS_INTER_CLOSEDPOLYLINE     1
#define PARTS_INTER_PSPLINE            3
#define PARTS_INTER_CLOSEDPSPLINE      4



extern bool getpartsname(int id, std::string *out);
extern bool gettextfacename(int id, std::string *out);
extern bool gettextjpfacename(int id, std::string *out);


class parts:public plobj
{
 public:
  parts();
  ~parts();
  parts& operator = (const parts& pr);

  bool get(std::string* member,  tokenbuff* outvalue);  
  bool set(std::string* member,  std::string* setvalue);  
  bool exec(std::string* function, buffarray *argument, tokenbuff* outvalue);  

  bool paint(bool flag,  FILE *f);
  void drawtext(int flag, FILE *f, intrect *rect, int *textwidth);
  void drawparts(int flag, FILE *f);
  void drawarrow(int flag,  FILE *f, doublepoint *start, doublepoint *end);
  void drawpattern(int flag,  FILE *f, int x, int y, int w, int h, int dataid);
  int getcharwidth(Capfont *lf, unsigned int c, bool iskanji);
  int getascend(Capfont *lf, bool iskanji);
  void getBBbox(intrect *r);
  int gettextwidth();
  bool readparts(FILE *fp);
  bool writeparts(FILE *fp);

  int id;
  std::string *label;
  int selected;
  int type;

  /* font */
  std::string *text;
  Capfont fo;
  Capcolor fontcolor;
  double linespacing;
  
  /* common */
  Capcolor linecolor, bodycolor;
  Caplinestyle linestyle;
  Capfillstyle bodystyle;
  int legendlength;
  int interpolate;
  int interpolatediv;
  bool p1_arrow, p2_arrow;
  intpoint p1, p2;
  int arrowlength;
  double arrowhalfangle;

  intpointarray *nodes;

};


#endif



