/*----------------------------------------------------------------------------
                          paths.cc (handling pathname)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

char *tofilename(char *in)
{
  char *p, *slash;
  char buff[1000];
  strcpy(buff, in);
  p = buff;
  slash = p;
  bool flag = false;
  while(*p != 0)
    {
      if (*p == '/')
	{
	  slash = p;
	  flag = true;
	}
      p++;
    }
  if (flag)
    return slash + 1;
  else
    return (slash = buff);
}

bool isabsolutepath(char *path)
{
  char *p = path;
  while(*p != 0)
    {
      if (*p != ' ')
	{
	  if (*p == '/')
	    return true;
	  else
	    return false;
	}
      p++;
    }
  return false;
}

char *toabsolutepath(char *file)
{
  char cpath[1024];
  char *ret;
  getcwd(cpath, sizeof(cpath));

  if (isabsolutepath(file))
    return file;
  else
    {
      strcat(cpath, "/");
      strcat(cpath, file);
      return (ret = cpath);
    }
}



bool abspath(const char *src, const char *ref, char *outpath)
{
  char *p, *q;
  char buffer[1024], stemp[3], tmpout[1024];

  // src is not relative path
  if (*src == '/')
    return false;
  // ref is not absolute path
  if (*ref != '/')
    return false;

  // extract path from ref
  p = (char *)ref + strlen(ref);
  while(*p != '/')
    p--;
  q = (char *)ref;
  strcpy(tmpout, "");
  while(q <= p)
    {
      stemp[0] = *q; stemp[1] = 0;
      strcat(tmpout, stemp);
      q++;
    }

  p = (char *)src;
  strcpy(buffer, "");

  while (*p != 0)
    {
      if(*p == '/')
	{
	  // up
	  if (strcmp(buffer, "..") == 0)
	    {
	      q = tmpout + strlen(tmpout);
	      if (strlen(tmpout) <= 1)
		return false;
	      q--;q--;
	      while(*q != '/')
		{
		  *q = 0;
		  q--;
		}
	    }
	  // stay
	  else if (strcmp(buffer, ".") == 0)
	    {
	    }
	  // down
	  else
	    {
	      strcat(tmpout, buffer);
	      strcat(tmpout, "/");
	    }
	  strcpy(buffer, "");
	}
      else
	{
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buffer, stemp);
	}
      p++;
    }
  strcpy(outpath, tmpout);
  strcat(outpath, buffer);
  return true;
}

bool pathmatch(const char *data, const char *graph)
{
  if (strlen(data) < strlen(graph))
    return 0;

  for (int i = 0; i < (int)strlen(graph); i++)
    if (graph[i] != data[i])
      return false;
  return true;
}

char *relpath(const char *src, const char *ref)
{
  char buff[1024], tmpout[10000], stemp[3];
  char *p, *q , *out;

  strcpy(tmpout, "");
  p = (char *)src + strlen(src);
  
  //remove srcpath
  while((*p != '/') && p > src)
    {
      p--;
    }
  
  while (p >= src)
    {
      q = (char *)src;
      strcpy(buff, "");
      while(q <= p)
	{
	  stemp[0] = *q; stemp[1] = 0;
	  strcat(buff, stemp);
	  q++;
	}
      
      if (pathmatch(ref, buff))
	{
	  strcat(tmpout, ref + strlen(buff));
	  return (out = tmpout);
	}
      p--;
      while((*p != '/') && p > src)
	{
	  p--;
	}
      strcat(tmpout, "../");
    }
  return (char *)ref;
}


