/*----------------------------------------------------------------------------
                        pipecall.cc (pipe call routine)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include "pipecall.h"
#include "convtxt.h"

extern void message(const char *msg);

int detectpipedirec(std::string *str, std::string *out)
{
  int len = strlen(str->c_str());
  std::string str1;
  
  if (len == 0)
    return ERROR;
  
  // start char
  switch (str->c_str()[0])
    {
    case '<':
      str1 = str->substr(1);
      *out = str1;
      return READFROMFILE;
      break;
    case '>':
      // >>
      if (str->c_str()[1] == '>')
	{
	  str1 = str->substr(2);
	  *out = str1;
	  return APPENDTOFILE;
	}
      // >
      else
	{
	  str1 = str->substr(1);
	  *out = str1;
	  return WRITETOFILE;
	}	
      break;
    case '+':
      if (str->c_str()[1] == '>')
	{
	  if (strlen(str->c_str()) >= 3)
	    {
	      // +>>
	      if (str->c_str()[2] == '>')
		{
		  str1 = str->substr(2);
		  *out = str1;
		  return READWRITEEND;
		}
	      // +>
	      else
		{
		  str1 = str->substr(2);
		  *out = str1;
		  return READWRITETRUNCATE;
		}
	    }
	  // +>
	  else
	    {
	      str1 = str->substr(2);
	      *out = str1;
	      return READWRITETRUNCATE;
	    }
	}
      // +<
      else if (str->c_str()[1] == '<')
	{
	  str1 = str->substr(2);
	  *out = str1;
	  return READWRITE;
	}
      else
	return ERROR;
      break;
    case '|':
      str1 = str->substr(1);
      *out = str1;
      return WRITETOPIPE;
      break;
    default:
      switch(str->c_str()[strlen(str->c_str()) - 1])
	{
	case '|':
	  str1 = str->substr(0, (int)strlen(str->c_str()) - 1);
	  *out = str1;
	  return READFROMPIPE;
	  break;
	default:
	  str1 = *str;
	  *out = str1;
	  return READFROMFILE;
	  break;
	}
      break;
    }
}



FILE *PipeOpen(std::string *str, int *type)
{
  int opentype;
  std::string out;
  FILE *fp_rw = 0;

  opentype = detectpipedirec(str, &out);
  if (opentype == ERROR)
    return NULL;

  switch(opentype)
    {
    case READFROMFILE:
      fp_rw = fopen(out.c_str(), "r");
      break;
    case WRITETOFILE:
      fp_rw = fopen(out.c_str(), "w");
      break;
    case APPENDTOFILE:
      fp_rw = fopen(out.c_str(), "a");
      break;
    case READWRITE:
      fp_rw = fopen(out.c_str(), "r+");
      break;
    case READWRITETRUNCATE:
      fp_rw = fopen(out.c_str(), "w+");
      break;
    case READWRITEEND:
      fp_rw = fopen(out.c_str(), "a+");
      break;
    case READFROMPIPE:
      fp_rw = popen(out.c_str(), "r");
      if (fp_rw == (FILE *)NULL || fp_rw == (FILE *)(-1))
	fp_rw = NULL;
      break;
    case WRITETOPIPE:
      fp_rw = popen(out.c_str(), "w");
      if (fp_rw == (FILE *)NULL || fp_rw == (FILE *)(-1))
 				fp_rw = NULL;

      break;
    }
  *type = opentype;
  return fp_rw;
}

bool PipeClose(FILE *h, int *type)
{

  switch(*type)
    {
    case READFROMFILE:
    case WRITETOFILE:
    case APPENDTOFILE:
      fclose(h);
      break;
    case READFROMPIPE:
    case WRITETOPIPE:
      pclose(h);
      break;
    default:
      return false;
    }

  return true;
}

bool ReadLine(FILE *h, std::string *buf)
{
  char buffer[10000];
  if (h == NULL)
    return false;
  
  if (topazfgets(buffer, sizeof(buffer), h) == NULL)
    return false;
  else
    *buf = std::string(buffer);

  return true;
}

bool WriteLine(FILE *h, std::string *buf)
{
  if (h == NULL)
    return false;
  
  fprintf(h, "%s\n", buf->c_str());
  return true;
}

bool ReadAllLines(FILE *h, buffarray *barray)
{
  char buffer[10000];
  if (h == NULL)
    return false;
  
  barray->setbuff(0);
  while (topazfgets(buffer, sizeof(buffer), h) != NULL)
      barray->add(buffer);

  return true;
}

bool WriteAllLines(FILE *h, buffarray *barray)
{
  if (h == NULL)
    return false;
  
  for (int i = 0; i < barray->getarraylength(); i++)
      fprintf(h, "%s\n", (*barray)[i].c_str());

  return true;
}


