/*----------------------------------------------------------------------------
                         plotdata.cc (plotting routine)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "data.h"
#include "frame.h"
#include "axis.h"
#include "vectdefs.h"
#include "pointdefs.h"
#include "gdi.h"
#include "spline.h"
#include "topazvalues.h"
#include "col.h"

extern void StyleLinePath(bool flag, FILE* f, doublepointarray *array, int no_of_point, int style, int width,
			  bool Clip, intrect *AxisRect, intrect *boundrect);
extern void ClipLine(bool flag, doublepoint *p1, doublepoint *p2, intrect *AxisRect, bool clip, FILE* f);
extern int CalcFormulastring(char *s, double *n, int InitFormula, data* dat, int cellno, int seriesno, int Isminmax);
extern bool isabort();
			  
void data::drawmark(int flag, FILE* f, int x, int y, intpointarray *ia)
{
  int rad;
  int recrad;             //ա  rad*1.4/2
  int trirad;             //ѷ  rad*1.7/2
  int dotrad;             //dotȾ

  frame *fr = (frame *)parent;
  rad = fr->scale(markersize);
  
  if (rad == 0)
    rad = 1;
  
  recrad = rad * 14 / 20;
  trirad = rad * 17 / 20;
  dotrad = 8;


  // body
  switch (markerstyle)
    {
      //NULL
    case MARKER_NULL:
      break;
      //rectangle
    case MARKER_SQUARE:
      (*ia)[0] = intpoint(x - recrad, y - recrad);
      (*ia)[1] = intpoint(x + recrad, y - recrad);
      (*ia)[2] = intpoint(x + recrad, y + recrad);
      (*ia)[3] = intpoint(x - recrad, y + recrad);
      ppolygon(flag, f, ia);
      break;
      //rev.triangle
    case MARKER_REVTRIANGLE:
      (*ia)[0] = intpoint(x         , y + rad);
      (*ia)[1] = intpoint(x + trirad, y - rad / 2);
      (*ia)[2] = intpoint(x - trirad, y - rad / 2);
      ppolygon(flag, f, ia);
      break;
      
      //triangle
    case MARKER_TRIANGLE:
      (*ia)[0] = intpoint(x         , y - rad);
      (*ia)[1] = intpoint(x + trirad, y + rad / 2);
      (*ia)[2] = intpoint(x - trirad, y + rad / 2);
      ppolygon(flag, f, ia);
      break;
      
      //diamond
    case MARKER_DIAMOND:
      (*ia)[0] = intpoint(x         , y - rad);
      (*ia)[1] = intpoint(x +    rad, y);
      (*ia)[2] = intpoint(x         , y + rad);
      (*ia)[3] = intpoint(x -    rad, y);
      ppolygon(flag, f, ia);
      break;
      
      //circle
    case MARKER_CIRCLE:
      pellipse(flag, f, x - rad, y - rad, x + rad, y + rad);
      break;
      //cross
    case MARKER_CROSS:
      pbeginpath(flag, f);
      pmoveto(flag, f, x - rad, y);
      plineto(flag, f, x + rad, y),
      pmoveto(flag, f, x      , y - rad);
      plineto(flag, f, x      , y + rad);
      pstroke(flag, f);
      break;
      //diag. cross
    case MARKER_DCROSS:
      pbeginpath(flag, f);
      pmoveto(flag, f, x - rad, y - rad);
      plineto(flag, f, x + rad, y + rad),
      pmoveto(flag, f, x - rad, y + rad);
      plineto(flag, f, x + rad, y - rad);
      pstroke(flag, f);
      break;
    }

  
  //dot
  if (markercenterdot && markerstyle)
      pellipse(flag, f, x - dotrad, y - dotrad, x + dotrad, y + dotrad);

}


void data::drawplot(int flag, FILE *f)
{
  int NumPoint, NumBufferPoint, NumInterpolatedPoint = 0;
  int i, j;
  earray *BufferX = 0, *BufferY = 0, *BufferMinX = 0, *BufferMaxX = 0, *BufferMinY = 0, *BufferMaxY = 0;
  doublearray *WorldX = 0, *WorldY = 0, *InterpolatedX = 0, *InterpolatedY = 0;
  intpointarray *AreaPoint = 0;
  doublepointarray *InterpolatedPoint = 0;
  intpoint *TempPoint;
  doublepoint *TempDoublePoint;
  intpoint *TempPoint1 = 0, *TempPoint2 = 0;
  doublepoint TempDoublePoint1, TempDoublePoint2;
  double fa;
  intrect  AxisRect, BarRect, ClippedRect;
  std::string temp_pre_formula;
  double TempMinX, TempMaxX, TempMinY, TempMaxY;
  intrect TempRect;
  dataelement tempelement;
  int buffsize = 1000;
  frame *fr = (frame *)parent;
  axis *ax;
  double ftemp;
  int plotstatus;
  intpointarray *markerbuff;
  int refx, refy;


  if (!display)
    return;

  if ((refx = fr->getaxisindex(refaxis_x)) == -1)
    return;
  if ((refy = fr->getaxisindex(refaxis_y)) == -1)
    return;


  if (refx >= fr->axarray->getarraylength() || refx < 0 ||
      refy >= fr->axarray->getarraylength() || refy < 0)
    return;

  
  AxisRect.left   = fr->wtov_x(0);
  AxisRect.right  = fr->wtov_x(fr->p2.x - fr->p1.x);
  AxisRect.top    = fr->wtov_y(0);
  AxisRect.bottom = fr->wtov_y(fr->p2.y - fr->p1.y);
  
  //calc data number
  NumPoint = x->getarraylength() - 1;
  
  BufferX = new earray;
  if (BufferX == 0)
    goto j1;
  if (!BufferX->setbuff(NumPoint + 1))
    goto j1;

  BufferY = new earray;
  if (BufferY == 0)
    goto j1;
  if (!BufferY->setbuff(NumPoint + 1))
    goto j1;

  BufferMinX = new earray;
  if (BufferMinX == 0)
    goto j1;
  if (!BufferMinX->setbuff(buffsize))
    goto j1;

  BufferMaxX = new earray;
  if (BufferMaxX == 0)
    goto j1;
  if (!BufferMaxX->setbuff(buffsize))
    goto j1;

  BufferMinY = new earray;
  if (BufferMinY == 0)
    goto j1;
  if (!BufferMinY->setbuff(buffsize))
    goto j1;

  BufferMaxY = new earray;
  if (BufferMaxY == 0)
    goto j1;
  if (!BufferMaxY->setbuff(buffsize))
    goto j1;


  WorldX = new doublearray;
  if (WorldX == 0)
    goto j1;
  if (!WorldX->setbuff(NumPoint + 1))
    goto j1;

  WorldY = new doublearray;
  if (WorldY == 0)
    goto j1;
  if (!WorldY->setbuff(NumPoint + 1))
    goto j1;

  InterpolatedX = new doublearray;
  if (InterpolatedX == 0)
    goto j1;
  if (!InterpolatedX->setbuff(NumPoint + 1))
    goto j1;

  InterpolatedY = new doublearray;
  if (InterpolatedY == 0)
    goto j1;
  if (!InterpolatedY->setbuff(NumPoint + 1))
    goto j1;

  InterpolatedPoint = new doublepointarray;
  if (InterpolatedPoint == 0)
    goto j1;
  if (!InterpolatedPoint->setbuff(NumPoint + 1))
    goto j1;

  AreaPoint = new intpointarray;
  if (AreaPoint == 0)
    goto j1;
  if (!AreaPoint->setbuff(buffsize))
    goto j1;
  
  //calc script
  //x-data
  if (trans_x)
    temp_pre_formula =  *format_x;
  else
    temp_pre_formula = std::string("x");
  
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 0);
  for (i = 0;i <= NumPoint; i++)
    {
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 0);
      if (!BufferX->add(new dataelement(fa, (*x)[i].status)))
	goto j1;
    }
  
  //y-data
  if (trans_y)
    temp_pre_formula = *format_y;
  else
    temp_pre_formula = std::string("y");
  
  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 0);
  for (i = 0;i <= NumPoint; i++)
    {
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 0);
      if (!BufferY->add(new dataelement(fa, (*y)[i].status)))
	goto j1;
    }
  
  if (bar_x)
    {
      //xmin-data
      if (trans_x)
	temp_pre_formula =  *format_x;
      else
	temp_pre_formula = std::string("x");
      
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 1);
      for (i = 0;i <= NumPoint; i++)
	{
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 1);
	  if (!BufferMinX->add(new dataelement(fa, (*xmin)[i].status)))
	    goto j1;
	}
      
      //xmax-data
      if (trans_x)
	temp_pre_formula =  *format_x;
      else
	temp_pre_formula = std::string("x");
      
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 2);
      for (i = 0;i <= NumPoint; i++)
	{
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 2);
	  if (!BufferMaxX->add(new dataelement(fa, (*xmax)[i].status)))
	    goto j1;
	}
    }
  
  if (bar_y)
    {
      //ymin-data
      if (trans_y)
	temp_pre_formula = *format_y;
      else
	temp_pre_formula = std::string("y");
      
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 1);
      for (i = 0;i <= NumPoint; i++)
	{
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 1);
	  if (!BufferMinY->add(new dataelement(fa, (*ymin)[i].status)))
	    goto j1;
	}
      
      //ymax-data
      if (trans_y)
	temp_pre_formula =  *format_y;
      else
	temp_pre_formula = std::string("y");
      
      CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, true ,this, 0, fr->getdataindex(id), 2);
      for (i = 0;i <= NumPoint; i++)
	{
	  CalcFormulastring((char *)temp_pre_formula.c_str(), &fa, false ,this, i, fr->getdataindex(id), 2);
	  if (!BufferMaxY->add(new dataelement(fa, (*ymax)[i].status)))
	    goto j1;
	}
    }
  
  for (i = 0;i <= NumPoint + 1; i++)
    {
      if (i <= NumPoint) 
	plotstatus = (*BufferX)[i].status | (*BufferY)[i].status;
      else
	plotstatus = 0;

      //numerical data -> stack
      if (  plotstatus == 0 && i <= NumPoint)
	{
	  //Virtual to calculated  and calculated to World
	  
	  ax = &(*fr->axarray)[refx];
	  ftemp = (*BufferX)[i].data;
	  ftemp = ax->virtual_to_doubleworld(ftemp).x;
	  ftemp = fr->doubleworld_to_view_x(ftemp);
	  if (!WorldX->add(ftemp))
	    goto j1;

	  ax = &(*fr->axarray)[refy];
	  ftemp = (*BufferY)[i].data;
	  ftemp = ax->virtual_to_doubleworld(ftemp).y;
	  ftemp = fr->doubleworld_to_view_y(ftemp);
	  if (!WorldY->add(ftemp))
	    goto j1;

	}
      //invalid data ->draw
      else if ((plotstatus & 2) != 0 || i > NumPoint)
	{
	  NumBufferPoint = WorldX->getarraylength() - 1;
	  
	  //world data to interpolated data
	  switch(interpolate)
	    {
	    case INTER_POLYLINE:
	      NumInterpolatedPoint = NumBufferPoint;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*WorldX)[j];
		  TempDoublePoint->y = (*WorldY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;
	    case INTER_CLOSEDPOLYLINE:
	      NumInterpolatedPoint = NumBufferPoint;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*WorldX)[j];
		  TempDoublePoint->y = (*WorldY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      TempDoublePoint = new doublepoint;
	      TempDoublePoint->x = (*WorldX)[0];
	      TempDoublePoint->y = (*WorldY)[0];
	      if (!InterpolatedPoint->add(TempDoublePoint))
		goto j1;
	      NumInterpolatedPoint++;
	      break;
	    case INTER_SPLINE:
	      CalcSpline(WorldX, WorldY, InterpolatedX, InterpolatedY, interpolatediv);
	      NumInterpolatedPoint = InterpolatedX->getarraylength() - 1;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*InterpolatedX)[j];
		  TempDoublePoint->y = (*InterpolatedY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;
	    case INTER_PSPLINE:
	      CalcPSpline(WorldX, WorldY, InterpolatedX, InterpolatedY, 1, interpolatediv);
	      NumInterpolatedPoint = InterpolatedX->getarraylength() - 1;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*InterpolatedX)[j];
		  TempDoublePoint->y = (*InterpolatedY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;
	    case INTER_CLOSEDPSPLINE:
	      CalcPSpline(WorldX, WorldY, InterpolatedX, InterpolatedY, 2, interpolatediv);
	      NumInterpolatedPoint = InterpolatedX->getarraylength() - 1;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*InterpolatedX)[j];
		  TempDoublePoint->y = (*InterpolatedY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;
	    case INTER_BSPLINE:
	      CalcBSpline(WorldX, WorldY ,InterpolatedX, InterpolatedY, 1, interpolatediv);
	      NumInterpolatedPoint = InterpolatedX->getarraylength() - 1;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*InterpolatedX)[j];
		  TempDoublePoint->y = (*InterpolatedY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;
	    case INTER_CLOSEDBSPLINE:
	      CalcBSpline(WorldX, WorldY ,InterpolatedX, InterpolatedY, 2, interpolatediv);
	      NumInterpolatedPoint = InterpolatedX->getarraylength() - 1;
	      for (j = 0; j <= NumInterpolatedPoint; j++)
		{
		  TempDoublePoint = new doublepoint;
		  TempDoublePoint->x = (*InterpolatedX)[j];
		  TempDoublePoint->y = (*InterpolatedY)[j];
		  if (!InterpolatedPoint->add(TempDoublePoint))
		    goto j1;
		}
	      break;

	    }
	  
	  
	  
	  //------------------ gdi start -----------------------------
	  
	  //-----------  Area -----------------------------------
	  if (markerstyle == MARKER_VAREA || markerstyle == MARKER_HAREA)
	    {
	      //exclude invalid(MAXSHORT) point
	      for (j = 0;j <= NumInterpolatedPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  TempPoint = new intpoint;
		  TempPoint->x = fr->doubleworldtoshortworld((*InterpolatedPoint)[j].x);
		  TempPoint->y = fr->doubleworldtoshortworld((*InterpolatedPoint)[j].y);
		  if (abs(TempPoint->x) < MAXSHORT & abs(TempPoint->y) < MAXSHORT)
		    {
		      
		      if (!AreaPoint->add(TempPoint))
			goto j1;
		    }
		}
	      
	      //additonal point
	      switch (interpolate)//open terminate
		{
		  //open terminate
		case INTER_POLYLINE:
		case INTER_SPLINE:
		case INTER_PSPLINE:
		case INTER_BSPLINE:

		  if (markerstyle == MARKER_VAREA)
		    {
		      //X-coordinate
		      TempPoint1 = new intpoint;
		      TempPoint1->x = (*AreaPoint)[AreaPoint->getarraylength() - 1].x;
		      TempPoint2 = new intpoint;
		      TempPoint2->x = (*AreaPoint)[0].x;
		      //Y-coordinate
		      
		      ax = &(*fr->axarray)[refy];
		      ftemp = basevalue;
		      ftemp = ax->virtual_to_doubleworld(ftemp).y;
		      ftemp = fr->doubleworld_to_view_y(ftemp);
		      
		      TempPoint1->y = fr->doubleworldtoshortworld(ftemp);
		      TempPoint2->y = TempPoint1->y;
		    }
		  else if (markerstyle == MARKER_HAREA)
		    {
		      //Y-coordinate
		      TempPoint1 = new intpoint;
		      TempPoint1->y = (*AreaPoint)[AreaPoint->getarraylength() - 1].y;
		      TempPoint2 = new intpoint;
		      TempPoint2->y = (*AreaPoint)[0].y;
		      //X-coordinate
		      
		      ax = &(*fr->axarray)[refx];
		      ftemp = basevalue;
		      ftemp = ax->virtual_to_doubleworld(ftemp).x;
		      ftemp = fr->doubleworld_to_view_x(ftemp);

		      TempPoint1->x = fr->doubleworldtoshortworld(ftemp);
		      TempPoint2->x = TempPoint1->x;
		    }

		  if (!AreaPoint->add(TempPoint1))
		    goto j1;
		  if (!AreaPoint->add(TempPoint2))
		    goto j1;
		  
		  break;
		  //closed terminate
		case INTER_CLOSEDPOLYLINE:
		case INTER_CLOSEDPSPLINE:
		case INTER_CLOSEDBSPLINE:
		  break;
		}
	      
	      psavegdc(flag, f);
	      //draw area
	      plinestyle(flag, f, fr->scale(markeredgestyle.width),
			 markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
	      psetforecolor(flag, f, markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
	      psetbackcolor(flag, f, markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
	      pfillstyle(flag, f, markerbodyfillstyle.fillrule, markerbodyfillstyle.style);


	      if (clipping)
		psetcliprect(flag, f, &AxisRect);
	      ppolygon(flag, f, AreaPoint);
	      if (clipping)
		pinitcliprect(flag, f);
	      presoregdc(flag, f);

	    }

	  //-----------  Bar -----------------------------------
	  else if (markerstyle == MARKER_VBAR || markerstyle == MARKER_HBAR)
	    {
	      plinestyle(flag, f, fr->scale(markeredgestyle.width),
			 markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
	      psetforecolor(flag, f, markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
	      psetbackcolor(flag, f, markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
	      pfillstyle(flag, f, markerbodyfillstyle.fillrule, markerbodyfillstyle.style);

	      for (j = 0;j < NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  if (markerstyle == MARKER_VBAR)
		    {
		      //X-coordinate
		      BarRect.left  = fr->doubleworldtoshortworld((*WorldX)[j]);
		      BarRect.right = fr->doubleworldtoshortworld((*WorldX)[j + 1]);
		      
		      //Y-coordinate
		      ax = &(*fr->axarray)[refy];
		      fa = basevalue;
		      fa = ax->virtual_to_doubleworld(fa).y;
		      fa = fr->doubleworld_to_view_y(fa);
		      
		      BarRect.top = (int)fa;
		      BarRect.bottom = fr->doubleworldtoshortworld((*WorldY)[j]);
		      BarRect.normalized();
		    }
		  else if (markerstyle == MARKER_HBAR)
		    {
		      //Y-coordinate
		      BarRect.top 	  = fr->doubleworldtoshortworld((*WorldY)[j]);
		      BarRect.bottom  = fr->doubleworldtoshortworld((*WorldY)[j + 1]);
		      
		      //X-coordinate
		      ax = &(*fr->axarray)[refx];
		      fa = basevalue;
		      fa = ax->virtual_to_doubleworld(fa).x;
		      fa = fr->doubleworld_to_view_x(fa);
		  
		      BarRect.left = (int)fa;
		      BarRect.right = fr->doubleworldtoshortworld((*WorldX)[j]);
		      BarRect.normalized();
		    }
		  
		  if (clipping)
		    ClippedRect = AxisRect & BarRect;
		  else
		    ClippedRect = BarRect;
		  
		  if (!ClippedRect.isempty())
		    {
		      AreaPoint->setbuff(0);
		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.left;
		      TempPoint->y = ClippedRect.top;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.right;
		      TempPoint->y = ClippedRect.top;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.right;
		      TempPoint->y = ClippedRect.bottom;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.left;
		      TempPoint->y = ClippedRect.bottom;
		      AreaPoint->add(TempPoint);
		      
		      ppolygon(flag, f, AreaPoint);
		    }
		}
	    }
	  //-----------  Step -----------------------------------
	  else if (markerstyle == MARKER_VSTEP || markerstyle == MARKER_HSTEP)
	    {

	      // inner area
	      plinestyle(flag, f, fr->scale(markeredgestyle.width),
			 markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
	      psetforecolor(flag, f, markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
	      psetbackcolor(flag, f, markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
	      pfillstyle(flag, f, markerbodyfillstyle.fillrule, markerbodyfillstyle.style);

	      if (markerstyle == MARKER_VSTEP)
		{
		  //Y-coordinate
		  ax = &(*fr->axarray)[refy];
		  fa = basevalue;
		  fa = ax->virtual_to_doubleworld(fa).y;
		  fa = fr->doubleworld_to_view_y(fa);
		  
		  for (j = 0;j < NumBufferPoint; j++)
		    {
		      if (isabort())
			goto j1;

		      TempPoint = new intpoint;
		      TempPoint->x = (int)(*WorldX)[j];
		      TempPoint->y = (int)(*WorldY)[j];
		      AreaPoint->add(TempPoint);
		      
		      TempPoint = new intpoint;
		      TempPoint->x = (int)(*WorldX)[j + 1];
		      TempPoint->y = (int)(*WorldY)[j];
		      AreaPoint->add(TempPoint);
		    }
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)(*WorldX)[NumBufferPoint];
		  TempPoint->y = (int)fa;
		  AreaPoint->add(TempPoint);
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)(*WorldX)[0];
		  TempPoint->y = (int)fa;
		  AreaPoint->add(TempPoint);
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)(*WorldX)[0];
		  TempPoint->y = (int)(*WorldY)[0];
		  AreaPoint->add(TempPoint);
		}
	      else if (markerstyle == MARKER_HSTEP)
		{
		  //X-coordinate
		  ax = &(*fr->axarray)[refx];
		  fa = basevalue;
		  fa = ax->virtual_to_doubleworld(fa).x;
		  fa = fr->doubleworld_to_view_x(fa);
		  
		  for (j = 0;j < NumBufferPoint; j++)
		    {
		      TempPoint = new intpoint;
		      TempPoint->x = (int)(*WorldX)[j];
		      TempPoint->y = (int)(*WorldY)[j];
		      AreaPoint->add(TempPoint);
		      
		      TempPoint = new intpoint;
		      TempPoint->x = (int)(*WorldX)[j];
		      TempPoint->y = (int)(*WorldY)[j + 1];
		      AreaPoint->add(TempPoint);
		    }
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)fa;
		  TempPoint->y = (int)(*WorldY)[NumBufferPoint];
		  AreaPoint->add(TempPoint);
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)fa;
		  TempPoint->y = (int)(*WorldY)[0];
		  AreaPoint->add(TempPoint);
		  
		  TempPoint = new intpoint;
		  TempPoint->x = (int)(*WorldX)[0];
		  TempPoint->y = (int)(*WorldY)[0];
		  AreaPoint->add(TempPoint);
		}

	      psavegdc(flag, f);
	      if (clipping)
		psetcliprect(flag, f, &AxisRect);
	      ppolygon(flag, f, AreaPoint);
	      if (clipping)
		pinitcliprect(flag, f);
	      presoregdc(flag, f);
	    }
	  //----------- Bar2 -----------------------------------
	  else if (markerstyle == MARKER_VBAR2 || markerstyle == MARKER_HBAR2)
	    {
	      plinestyle(flag, f, fr->scale(markeredgestyle.width),
			 markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
	      psetforecolor(flag, f, markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
	      psetbackcolor(flag, f, markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
	      pfillstyle(flag, f, markerbodyfillstyle.fillrule, markerbodyfillstyle.style);
	      
	      for (j = 0;j <= NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  if (markerstyle == MARKER_VBAR2)
		    {
		      //X-coordinate
		      BarRect.left  = fr->doubleworldtoshortworld((*WorldX)[j] - markersize / 2);
		      BarRect.right = fr->doubleworldtoshortworld((*WorldX)[j] + markersize / 2);
		      
		      //Y-coordinate
		      ax = &(*fr->axarray)[refy];
		      fa = basevalue;
		      fa = ax->virtual_to_doubleworld(fa).y;
		      fa = fr->doubleworld_to_view_y(fa);
		      
		      BarRect.top = (int)fa;
		      BarRect.bottom = fr->doubleworldtoshortworld((*WorldY)[j]);
		      BarRect.normalized();
		    }
		  else if (markerstyle == MARKER_HBAR2)
		    {
		      //Y-coordinate
		      BarRect.top     = fr->doubleworldtoshortworld((*WorldY)[j] - markersize / 2);
		      BarRect.bottom  = fr->doubleworldtoshortworld((*WorldY)[j] + markersize / 2);
		      
		      //X-coordinate
		      ax = &(*fr->axarray)[refx];
		      fa = basevalue;
		      fa = ax->virtual_to_doubleworld(fa).x;
		      fa = fr->doubleworld_to_view_x(fa);
		  
		      BarRect.left = (int)fa;
		      BarRect.right = fr->doubleworldtoshortworld((*WorldX)[j]);
		      BarRect.normalized();
		    }
		  
		  if (clipping)
		    ClippedRect = AxisRect & BarRect;
		  else
		    ClippedRect = BarRect;
		  
		  if (!ClippedRect.isempty())
		    {
		      AreaPoint->setbuff(0);
		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.left;
		      TempPoint->y = ClippedRect.top;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.right;
		      TempPoint->y = ClippedRect.top;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.right;
		      TempPoint->y = ClippedRect.bottom;
		      AreaPoint->add(TempPoint);

		      TempPoint = new intpoint;
		      TempPoint->x = ClippedRect.left;
		      TempPoint->y = ClippedRect.bottom;
		      AreaPoint->add(TempPoint);
		      
		      ppolygon(flag, f, AreaPoint);
		    }
		}
	    }
	  //----------- draw line -------------------
	  
	  plinestyle(flag, f, fr->scale(linestyle.width), linestyle.cap, linestyle.join, linestyle.style);
	  psetforecolor(flag, f, linecolor.red, linecolor.green, linecolor.blue);
	  
	  if (clipping)
	    {
	      switch(linestyle.style)
		{
		case LINESTYLENULL:
		  break;
		default:
		  pbeginpath(flag, f); 
		  StyleLinePath(flag, f, InterpolatedPoint, InterpolatedPoint->getarraylength() - 1,
				linestyle.style, linestyle.width,
				true, &AxisRect, &AxisRect);
		  pstroke(flag, f); 
		  break;
		}
	    }
	  else
	    {
	      switch(linestyle.style)
		{
		case LINESTYLENULL:
		  break;
		default:
		  pbeginpath(flag, f); 
		  StyleLinePath(flag, f,InterpolatedPoint, InterpolatedPoint->getarraylength() - 1,
				linestyle.style, linestyle.width,
				false, &AxisRect, &AxisRect);
		  pstroke(flag, f); 
		  break;
		}
	    }
	  
	  
	  //------------------------ Error-bar --------------------------------------
	  plinestyle(flag, f, fr->scale(linestyle.width), linestyle.cap, linestyle.join, LINESTYLESOLID);
	  psetforecolor(flag, f, linecolor.red, linecolor.green, linecolor.blue);

	  if (bar_x)
	    {
	      for (j = 0;j <= NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  ax = &(*fr->axarray)[refx];
		  TempMinX = (*xmin)[j].data;
		  TempMinX = ax->virtual_to_doubleworld(TempMinX).x;
		  TempMinX = fr->doubleworld_to_view_x(TempMinX);

		  TempMaxX = (*xmax)[j].data;
		  TempMaxX = ax->virtual_to_doubleworld(TempMaxX).x;
		  TempMaxX = fr->doubleworld_to_view_x(TempMaxX);
		  
		  TempDoublePoint1.x = TempMinX;
		  TempDoublePoint1.y = (*WorldY)[j];
		  TempDoublePoint2.x = TempMaxX;
		  TempDoublePoint2.y = (*WorldY)[j];
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f);
		    }
		  
		  TempDoublePoint1.x = TempMinX;
		  TempDoublePoint1.y = (*WorldY)[j] - markersize;
		  TempDoublePoint2.x = TempMinX;
		  TempDoublePoint2.y = (*WorldY)[j] + markersize;
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f); 
		    }
		  
		  TempDoublePoint1.x = TempMaxX;
		  TempDoublePoint1.y = (*WorldY)[j] - markersize;
		  TempDoublePoint2.x = TempMaxX;
		  TempDoublePoint2.y = (*WorldY)[j] + markersize;
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f);
		    }
		}
	    }

	  if (bar_y)
	    {
	      for (j = 0;j <= NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  ax = &(*fr->axarray)[refy];
		  TempMinY = (*ymin)[j].data;
		  TempMinY = ax->virtual_to_doubleworld(TempMinY).y;
		  TempMinY = fr->doubleworld_to_view_y(TempMinY);

		  TempMaxY = (*ymax)[j].data;
		  TempMaxY = ax->virtual_to_doubleworld(TempMaxY).y;
		  TempMaxY = fr->doubleworld_to_view_y(TempMaxY);
		  
		  TempDoublePoint1.x = (*WorldX)[j];
		  TempDoublePoint1.y = TempMinY;
		  TempDoublePoint2.x = (*WorldX)[j];
		  TempDoublePoint2.y = TempMaxY;
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f);
		    }
		  
		  TempDoublePoint1.x = (*WorldX)[j] - markersize;
		  TempDoublePoint1.y = TempMinY;
		  TempDoublePoint2.x = (*WorldX)[j] + markersize;
		  TempDoublePoint2.y = TempMinY;
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f); 
		    }
		  
		  TempDoublePoint1.x = (*WorldX)[j] - markersize;
		  TempDoublePoint1.y = TempMaxY;
		  TempDoublePoint2.x = (*WorldX)[j] + markersize;
		  TempDoublePoint2.y = TempMaxY;
		  if (fabs(TempDoublePoint1.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint1.y) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.x) < (double)MAXSHORT
		      && fabs(TempDoublePoint2.y) < (double)MAXSHORT)
		    {
		      pbeginpath(flag, f); 
		      ClipLine(flag, &TempDoublePoint1, &TempDoublePoint2, &AxisRect, clipping, f);
		      pstroke(flag, f);
		    }
		}
	    }
	  
	  //-------------- draw marker ------------------------------
	  
	  
	  markerbuff = new intpointarray;
	  switch (markerstyle)
	    {
	    case MARKER_SQUARE:
	    case MARKER_DIAMOND:
	      markerbuff->setarray(4);
	      break;
	    case MARKER_REVTRIANGLE:
	    case MARKER_TRIANGLE:
	      markerbuff->setarray(3);
	      break;
	    default:
	      break;
	    }
	  

	  psetforecolor(flag, f, markeredgecolor.red, markeredgecolor.green, markeredgecolor.blue);
	  psetbackcolor(flag, f, markerbodycolor.red, markerbodycolor.green, markerbodycolor.blue);
	  plinestyle(flag, f, fr->scale(markeredgestyle.width),
		     markeredgestyle.cap, markeredgestyle.join, markeredgestyle.style);
	  pfillstyle(flag, f, markerbodyfillstyle.fillrule, markerbodyfillstyle.style);
	  if (clipping)
	    {
	      TempRect = AxisRect;
	      for (j = 0;j <= NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  if (fabs((*WorldX)[j]) < MAXSHORT && fabs((*WorldY)[j]) < MAXSHORT)
		    {
		      if (TempRect.contains((int)(*WorldX)[j], (int)(*WorldY)[j]))
			drawmark(flag, f, (int)(*WorldX)[j], (int)(*WorldY)[j], markerbuff);
		    }
		}
	    }
	  else
	    {
	      for (j = 0;j <= NumBufferPoint; j++)
		{
		  if (isabort())
		    goto j1;

		  if (fabs((*WorldX)[j]) < MAXSHORT && fabs((*WorldY)[j]) < MAXSHORT)
		    {
		      drawmark(flag, f, (int)(*WorldX)[j], (int)(*WorldY)[j], markerbuff);
		    }
		}
	    }
	  delete markerbuff;

	  //*********************** gdi end *********************************
	  //datainit
	  WorldX->setbuff(0);
	  WorldY->setbuff(0);
	  InterpolatedX->setbuff(0);
	  InterpolatedY->setbuff(0);
	  InterpolatedPoint->setbuff(0);
	  AreaPoint->setbuff(0);
	}
    }
  
 j1:
  
  if (BufferX != 0)
    delete BufferX;
  if (BufferY != 0)
    delete BufferY;
  if (BufferMinX != 0)
    delete BufferMinX;
  if (BufferMaxX != 0)
    delete BufferMaxX;
  if (BufferMinY != 0)
    delete BufferMinY;
  if (BufferMaxY != 0)
    delete BufferMaxY;
  if (WorldX != 0)
    delete WorldX;
  if (WorldY != 0)
    delete WorldY;
  if (InterpolatedX != 0)
    delete InterpolatedX;
  if (InterpolatedY != 0)
    delete InterpolatedY;
  if (InterpolatedPoint != 0)
    delete InterpolatedPoint;
  if (AreaPoint != 0)
    delete AreaPoint;
  
}

