/*----------------------------------------------------------------------------
                      pspline.cc (parametric spline curve)
                       This file is a part of topaz systems
       This module is based on the codes freely distributed by Mr. Okumura.
                 This module is free from copyright protection.


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include "spline.h"
#include "topazvalues.h"

int Omaketable(doublearray *x, doublearray *y, doublearray *z, int N)
{
  int i;
  double t;
  doublearray *h = 0, *d = 0;
  
  h = new doublearray;
  if (h == 0)
    return false;
  h->setbuff(1000);
  
  d = new doublearray;
  if (d == 0)
    return false;
  d->setbuff(1000);
  
  (*z)[0] = 0;  (*z)[N - 1] = 0;
  for (i = 0; i < N - 1; i++)
    {
      (*h)[i] =  (*x)[i + 1] - (*x)[i];
      if ((*h)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  return false;
	}
      (*d)[i + 1] = ((*y)[i + 1] - (*y)[i]) / (*h)[i];
    }
  (*z)[1] = (*d)[2] - (*d)[1] - (*h)[0] * (*z)[0];
  (*d)[1] = 2 * ((*x)[2] - (*x)[0]);
  for (i = 1; i < N - 2; i++)
    {
      if ((*d)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  return false;
	}
      t = (*h)[i] / (*d)[i];
      (*z)[i + 1] = (*d)[i + 2] - (*d)[i + 1] - (*z)[i] * t;
      (*d)[i + 1] = 2 * ((*x)[i + 2] - (*x)[i]) - (*h)[i] * t;
    }
  (*z)[N - 2] -= (*h)[N - 2] * (*z)[N - 1];
  for (i = N - 2; i > 0; i--)
    {
      if ((*d)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  return false;
	}
      (*z)[i] = ((*z)[i] - (*h)[i] * (*z)[i + 1]) / (*d)[i];
    }
	delete h;
  delete d;
  return true;
  
}

double Ospline(double t, doublearray *x, doublearray *y, doublearray *z , int N)
{
  int i, j, k;
  double d, h;
  
  i = 0;  j = N - 1;
  while (i < j)
  {
    k = (i + j) / 2;
    if ((*x)[k] < t)
      i = k + 1;
    else
      j = k;
  }

  if (i > 0)
    i--;

  h = (*x)[i + 1] - (*x)[i];
  d = t - (*x)[i];

  if (h == 0.0)
    return HUGE_VAL;
  
  return ((((*z)[i + 1] - (*z)[i]) * d / h + (*z)[i] * 3) * d
	  + (((*y)[i + 1] - (*y)[i]) / h
	     - ((*z)[i] * 2 + (*z)[i + 1]) * h)) * d + (*y)[i];
}

int Omaketable2(doublearray *p, doublearray *x, doublearray *y,
		doublearray *a, doublearray *b, int N)
{
  int i;
  
  (*p)[0] = 0.0;
  for (i = 1; i < N; i++)
      (*p)[i] = (*p)[i - 1] + 1.0;

  for (i = 1; i < N; i++)
    {
      if ((*p)[N - 1]==0.0)
	return false;

      (*p)[i] /= (*p)[N - 1];
    }
  if (!Omaketable(p, x, a, N))
    return false;
  if (!Omaketable(p, y, b, N))
    return false;
  
  return true;
}

void Ospline2(double t, double *px, double *py,
		doublearray *p, doublearray *x, doublearray *y,
		doublearray *a, doublearray *b,int N)
{
  *px = Ospline(t, p, x, a, N);
  *py = Ospline(t, p, y, b, N);
}


int CalcPSpline(doublearray *inx, doublearray *iny,
		doublearray *outx, doublearray *outy
		, int contr_kind, int divno)
{
  doublearray *xy[3],*p = 0,*a = 0,*b = 0;
  int ErrorFlag = false;
  int n, normalcount,i,outcount;
  double t, px, py;
  xy[1] = 0 ;xy[2] = 0;

  xy[1] = new doublearray;
  if (xy[1] == 0)
    {
      ErrorFlag = true;
      goto j1;
    }

  xy[2] = new doublearray;
  if (xy[2] == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  
  n = inx->getarraylength() - 1;

  // remove invalid point
  normalcount = 0;
  for (i = 0; i <= n; i++)
    {
      if (fabs((*inx)[i]) < HUGE_VAL	&& fabs((*iny)[i]) < HUGE_VAL)
	{
	  normalcount++;
	  xy[1]->add((*inx)[i]);
	  xy[2]->add((*iny)[i]);
	}
    }
  //too small data number
  n = normalcount;
  if (n == 2)
    {
      outx->add((*xy[1])[0]); outy->add((*xy[2])[0]);
      outx->add((*xy[1])[1]); outy->add((*xy[2])[1]);
      ErrorFlag = false;
      goto j1;
    }
  else if (n == 1)
    {
      outx->add((*xy[1])[0]); outy->add((*xy[2])[0]);
      ErrorFlag = false;
      goto j1;
    }
  else if (n < 1)
    {
      ErrorFlag = true;
      goto j1;
    }


  //when clesed curve, add last point
  if (contr_kind == 2)
    {
      if (!xy[1]->add((*xy[1])[0]))
	{
	  ErrorFlag=true;
	  goto j1;
	}
      if (!xy[2]->add((*xy[2])[0]))
	{
	  ErrorFlag=true;
	  goto j1;
	}
    }

  p = new doublearray;
  if (p == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  p->setbuff(n - 1);

  a = new doublearray;
  if (a == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  a->setbuff(n - 1);

  b = new doublearray;
  if (b == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  b->setbuff(n - 1);

  //in case of open curve
  if (contr_kind == 1)
    {
      if (!Omaketable2(p,  xy[1],  xy[2], a, b, normalcount))
	{
	  ErrorFlag = true;
	  goto j1;
	}
      
      outcount = (normalcount - 1) * divno;
      for (i = 0; i <= outcount; i++)
	{
	  t= 1.0 / (double)outcount * (double)i;
	  
	  Ospline2( t, &px, &py, p, xy[1], xy[2], a,  b, normalcount);
	  if (px == HUGE_VAL || py == HUGE_VAL)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  if (!outx->add(px))
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  if (!outy->add(py))
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	}
    }
  // in case of closed curve
  else if (contr_kind == 2)
    {
      if (!Cmaketable2(p, xy[1],  xy[2], a, b, normalcount))
	{
	  ErrorFlag = true;
	  goto j1;
	}
      
      outcount = (normalcount - 1) * divno;
      for (i = 0; i <= outcount; i++)
	{
	  t = 1.0 / (double)outcount * (double)i;
	  
	  Cspline2( t, &px, &py, p, xy[1], xy[2], a, b, normalcount);
	  if (px == HUGE_VAL || py == HUGE_VAL)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  
	  if (!outx->add(px))
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  if (!outy->add(py))
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	}
    }
  
  
 j1:;
  if (p != 0)
    delete p;
  if (a != 0)
    delete a;
  if (b != 0)
    delete b;
  
  if (xy[1] != 0)
    delete xy[1];
  if (xy[2] != 0)
    delete xy[2];
  
  if (ErrorFlag)
    return false;
  else
    return true;
}


int Cmaketable(doublearray *x, doublearray *y, doublearray *z, int N)
{
  int i;
  double t;
  doublearray *h = 0, *d = 0, *w = 0;
  
  h = new doublearray;
  if (h == 0)
    return false;
  h->setbuff(N);

  d = new doublearray;
  if (d == 0)
    {
      delete h;
      return false;
    }
  d->setbuff(N);

  w = new doublearray;
  if (w == 0)
    {
      delete h;
      delete d;
      return false;
    }
  w->setbuff(N);

  for (i = 0; i < N; i++)
    {
      (*h)[i] = (*x)[i + 1] - (*x)[i];
      if ((*h)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  delete w;
	  return false;
	}
      
      (*w)[i] = ((*y)[i + 1] - (*y)[i]) / (*h)[i];
    }
  (*w)[N] = (*w)[0];
  for (i = 1; i < N; i++)
    (*d)[i] = 2 * ((*x)[i + 1] - (*x)[i - 1]);
  (*d)[N] = 2 * ((*h)[N - 1] + (*h)[0]);

  for (i = 1; i <= N; i++)
    (*z)[i] = (*w)[i] - (*w)[i - 1];
  (*w)[1] = (*h)[0];
  (*w)[N - 1] = (*h)[N - 1];
  (*w)[N] = (*d)[N];

  for (i = 2; i < N - 1; i++)
    (*w)[i] = 0;

  for (i = 1; i < N; i++)
    {
      if ((*d)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  delete w;
	  return false;
	}
      t = (*h)[i] / (*d)[i];
      (*z)[i + 1] = (*z)[i + 1] - (*z)[i] * t;
      (*d)[i + 1] = (*d)[i + 1] - (*h)[i] * t;
      (*w)[i + 1] = (*w)[i + 1] - (*w)[i] * t;
    }
  (*w)[0] = (*w)[N];
  (*z)[0] = (*z)[N];
  for (i = N - 2; i >= 0; i--)
    {
      if ((*h)[i + 1] == 0.0)
	{
	  delete h;
	  delete d;
	  delete w;
	  return false;
	}
      t = (*h)[i] / (*d)[i + 1];
      (*z)[i] = (*z)[i] - (*z)[i + 1] * t;
      (*w)[i] = (*w)[i] - (*w)[i + 1] * t;
    }
  
  if ((*w)[0] == 0.0)
    {
      delete h;
      delete d;
      delete w;
      return false;
    }
  
  t = (*z)[0] / (*w)[0];
  (*z)[0] = t;
  (*z)[N] = t;
  for (i = 1; i < N; i++)
    {
      if ((*d)[i] == 0.0)
	{
	  delete h;
	  delete d;
	  delete w;
	  return false;
	}
      (*z)[i] = ((*z)[i] - (*w)[i] * t) / (*d)[i];
    }
  
  delete h;
  delete d;
  delete w;
  return true;
}

double Cspline(double t, doublearray *x, doublearray *y, doublearray *z,int N)
{
  int i, j, k;
  double d, h, period;
  
  period = (*x)[N] - (*x)[0];
  while (t > (*x)[N])
    t -= period;

  while (t < (*x)[0])
    t += period;
  i = 0;
  j = N;
  while (i < j)
  {
    k = (i + j) / 2;
    if ((*x)[k] < t) i = k + 1;  else j = k;
  }

  if (i > 0)
    i--;

  h = (*x)[i + 1] - (*x)[i];

  if (h == 0.0)
    return HUGE_VAL;
  
  d = t - (*x)[i];
  return ((((*z)[i + 1] - (*z)[i]) * d / h + (*z)[i] * 3) * d
	  + (((*y)[i + 1] - (*y)[i]) / h
	     - ((*z)[i] * 2 + (*z)[i + 1]) * h)) * d + (*y)[i];
}

int Cmaketable2(doublearray *p, doublearray *x, doublearray *y,
		doublearray *a, doublearray *b,int N)
{
  int i;
  
  (*p)[0] = 0;
  for (i = 1; i <= N; i++)
      (*p)[i] = (*p)[i - 1] + 1.0;
  
  if ((*p)[N] == 0.0)
    return false;
  
  for (i = 1; i <= N; i++) (*p)[i] /= (*p)[N];
  
  if (!Cmaketable(p, x, a, N))
    return false;
  if (!Cmaketable(p, y, b, N))
    return false;
  
  return true;
}

void Cspline2(double t, double *px, double *py,
	      doublearray *p, doublearray *x, doublearray * y,
	      doublearray *a, doublearray *b,int N)
{
  *px = Cspline(t, p, x, a, N);
  *py = Cspline(t, p, y, b, N);
}



int CalcSpline(doublearray *inx, doublearray *iny,
		doublearray *outx, doublearray *outy
		, int divno)
{
  doublearray *xy[3],*a = 0, *tx = 0, *ty = 0;
  int ErrorFlag = false;
  int n, normalcount,i,outcount;
  double t, py;
  
  xy[1] = 0; xy[2] = 0;

  xy[1] = new doublearray;
  if (xy[1] == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  xy[1]->setbuff(1000);

  xy[2] = new doublearray;
  if (xy[2] == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  xy[2]->setbuff(1000);

  tx = new doublearray;
  if (tx == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  tx->setbuff(1000);

  ty = new doublearray;
  if (ty == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  ty->setbuff(1000);

  n = inx->getarraylength() - 1;

  // remove invalid points
  normalcount = 0;
  for (i = 0; i <= n; i++)
    {
      if (fabs((*inx)[i]) < HUGE_VAL	&& fabs((*iny)[i]) < HUGE_VAL)
	{
	  normalcount++;
	  tx->add((*inx)[i]);
	  ty->add((*iny)[i]);
	}
    }
  
  // too small data number
  n = normalcount;
  if (n == 2)
    {
      outx->add((*tx)[0]); outy->add((*ty)[0]);
      outx->add((*tx)[1]); outy->add((*ty)[1]);
      ErrorFlag = false;
      goto j1;
    }
  else if (n == 1)
    {
      outx->add((*tx)[0]); outy->add((*ty)[0]);
      ErrorFlag = false;
      goto j1;
    }
  else if (n < 1)
    {
      ErrorFlag = true;
      goto j1;
    }

  // is correct order?
  if ((*tx)[normalcount - 1] - (*tx)[0] > 0.0)
    {
      for (i = 0; i <= normalcount - 2; i++)
	{
	  if ((*tx)[i + 1] - (*tx)[i] <= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  else
	    {
	      xy[1]->add((*tx)[i]);
	      xy[2]->add((*ty)[i]);
	    }
	}
      xy[1]->add((*tx)[normalcount - 1]);
      xy[2]->add((*ty)[normalcount - 1]);
    }
  else if ((*tx)[normalcount - 1] - (*tx)[0] < 0.0)
    {
      for (i = normalcount - 1; i >= 1; i--)
	{
	  if ((*tx)[i] - (*tx)[i - 1] >= 0.0)
	    {
	      ErrorFlag = true;
	      goto j1;
	    }
	  else
	    {
	      xy[1]->add((*tx)[i]);
	      xy[2]->add((*ty)[i]);
	    }
	}
      xy[1]->add((*tx)[0]);
      xy[2]->add((*ty)[0]);
    }
  else
    {
      ErrorFlag = true;
      goto j1;
    }
  
  a = new doublearray;
  if (a == 0)
    {
      ErrorFlag = true;
      goto j1;
    }
  a->setbuff(1000);

  if (!Omaketable(xy[1],  xy[2], a, normalcount))
    {
      ErrorFlag = true;
      goto j1;
    }

  outcount = (normalcount - 1) * divno;
  for (i = 0; i <= outcount; i++)
    {
      t= ((*xy[1])[normalcount - 1] - (*xy[1])[0])/ (double)outcount * (double)i + (*xy[1])[0];
      
      py = Ospline( t, xy[1], xy[2], a, normalcount);
      if (t == HUGE_VAL || py == HUGE_VAL)
	{
	  ErrorFlag = true;
	  goto j1;
	}
      
      if (!outx->add(t))
	{
	  ErrorFlag = true;
	  goto j1;
	}
      if (!outy->add(py))
	{
	  ErrorFlag = true;
	  goto j1;
	}
    }
  

j1:;
  if (a != 0)
    delete a;
  
  if (xy[1] != 0)
    delete xy[1];
  if (xy[2] != 0)
    delete xy[2];
  if (tx != 0)
    delete tx;
  if (ty != 0)
    delete ty;
  
  if (ErrorFlag)
    return false;
  else
    return false;
}
