/*----------------------------------------------------------------------------
                  regexpress.cc (regular expression module)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 99


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#if defined(TOPAZWIN32)
extern "C" {
#include <sys/types.h>
#include <regex.h>
};
#else
#include <sys/types.h>
#include <regex.h>
#endif
#include "regexpress.h"
#include "valbuff.h"

bool groupmatch(const char *text, const char *pattern, VariantBuff *vbuff)
{
  int     status;
  regex_t reg;
  int     nsub;
  regmatch_t *pmatch;
  std::string str1, str2;
  char stemp[100], stemp1[10];

  if (regcomp(&reg, pattern, REG_EXTENDED) != 0)
    return false;
  nsub = reg.re_nsub + 1;
  pmatch = new regmatch_t [nsub];
  status = regexec(&reg, text, nsub, pmatch, 0);


  if (status == 0)
    {
      // $&
      str1 = std::string("$&");
      str2 = std::string("");
      for (int j = 0; j < pmatch[0].rm_eo - pmatch[0].rm_so; j++)
	{
	  stemp1[0] = text[j + pmatch[0].rm_so]; stemp1[1] = 0;
	  str2 += std::string(stemp1);
	}
      vbuff->SetSystemScalar(&str1, &str2);
      
      // $`
      str1 = std::string("$`");
      str2 = std::string("");
      for (int j = 0; j < pmatch[0].rm_so; j++)
	{
	  stemp1[0] = text[j]; stemp1[1] = 0;
	  str2 += std::string(stemp1);
	}
      vbuff->SetSystemScalar(&str1, &str2);
      
      // $'
      str1 = std::string("$'");
      str2 = std::string("");
      for (int j =  pmatch[0].rm_eo; j < (int)strlen(text); j++)
	{
	  stemp1[0] = text[j]; stemp1[1] = 0;
	  str2 += std::string(stemp1);
	}
      vbuff->SetSystemScalar(&str1, &str2);
      
      // $number
      for (int i = 1; i < nsub; i++)
	{
	  str1 = std::string("$");
	  sprintf(stemp, "%d", i);
	  str1 += std::string(stemp);
	  str2 = std::string("");
	  for (int j = 0; j < pmatch[i].rm_eo - pmatch[i].rm_so; j++)
	    {
	      stemp1[0] = text[j + pmatch[i].rm_so]; stemp1[1] = 0;
	      str2 += std::string(stemp1);
	    }
	  vbuff->SetSystemScalar(&str1, &str2);
	}
    }
  
  regfree(&reg);
  if (status != 0)
    return false;

  return true;
}

bool simplematch(const char *string, const char *pattern)
{
  int     status;
  regex_t reg;
  
  if (regcomp(&reg, pattern, REG_EXTENDED | REG_NOSUB) != 0)
    return false;
  status = regexec(&reg, string, (size_t) 0, NULL, 0);
  regfree(&reg);
  if (status != 0)
    return false;

  return true;
}



