/*----------------------------------------------------------------------------
                  script.cc (main routine of script execution)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <math.h>
#include  <stdlib.h>
#include  <string>
#include  <ctype.h>
#include  <string.h>
#include  <unistd.h>
#include  <time.h>

#include "script.h"
#include "system.h"
#include "tokenbuff.h"
#include "graph.h"
#include "buffarray.h"
#include "cui.h"
#include "browse.h"
#include "objectlist.h"
#include "pipecall.h"
#include "minmax.h"
#include "gamma.h"
#include "filepoint.h"
#include "convtxt.h"
#include "msgbuff.h"
#include "topazvalues.h"
#include "regexpress.h"

extern bool dinvq(double in, double *out);
extern WINDOW *menuwin, *msgwin, *sepwin ,*titlewin;
extern FileHandleArray *fhvuff;
extern void cuimenu();
extern void showsepwin();
extern void message(const char *msg);
extern msgbuff* msgb;
extern int emessage(int lineno, char *mes, const char *com, char *scrname);

#define GRAPHOBJ_INVALID  0
#define GRAPHOBJ_SCALAR   1
#define GRAPHOBJ_ARRAY    2

script::script(graph *g)
{
  scriptname = new std::string("");
  stack   = new vstackarray(0);
  polish  = new vstackarray(0);
  v       = new stackarray(0);
  buff    = new stackarray(0);
  vbuff   = new VariantBuff;
  gra = g;
  prevscr = 0;

  int i;
  for (i = 0; i < STRSTATUS_BUFF_LEN; i++) 
    strstatusbuff[i] = 0;

  for (i = 0; i < BRACESTATUS_BUFF_LEN; i++) 
    bracestatusbuff[i] = 0;

}

script::script(graph *g, script *scr)
{
  prevscr = scr;
  scriptname = new std::string("");
  stack   = new vstackarray(0);
  polish  = new vstackarray(0);
  v       = new stackarray(0);
  buff    = new stackarray(0);
  vbuff   = prevscr->vbuff;
  gra = g;

  int i;
  for (i = 0; i < STRSTATUS_BUFF_LEN; i++) 
    strstatusbuff[i] = 0;

  for (i = 0; i < BRACESTATUS_BUFF_LEN; i++) 
    bracestatusbuff[i] = 0;

}

script::~script()
{
  delete scriptname;
  delete stack;
  delete polish;
  delete v;
  delete buff;
  
  if (prevscr == 0)
    delete vbuff;
}



bool script::GetValue(tokenbuff *token, std::string *outstring)
{
  std::string tempvalue, tempstring, tempstring1;

  if (token->strstatus == STR_UQUOTE)
    {
      switch(token->status)
	{
	case NUM:
	  if (execchildprocess(&(*token->value)[0], outstring) == EXIT_SUCCESS)
	  {
	    token->strstatus = STR_NORMAL;
	    return true;
	  }
	  else
	    return false;
	  break;
	}
    }
  else
    return GetImplicitValue(token, outstring);
  return true;
}

bool script::GetImplicitValue(tokenbuff *token,  std::string *outstring)
{
  std::string tempvalue, tempstring, tempstring1;
  tokenbuff temptoken;
  std::string out;
  int id;

  switch(token->status)
    {
    case NUM:
      *outstring = (*token->value)[0];
      return true;
      break;
    case VARIANT:
      if (vbuff->GetScalar(token->label, &tempvalue) == true)
	{
	  *outstring = tempvalue;
	  return true;
	}
      else
	return errormessage(token->lineno, "not defined", token->label->c_str());
      break;
    case ARRAYVARIANT:
      tempstring = std::string("@");
      tempstring1 = std::string((const char *)(token->label->c_str() + 1));
      tempstring += tempstring1;
      if (vbuff->Get(&tempstring, atoi((*token->value)[0].c_str()), &tempvalue) == true)
	{
	  *outstring = tempvalue;
	  return true;
	}
      else
	{
	  tempstring = *token->label + std::string("[") + (*token->value)[0] + std::string("]");
	  return errormessage(token->lineno, "not defined", tempstring.c_str());
	}
      break;
    case OBJ_MEMBER:
       if (gra->topobject(token->label, &out, &id) == true)
	{
	  if (out == std::string("$graph") && id == -1)
	    {
	      if (gra->shrinkobject(token->label, &out) == true)
		{
		  if (gra->get(&out,  &temptoken))  
		  {
		    *outstring = (*temptoken.value)[0];
		    return true;
		  }
		  else
		    return errormessage(token->lineno, "not defined", token->label->c_str());
		}
	      else
		return errormessage(token->lineno, "not defined", token->label->c_str());
	    }
	  else
	    return errormessage(token->lineno, "not defined", token->label->c_str());
	}
      else
	return errormessage(token->lineno, "not defined", token->label->c_str());
      break;
    default:
      return false;
    }
  return true;
}

int script::errormessage(int lineno, char *mes, const char *com)
{
  return emessage(lineno, mes, com, (char *)scriptname->c_str());
}



//*****************************************
//************* objpointer ****************
//*****************************************
int o_objpointer(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  char stemp[1024];
  std::string out;
  int id;
  int graphobjtype;
  buffarray tempbuffarray;
  tokenbuff temptoken;
  int itemp;

  if (st != OBJ_POINTER)
    return scr->errormessage(lineno, "invalid operator", "->");

  if (v->getarraylength() > 1 )
    {
      // object member
      if ((*v)[v->getarraylength() - 2].status == OBJ_MEMBER
	  &&
	  ((*v)[v->getarraylength() - 1].status == OBJ_ARRAYMEMBER
	   || (*v)[v->getarraylength() - 1].status == OBJ_MEMBER))
	{
	  (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	  
	  if ((*v)[v->getarraylength() - 1].status == OBJ_MEMBER)
	    {
	      *(*v)[v->getarraylength() - 2].label += std::string("->") + *(*v)[v->getarraylength() - 1].label;
	      (*v)[v->getarraylength() - 2].status = OBJ_MEMBER;
	      v->pop();
	    }
	  else if ((*v)[v->getarraylength() - 1].status == OBJ_ARRAYMEMBER)
	    {
	      *(*v)[v->getarraylength() - 2].label += std::string("->")
		+ *(*v)[v->getarraylength() - 1].label
		+ std::string("[")
		+(*(*v)[v->getarraylength() - 1].value)[0]
		+ std::string("]");
	      (*v)[v->getarraylength() - 2].status = OBJ_MEMBER;
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", "->");
	}
      else if (v->getarraylength() > 2)
	{
	  // object method
	  if ((*v)[v->getarraylength() - 3].status == OBJ_MEMBER
	      && (*v)[v->getarraylength() - 1].status == OBJ_METHOD)
	    {
	      *(*v)[v->getarraylength() - 3].label += std::string("->") + *(*v)[v->getarraylength() - 1].label;
	      
	      //must execute method here!!
	      if ((*v)[v->getarraylength() - 2].status == NUM
		  || (*v)[v->getarraylength() - 2].status == VARIANT
		  || (*v)[v->getarraylength() - 2].status == ARRAYVARIANT
		  || (*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
		{
		  if (scr->gra->topobject((*v)[v->getarraylength() - 3].label, &out, &id) == true)
		    {
		      if (out == std::string("$graph"))
			graphobjtype = GRAPHOBJ_SCALAR;
		      else if (out == std::string("@graph"))
			graphobjtype = GRAPHOBJ_ARRAY;
		      else
			graphobjtype = GRAPHOBJ_INVALID;
		      
		      if (graphobjtype != GRAPHOBJ_INVALID && id == -1)
			{
			  if (scr->gra->shrinkobject((*v)[v->getarraylength() - 3].label, &out) == true)
			    {
			      if (scr->GetValue(&(*v)[v->getarraylength() - 2], &str1))
				{
				  tempbuffarray.setarray(0);
				  tempbuffarray.add(str1.c_str());
				  
				  temptoken.value->setbuff(0);
				  if (scr->gra->exec(&out, &tempbuffarray, &temptoken))
				    {
				      (*v)[v->getarraylength() - 3] = temptoken;
				      *(*v)[v->getarraylength() - 3].label = std::string("");
				      
				      if (temptoken.value->getarraylength() == 0)
					v->pop();
				      else if (temptoken.value->getarraylength() == 1 && graphobjtype == GRAPHOBJ_SCALAR)
					(*v)[v->getarraylength() - 3].status = NUM;
				      else if (temptoken.value->getarraylength() > 1 && graphobjtype == GRAPHOBJ_ARRAY)
					(*v)[v->getarraylength() - 3].status = IMPLICITARRAY;
				      else
					return scr->errormessage(lineno, "invalid return", out.c_str());
				    }
				  else
				    return scr->errormessage(lineno, "execution error", out.c_str());
				}
			      else
				return scr->errormessage(lineno, "invalid arguments", out.c_str());
			    }
			  else
			    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
			}
		      else
			return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		    }
		  else
		    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		}
	      else if ((*v)[v->getarraylength() - 2].status == IMPLICITARRAY)
		{
		  if (scr->gra->topobject((*v)[v->getarraylength() - 3].label, &out, &id) == true)
		    {
		      if (out == std::string("$graph"))
			graphobjtype = GRAPHOBJ_SCALAR;
		      else if (out == std::string("@graph"))
			graphobjtype = GRAPHOBJ_ARRAY;
		      else
			graphobjtype = GRAPHOBJ_INVALID;
		      
		      if (graphobjtype != GRAPHOBJ_INVALID && id == -1)
			{
			  if (scr->gra->shrinkobject((*v)[v->getarraylength() - 3].label, &out) == true)
			    {
			      temptoken.value->setbuff(0);
			      if (scr->gra->exec(&out, (*v)[v->getarraylength() - 2].value, &temptoken))
				{
				  (*v)[v->getarraylength() - 3] = temptoken;
				  *(*v)[v->getarraylength() - 3].label = std::string("");
				  
				  if (temptoken.value->getarraylength() == 0)
				    v->pop();
				  else if (temptoken.value->getarraylength() == 1 && graphobjtype == GRAPHOBJ_SCALAR)
				    (*v)[v->getarraylength() - 3].status = NUM;
				  else if (temptoken.value->getarraylength() > 1 && graphobjtype == GRAPHOBJ_ARRAY)
				    (*v)[v->getarraylength() - 3].status = IMPLICITARRAY;
				  else
				    return scr->errormessage(lineno, "invalid return", out.c_str());
				}
			      else
				return scr->errormessage(lineno, "execution error", out.c_str());
			    }
			  else
			    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
			}
		      else
			return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		    }
		  else
		    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		}
	      else if ((*v)[v->getarraylength() - 2].status == ARRAY)
		{
		  if (scr->gra->topobject((*v)[v->getarraylength() - 3].label, &out, &id) == true)
		    {
		      if (out == std::string("$graph"))
			graphobjtype = GRAPHOBJ_SCALAR;
		      else if (out == std::string("@graph"))
			graphobjtype = GRAPHOBJ_ARRAY;
		      else
			graphobjtype = GRAPHOBJ_INVALID;
		      
		      if (graphobjtype != GRAPHOBJ_INVALID && id == -1)
			{
			  if (scr->gra->shrinkobject((*v)[v->getarraylength() - 3].label, &out) == true)
			    {
			      tempbuffarray.setarray(0);					  
			      for (itemp = 0;
				   itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 2].label);
				   itemp++)
				{
				  scr->vbuff->Get((*v)[v->getarraylength() - 2].label, itemp, &str1);
				  tempbuffarray.add(str1.c_str());
				}
			      temptoken.value->setbuff(0);
			      if (scr->gra->exec(&out, &tempbuffarray, &temptoken))
				{
				  (*v)[v->getarraylength() - 3] = temptoken;
				  *(*v)[v->getarraylength() - 3].label = std::string("");
				  
				  if (temptoken.value->getarraylength() == 0)
				    v->pop();
				  else if (temptoken.value->getarraylength() == 1 && graphobjtype == GRAPHOBJ_SCALAR)
				    (*v)[v->getarraylength() - 3].status = NUM;
				  else if (temptoken.value->getarraylength() > 1 && graphobjtype == GRAPHOBJ_ARRAY)
				    (*v)[v->getarraylength() - 3].status = IMPLICITARRAY;
				  else
				    return scr->errormessage(lineno, "invalid return", out.c_str());
				}
			      else
				return scr->errormessage(lineno, "execution error", out.c_str());
			    }
			  else
			    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
			}
		      else
			return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		    }
		  else
		    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
		}
	      else
		return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
	      v->pop();
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid object", (*v)[v->getarraylength() - 3].label->c_str());
	}
      else
	return scr->errormessage(lineno, "invalid object", "");
    }
  else
    return scr->errormessage(lineno, "invalid object", "");
  
  return true;
}

//*****************************************
//************* set ***********************
//*****************************************
int o_set(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, str4, tempstring, tempstring1;
  bool b1;
  char stemp[1024];
  FILE *h;
  std::string out;
  int id, itemp;
  
  if (st != SET)
    return scr->errormessage(lineno, "invalid operator", "=");

  if (v->getarraylength() > 1 )
    {
      if ((*v)[v->getarraylength() - 2].status == VARIANT)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
	  if (b1)
	    {
	      if (!scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label, &str1))
		return false;
	      (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	      *(*v)[v->getarraylength() - 2].label = std::string("");
	      (*v)[v->getarraylength() - 2].status = NUM;
	      v->pop();
	    }
	  else
	    {
	      if ((*v)[v->getarraylength() - 1].status == ARRAY)
		{
		  sprintf(stemp, "%d", scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label));
		  str4 = std::string(stemp);
		  scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label
				   , &str4);
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
		{
		  sprintf(stemp, "%d", (*v)[v->getarraylength() - 2].value->getarraylength());
		  str4 = std::string(stemp);
		  scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label
				   , &str4);
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      //read a line from file
	      else if ((*v)[v->getarraylength() - 1].status == READPIPE)
		{
		  if (fhvuff->get((*v)[v->getarraylength() - 1].label, &h, &itemp, &str3))
		    {
		      if (ReadLine(h, &str2))
			{
			  if (!scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label, &str2))
			    return false;
			  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
			}
		      else
			(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
		      
		      
		      *(*v)[v->getarraylength() - 2].label = std::string("");
		      (*v)[v->getarraylength() - 2].status = NUM;
		      v->pop();
		    }
		  else
		    return scr->errormessage(lineno, "could not get file handle", "=");
		}
	      else
		return scr->errormessage(lineno, "invalid operator", "=");
	    }
	}
      else if ((*v)[v->getarraylength() - 2].status == ARRAY)
	{
	  if ((*v)[v->getarraylength() - 1].status == NUM)
	    {
	      scr->vbuff->SetArray((*v)[v->getarraylength() - 2].label
			      , atoi((*(*v)[v->getarraylength() - 1].value)[0].c_str()));
	      
	      (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	      *(*v)[v->getarraylength() - 2].label = std::string("");
	      (*v)[v->getarraylength() - 2].status = NUM;
	      v->pop();
	    }
	  else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	    {
	      if (scr->vbuff->SetArray((*v)[v->getarraylength() - 2].label
				  , (*v)[v->getarraylength() - 1].value->getarraylength()) == true)
		{
		  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		    {
		      str1 = (*v)[v->getarraylength() - 1].value->getdata(itemp);
		      scr->vbuff->Set((*v)[v->getarraylength() - 2].label,
				 itemp,
				 &str1);
		    }
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return scr->errormessage(lineno, "could not allocate memory", "=");
	    }
	  else if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      if (scr->vbuff->SetArray((*v)[v->getarraylength() - 2].label
				  , scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label)) == true)
		{
		  for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label); itemp++)
		    {
		      scr->vbuff->Get((*v)[v->getarraylength() - 1].label, itemp, &str1);
		      scr->vbuff->Set((*v)[v->getarraylength() - 2].label, itemp, &str1);
		    }
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return scr->errormessage(lineno, "could not allocate memory", "=");
	    }
	  //read all lines from file
	  else if ((*v)[v->getarraylength() - 1].status == READPIPE)
	    {
	      if (fhvuff->get((*v)[v->getarraylength() - 1].label, &h, &itemp, &str3))
		{
		  if (ReadAllLines(h, (*v)[v->getarraylength() - 1].value))
		    {
		      if (scr->vbuff->SetArray((*v)[v->getarraylength() - 2].label
					  , (*v)[v->getarraylength() - 1].value->getarraylength()) == true)
			{
			  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
			    {
			      str1 = (*v)[v->getarraylength() - 1].value->getdata(itemp);
			      scr->vbuff->Set((*v)[v->getarraylength() - 2].label,
					 itemp,
					 &str1);
			    }
			  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
			  *(*v)[v->getarraylength() - 2].label = std::string("");
			  (*v)[v->getarraylength() - 2].status = NUM;
			  v->pop();
			}
		      else
			return scr->errormessage(lineno, "could not allocate memory", "=");
		    }
		  else
		    return scr->errormessage(lineno, "could not read lines", "=");
		}
	      else
		return scr->errormessage(lineno, "could not get file handle", "=");
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", "=");
	}
      else if ((*v)[v->getarraylength() - 2].status == ARRAYVARIANT)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
	  if (b1)
	    {
	      tempstring = std::string("@");
	      tempstring1 = std::string((const char *)((*v)[v->getarraylength() - 2].label->c_str()) + 1);
	      tempstring += tempstring1;
	      
	      if (scr->vbuff->Set(&tempstring
			     , atoi((*(*v)[v->getarraylength() - 2].value)[0].c_str())
			     , &str1))
		{
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return scr->errormessage(lineno, "could not allocate memory", "=");
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", "=");
	}
      else if ((*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
	  if (b1)
	    {
	      if (scr->gra->topobject((*v)[v->getarraylength() - 2].label, &out, &id) == true)
		{
		  if (out == std::string("$graph") && id == -1)
		    {
		      if (scr->gra->shrinkobject((*v)[v->getarraylength() - 2].label, &out) == true)
			{
			  if (!scr->gra->set(&out, &str1))  
			    return scr->errormessage(lineno, "invalid object", "=");
			  
			  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
			  *(*v)[v->getarraylength() - 2].label = std::string("");
			  (*v)[v->getarraylength() - 2].status = NUM;
			}
		      else
			return scr->errormessage(lineno, "invalid object", "=");
		    }
		  else
		    return scr->errormessage(lineno, "invalid object", "=");
		}
	      else
		return scr->errormessage(lineno, "invalid object", "=");
	      
	      v->pop(); v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid arguments", "=");
	}
      else
	return scr->errormessage(lineno, "invalid arguments", "=");
    }
  else
    return scr->errormessage(lineno, "invalid arguments", "=");

  return true;
}

//*****************************************
//************* to_operand ****************
//*****************************************
int o_to_operand(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string tempstring;

  if (st != TO_OPERAND)
    return scr->errormessage(lineno, "invalid operator", "");
  
  if (v->getarraylength() > 0)
    {
      tempstring = std::string("$");
      tempstring += (*(*v)[v->getarraylength() - 1].value)[0];
      (*v)[v->getarraylength() - 1].priority      = PRI_OPERAND;
      (*v)[v->getarraylength() - 1].status        = VARIANT;  
      (*v)[v->getarraylength() - 1].value->add("");   
      *(*v)[v->getarraylength() - 1].label         = tempstring;   
    }
  else
    return scr->errormessage(lineno, "invalid array", "");
  
  return true;
}

//*****************************************
//************* to_array ****************
//*****************************************
int o_to_array(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string tempstring;
  
  if (st != TO_ARRAY)
    return scr->errormessage(lineno, "invalid operator", "");

  if (v->getarraylength() > 0)
    {
      tempstring = std::string("@");
      tempstring += (*(*v)[v->getarraylength() - 1].value)[0];
      (*v)[v->getarraylength() - 1].priority      = PRI_OPERAND;
      (*v)[v->getarraylength() - 1].status        = ARRAY;  
      (*v)[v->getarraylength() - 1].value->add("");   
      *(*v)[v->getarraylength() - 1].label         = tempstring;   
    }
  else
    return scr->errormessage(lineno, "invalid array", "");
  
  return true;
}

//*****************************************
//***************** + *********************
//*****************************************
int o_plus(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != PLUS)
    return scr->errormessage(lineno, "invalid operator", "+");

  if (v->getarraylength() > 1 )
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  ftemp = atof(str1.c_str()) + atof(str2.c_str());
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "+");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "+");
  
  return true;
}

//*****************************************
//***************** pre + *****************
//*****************************************
int o_preplus(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  bool b1;
  double ftemp;
  char stemp[1024];
  
  if (st != PREPLUS)
    return scr->errormessage(lineno, "invalid operator", "+");

  if (v->getarraylength() > 0)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
      if (b1)
	{
	  ftemp = atof(str1.c_str());
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid sign", "+");
    }
  else
    return scr->errormessage(lineno, "invalid sign", "+");
  
  return true;
}

//*****************************************
//***************** . *********************
//*****************************************
int o_strplus(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, str4, str5;
  bool b1, b2, b3;
  int itemp;
  tokenbuff arrayinstringseparator;
  
  // arrayseparator in string $
  arrayinstringseparator.priority = PRI_OPERAND;
  arrayinstringseparator.status = VARIANT;
  *arrayinstringseparator.label = std::string("$\"");
  arrayinstringseparator.value->add("");

  if (v->getarraylength() > 1)
    {
      if (st == STR_PLUS)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
	  b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
	}
      else
	{
	  b1 = scr->GetImplicitValue(&(*v)[v->getarraylength() - 2], &str1);
	  b2 = scr->GetImplicitValue(&(*v)[v->getarraylength() - 1], &str2);
	}
      
      if (b1 && b2)
	{
	  (*(*v)[v->getarraylength() - 2].value)[0] = str1 + str2;
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else if (!b1 && b2)
	{
	  if ((*v)[v->getarraylength() - 2].status == ARRAY)
	    {
	      b3 = scr->GetValue(&arrayinstringseparator, &str3);
	      if (b3)
		{
		  str4 = std::string("");
		  for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 2].label) - 1; itemp++)
		    {
		      scr->vbuff->Get((*v)[v->getarraylength() - 2].label, itemp, &str5);
		      str4 += str5 + str3;
		    }
		  scr->vbuff->Get((*v)[v->getarraylength() - 2].label, 
			     scr->vbuff->ArrayLength((*v)[v->getarraylength() - 2].label) - 1, &str5);
		  str4 += str5;
		  
		  (*(*v)[v->getarraylength() - 2].value)[0] = str4 + str2;
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
			      v->pop();
		}
	      else
		return scr->errormessage(lineno, "invalid operator", ".");
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", ".");
	}
      else if (b1 && !b2)
	{
	  if ((*v)[v->getarraylength() - 1].status == ARRAY)
	    {
	      b3 = scr->GetValue(&arrayinstringseparator, &str3);
	      if (b3)
		{
		  str4 = std::string("");
		  for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1; itemp++)
		    {
		      scr->vbuff->Get((*v)[v->getarraylength() - 1].label, itemp, &str5);
		      str4 += str5 + str3;
		    }
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, 
				  scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label) - 1, &str5);
		  str4 += str5;
		  
		  (*(*v)[v->getarraylength() - 2].value)[0] = str1 + str4;
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return scr->errormessage(lineno, "invalid operator", ".");
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", ".");
	}
      else
	return scr->errormessage(lineno, "invalid operator", ".");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ".");
  
  return true;
}

//*****************************************
//***************** - *********************
//*****************************************
int o_minus(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != MINUS)
    return scr->errormessage(lineno, "invalid operator", "-");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  ftemp = atof(str1.c_str()) - atof(str2.c_str());
	  
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "-");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "-");
  
  return true;
}

//*****************************************
//***************** pre - *****************
//*****************************************
int o_preminus(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  bool b1;
  double ftemp;
  char stemp[1024];
  
  if (st != PREMINUS)
    return scr->errormessage(lineno, "invalid operator", "-");

  if (v->getarraylength() > 0)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
      if (b1)
	{
	  ftemp = -atof(str1.c_str());
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 1].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid sign", "-");
    }
  else
    return scr->errormessage(lineno, "invalid sign", "-");
  
  return true;
}

//*****************************************
//***************** / *********************
//*****************************************
int o_divide(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != DIVIDE)
    return scr->errormessage(lineno, "invalid operator", "/");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str2.c_str()) != 0)
	    {
	      ftemp = atof(str1.c_str()) / atof(str2.c_str());
	      
	      sprintf(stemp, "%.15g", ftemp);		  
	      (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	      *(*v)[v->getarraylength() - 2].label = std::string("");
			  (*v)[v->getarraylength() - 2].status = NUM;
			  v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "zero division", "/");
	}
      else
	return scr->errormessage(lineno, "invalid operator", "/");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "/");
  
  return true;
}

//*****************************************
//***************** % *********************
//*****************************************
int o_intdivide(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != INTDIVIDE)
    return scr->errormessage(lineno, "invalid operator", "%");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if ((int)atof(str2.c_str()) != 0)
	    {
	      ftemp = (double)((int)atof(str1.c_str()) % (int)atof(str2.c_str()));
	      
	      sprintf(stemp, "%.15g", ftemp);		  
	      (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	      *(*v)[v->getarraylength() - 2].label = std::string("");
	      (*v)[v->getarraylength() - 2].status = NUM;
	      v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "zero division", "%");
	}
      else
	return scr->errormessage(lineno, "invalid operator", "%");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "%");
  
  return true;
}

//*****************************************
//***************** * *********************
//*****************************************
int o_times(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != TIMES)
    return scr->errormessage(lineno, "invalid operator", "*");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  ftemp = atof(str1.c_str()) * atof(str2.c_str());
	  
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "*");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "*");
  
  return true;
}

//*****************************************
//***************** , *********************
//*****************************************
int o_comma(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  int itemp;
  
  if (st != COMMA)
    return scr->errormessage(lineno, "invalid operator", ",");

  if (v->getarraylength() > 1)
    {
      if ((*v)[v->getarraylength() - 2].status == ARRAY)
	{
	  if (scr->vbuff->search((*v)[v->getarraylength() - 2].label, &itemp))
	    {
	      (*v)[v->getarraylength() - 2].value->setbuff(0);
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 2].label); itemp++)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 2].label, itemp, &str1);
		  (*v)[v->getarraylength() - 2].value->add(str1.c_str());
		}
	      *(*v)[v->getarraylength() - 2].label = std::string("");
	      (*v)[v->getarraylength() - 2].status = IMPLICITARRAY;
	    }
	  else
	    return scr->errormessage(lineno, "can not fine variable", (*v)[v->getarraylength() - 2].label->c_str());
	}
      
      if ((*v)[v->getarraylength() - 1].status == ARRAY)
	{
	  if (scr->vbuff->search((*v)[v->getarraylength() - 1].label, &itemp))
	    {
	      (*v)[v->getarraylength() - 1].value->setbuff(0);
	      for (itemp = 0; itemp < scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label); itemp++)
		{
		  scr->vbuff->Get((*v)[v->getarraylength() - 1].label, itemp, &str1);
		  (*v)[v->getarraylength() - 1].value->add(str1.c_str());
		}
	      *(*v)[v->getarraylength() - 1].label = std::string("");
	      (*v)[v->getarraylength() - 1].status = IMPLICITARRAY;
	    }
	  else
	    return scr->errormessage(lineno, "can not find variable", (*v)[v->getarraylength() - 1].label->c_str());
	}
      
      
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      
      if (b1 == true && b2 == true)
	{
	  (*v)[v->getarraylength() - 2].value->setarray(0);
	  (*v)[v->getarraylength() - 2].value->add(str1.c_str());
	  (*v)[v->getarraylength() - 2].value->add(str2.c_str());
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = IMPLICITARRAY;
		      v->pop();
	}
      else if (b1 == false && b2 == true && (*v)[v->getarraylength() - 2].status == IMPLICITARRAY)
	{
	  (*v)[v->getarraylength() - 2].value->add(str2.c_str());
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = IMPLICITARRAY;
	  
	  v->pop();
	}
      else if (b1 == true && b2 == false && (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  (*v)[v->getarraylength() - 2].value->setarray(0);
	  (*v)[v->getarraylength() - 2].value->add(str1.c_str());
	  
	  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
	    (*v)[v->getarraylength() - 2].value->add((*(*v)[v->getarraylength() - 1].value)[itemp].c_str());
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = IMPLICITARRAY;
	  
	  v->pop();
	}
      else if (b1 == false && b2 == false 
	       && (*v)[v->getarraylength() - 2].status == IMPLICITARRAY
	       && (*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
	{
	  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
	    (*v)[v->getarraylength() - 2].value->add((*(*v)[v->getarraylength() - 1].value)[itemp].c_str());
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = IMPLICITARRAY;
	  
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", ",");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ",");
  
  return true;
}

//*****************************************
//***************** exp *******************
//*****************************************
int o_exp(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  double ftemp;
  char stemp[1024];
  
  if (st != EXP)
    return scr->errormessage(lineno, "invalid operator", "** / ^");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  ftemp = pow(atof(str1.c_str()), atof(str2.c_str()));
	  
	  sprintf(stemp, "%.15g", ftemp);		  
	  (*(*v)[v->getarraylength() - 2].value)[0] = std::string(stemp);
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "** / ^");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "** / ^");
  
  return true;
}

//*****************************************
//***************** >= *******************
//*****************************************
int o_larger_equal_than(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  
  if (st != LARGER_EQUAL_THAN)
    return scr->errormessage(lineno, "invalid operator", ">= / =>");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) >= atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", ">= / =>");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ">= / =>");
  
  return true;
}

//*****************************************
//***************** > *********************
//*****************************************
int o_larger_than(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  
  if (st != LARGER_THAN)
    return scr->errormessage(lineno, "invalid operator", ">");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) > atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", ">");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ">");
  
  return true;
}

//*****************************************
//***************** <= ********************
//*****************************************
int o_smaller_equal_than(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  
  if (st != SMALLER_EQUAL_THAN)
    return scr->errormessage(lineno, "invalid operator", "<= / =<");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) <= atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "<= / =<");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "<= / =<");
  
  return true;
}

//*****************************************
//***************** < *********************
//*****************************************
int o_smaller_than(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;
  
  if (st != SMALLER_THAN)
    return scr->errormessage(lineno, "invalid operator", "<");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) < atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "<");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "<");
  
  return true;
}

//*****************************************
//*********** string operation ************
//*****************************************
int o_strop(script *scr, int status, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3, str4;
  bool b1, b2;

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (status == STR_EQUAL)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) == 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	    }
	  else if (status == STR_NOTEQUAL)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) == 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  else if (status == STR_LESSTHAN)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) < 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  else if (status == STR_GREATERTHAN)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) > 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  else if (status == STR_LESSEQUALTHAN)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) <= 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  else if (status == STR_GREATEREQUALTHAN)
	    {
	      if (strcmp(str1.c_str(), str2.c_str()) >= 0)
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  else if (status == MATCH)
	    {

	      if (groupmatch(str1.c_str(), str2.c_str(), scr->vbuff))
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	      else
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	    }
	  else if (status == NOTMATCH)
	    {
	      if (groupmatch(str1.c_str(), str2.c_str(), scr->vbuff))
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	    }
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "string operation");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "string operation");
  
  return true;
}

//*****************************************
//***************** inc *******************
//*****************************************
int o_inc(script *scr, int status, int lineno)
{
  stackarray *v = scr->v;
  std::string tempstring, tempstring1, str1, str2, str3, str4;
  bool b1;
  tokenbuff temptoken;
  int itemp;
  char stemp[1024];
  
  if (v->getarraylength() > 0)
    {
      temptoken = (*v)[v->getarraylength() - 1];
      str3 = *temptoken.label;
      itemp = temptoken.status;
      if (temptoken.status == ARRAYVARIANT)
	str4 = (*temptoken.value)[0];
      
      b1 = scr->GetValue(&temptoken, &str1);
      if (b1)
	{
	  //pre increment/decrement
	  if (status == BEFORE_INC
	      || status == BEFORE_DEC)
	    {
	      if (status == BEFORE_INC)
		sprintf(stemp, "%.15g", (atof(str1.c_str()) + 1.0));
	      else if (status == BEFORE_DEC)
		sprintf(stemp, "%.15g", (atof(str1.c_str()) - 1.0));
			  
	      str2 = std::string(stemp);
	      
	      if (itemp == VARIANT)
		{
		  if (!scr->vbuff->SetScalar(&str3, &str2))
				return false;
		}
	      else if (itemp == ARRAYVARIANT)
		{
		  tempstring = std::string("@");
		  tempstring1 = std::string((const char *)(str3.c_str() + 1));
		  tempstring += tempstring1;
			      
		  if (!scr->vbuff->Set(&tempstring
				  ,atoi(str4.c_str())
					      ,&str2))
		    return false;
		}
	      else
		return scr->errormessage(lineno, "invalid operator", "++ / --");
	    }
	  //end of pre increment/decrement
	  
	  
	  //getvalue
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
	  (*(*v)[v->getarraylength() - 1].value)[0] = str1;
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	  
		      
	  //post increment/decrement
	  if (status == AFTER_INC
	      || status == AFTER_DEC)
	    {
	      if (status == AFTER_INC)
		sprintf(stemp, "%.15g", (atof(str1.c_str()) + 1.0));
	      else if (status == AFTER_DEC)
		sprintf(stemp, "%.15g", (atof(str1.c_str()) - 1.0));
	      
	      str2 = std::string(stemp);

	      if (itemp == VARIANT)
		{
		  if (!scr->vbuff->SetScalar(&str3, &str2))
		    return false;
		}
	      else if (itemp == ARRAYVARIANT)
		{
		  tempstring = std::string("@");
		  tempstring1 = std::string((const char *)(str3.c_str() + 1));
		  tempstring += tempstring1;
			      
		  if (!scr->vbuff->Set(&tempstring
				  ,atoi(str4.c_str())
				  ,&str2))
		    return false;
		}
	      else
		return scr->errormessage(lineno, "invalid operator", "++ / --");
	    }
	  //end of post increment/decrement
	  
	}
      else
	return scr->errormessage(lineno, "invalid operator", "++ / --");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "++ / --");
  
  return true;
}

//*****************************************
//***************** or/and ****************
//*****************************************
int o_or_and(script *scr, int status, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (status == OR)
	    {
	      if (atof(str1.c_str()) < atof(str2.c_str()))
		(*(*v)[v->getarraylength() - 2].value)[0] = str2;
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = str1;
	    }  
	  else if (status == AND)
	    {
	      if (atof(str1.c_str()) < atof(str2.c_str()))
		(*(*v)[v->getarraylength() - 2].value)[0] = str1;
	      else
		(*(*v)[v->getarraylength() - 2].value)[0] = str2;
			}
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	  
	}
      else
	return scr->errormessage(lineno, "invalid operator", "|| / &&");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "|| / &&");
  
  return true;
}

//*****************************************
//***************** == ********************
//*****************************************
int o_equal(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;

  if (st != EQUAL)
    return scr->errormessage(lineno, "invalid operator", "==");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) == atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "==");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "==");
  
  return true;
}

//*****************************************
//***************** != ********************
//*****************************************
int o_not_equal(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2;
  bool b1, b2;

  if (st != NOTEQUAL)
    return scr->errormessage(lineno, "invalid operator", "!=");

  if (v->getarraylength() > 1)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
      b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
      if (b1 && b2)
	{
	  if (atof(str1.c_str()) != atof(str2.c_str()))
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
	  else
	    (*(*v)[v->getarraylength() - 2].value)[0] = std::string("0");
	  
	  *(*v)[v->getarraylength() - 2].label = std::string("");
	  (*v)[v->getarraylength() - 2].status = NUM;
	  v->pop();
	}
      else
	return scr->errormessage(lineno, "invalid operator", "!=");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "!=");
  
  return true;
}

//*****************************************
//***************** ! *********************
//*****************************************
int o_not(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  bool b1;

  if (st != NOT)
    return scr->errormessage(lineno, "invalid operator", "!");

  if (v->getarraylength() > 0)
    {
      b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
      if (b1)
	{
	  if ((int)atof(str1.c_str()) == 1)
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("0");
	  else if ((int)atof(str1.c_str()) == 0)
	    (*(*v)[v->getarraylength() - 1].value)[0] = std::string("1");
	  else
	    return false;
	  
	  *(*v)[v->getarraylength() - 1].label = std::string("");
	  (*v)[v->getarraylength() - 1].status = NUM;
	}
      else
	return scr->errormessage(lineno, "invalid operator", "!");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "!");
  
  return true;
}

//*****************************************
//***************** ,= ********************
//*****************************************
int o_comma_equal(script *scr, int st, int lineno)
{
  stackarray *v = scr->v;
  std::string str1;
  bool b1;
  int itemp;

  if (st != COMMAEQUAL)
    return scr->errormessage(lineno, "invalid operator", ",=");

  if (v->getarraylength() > 1)
    {
      if ((*v)[v->getarraylength() - 2].status == ARRAY)
	{
	  
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str1);
	  if (b1)
	    {
	      if (scr->vbuff->Add((*v)[v->getarraylength() - 2].label, &str1))
		{
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return false;
	    }
	  else
	    {
	      if ((*v)[v->getarraylength() - 1].status == ARRAY)
		{
		  itemp = scr->vbuff->ArrayLength((*v)[v->getarraylength() - 1].label);
		  for (int i = 0; i < itemp; i++)
		    {
		      scr->vbuff->Get((*v)[v->getarraylength() - 1].label, i, &str1);
		      scr->vbuff->Add((*v)[v->getarraylength() - 2].label, &str1);
		    }
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else if ((*v)[v->getarraylength() - 1].status == IMPLICITARRAY)
		{
		  for (itemp = 0; itemp < (*v)[v->getarraylength() - 1].value->getarraylength(); itemp++)
		    {
		      str1 = (*v)[v->getarraylength() - 1].value->getdata(itemp);
		      scr->vbuff->Add((*v)[v->getarraylength() - 2].label, &str1);
		    }
		  (*(*v)[v->getarraylength() - 2].value)[0] = std::string("1");
		  *(*v)[v->getarraylength() - 2].label = std::string("");
		  (*v)[v->getarraylength() - 2].status = NUM;
		  v->pop();
		}
	      else
		return scr->errormessage(lineno, "invalid operator", ",=");
	    }
	}
      else
	return scr->errormessage(lineno, "invalid operator", ",=");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ",=");
  
  return true;
}

//*****************************************
//***************** op = ******************
//*****************************************
int o_op_equal(script *scr, int status, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  bool b1, b2;
  char stemp[1024];
  double ftemp = 0.0;
  std::string out;
  int id;

  if (v->getarraylength() > 1 )
    {
      if ((*v)[v->getarraylength() - 2].status == VARIANT
	  || (*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
	  b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
	  if (b1 && b2)
	    {
	      switch(status)
		{
		case PLUSEQUAL:
		  ftemp = atof(str1.c_str()) + atof(str2.c_str());
		  break;
		case MINUSEQUAL:
		  ftemp = atof(str1.c_str()) - atof(str2.c_str());
		  break;
		case TIMESEQUAL:
		  ftemp = atof(str1.c_str()) * atof(str2.c_str());
		  break;
		case DIVIDEEQUAL:
		  if (atof(str2.c_str()) != 0.0)
		    ftemp = atof(str1.c_str()) / atof(str2.c_str());
		  else
		    return scr->errormessage(lineno, "zero division", "/=");
		  break;
		case INTDIVIDEEQUAL:
		  if ((int)atof(str2.c_str()) != 0)
		    ftemp = (double)((int)atof(str1.c_str()) % (int)atof(str2.c_str()));
		  else
		    return scr->errormessage(lineno, "zero division", "%=");
		  break;
		case EXPEQUAL:
		  ftemp = pow(atof(str1.c_str()), atof(str2.c_str()));
		  break;
		}
	      
	      sprintf(stemp, "%.15g", ftemp);		  
	      if ((*v)[v->getarraylength() - 2].status == VARIANT)
		{
		  str3 = std::string(stemp);
		  if (!scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label, &str3))
		    return false;
		}
	      else if ((*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
		{
		  if (scr->gra->topobject((*v)[v->getarraylength() - 2].label, &out, &id) == true)
		    {
		      if (out == std::string("$graph") && id == -1)
			{
			  if (scr->gra->shrinkobject((*v)[v->getarraylength() - 2].label, &out) == true)
			    {
			      str3 = std::string(stemp);
			      if (!scr->gra->set(&out, &str3))  
				return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
			    }
			  else
			    return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
			}
		      else
			return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
		    }
		  else
				return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
		}
	      v->pop(); v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
	  
	}
      else
	return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
    }
  else
    return scr->errormessage(lineno, "invalid operator", "+= / -= / *= / . . .");
  
  return true;
}

//*****************************************
//***************** .= ********************
//*****************************************
int o_strplus_equal(script *scr, int status, int lineno)
{
  stackarray *v = scr->v;
  std::string str1, str2, str3;
  bool b1, b2;
  std::string out;
  int id;

  if (v->getarraylength() > 1 )
    {
      if ((*v)[v->getarraylength() - 2].status == VARIANT
	  || (*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
	{
	  b1 = scr->GetValue(&(*v)[v->getarraylength() - 2], &str1);
	  b2 = scr->GetValue(&(*v)[v->getarraylength() - 1], &str2);
	  if (b1 && b2)
	    {
	      switch(status)
		{
		case STRPLUSEQUAL:
		  str3 = str1 + str2;
		  break;
		}
	      
	      if ((*v)[v->getarraylength() - 2].status == VARIANT)
		{
		  if (!scr->vbuff->SetScalar((*v)[v->getarraylength() - 2].label, &str3))
		    return false;
		}
	      else if ((*v)[v->getarraylength() - 2].status == OBJ_MEMBER)
		{
		  if (scr->gra->topobject((*v)[v->getarraylength() - 2].label, &out, &id) == true)
		    {
		      if (out == std::string("$graph") && id == -1)
			{
			  if (scr->gra->shrinkobject((*v)[v->getarraylength() - 2].label, &out) == true)
			    {
			      if (!scr->gra->set(&out, &str3))  
				return scr->errormessage(lineno, "invalid operator", ".=");
			    }
			  else
			    return scr->errormessage(lineno, "invalid operator", ".=");
			}
		      else
			return scr->errormessage(lineno, "invalid operator", ".=");
		    }
		  else
		    return scr->errormessage(lineno, "invalid operator", ".=");
		}
	      
	      v->pop(); v->pop();
	    }
	  else
	    return scr->errormessage(lineno, "invalid operator", ".=");
	}
      else
	return scr->errormessage(lineno, "invalid operator", ".=");
    }
  else
    return scr->errormessage(lineno, "invalid operator", ".=");
  
  return true;
}



typedef struct funcsmap
{
  int op;
  int (*func)(script *scr, int status, int lineno);
}OPMAP;

OPMAP opmap[] = {
  {OBJ_POINTER, o_objpointer},
  {SET, o_set},
  {TO_OPERAND, o_to_operand},
  {TO_ARRAY, o_to_array},
  {BEFORE_INC, o_inc},
  {AFTER_INC, o_inc},
  {BEFORE_DEC, o_inc},
  {AFTER_DEC, o_inc},
  {PLUS, o_plus},
  {PREPLUS, o_preplus},
  {STR_PLUS, o_strplus},
  {IMPLICITSTRPLUS, o_strplus},
  {MINUS, o_minus},
  {PREMINUS, o_preminus},
  {DIVIDE, o_divide},
  {INTDIVIDE, o_intdivide},
  {TIMES, o_times},
  {COMMA, o_comma},
  {EXP, o_exp},
  {LARGER_EQUAL_THAN, o_larger_equal_than},
  {LARGER_THAN, o_larger_than},
  {SMALLER_EQUAL_THAN, o_smaller_equal_than},
  {SMALLER_THAN, o_smaller_than},
  {STR_EQUAL, o_strop},
  {STR_NOTEQUAL, o_strop},
  {STR_LESSTHAN, o_strop},
  {STR_GREATERTHAN, o_strop},
  {STR_LESSEQUALTHAN, o_strop},
  {STR_GREATEREQUALTHAN, o_strop},
  {MATCH, o_strop},
  {NOTMATCH, o_strop},
  {OR, o_or_and},
  {AND, o_or_and},
  {EQUAL, o_equal},
  {NOTEQUAL, o_not_equal},
  {NOT, o_not},
  {COMMAEQUAL, o_comma_equal},
  {PLUSEQUAL, o_op_equal},
  {MINUSEQUAL, o_op_equal},
  {TIMESEQUAL, o_op_equal},
  {DIVIDEEQUAL, o_op_equal},
  {INTDIVIDEEQUAL, o_op_equal},
  {EXPEQUAL, o_op_equal},
  {STRPLUSEQUAL, o_strplus_equal},
  {0, 0}
}; 


int script::calc ()
{
  int itemp;
  std::string tempstring, tempstring1, str1, str2, str3, str4, str5, str6;//, chompstr, flushstr;
  int parsepoint;
  tokenbuff temptoken;
  std::string out;
  buffarray tempbuffarray;
  int lineno;
  int status;

  OPMAP *om;
  int itemno;
  int (*execfunc)(script *scr, int st, int linenumber);

  /*
  // chompstr
  str1 = std::string("$/");
  vbuff->GetScalar(&str1, &chompstr);

  // flushstr
  str1 = std::string("$|");
  vbuff->GetScalar(&str1, &flushstr);
  */

  for (parsepoint = 0; parsepoint < polish->getarraylength(); parsepoint++)
    {
      lineno = (*polish)[parsepoint].lineno;
      status = (*polish)[parsepoint].status;

      //  NUM
      if (status    == NUM 
	  || status == VARIANT
	  || status == OBJ_MEMBER
	  || status == OBJ_METHOD
	  || status == ARRAYVARIANT
	  || status == READPIPE
	  || status == ARRAY )
	{
	  v->push1(&(*polish)[parsepoint]);
	}
      // FUNC
      else if (status == FUNC)
	{
	  itemp = funcs(parsepoint);
	  if (itemp != true)
	    return itemp;
	}
      //  TOARRAY
      else if (status == TOARRAY)
        {
	  if (v->getarraylength() > 0)
	    {
	      if (GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
		{
		  (*(*v)[v->getarraylength() - 1].value)[0] = str1;
		  *(*v)[v->getarraylength() - 1].label = *(*polish)[parsepoint].label;
		  (*v)[v->getarraylength() - 1].status = ARRAYVARIANT;
		}
	      else 
		return errormessage(lineno, "invalid arguments", "toarray");
	    }
	  else
	    return errormessage(lineno, "invalid arguments", "toarray");
        }
      //  OBJ_TOARRAY
      else if (status == OBJ_TOARRAY)
        {
	  if (v->getarraylength() > 0)
	    {
	      if (GetValue(&(*v)[v->getarraylength() - 1], &str1) == true)
		{
		  (*(*v)[v->getarraylength() - 1].value)[0] = str1;
		  *(*v)[v->getarraylength() - 1].label = *(*polish)[parsepoint].label;
		  (*v)[v->getarraylength() - 1].status = OBJ_ARRAYMEMBER;
		}
	      else 
		return errormessage(lineno, "invalid arguments", "");
	    }
	  else
	    return errormessage(lineno, "invalid arguments", "");
        }
      else                            
	{
	  om = opmap;
	  for (; om->op; om++)
	    {
	    }
	  itemno = om - opmap;
	  for (int i = 0; i < itemno; i++)
	    {
	      if (status == opmap[i].op)
		{
		  execfunc = opmap[i].func;
		  if (execfunc(this, status, lineno) == false)
		    return false;
		}
	    }
	}
    }

  return true;
}











