/*----------------------------------------------------------------------------
                    script.h (definitions of script object)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__script_h)
#define __script_h

#include <stdio.h>
#include <string>
#include "valbuff.h"
#include "tokenbuff.h"
#include "graph.h"
#include "buffarray.h"

/*   status   */
#define MINUS 						1
#define TIMES 						2
#define DIVIDE 						3
#define COLON 						5
#define CPAREN 						7
#define NUM 						8
#define CELL 						9
#define EOLN 						11
#define BAD 						12
#define PREPLUS 					14
#define PREMINUS 					15
#define DATAROW 					16
#define CELLNO 						17
#define OTHERDATAROW   				26
#define PARAMDATAROW   				27
#define SEMICOLON                               29
#define VARIANT                              30
#define STR_PLUS                              31
#define TO_OPERAND                    32
#define IF                            33
#define ELSEIF                        34
#define ELSE                          35
#define WHILE                         36
#define UNTIL                         37
#define FOR                           38
#define NEXTFOR                       39
#define TOARRAY                       40
#define ARRAY                         41
#define ARRAYVARIANT                  42
#define IMPLICITSTRPLUS               43
#define IMPLICITARRAY                 44       // (1,4,2)
#define FOREACH                       45
#define CONTINUE                      46
#define LAST                          47
#define NEXT                          48
#define INTDIVIDE                     49
#define NOT                           50
#define BEFORE_INC                    51
#define AFTER_INC                     52
#define BEFORE_DEC                    53
#define AFTER_DEC                     54
#define STR_EQUAL                     55
#define STR_NOTEQUAL                  56
#define STR_LESSTHAN                  57
#define STR_GREATERTHAN               58
#define STR_LESSEQUALTHAN             59
#define STR_GREATEREQUALTHAN          60
#define OBJ_POINTER                   61
#define OBJ_MEMBER                    62
#define OBJ_METHOD                    63
#define OBJ_TOARRAY                   64
#define OBJ_ARRAYMEMBER               65
#define MATCH                         66
#define NOTMATCH                      67
#define TO_ARRAY                      68
#define READPIPE                      69
#define NEXTFOREACH                   70
#define PLUS  			      71




#define OPAREN                      1001
#define COMMA                       1002
#define EXP                         1003
#define OR                          1004
#define AND                         1005
#define EQUAL                       1006
#define NOTEQUAL                    1007
#define SMALLER_THAN                1008
#define SMALLER_EQUAL_THAN          1009
#define LARGER_THAN                 1010
#define LARGER_EQUAL_THAN           1011
#define SET                         1012
#define FUNC                        1013
#define PLUSEQUAL                   1014
#define MINUSEQUAL                  1015
#define TIMESEQUAL                  1016
#define DIVIDEEQUAL                 1017
#define INTDIVIDEEQUAL              1018
#define EXPEQUAL                    1019
#define STRPLUSEQUAL                1020
#define COMMAEQUAL                  1021








#define MAXFUNCNAMELEN 	5



/*   priority   */
#define PRI_SEMICOLON                           -30
#define PRI_SET                                 -5
#define PRI_CPAREN                              -10

#define PRI_PRINT                                 0  

#define PRI_COMMA                                10
#define PRI_OR                                   20
#define PRI_AND                                  30
#define	PRI_STR_EQUAL                            35
#define	PRI_EQUAL                                40
#define	PRI_NOTEQUAL                             40
#define PRI_STR_LESSTHAN                         45
#define PRI_SMALLER_THAN                         50
#define PRI_SMALLER_EQUAL_THAN                   50
#define PRI_LARGER_THAN                          50
#define PRI_LARGER_EQUAL_THAN                    50
#define PRI_FUNCTION                             60  
#define PRI_PRUSMINUS                            80
#define PRI_MULDIV                               90
#define PRI_MATCH                                93
#define PRI_NOT                                  95
#define PRI_EXP                                 100
#define PRI_INCDEC                              105
#define PRI_OBJPOINTER                          108
#define PRI_OPERAND                             110
#define PRI_DATAROW                             110
#define PRI_PARENFUNCTION                       120  
#define PRI_OPAREN                              130


/*   string status   */
#define STR_NORMAL       0
#define STR_SQUOTE       1
#define STR_DQUOTE       2
#define STR_UQUOTE       3
#define STR_REGEX        4
#define STRSTATUS_BUFF_LEN      100

/*   brace status    */
#define BRC_NORMAL         0      /* this has no meanings  */
#define BRC_TOOPERAND      1
#define BRC_FOR            2
#define BRC_WHILE          3
#define BRC_IF             4
#define BRC_ELSEIF         5
#define BRC_ELSE           6
#define BRC_FOREACH        7
#define BRC_CONTINUE       8
#define BRC_TOARRAY        9
#define BRC_TOARRAYVARIANT 10

/* bracetype */
#define BRC_OPENPAREN      1
#define BRC_CLOSEDPAREN    2
#define BRC_OPENBRACE      3
#define BRC_CLOSEDBRACE    4
#define BRC_BRA            5
#define BRC_KET            6



#define BRACESTATUS_BUFF_LEN      100


#define PARSERSTACKSIZE 100
#define MAXINPUT 79
#define TOKENBUFFSIZE      10000

/* maxpoint number in path  */
#define  MAXPOINT_IN_PATH  1000

#define exitscriptstatus   2
#define NOTEXISTSCRIPT   20


class script
{
 public:
  script(graph *g);
  script(graph *g, script *scr);
  ~script();
  
  std::string *scriptname;  
  int errormessage(int lineno, char *mes, const char *com);
  int execscript(std::string *inputstr, std::string* out, buffarray* arg);
  int execscriptfile(std::string *inputfile, std::string *out, buffarray* arg);

// private:
  
  graph* gra;
  stackarray *v, *buff;
  vstackarray *stack , *polish;
  bool addbuff(int priority, int status, const char *value, const char *label);
  script *prevscr;

//  int sp1,sp2;
  VariantBuff *vbuff;

  bool parsetoken(std::string *inputstring);
  int calc ();
  int funcs(int);
  int exectoken(tokenbuff *token);
  int controltoken();
  bool GetValue(tokenbuff *token, std::string *outstring);
  bool GetImplicitValue(tokenbuff *token, std::string *outstring);
  void initparse(buffarray* arg, std::string *scrfile);
  void getenvstring();

/*  status for strings and braces */
  int strstatusbuffpoint;
  int strstatusbuff[STRSTATUS_BUFF_LEN];
  int bracestatusbuffpoint;
  int bracestatusbuff[BRACESTATUS_BUFF_LEN];
  int parentheseslevel;
  int lineno;
  int bracetype;
};


#endif




