/*----------------------------------------------------------------------------
                      styledisp.cc (give style name)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string>
#include "col.h"
#include "styledisp.h"
#include "data.h"
#include "frame.h"

bool getlinestylename(int id, std::string *out)
{
  switch(id)
    {
    case LINESTYLESOLID:
      *out = std::string("solid");
      break;
    case LINESTYLEDASH:
      *out = std::string("dash");
      break;
    case LINESTYLELONGDASH:
      *out = std::string("longdash");
      break;
    case LINESTYLESHORTDASH:
      *out = std::string("shortdash");
      break;
    case LINESTYLEDASHDOTDASH:
      *out = std::string("-*-");
      break;
    case LINESTYLEDASHDOTDOTDASH:
      *out = std::string("-**-");
      break;
    case LINESTYLEDOT:
      *out = std::string("dot");
      break;
    case LINESTYLENULL:
      *out = std::string("null");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

bool getcolorname(unsigned int r, unsigned int g,unsigned int b, std::string *out)
{
  char stemp[100];

  if (r == 0 && g == 0 && b == 0)
    *out = std::string("black");

  else if (r == 32768 && g == 32768 && b == 32768)
    *out = std::string("gray");

  else if (r == 65535 && g == 65535 && b == 65535)
    *out = std::string("white");

  else if (r == 65535 && g == 0 && b == 0)
    *out = std::string("red");

  else if (r == 0 && g == 65535 && b == 0)
    *out = std::string("green");

  else if (r == 0 && g == 0 && b == 65535)
    *out = std::string("blue");

  else if (r == 65535 && g == 65535 && b == 0)
    *out = std::string("yellow");

  else if (r == 65535 && g == 0 && b == 65535)
    *out = std::string("magenta");

  else if (r == 0 && g == 65535 && b == 65535)
    *out = std::string("cyan");

  else if (r == 65535 && g == 65535 && b == 32768)
    *out = std::string("L-yellow");

  else if (r == 65535 && g == 32768 && b == 65535)
    *out = std::string("L-magenta");

  else if (r == 32768 && g == 65535 && b == 65535)
    *out = std::string("L-cyan");

  else if (r == 65535 && g == 32768 && b == 32768)
    *out = std::string("L-red");

  else if (r == 32768 && g == 65535 && b == 32768)
    *out = std::string("L-green");

  else if (r == 32768 && g == 32768 && b == 65535)
    *out = std::string("L-blue");

  else if (r == 65535 && g == 32768 && b == 0)
    *out = std::string("orange");

  else if (r == 65535 && g == 0 && b == 32768)
    *out = std::string("wine");

  else if (r == 32768 && g == 32768 && b == 0)
    *out = std::string("D-yellow");

  else if (r == 32768 && g == 0 && b == 32768)
    *out = std::string("D-magenta");

  else if (r == 0 && g == 32768 && b == 32768)
    *out = std::string("D-cyan");

  else if (r == 32768 && g == 0 && b == 0)
    *out = std::string("D-red");

  else if (r == 0 && g == 32768 && b == 0)
    *out = std::string("D-green");

  else if (r == 0 && g == 0 && b == 32768)
    *out = std::string("D-blue");

  else
    {
      sprintf(stemp, "%u,%u,%u", r, g, b);
      *out = std::string(stemp);
    }
  return true;
}

bool getmarkerstylename(int id, std::string *out)
{
  switch(id)
    {
    case MARKER_NULL:
      *out = std::string("null");
      break;
    case MARKER_SQUARE:
      *out = std::string("square");
      break;
    case MARKER_TRIANGLE:
      *out = std::string("triangle");
      break;
    case MARKER_REVTRIANGLE:
      *out = std::string("rev. tri.");
      break;
    case MARKER_DIAMOND:
      *out = std::string("diamond");
      break;
    case MARKER_CROSS:
      *out = std::string("cross");
      break;
    case MARKER_DCROSS:
      *out = std::string("diag. cr.");
      break;
    case MARKER_VAREA:
      *out = std::string("v. area");
      break;
    case MARKER_HAREA:
      *out = std::string("h. area");
      break;
    case MARKER_VBAR:
      *out = std::string("v. bar");
      break;
    case MARKER_HBAR:
      *out = std::string("h. bar");
      break;
    case MARKER_VSTEP:
      *out = std::string("v. step");
      break;
    case MARKER_HSTEP:
      *out = std::string("h. step");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}

bool getframetypename(int type, std::string *out)
{
  switch(type)
    {
    case FRAME_BOX:
      *out = std::string("box");
      break;
    case FRAME_XY:
      *out = std::string("x-y");
      break;
    case FRAME_INDIV:
      *out = std::string("independent");
      break;
    default:
      *out = std::string("");
      return false;
    }
  return true;
}





