/*----------------------------------------------------------------------------
                     styleline.cc (draw lines with styles)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "vectdefs.h"
#include "pointdefs.h"
#include "gdi.h"
#include "graph.h"
#include "naninf.h"
#include "topazvalues.h"

extern graph *gra;
extern bool isabort();

int GetNextNode(double array[], int arraysize, double position, int style)
{
  double length = 0;
  if (style == LINESTYLESOLID)
    return 0;
  for (int i = 0; i < arraysize; i++)
    {
      length += array[i];
      if (position < length)
	return i;
    }
  return -1;
}

double GetLengthFromStartNode(double array[], int suffix, double OldAmari)
{
  int i;
  double length = 0.0;
  for (i = 0; i <= suffix; i++)
    length += array[i];
  
  return (length - OldAmari);
}

int code(double x, double y, intrect *AxisRect)
{
  return (x<(double)AxisRect->left)     | (x>(double)AxisRect->right)<<1 |
    (y<(double)AxisRect->top)<<2 | (y>(double)AxisRect->bottom)<<3;
}

void ClipLine(bool flag, doublepoint *p1, doublepoint *p2, intrect *AxisRect, bool clip, FILE* f)
{
  double x1, y1, x2, y2, cx1, cx2, cy1, cy2;
  int c1, c2;
  double dx, dy;
  intrect boundrect;
  
  x1 = p1->x;
  y1 = p1->y;
  x2 = p2->x;
  y2 = p2->y;
  
  if (
      f_inf(x1) || f_nan(x1) ||
      f_inf(x2) || f_nan(x2) ||
      f_inf(y1) || f_nan(y1) ||
      f_inf(y2) || f_nan(y2)
      )
    return ;
  
  
  if (clip != true)
    {
      pmoveto(flag, f, (int)(x1+.5), (int)(y1+.5));
      plineto(flag, f, (int)(x2+.5), (int)(y2+.5));
      boundrect.left   = -MAXSHORT;
      boundrect.right  = MAXSHORT;
      boundrect.top    = -MAXSHORT;
      boundrect.bottom = MAXSHORT;
    }
  else
    {
      boundrect = *AxisRect;
    }

  
  cx1 = (double)boundrect.left;
  cx2 = (double)boundrect.right;
  cy1 = (double)boundrect.top;
  cy2 = (double)boundrect.bottom;

  c1 = code(x1, y1, &boundrect);
  c2 = code(x2, y2, &boundrect);
  
  
  while ((c1 | c2) != 0)
    {
      
      if ((c1 & c2) != 0)
	return;
      
      dx = x2 - x1;
      dy = y2 - y1;
      if (c1 != 0)
	{
	  if ((c1 & 1) != 0)
	    {
	      y1 += dy*(cx1-x1)/dx;
	      x1 = cx1;
	    }
	  else if ((c1 & 2) != 0)
	    {
	      y1 += dy*(cx2-x1)/dx;
	      x1 = cx2;
	    }
	  else if ((c1 & 4) != 0)
	    {
	      x1 += dx*(cy1-y1)/dy;
	      y1 = cy1;
	    }
	  else if ((c1 & 8) != 0)
	    {
	      x1 += dx*(cy2-y1)/dy;
	      y1 = cy2;
	    }
	  c1=code(x1,y1,&boundrect);
	}
      else
	{
	  if ((c2 & 1) != 0)
	    {
	      y2 += dy*(cx1-x2)/dx;
	      x2 = cx1;
	    }
	  else if ((c2 & 2) != 0)
	    {
	      y2 += dy*(cx2-x2)/dx;
	      x2 = cx2;
	    }
	  else if ((c2 & 4) != 0)
	    {
	      x2 += dx*(cy1-y2)/dy;
	      y2 = cy1;
	    }
	  else if ((c2 & 8) != 0)
	    {
	      x2 += dx*(cy2-y2)/dy;
	      y2 = cy2;
	    }
	  c2=code(x2,y2,&boundrect);
	  
	}
    }

   pmoveto(flag, f, (int)(x1+.5), (int)(y1+.5));
   plineto(flag, f, (int)(x2+.5), (int)(y2+.5));
}

int ClipLinePath(doublepoint *p1, doublepoint *p2, intrect *AxisRect)
{
  //return value  0:noclipped  1: starting point clipped  2:end point clipped 3: both clipped  4:out of range -1:false
  double x1,y1,x2,y2,cx1,cx2,cy1,cy2;
  int c1,c2;
  double dx,dy;
  bool startflag, endflag;
  
  startflag  	= false;
  endflag 	= false;
  x1=(double)p1->x;
  y1=(double)p1->y;
  x2=(double)p2->x;
  y2=(double)p2->y;

  if (
      f_inf(x1) || f_nan(x1) ||
      f_inf(x2) || f_nan(x2) ||
      f_inf(y1) || f_nan(y1) ||
      f_inf(y2) || f_nan(y2)
      )
    return -1;
  
  cx1=(double)AxisRect->left;
  cx2=(double)AxisRect->right;
  cy1=(double)AxisRect->top;
  cy2=(double)AxisRect->bottom;
  
  c1=code(x1,y1,AxisRect);
  c2=code(x2,y2,AxisRect);
  
  
  while ((c1 | c2) != 0)
    {
      
      if ((c1 & c2) != 0) return 4;
      dx=x2-x1;
      dy=y2-y1;
      if (c1 != 0)
	{
	  if ((c1 & 1) != 0)
	    {
	      y1 += dy*(cx1-x1)/dx;
	      x1 = cx1;
	    }
	  else if ((c1 & 2) != 0)
	    {
	      y1 += dy*(cx2-x1)/dx;
	      x1 = cx2;
	    }
	  else if ((c1 & 4) != 0)
	    {
	      x1 += dx*(cy1-y1)/dy;
	      y1 = cy1;
	    }
	  else if ((c1 & 8) != 0)
	    {
	      x1 += dx*(cy2-y1)/dy;
	      y1 = cy2;
	    }
	  c1=code(x1,y1,AxisRect);
	  startflag = true;
	}
      else
	{
	  if ((c2 & 1) != 0)
	    {
	      y2 += dy*(cx1-x2)/dx;
	      x2 = cx1;
	    }
	  else if ((c2 & 2) != 0)
	    {
	      y2 += dy*(cx2-x2)/dx;
	      x2 = cx2;
	    }
	  else if ((c2 & 4) != 0)
	    {
	      x2 += dx*(cy1-y2)/dy;
	      y2 = cy1;
	    }
	  else if ((c2 & 8) != 0)
	    {
	      x2 += dx*(cy2-y2)/dy;
	      y2 = cy2;
	    }
	  c2=code(x2,y2,AxisRect);
	  endflag = true;
	}
    }
  
  p1->x = (int)(x1+.5);
  p1->y = (int)(y1+.5);
  p2->x = (int)(x2+.5);
  p2->y = (int)(y2+.5);
  
  if (startflag == true)
    {
      if (endflag == true)
	return 3;
      else
	return 1;
    }
  else
    {
      if (endflag == true)
	return 2;
      else
	return 0;
    }
}

void StyleLinePath(bool flag, FILE* f, doublepointarray *array, int no_of_point, int style, int width, bool Clip, intrect *AxisRect, intrect *boundrect)
{
  int LineSpace[20];                    // 0,2,4... Line   1,3,5... Space
  int LineSpaceNo;                      // must be even no
  double RealLineSpace[20];
  double CycleLength;                   // length of 1 cycle
  double SegmentLength;                 //length of segment
  double PositionInSegment;	        //position in segment
  double PhaseInSegment;
  double OldAmari;
  double cosT, sinT;
  double ftemp1, ftemp2;
  int    suffix;
  int i, clipflag;
  intpoint TempPoint, PrevPoint;
  doublepoint TempDoublePoint, TempPrevDoublePoint, TempCurDoublePoint, StartInSegment, EndInSegment;
  bool InitPointflag = false;
  intrect ClipRect;

  if (boundrect != 0)
    ClipRect = *boundrect;

  //definition of LineSpace
  switch (style)
    {
    case LINESTYLENULL:
      return;
    case LINESTYLESOLID:
      LineSpace[0] = 10000; LineSpace[1] = 0; LineSpaceNo = 2;
      break;
    case LINESTYLELONGDASH:
      LineSpace[0] = 10; LineSpace[1] = 10; LineSpaceNo = 2;
      break;
    case LINESTYLEDASH:
      LineSpace[0] = 5; LineSpace[1] = 5; LineSpaceNo = 2;
      break;
    case LINESTYLESHORTDASH:
      LineSpace[0] = 3 ; LineSpace[1] = 3 ; LineSpaceNo = 2;
      break;
    case LINESTYLEDOT:
      LineSpace[0] = 1 ; LineSpace[1] = 3 ; LineSpaceNo = 2;
      break;
    case LINESTYLEDASHDOTDASH:
      LineSpace[0] = 10; LineSpace[1] = 3;
      LineSpace[2] = 1 ; LineSpace[3] = 3;
      LineSpaceNo = 4;
      break;
    case LINESTYLEDASHDOTDOTDASH:
      LineSpace[0] = 10; LineSpace[1] = 3;
      LineSpace[2] = 1 ; LineSpace[3] = 3;
      LineSpace[4] = 1 ; LineSpace[5] = 3;
      LineSpaceNo = 6;
      break;
    default:
      return;
    }

  //transform to real line and space
  for (i = 0; i < LineSpaceNo; i++)
    {
      if (width > 0)
	RealLineSpace[i] = (double)LineSpace[i] * (double)width;
      else
	RealLineSpace[i] = (double)LineSpace[i] * 20.0;
    }
  
  //calculate 1 cycle length
  CycleLength = 0.0;
  for (i = 0; i < LineSpaceNo; i++)
    CycleLength += RealLineSpace[i];

  // setcliprect
  if (Clip)
    ClipRect = *AxisRect;
  else
    {
      ClipRect.left = -30000; ClipRect.top = -30000;
      ClipRect.right = 30000; ClipRect.bottom = 30000;
    }

  //	tdc.BeginPath();
  
  OldAmari = 0.0;
  
  for (i = 0; i < no_of_point; i++)
    {
      if (isabort())
	goto j2;

      //determin initpoint and moveto initpoint
      if (InitPointflag == false)
        {
	  if (
	      !f_inf((*array)[i].x)     && !f_nan((*array)[i].x) &&
	      !f_inf((*array)[i].y)     && !f_nan((*array)[i].y)
	      )
            {
	      TempPoint.x = (int)(*array)[i].x;
	      TempPoint.y = (int)(*array)[i].y;

	      pmoveto(flag, f, TempPoint.x, TempPoint.y);
	      
	      InitPointflag = true;

            }
	  else
	    goto j2;
        }
      
      if (
	  !f_inf((*array)[i].x)     && !f_nan((*array)[i].x) &&
	  !f_inf((*array)[i + 1].x) && !f_nan((*array)[i + 1].x) &&
	  !f_inf((*array)[i].y)     && !f_nan((*array)[i].y) &&
	  !f_inf((*array)[i + 1].y) && !f_nan((*array)[i + 1].y)
	  )
	{
	  StartInSegment = (*array)[i];
	  EndInSegment = (*array)[i + 1];
	  clipflag = ClipLinePath(&StartInSegment, &EndInSegment, &ClipRect);
	  
	  TempPrevDoublePoint = StartInSegment;
	  switch(clipflag)
            {
	    case -1:
	    case 4:	//out of range
	      goto j1;
	    case 0:	//no clipped
	    case 2:     //end point clipped
	      break;
	    case 1:    	//start point clipped
	    case 3:     //both points clipped
	      TempPoint.x = (int)StartInSegment.x;
	      TempPoint.y = (int)StartInSegment.y;
	      pmoveto(flag, f, TempPoint.x, TempPoint.y);
	      break;
            }
	  
	  TempPrevDoublePoint = StartInSegment;
	  
	  ftemp1 = EndInSegment.x - StartInSegment.x;
	  ftemp2 = EndInSegment.y - StartInSegment.y;
	  SegmentLength = sqrt(ftemp1 * ftemp1 + ftemp2 * ftemp2);
	  PositionInSegment = 0.0;
	  PhaseInSegment = OldAmari;
	  
	  if (SegmentLength > 0)
	    {
	      cosT = ftemp1 / SegmentLength;
	      sinT = ftemp2 / SegmentLength;
	      
	      suffix = GetNextNode(RealLineSpace, LineSpaceNo, PhaseInSegment, style);
	      PositionInSegment = GetLengthFromStartNode(RealLineSpace, suffix, OldAmari);
	      while (PositionInSegment < SegmentLength)
		{

		  TempDoublePoint.x = (double)StartInSegment.x + PositionInSegment * cosT;
		  TempDoublePoint.y = (double)StartInSegment.y + PositionInSegment * sinT;
		  
		  TempCurDoublePoint = TempDoublePoint;
		  PrevPoint.x		  = (int)TempPrevDoublePoint.x;
		  PrevPoint.y		  = (int)TempPrevDoublePoint.y;
		  TempPoint.x		  = (int)TempCurDoublePoint.x;
		  TempPoint.y		  = (int)TempCurDoublePoint.y;
		  
		  if (suffix % 2 == 0)  //line
		      plineto(flag, f, TempPoint.x, TempPoint.y);
		  else				//space
		      pmoveto(flag, f, TempPoint.x, TempPoint.y);
		  
		  suffix ++;
		  suffix = suffix % LineSpaceNo;
		  PositionInSegment += RealLineSpace[suffix];
		  PrevPoint = TempPoint;
		  TempPrevDoublePoint = TempDoublePoint;
		}
	      
	      //next node
	      TempDoublePoint.x = (double)EndInSegment.x;
	      TempDoublePoint.y = (double)EndInSegment.y;
	      
	      TempCurDoublePoint = TempDoublePoint;
	      
	      PrevPoint.x		  = (int)TempPrevDoublePoint.x;
	      PrevPoint.y		  = (int)TempPrevDoublePoint.y;
	      TempPoint.x		  = (int)TempCurDoublePoint.x;
	      TempPoint.y		  = (int)TempCurDoublePoint.y;
	      
	      if (suffix % 2 == 0)  //line
		  plineto(flag, f, TempPoint.x, TempPoint.y);
	      else				//space
		  pmoveto(flag, f, TempPoint.x, TempPoint.y);
	      
	      OldAmari = SegmentLength - floor(SegmentLength / CycleLength) * CycleLength + OldAmari;
	      OldAmari = OldAmari - floor(OldAmari / CycleLength) * CycleLength;
	      PrevPoint = TempPoint;
	      TempPrevDoublePoint = TempDoublePoint;
	    }
	  
	j1:;
        }
      else if (
	       !f_inf((*array)[i + 1].x) && !f_nan((*array)[i + 1].x) &&
	       !f_inf((*array)[i + 1].y) && !f_nan((*array)[i + 1].y)
	       )
	{
	  TempPoint.x = (int)(*array)[i + 1].x;
	  TempPoint.y = (int)(*array)[i + 1].y;
	  if (ClipRect.contains(TempPoint.x, TempPoint.y))
	      pmoveto(flag, f, TempPoint.x, TempPoint.y);
	  else
	      pmoveto(flag, f, TempPoint.x, TempPoint.y);

	  TempPrevDoublePoint = (*array)[i + 1];
        }
    j2:;
    }
}












