/*----------------------------------------------------------------------------
                        tokenbuff.cc (buffer for token)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <string>
#include "tokenbuff.h"

tokenbuff::tokenbuff()
{
  priority = 0;
  status = 0;
  value = new buffarray();
  label = new std::string();
  strstatus = 0;
  bracestatus = 0;
  bracenestlevel = 0;
  parentheseslevel = 0;
  jumpindex = 0;
  lineno = 0;
  count = -1;
}

tokenbuff::tokenbuff(tokenbuff *buff)
{
  priority = buff->priority;
  status = buff->status;
  value = new buffarray();
  *value = *buff->value;
  label = new std::string();
  *label = *buff->label;
  strstatus = buff->strstatus;
  bracestatus = buff->bracestatus;
  bracenestlevel = buff->bracenestlevel;
  parentheseslevel = buff->parentheseslevel;
  jumpindex = buff->jumpindex;
  lineno = buff->lineno;
  count = buff->count;
}

tokenbuff::~tokenbuff()
{
  delete value;
  delete label;
}

tokenbuff& tokenbuff::operator = (const tokenbuff&  buff)
{
  priority = buff.priority;
  status = buff.status;
  *value = *buff.value;
  *label = *buff.label;
  strstatus = buff.strstatus;
  bracestatus = buff.bracestatus;
  bracenestlevel = buff.bracenestlevel;
  parentheseslevel = buff.parentheseslevel;
  jumpindex = buff.jumpindex;
  lineno = buff.lineno;
  count = buff.count;
  return *this;
}


stackarray::stackarray(int len)
{
  int i;
  
  arraylength = len;
  increment = 100;

  bufflength = (arraylength / increment + 1) * increment;

  buff =   new tokenbuff * [bufflength];
  memset(buff, 0, bufflength * sizeof(int));
  for (i = 0; i < arraylength; i++)
    buff[i] = new tokenbuff();
}

stackarray::~stackarray()
{
  int i;
  for (i = 0; i < arraylength; i++)
    delete buff[i];

  delete [] buff;

}

tokenbuff& stackarray::operator [] (int pos)
{
  return *buff[pos];
}


bool stackarray::setdata(int pos, tokenbuff *val)
{
  if (pos < arraylength)
    {
      delete buff[pos];
      buff[pos] =  new tokenbuff(val);
      return true;
    }
  else
    return false;
  return true;
}

tokenbuff stackarray::getdata(int pos)
{
  if (pos < arraylength)
    return *buff[pos];
  else
    return 0;
}

int stackarray::getarraylength()
{
  return arraylength;
}

bool stackarray::setarray(int length)
{
  int i;

  setbuff(length);
  arraylength = length;

  for (i = 0; i < arraylength; i++)
    buff[i] = new tokenbuff();

  return true;
}

bool stackarray::setbuff(int length)
{
  int i;

  for (i = 0; i < arraylength; i++)
    delete buff[i];
  delete [] buff;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  buff =  new tokenbuff * [bufflength];
  memset(buff, 0, bufflength * sizeof(int));
  return true;
}


bool stackarray::push(int priority,
		      int status,
		      std::string *value,
		      std::string *label,
		      char strstatus,
		      char bracestatus,
		      int bracenestlevel,
		      int parentheseslevel,
		      int jumpindex,
		      int lineno,
		      int count
		      )
{
  if (arraylength + 1 >= bufflength)
    changebufflength(arraylength + 1);
  arraylength++;
//printf("%d, %d\n", arraylength, bufflength);

  buff[arraylength - 1] = new tokenbuff();
  
  buff[arraylength - 1]->priority         = priority;
  buff[arraylength - 1]->status           = status;
  buff[arraylength - 1]->value->add(value->c_str());
  *buff[arraylength - 1]->label           = *label;
  buff[arraylength - 1]->strstatus        = strstatus;
  buff[arraylength - 1]->bracestatus      = bracestatus;
  buff[arraylength - 1]->bracenestlevel   = bracenestlevel;
  buff[arraylength - 1]->parentheseslevel = parentheseslevel;
  buff[arraylength - 1]->jumpindex        = jumpindex;
  buff[arraylength - 1]->lineno           = lineno;
  buff[arraylength - 1]->count            = count;
  return true;
}

bool stackarray::push1(tokenbuff *val)
{
  if (arraylength < bufflength)
    arraylength++;
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
    }
  
  buff[arraylength - 1] = new tokenbuff();
  *buff[arraylength - 1] = *val;
  return true;
}


bool stackarray::pop()
{
  remove (arraylength - 1);
  return true;
}

bool stackarray::remove(int pos)
{
  int i;
  tokenbuff temp;

  for (i = pos; i < arraylength - 1; i++)
    {
      //setdata(i, &getdata(i + 1));
      *buff[i] = *buff[i + 1];
    }
  delete buff[arraylength - 1];
  arraylength--;
  return true;
}

bool stackarray::changebufflength(int newlength)
{
  int i;
  tokenbuff **temptokenarray;

  if (newlength <= bufflength)
    return true;

  bufflength = (newlength / increment + 1) * increment;
  temptokenarray =   new tokenbuff * [bufflength];
  memset(temptokenarray, 0, bufflength * sizeof(int));
  
  for (i = 0; i < arraylength; i++)
    temptokenarray[i] = buff[i];

  buff = temptokenarray;
  return true;
}


vstackarray::vstackarray(int len)
{
  arraylength = len;
  increment = 100;

  bufflength = (arraylength / increment + 1) * increment;

  buff =   new tokenbuff * [bufflength];
  memset(buff, 0, bufflength * sizeof(int));
}

vstackarray::~vstackarray()
{
  delete [] buff;
}

tokenbuff& vstackarray::operator [] (int pos)
{
  return *buff[pos];
}

tokenbuff vstackarray::getdata(int pos)
{
  if (pos < arraylength)
    return *buff[pos];
  else
    return 0;
}

int vstackarray::getarraylength()
{
  return arraylength;
}

bool vstackarray::setbuff(int length)
{
  delete [] buff;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  buff =  new tokenbuff * [bufflength];
  memset(buff, 0, bufflength * sizeof(int));
  return true;
}

bool vstackarray::push1(tokenbuff *val)
{
  if (arraylength < bufflength)
    arraylength++;
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
    }
  
  buff[arraylength - 1] = val;
  return true;
}


bool vstackarray::pop()
{
  remove (arraylength - 1);
  return true;
}

bool vstackarray::remove(int pos)
{
  int i;
  tokenbuff temp;

  for (i = pos; i < arraylength - 1; i++)
      buff[i] = buff[i + 1];

  arraylength--;
  return true;
}

bool vstackarray::changebufflength(int newlength)
{
  int i;
  tokenbuff **temptokenarray;

  if (newlength <= bufflength)
    return true;

  bufflength = (newlength / increment + 1) * increment;
  temptokenarray =   new tokenbuff * [bufflength];
  memset(temptokenarray, 0, bufflength * sizeof(int));
  
  for (i = 0; i < arraylength; i++)
    temptokenarray[i] = buff[i];

  buff = temptokenarray;
  return true;
}






