/*----------------------------------------------------------------------------
                  tokenbuff.h (definitions of token structure)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__tokenbuff_h)
#define __tokenbuff_h

#include <string>
#include "buffarray.h"

class tokenbuff
{
 public:

  tokenbuff();
  tokenbuff(tokenbuff *buff);
  ~tokenbuff();
    

  int priority;   	/*  priority of operation                  */
  int status;           /*  token type                             */
  buffarray *value;     /*  token value or arrayvariant index      */
  std::string *label;        /*  variant name or function name or       */
  char strstatus;       /*  string status                          */
  char bracestatus;     /*  brace status                           */
  int bracenestlevel;   /*  brace nest level                       */
  int parentheseslevel; /*  parentheses nest level                 */
  int jumpindex;        /*  jump index in case of '}'              */
  int lineno;           /*  lineno                                 */
  int count;            /*  count  used for foreach                */

  tokenbuff& operator = (const tokenbuff& buff);
};

class stackarray
{
 public:

  stackarray(int len);
  ~stackarray();

  bool setdata(int pos, tokenbuff *val);
  tokenbuff getdata(int pos);
  int getarraylength();     
  bool setarray(int length);
  bool setbuff(int length);
  bool push(int priority,
	    int status,
	    std::string *value,
	    std::string *label,
	    char strstatus,
	    char bracestatus,
	    int bracenestlevel,
	    int parentheseslevel,
	    int jumpindex,
	    int lineno,
	    int count
	    );
  bool push1(tokenbuff *val);
  bool pop();
  bool remove(int pos);

  tokenbuff& operator [] (int pos);

 private:
  int arraylength;          /* array length of filled array */
  int bufflength;           /* buffer length of array */
  tokenbuff **buff;
  int increment;
  bool changebufflength(int length);
};


class vstackarray
{
 public:

  vstackarray(int len);
  ~vstackarray();

  tokenbuff getdata(int pos);
  int getarraylength();     
  bool setbuff(int length);
  bool push1(tokenbuff *val);
  bool pop();
  bool remove(int pos);

  tokenbuff& operator [] (int pos);

 private:
  int arraylength;          /* array length of filled array */
  int bufflength;           /* buffer length of array */
  tokenbuff **buff;
  int increment;
  bool changebufflength(int length);
};


#endif
