/*----------------------------------------------------------------------------
                           topaz.cc (main routine of topaz)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 99


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "topazcurses.h"
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <signal.h>
#include <getopt.h>
//char *optarg;
#include "script.h"
#include "frame.h"
#include "menudef.h"
#if defined(TOPAZ_SJIS)
#include "jmenudef_sjis.h"
#else
#include "jmenudef_euc.h"
#endif
#include "cui.h"
#include "msgbuff.h"
#include "filepoint.h"

extern char *toshortfilename(char *in);
extern int linelength(char *str, int offset, int displength);
extern bool getmacroname(std::string *file, std::string *out);

// sub-window pointer
WINDOW  *menuwin = 0, *msgwin = 0, *sepwin = 0;
// mainwindow size
int mainwinlines, mainwincols;
// window lines
# define MINMENUWINLINES 17
int menuwinlines = MINMENUWINLINES;
int sepwinlines =1;
int tabnumber = 5;
// menuitem
int itemwidth = 11;
int itemspace = 2;
// maxrow
int maxrow;
int leftmargin = 1;
int topmargin = 1;
// buffersize 
int msgbuffsize = 100;


// pointer to graph object
graph* gra = 0;

// pointer to msgbuff
msgbuff* msgb = 0;

// path
char confpath[1000];
char tmppath[1000];
char pid[100];

// filehandle buff
FileHandleArray* fhvuff = 0;

// exist cuimenu
bool isexistcuimenu = false;

// pointer to environment variables
char **envp;

// japanesecuimenu
bool japanesecui = false;

#if defined(TOPAZ_PDCURSES)
#define TABSIZE  0
#endif

#define MENU_HOTKEY_FORECOLOR 11
#define MENU_HOTKEY_BACKCOLOR 12
#define MENU_HOTKEY_COLORPAIR 1
bool has_color = true;
int hotkeyattr;

void updatemenu();
int mainmenuitempos(int i);
void updatepop();


int topazgetch(WINDOW *win)
{
  return wgetch(win);
}

void clearmenuwin()
{
  for (int i = 0; i < mainwinlines; i++)
    {
      for (int j = 0; j < mainwincols; j++)
	{
	  wmove(menuwin, i, j); wprintw(menuwin, " ");
	}
    }
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);
  wrefresh(menuwin);
}


void exitcuimenu()
{
  if (msgb != 0)
    {
      def_prog_mode();
      endwin();
    }
}

void restorecuimenu()
{
  if (msgb != 0)
    {
      wnoutrefresh(menuwin); wnoutrefresh(msgwin);
      wnoutrefresh(sepwin);
      wnoutrefresh(stdscr);
      doupdate();
    }
}

int abs_wtov_x(int wx)
{
  return wx;
}

int abs_wtov_y(int wy)
{
  return wy;
}

int abs_scale(int x)
{
  return x;
}

bool fileexist(char *file)
{
  FILE *h;
  if ((h = fopen(file, "r")) != NULL)
    {
      fclose(h);
      return true;
    }
  return false;
}

void message(const char *msg)
{
  if (msgb == 0)
    fprintf(stdout, (char *)msg);
  else
    msgb->print(msg);
}

int emessage(int lineno, char *mes, const char *com, char *scrname)
{
  std::string tempstr;
  char stemp[100];

  tempstr = std::string(scrname);
  tempstr += std::string(":");
  sprintf(stemp, "%d: ", lineno);
  tempstr += std::string(stemp);
  tempstr += std::string("Error!: ");
  tempstr += std::string(mes);
  tempstr += std::string(" '");
  tempstr += std::string(com);
  tempstr += std::string("'\n");

  if (msgb == 0)
    fprintf(stderr, (char *)tempstr.c_str());
  else
    msgb->print(tempstr.c_str());

  return false;
}


void showsepwin()
{
  char stemp[1000], stemp1[1000];
  char dest[] = " Topaz | ";
  int len = strlen(dest);
  strcpy(stemp, dest);
  strncpy(stemp1, toshortfilename(gra->graphname), 75 - len);
  stemp[sizeof(stemp1) - 1] = 0;
  strcat(stemp, stemp1);

  if (sepwin == 0)
    return;

  wattron(sepwin, A_REVERSE);
  for (int i = 0; i < mainwincols; i++)
    {
      wmove(sepwin, 0, i);
      if (i < (int)strlen(stemp))
          wprintw(sepwin, "%c", stemp[i]);
      else
          wprintw(sepwin, " ");
    }
  wattroff(sepwin, A_REVERSE);
  wrefresh(sepwin);

}

int truetitlelength(const char *s)
{
  int count = 0;
  char *p = (char *)s;
  while (*p != 0)
    {
      if (*p != '&')
	count++;
      p++;
    }
  return count;
}

void itemshowcurrentmenu(CUIMENU *menu, int maxitem, int addoffset)
{
  char mask[100];
  int x, y;
  int width = 3 * itemwidth + 3 * itemspace;
  int oset, count = 0;
  char stemp[20];
  char *p;
  bool keyflag = false;
  script *scr;
  std::string str1, str2;
  int casoffsety = 2;
  char menuline1[] = "-----------------------------------------------------------------------------";
  char menuline2[] = "=============================================================================";
  int labelwidth = itemwidth * 3 / 2;

  if (menuwin == 0)
    return;

  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");

  oset = 0;

  clearmenuwin();
  //wclear(menuwin);
  for (int i = 1; i <= maxitem; i++)
    {
      keyflag = false;
      y = (i - 1) / (maxrow / 3);
      x = (i - 1) - (maxrow / 3) * y;

      wmove(menuwin, y + topmargin + addoffset + casoffsety,
	    x * width + leftmargin);
      wprintw(menuwin, mask);
      
      p = (char *)menu[i].item;
      count = 0;

      while (*p != 0)
	{
	  if (*p != '&')
	    {
	      stemp[0] = *p; stemp[1] = 0;
	      wmove(menuwin, y + topmargin + addoffset + casoffsety,
		    x * width + leftmargin + oset + count++);
	      wprintw(menuwin, stemp);
	      
	      if (keyflag)
		{
		  wattroff(menuwin, hotkeyattr);
		  keyflag = false;
		}
	    }
	  else
	    {
	      wattron(menuwin, hotkeyattr);
	      keyflag = true;
	    }
	  p++;
	}
      
      // display value
      str1 = std::string ("");
      scr = new script(gra);
      str2 = std::string(menu[i].getvalue);
      scr->execscript(&str2, &str1, 0);
      delete scr;
      strncpy(stemp, str1.c_str(), sizeof(stemp));
      stemp[linelength(stemp, 0, sizeof(stemp))] = 0;

      wattron(menuwin, A_BOLD);
      wmove(menuwin, y + topmargin + addoffset + casoffsety,
	    x * width + oset + (labelwidth + 1) + leftmargin + oset);
      printtext(menuwin, stemp);
      wattroff(menuwin, A_BOLD);

    }

  wattron(menuwin, A_BOLD);
  wmove(menuwin, 1, leftmargin);
  wprintw(menuwin, menu[0].item);
  wattroff(menuwin, A_BOLD);
  wmove(menuwin, 0, leftmargin);
  wprintw(menuwin, menuline1);
  wmove(menuwin, 2, leftmargin);
  wprintw(menuwin, menuline2);

  wrefresh(menuwin);
}


void itemshowcursor(CUIMENU *menu, int itemno, bool flag, int addoffset)
{
  char mask[100];
  int x, y;
  int width = itemwidth * 3 + itemspace * 3;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;
  int casoffsety = 2;
  int labelwidth = itemwidth * 3 / 2;

  if (menuwin == 0)
    return;

  y = (itemno - 1) / (maxrow / 3);
  x = (itemno - 1) - (maxrow / 3) * y;

  strcpy(mask, "");
  for (int i = 0; i < labelwidth; i++)
    strcat(mask, " ");

  oset = 0;
  
  if (flag)
      wattron(menuwin, A_REVERSE);
  
  wmove(menuwin, y + topmargin + addoffset + casoffsety,
	x * width + oset + leftmargin);
  wprintw(menuwin, mask);

  p = (char *)menu[itemno].item;
  count = 0;
  while (*p != 0)
    {
      if (*p != '&')
	{
	  stemp[0] = *p; stemp[1] = 0;
	  wmove(menuwin, y + topmargin + addoffset + casoffsety,
		x * width + leftmargin + oset + count++);
	  wprintw(menuwin, stemp);

	  if (keyflag)
	    {
	      if (!flag)
		wattroff(menuwin, hotkeyattr);
	      keyflag = false;
	    }
	}
      else
	{
	  if (!flag)
	    wattron(menuwin, hotkeyattr);
	  keyflag = true;
	}
      p++;
    }
  
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);

  if (flag)
      wattroff(menuwin, A_REVERSE);

  wrefresh(menuwin);

}


int itemcursormove (CUIMENU *menu, int *itemno, int maxitem, int addoffset)
{
  int key, y;
  char *p;
  script *scr;
  std::string str;
  
  if (menuwin == 0)
    return LOCAL_ESCAPE_KEY;

  cbreak();
  noecho();
  if (*itemno < 0 || *itemno > maxitem)
    *itemno = 0;

  while (1)
    {

      y = *itemno;
      itemshowcursor(menu, y, true, addoffset);
      key = topazgetch(menuwin);
      itemshowcursor(menu, y, false, addoffset);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  *itemno = (*itemno > 1) ? *itemno - 1: maxitem;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  *itemno = (*itemno < maxitem) ? *itemno + 1: 1;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  *itemno = (*itemno + maxrow / 3 <= maxitem) ? *itemno + maxrow / 3: 1;
	  break;
	case KEY_UP:
	case CTRL('p'):
	  *itemno = (*itemno - maxrow / 3 >= 1) ? *itemno - maxrow / 3: maxitem;
	  break;
	case CTRL('a'):
	  *itemno = 1;
	  break;
	case CTRL('e'):
	  *itemno = maxitem;
	  break;
	  //select
	case ' ':
	  //paint
	case CTRL('v'):
	  scr = new script(gra);
	  str = std::string("$graph->paintall(V0);");
	  scr->execscript(&str, 0, 0);
	  delete scr;
	  break;
	  // return to top menu
	case CTRL('t'):
	  *itemno = -1;
	  return LOCAL_LF_KEY;
	  // scrolldown messagewindow
 	case '[':
	  msgb->previous(1);
	  break;
	  // scrollup messagewindow
	case ']':
	  msgb->next(1);
	  break;
	  // return to previous menu
	case CTRL('g'):
	case '.':
	  return LOCAL_ESCAPE_KEY;
	}

      if ( 'a' < key < 'z')
	{
	  for (int i = 1; i <= maxitem; i++)
	    {
	      p = (char *)menu[i].item;
	      while (*p != 0)
		{
		  if (*p == '&')
		    {
		      p++;
		      if ((*p == key) || ((*p - 'A' + 'a') == key))
			{
			  *itemno = i;
			  return (LOCAL_LF_KEY);
			}
		    }
		  p++;
		}
	    }
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }
  itemshowcursor(menu, y, true, addoffset);

  return (key);
}





/*                                            */
/*   base of various menus for selecting item */
/*                                            */

int itemmenuexec(CUIMENU *menu, int *selectno)
{
  CUIMENU *m;
  int itemlength = 0;
  int itemno;
  int flag1;
  int selectedkey = 1;

  if (menuwin == 0)
    return false;
  
  m = menu;
  for (; m->item; m++)
    {
      if (itemlength < (int)strlen(m->item))
	itemlength = strlen(m->item);
    }
  itemno = m - menu - 1;
  if (itemno < 1)
    return false;

  do
    {

      itemshowcurrentmenu(menu, itemno, 0);
      flag1 = itemcursormove (menu, &selectedkey, itemno, 0);


      //return to top menu
      if (selectedkey == -1)
	{
	    *selectno = M_MAIN;
	    return LOCAL_LF_KEY;
	}
      else if (flag1 == LOCAL_LF_KEY && !menu[selectedkey].submenu)
	{
	  execitem(menu, &selectedkey);

	  if (menu[selectedkey].id == M_QUIT)
	  {
	    *selectno = M_QUIT;
	    return LOCAL_LF_KEY;
	  }
	}
      *selectno = menu[selectedkey].id;

    }while (flag1 != LOCAL_ESCAPE_KEY);
  return flag1;
}

int framemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JFrame, &selectno);
  else
    return itemmenuexec(Frame, &selectno);
}

int xbasemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JXBase, &selectno);
  else
    return itemmenuexec(XBase, &selectno);
}
int xrangemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JXRange, &selectno);
  else
    return itemmenuexec(XRange, &selectno);
}
int xtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JXMajortick, &selectno);
  else
    return itemmenuexec(XMajortick, &selectno);
}
int xsubtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JXMinortick, &selectno);
  else
    return itemmenuexec(XMinortick, &selectno);
}
int xlabelmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JXLabel, &selectno);
  else
    return itemmenuexec(XLabel, &selectno);
}

int ybasemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JYBase, &selectno);
  else
    return itemmenuexec(YBase, &selectno);
}
int yrangemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JYRange, &selectno);
  else
    return itemmenuexec(YRange, &selectno);
}
int ytickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JYMajortick, &selectno);
  else
    return itemmenuexec(YMajortick, &selectno);
}
int ysubtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JYMinortick, &selectno);
  else
    return itemmenuexec(YMinortick, &selectno);
}
int ylabelmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JYLabel, &selectno);
  else
    return itemmenuexec(YLabel, &selectno);
}

int tbasemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTBase, &selectno);
  else
    return itemmenuexec(TBase, &selectno);
}
int trangemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTRange, &selectno);
  else
    return itemmenuexec(TRange, &selectno);
}
int ttickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTMajortick, &selectno);
  else
    return itemmenuexec(TMajortick, &selectno);
}
int tsubtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTMinortick, &selectno);
  else
    return itemmenuexec(TMinortick, &selectno);
}
int tlabelmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTLabel, &selectno);
  else
    return itemmenuexec(TLabel, &selectno);
}

int rbasemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JRBase, &selectno);
  else
    return itemmenuexec(RBase, &selectno);
}
int rrangemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JRRange, &selectno);
  else
    return itemmenuexec(RRange, &selectno);
}
int rtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JRMajortick, &selectno);
  else
    return itemmenuexec(RMajortick, &selectno);
}
int rsubtickmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JRMinortick, &selectno);
  else
    return itemmenuexec(RMinortick, &selectno);
}
int rlabelmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JRLabel, &selectno);
  else
    return itemmenuexec(RLabel, &selectno);
}

int plotmappingmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JMapping, &selectno);
  else
    return itemmenuexec(Mapping, &selectno);
}
int plotlinemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JPlotline, &selectno);
  else
    return itemmenuexec(Plotline, &selectno);
}
int plotmarkermenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JPlotmarker, &selectno);
  else
    return itemmenuexec(Plotmarker, &selectno);
}
int plotmemdatamenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JMemdata, &selectno);
  else
    return itemmenuexec(Memdata, &selectno);
}
int loaddatamenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JLoaddata, &selectno);
  else
    return itemmenuexec(Loaddata, &selectno);
}

int partsstylemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JPartsstyle, &selectno);
  else
    return itemmenuexec(Partsstyle, &selectno);
}

int textstylemenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JTextstyle, &selectno);
  else
    return itemmenuexec(Textstyle, &selectno);
}

int functionmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JFunction, &selectno);
  else
    return itemmenuexec(Function, &selectno);
}

int linearfitmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JLinear, &selectno);
  else
    return itemmenuexec(Linear, &selectno);
}

int nonlinearfitmenu()
{
  int selectno;
  if (japanesecui)
    return itemmenuexec(JNonlinear, &selectno);
  else
    return itemmenuexec(Nonlinear, &selectno);
}



/********************************************************************/
/**************** popup menu ****************************************/
/********************************************************************/

typedef struct popstack
{
  CUIMENU *menu;
  int item;
  int maxitem;
}POPSTACK;

POPSTACK pop[20];
int popstackno = -1;

int getmaxpopupitem(CUIMENU *menu, int maxitem)
{
  int length;
  int maxlength = 0;
  char *p;

  for (int i = 1; i <= maxitem; i++)
    {
      p = (char *)menu[i].item;
      length = 0;
      while (*p != 0)
	{
	  if (*p != '&')
	    length++;
	  p++;
	}
      if (length > maxlength)
	maxlength = length;
    }
  return maxlength;
}


void showpopcurrentmenu(CUIMENU *menu, int pos, int maxitem, int poplevel, int addoffset)
{
  char mask[100];
  int x, y;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;
  int maxlength;
  char itemback[100], nextitemback[100], endback[100], sepback[100], startback[100];
  int casoffsetx, casoffsety;

  if (menuwin == 0)
    return;

  maxlength = getmaxpopupitem(menu, maxitem);
  casoffsetx = 1;
  for(int i = 1; i < poplevel; i++)
    casoffsetx += getmaxpopupitem(pop[i].menu, pop[i].maxitem) + 5 - 2;

  // itemback
  strcpy(itemback, "| ");
  for (int i = 0; i < maxlength; i++)
    strcat(itemback, " ");
  strcat(itemback, "  |");

  // nextitemback
  strcpy(nextitemback, "| ");
  for (int i = 0; i < maxlength; i++)
    strcat(nextitemback, " ");
  strcat(nextitemback, " >|");
  
  // sepback
  strcpy(sepback, "|-");
  for (int i = 0; i < maxlength; i++)
    strcat(sepback, "-");
  strcat(sepback, "--|");

  // startback
  strcpy(startback, "+-");
  for (int i = 0; i < maxlength; i++)
    strcat(startback, "-");
  strcat(startback, "--+");

  // endback
  strcpy(endback, "+-");
  for (int i = 0; i < maxlength; i++)
    strcat(endback, "-");
  strcat(endback, "--+");

  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");

  if (poplevel > 1)
    {
      if (pop[poplevel - 1].item + maxitem + topmargin + addoffset + 2 <=  menuwinlines)
	casoffsety = pop[poplevel - 1].item - 1;
      else
	casoffsety = menuwinlines - maxitem - topmargin - addoffset - 3;
    }
  else
    casoffsety = 0;

  //  wclear(menuwin);

  wmove(menuwin, 1 + topmargin + addoffset + casoffsety,
	mainmenuitempos(pos) + leftmargin + casoffsetx);
  wprintw(menuwin, startback);

  for (int i = 1; i <= maxitem; i++)
    {
      keyflag = false;
      y = i + 1;
      x = mainmenuitempos(pos);

      wmove(menuwin, y + topmargin + addoffset + casoffsety,
	    x + leftmargin + casoffsetx);

      if (menu[i].id != M_SEP)
	{
	  if (menu[i].submenu)
	    wprintw(menuwin, nextitemback);
	  else
	    wprintw(menuwin, itemback);
	}
      else
	wprintw(menuwin, sepback);
      
      oset = 2;
      p = (char *)menu[i].item;
      count = 0;

      if (menu[i].id != M_SEP)
	{
	  while (*p != 0)
	    {
	      if (*p != '&')
		{
		  stemp[0] = *p; stemp[1] = 0;
		  wmove(menuwin, y + topmargin + addoffset + casoffsety,
			x + leftmargin + oset + count++ + casoffsetx);
		  wprintw(menuwin, stemp);

		  if (keyflag)
		    {
		      wattroff(menuwin, hotkeyattr);
		      keyflag = false;
		    }
		}
	      else
		{
		  wattron(menuwin, hotkeyattr);
		  keyflag = true;
		}
	      p++;
	    }
	  
	}
    }

  y = maxitem + 2;
  x = mainmenuitempos(pos);
  wmove(menuwin, y + topmargin + addoffset + casoffsety,
	x + leftmargin + casoffsetx);
  wprintw(menuwin, endback);

  //  wrefresh(menuwin);
}

void showpopcursor(CUIMENU *menu, int pos, int maxitem, int poplevel, int itemno, bool flag, int addoffset)
{
  char mask[100];
  int x, y;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;
  int maxlength;
  int casoffsetx, casoffsety;

  if (menuwin == 0)
    return;

  maxlength = getmaxpopupitem(menu, maxitem);
  casoffsetx = 1;
  for(int i = 1; i < poplevel; i++)
    casoffsetx += getmaxpopupitem(pop[i].menu, pop[i].maxitem) + 5 - 2;

  y = itemno + 1;
  x = mainmenuitempos(pos);

  if (poplevel > 1)
    {
      if (pop[poplevel - 1].item + maxitem + topmargin + addoffset + 2 <=  menuwinlines)
	casoffsety = pop[poplevel - 1].item - 1;
      else
	casoffsety = menuwinlines - maxitem - topmargin - addoffset - 3;
    }
  else
    casoffsety = 0;
  
  strcpy(mask, "");
  for (int i = 0; i < maxlength + 1; i++)
    strcat(mask, " ");
  
  if (flag)
    wattron(menuwin, A_REVERSE);
  
  oset = 2;

  wmove(menuwin, y + topmargin + addoffset + casoffsety
	, x + leftmargin + oset + casoffsetx);
  wprintw(menuwin, mask);
  
  p = (char *)menu[itemno].item;
  count = 0;
  while (*p != 0)
  {
    if (*p != '&')
      {
	stemp[0] = *p; stemp[1] = 0;
	wmove(menuwin, y + topmargin + addoffset + casoffsety
	      , x + leftmargin + oset + count++ + casoffsetx);
	wprintw(menuwin, stemp);
	
	if (keyflag)
	  {
	    if (!flag)
	      wattroff(menuwin, hotkeyattr);
	    keyflag = false;
	  }
      }
    else
      {
	  if (!flag)
	    wattron(menuwin, hotkeyattr);
	  keyflag = true;
      }
    p++;
  }
  
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);

  if (flag)
      wattroff(menuwin, A_REVERSE);

  //  wrefresh(menuwin);

}


int popcursormove (CUIMENU *menu, int pos, int *itemno, int maxitem, int addoffset)
{
  int key, y;
  char *p;
  script *scr;
  std::string str;
  
  if (menuwin == 0)
    return LOCAL_ESCAPE_KEY;

  cbreak();
  noecho();
  if (*itemno < 0 || *itemno > maxitem)
    *itemno = 0;

  while (1)
    {

      y = *itemno;
      showpopcursor(menu, pos, maxitem, popstackno, y, true, addoffset);
      key = topazgetch(menuwin);
      showpopcursor(menu, pos, maxitem, popstackno, y, false, addoffset);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  key = MENU_LEFT;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  key = MENU_RIGHT;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  if (*itemno < maxitem)
	    {
	      if (menu[*itemno + 1].id != M_SEP)
		*itemno = *itemno + 1;
	      else
		{
		  if (*itemno + 1 < maxitem && menu[*itemno + 2].id != M_SEP)
		    *itemno = *itemno + 2;
		  else
		    *itemno = 1;
		}
	    }
	  else
	    *itemno = 1;
	  break;
	case KEY_UP:
	case CTRL('p'):
	  if (*itemno > 1)
	    {
	      if (menu[*itemno - 1].id != M_SEP)
		*itemno = *itemno - 1;
	      else
		{
		  if (*itemno - 1 > 1 && menu[*itemno - 2].id != M_SEP)
		    *itemno = *itemno - 2;
		  else
		    *itemno = maxitem;
		}
	    }
	  else
	    *itemno = maxitem;
	  break;
	case CTRL('a'):
	  *itemno = 1;
	  break;
	case CTRL('e'):
	  *itemno = maxitem;
	  break;
	  //paint
	case CTRL('v'):
	case ' ':
	  scr = new script(gra);
	  str = std::string("$graph->paintall(V0);");
	  scr->execscript(&str, 0, 0);
	  delete scr;
	  break;
	  // return to top menu
	case CTRL('t'):
	case '/':
	  *itemno = -1;
	  return LOCAL_LF_KEY;
	  // scrolldown messagewindow
 	case '[':
	  msgb->previous(1);
	  break;
	  // scrollup messagewindow
	case ']':
	  msgb->next(1);
	  break;
	  // return to previous menu
	case CTRL('g'):
	case '.':
	  return LOCAL_ESCAPE_KEY;
	}

      if ( 'a' < key < 'z')
	{
	  for (int i = 1; i <= maxitem; i++)
	    {
	      p = (char *)menu[i].item;
	      while (*p != 0)
		{
		  if (*p == '&')
		    {
		      p++;
		      if ((*p == key) || ((*p - 'A' + 'a') == key))
			{
			  *itemno = i;
			  return (LOCAL_LF_KEY);
			}
		    }
		  p++;
		}
	    }
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY
	  || key == MENU_RIGHT || key == MENU_LEFT)
	break;
    }
  showpopcursor(menu, pos, maxitem, popstackno, y, true, addoffset);

  return (key);
}

/********************************************************************/
/**************** main  menu ****************************************/
/********************************************************************/

int mainmenuitempos(int i)
{
  int pos = 0;
  if (i == 1)
    return 3;
  else
    {
      pos = 3;
      for(int j = 1; j < i; j++)
	{
	  if (japanesecui)
	    pos += truetitlelength(JMain[j].item) + 6;
	  else
	    pos += truetitlelength(Main[j].item) + 6;
	}
    }
  return pos;
}

void showmaincurrentmenu(CUIMENU *menu, int maxitem, int addoffset)
{
  char mask[20];
  int x, y;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;
  int maxlength;
  char menuline1[] = "-----------------------------------------------------------------------------";
  char menuline2[] = "=============================================================================";

  if (menuwin == 0)
    return;

  maxlength = getmaxpopupitem(menu, maxitem);

  strcpy(mask, "   ");
  for (int i = 0; i < maxlength; i++)
    strcat(mask, " ");
  strcat(mask, "   ");

  wmove(menuwin, 0, leftmargin);
  wprintw(menuwin, menuline1);
  wmove(menuwin, 2, leftmargin);
  wprintw(menuwin, menuline2);

  for (int i = 1; i <= maxitem; i++)
    {
      keyflag = false;
      y = 0;
      x = mainmenuitempos(i);

      wmove(menuwin, y + topmargin + addoffset, x + leftmargin);
      wprintw(menuwin, mask);
      
      oset = 3;
      p = (char *)menu[i].item;
      count = 0;

      while (*p != 0)
	{
	  if (*p != '&')
	    {
	      stemp[0] = *p; stemp[1] = 0;
	      wmove(menuwin, y + topmargin + addoffset, x + leftmargin + oset + count++);
	      wprintw(menuwin, stemp);

	      if (keyflag)
		{
		  wattroff(menuwin, hotkeyattr);
		  keyflag = false;
		}
	    }
	  else
	    {
	      wattron(menuwin, hotkeyattr);
	      keyflag = true;
	    }
	  p++;
	}

    }

  //  wrefresh(menuwin);
}


void showmaincursor(CUIMENU *menu, int itemno, bool flag, int addoffset)
{
  char mask[20];
  int x, y;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;

  if (menuwin == 0)
    return;
  
  strcpy(mask, "   ");
  for (int i = 0; i < truetitlelength(menu[itemno].item); i++)
    strcat(mask, " ");
  strcat(mask, "   ");
  
  y = 0;
  x = mainmenuitempos(itemno);

  if (flag)
    wattron(menuwin, A_REVERSE);
  
  wmove(menuwin, y + topmargin + addoffset, x + leftmargin);
  wprintw(menuwin, mask);
  
  oset = 3;
  p = (char *)menu[itemno].item;
  count = 0;
  while (*p != 0)
  {
    if (*p != '&')
      {
	stemp[0] = *p; stemp[1] = 0;
	wmove(menuwin, y + topmargin + addoffset, x + leftmargin + oset + count++);
	wprintw(menuwin, stemp);
	
	if (keyflag)
	  {
	    if (!flag)
	      wattroff(menuwin, hotkeyattr);
	    keyflag = false;
	  }
      }
    else
      {
	  if (!flag)
	    wattron(menuwin, hotkeyattr);
	  keyflag = true;
      }
    p++;
  }
  
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);

  if (flag)
      wattroff(menuwin, A_REVERSE);

  //  wrefresh(menuwin);

}


int maincursormove (CUIMENU *menu, int *itemno, int maxitem, int addoffset)
{
  int key, y;
  char *p;
  script *scr;
  std::string str;
  
  if (menuwin == 0)
    return LOCAL_ESCAPE_KEY;

  cbreak();
  noecho();
  if (*itemno < 0 || *itemno > maxitem)
    *itemno = 0;

  while (1)
    {

      y = *itemno;
      showmaincursor(menu, y, true, addoffset);
      key = topazgetch(menuwin);
      showmaincursor(menu, y, false, addoffset);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  *itemno = (*itemno > 1) ? *itemno - 1: maxitem;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  *itemno = (*itemno < maxitem) ? *itemno + 1: 1;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  *itemno = (*itemno + maxrow <= maxitem) ? *itemno + maxrow: 1;
	  break;
	case KEY_UP:
	case CTRL('p'):
	  *itemno = (*itemno - maxrow >= 1) ? *itemno - maxrow: maxitem;
	  break;
	case CTRL('a'):
	  *itemno = 1;
	  break;
	case CTRL('e'):
	  *itemno = maxitem;
	  break;
	  //paint
	case CTRL('v'):
	case ' ':
	  scr = new script(gra);
	  str = std::string("$graph->paintall(V0);");
	  scr->execscript(&str, 0, 0);
	  delete scr;
	  break;
	  // scrolldown messagewindow
 	case '[':
	  msgb->previous(1);
	  break;
	  // scrollup messagewindow
	case ']':
	  msgb->next(1);
	  break;
	  // return to previous menu
	case CTRL('g'):
	  return LOCAL_ESCAPE_KEY;
	}

      if ( 'a' < key < 'z')
	{
	  for (int i = 1; i <= maxitem; i++)
	    {
	      p = (char *)menu[i].item;
	      while (*p != 0)
		{
		  if (*p == '&')
		    {
		      p++;
		      if ((*p == key) || ((*p - 'A' + 'a') == key))
			{
			  *itemno = i;
			  return (LOCAL_LF_KEY);
			}
		    }
		  p++;
		}
	    }
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }
  showmaincursor(menu, y, true, addoffset);

  return (key);
}


/********************************************************************/
/**************** common menu ***************************************/
/********************************************************************/

void showcurrentmenu(CUIMENU *menu, int maxitem, int addoffset)
{
  char mask[20];
  int x, y;
  int width = itemwidth + itemspace;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;

  if (menuwin == 0)
    return;

  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");


  clearmenuwin();
  //wclear(menuwin);
  for (int i = 1; i <= maxitem; i++)
    {
      keyflag = false;
      y = (i - 1) / maxrow;
      x = (i - 1) - maxrow * y;

      wmove(menuwin, y + topmargin + addoffset, x * width + leftmargin);
      wprintw(menuwin, mask);
      
      oset = (itemwidth - truetitlelength(menu[i].item)) / 2;
      p = (char *)menu[i].item;
      count = 0;

      while (*p != 0)
	{
	  if (*p != '&')
	    {
	      stemp[0] = *p; stemp[1] = 0;
	      wmove(menuwin, y + topmargin + addoffset, x * width + leftmargin + oset + count++);
	      wprintw(menuwin, stemp);

	      if (keyflag)
		{
		  wattroff(menuwin, hotkeyattr);
		  keyflag = false;
		}
	    }
	  else
	    {
	      wattron(menuwin, hotkeyattr);
	      keyflag = true;
	    }
	  p++;
	}

    }

  wrefresh(menuwin);
}


void showcursor(CUIMENU *menu, int itemno, bool flag, int addoffset)
{
  char mask[20];
  int x, y;
  int width = itemwidth + itemspace;
  int oset, count = 0;
  char stemp[10];
  char *p;
  bool keyflag = false;

  if (menuwin == 0)
    return;

  y = (itemno - 1) / maxrow;
  x = (itemno - 1) - maxrow * y;
  
  strcpy(mask, "");
  for (int i = 0; i < itemwidth; i++)
    strcat(mask, " ");
  
  
  if (flag)
    wattron(menuwin, A_REVERSE);
  
  
  wmove(menuwin, y + topmargin + addoffset, x * width + leftmargin);
  wprintw(menuwin, mask);
  
  oset = (itemwidth - truetitlelength(menu[itemno].item)) / 2;
  p = (char *)menu[itemno].item;
  count = 0;
  while (*p != 0)
  {
    if (*p != '&')
      {
	stemp[0] = *p; stemp[1] = 0;
	wmove(menuwin, y + topmargin + addoffset, x * width + leftmargin + oset + count++);
	wprintw(menuwin, stemp);
	
	if (keyflag)
	  {
	    if (!flag)
	      wattroff(menuwin, hotkeyattr);
	    keyflag = false;
	  }
      }
    else
      {
	  if (!flag)
	    wattron(menuwin, hotkeyattr);
	  keyflag = true;
      }
    p++;
  }
  
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);

  if (flag)
      wattroff(menuwin, A_REVERSE);

  wrefresh(menuwin);

}


int cursormove (CUIMENU *menu, int *itemno, int maxitem, int addoffset)
{
  int key, y;
  char *p;
  script *scr;
  std::string str;
  
  if (menuwin == 0)
    return LOCAL_ESCAPE_KEY;

  cbreak();
  noecho();
  if (*itemno < 0 || *itemno > maxitem)
    *itemno = 0;

  while (1)
    {

      y = *itemno;
      showcursor(menu, y, true, addoffset);
      key = topazgetch(menuwin);
      showcursor(menu, y, false, addoffset);

      switch (key)
	{
	case KEY_LEFT:
	case CTRL('b'):
	  *itemno = (*itemno > 1) ? *itemno - 1: maxitem;
	  break;
	case KEY_RIGHT:
	case CTRL('f'):
	  *itemno = (*itemno < maxitem) ? *itemno + 1: 1;
	  break;
	case KEY_DOWN:
	case CTRL('n'):
	  *itemno = (*itemno + maxrow <= maxitem) ? *itemno + maxrow: 1;
	  break;
	case KEY_UP:
	case CTRL('p'):
	  *itemno = (*itemno - maxrow >= 1) ? *itemno - maxrow: maxitem;
	  break;
	case CTRL('a'):
	  *itemno = 1;
	  break;
	case CTRL('e'):
	  *itemno = maxitem;
	  break;
	  //paint
	case CTRL('v'):
	case ' ':
	  scr = new script(gra);
	  str = std::string("$graph->paintall(V0);");
	  scr->execscript(&str, 0, 0);
	  delete scr;
	  break;
	  // return to top menu
	case CTRL('t'):
	  *itemno = -1;
	  return LOCAL_LF_KEY;
	  // scrolldown messagewindow
 	case '[':
	  msgb->previous(1);
	  break;
	  // scrollup messagewindow
	case ']':
	  msgb->next(1);
	  break;
	  // return to previous menu
	case CTRL('g'):
	  return LOCAL_ESCAPE_KEY;
	}

      if ( 'a' < key < 'z')
	{
	  for (int i = 1; i <= maxitem; i++)
	    {
	      p = (char *)menu[i].item;
	      while (*p != 0)
		{
		  if (*p == '&')
		    {
		      p++;
		      if ((*p == key) || ((*p - 'A' + 'a') == key))
			{
			  *itemno = i;
			  return (LOCAL_LF_KEY);
			}
		    }
		  p++;
		}
	    }
	}
      
      if (key == LOCAL_ESCAPE_KEY || key == LOCAL_LF_KEY)
	break;
    }
  showcursor(menu, y, true, addoffset);

  return (key);
}

void getwininfo()
{
  mainwinlines = LINES;
  mainwincols  = COLS;
  tabnumber    = TABSIZE;
}

int menuexec(CUIMENU *menu, int *selectno, int item)
{
  CUIMENU *m;
  int itemlength = 0;
  int itemno;
  int flag1 , flag2;
  int selectedkey = 1;

  if (menuwin == 0)
    return false;

  popstackno++;
  pop[popstackno].menu = menu;

  m = menu;
  for (; m->item; m++)
    {
      if (itemlength < (int)strlen(m->item))
	itemlength = strlen(m->item);
    }
  itemno = m - menu - 1;
  if (itemno < 1)
    return false;

  pop[popstackno].maxitem = itemno;
  pop[popstackno].item = selectedkey;

  do
    {
      if (menu[*selectno].submenu)
	{
	  updatepop();
	  flag1 = popcursormove (menu, item, &selectedkey, itemno, 0);
	}
      else
	{
	  updatepop();
	  flag1 = popcursormove (menu, item, &selectedkey, itemno, 0);
	}

      pop[popstackno].item = selectedkey;

      //return to top menu
      if (selectedkey == -1)
	{
	  popstackno = 0;
	  pop[popstackno].item = 1;
	  *selectno = M_MAIN;
	  return LOCAL_LF_KEY;
	}
      else if ((flag1 == LOCAL_LF_KEY)
	       && menu[selectedkey].submenu)
	{
	  flag2 = menuexec(menu[selectedkey].submenu, selectno, item);
	  if ((*selectno == M_QUIT || *selectno == M_MAIN)
	      && flag2 == LOCAL_LF_KEY)
	    {
	      popstackno = 0;
	      pop[popstackno].item = 1;
	      return LOCAL_LF_KEY;
	    }
	}
      else if ((flag1 == LOCAL_LF_KEY)
	       && !menu[selectedkey].submenu)
	{
	  execitem(menu, &selectedkey);

	  if (menu[selectedkey].id == M_QUIT)
	  {
	    popstackno = 0;
	    pop[popstackno].item = 1;
	    *selectno = M_QUIT;
	    return LOCAL_LF_KEY;
	  }
	}

      *selectno = menu[selectedkey].id;

    }while(!(flag1 == LOCAL_ESCAPE_KEY ||
	     ((flag1 == MENU_LEFT || flag1 == MENU_RIGHT)
	      && popstackno == 1)));

  popstackno--;
  return flag1;

} 



int mainmenuexec(CUIMENU *menu, int *selectno)
{
  CUIMENU *m;
  int itemlength = 0;
  int itemno;
  int flag1 , flag2;
  int selectedkey = 1;
  std::string str, str1;
  bool bufferflashflag = true;

  if (menuwin == 0)
    return false;

  popstackno++;
  m = menu;
  for (; m->item; m++)
    {
      if (itemlength < (int)strlen(m->item))
	itemlength = strlen(m->item);
    }
  itemno = m - menu - 1;
  if (itemno < 1)
    return false;

  pop[popstackno].menu = menu;
  pop[popstackno].maxitem = itemno;
  pop[popstackno].item = selectedkey;

  updatemenu();
  do
    {
      updatepop();
      flag1 = maincursormove (menu, &selectedkey, itemno, 0);
      pop[popstackno].item = selectedkey;

      //message area flush. only one time 
      if (bufferflashflag)
	{
	  msgb->flush();
	  bufferflashflag = false;
	}
      
      //return to top menu
      if (selectedkey == -1)
	{
	    *selectno = M_MAIN;
	    selectedkey = 1;
	}
      else if (flag1 == LOCAL_LF_KEY  && menu[selectedkey].submenu)
	{
	  do
	    {
	      flag2 = menuexec(menu[selectedkey].submenu, selectno, selectedkey);
	      if (*selectno == M_QUIT && flag2 == LOCAL_LF_KEY)
		{
		  if (gra->updated)
		    {
		      str = std::string("0");
		      str1 = std::string("The graph is updated. Exit Topaz?");
		      exec_yesno_select(&str, &str1);
		      if (str == std::string("1"))
			return LOCAL_LF_KEY;
		    }
		  else
		    {
		      str = std::string("0");
		      str1 = std::string("Exit Topaz?");
		      exec_yesno_select(&str, &str1);
		      if (str == std::string("1"))
			return LOCAL_LF_KEY;
		    }
		}
	      else if (*selectno == M_MAIN && flag2 == LOCAL_LF_KEY)
		{
		  *selectno = M_MAIN;
		  selectedkey = 1;
		}
	      else
		*selectno = menu[selectedkey].id;

	      switch (flag2)
		{
		case MENU_LEFT:
		  if (selectedkey > 1)
		    selectedkey--;
		  break;
		case MENU_RIGHT:
		  if (selectedkey < itemno)
		    selectedkey++;
		  break;
		}
	      pop[popstackno].item = selectedkey;
	      //	      showmaincursor(menu, *selectno, true, 0);
	    }
	  while(flag2 == MENU_LEFT || flag2 == MENU_RIGHT);

	}
      else if (flag1 == LOCAL_LF_KEY && !menu[selectedkey].submenu)
	{
	  execitem(menu, &selectedkey);

	  if (menu[selectedkey].id == M_QUIT)
	  {
	    *selectno = M_QUIT;
	    return LOCAL_LF_KEY;
	  }
	  *selectno = menu[selectedkey].id;
	}
    }while (1);
  return flag1;

} 

void updatemenu()
{
  wclear(menuwin);
  showmaincurrentmenu(pop[0].menu, pop[0].maxitem, 0);
  showmaincursor(pop[0].menu, pop[0].item, true, 0);
  for (int i = 1; i <= popstackno; i++)
    {
      showpopcurrentmenu(pop[i].menu, pop[0].item, pop[i].maxitem, i, 0);
      showpopcursor(pop[i].menu, pop[0].item, pop[i].maxitem, i, pop[i].item, true, 0);
    }
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);
  wrefresh(menuwin);
}

void updatepop()
{
  for (int i = 0; i < mainwinlines; i++)
    {
      for (int j = 0; j < mainwincols; j++)
	{
	  wmove(menuwin, i, j); wprintw(menuwin, " ");
	}
    }
  showmaincurrentmenu(pop[0].menu, pop[0].maxitem, 0);
  showmaincursor(pop[0].menu, pop[0].item, true, 0);
  for (int i = 1; i <= popstackno; i++)
    {
      showpopcurrentmenu(pop[i].menu, pop[0].item, pop[i].maxitem, i, 0);
      showpopcursor(pop[i].menu, pop[0].item, pop[i].maxitem, i, pop[i].item, true, 0);
    }
  wmove(menuwin, 0, 0);
  wprintw(menuwin, " ");
  wmove(menuwin, 0, 0);
  wrefresh(menuwin);
}

bool iscommentline(const char *s)
{
  char *p;
  
  p = (char *)s;
  while (*p != 0)
    {
      if ((*p != ' ') && (*p != '\t'))
        {
	  if (*p == '#')
	    return true;
	  else
	    return false;
        }
      else
	p++;
    }
  return false;
}

bool separatercitem(char *line)
{
  char *p = line;
  char tmpline[1024], item[1000], value[1000], stemp[3];
  int count = 0;
  bool flag = false;
  int fcolor, bcolor, itemp;

  //remove space, tab, '\r' and '\n'
  strcpy(tmpline, "");
  while(*p != 0)
    {
      if (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n')
	{
	  stemp[0] = *p;
	  stemp[1] = 0;
	  strcat(tmpline, stemp);
	}
      p++;
    }

  // commentline
  if (iscommentline(tmpline))
    return true;

  // null line
  if (strcmp(tmpline, "") == 0)
    return true;

  // count number of '='
  count = 0;
  p = tmpline;
  while(*p != 0)
    {
      if (*p == '=')
	count++;
      p++;
    }

  if (count != 1)
    return false;

  //separate tmpline into item and value
  strcpy(item, ""); strcpy(value, "");
  p = tmpline;
  while(*p != 0)
    {
      if (*p == '=')
	  flag = true;
      else
	{
	  stemp[0] = *p;
	  stemp[1] = 0;
	  if (flag)
	    strcat(value, stemp);
	  else
	    strcat(item, stemp);
	}
      p++;
    }

  if (strcmp(item, "menu_hotkey_color") == 0)
    {
      if (sscanf(value, "%d/%d", 
		 &fcolor, &bcolor) != 2)
	return false;
      init_pair(MENU_HOTKEY_COLORPAIR, fcolor, bcolor);
    }
  else if (strcmp(item, "menu_color_enable") == 0)
    {
      if (sscanf(value, "%d", 
		 &itemp) != 1)
	return false;

      switch(itemp)
	{
	case 1:
	  if (has_color)
	    has_color = true;
	  break;
	default:
	    has_color = false;
	  break;
	}
    }
  else if (strcmp(item, "menuarea_height") == 0)
    {
      if (sscanf(value, "%d", 
		 &itemp) != 1)
	return false;

      if (17 <= itemp && itemp <= 20)
	menuwinlines = itemp;
      else
	menuwinlines    = 17;
    }
  else if (strcmp(item, "messagebuffer_size") == 0)
    {
      if (sscanf(value, "%d", 
		 &itemp) != 1)
	return false;

      if (100 <= itemp && itemp <= 10000)
	msgbuffsize = itemp;
      else
	msgbuffsize = 100;
    }
  else if (strcmp(item, "kanjimenu") == 0)
    {
      if (sscanf(value, "%d", 
		 &itemp) != 1)
	return false;

      if (itemp == 1)
	japanesecui = true;
      else
	japanesecui = false;
    }
  return true;
}

bool loadrc()
{
  char buffer[1024];
  FILE *f;
  std::string out, str;

  str = std::string("cui.rc");
  if (!getmacroname(&str, &out))
    return false;

  if ((f = fopen((char *)out.c_str(), "r")) == NULL)
    return false;

  while(fgets(buffer, sizeof(buffer), f) != NULL)
    {
      if (!separatercitem(buffer))
	{
	  fclose(f);
	  return false;
	}
    }

  fclose (f);
  return true;
}

void cuimenu()
{
  int selectno = 0;
  char stemp1[200], stemp2[200];
  FILE *ttyin = NULL, *ttyout = NULL;

  ttyin = fopen("/dev/tty", "r");
  ttyout = fopen("/dev/tty", "w");

#if !defined(TOPAZ_PDCURSES)
  set_term(newterm(NULL, ttyout, ttyin));
#else
  initscr();
#endif

  if (has_colors())
    has_color = true;
  else
    has_color = false;


  if (has_color)
    {
      if (start_color() != OK)
	{
	  endwin();
	  fclose(ttyin); fclose(ttyout);
	  fprintf(stderr, "Error!!: could not initialize colors!!\n");
	  return;
	}
    }

  if (!loadrc())
    {
      endwin();
      fclose(ttyin); fclose(ttyout);
      fprintf(stderr, "Error!!: could not load resource file!!\n");
      return;
    }

  if (has_color)
    hotkeyattr = COLOR_PAIR(MENU_HOTKEY_COLORPAIR);
  else
    hotkeyattr = A_UNDERLINE;

  getwininfo();
  if (mainwincols < WINDOWCOL)
    {
      endwin();
      fclose(ttyin); fclose(ttyout);
      fprintf(stderr, "Error!!: Terminal width must be 78 characters or more.\naborted!!\n");
      return;
    }
  else
    mainwincols = WINDOWCOL;
  
  if (mainwinlines < 24)
    {
      endwin();
      fclose(ttyin); fclose(ttyout);
      fprintf(stderr, "Error!!: Terminal height must be 25 lines or more.\naborted!!\n");
      return;
    }
  
  maxrow = mainwincols / (itemwidth + itemspace);
  crmode();

#if !defined(TOPAZ_PDCURSES)
  ESCDELAY = 0;
#endif

  menuwin = newwin(menuwinlines, mainwincols, 0, 0);
  sepwin = newwin(sepwinlines, mainwincols, menuwinlines, 0);
  msgwin = newwin(mainwinlines - menuwinlines - sepwinlines, mainwincols
		    , menuwinlines + sepwinlines, 0);
  msgb = new msgbuff(msgwin, mainwinlines - menuwinlines - sepwinlines, msgbuffsize);
  
  scrollok(msgwin, true);
  cbreak();
  noecho();
  showsepwin();
  
  keypad(menuwin, TRUE);
  keypad(stdscr, TRUE);
  
  gra->versionstring(stemp1);
  sprintf(stemp2, "Welcome to the Topaz system version %s!!\n", stemp1);
  strcat(stemp2, "Topaz comes with ABSOLUTELY NO WARRANTY.\n");
  if (japanesecui)
    strcat(stemp2, "Kanji cuimenu is ready.\n");
  msgb->print(stemp2);

  if (japanesecui)
    mainmenuexec(JMain, &selectno);
  else
    mainmenuexec(Main, &selectno);
  
  delete msgb;
  msgb = 0;

  delwin(msgwin);
  msgwin = 0;

  delwin(sepwin);
  sepwin = 0;

  delwin(menuwin);
  menuwin = 0;

  endwin();
  fclose(ttyin); fclose(ttyout);

  popstackno = -1;

}

int intrflag = false;

void setintrflag(int sig)
{
  signal(sig, SIG_IGN);
  message("User interruption\n");
  intrflag = true;
  (void) signal(sig, setintrflag);
}

void floatingerror(int sig)
{
  signal(sig, SIG_IGN);
  (void) signal(sig, floatingerror);
}

void pipeerror(int sig)
{
  signal(sig, SIG_IGN);
  message("Error: cannot create pipe\n");
  (void) signal(sig, pipeerror);
}

bool isabort()
{
  return intrflag;
}

void resetabort()
{
  intrflag = false;
}


int main (int argc, char *argv[], char *env[])
{
  int c;
  extern int optind;
  std::string out, initscript, argscript = std::string(""), str1;
  buffarray tempbuffarray;
  DIR *dp;
  

  //check environ
  if (getenv("TOPAZINIDIR") == NULL)
    {
      fprintf(stderr, "Error: 'TOPAZINIDIR' is not defined.");
      exit(1);
    }
  if (getenv("TOPAZDIR") == NULL)
    {
      fprintf(stderr, "Error: 'TOPAZDIR' is not defined.");
      exit(1);
    }

  // initialization
  strcpy(confpath, getenv("TOPAZINIDIR"));
  strcpy(tmppath, confpath);
  strcat(tmppath, "/tmp");
  sprintf(pid, "%u_", (unsigned int)getpid());

  //check directories
  if ((dp = opendir(confpath)) == NULL)
    {
      mkdir(confpath, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IROTH);
    }
  else
    closedir (dp);

  if ((dp = opendir(tmppath)) == NULL)
    {
      mkdir(tmppath, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IROTH);
    }
  else
    closedir (dp);
    


  // options
  while ((c = getopt(argc, argv, "vh:b:")) != -1)
    {
      switch(c)
	{
	  // show version and exit;
	case 'v':
	  fprintf(stderr, "\n");
	  fprintf(stderr, "    Topaz Version 3.40 /  Copyright: Hisao Kawaura\n");
	  fprintf(stderr, "\n");
	  fprintf(stderr, "    Options:\n");
	  fprintf(stderr, "       -v:    show current version and options\n");
	  fprintf(stderr, "\n");
	  exit(0);
	  break;
	}
    }
  
  // script exists
  if (optind < argc)
    {
      argscript = std::string(argv[optind++]);
      tempbuffarray.setbuff(0);
      for (; optind < argc; optind++)
	tempbuffarray.add(argv[optind]);
    }

  // environment variables
  envp = env;

  // interrupt
  (void) signal(SIGINT, setintrflag);
  // floating point error
  (void) signal(SIGFPE, floatingerror);
  // write on a pipe with no reader
  (void) signal(SIGPIPE, pipeerror);
  
  // creaete filehandle buffer
  fhvuff  = new FileHandleArray;

  // create graph
  gra = new graph;
  
  script *scr = new script(gra);

  // initialize objects
  if (argscript == std::string(""))
    {
      initscript = std::string("inittopaz");
      switch (scr->execscriptfile(&initscript, &out, 0))
	{
	case false:
	case NOTEXISTSCRIPT:
	  str1 = std::string("Error!!: cannot execute '");
	  str1 += initscript;
	  str1 += std::string("'\n");
	  message((char *)str1.c_str());
	  break;
	}
    }
  // execute loaded script
  else
    {
      switch (scr->execscriptfile(&argscript, &out, &tempbuffarray))
	{
	case NOTEXISTSCRIPT:
	case false:
	  str1 = std::string("Error!!: cannot execute '");
	  str1 += argscript;
	  str1 += std::string("'\n");
	  message((char *)str1.c_str());
	  break;
	}
    }

  delete scr;

  // flush tmp files in '$HOME/.topaz/tmp'
  gra->flushtmpfiles();
  delete gra;
  delete fhvuff;

  exit (0);
}





