/*----------------------------------------------------------------------------
                       tpv2ps.cc (filter from tpv to eps)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 2005


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <math.h>
#include <string>
#include <time.h>
#include <getopt.h>
//char *optarg;

extern int errno;
#include "vectdefs.h"
#include "pointdefs.h"
#include "col.h"
#include "kanji.h"
#include "convtxt.h"

class pscolor
{
public:
  pscolor(){init();}
  double red;
  double green;
  double blue;
  void init(){red = 0.0; green = 0.0; blue = 0.0;}
  pscolor& operator = (const pscolor& p)
    {red = p.red; green = p.green; blue = p.blue; return *this;}
  int operator == ( const pscolor& obj ) const
    { return this == &obj;}
};


double unit = .05;

FILE* in = NULL;
FILE* out = NULL;

#define buffersize 1024

static char entry[13][100] = {
  {"Times-Roman-Q"},
  {"Times-Bold-Q"},
  {"Times-Italic-Q"},
  {"Times-BoldItalic-Q"},
  {"Helvetica-Q"},
  {"Helvetica-Bold-Q"},
  {"Helvetica-Oblique-Q"},
  {"Helvetica-BoldOblique-Q"},
  {"Courier-Q"},
  {"Courier-Bold-Q"},
  {"Courier-Oblique-Q"},
  {"Courier-BoldOblique-Q"},
  {"Symbol"}
};

static char kentry[8][100] = {
  {"Ryumin-Light-H"},
  {"Ryumin-Light-H"},
  {"Ryumin-Light-H"},
  {"Ryumin-Light-H"},
  {"GothicBBB-Medium-H"},
  {"GothicBBB-Medium-H"},
  {"GothicBBB-Medium-H"},
  {"GothicBBB-Medium-H"},
};

double papery;

double convy(double y)
{
  return (papery - y);
}


bool removeret(char *in, char *out)
{
  char buff[1000], stemp[20];
  char *p;
  
  p = in;
  strcpy(buff, "");
  while (*p != 0)
    {
     if ((*p == 0xd) || (*p == 0xa))
        {
	  strcpy (out, buff);
	  return true;
        }
      else
        {
	  stemp[0] = *p;
	  stemp[1] = 0;
	  strcat(buff, stemp);
        }
      p++;
    }
  return false;
}


int vfpf(char *fmt, ...)
{
   va_list argptr;
   int cnt;
   
   va_start(argptr, fmt);
   cnt = vfprintf(out, fmt, argptr);
   va_end(argptr);
   
   return(cnt);
}

void convminus(const char *s)
{
  char *p;
  p = (char *)s;
  while(*p != 0)
    {
      if (*p == '-')
	*p = '~';
      p++;
    }
}

bool iscommentline(const char *s)
{
  char *p;
  
  p = (char *)s;
  while (*p != 0)
    {
      if ((*p != ' ') && (*p != '\t'))
        {
	  if (*p == '#')
	    return true;
	  else
	    return false;
        }
      else
	p++;
    }

  return false;
}

bool getcapcolor(const char *str, pscolor *col)
{
  char *p;
  pscolor tempcol;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;
  do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tempcol.red = atof(buff) / 65535.0;
	  else if(tokencount == 1)
	    tempcol.green = atof(buff) / 65535.0;
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
	  p++;
        }
      else
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
	  p++;
        }
    }while(*p != 0);
  
    if(tokencount == 2)
      {
        tempcol.blue = atof(buff) / 65535.0;
        *col = tempcol;
        return true;
      }
    else
      return false;
}

bool getcapfont(const char *str, Capfont *fo)
{
  char *p;
  Capfont tempfo;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;
  do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tempfo.face = atoi(buff);
	  else if(tokencount == 1)
	    tempfo.size = atoi(buff);
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
	  p++;
        }
      else
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
	  p++;
        }
    }while(*p != 0);
  
    if(tokencount == 2)
      {
        tempfo.angle = atof(buff);
        *fo = tempfo;
        return true;
      }
    else
      return false;
}


bool getcappoint(const char *str, doublepoint *dp)
{
  char *p;
  doublepoint tdp;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;

    do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tdp.x = atof(buff) * unit;
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
	  p++;
        }
      else
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
	  p++;
        }
    }while(*p != 0);
  
    if(tokencount == 1)
    {
        tdp.y = atof(buff) * unit;
        *dp = tdp;
      	return true;
    }
    else
       	return false;

}



bool interpret(void)
{
  int  no;
  char *p;
  char temp[10];
  doublepoint dp;
  bool lineflag = false;
  double arc_x1, arc_x2, arc_y1, arc_y2, arc_left, arc_right, arc_top, arc_bottom;
  double rect_left, rect_right, rect_top, rect_bottom;
  std::string text, temptext;
  char stemp[1000], buff[10000];
  unsigned int kanji;

  //current status
  pscolor forecol, backcol;
  Capfont currentfo;
  double currentlinewidth = 0.0;
  int currentcapstyle = 1;
  int currentjoinstyle = 0;
  int linestyle = LINESTYLESOLID;
  int fillstyle = FILLSTYLESOLID;
  int currentfillrule = FILLRULEEVENODD;

  doublepointarray dpa;

  while (topazfgets(buff, sizeof(buff), in) != NULL)
    {
      if (!iscommentline(buff))
	{
	  p = buff;
	  //command
	  if (*p == '/')
	    {
	      p++;
	      if (strlen(p) == 4)
	        {
		  temp[0] = *p;
		  temp[1] = *(p + 1);
		  temp[2] = *(p + 2);
		  temp[3] = 0;

		  //Moveto
		  if (strcmp(temp, "MVT") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			    return false;
			  else
			    {
			      if (lineflag == false)
				  dpa.add(new doublepoint(dp.x, convy(dp.y)));
			    }
			}
		    }
		  //Lineto
		  else if (strcmp(temp, "LNT") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			      return false;
			  else
			    {
			      if (dpa.getarraylength() < 1000)
				{
				  dpa.add(new doublepoint(dp.x, convy(dp.y)));
				  lineflag = true;
				}
			      else
				{
				  if (dpa.getarraylength() > 0)
				    {
				      vfpf("%f %f M\n", dpa[0].x, dpa[0].y);
				      for (int i = 1; i < dpa.getarraylength(); i++)
					vfpf("%f %f L\n", dpa[i].x, dpa[i].y);
				      vfpf("SRP\n");
				      dpa.setbuff(0);
				      lineflag = false;
				    }
				}
			    }
			}
		    }
		  //Polygon
		  else if (strcmp(temp, "PLY") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  no = atoi(buff);
			  vfpf("BGP\n");
			  for (int i = 0; i < no; i++)
			    {
			      if(topazfgets(buff, sizeof(buff), in) == NULL)
				return false;
			      else
				{
				  if (!getcappoint(buff, &dp))
				    return false;

				  if ( i == 0)
				    vfpf("%f %f M\n", dp.x, convy(dp.y));
				  else
				    vfpf("%f %f L\n", dp.x, convy(dp.y));
				}
			    }
			  vfpf("CLP\n");
			  vfpf("FLP\n");
			  vfpf("SRP\n");
			}
		    }
		  //Ellipse
		  else if (strcmp(temp, "ELP") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			    return false;
			  else
			    {
                              arc_x1 = dp.x;
                              arc_y1 = convy(dp.y);
			      if(topazfgets(buff, sizeof(buff), in) == NULL)
				return false;
			      if (getcappoint(buff, &dp) == false)
				return false;
			      else
				{
				  arc_x2 = dp.x;
				  arc_y2 = convy(dp.y);

				  if (arc_x1 < arc_x2)
				    {
				      arc_left  = arc_x1;
				      arc_right = arc_x2;
				    }
				  else
				    {
				      arc_left  = arc_x2;
				      arc_right = arc_x1;
				    }

				  if (arc_y1 < arc_y2)
				    {
				      arc_top    = arc_y1;
				      arc_bottom = arc_y2;
				    }
				  else
				    {
				      arc_top    = arc_y2;
				      arc_bottom = arc_y1;
				    }

				  vfpf("%f %f %f %f %f \n",
				       0.0,
				       0.0,
				       fabs(arc_right - arc_left) / 2.0,
				       0.0,
				       360.0);
				  vfpf("1 %f \n",
				       fabs((arc_bottom - arc_top) / (arc_right - arc_left)));
				  vfpf("%f %f \n",(arc_left + arc_right) / 2.0,
				       (arc_top + arc_bottom) / 2.0);
				  vfpf("ELS\n");
				}
			    }
			}
		    }
		  //Begin Path
		  else if (strcmp(temp, "BGP") == 0)
		    {
		      vfpf("BGP\n");
		      lineflag = false;
		    }
		  //Close Path
		  else if (strcmp(temp, "CSP") == 0)
		    {
		      // not implemented
		    }
		  //Stroke Path
		  else if (strcmp(temp, "SRP") == 0)
		    {
		      if (dpa.getarraylength() > 0)
			{
			  vfpf("%f %f M\n", dpa[0].x, dpa[0].y);
			  for (int i = 1; i < dpa.getarraylength(); i++)
			    vfpf("%f %f L\n", dpa[i].x, dpa[i].y);
				      
			}
		      vfpf("SRP\n");
		      dpa.setbuff(0);
		      lineflag = false;
		    }
		  // select foreground color
		  else if (strcmp(temp, "FCL") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcapcolor(buff, &forecol) == false)
			    return false;
			  else
			    vfpf("%f %f %f SPC\n"
				 , forecol.red
				 , forecol.green
				 , forecol.blue);
			}
		    }
		  // select background color
		  else if (strcmp(temp, "BCL") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcapcolor(buff, &backcol) == false)
			    return false;
			  {
			    vfpf("%f %f %f SBC\n"
				 , backcol.red
				 , backcol.green
				 , backcol.blue);
			  }
			}
		    }
		  // select font
		  else if (strcmp(temp, "LFT") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcapfont(buff, &currentfo) == false)
			    return false;
			  {
			    if (currentfo.face < 1000)
			      {
				vfpf("/%s findfont [%f 0 0 %f 0 0] makefont setfont\n",
				     entry[currentfo.face],
				     (double)currentfo.size * unit,
				     (double)currentfo.size * unit);
				vfpf("/CFONTFACE /%s def\n", entry[currentfo.face]);
			      }
			    else
			      {
				vfpf("/%s findfont [%f 0 0 %f 0 0] makefont setfont\n",
				     kentry[currentfo.face - 1000],
				     (double)currentfo.size * unit,
				     (double)currentfo.size * unit);
				vfpf("/CFONTFACE /%s def\n", kentry[currentfo.face - 1000]);
			      }

			    vfpf("/CFONTSIZE %f def\n", (double)currentfo.size * unit);
			    vfpf("/CFONTESCAPE %f def\n",currentfo.angle);
			  }
			}
		    }
		  // draw text
		  else if (strcmp(temp, "DTX") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			    return false;
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  removeret((char  *)buff, stemp);

			  if (strlen(stemp) == 1)
			    {
			      if (strcmp(stemp, "(") == 0)
				vfpf(" (\\() %f %f TO\n", dp.x, convy(dp.y));
			      else if (strcmp(stemp, ")") == 0)
				vfpf(" (\\)) %f %f TO\n", dp.x, convy(dp.y));
			      else if (strcmp(stemp, "\\") == 0)
				vfpf(" (\\\\) %f %f TO\n", dp.x, convy(dp.y));
			      else
				vfpf(" (%s) %f %f TO\n", stemp, dp.x, convy(dp.y));
			    }
			  else
			    {
#if defined(TOPAZ_SJIS)
			      sjistojis((unsigned int*)stemp);
#else
			      tojis((unsigned int*)stemp);
#endif
			      kanji = *(unsigned short*)stemp;
			      vfpf(" <%X> %f %f TO\n", kanji, dp.x, convy(dp.y));
			    }
			}
		    }
		  //set line style
		  else if (strcmp(temp, "LSY") == 0)
		    {
			  //thickness
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  currentlinewidth = atof(buff) * unit;
			  //capstyle
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  switch(atoi(buff))
			    {
			    case LINECAPBUTT:
			      currentcapstyle = LINECAPBUTT;
			      break; 
			    case LINECAPROUND:
			      currentcapstyle = LINECAPROUND;
			      break; 
			    case LINECAPPROJECTING:
			      currentcapstyle = LINECAPPROJECTING;
			      break; 
			    }
			  //joinstyle
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  switch(atoi(buff))
			    {
			    case LINEJOINMITER:
			      currentjoinstyle = LINEJOINMITER;
			      break;
			    case LINEJOINROUND:
			      currentjoinstyle = LINEJOINROUND;
			      break;
			    case LINEJOINBEVEL:
			      currentjoinstyle = LINEJOINBEVEL;
			      break;
			    }
			  //linestyle
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  linestyle  = atoi(buff);
			  
			  if (linestyle == LINESTYLENULL)
			    vfpf("1 SNP\n");
			  else
			    vfpf("0 SNP\n");
			    

			  vfpf("%f %d %d LSY\n",
			       currentlinewidth,
			       currentcapstyle,
			       currentjoinstyle);
		    }
		  //set fill style
		  else if (strcmp(temp, "FSY") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  // fillrule
			  switch(atoi(buff))
			    {
			    case 0:
			      currentfillrule  = 0;
			      break;
			    case 1:
			      currentfillrule  = 1;
			      break;
			    }
			  
			  //fillstyle
			  if(topazfgets(buff, sizeof(buff), in) == NULL)
			    return false;
			  else
			    fillstyle  = atoi(buff);
			  
			  if (fillstyle == FILLSTYLENULL)
			    vfpf("1 SNB\n");
			  else
			    vfpf("0 SNB\n");
			}
		    }

		  //setpapersize
		  else if (strcmp(temp, "PAR") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			    return false;
			  else
			    {
			      papery = dp.y;
			      vfpf("%f %f PAI\n", dp.x, dp.y);
			    }
			}
		    }
		  //releaseclip
		  else if (strcmp(temp, "RCP") == 0)
		    {
		      vfpf("RCP\n");
		    }
		  //setcliprect
		  else if (strcmp(temp, "CRT") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp) == false)
			    return false;
			  else
			    {
                              rect_left   = dp.x;
                              rect_top  = convy(dp.y);
			      
			      if(topazfgets(buff, sizeof(buff), in) == NULL)
				return false;
			      if (getcappoint(buff, &dp) == false)
				return false;
			      rect_right  = dp.x;
			      rect_bottom = convy(dp.y);
			      
			      vfpf("%f %f \n", rect_left, rect_top);
			      vfpf("%f %f \n", rect_right, rect_top);
			      vfpf("%f %f \n", rect_right, rect_bottom);
			      vfpf("%f %f \n", rect_left, rect_bottom);
			      vfpf("SCP\n");
			    }
			}
		    }
		}
            }
        }
    }
  return true;
}


bool getpaper()
{
  char buff[1000], temp[20];
  char *p;
  doublepoint dp1, dp2;
  
  while (topazfgets(buff, sizeof(buff), in) != NULL)
    {
      if (!iscommentline(buff))
	{
	  p = buff;
	  //command
	  if (*p == '/')
	    {
	      p++;
	      if (strlen(p) == 4)
		{
		  temp[0] = *p;
		  temp[1] = *(p + 1);
		  temp[2] = *(p + 2);
		  temp[3] = 0;
		  if (strcmp(temp, "PAR") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      else
			{
			  if (getcappoint(buff, &dp1) == false)
			    return false;
			  else
			    {
			      papery = dp1.y;
			      return true;
			    }
			}
		    }
		}
	    }
	}
    }
  return false;
}

bool getbounding(doublerect *outrect)
{
  char buff[1000], temp[20];
  char *p;
  doublepoint dp1, dp2;
  double x1, y1, x2, y2;
  
  while (topazfgets(buff, sizeof(buff), in) != NULL)
    {
      if (!iscommentline(buff))
	{
	  p = buff;
	  //command
	  if (*p == '/')
	    {
	      p++;
	      if (strlen(p) == 4)
		{
		  temp[0] = *p;
		  temp[1] = *(p + 1);
		  temp[2] = *(p + 2);
		  temp[3] = 0;
		  if (strcmp(temp, "BBX") == 0)
		    {
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      if (getcappoint(buff, &dp1) == false)
			return false;
		      if(topazfgets(buff, sizeof(buff), in) == NULL)
			return false;
		      if (getcappoint(buff, &dp2) == false)
			return false;
		      
		      x1 = dp1.x;
		      y1 = convy(dp1.y);
		      x2 = dp2.x;
		      y2 = convy(dp2.y);
		      
		      if (x1 < x2)
			{
			  outrect->left      = x1;
			  outrect->right     = x2;
			}
		      else
			{
			  outrect->left      = x2;
			  outrect->right     = x1;
			}
		      
		      if (y1 < y2)
			{
			  outrect->top       = y1;
			  outrect->bottom    = y2;
			}
		      else
			{
			  outrect->top       = y2;
			  outrect->bottom    = y1;
			}
		      return true;
		    }
		}
	    }
	}
    }
  return false;
}

int main(int argc, char *argv[])
{
  char headerfile[100], buff[10000];
  FILE* headp;
  doublerect brect;
  double bbmargin = 0.0;
  time_t timeval;
  extern int optind;
  int c;
  bool inflag = false; 
  bool outflag = false;
  bool isepsfile = false;

  in  = stdin;
  out = stdout;


  if (getenv("TOPAZDIR") == NULL)
    {
      fprintf(stderr, "TOPAZDIR is not defined.\n");
      exit(1);
    }

  // options
  while ((c = getopt(argc, argv, "o:m:ve")) != -1)
    {
      switch(c)
	{
	  // outputfile
	case 'o':
	  if ((out = fopen(optarg, "w")) == NULL)
	    {
	      fprintf(stderr, "Error: cannot open output file %s\n", optarg);
	      exit(1);
	    }
	  else
	    outflag = true;

	  // bounding box margin
	case 'm':
	  bbmargin = atof(optarg);
	  break;
	  // epsfile
	case 'e':
	  isepsfile = true;
	  break;
	  // help
	case 'v':
	  fprintf(stderr, "tpv2ps Ver. 1.02 (Oct. 11, 2005)\n");
	  fprintf(stderr, "Written by Hisao Kawaura\n");
	  fprintf(stderr, "Output file:         -o (filename)\n");
	  fprintf(stderr, "Bounding box margin: -m (margin [points])\n");
	  fprintf(stderr, "Eps format:          -e \n");
	  exit(0);
	  break;
	}
    }

  // input file exists
  if (optind < argc)
    {
      if ((in = fopen(argv[optind], "r")) == NULL)
	{
	  fprintf(stderr, "Error: cannot open input file '%s'\n", argv[optind]);
	  exit(1);
	}
      else
	  inflag = true;
    }

  strcpy(headerfile, getenv("TOPAZDIR"));
  strcat(headerfile, "/head.ps");
  
  //out header
  vfpf("%%!PS-Adobe-3.0 EPSF-3.0\n");

  // get bounding box
  if (!getpaper())
    {
      fprintf(stderr, "Error: cannot get paper size\n");
      exit(1);
    }
  if (!getbounding(&brect))
    {
      fprintf(stderr, "Error: cannot get bounding box\n");
      exit(1);
    }

  vfpf("%%%%BoundingBox: %d %d %d %d\n", 
       (int)(brect.left - bbmargin), (int)(brect.top - bbmargin),
       (int)(brect.right + bbmargin) + 1, (int)(brect.bottom + bbmargin) + 1);
  vfpf("%%%%Creator: tpv2ps Version 1.02\n");
  vfpf("%%%%Title: topaz\n");

  (void)time(&timeval);
  vfpf("%%%%CreationDate: %s", ctime(&timeval));

  vfpf("%%%%Pages: 1\n");
  vfpf("%%%%EndComments\n");
  
  //input header
  if ((headp = fopen(headerfile, "r")) == NULL)
    {
      fprintf(stderr, "Error: Cannot open header file!!\n");
      return 1;
    }


  while (topazfgets(buff, sizeof(buff), headp) != NULL)
    vfpf(buff);
  fclose(headp);
     
  vfpf("%%%%Page: Graph 1\n");
  vfpf("%%%%BeginPageSetup\n");
  vfpf("TopazDict begin\n");
  vfpf("SavePage\n");
  vfpf("StartPage\n");
  vfpf("%%%%EndPageSetup\n");


  if (interpret() == false)
    {
      if (outflag)
	fclose(out);
      if (inflag)
	fclose(in);
      exit(1);
    }
  
  vfpf("EndPage\n");
  vfpf("RestorePage\n");
  if(!isepsfile)
    vfpf("showpage\n");
  vfpf("end\n");
  vfpf("%%%%Trailer\n");
  vfpf("%%%%EOF\n");

  if (outflag)
    fclose(out);
  if (inflag)
    fclose(in);

  exit(0);
}





