/*----------------------------------------------------------------------------
                        tpvview.cc (viewer for tpv file)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <math.h>
#include <string>
#include <unistd.h>
#include <X11/Xlib.h>

extern int errno;
#include "topazvalues.h"
#include "vectdefs.h"
#include "pointdefs.h"
#include "Xparray.h"
#include "col.h"
#include "kanji.h"
#include "convtxt.h"

Display *disp;
int scr;
Window win, root;
GC gc = (GC)0;
Colormap cm;
XSetWindowAttributes at;

unsigned int w = 640, h = 450;
int x = 100, y = 100;
unsigned int border = 1;
double unit = .05;
Xparray xpa;
XFontStruct *fst;

//current status
Capcolor currentcol, forecol, backcol;
Capfont currentfo;
unsigned int forecolor, backcolor;  
int currentlinewidth = 0;
int currentcapstyle = 1;
int currentjoinstyle = 0;
int linestyle = LINESTYLESOLID;
int fillstyle = FILLSTYLESOLID;
int currentfillrule = FILLRULEEVENODD;
int paintmode = 0;
std::string title;
int paperx, papery;
FILE *in;

double mag = 1;
int shiftx = 0, shifty = 0;
int cdepth = 2;
XColor wcol, bcol;


unsigned int reducedcolor(unsigned int c);

#define buffersize 1024

typedef struct xlfd_table
{
  char name[256];
}XLFDENTRY;

XLFDENTRY entry[] = {
  {"-adobe-times-medium-r-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-times-bold-r-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-times-medium-i-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-times-bold-i-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-helvetica-medium-r-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-helvetica-bold-r-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-helvetica-medium-o-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-helvetica-bold-o-normal--[%s %s %s %s]-*-100-100-p-*-iso8859-1"},
  {"-adobe-courier-medium-r-normal--[%s %s %s %s]-*-100-100-m-*-iso8859-1"},
  {"-adobe-courier-bold-r-normal--[%s %s %s %s]-*-100-100-m-*-iso8859-1"},
  {"-adobe-courier-medium-o-normal--[%s %s %s %s]-*-100-100-m-*-iso8859-1"},
  {"-adobe-courier-bold-o-normal--[%s %s %s %s]-*-100-100-m-*-iso8859-1"},
  {"-adobe-symbol-medium-r-normal--[%s %s %s %s]-*-100-100-p-*-adobe-fontspecific"}
};

char topazfontmap[13][256];
char topazkfontmap[8][256];

char dash1[] = {2,2};

double posix(double x)
{
  double f = (x - shiftx) * mag * unit;
  
  if (f > (double)MAXSHORT)
    f = (double)MAXSHORT;
  else if (f < -(double)MAXSHORT)
    f = -(double)MAXSHORT;

  return f;
}

double posiy(double y)
{
  double f = (y - shifty) * mag * unit;

  if (f > (double)MAXSHORT)
    f = (double)MAXSHORT;
  else if (f < -(double)MAXSHORT)
    f = -(double)MAXSHORT;

  return f;
}


void convminus(const char *s)
{
  char *p;
  p = (char *)s;
  while(*p != 0)
    {
      if (*p == '-')
	*p = '~';
      p++;
    }
}

bool iscommentline(const char *s)
{
  char *p;
  
  p = (char *)s;
  while (*p != 0)
    {
      if ((*p != ' ') && (*p != '\t'))
        {
	  if (*p == '#')
	    return true;
	  else
	    return false;
        }
      else
	p++;
    }
  return false;
}

bool getcapcolor(char *str, Capcolor *col)
{
  char *p;
  Capcolor tempcol;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;
  do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tempcol.red = (unsigned long)atol(buff);
	  else if(tokencount == 1)
	    tempcol.green = (unsigned long)atol(buff);
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
        }
      else if (*p != '\n')
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
        }
      p++;
    }while(*p != 0);
  
    if(tokencount == 2)
      {
        tempcol.blue = (unsigned long)atol(buff);
        *col = tempcol;
        return true;
      }
    else
      return false;
}

bool getcapfont(char *str, Capfont *fo)
{
  char *p;
  Capfont tempfo;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;
  do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tempfo.face = atoi(buff);
	  else if(tokencount == 1)
	    tempfo.size = (int)(atof(buff) * unit * mag);
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
        }
      else if (*p != '\n')
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
        }
      p++;
    }while(*p != 0);
  
    if(tokencount == 2)
      {
        tempfo.angle = atof(buff);
        *fo = tempfo;
        return true;
      }
    else
      return false;
}


bool getcappoint(char *str, doublepoint *dp)
{
  char *p;
  doublepoint tdp;
  char buff[buffersize], stemp[3];
  int tokencount;
  p = (char *)str;
  strcpy(buff, "");
  tokencount = 0;

    do
    {
      if (*p == ',')
        {
	  if(tokencount == 0)
	    tdp.x = posix(atof(buff));
	  else
	    return false;
	  
	  tokencount++;
	  strcpy(buff, "");
        }
      else if (*p != '\n')
        {
	  stemp[0] = *p; stemp[1] = 0;
	  strcat(buff, stemp);
        }
      p++;
    }while(*p != 0);
  
    if(tokencount == 1)
    {
        tdp.y = posiy(atof(buff));
        *dp = tdp;
      	return true;
    }
    else
       	return false;

}


void setforcolor(unsigned int col)
{
  if (paintmode == 0)
    XSetForeground(disp, gc, col);
  else if (paintmode == 1)
    XSetForeground(disp, gc, wcol.pixel  ^ col);
}

bool painting(char* str)
{
  int i, itemp;
  char *p;
  char temp[10];
  doublepoint dp;
  bool lineflag = false;
  int arc_x, arc_y, arc_x1, arc_x2, arc_y1, arc_y2;
  unsigned int arc_w, arc_h;
  std::string text;
  double ftemp1, ftemp2, ftemp3, ftemp4;
  char s1[100], s2[100], s3[100], s4[100], stemp[1000]; 
  char buffer[buffersize];
  XChar2b kanji[2];


  XPoint txp;
  XColor xcol;
  XRectangle xrect;

  if (!iscommentline(str))
    {
      p = str;
      //command
      if (*p == '/')
	{
	  p++;
	  if (strlen(p) == 4)
	    {
	      temp[0] = *p;
	      temp[1] = *(p + 1);
	      temp[2] = *(p + 2);
	      temp[3] = 0;
	      
	      //Moveto
	      if (strcmp(temp, "MVT") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  
		  if (lineflag == false)
		    {
		      txp.x = (short)dp.x;
		      txp.y = (short)dp.y;
		      xpa.add(&txp);
		    }
		}
	      //Lineto
	      else if (strcmp(temp, "LNT") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  
		  txp.x = (short)dp.x;
		  txp.y = (short)dp.y;
		  xpa.add(&txp);
		  lineflag = true;
		}
	      //Polygon
	      else if (strcmp(temp, "PLY") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  
		  itemp = atoi(buffer);
		  xpa.setbuff(0);
		  
		  for (i = 0; i < itemp; i++)
		    {
		      if (topazfgets(buffer, sizeof(buffer), in) == NULL)
			return false;
		      if (getcappoint(buffer, &dp) == false)
			return false;
		      
		      txp.x = (short)dp.x;
		      txp.y = (short)dp.y;
		      xpa.add(&txp);
			  
		    }
		  
		  // fill inside area
		  if (fillstyle == FILLSTYLESOLID)
		    {
		      XSetForeground(disp, gc, backcolor);
		      XFillPolygon(disp, win, gc, xpa.array, xpa.getarraylength(), Complex, CoordModeOrigin);
		    }
		  
		  // draw edge
		  if (linestyle != LINESTYLENULL)
		    {
		      setforcolor(forecolor);
		      
		      // close path
		      txp.x = xpa.array[0].x;
		      txp.y = xpa.array[0].y;
		      xpa.add(&txp);
		      
		      XDrawLines(disp, win, gc, xpa.array, xpa.getarraylength(), CoordModeOrigin);
		    }
		  
		  // flush buffer
		  XFlush(disp);
		  xpa.setbuff(0);
		}
	      //Ellipse
	      else if (strcmp(temp, "ELP") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  
		  arc_x1 = (short)dp.x;
		  arc_y1 = (short)dp.y;
		  
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  
		  arc_x2 = (short)dp.x;
		  arc_y2 = (short)dp.y;
		  
		  if (arc_x1 < arc_x2)
		    {
		      arc_x = arc_x1;
		      arc_w = arc_x2 - arc_x1;
		    }
		  else
		    {
		      arc_x = arc_x2;
		      arc_w = arc_x1 - arc_x2;
		    }
		  
		  if (arc_y1 < arc_y2)
		    {
		      arc_y = arc_y1;
		      arc_h = arc_y2 - arc_y1;
		    }
		  else
		    {
		      arc_y = arc_y2;
		      arc_h = arc_y1 - arc_y2;
		    }
		  
		  // fill inside area
		  if (fillstyle == FILLSTYLESOLID)
		    {
		      XSetForeground(disp, gc, backcolor);
		      XFillArc(disp, win, gc, arc_x, arc_y, arc_w, arc_h, 0 * 64, 360 * 64);
		    }
		  // draw edge
		  if (linestyle != LINESTYLENULL)
		    {
		      setforcolor(forecolor);
		      XDrawArc(disp, win, gc, arc_x, arc_y, arc_w, arc_h, 0 * 64, 360 * 64);
		    }
		}
	      //Begin Path
	      else if (strcmp(temp, "BGP") == 0)
		{
		  lineflag = false;
		}
	      //Close Path
	      else if (strcmp(temp, "CSP") == 0)
		{
		  if (xpa.getarraylength() > 0)
		    {
		      txp.x = xpa.array[0].x;
		      txp.y = xpa.array[0].y;
		      xpa.add(&txp);
		    }
		}
	      //Stroke Path
	      else if (strcmp(temp, "SRP") == 0)
		{
		  setforcolor(forecolor);
		  XDrawLines(disp, win, gc, xpa.array, xpa.getarraylength(), CoordModeOrigin);
		  XFlush(disp);
		  xpa.setbuff(0);
		  lineflag = false;
		}
	      // select foreground color
	      else if (strcmp(temp, "FCL") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcapcolor(buffer, &forecol) == false)
		    return false;
		  
		  xcol.red   = reducedcolor(forecol.red);
		  xcol.green = reducedcolor(forecol.green);
		  xcol.blue  = reducedcolor(forecol.blue);
		  XAllocColor(disp, cm, &xcol);
		  forecolor = xcol.pixel;
		  setforcolor(forecolor);
		  XFlush(disp);
		}
	      // select background color
	      else if (strcmp(temp, "BCL") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcapcolor(buffer, &backcol) == false)
		    return false;
		  
		  xcol.red   = reducedcolor(backcol.red);
		  xcol.green = reducedcolor(backcol.green);
		  xcol.blue  = reducedcolor(backcol.blue);
		  XAllocColor(disp, cm, &xcol);
		  backcolor = xcol.pixel;
		  XFlush(disp);
		}
	      // select font
	      else if (strcmp(temp, "LFT") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcapfont(buffer, &currentfo) == false)
		    return false;
		  ftemp1 = currentfo.size * cos(currentfo.angle / 180.0 * M_PI);
		  ftemp2 = currentfo.size * sin(currentfo.angle / 180.0 * M_PI);
		  ftemp3 = -currentfo.size * sin(currentfo.angle / 180.0 * M_PI);
		  ftemp4 = currentfo.size * cos(currentfo.angle / 180.0 * M_PI);
		  
		  sprintf(s1, "%f", ftemp1); convminus(s1);
		  sprintf(s2, "%f", ftemp2); convminus(s2);
		  sprintf(s3, "%f", ftemp3); convminus(s3);
		  sprintf(s4, "%f", ftemp4); convminus(s4);
		  
		  if (currentfo.face < 1000)
		    sprintf(stemp, topazfontmap[currentfo.face], s1, s2, s3, s4);
		  else
		      sprintf(stemp, topazkfontmap[currentfo.face - 1000], s1, s2, s3, s4);

		  if ((fst = XLoadQueryFont(disp, stemp)) != NULL)
		    {
		      XSetFont(disp, gc, fst->fid);
		      XFlush(disp);
		    }
		  else
		    fprintf(stderr, "cannot find font: %s\n", stemp);
		}
	      // draw text
	      else if (strcmp(temp, "DTX") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  
		  if (fst != NULL)
		    {
		      if (strlen(buffer) == 2)
			{
			  buffer[1] = 0;
			  XDrawString(disp, win, gc, (short)dp.x, (short)dp.y, buffer, strlen(buffer));
			}
		      else
			{
			  tojis((unsigned int *)buffer);
			  kanji[0].byte1 = buffer[1];
			  kanji[0].byte2 = buffer[0];
			  XDrawString16(disp, win, gc, (short)dp.x, (short)dp.y, kanji, 1);
			}

		    }
		      XFlush(disp);
		}
	      //set paint mode 
	      else if (strcmp(temp, "PMD") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  
		  //thickness
		  paintmode = atoi(buffer);
		  
		  if (paintmode == 0)
		    XSetFunction(disp, gc, GXcopy);
		  else if (paintmode == 1)
		    XSetFunction(disp, gc, GXxor);
		}
	      //set line style
	      else if (strcmp(temp, "LSY") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //thickness
		  currentlinewidth = (int)((double)atoi(buffer) * unit * mag);
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //capstyle
		  switch(atoi(buffer))
		    {
		    case LINECAPBUTT:
		      currentcapstyle = CapButt;
		      break; 
		    case LINECAPROUND:
		      currentcapstyle = CapRound;
		      break; 
		    case LINECAPPROJECTING:
		      currentcapstyle = CapProjecting;
		      break; 
		    }
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //joinstyle
		  switch(atoi(buffer))
		    {
		    case LINEJOINMITER:
		      currentjoinstyle = JoinMiter;
		      break;
		    case LINEJOINROUND:
		      currentjoinstyle = JoinRound;
		      break;
		    case LINEJOINBEVEL:
		      currentjoinstyle = JoinBevel;
		      break;
		    }
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //linestyle
		  linestyle  = atoi(buffer);
		  
		  XSetLineAttributes(disp, gc,
				     currentlinewidth,
				     LineSolid,
				     currentcapstyle,
				     currentjoinstyle);
		}
	      //set fill style
	      else if (strcmp(temp, "FSY") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  // fillrule
		  switch(atoi(buffer))
		    {
		    case FILLRULEEVENODD:
		      currentfillrule  = EvenOddRule;
		      break;
		    case FILLRULEWINDING:
		      currentfillrule  = WindingRule;
		      break;
		    }
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //fillstyle
		  fillstyle  = atoi(buffer);
		  
		  XSetFillRule(disp, gc, currentfillrule);
		}
	      //setcliprect
	      else if (strcmp(temp, "CRT") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  xrect.x = (short)dp.x;
		  xrect.y = (short)dp.y;
		  
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  if (getcappoint(buffer, &dp) == false)
		    return false;
		  xrect.width  = (unsigned short)dp.x - xrect.x;
		  xrect.height = (unsigned short)dp.y - xrect.y;
		  
		  XSetClipRectangles(disp, gc, 0, 0, &xrect, 1, Unsorted);
		}
	      //initcliprect
	      else if (strcmp(temp, "RCP") == 0)
		{
		  xrect.x = (short)0;
		  xrect.y = (short)0;
		  xrect.width  = (unsigned short)MAXSHORT;
		  xrect.height = (unsigned short)MAXSHORT;
		  
		  XSetClipRectangles(disp, gc, 0, 0, &xrect, 1, Unsorted);
		}
	      // paper
	      else if (strcmp(temp, "PAR") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;

		  sscanf(buffer, "%d,%d\n", &paperx, &papery);
		}
	      //clear window
	      else if (strcmp(temp, "CLR") == 0)
		{
		  XClearWindow(disp, win);
		  XSetLineAttributes(disp, gc,
				     0,
				     LineSolid,
				     CapButt,
				     JoinMiter);
		  xcol.red   = 0;
		  xcol.green = 0;
		  xcol.blue  = 0;
		  XAllocColor(disp, cm, &xcol);
		  forecolor = xcol.pixel;
		  XSetForeground(disp, gc, forecolor);
		  XSetFunction(disp, gc, GXcopy);

		  XDrawLine(disp, win, gc, (int)posix(0), (int)posiy(0), (int)posix(paperx), (int)posiy(0));
		  XDrawLine(disp, win, gc, (int)posix(paperx), (int)posiy(0), (int)posix(paperx), (int)posiy(papery));
		  XDrawLine(disp, win, gc, (int)posix(paperx), (int)posiy(papery), (int)posix(0), (int)posiy(papery));
		  XDrawLine(disp, win, gc, (int)posix(0), (int)posiy(papery), (int)posix(0), (int)posiy(0));
		  XFlush(disp);
		}
	      //set temporary line style (for xor draw)
	      else if (strcmp(temp, "STD") == 0)
		{
		  XSetLineAttributes(disp, gc,
				     currentlinewidth,
				     LineOnOffDash,
				     currentcapstyle,
				     currentjoinstyle);
		  XSetDashes(disp, gc, 0, dash1, 2);
		}
	      //restore line style (fromr xor draw mode)
	      else if (strcmp(temp, "RLS") == 0)
		{
		  XSetLineAttributes(disp, gc,
				     currentlinewidth,
				     LineSolid,
				     currentcapstyle,
				     currentjoinstyle);
		}
	      // mag
	      else if (strcmp(temp, "MAG") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  //thickness
		  mag = atof(buffer);
		}
	      // shift
	      else if (strcmp(temp, "SFT") == 0)
		{
		  if (topazfgets(buffer, sizeof(buffer), in) == NULL)
		    return false;
		  sscanf(buffer, "%d,%d\n", &shiftx, &shifty);
		}
	    }
	}
    }
  return true;
}


int readdata(void)
{
  char buffer[buffersize];

  while (topazfgets(buffer, sizeof(buffer), in))
    {
      if (strcmp(buffer, "PNT\n") == 0)
          XFlush(disp);
      // reading
      else
	  painting(buffer);
    }
  return true;
 }

int loaddata()
{
  char buffer[buffersize];

  while (topazfgets(buffer, sizeof(buffer), in) != NULL)
    {
      if (strcmp(buffer, "PNT\n") == 0)
	XFlush(disp);
      // reading
      else
	painting(buffer);
    }
  return true;
}

unsigned int reducedcolor(unsigned int c)
{
  double cwidth = 65535.0 / (double)cdepth;
  return (unsigned int)(floor(((double)c + .1)/ cwidth) * cwidth);
}

bool getusecolor()
{
  XColor xcol;
  double cwidth = 65535.1 / (double)cdepth;
  for (int r = 0; r <= cdepth; r++)
    {
      for (int g = 0; g <= cdepth; g++)
	{
	  for (int b = 0; b <= cdepth; b++)
	    {
	      xcol.red   = (unsigned int)(cwidth * (double)r);
	      xcol.green = (unsigned int)(cwidth * (double)g);
	      xcol.blue  = (unsigned int)(cwidth * (double)b);
	      XAllocColor(disp, cm, &xcol);
	    }
	}
    }
  return true;
}

void preloop()
{
  char *display = getenv("DISPLAY");
  static XSizeHints hints;
  
  signal(SIGINT, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  disp = XOpenDisplay(display); 
  if (!disp)
    {
      fprintf(stderr, "can't open display\n");
      exit(1);
    }
  cm = DefaultColormap(disp, 0);

  wcol.red   = 65535;
  wcol.green = 65535;
  wcol.blue  = 65535;
  XAllocColor(disp, cm, &wcol);

  bcol.red   = 65535;
  bcol.green = 65535;
  bcol.blue  = 65535;
  XAllocColor(disp, cm, &bcol);

  root = XDefaultRootWindow(disp);
  win = XCreateSimpleWindow(disp, root, x, y, w, h, border, bcol.pixel, wcol.pixel);
  at.backing_store = Always;
  XChangeWindowAttributes(disp, win, CWBackingStore, &at);
  
  hints.flags = USPosition |  USSize;
  XSetWMNormalHints(disp, win, &hints);

  XSelectInput(disp, win, StructureNotifyMask);

  XMapWindow(disp, win);
  XStoreName(disp, win, (_Xconst char*)title.c_str());
  XFlush(disp);
  
  if (gc)
    XFreeGC(disp, gc);
  gc = XCreateGC(disp, win, 0, 0);

  XSetForeground(disp, gc, 0);
  XSetBackground(disp, gc, 0);

  //  getusecolor();
} 

bool fileexist(char *file)
{
  FILE *h;
  if ((h = fopen(file, "r")) != NULL)
    {
      fclose(h);
      return true;
    }
  return false;
}

bool searchfontmap(std::string *file, std::string *out)
{
  std::string str;
  if (getenv("TOPAZINIDIR") == NULL)
    return false;
  str = std::string(getenv("TOPAZINIDIR")) + std::string("/") + *file;
  if (fileexist((char *)str.c_str()))
    {
      *out = str;
      return true;
    }
  else
    {
      if (getenv("TOPAZDIR") == NULL)
	return false;
      str = std::string(getenv("TOPAZDIR")) + std::string("/_topaz/") + *file;
      if (fileexist((char *)str.c_str()))
	{
	  *out = str;
	  return true;
	}
      else
	return false;
    }
  return true;
}

bool loadfontmap()
{
  FILE* f;
  char fontmapfile[1024];
  char kfontmapfile[1024];
  char stemp[256], stemp1[3];
  char *p;
  std::string str, out;

  str = std::string("topazfontmap");
  if (!searchfontmap(&str, &out))
    return false;
  strcpy(fontmapfile, out.c_str());
  
  str = std::string("topazkfontmap");
  if (!searchfontmap(&str, &out))
    return false;
  strcpy(kfontmapfile, out.c_str());

  // ascii
  if ((f = fopen(fontmapfile, "r")) == NULL)
    return false;
  
  for (int i = 0; i < 13; i++)
    {
      if (topazfgets(stemp, sizeof(stemp), f) == NULL)
	{
	  fclose(f);
	  return false;
	}
      stemp[sizeof(stemp) - 1] = 0;
      
      p = stemp;
      strcpy(topazfontmap[i], "");
      while (*p != 0)
	{
	  if (*p != '\n' && *p != '\r')
	    {
	      stemp1[0] = *p; stemp1[1] = 0;
	      strcat(topazfontmap[i], stemp1);
	    }
	  p++;
	}
    }
  fclose(f);

  // kanji
  if ((f = fopen(kfontmapfile, "r")) == NULL)
    return false;
  
  for (int i = 0; i < 8; i++)
    {
      if (topazfgets(stemp, sizeof(stemp), f) == NULL)
	{
	  fclose(f);
	  return false;
	}
      stemp[sizeof(stemp) - 1] = 0;
      
      p = stemp;
      strcpy(topazkfontmap[i], "");
      while (*p != 0)
	{
	  if (*p != '\n' && *p != '\r')
	    {
	      stemp1[0] = *p; stemp1[1] = 0;
	      strcat(topazkfontmap[i], stemp1);
	    }
	  p++;
	}
    }
  fclose(f);
  return true;
}

int main(int argc, char *argv[])
{
  //  extern char *optarg;
  extern int optind;
  in = stdin;
  bool inflag = false;
  int c;
  
  title = std::string("");

  // options
  while ((c = getopt(argc, argv, "x:y:w:h:t:c:v")) != -1)
    {
      switch(c)
	{
	  // x
	case 'x':
	  x = atoi(optarg);
	  break;
	  // y
	case 'y':
	  y = atoi(optarg);
	  break;
	  // w
	case 'w':
	  w = atoi(optarg);
	  break;
	  // h
	case 'h':
	  h = atoi(optarg);
	  break;
	  // title
	case 't':
	  title = std::string(optarg);
	  // color depth
	case 'c':
	  cdepth = atoi(optarg);
	  break;
	case 'v':
	  fprintf(stderr, "tpvview Ver. 1.00\n");
	  fprintf(stderr, "Written by Hisao Kawaura\n");
	  fprintf(stderr, "-x (x position)\n");
	  fprintf(stderr, "-y (y position)\n");
	  fprintf(stderr, "-w (width)\n");
	  fprintf(stderr, "-h (height)\n");
	  fprintf(stderr, "-t (caption)\n");
	  fprintf(stderr, "-c (color number)\n");
	  exit(0);
	  break;
	}
    }

  // input file exists
  if (optind < argc)
    {
      if ((in = fopen(argv[optind], "r")) == NULL)
	{
	  fprintf(stderr, "Error: cannot open input file '%s'\n", argv[optind]);
	  exit(1);
	}
      else
	inflag = true;
    }

  if (!loadfontmap())
    {
      fprintf(stderr, "Error: cannot open topazfontmap\n");
      exit(1);
    }

  preloop();
  if (inflag)
    loaddata();
  else
    readdata();

  if (inflag)
    fclose(in);

  if (gc)
    XFreeGC(disp, gc);
  exit(0);
}







