/*----------------------------------------------------------------------------
                      valvuff.cc (vector for variables)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <string>
#include "valbuff.h"
#include "buffarray.h"

BuffElement::BuffElement()
{label = std::string(""); value = new buffarray(); isreadonly = false;}

BuffElement& BuffElement::operator = (const BuffElement&  ele)
{
  label = ele.label;
  *value = *ele.value;
  isreadonly = ele.isreadonly;
  return *this;
}

VariantBuff::VariantBuff()
{
  int i;
  arraylength = 0;
  increment = 10;

  bufflength = (arraylength / increment + 1) * increment;
  element =   new BuffElement * [bufflength];
  memset(element, 0, bufflength * sizeof(int));
  for (i = 0; i < arraylength; i++)
    element[i] = new BuffElement();

}

VariantBuff::~VariantBuff()
{
  int i;
  for (i = 0; i < arraylength; i++)
    delete element[i];

  delete [] element;

}

bool VariantBuff::SetArray (std::string *label, int length)
{
  int itemp;
  std::string str;
  
  if (search(label, &itemp) == true)  //exist
    {
      if (element[itemp]->isreadonly)
	return true;

	element[itemp]->value->setarray(length);
      return true;
    }
  else                                //new 
    {
      str = std::string("");
      add(label, &str);
      element[getarraylength() - 1]->value->setarray(length);
      return true;
    }
}

bool VariantBuff::SystemSetArray (std::string *label, int length)
{
  int itemp;
  std::string str;
  
  if (search(label, &itemp) == true)  //exist
    {
	element[itemp]->value->setarray(length);
	element[itemp]->isreadonly = true;
     return true;
    }
  else                                //new 
    {
      str = std::string("");
      add(label, &str);
      element[getarraylength() - 1]->value->setarray(length);
      element[arraylength - 1]->isreadonly = true;
      return true;
    }
}

bool VariantBuff::search(std::string *label, int *SeachedIndex)
{
  int i;

  for (i = 0; i < arraylength; i++)
    {
      if (*label == element[i]->label)
	{
	  *SeachedIndex = i;
	  return true;
	}
    }
  return false;

}

bool VariantBuff::Set(std::string *label, int pos, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)    //overwrite
    {
      if (element[itemp]->isreadonly)
	return true;
      element[itemp]->value->setdata(pos, value->c_str());
    }
  else                                  //append to last
    {
      arraylength++;
      element[arraylength - 1] = new BuffElement();
      element[arraylength - 1]->label = *label;
      element[arraylength - 1]->value->setdata(pos, value->c_str());
    }
  return true;
}

bool VariantBuff::SystemSet(std::string *label, int pos, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)    //overwrite
    {
      element[itemp]->value->setdata(pos, value->c_str());
      element[itemp]->isreadonly = true;
    }
  else                                  //append to last
    {
      arraylength++;
      element[arraylength - 1] = new BuffElement();
      element[arraylength - 1]->label = *label;
      element[arraylength - 1]->value->setdata(pos, value->c_str());
      element[arraylength - 1]->isreadonly = true;
    }
  return true;
}

bool VariantBuff::Add(std::string *label, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)
    {
      element[itemp]->value->add(value->c_str());
      return true;
    }
  else
    return false;
}


bool VariantBuff::Get(std::string *label, int pos, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)
    {
      if (pos < element[itemp]->value->getarraylength() && 0 <= pos)
      {
	*value = element[itemp]->value->getdata(pos);
	return true;
      }
      else
	return false;
    }
  else
   {
     return false;
   }
}

int VariantBuff::getarraylength()
{
  return arraylength;
}

int VariantBuff::ArrayLength(std::string *label)
{
  int itemp;  
  if (search(label, &itemp) == true)
      return element[itemp]->value->getarraylength();
  else
    return -1;
}

bool VariantBuff::setbuff(int length)
{
  int i;

  for (i = 0; i < arraylength; i++)
    delete element[i];
  delete [] element;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  element =  new BuffElement * [bufflength];
  memset(element, 0, bufflength * sizeof(int));

  return true;
}

bool VariantBuff::add(std::string *label, std::string *val)
{
  if (arraylength + 1 >= bufflength)
      changebufflength(arraylength + 1);

  arraylength++;
  element[arraylength - 1] = new BuffElement();
  element[arraylength - 1]->label = *label;
  element[arraylength - 1]->value->add(val->c_str());
  return true;
}

bool VariantBuff::add(BuffElement *buffele)
{
  if (arraylength + 1 >= bufflength)
      changebufflength(arraylength + 1);

  arraylength++;
  element[arraylength - 1] = new BuffElement();
  *element[arraylength - 1] = *buffele;
  return true;
}


bool VariantBuff::addat(std::string *label, int pos, std::string *val)
{
  int i;
  if (pos < arraylength)
    {
      if (arraylength >=  bufflength)
	  changebufflength(++arraylength);
      for (i = arraylength - 1; i > pos; i--)
	  *element[i] = *element[i - 1];

      element[pos]->label = *label;
      element[pos]->value->add(val->c_str());
    }
  else
    add(label, val);
  return true;
}

bool VariantBuff::remove(int pos)
{
  int i;
  for (i = pos; i < arraylength - 1; i++)
    *element[i] = *element[i + 1];

  delete element[arraylength - 1];
  arraylength--;
  return true;
}

bool VariantBuff::changebufflength(int newlength)
{

  int i;
  BuffElement **tempstrarray;

  if (newlength <= bufflength)
    return true;

  bufflength = (newlength / increment + 1) * increment;
  tempstrarray =  new BuffElement * [bufflength];
  memset(tempstrarray, 0, bufflength * sizeof(int));

  for (i = 0; i < arraylength; i++)
    tempstrarray[i] = element[i];

  delete [] element;
  element = tempstrarray;

  return true;
}

bool VariantBuff::SetScalar(std::string *label, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)    //overwrite
    {
      if (element[itemp]->isreadonly)
	return true;
      else
	{
	  element[itemp]->value->setdata(0, value->c_str());
	  return true;
	}
    }
  else                                  //append to last
    {
      add(label, value);
      return true;
    }
  return true;
}

bool VariantBuff::SetSystemScalar(std::string *label, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)    //overwrite
    {
      element[itemp]->isreadonly = true;
      element[itemp]->value->setdata(0, value->c_str());
      return true;
    }
  else                                  //append to last
    {
      add(label, value);
      element[arraylength - 1]->isreadonly = true;
      return true;
    }
  return true;
}

bool VariantBuff::GetScalar(std::string *label, std::string *value)
{
  int itemp;

  if (search(label, &itemp) == true)
    {
      *value = element[itemp]->value->getdata(0);
      return true;
    }
  else
   {
     return false;
   }
}


VariantBuff& VariantBuff::operator = (const VariantBuff&  buff)
{
  int i;  
  setbuff(0);
  for (i = 0; i < buff.arraylength; i++)
    add(buff.element[i]);
  return *this;
}

BuffElement& VariantBuff::operator [] (int pos)
{
    return *element[pos];
}







