/*----------------------------------------------------------------------------
                       vect.h (templates of vector class)
                       This file is a part of topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#if !defined(__vect_h)
#define __vect_h
#include <stdio.h>
#include <string.h>

template <class T>
class vect
{
 public:
  
  vect();
  vect(int inc);
  ~vect();
  bool setdata(int pos, T* val);
  T& getdata(int pos);
  int getarraylength();     
  bool setarray(int length);
  bool setbuff(int length);
  bool add(T* val);
  bool addat(int pos, T* val);
  bool remove(int pos);
  bool up(int pos);
  bool down(int pos);
  bool top(int pos);
  bool bottom(int pos);
  
  vect& operator = (const vect& buff);
  T& operator [] (int pos);
  

  int arraylength; 
  int bufflength;  
  T **array;
  int increment;
  bool changebufflength(int length);

};


template <class T>
vect<T>::vect()
{
  int i;
  arraylength = 0;
  increment = 100;

  bufflength = (arraylength / increment + 1) * increment;
  array =   new T * [bufflength];
  memset(array, 0, bufflength * sizeof(int));
  for (i = 0; i < arraylength; i++)
    array[i] = new T;

}

template <class T>
vect<T>::vect(int inc)
{
  int i;
  arraylength = 0;
  increment = inc;

  bufflength = (arraylength / increment + 1) * increment;
  array =   new T * [bufflength];
  memset(array, 0, bufflength * sizeof(int));
  for (i = 0; i < arraylength; i++)
    array[i] = new T;
}


template <class T>
vect<T>::~vect()
{
  int i;
  for (i = 0; i < arraylength; i++)
    delete array[i];

  delete [] array;
}


template <class T>
bool vect<T>::setdata(int pos, T* val)
{
  int i;

  if (pos < arraylength)
    {
      delete array[pos];
      array[pos] =  val;

      if (array[pos] == 0)
	return false;
      else
	return true;
    }
  else
    return false;
}

template <class T>
T& vect<T>::getdata(int pos)
{
  
  if (pos < arraylength)
    return *array[pos];
  else
    return *array[0];
}

template <class T>
int vect<T>::getarraylength()
{
  return arraylength;
}

template <class T>
bool vect<T>::setbuff(int length)
{
  int i;

  for (i = 0; i < arraylength; i++)
    delete array[i];
  delete [] array;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  array =  new T * [bufflength];
  if (array ==  0)
    return false;

  memset(array, 0, bufflength * sizeof(int));
  return true;
}

template <class T>
bool vect<T>::setarray(int length)
{
  int i;

  if (setbuff(length) == false)
    return false;

  arraylength = length;

  for (i = 0; i < arraylength; i++)
    array[i] = new T;
  
  return true;
}


template <class T>
bool vect<T>::add(T* val)
{

  if (arraylength < bufflength)
    {
      arraylength++;
      array[arraylength - 1] = val;
    }
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
      array[arraylength - 1] = val;
    }
  
  if (array[arraylength - 1] == 0)
    return false;
  else
    return true;
}

template <class T>
bool vect<T>::addat(int pos, T* val)
{
  int i;
  if (pos < arraylength)
    {
      if (arraylength >=  bufflength)
	{
	  if (changebufflength(++arraylength) == false)
	  return false;
	}
      else
	arraylength++;

      for (i = arraylength - 1; i > pos; i--)
	  array[i] = array[i - 1];
      
      if (setdata(pos, val) == false)
	return false;
      
      return true;
    }
  else
    return add(val);
}

template <class T>
bool vect<T>::remove(int pos)
{
  int i;
  
  if (pos < arraylength)
    {
      delete array[pos];
      for (i = pos; i < arraylength - 1; i++)
	array[i] = array[i + 1];
      
      array[arraylength - 1] = 0;
      arraylength--;
      return true;
    }
  else
    return false;
}

template <class T>
bool vect<T>::up(int pos)
{
  T* temp;
  
  if (0 <= pos && pos < arraylength - 1)
    {
      temp = array[pos];
      array[pos] = array[pos + 1];
      array[pos + 1] = temp;
      return true;
    }
  else if (pos == arraylength - 1)
    return true;
  else
    return false;
}

template <class T>
bool vect<T>::down(int pos)
{
  T* temp;
  
  if (0 < pos && pos <= arraylength - 1)
    {
      temp = array[pos];
      array[pos] = array[pos - 1];
      array[pos - 1] = temp;
      return true;
    }
  else if (pos == 0)
    return true;
  else
    return false;
}

template <class T>
bool vect<T>::top(int pos)
{
  int i;
  T* temp;
  
  if (0 <= pos && pos < arraylength - 1)
    {
      temp = array[pos];
      for (i = pos; i < arraylength - 1; i++)
	array[i] = array[i + 1];
      
      array[arraylength - 1] = temp;
      return true;
    }
  else if (pos == arraylength - 1)
    return true;
  else
    return false;
}

template <class T>
bool vect<T>::bottom(int pos)
{
  int i;
  T* temp;
  
  if (0 < pos && pos <= arraylength - 1)
    {
      temp = array[pos];
      for (i = pos; i > 0; i--)
	array[i] = array[i - 1];
      
      array[0] = temp;
      return true;
    }
  else if (pos == 0)
    return true;
  else
    return false;
}


template <class T>
bool vect<T>::changebufflength(int newlength)
{
  
  int i;
  T **temparray;

  if (newlength <= bufflength)
    return true;

  bufflength = (newlength / increment + 1) * increment;
  temparray =   new T * [bufflength];
  if (temparray == 0)
    return false;

  memset(temparray, 0, bufflength * sizeof(int));
  
  for (i = 0; i < arraylength; i++)
    temparray[i] = array[i];

  delete[] array;
  array = temparray;
  return true;
}

template <class T>
vect<T>& vect<T>::operator = (const vect<T>&  buff)
{
  int i;  
  T *tempt;

  setbuff(buff.bufflength);
  for (i = 0; i < buff.arraylength; i++)
    {
      tempt = new T;
      *tempt = *buff.array[i];
      add(tempt);
    }
  return *this;
}

template <class T>
T& vect<T>::operator [] (int pos)
{
    return *array[pos];
}

/****************** rvect ********************************/

template <class T>
class rvect
{
 public:
  
  rvect();
  rvect(int inc);
  ~rvect();
  bool setdata(int pos, T val);
  T& getdata(int pos);
  int getarraylength();     
  bool setarray(int length);
  bool setbuff(int length);
  bool add(T val);
  bool addat(int pos, T val);
  bool remove(int pos);
  
  rvect& operator = (const rvect& buff);
  T& operator [] (int pos);
  

  int arraylength; 
  int bufflength;  
  T *array;
  int increment;
  bool changebufflength(int length);

};


template <class T>
rvect<T>::rvect()
{
  arraylength = 0;
  increment = 100;

  bufflength = (arraylength / increment + 1) * increment;

  array =   new T [sizeof (T) * bufflength];
  memset(array, 0, sizeof (T) * bufflength);
}

template <class T>
rvect<T>::rvect(int inc)
{
  int i;
  arraylength = 0;
  increment = inc;

  bufflength = (arraylength / increment + 1) * increment;

  array =   new T [sizeof (T) * bufflength];
  memset(array, 0, sizeof (T) * bufflength);

}


template <class T>
rvect<T>::~rvect()
{
    delete array;
}


template <class T>
bool rvect<T>::setdata(int pos, T val)
{
  if (pos < arraylength)
    {
      array[pos] =  val;
      return true;
    }
  else
    return false;
}

template <class T>
T& rvect<T>::getdata(int pos)
{
  
  if (pos < arraylength)
    return array[pos];
  else
    return array[0];
}

template <class T>
int rvect<T>::getarraylength()
{
  return arraylength;
}

template <class T>
bool rvect<T>::setbuff(int length)
{
  delete array;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  array =   new T [sizeof (T) * bufflength];
  memset(array, 0, sizeof (T) * bufflength);
  return true;
}

template <class T>
bool rvect<T>::setarray(int length)
{
  if (setbuff(length) == false)
    return false;
  arraylength = length;
  return true;
}


template <class T>
bool rvect<T>::add(T val)
{

  if (arraylength < bufflength)
    {
      arraylength++;
      array[arraylength - 1] = val;
    }
  else
    {
      if (changebufflength(arraylength + 1))
	{
	  arraylength++;
	  array[arraylength - 1] = val;
	}
      else
	return false;
    }
  
    return true;
}

template <class T>
bool rvect<T>::addat(int pos, T val)
{
  int i;
  if (pos < arraylength)
    {
      if (arraylength >=  bufflength)
	{
	  if (changebufflength(++arraylength) == false)
	  return false;
	}
      else
	arraylength++;

      for (i = arraylength - 1; i > pos; i--)
	  array[i] = array[i - 1];
      
      if (setdata(pos, val) == false)
	return false;
      
      return true;
    }
  else
    return add(val);
}

template <class T>
bool rvect<T>::remove(int pos)
{
  int i;
  
  if (pos < arraylength)
    {
      for (i = pos; i < arraylength - 1; i++)
	array[i] = array[i + 1];
      
      array[arraylength - 1] = 0;
      arraylength--;
      return true;
    }
  else
    return false;
  
}

template <class T>
bool rvect<T>::changebufflength(int newlength)
{
  
  int i;
  T *temparray;
  int tempbufflength, temparraylength;

  if (newlength <= bufflength)
    return true;

  tempbufflength = (newlength / increment + 1) * increment;
  temparray =   new T [sizeof (T) * bufflength];
  if (temparray == 0)
    return false;
  memset(temparray, 0, sizeof (T) * bufflength);

  temparraylength = arraylength;
  
  for (i = 0; i < arraylength; i++)
    temparray[i] = array[i];
  
  delete array;
  array = temparray;
  bufflength = tempbufflength;

  return true;
}

template <class T>
rvect<T>& rvect<T>::operator = (const rvect<T>&  buff)
{
  int i;  

  setbuff(buff.bufflength);
  for (i = 0; i < buff.arraylength; i++)
      add(buff.array[i]);

}

template <class T>
T& rvect<T>::operator [] (int pos)
{
    return array[pos];
}


#endif










