/*----------------------------------------------------------------------------
                      parray.c (definitions of POINT array)
                       This file is a part of Topaz systems
                  Copyright: Hisao Kawaura, All rights reserved
                                   1997 - 98


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

----------------------------------------------------------------------------*/

#include <memory.h>
#include "Parray.h"

Parray::Parray()
{
  arraylength = 0;
  increment = 10000;

  bufflength = (arraylength / increment + 1) * increment;

  array =   new POINT[bufflength];
  memset(array, bufflength, 0);
}


Parray::~Parray()
{
  delete [] array;
}

int Parray::getarraylength()
{
  return arraylength;
}

bool Parray::setbuff(int length)
{
  delete [] array;

  arraylength = 0;
  bufflength = (length / increment + 1) * increment;
  array =   new POINT[bufflength];
  if (array == 0)
    return false;
  
  memset(array, bufflength, 0);
  
  return true;
}


bool Parray::add(POINT* val)
{

  if (arraylength < bufflength)
    {
      arraylength++;
      array[arraylength - 1] = *val;
    }
  else
    {
      changebufflength(arraylength + 1);
      arraylength++;
      array[arraylength - 1] = *val;
    }
  
    return true;
}

bool Parray::changebufflength(int newlength)
{
  
  POINT *temparray;
  int temparraylength;

  if (newlength <= arraylength)
    return true;

  temparray =   new POINT[bufflength];
  if (temparray == 0)
    return false;

  temparraylength = arraylength;
  
  memcpy(temparray, array, arraylength * sizeof(POINT));

  delete[] array;

  array = temparray;

  return true;
}

POINT& Parray::operator [] (int pos)
{
    return array[pos];
}
